************************************************************************
      SUBROUTINE MERLIN ( NOV, NOSQ, VERSIM, MAXW, IQUIT )
************************************************************************
*                                                                      *
*       The Merlin optimization package, version 3.0.4 / May 2002      *
*                                                                      *
*  Implemented by:                                                     *
*         D. G. Papageorgiou                                           *
*         Dept. of Chemistry, University of Ioannina, GREECE.          *
*         Email: jimmy@nrt.cs.uoi.gr                                   *
*                                                                      *
*         I. E. Lagaris                                                *
*         Dept. of Computer Science, University of Ioannina, GREECE.   *
*         Email: lagaris@cs.uoi.gr                                     *
*                                                                      *
*         I. N. Demetropoulos                                          *
*         Dept. of Chemistry, University of Ioannina, GREECE.          *
*         Email: idimitr@cc.uoi.gr                                     *
*                                                                      *
*  All authors can be reached at: merlin@nrt.cs.uoi.gr                 *
*  More information is available at: http://nrt.cs.uoi.gr/merlin       *
*                                                                      *
************************************************************************
C
C  Input arguments:
C    NOV        Number of variables of the objective function.
C    NOSQ       Number of terms, in case the objective function is a
C               sum of squares. Must be set to 0 otherwise.
C    VERSIM     Array, used for temporary storage for routines that
C               require storage of the order NOV**2 or NOV*NOSQ, such as
C               BFGS, SIMPLEX, TOLMIN, LEVE etc.)
C    MAXW       Length of array VERSIM.
C
C  Output arguments:
C    IQUIT      If an error occurs, this variable is set to a negative
C               value as follows:
C                  -1  --> Not enough (one-dimensional) storage for
C                          this problem. You must redimension and
C                          recompile Merlin.
C                  -2  --> Initial unit numbers for the configuration
C                          file cannot be allocated.
C                  -3  --> The output file/unit cannot be opened.
C                          Moreover, some errors occurred while
C                          parsing the configuration file.
C                  -4  --> Some error while parsing the configuration
C                          file. The error is related to directives
C                          that set the output unit/file.
C                  -5  --> The output file/unit cannot be opened.
C                          The configuration file has been successfully
C                          parsed.
C                  -6  --> Some error while parsing the configuration
C                          file. The error is not related to the
C                          output file/unit which has been successfully
C                          opened.
C                  -7  --> Some error while parsing the panel
C                          description file.
C                  -8  --> The input unit cannot be opened.
C                 -10  --> EOF in input.
C                 -20  --> A run time error, other than EOF has
C                          occurred. (Incorrect command, MCL runtime
C                          error, etc.) Will be returned only if the
C                          BATCH mode was set.
C               If a RETURN command is issued to Merlin, IQUIT is set
C               to 0.
C               If a QUIT command is issued, IQUIT is set to the return
C               code supplied by the user. Note however that QUIT will
C               not accept non positive values for IQUIT.
C
C  ---------------------------------------------------------------------
      IMPLICIT $(TYPE) (A-H,O-Z)
C
C  Name of the Merlin configuration file.
      CHARACTER*(*) CFILE
      PARAMETER ( CFILE = '$(CONFIG)' )
C  ---------------------------------------------------------------------
C  Arguments:
      DIMENSION VERSIM(*)
C  ---------------------------------------------------------------------
C  Common blocks:
      LOGICAL STRO, WEAK
      COMMON /PRILOG/ STRO, WEAK
      COMMON /UNITS/ IUINP, IUOUT
      COMMON /DIMEN/ NOD, NOF
      PARAMETER ( MXFILE = 80 )
      CHARACTER*(MXFILE) PROLO, EPILO, IN, OUT, HFILE, MFILE, PDFILE
      COMMON /CP2/ PROLO, EPILO, IN, OUT, HFILE, MFILE, PDFILE
      COMMON /GETOUT/ IRET, IRET2, IQU
      PARAMETER ( NMCOM = 98, MXPLUG = $(MXPLUG), NCOM = NMCOM+MXPLUG )
      CHARACTER*10 COMMAN(NCOM)
      COMMON /COM1/ COMMAN
      INTEGER IND(NCOM)
      COMMON /COM2/ IND
      CHARACTER*10 DMCOMS(NCOM)
      COMMON /DCCOM/ DMCOMS
      PARAMETER ( LINLEN = 80 )
      INTEGER IS(LINLEN/2), IE(LINLEN/2)
      COMMON /INDAT1/ IS, IE, NIT, LEC
      COMMON /CNTRL/ IDISK, MACECH, LIMCLK, IECOMA, IPROEC, MCLUBU
      PARAMETER ( MAXINL = 20 )
      INTEGER BACKIN(MAXINL), WHATIN(MAXINL), REDIN(MAXINL)
      COMMON /INNER1/ LEVIN, BACKIN, WHATIN, REDIN
      COMMON /MODE3/ MODEV, IFFORM, INMCL
C  ---------------------------------------------------------------------
C  Local variables:
C  COM is set to this length, to allow full pathnames in a a macro call.
      CHARACTER*(LINLEN) COM, ACTCOM
      CHARACTER*(MXFILE) CONFIL
      LOGICAL ERRORO, INPROL, INEPIL, FERROR, ISREDI
C  ---------------------------------------------------------------------
C
      NOD = NOV
      NOF = NOSQ
C
C  Do we have enough local storage ?
      CALL STOREN(NOD,NOF,IERR)
      IF (IERR.NE.0) THEN
         IQUIT = -1
         RETURN
      END IF
C
C  Copy commands to the command array.
      DO 200,I=1,NMCOM
200      COMMAN(I) = DMCOMS(I)
C
C  Initialize variables needed specifically for parsing the
C  configuration file.
      CALL CPINIT
C
C  Initialize variables in common.
      CALL INIT(VERSIM,MAXW)
C
C  Add empty plug-in entries to the command array.
      CALL ADDPLG
C
C  Read and process the configuration file.
      ICONU = NUNIT()
      IFILEU = NUNIT()
      IERRU = NUNIT()
      IF (ICONU.EQ.-1 .OR. IFILEU.EQ.-1 .OR. IERRU.EQ.-1) THEN
         IQUIT = -2
         RETURN
      END IF
      CONFIL = CFILE
      CALL CONPRO(CONFIL,ICONU,IFILEU,IERRU,IERROR)
      IF (IERROR.EQ.2) THEN
         CALL CONECP(CONFIL,IERRU)
         IQUIT = -4
         RETURN
      END IF
C
C  Open the output file.
      CALL OPOUT(ERRORO)
      IF (ERRORO) THEN
         IF (IERROR.NE.0) THEN
            CALL CONECP(CONFIL,IERRU)
            IQUIT = -3
         ELSE
            IQUIT = -5
         END IF
         RETURN
      END IF
C
C  Copy any errors encountered while reading the configuration file,
C  to the output unit.
      IF (IERROR.NE.0) THEN
         CALL CONEOU(CONFIL,IUOUT,IERRU)
         CALL CLOUT
         IQUIT = -6
         RETURN
      END IF
C
C  Initialize variables that depend on configuration values.
      CALL ZINIT
C
C  Determine machine's accuracy
      CALL MACAC
C
C  Open the Merlin files.
      CALL OPIN(FERROR)
      IF (FERROR) THEN
         IQUIT = -8
         RETURN
      END IF
C
C  Sort the command array, adjusting command indices.
      DO 222,I=1,NCOM
222      IND(I) = I
      CALL CMSORT(COMMAN,IND,NCOM)
C
C  Parse the panel description file.
      CALL PARSP(PDFILE,IUOUT,STRO,WEAK,IERR)
      IF (IERR.NE.0) THEN
         CALL DASH
         IQUIT = -7
         RETURN
      END IF
      IQU = 0
      IRET = 0
C
C  Informative printout
      CALL AUTIDA
C
C  Start backup and history.
      CALL STBALO
C
C  Take appropriate action if a prolog command is defined.
      INPROL = .FALSE.
      INEPIL =.FALSE.
      IPROLE = 0
      IEPILE = 0
      IF (PROLO.NE.' ') THEN
         INPROL = .TRUE.
         IPROLE = LEVIN
         IF (IPROEC.NE.0)
     &          CALL MID1('Prolog execution has been initiated')
         CALL READCO(COM,MERR,1,PROLO)
         GOTO 700
      END IF
C  ---------------------------------------------------------------------
C  The main loop: Command identification and execution starts here.
C  ---------------------------------------------------------------------
1     CONTINUE
C
C  Print the Merlin prompt.
      CALL PROMPT
C
C  Read the next command
      CALL READCO(COM,MERR,0,' ')
C
C  Check if errors occurred, and if so, issue error message
700   CONTINUE
      IF (MERR.EQ.1) THEN
         GOTO 2
      ELSE IF (MERR.NE.0) THEN
         CALL ERHAND(MERR,IDO)
         IF (IDO.NE.0) THEN
            IQUIT = IDO
            RETURN
         ELSE
            GOTO 2
         END IF
      END IF
      IF (COM(1:1).EQ.'$' .AND. INMCL.NE.0) THEN
         CALL USEMCL(COM,MERR)
         IF (MERR.NE.0) THEN
            CALL ERHAND(MERR,IDO)
            IF (IDO.NE.0) THEN
               IQUIT = IDO
               RETURN
            END IF
         END IF
         GOTO 2
      END IF
C
C  Recognize command and transfer the control appropriately.
C  Check for an alias.
      IF (INMCL.EQ.0) THEN
         CALL MATCHA(COM,LEC,ACTCOM,IAL)
         IF (IAL.NE.0) THEN
            COM = ACTCOM
            CALL SUBST(ACTCOM)
         END IF
      END IF
      CALL MATCH(COMMAN,NCOM,COM,LEC,MINF,ICOM,IEND)
      IF (MINF.EQ.-2) THEN
         IF (LEC.GT.10) THEN
            IF (WEAK) WRITE (IUOUT,100) COM//'...'
         ELSE
            IF (WEAK) WRITE (IUOUT,100) COM(1:LEC)
         END IF
         MERR = -2
         CALL ERHAND(MERR,IDO)
         IF (IDO.NE.0) THEN
            IQUIT = IDO
            RETURN
         END IF
      ELSE IF (MINF.EQ.-1) THEN
         CALL AMBI(COM,LEC,ICOM,IEND,NPOS)
         IF (NPOS.EQ.0) THEN
            IF (LEC.GT.10) THEN
               IF (WEAK) WRITE (IUOUT,100) COM//'...'
            ELSE
               IF (WEAK) WRITE (IUOUT,100) COM(1:LEC)
            END IF
         END IF
         MERR = -2
         CALL ERHAND(MERR,IDO)
         IF (IDO.NE.0) THEN
            IQUIT = IDO
            RETURN
         END IF
      ELSE
         COM = COMMAN(ICOM)
         CALL SUBST(COM)
         CALL REDIR(NU,IERR)
         IF (IERR.NE.0) GOTO 2
         ISREDI = NU.NE.-1
         IF (STRO .AND. IECOMA.NE.0) WRITE (IUOUT,10) COMMAN(ICOM)
         IF (ISREDI) THEN
            CALL INCOLE(NU,1,IERR)
            IF (IERR.NE.0) THEN
               CLOSE (NU)
               ISREDI = .FALSE.
            END IF
         END IF
         CALL EXECOM(ICOM,VERSIM,MAXW,COM,ISREDI,MERR)
         IF (ISREDI) CALL DECOLE(1)
         IF (MERR.NE.0) THEN
            CALL ERHAND(MERR,IDO)
            IF (IDO.NE.0) THEN
               IQUIT = IDO
               RETURN
            END IF
         END IF
         IF (IRET.NE.0) THEN
            IF (INEPIL) THEN
C  We are already running our epilog. A RETURN command has been
C  issued from within the epilog. Return immediately.
               IQUIT = IQU
               CALL MEREND(0)
               IF (IRET2.EQ.2) STOP
               RETURN
            ELSE
C  A RETURN/QUIT issued for the first time.
C  Do we have an epilog to execute ?
               IF (EPILO.NE.' ') THEN
                  IRET = 0
                  INEPIL = .TRUE.
                  IEPILE = LEVIN
                  IF (IPROEC.NE.0)
     &             CALL MID1('Epilog execution has been initiated')
                  CALL READCO(COM,MERR,1,EPILO)
                  GOTO 700
               ELSE
C  No epilog. Return at once.
                  IQUIT = IQU
                  CALL MEREND(0)
                  IF (IRET.EQ.2) STOP
                  RETURN
               END IF
            END IF
         END IF
      END IF
C
2     CONTINUE
      IF (LEVIN.EQ.IPROLE .AND. INPROL) THEN
         INPROL = .FALSE.
         IF (IPROEC.NE.0)
     &          CALL MID1('Prolog execution has been completed')
      END IF
      IF (LEVIN.EQ.IEPILE .AND. INEPIL) THEN
         INEPIL = .FALSE.
         IF (IPROEC.NE.0)
     &          CALL MID1('Epilog execution has been completed')
         IQUIT = IQU
         CALL MEREND(0)
         IF (IRET2.EQ.2) STOP
         RETURN
      END IF
      GOTO 1
C
10    FORMAT ($(FC)A)
100   FORMAT (/'"',A,'" is not a valid command.')
150   FORMAT ('The following error(s) have been detected while '/
     &        'processing the Merlin configuration file "',A,'":')
      END
C  ---------------------------------------------------------------------
      BLOCK DATA COMINI
C  ---------------------------------------------------------------------
C
C  Description:
C    The names of all Merlin commands (excluding Plug-in modules) are
C    defined here.
C
C  ---------------------------------------------------------------------
C  Common blocks:
      PARAMETER ( NMCOM = 98, MXPLUG = $(MXPLUG), NCOM = NMCOM+MXPLUG )
      CHARACTER*10 DMCOMS(NCOM)
      COMMON /DCCOM/ DMCOMS
C  ---------------------------------------------------------------------
C
      DATA (DMCOMS(I),I=1,30) /
     &    'ACCUM     ',   'ADJUST    ',   'ANAL      ',
     &    'AUTO      ',   'BATCH     ',   'TRUST     ',
     &    'PDUMP     ',   'CLEAR     ',   'CONGRA    ',
     &    'BFGS      ',   'DFP       ',   'DISCARD   ',
     &    'FIX       ',   'FIXALL    ',   'FULLBACK  ',
     &    'FULLPRINT ',   'GODFATHER ',   'GRADCHECK ',
     &    'GRADDIS   ',   'GRAPH     ',   'HALFPRINT ',
     &    'HELP      ',   'HIDEOUT   ',   'IAF       ',
     &    'DELETE    ',   'INIT      ',   'INSPECT   ',
     &    'EPILOG    ',   'LASTBACK  ',   'LIST      '/
      DATA (DMCOMS(I),I=31,60) /
     &    'LOOSALL   ',   'LOOSE     ',   'MACRO     ',
     &    'PSGRAPH   ',   'LMARGIN   ',   'MEMO      ',
     &    'MODEDIS   ',   'RMARGIN   ',   'NOBACK    ',
     &    'NOPRINT   ',   'NUMER     ',   'PANELOFF  ',
     &    'PANELON   ',   'PICK      ',   'POINT     ',
     &    'QUIT      ',   'HESSIAN   ',   'RESET     ',
     &    'RETURN    ',   'REVEAL    ',   'REWIND    ',
     &    'LDEMARGIN ',   'RUNMCL    ',   'SHORTDIS  ',
     &    'CONFIDENCE',   'STEP      ',   'STEPALL   ',
     &    'STOP      ',   'VALDIS    ',   'NOEVAL    '/
      DATA (DMCOMS(I),I=61,90) /
     &    'FAST      ',   'QUAD      ',   'MIXED     ',
     &    'EVALUATE  ',   'COVARIANCE',   'MAD       ',
     &    'TARGET    ',   'NOTARGET  ',   'SIMPLEX   ',
     &    'CONTROL   ',   'TOLMIN    ',   'ECHO      ',
     &    'STEPDIS   ',   'ALIAS     ',   'ALIASDIS  ',
     &    'UNALIAS   ',   'GOEOF     ',   'TITLE     ',
     &    'LIMITS    ',   'TERMDIS   ',   'GNORM     ',
     &    'FLAG      ',   'FLAGDIS   ',   'CFLAG     ',
     &    'CFLAGDIS  ',   'HISTORY   ',   'PSTATUS   ',
     &    'RDEMARGIN ',   'LEVE      ',   'GENERAL   '/
      DATA (DMCOMS(I),I=91,NMCOM) /
     &    'SOS       ',   'ROLL      ',   'BACKUP    ',
     &    'NONAME    ',   'DUMP      ',   'JANAL     ',
     &    'JNUMER    ',   'JCOMPARE  ' /
C
      END
C  ---------------------------------------------------------------------
      FUNCTION ACSQ ( X, N )
C  ---------------------------------------------------------------------
C
C  Description:
C    All Merlin minimization methods that operate on general functions
C    (not on sum of squares) call this routine. It contains a counter
C    and updates some common block parameters. Its value is set equal
C    to the user supplied function FUNMIN or subroutine SUBSUB,
C    according to the functional form (GENERAL or SOS).
C
C  Input arguments:
C    X          The minimization variables.
C    N          Dimensionality of the objective function.
C
C  ---------------------------------------------------------------------
      IMPLICIT $(TYPE) (A-H,O-Z)
C  Arguments:
      DIMENSION X(N)
C  ---------------------------------------------------------------------
C  Common blocks:
      COMMON /UPDATE/ IGRADU, IHESSU
      COMMON /TOTCAL/ ITNOCS, IPARTC, ITDER, IPDER, ITHESS, IPHESS,
     &                ITJAC, IPJAC
      COMMON /MODE3/ MODEV, IFFORM, INMCL
      COMMON /DIMEN/ NOD, NOF
      PARAMETER ( MXT = $(MXT) )
      DIMENSION F(MXT)
C  ---------------------------------------------------------------------
C
C  Something must be wrong if the objective function is called while
C  NOEVAL is in effect.
      IF (MODEV.EQ.0)
     &    CALL MABORT('ACSQ was called, while evaluation was off.',
     &                'ACSQ')
C
C  Call the user supplied routine.
      IF (IFFORM.EQ.0) THEN
C  Increment counters.
         ITNOCS = ITNOCS+1
         IPARTC = IPARTC+1
         IHESSU = 0
         IGRADU = 0
         ACSQ = FUNMIN(X,N)
      ELSE
         CALL LSQFCN(NOF,N,X,F)
         CALL VVMULT(F,F,SUM,NOF)
         ACSQ = SUM
      END IF
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE LSQFCN ( M, N, X, F )
C  ---------------------------------------------------------------------
C
C  Description:
C    All Merlin minimization methods that operate on functions that are
C    sum of squares, call this routine. It contains a counter
C    and updates some common block parameters. Its value is set equal
C    to the user subroutine SUBSUM.
C
C  Input arguments:
C    M          Number of terms in the sum-of-squares.
C    N          Dimensionality of the objective function.
C    X          The minimization variables.
C
C  Output arguments:
C    F          F(I) contains the value of term I.
C
C  ---------------------------------------------------------------------
      IMPLICIT $(TYPE) (A-H,O-Z)
C  Arguments:
      DIMENSION X(N), F(M)
C  ---------------------------------------------------------------------
C  Common blocks:
      COMMON /UPDATE/ IGRADU, IHESSU
      COMMON /TOTCAL/ ITNOCS, IPARTC, ITDER, IPDER, ITHESS, IPHESS,
     &                ITJAC, IPJAC
      COMMON /MODE3/ MODEV, IFFORM, INMCL
      PARAMETER ( MXT = $(MXT) )
      COMMON /TERMS/ VECLS(MXT)
C  ---------------------------------------------------------------------
C  Local variables:
C  ---------------------------------------------------------------------
C
C  Something must be wrong if the objective function is called while
C  NOEVAL is in effect.
      IF (MODEV.EQ.0)
     &    CALL MABORT('LSQFCN was called, while evaluation was off.',
     &                'LSQFCN')
C
C  Increment counters.
      ITNOCS = ITNOCS+1
      IPARTC = IPARTC+1
      IHESSU = 0
      IGRADU = 0
C
C  Call the user supplied routine.
      CALL SUBSUM(M,N,X,VECLS)
      DO 10,I=1,M
10       F(I) = VECLS(I)
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE ADDPLG
C  ---------------------------------------------------------------------
C
C  Description:
C    Adds the default plug-in entries in the Merlin command array.
C    The entries are added at the end of the command array and are
C    of the form: PLUGxx with xx ranging from 1 to MXPLUG.
C
C  ---------------------------------------------------------------------
C  Common blocks:
      PARAMETER ( NMCOM = 98, MXPLUG = $(MXPLUG), NCOM = NMCOM+MXPLUG )
      CHARACTER*10 COMMAN(NCOM)
      COMMON /COM1/ COMMAN
C  ---------------------------------------------------------------------
C  Local variables:
      CHARACTER*10 STR
C  ---------------------------------------------------------------------
C
      DO 10,I=1,MXPLUG
         CALL I2STR(I,STR,LES)
         COMMAN(NMCOM+I) = 'PLUG'//STR(1:LES)
10    CONTINUE
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE AMBI ( COM, LEC, IST, IEND, NPOS )
C  ---------------------------------------------------------------------
C
C  Description:
C    This subroutine will be called when the user enters an ambiguous
C    abbreviation. It will print a message describing the various
C    possibilities.
C
C  Input arguments:
C    COM        The ambiguous command.
C    LEC        Effective length of COM.
C    IST        Starting position in the Merlin command array.
C    IEND       Ending position in the Merlin command array.
C               All commands from IST to IEND match COM.
C
C  Output arguments:
C    NPOS       Number of possible commands matching the user input.
C               PLUGxx commands are excluded.
C
C  ---------------------------------------------------------------------
C  Arguments:
      CHARACTER*(*) COM
C  ---------------------------------------------------------------------
C  Common blocks:
      PARAMETER ( NMCOM = 98, MXPLUG = $(MXPLUG), NCOM = NMCOM+MXPLUG )
      CHARACTER*10 COMMAN(NCOM)
      COMMON /COM1/ COMMAN
      LOGICAL STRO, WEAK
      COMMON /PRILOG/ STRO, WEAK
      COMMON /UNITS/ IUINP, IUOUT
C  ---------------------------------------------------------------------
C  Local variables:
      CHARACTER*75 LINE
C  ---------------------------------------------------------------------
C
      IF (.NOT.STRO) RETURN
C
      WRITE (IUOUT,10) COM(1:LEC)
C
      NC = 0
      NPOS = 0
      LEL = LEN(LINE)
      DO 20,I=IST,IEND
         IF (COMMAN(I)(1:4).NE.'PLUG') THEN
            LL = LENGTH(COMMAN(I))
            IF (I.EQ.IST) THEN
               LINE(NC+1:NC+LL) = COMMAN(I)(1:LL)
               NC = NC+LL
               NPOS = NPOS+1
            ELSE IF (NC+LL+3.GT.LEL) THEN
               LINE(NC+1:NC+4) = ' ...'
               NC = NC+4
               GOTO 30
            ELSE
               LINE(NC+1:NC+LL+2) = ', '//COMMAN(I)(1:LL)
               NC = NC+LL+2
               NPOS = NPOS+1
            END IF
         END IF
20    CONTINUE
      NC = NC+1
      LINE(NC:NC) = '.'
C
30    CONTINUE
C
C  Issue message only when there are some possible matches.
C  Otherwise the message is issued by the calling routine.
      IF (NPOS.NE.0) WRITE (IUOUT,40) LINE(1:NC)
C
10    FORMAT (/$(FC)'Ambiguous command "',A,'". Possibilities are:')
40    FORMAT (/$(FC)A)
      END
C  ---------------------------------------------------------------------
      SUBROUTINE AUTIDA
C  ---------------------------------------------------------------------
C
C  Description:
C    Subroutine AUTIDA is called during Merlin startup. It issues
C    information about the authors of the program, version, e-mail etc.
C    It also checks the existence of the help and panel description
C    files, and issues appropriate messages.
C
C  ---------------------------------------------------------------------
      IMPLICIT $(TYPE) (A-H,O-Z)
C  Common blocks:
      LOGICAL HEADER, IPRECO, OPRECO, HASAPP, OUTAPP
      COMMON /CP1/ HEADER, IPRECO, OPRECO, HASAPP, OUTAPP
      COMMON /UNITS/ IUINP, IUOUT
      LOGICAL STRO, WEAK
      COMMON /PRILOG/ STRO, WEAK
      PARAMETER ( MXFILE = 80 )
      CHARACTER*(MXFILE) PROLO, EPILO, IN, OUT, HFILE, MFILE, PDFILE
      COMMON /CP2/ PROLO, EPILO, IN, OUT, HFILE, MFILE, PDFILE
      COMMON /DIMEN/ NOD, NOF
      COMMON /ACCURA/ COMACC
C  ---------------------------------------------------------------------
C  Local variables:
      LOGICAL EXI, OPE
      CHARACTER*30 STR
C  ---------------------------------------------------------------------
C
      IF (.NOT.STRO) RETURN
C
C  Print the standard Merlin header.
      IF (HEADER) THEN
         CALL DASH
         WRITE (IUOUT,10)
         WRITE (IUOUT,20)
         WRITE (IUOUT,10)
         WRITE (IUOUT,30)
         WRITE (IUOUT,40)
         WRITE (IUOUT,50)
         WRITE (IUOUT,10)
         WRITE (IUOUT,60)
         WRITE (IUOUT,70)
         WRITE (IUOUT,10)
      END IF
C
C  Check the help file.
      CALL DASH
      LEF = LENGTH(HFILE)
      IF (LEF.NE.0) THEN
         INQUIRE (FILE=HFILE,EXIST=EXI,OPENED=OPE,NUMBER=NUM,
     &               IOSTAT=IOE)
         IF (IOE.EQ.0) THEN
            IF (EXI) THEN
               IF (OPE) THEN
                  INFO = 1
                  CALL I2STR(NUM,STR,LES)
                  WRITE (IUOUT,100) HFILE(1:LEF), STR(1:LES)
               ELSE
                  INFO = 2
                  WRITE (IUOUT,80) HFILE(1:LEF)
               END IF
            ELSE
               INFO = 0
               WRITE (IUOUT,90) HFILE(1:LEF)
            END IF
         ELSE
            INFO = 1
            CALL I2STR(IOE,STR,LES)
            WRITE (IUOUT,110) HFILE(1:LEF), STR(1:LES)
         END IF
      ELSE
         INFO = 0
         WRITE (IUOUT,120)
      END IF
C
C  Check the panel description file.
      LEF = LENGTH(PDFILE)
      IF (LEF.NE.0) THEN
         INQUIRE (FILE=PDFILE,EXIST=EXI,OPENED=OPE,NUMBER=NUM,
     &               IOSTAT=IOE)
         IF (IOE.EQ.0) THEN
            IF (EXI) THEN
               IF (OPE) THEN
                  IP = 1
                  CALL I2STR(NUM,STR,LES)
                  WRITE (IUOUT,300) PDFILE(1:LEF), STR(1:LES)
               ELSE
                  IP = 2
                  WRITE (IUOUT,380) PDFILE(1:LEF)
               END IF
            ELSE
               IP = 0
               WRITE (IUOUT,390) PDFILE(1:LEF)
            END IF
         ELSE
            IP = 1
            CALL I2STR(IOE,STR,LES)
            WRITE (IUOUT,310) PDFILE(1:LEF), STR(1:LES)
         END IF
      ELSE
         IP = 0
         WRITE (IUOUT,120)
      END IF
C
C  Can the user obtain any help ?
      IF (INFO.EQ.0 .AND. IP.EQ.0) THEN
         IF (STRO) WRITE (IUOUT,230)
      ELSE IF (INFO.EQ.2 .AND. IP.EQ.2) THEN
         IF (STRO) WRITE (IUOUT,250)
      ELSE
         IF (STRO) WRITE (IUOUT,240)
      END IF
      CALL DASH
C
C  Print # of variables, terms and machine accuracy.
      CALL I2STR(NOF,STR,LES)
      WRITE (IUOUT,140) STR(1:LES)
      CALL I2STR(NOD,STR,LES)
      WRITE (IUOUT,130) STR(1:LES)
      CALL R2STR(COMACC,STR,LES)
      WRITE (IUOUT,150) STR(1:LES)
      IF (NOF.EQ.0) THEN
         WRITE (IUOUT,160) '"FUNCTION FUNMIN"'
      ELSE
         WRITE (IUOUT,160) '"SUBROUTINE SUBSUM"'
      END IF
C
10    FORMAT ($(FC)27('.'))
20    FORMAT ($(FC)27('.'),14X,'M E R L I N - 3.0.4 / May 2002')
30    FORMAT ($(FC)27('.'),13X,'D.G. Papageorgiou, I.E. Lagaris',
     &       /$(FC)27('.'),19X,'I.N. Demetropoulos')
40    FORMAT ($(FC)27('.'),17X,'University of Ioannina')
50    FORMAT ($(FC)27('.'),23X,'G R E E C E')
60    FORMAT ($(FC)27('.'),15X,'Email: merlin@nrt.cs.uoi.gr')
70    FORMAT ($(FC)27('.'),12X,'Web: http://nrt.cs.uoi.gr/merlin')
C
80    FORMAT ($(FC)'The Merlin help file "',A,'" is present.')
90    FORMAT ($(FC)'The Merlin help file "',A,'" is not present.')
100   FORMAT ($(FC)'The Merlin help file "',A,'" is present ',
     &        'but opened. UNIT = ',A,'.')
110   FORMAT ($(FC)'Cannot inquire the Merlin help file "',A,'". ',
     &        'IOSTAT = ',A,'.')
120   FORMAT ($(FC)'Sorry, there is no help file.')
C
300   FORMAT ($(FC)'The panel description file "',A,'" is present ',
     &        'but opened. UNIT = ',A,'.')
380   FORMAT ($(FC)'The panel description file "',A,'" is present.')
390   FORMAT ($(FC)'The panel description file "',A,'" is not ',
     &        'present.')
310   FORMAT ($(FC)'Cannot inquire the panel description file "',
     &        A,'". IOSTAT = ',A,'.')
C
230   FORMAT ($(FC)'You will not be able to obtain on-line ',
     &        'information.')
240   FORMAT ($(FC)'You may not be able to obtain on-line ',
     &        'information.')
250   FORMAT ($(FC)'Use the "help" command to obtain on-line ',
     &        'information.')
C
140   FORMAT ($(FC)'Number of terms:     ',A)
130   FORMAT ($(FC)'Number of variables: ',A)
150   FORMAT ($(FC)'Estimated machine''s accuracy: ',A)
160   FORMAT ($(FC)'Merlin uses ',A,' as the objective function.')
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE CMSORT ( COMMAN, IND, NCOM )
C  ---------------------------------------------------------------------
C
C  Description:
C    Subroutine CMSORT sorts the Merlin command array in ascending
C    order. At the same time it readjusts the contents of array IND.
C    For information on the sorting method (bubble sort) see:
C      N. Wirth, Algorithms & Data Structures,
C      Prentice-Hall, 1986.
C      Knuth D.E., The art of computer programming, vol 3,
C      Addison-Wesley, Reading, Mass., 1973.
C
C  Input arguments:
C    NCOM       Number of commands in the Merlin command array.
C
C  Input / output arguments:
C    COMMAN     The Merlin command array. On output it is sorted.
C    IND        Indices of the Merlin command array. On input contains
C               the sequence 1,2,3, ... NCOM. On output IND(i) contains
C               the position of COMMAN(i) in the original, unsorted
C               array.
C
C  ---------------------------------------------------------------------
C  Arguments:
      CHARACTER*(*) COMMAN(NCOM)
      INTEGER IND(NCOM)
C  ---------------------------------------------------------------------
C  Local variables:
      LOGICAL SORTED
      CHARACTER*15 CTEMP
C  ---------------------------------------------------------------------
C
      DO 10,K=1,NCOM-1
         SORTED = .TRUE.
         DO 20,I=NCOM,K+1,-1
            IF (COMMAN(I).LT.COMMAN(I-1)) THEN
               SORTED = .FALSE.
               IM1 = I-1
               CTEMP = COMMAN(I)
               COMMAN(I) = COMMAN(IM1)
               COMMAN(IM1) = CTEMP
               ITEMP = IND(I)
               IND(I) = IND(IM1)
               IND(IM1) = ITEMP
            END IF
20       CONTINUE
         IF (SORTED) RETURN
10    CONTINUE
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE CNTDIS
C  ---------------------------------------------------------------------
C
C  Description:
C    Displays the partial and total counters for function, gradient
C    Hessian and Jacobian calls.
C
C  ---------------------------------------------------------------------
C  Common blocks:
      LOGICAL STRO, WEAK
      COMMON /PRILOG/ STRO, WEAK
      COMMON /UNITS/ IUINP, IUOUT
      COMMON /TOTCAL/ ITNOCS, IPARTC, ITDER, IPDER, ITHESS, IPHESS,
     &                ITJAC, IPJAC
C  ---------------------------------------------------------------------
C
      IF (STRO) THEN
         CALL DASH
         WRITE (IUOUT,10)
         WRITE (IUOUT,20) ITNOCS, ITDER, ITHESS, ITJAC
         WRITE (IUOUT,30) IPARTC, IPDER, IPHESS, IPJAC
         CALL DASH
      END IF
C
10    FORMAT ($(FC)'Number of evaluations:    Function       Gradient',
     &        '       Hessian      Jacobian')
20    FORMAT ($(FC)'Total:',20X,I8,7X,I8,6X,I8,6X,I8)
30    FORMAT ($(FC)'Since last reset:',9X,I8,7X,I8,6X,I8,6X,I8)
      END
C  ---------------------------------------------------------------------
      SUBROUTINE CONPRO ( CFILE, ICONU, IFILEU, IERRU, IERROR )
C  ---------------------------------------------------------------------
C
C  Description:
C    Subroutine CONPRO acts as a front end to subroutine MCONF that
C    parses the Merlin configuration file.
C
C  Input arguments:
C    CFILE      Name of the Merlin configuration file.
C    ICONU      The configuration file will be opened with this unit
C               number.
C    IFILEU     Unit number to be used for FILE directives.
C    IERRU      Unit number for the temporary file that accumulates
C               error messages during the run of subroutine MCONF.
C
C  Output arguments:
C    IERROR     Error indicator:
C               IERROR = 0 -> No error.
C               IERROR = 1 -> Some error.
C               IERROR = 2 -> Some error related to directives that
C                             alter I/O units/files.
C
C  ---------------------------------------------------------------------
      IMPLICIT $(TYPE) (A-H,O-Z)
C  Arguments:
      CHARACTER*(*) CFILE
C  ---------------------------------------------------------------------
C  Common blocks:
      CHARACTER*10 DERIVA, PRINTO, PROCES, JACOMO
      COMMON /MODEIN/ DERIVA, PRINTO, PROCES, JACOMO
      COMMON /BIGCOM/ BIGNUM, BLIM, DWARF, DWALOG
      LOGICAL HEADER, IPRECO, OPRECO, HASAPP, OUTAPP
      COMMON /CP1/ HEADER, IPRECO, OPRECO, HASAPP, OUTAPP
      PARAMETER ( MXFILE = 80 )
      CHARACTER*(MXFILE) PROLO, EPILO, IN, OUT, HFILE, MFILE, PDFILE
      COMMON /CP2/ PROLO, EPILO, IN, OUT, HFILE, MFILE, PDFILE
      COMMON /CP3/ NEOFA
      LOGICAL STRO, WEAK
      COMMON /PRILOG/ STRO, WEAK
      COMMON /UNITS/ IUINP, IUOUT
      COMMON /UALLOC/ MINU, MAXU, LASTU
      PARAMETER ( NURANG = 15 )
      PARAMETER ( NMCOM = 98, MXPLUG = $(MXPLUG), NCOM = NMCOM+MXPLUG )
      CHARACTER*10 COMMAN(NCOM)
      COMMON /COM1/ COMMAN
      COMMON /VSIZES/ NSCHA, NSINT, NSREAL
      COMMON /ACCUR1/ NDRTF
C  ---------------------------------------------------------------------
C  Local variables:
      CHARACTER EXCLU(1), MCLEF, MCLOF
C  ---------------------------------------------------------------------
C
      CALL MCONF(CFILE,ICONU,IFILEU,IERRU,NMCOM,MXPLUG,COMMAN,EXCLU,1,
     &           NURANG,IERROR,HEADER,PROCES,PROLO,EPILO,IN,OUT,PDFILE,
     &           STRO,WEAK,IPRECO,OPRECO,BIGNUM,DWARF,IUINP,IUOUT,MINU,
     &           MAXU,HASAPP,OUTAPP,NSCHA,NSINT,NSREAL,HFILE,MFILE,
     &           MCLEF,MCLOF,NDRTF,NEOFA)
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE CVSEL ( H, LD, ISEL, NSEL, NCOV )
C  ---------------------------------------------------------------------
C
C  Description:
C    Shrinks an NCOV x NCOV matrix, retaining the rows and columns
C    contained in the input array ISEL. The resulting matrix
C    is an NSEL x NSEL matrix. This routine is used to select some
C    rows and columns of a covariance matrix, in order to determine
C    the confidence intervals.
C    Note that ISEL must be sorted in ascending order. This is
C    guaranteed if subroutine CVSEL is called using the output of one
C    of the CHI routines.
C
C  Input arguments:
C    LD         Leading dimension of matrix H.
C    NCOV       Initial dimensionality of H.
C
C  Input / Output arguments:
C    H          On input an NCOV x NCOV matrix.
C               On output an NSEL x NSEL matrix.
C    ISEL       Integer array of NSEL elements. On input it contains
C               the numbers of rows and columns to be retained in H.
C               If ISEL contains the indices of any fixed variables,
C               these are removed on output.
C    NSEL       On input is the number of elements in array ISEL.
C               On output is the number of non-fixed variables that
C               were present in array ISEL.
C
C  ---------------------------------------------------------------------
      IMPLICIT $(TYPE) (A-H,O-Z)
C  Arguments:
      INTEGER ISEL(*)
      DIMENSION H(LD,*)
C  ---------------------------------------------------------------------
C  Common blocks:
      PARAMETER ( MXV = $(MXV) )
      COMMON /DETAIL/ IXAT(MXV), MARG(MXV)
      LOGICAL STRO, WEAK
      COMMON /PRILOG/ STRO, WEAK
      COMMON /UNITS/ IUINP, IUOUT
C  ---------------------------------------------------------------------
C  Local variables:
      CHARACTER*30 STR
C  ---------------------------------------------------------------------
C
C  Remove fixed variables.
      NOK = 0
      DO 10,I=1,NSEL
         IND = ISEL(I)
         IF (IXAT(IND).EQ.0) THEN
            IF (WEAK) THEN
               CALL VNAME(IND,STR,LES)
               WRITE (IUOUT,70) STR(1:LES)
            END IF
         ELSE
            NOK = NOK+1
            IF (NOK.NE.I) ISEL(NOK) = ISEL(I)
         END IF
10    CONTINUE
      NSEL = NOK
C
C  Change numbering. ISEL now contains the actual covariance matrix
C  indices. Note that fixed variables are removed from the covariance
C  matrix, resulting thus in smaller dimensionality.
      DO 20,I=1,NSEL
         IND = ISEL(I)
         NF = 0
         DO 30,J=1,IND
30          IF (IXAT(J).NE.0) NF = NF+1
         ISEL(I) = NF
20    CONTINUE
C
C  Shrink the covariance matrix.
      DO 60,I=1,NSEL
         IND = ISEL(I)
         IF (IND.NE.I) THEN
            DO 40,J=1,NCOV
40             H(J,I) = H(J,IND)
         END IF
         DO 50,J=1,NCOV
            IX = ISEL(J)
            IF (IX.NE.J) H(J,I) = H(IX,I)
50       CONTINUE
60    CONTINUE
C
70    FORMAT (/$(FC)'Warning: Variable ',A,' is fixed. Ignored.')
      END
C  ---------------------------------------------------------------------
      SUBROUTINE DETUNT ( IUDEF, IU, IERR )
C  ---------------------------------------------------------------------
C
C  Description:
C    Subroutine DETUNT determines a unit number (IU) to be used in I/O
C    operations during an MCL program.
C    The MOC file is expected to contain:
C    <File_Name>  -> Name of the file or .STANDARD for the default
C                    input/output unit.
C
C  Input arguments:
C    IUDEF      Unit number to be used when the default input/output
C               file is specified.
C
C  Output arguments:
C    IU         The unit number to be used. The corresponding file is
C               open and  ready for I/O.
C    IERR       Error indicator.
C
C  ---------------------------------------------------------------------
      IMPLICIT $(TYPE) (A-H,O-Z)
C  Common blocks:
      PARAMETER ( MAXMFI = 20 )
      COMMON /MCLV2/ ISTACK, INFILE, MBASE, INAD, MPC, MCLLNO, LENBUF,
     &               IBUF, MCLUNI(MAXMFI), NUMFI
      LOGICAL STRO, WEAK
      COMMON /PRILOG/ STRO, WEAK
      COMMON /UNITS/ IUINP, IUOUT
C  ---------------------------------------------------------------------
C  Local variables:
      PARAMETER ( MXFILE = 80 )
      CHARACTER*(MXFILE) NAME
      CHARACTER*10 STR
      LOGICAL OPE
C  ---------------------------------------------------------------------
C
      IERR = 0
C
C  Get the filename.
      CALL INNOCR(NAME,.FALSE.,IERR)
      IF (IERR.NE.0) RETURN
      IF (NAME.EQ.'.STANDARD') THEN
         IU = IUDEF
         RETURN
      END IF
C
C  Get info for the file.
      INQUIRE (FILE=NAME,OPENED=OPE,NUMBER=IU,IOSTAT=IERR)
      IF (IERR.NE.0) THEN
         CALL ERRINQ(NAME,IERR)
         IERR = -2
         RETURN
      END IF
C
C  Is it open ?
      IF (OPE) THEN
         DO 10,I=1,NUMFI
10          IF (IU.EQ.MCLUNI(I)) GOTO 15
C
C  Not opened by MCL. We are not allowed to read/write files not
C  opened by MCL.
         IF (WEAK) THEN
            LENA = LENGTH(NAME)
            WRITE (IUOUT,20) NAME(1:LENA)
         END IF
         IERR = -2
         RETURN
15       CONTINUE
      ELSE
C  Not open. Assign new unit and open the file.
         IU = NUNIT()
         IF (IU.EQ.-1) THEN
            CALL ERRUAL(NAME)
            IERR = -2
            RETURN
         END IF
         OPEN (UNIT=IU,FILE=NAME,IOSTAT=IERR)
         IF (IERR.NE.0) THEN
            CALL ERROPE(NAME,IERR)
            IERR = -2
            RETURN
         END IF
         REWIND IU
C  Increment MCL file counter.
         NUMFI = NUMFI+1
         IF (NUMFI.GT.MAXMFI) THEN
C  Too many files.
            IF (WEAK) THEN
               CALL I2STR(MAXMFI,STR,LES)
               LENA = LENGTH(NAME)
               WRITE (IUOUT,30) STR(1:LES), NAME(1:LENA)
            END IF
            IERR = -2
            CLOSE (IU)
            RETURN
         END IF
         MCLUNI(NUMFI) = IU
      END IF
C
20    FORMAT (/$(FC)'Error: File "',A,'" is already open; ',
     &        'not by MCL however.')
30    FORMAT (/$(FC)'Error: You have too many (',A,') MCL files open.'
     &        /$(FC)'       File "',A,'" cannot be opened.')
      END
C  ---------------------------------------------------------------------
      SUBROUTINE ERHAND ( MERR, IDO )
C  ---------------------------------------------------------------------
C
C  Description:
C    This is the Merlin error handling routine. When a command returns
C    a non-zero error indicator (MERR), this routine is called to
C    decipher the error code and act accordingly. Currently only
C    two error codes are considered lethal: -1 and -2. When in batch
C    mode, this routine always aborts; properly however, exiting
C    MCL programs and macros, and closing all files.
C
C  Input arguments:
C    MERR       The error code.
C               MERR =  0 -> No error.
C               MERR = -1 -> End-of-File has been encountered in
C                            the Merlin input file.
C               MERR = -2 -> Some other error.
C               Any other value is meaningless and the error
C               handler just returns.
C
C  Output arguments:
C    IDO        In case this routine returns, IDO determines what
C               merlin should do:
C               IDO = 0 -> Do nothing. Proceed with reading a new
C                          command.
C               IDO.NE.0 -> Merlin must return immediately to the
C                           calling program. The value of IDO will
C                           be used as the Merlin quit flag.
C
C  ---------------------------------------------------------------------
C  Common blocks:
      CHARACTER*10 DERIVA, PRINTO, PROCES, JACOMO
      COMMON /MODEIN/ DERIVA, PRINTO, PROCES, JACOMO
      COMMON /CP3/ NEOFA
      PARAMETER ( MAXMFI = 20 )
      COMMON /MCLV2/ ISTACK, INFILE, MBASE, INAD, MPC, MCLLNO, LENBUF,
     &               IBUF, MCLUNI(MAXMFI), NUMFI
      CHARACTER*80 MCLCLI
      COMMON /MCLV3/ MCLCLI
      COMMON /MODE3/ MODEV, IFFORM, INMCL
      LOGICAL STRO, WEAK
      COMMON /PRILOG/ STRO, WEAK
      COMMON /UNITS/ IUINP, IUOUT
C  ---------------------------------------------------------------------
C  Local variables:
      CHARACTER*10 STR
C  ---------------------------------------------------------------------
C
C  -1 and -2 are the only fatal errors.
      IDO = 0
      IF (MERR.NE.-1 .AND. MERR.NE.-2) RETURN
C
      IF (PROCES.EQ.'BATCH') THEN
C  Batch mode.
         IF (MERR.EQ.-1 .AND. WEAK) WRITE (IUOUT,10)
C  Close files, exit macros, MCL programs etc.
         IF (INMCL.NE.0) THEN
            IF (MCLLNO.NE.-1 .AND. WEAK) THEN
               CALL I2STR(MCLLNO,STR,LES)
               LL = LENGTH(MCLCLI)
               WRITE (IUOUT,20) STR(1:LES), MCLCLI(1:LL)
            END IF
            CALL MCLEND
         END IF
         CALL MEREND(1)
C  Set the return flag.
         IF (MERR.EQ.-1) THEN
            IDO = -10
         ELSE
            IDO = -20
         END IF
      ELSE
C
C  We are running interactively.
         IF (MERR.EQ.-1) THEN
C  Deal with EOF.
            IF (WEAK) WRITE (IUOUT,10)
            IF (NEOFA.EQ.1) THEN
               REWIND IUINP
            ELSE IF (NEOFA.EQ.3) THEN
               CALL MEREND(1)
               IDO = -10
            END IF
         ELSE
C  Any other error.
            IF (INMCL.NE.0) THEN
               IF (MCLLNO.NE.-1 .AND. WEAK) THEN
                  CALL I2STR(MCLLNO,STR,LES)
                  LL = LENGTH(MCLCLI)
                  WRITE (IUOUT,20) STR(1:LES), MCLCLI(1:LL)
               END IF
               CALL MCLEND
            END IF
            CALL TOPLEV(1)
         END IF
      END IF
C
10    FORMAT (/$(FC)'End-of-File encountered.')
20    FORMAT (/$(FC)'An error occurred in MCL source line ',A,'.'
     &        /$(FC)'The source line was: ',A)
      END
C  ---------------------------------------------------------------------
      SUBROUTINE ERRINQ ( FILE, IOE )
C  ---------------------------------------------------------------------
C
C  Description:
C    Part of the Merlin error handling routines. Issues an error message
C    for INQUIRE statements that fail.
C
C  Input arguments:
C    FILE       File that was INQUIREd.
C    IOE        Value of the IOSTAT specifier.
C
C  ---------------------------------------------------------------------
C  Arguments:
      CHARACTER*(*) FILE
C  ---------------------------------------------------------------------
C  Common blocks:
      LOGICAL STRO, WEAK
      COMMON /PRILOG/ STRO, WEAK
      COMMON /UNITS/ IUINP, IUOUT
C  ---------------------------------------------------------------------
C  Local variables:
      CHARACTER*10 STR
C  ---------------------------------------------------------------------
C
      IF (WEAK) THEN
         CALL I2STR(IOE,STR,LES)
         LEF = LENGTH(FILE)
         WRITE (IUOUT,10) FILE(1:LEF), STR(1:LES)
      END IF
C
10    FORMAT (/$(FC)'Error: Cannot inquire file "',A,'". IOSTAT = ',
     &       A,'.')
      END
C  ---------------------------------------------------------------------
      SUBROUTINE ERROPE ( FILE,  IOE )
C  ---------------------------------------------------------------------
C
C  Description:
C    Part of the Merlin error handling routines. Issues an error message
C    for OPEN statements that fail.
C
C  Input arguments:
C    FILE       File that was OPENed.
C    IOE        Value of the IOSTAT specifier.
C
C  ---------------------------------------------------------------------
C  Arguments:
      CHARACTER*(*) FILE
C  ---------------------------------------------------------------------
C  Common blocks:
      LOGICAL STRO, WEAK
      COMMON /PRILOG/ STRO, WEAK
      COMMON /UNITS/ IUINP, IUOUT
C  ---------------------------------------------------------------------
C  Local variables:
      CHARACTER*10 STR
C  ---------------------------------------------------------------------
C
      IF (WEAK) THEN
         CALL I2STR(IOE,STR,LES)
         LEF = LENGTH(FILE)
         WRITE (IUOUT,10) FILE(1:LEF), STR(1:LES)
      END IF
C
10    FORMAT (/$(FC)'Error: Cannot open file "',A,'". IOSTAT = ',A,'.')
      END
C  ---------------------------------------------------------------------
      SUBROUTINE ERRCLO ( FILE,  IOE )
C  ---------------------------------------------------------------------
C
C  Description:
C    Part of the Merlin error handling routines. Issues an error message
C    for CLOSE statements that fail.
C
C  Input arguments:
C    FILE       File that was CLOSEd.
C    IOE        Value of the IOSTAT specifier.
C
C  ---------------------------------------------------------------------
C  Arguments:
      CHARACTER*(*) FILE
C  ---------------------------------------------------------------------
C  Common blocks:
      LOGICAL STRO, WEAK
      COMMON /PRILOG/ STRO, WEAK
      COMMON /UNITS/ IUINP, IUOUT
C  ---------------------------------------------------------------------
C  Local variables:
      CHARACTER*10 STR
C  ---------------------------------------------------------------------
C
      IF (WEAK) THEN
         CALL I2STR(IOE,STR,LES)
         LEF = LENGTH(FILE)
         WRITE (IUOUT,10) FILE(1:LEF), STR(1:LES)
      END IF
C
10    FORMAT (/$(FC)'Error: Cannot close file "',A,'". IOSTAT = ',
     &        A,'.')
      END
C  ---------------------------------------------------------------------
      SUBROUTINE ERRREW ( FILE,  IOE )
C  ---------------------------------------------------------------------
C
C  Description:
C    Part of the Merlin error handling routines. Issues an error message
C    for REWIND statements that fail.
C
C  Input arguments:
C    FILE       File that was REWound.
C    IOE        Value of the IOSTAT specifier.
C
C  ---------------------------------------------------------------------
C  Arguments:
      CHARACTER*(*) FILE
C  ---------------------------------------------------------------------
C  Common blocks:
      LOGICAL STRO, WEAK
      COMMON /PRILOG/ STRO, WEAK
      COMMON /UNITS/ IUINP, IUOUT
C  ---------------------------------------------------------------------
C  Local variables:
      CHARACTER*10 STR
C  ---------------------------------------------------------------------
C
      IF (WEAK) THEN
         CALL I2STR(IOE,STR,LES)
         LEF = LENGTH(FILE)
         WRITE (IUOUT,10) FILE(1:LEF), STR(1:LES)
      END IF
C
10    FORMAT (/$(FC)'Error: Cannot rewind file "',A,'". IOSTAT = ',
     &        A,'.')
      END
C  ---------------------------------------------------------------------
      SUBROUTINE ERRBUS ( FILE,  IU )
C  ---------------------------------------------------------------------
C
C  Description:
C    Part of the Merlin error handling routines. Issues an error message
C    for operations attempted on already open files.
C
C  Input arguments:
C    FILE       File name.
C    IU         Unit number of the open file.
C
C  ---------------------------------------------------------------------
C  Arguments:
      CHARACTER*(*) FILE
C  ---------------------------------------------------------------------
C  Common blocks:
      LOGICAL STRO, WEAK
      COMMON /PRILOG/ STRO, WEAK
      COMMON /UNITS/ IUINP, IUOUT
C  ---------------------------------------------------------------------
C  Local variables:
      CHARACTER*10 STR
C  ---------------------------------------------------------------------
C
      IF (WEAK) THEN
         CALL I2STR(IU,STR,LES)
         LEF = LENGTH(FILE)
         WRITE (IUOUT,10) FILE(1:LEF), STR(1:LES)
      END IF
C
10    FORMAT (/$(FC)'Error: File "',A,'" is already open. ',
     &        'UNIT = ',A,'.')
      END
C  ---------------------------------------------------------------------
      SUBROUTINE ERREXI ( FILE )
C  ---------------------------------------------------------------------
C
C  Description:
C    Part of the Merlin error handling routines. Issues an error message
C    for operations attempted on already existing files.
C
C  Input arguments:
C    FILE       File name.
C
C  ---------------------------------------------------------------------
C  Arguments:
      CHARACTER*(*) FILE
C  ---------------------------------------------------------------------
C  Common blocks:
      LOGICAL STRO, WEAK
      COMMON /PRILOG/ STRO, WEAK
      COMMON /UNITS/ IUINP, IUOUT
C  ---------------------------------------------------------------------
C
      IF (WEAK) THEN
         LEF = LENGTH(FILE)
         WRITE (IUOUT,10) FILE(1:LEF)
      END IF
C
10    FORMAT (/$(FC)'Error: File "',A,'" does not exist.')
      END
C  ---------------------------------------------------------------------
      SUBROUTINE ERRUAL ( FILE )
C  ---------------------------------------------------------------------
C
C  Description:
C    Part of the Merlin error handling routines. Issues an error message
C    when the unit allocator (function NUNIT) fails.
C
C  Input arguments:
C    FILE       File name.
C
C  ---------------------------------------------------------------------
C  Arguments:
      CHARACTER*(*) FILE
C  ---------------------------------------------------------------------
C  Common blocks:
      LOGICAL STRO, WEAK
      COMMON /PRILOG/ STRO, WEAK
      COMMON /UNITS/ IUINP, IUOUT
C  ---------------------------------------------------------------------
C
      IF (WEAK) THEN
         LEF = LENGTH(FILE)
         WRITE (IUOUT,10) FILE(1:LEF)
      END IF
C
10    FORMAT (/$(FC)'Error: Cannot allocate unit number for file "',
     &       A,'".')
      END
C  ---------------------------------------------------------------------
      SUBROUTINE ERRNAR
C  ---------------------------------------------------------------------
C
C  Description:
C    Part of the Merlin error handling routines. Issues an error message
C    when an incorrect number of arguments is entered in the command
C    line.
C
C  ---------------------------------------------------------------------
C  Common blocks:
      LOGICAL STRO, WEAK
      COMMON /PRILOG/ STRO, WEAK
      COMMON /UNITS/ IUINP, IUOUT
C  ---------------------------------------------------------------------
C
      IF (WEAK) WRITE (IUOUT,10)
C
10    FORMAT (/$(FC)'Error: Incorrect number of arguments.')
      END
C  ---------------------------------------------------------------------
      SUBROUTINE ERRARG ( ARG )
C  ---------------------------------------------------------------------
C
C  Description:
C    Part of the Merlin error handling routines. Issues an error message
C    when an incorrect argument is entered in a Merlin command.
C
C  Input arguments:
C    ARG        The incorrect argument.
C
C  ---------------------------------------------------------------------
C  Arguments:
      CHARACTER*(*) ARG
C  ---------------------------------------------------------------------
C  Common blocks:
      LOGICAL STRO, WEAK
      COMMON /PRILOG/ STRO, WEAK
      COMMON /UNITS/ IUINP, IUOUT
C  ---------------------------------------------------------------------
C
      IF (WEAK) THEN
         LEA = LENGTH(ARG)
         WRITE (IUOUT,10) ARG(1:LEA)
      END IF
C
10    FORMAT (/$(FC)'Error: Argument "',A,'" is not permitted.')
      END
C  ---------------------------------------------------------------------
      SUBROUTINE ERRRD ( FILE, IOE )
C  ---------------------------------------------------------------------
C
C  Description:
C    Part of the Merlin error handling routines. Issues an error message
C    when a READ statement fails.
C
C  Input arguments:
C    FILE       File name.
C    IOE        Value of the IOSTAT specifier.
C
C  ---------------------------------------------------------------------
C  Arguments:
      CHARACTER*(*) FILE
C  ---------------------------------------------------------------------
C  Common blocks:
      LOGICAL STRO, WEAK
      COMMON /PRILOG/ STRO, WEAK
      COMMON /UNITS/ IUINP, IUOUT
C  ---------------------------------------------------------------------
C  Local variables:
      CHARACTER*10 STR
C  ---------------------------------------------------------------------
C
      IF (WEAK) THEN
         CALL I2STR(IOE,STR,LES)
         LEF = LENGTH(FILE)
         WRITE (IUOUT,10) FILE(1:LEF), STR(1:LES)
      END IF
C
10    FORMAT (/$(FC)'Error: Cannot read from file "',A,'". IOSTAT = ',
     &       A,'.')
      END
C  ---------------------------------------------------------------------
      SUBROUTINE ERRWR ( FILE, IOE )
C  ---------------------------------------------------------------------
C
C  Description:
C    Part of the Merlin error handling routines. Issues an error message
C    when a WRITE statement fails.
C
C  Input arguments:
C    FILE       File name.
C    IOE        Value of the IOSTAT specifier.
C
C  ---------------------------------------------------------------------
C  Arguments:
      CHARACTER*(*) FILE
C  ---------------------------------------------------------------------
C  Common blocks:
      LOGICAL STRO, WEAK
      COMMON /PRILOG/ STRO, WEAK
      COMMON /UNITS/ IUINP, IUOUT
C  ---------------------------------------------------------------------
C  Local variables:
      CHARACTER*10 STR
C  ---------------------------------------------------------------------
C
      IF (WEAK) THEN
         CALL I2STR(IOE,STR,LES)
         LEF = LENGTH(FILE)
         WRITE (IUOUT,10) FILE(1:LEF), STR(1:LES)
      END IF
C
10    FORMAT (/$(FC)'Error: Cannot write to file "',A,'". IOSTAT = ',
     &       A,'.')
      END
C  ---------------------------------------------------------------------
      SUBROUTINE ERRZFI
C  ---------------------------------------------------------------------
C
C  Description:
C    Part of the Merlin error handling routines. Issues an error message
C    when a file name is missing or has zero length.
C
C  ---------------------------------------------------------------------
C  Common blocks:
      LOGICAL STRO, WEAK
      COMMON /PRILOG/ STRO, WEAK
      COMMON /UNITS/ IUINP, IUOUT
C  ---------------------------------------------------------------------
C
      IF (WEAK) WRITE (IUOUT,10)
C
10    FORMAT (/$(FC)'Error: You must specify a file name.')
      END
C  ---------------------------------------------------------------------
      SUBROUTINE WRNEXI ( FILE )
C  ---------------------------------------------------------------------
C
C  Description:
C    Part of the Merlin error handling routines. Issues a warning
C    message for a file that does not exist.
C
C  Input arguments:
C    FILE       File name.
C
C  ---------------------------------------------------------------------
C  Arguments:
      CHARACTER*(*) FILE
C  ---------------------------------------------------------------------
C  Common blocks:
      LOGICAL STRO, WEAK
      COMMON /PRILOG/ STRO, WEAK
      COMMON /UNITS/ IUINP, IUOUT
C  ---------------------------------------------------------------------
C
      IF (WEAK) THEN
         LEF = LENGTH(FILE)
         WRITE (IUOUT,10) FILE(1:LEF)
      END IF
C
10    FORMAT (/$(FC)'Warning: File "',A,'" does not exist.')
      END
C  ---------------------------------------------------------------------
      SUBROUTINE EXECOM ( ICOM, VERSIM, MAXW, COM, ISREDI, MERR )
C  ---------------------------------------------------------------------
C
C  Description:
C    This is the Merlin command dispatcher. According to the input
C    argument ICOM, it calls a subroutine that implements the user
C    command COM.
C
C  Input arguments:
C    ICOM       Index of the command to be executed.
C    VERSIM     The two-dimensional Merlin work array.
C    MAXW       Number of storage locations available in array VERSIM.
C    COM        The command entered by the user.
C
C  Output arguments:
C    MERR       Error indicator. Every routine that implements a Merlin
C               command is required to return an error indicator:
C               Any value may be used; the following have special
C               meaning for the Merlin error handler routine ERHAND.
C               MERR =  0 -> No error.
C               MERR = -1 -> EOF in the Merlin input file.
C               MERR = -2 -> Some other error.
C
C  Input / Output arguments:
C    ISREDI     Logical variable. On input, ISREDI = .TRUE. indicates
C               that output redirection (> or >>) has been requested
C               by the user. When subroutine EXECOM returns, Merlin
C               uses the setting of ISREDI to undo the effect of
C               redirection. Some commands (RUNMCL) that handle
C               redirection directly, set ISREDI to .FALSE., thus
C               prohibiting Merlin to undo redirection.
C
C  ---------------------------------------------------------------------
      IMPLICIT $(TYPE) (A-H,O-Z)
C  Arguments:
      DIMENSION VERSIM(*)
      CHARACTER*(*) COM
      LOGICAL ISREDI
C  ---------------------------------------------------------------------
C  Common blocks:
      PARAMETER ( NMCOM = 98, MXPLUG = $(MXPLUG), NCOM = NMCOM+MXPLUG )
      INTEGER IND(NCOM)
      COMMON /COM2/ IND
      PARAMETER ( MXV = $(MXV) )
      COMMON /PARAMS/ POINT(MXV), VAL
      COMMON /MODE3/ MODEV, IFFORM, INMCL
      LOGICAL STRO, WEAK
      COMMON /PRILOG/ STRO, WEAK
      CHARACTER*10 DERIVA, PRINTO, PROCES, JACOMO
      COMMON /MODEIN/ DERIVA, PRINTO, PROCES, JACOMO
      COMMON /CNTRL/ IDISK, MACECH, LIMCLK, IECOMA, IPROEC, MCLUBU
      CHARACTER*10 CURCOM
      COMMON /CURC/ CURCOM
C  ---------------------------------------------------------------------
C  Local variables:
      CHARACTER*10 TMP
      LOGICAL LO
      LO(I1,I2) = I.GE.I1 .AND. I.LE.I2
C  ---------------------------------------------------------------------
C
C  Save the current command.
      CURCOM = COM
C
C  Default value for the error code.
      MERR = 0
C
C  Dispatch.
      I = IND(ICOM)
      IF (LO(1,10))  GOTO (61,62,35,77,44,74,9,58,76,24),I
      IF (LO(11,20)) GOTO (75,26,14,64,31,36,13,49,46,50),I-10
      IF (LO(21,30)) GOTO (37,54,89,43,41,60,6,56,32,51),I-20
      IF (LO(31,40)) GOTO (27,15,57,16,59,22,45,42,33,38),I-30
      IF (LO(41,50)) GOTO (34,47,23,12,20,55,73,91,5,90),I-40
      IF (LO(51,60)) GOTO (88,72,63,48,71,18,29,4,65,66),I-50
      IF (LO(61,70)) GOTO (67,68,82,83,84,85,86,81,69,70),I-60
      IF (LO(71,80)) GOTO (78,79,80,87,39,40,190,191,192,193),I-70
      IF (LO(81,90)) GOTO (194,195,196,197,198,199,200,301,302,303),I-80
      IF (LO(91,NMCOM)) GOTO (304,305,306,307,308,309,310,311),I-90
      IF (LO(NMCOM+1,NCOM)) GOTO 700
      CALL I2STR(I,TMP,LET)
      CALL MABORT('Index '//TMP(1:LET)//' not valid.','EXECOM')
C  .........................................................  STOP
4     CONTINUE
      CALL RETSTP(2,MERR)
      GOTO 1
C  .........................................................  RETURN
5     CONTINUE
      CALL RETSTP(1,MERR)
      GOTO 1
C  .........................................................  INSPECT
6     CONTINUE
      CALL INSPE(MERR)
      GOTO 1
C  .........................................................  PDUMP
9     CONTINUE
      CALL PDUMP(MERR)
      GOTO 1
C  .........................................................  PICK
12    CONTINUE
      CALL PICK(MERR)
      GOTO 1
C  .........................................................  GODFATHER
13    CONTINUE
      CALL GODFA(MERR)
      GOTO 1
C  .........................................................  FIX
14    CONTINUE
      ICFIX = 0
      CALL FIXLOO(ICFIX,MERR)
      GOTO 1
C  .........................................................  LOOSE
15    CONTINUE
      ICFIX = 1
      CALL FIXLOO(ICFIX,MERR)
      GOTO 1
C  .........................................................  PSGRAPH
16    CONTINUE
      CALL PSGRAF(MERR)
      GOTO 1
C  .........................................................  STEP
18    CONTINUE
      CALL CHSTEP(MERR)
      GOTO 1
C  .........................................................  POINT
20    CONTINUE
      CALL POICOM(MERR)
      GOTO 1
C  .........................................................  MEMO
22    CONTINUE
      CALL MEMO(MERR)
      GOTO 1
C  .........................................................  PANELON
23    CONTINUE
      CALL PONOFF(1,MERR)
      GOTO 1
C  .........................................................  BFGS
24    CONTINUE
      CALL BFGS(VERSIM,MAXW,MERR)
      GOTO 1
C  .........................................................  DISCARD
26    CONTINUE
      CALL DISCA(MERR)
      GOTO 1
C  .........................................................  LOOSAL
27    CONTINUE
      CALL LOOSAL
      GOTO 1
C  .........................................................  STEPALL
29    CONTINUE
      CALL STEPAL
      GOTO 1
C  .........................................................  FULLBACK
31    CONTINUE
      CALL BACK('FULL',MERR)
      GOTO 1
C  .........................................................  LASTBACK
32    CONTINUE
      CALL BACK('LAST',MERR)
      GOTO 1
C  .........................................................  NOBACK
33    CONTINUE
      CALL BACK('NO',MERR)
      GOTO 1
C  .........................................................  NUMER
34    CONTINUE
      CALL SETDER('NUMER')
      GOTO 1
C  .........................................................  ANAL
35    CONTINUE
      CALL SETDER('ANAL')
      GOTO 1
C  .........................................................  FULLPRINT
36    CONTINUE
      PRINTO = 'FULLPRINT'
      WEAK = .TRUE.
      STRO = .TRUE.
      GOTO 1
C  .........................................................  HALFPRINT
37    CONTINUE
      PRINTO = 'HALFPRINT'
      STRO = .FALSE.
      WEAK = .TRUE.
      GOTO 1
C  .........................................................  NOPRINT
38    CONTINUE
      PRINTO = 'NOPRINT'
      WEAK = .FALSE.
      STRO = .FALSE.
      GOTO 1
C  .........................................................  ALIASDIS
39    CONTINUE
      CALL ALIDIS
      GOTO 1
C  .........................................................  UNALIAS
40    CONTINUE
      CALL UNALI(MERR)
      GOTO 1
C  .........................................................  DELETE
41    CONTINUE
      CALL DELETE(MERR)
      GOTO 1
C  .........................................................  RMARGIN
42    CONTINUE
      CALL MARGI(1,MERR)
      GOTO 1
C  .........................................................  IAF
43    CONTINUE
      PROCES = 'IAF'
      GOTO 1
C  .........................................................  BATCH
44    CONTINUE
      PROCES = 'BATCH'
      GOTO 1
C  .........................................................  MODEDIS
45    CONTINUE
      CALL MODIS
      GOTO 1
C  .........................................................  GRADDIS
46    CONTINUE
      CALL GRDIS(MERR)
      GOTO 1
C  .........................................................  PANELOFF
47    CONTINUE
      CALL PONOFF(0,MERR)
      GOTO 1
C  .........................................................  SHORTDIS
48    CONTINUE
      CALL SHDIS(MERR)
      GOTO 1
C  .........................................................  GRADCHECK
49    CONTINUE
      CALL GRADC(MERR)
      GOTO 1
C  .........................................................  GRAPH
50    CONTINUE
      CALL GRAPH(MERR)
      GOTO 1
C  .........................................................  LIST
51    CONTINUE
      CALL LIST
      GOTO 1
C  .........................................................  HELP
54    CONTINUE
      CALL HELP
      GOTO 1
C  .........................................................  QUIT
55    CONTINUE
      CALL QUIT(MERR)
      GOTO 1
C  .........................................................  EPILOG
56    CONTINUE
      CALL EPILOG(MERR)
      GOTO 1
C  .........................................................  MACRO
57    CONTINUE
      CALL MACRO(MERR)
      GOTO 1
C  .........................................................  CLEAR
58    CONTINUE
      CALL CLEAR(MERR)
      GOTO 1
C  .........................................................  LMARGIN
59    CONTINUE
      CALL MARGI(-1,MERR)
      GOTO 1
C  .........................................................  INIT
60    CONTINUE
      CALL INITCO(MERR)
      GOTO 1
C  .........................................................  ACCUM
61    CONTINUE
      CALL ACCUM(MERR)
      GOTO 1
C  .........................................................  ADJUST
62    CONTINUE
      CALL PRADJ
      GOTO 1
C  .........................................................  RUNMCL
63    CONTINUE
      CALL RUNMCL(ISREDI,MERR)
      GOTO 1
C  .........................................................  FIXALL
64    CONTINUE
      CALL FIXALL
      GOTO 1
C  .........................................................  VALDIS
65    CONTINUE
      CALL VALDIS
      GOTO 1
C  .........................................................  NOEVAL
66    CONTINUE
      MODEV = 0
      VAL = 0.
      GOTO 1
C  .........................................................  FAST
67    CONTINUE
      CALL SETDER('FAST')
      GOTO 1
C  .........................................................  QUAD
68    CONTINUE
      CALL SETDER('QUAD')
      GOTO 1
C  .........................................................  SIMPLEX
69    CONTINUE
      CALL SIMPLX(VERSIM,MAXW,MERR)
      GOTO 1
C  .........................................................  CONTROL
70    CONTINUE
      CALL CONTRL(MERR)
      GOTO 1
C  .........................................................  CONFIDENCE
71    CONTINUE
      CALL CONFID(VERSIM,MAXW,MERR)
      GOTO 1
C  .........................................................  LDEMARGIN
72    CONTINUE
      CALL DEMAR(-1,MERR)
      GOTO 1
C  .........................................................  HESSIAN
73    CONTINUE
      CALL HESCOM(VERSIM,MAXW,MERR)
      GOTO 1
C  .........................................................  TRUST
74    CONTINUE
      CALL TRUST(VERSIM,MAXW,MERR)
      GOTO 1
C  .........................................................  DFP
75    CONTINUE
      CALL DFP(VERSIM,MAXW,MERR)
      GOTO 1
C  .........................................................  CONGRA
76    CONTINUE
      CALL CONGRA(MERR)
      GOTO 1
C  .........................................................  AUTO
77    CONTINUE
      CALL AUTO(VERSIM,MAXW,MERR)
      GOTO 1
C  .........................................................  TOLMIN
78    CONTINUE
      CALL TOLMIN(VERSIM,MAXW,MERR)
      GOTO 1
C  .........................................................  ECHO
79    CONTINUE
      CALL ECHO
      GOTO 1
C  .........................................................  STEPDIS
80    CONTINUE
      CALL STPDIS(MERR)
      GOTO 1
C  .........................................................  NOTARGET
81    CONTINUE
      CALL NOTARG
      GOTO 1
C  .........................................................  MIXED
82    CONTINUE
      CALL MIXED(MERR)
      GOTO 1
C  .........................................................  EVALUATE
83    CONTINUE
      CALL EVAL
      GOTO 1
C  .........................................................  COVARIANCE
84    CONTINUE
      CALL COVAR(VERSIM,MAXW,MERR)
      GOTO 1
C  .........................................................  MAD
85    CONTINUE
      CALL MAD(MERR)
      GOTO 1
C  .........................................................  TARGET
86    CONTINUE
      CALL TARG(MERR)
      GOTO 1
C  .........................................................  ALIAS
87    CONTINUE
      CALL NEWALI(MERR)
      GOTO 1
C  .........................................................  REWIND
88    CONTINUE
      CALL REWEOF(1,MERR)
      GOTO 1
C  .........................................................  HIDEOUT
89    CONTINUE
      CALL HIDE(MERR)
      GOTO 1
C  .........................................................  REVEAL
90    CONTINUE
      CALL REVEAL(MERR)
      GOTO 1
C  .........................................................  RESET
91    CONTINUE
      CALL RESET
      GOTO 1
C  .........................................................  GOEOF
190   CONTINUE
      CALL REWEOF(2,MERR)
      GOTO 1
C  .........................................................  TITLE
191   CONTINUE
      CALL TITLE(MERR)
      GOTO 1
C  .........................................................  LIMITS
192   CONTINUE
      CALL LIMITS
      GOTO 1
C  .........................................................  TERMDIS
193   CONTINUE
      CALL TRMDIS(MERR)
      GOTO 1
C  .........................................................  GNORM
194   CONTINUE
      CALL GNORM
      GOTO 1
C  .........................................................  FLAG
195   CONTINUE
      CALL CHFLAG(MERR)
      GOTO 1
C  .........................................................  FLAGDIS
196   CONTINUE
      CALL FLADIS(MERR)
      GOTO 1
C  .........................................................  CFLAG
197   CONTINUE
      CALL CHCFLA(MERR)
      GOTO 1
C  .........................................................  CFLAGDIS
198   CONTINUE
      CALL CFLADI(MERR)
      GOTO 1
C  .........................................................  HISTORY
199   CONTINUE
      CALL HISTOR(0,MERR)
      GOTO 1
C  .........................................................  PSTATUS
200   CONTINUE
      CALL PSTAT(MERR)
      GOTO 1
C  .........................................................  RDEMARGIN
301   CONTINUE
      CALL DEMAR(1,MERR)
      GOTO 1
C  .........................................................  LEVE
302   CONTINUE
      CALL LEVE(VERSIM,MAXW,MERR)
      GOTO 1
C  .........................................................  GENERAL
303   CONTINUE
      CALL SETFFO(0,MERR)
      GOTO 1
C  .........................................................  SOS
304   CONTINUE
      CALL SETFFO(1,MERR)
      GOTO 1
C  .........................................................  ROLL
305   CONTINUE
      CALL ROLL(MERR)
      GOTO 1
C  .........................................................  BACKUP
306   CONTINUE
      CALL BACK(' ',MERR)
      GOTO 1
C  .........................................................  NONAME
307   CONTINUE
      CALL NONAME(MERR)
      GOTO 1
C  .........................................................  DUMP
308   CONTINUE
      CALL DUMP(MERR)
      GOTO 1
C  .........................................................  JANAL
309   CONTINUE
      JACOMO = 'JANAL'
      GOTO 1
C  .........................................................  JNUMER
310   CONTINUE
      JACOMO = 'JNUMER'
      GOTO 1
C  .........................................................  JCOMPARE
311   CONTINUE
      CALL JCOMP(VERSIM,MAXW,MERR)
      GOTO 1
C  .........................................................  PLUG-INs
700   CONTINUE
      IPLG = I-NMCOM
      CALL EXEPLG(IPLG,MERR)
      GOTO 1
C  ---------------------------------------------------------------------
C  Control is transfered here upon termination of each command.
C  ---------------------------------------------------------------------
1     CONTINUE
C
C  If this is a panel command, print keywords that can be returned
C  to an MCL program.
      IF (LIMCLK.NE.0) THEN
         CALL PLOCP(COM,IPA)
         IF (IPA.NE.0) CALL PPRO
      END IF
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE EXEPLG ( IPL, MERR )
C  ---------------------------------------------------------------------
C
C  Description:
C    Calls the appropriate plug-in routine, according to the input
C    argument IPL.
C
C  Input arguments:
C    IPL        Index of the plug-in to call.
C
C  Output arguments:
C    MERR       Error indicator, set by the plug-in. If a plug-in
C               does not set the error indicator, MERR = 0 is returned.
C
C  ---------------------------------------------------------------------
C  Common blocks:
      COMMON /PLGERR/ IPLGER
C  ---------------------------------------------------------------------
C  Local variables:
      CHARACTER*10 STR
      LOGICAL LO
      LO(I1,I2) = IPL.GE.I1 .AND. IPL.LE.I2
C  ---------------------------------------------------------------------
C
      IPLGER = 0
      IF (LO(1,10)) GOTO (1,2,3,4,5,6,7,8,9,10), IPL
      IF (LO(11,20)) GOTO (11,12,13,14,15,16,17,18,19,20), IPL-10
      IF (LO(21,30)) GOTO (21,22,23,24,25,26,27,28,29,30), IPL-20
      IF (LO(31,40)) GOTO (31,32,33,34,35,36,37,38,39,40), IPL-30
      IF (LO(41,50)) GOTO (41,42,43,44,45,46,47,48,49,50), IPL-40
      CALL I2STR(IPL,STR,LES)
      CALL MABORT('No such plug-in: '//STR(1:LES),'EXEPLG')
C
C  .........................................................  Plug1
1     CONTINUE
      CALL PLUG1
      GOTO 1000
C  .........................................................  Plug2
2     CONTINUE
      CALL PLUG2
      GOTO 1000
C  .........................................................  Plug3
3     CONTINUE
      CALL PLUG3
      GOTO 1000
C  .........................................................  Plug4
4     CONTINUE
      CALL PLUG4
      GOTO 1000
C  .........................................................  Plug5
5     CONTINUE
      CALL PLUG5
      GOTO 1000
C  .........................................................  Plug6
6     CONTINUE
      CALL PLUG6
      GOTO 1000
C  .........................................................  Plug7
7     CONTINUE
      CALL PLUG7
      GOTO 1000
C  .........................................................  Plug8
8     CONTINUE
      CALL PLUG8
      GOTO 1000
C  .........................................................  Plug9
9     CONTINUE
      CALL PLUG9
      GOTO 1000
C  .........................................................  Plug10
10    CONTINUE
      CALL PLUG10
      GOTO 1000
C  .........................................................  Plug11
11    CONTINUE
      CALL PLUG11
      GOTO 1000
C  .........................................................  Plug12
12    CONTINUE
      CALL PLUG12
      GOTO 1000
C  .........................................................  Plug13
13    CONTINUE
      CALL PLUG13
      GOTO 1000
C  .........................................................  Plug14
14    CONTINUE
      CALL PLUG14
      GOTO 1000
C  .........................................................  Plug15
15    CONTINUE
      CALL PLUG15
      GOTO 1000
C  .........................................................  Plug16
16    CONTINUE
      CALL PLUG16
      GOTO 1000
C  .........................................................  Plug17
17    CONTINUE
      CALL PLUG17
      GOTO 1000
C  .........................................................  Plug18
18    CONTINUE
      CALL PLUG18
      GOTO 1000
C  .........................................................  Plug19
19    CONTINUE
      CALL PLUG19
      GOTO 1000
C  .........................................................  Plug20
20    CONTINUE
      CALL PLUG20
      GOTO 1000
C  .........................................................  Plug21
21    CONTINUE
      CALL PLUG21
      GOTO 1000
C  .........................................................  Plug22
22    CONTINUE
      CALL PLUG22
      GOTO 1000
C  .........................................................  Plug23
23    CONTINUE
      CALL PLUG23
      GOTO 1000
C  .........................................................  Plug24
24    CONTINUE
      CALL PLUG24
      GOTO 1000
C  .........................................................  Plug25
25    CONTINUE
      CALL PLUG25
      GOTO 1000
C  .........................................................  Plug26
26    CONTINUE
      CALL PLUG26
      GOTO 1000
C  .........................................................  Plug27
27    CONTINUE
      CALL PLUG27
      GOTO 1000
C  .........................................................  Plug28
28    CONTINUE
      CALL PLUG28
      GOTO 1000
C  .........................................................  Plug29
29    CONTINUE
      CALL PLUG29
      GOTO 1000
C  .........................................................  Plug30
30    CONTINUE
      CALL PLUG30
      GOTO 1000
C  .........................................................  Plug31
31    CONTINUE
      CALL PLUG31
      GOTO 1000
C  .........................................................  Plug32
32    CONTINUE
      CALL PLUG32
      GOTO 1000
C  .........................................................  Plug33
33    CONTINUE
      CALL PLUG33
      GOTO 1000
C  .........................................................  Plug34
34    CONTINUE
      CALL PLUG34
      GOTO 1000
C  .........................................................  Plug35
35    CONTINUE
      CALL PLUG35
      GOTO 1000
C  .........................................................  Plug36
36    CONTINUE
      CALL PLUG36
      GOTO 1000
C  .........................................................  Plug37
37    CONTINUE
      CALL PLUG37
      GOTO 1000
C  .........................................................  Plug38
38    CONTINUE
      CALL PLUG38
      GOTO 1000
C  .........................................................  Plug39
39    CONTINUE
      CALL PLUG39
      GOTO 1000
C  .........................................................  Plug40
40    CONTINUE
      CALL PLUG40
      GOTO 1000
C  .........................................................  Plug41
41    CONTINUE
      CALL PLUG41
      GOTO 1000
C  .........................................................  Plug42
42    CONTINUE
      CALL PLUG42
      GOTO 1000
C  .........................................................  Plug43
43    CONTINUE
      CALL PLUG43
      GOTO 1000
C  .........................................................  Plug44
44    CONTINUE
      CALL PLUG44
      GOTO 1000
C  .........................................................  Plug45
45    CONTINUE
      CALL PLUG45
      GOTO 1000
C  .........................................................  Plug46
46    CONTINUE
      CALL PLUG46
      GOTO 1000
C  .........................................................  Plug47
47    CONTINUE
      CALL PLUG47
      GOTO 1000
C  .........................................................  Plug48
48    CONTINUE
      CALL PLUG48
      GOTO 1000
C  .........................................................  Plug49
49    CONTINUE
      CALL PLUG49
      GOTO 1000
C  .........................................................  Plug50
50    CONTINUE
      CALL PLUG50
      GOTO 1000
C  ---------------------------------------------------------------------
C
1000  CONTINUE
      MERR = IPLGER
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE FACT ( N, F, IERR )
C  ---------------------------------------------------------------------
C
C  Description:
C    Subroutine FACT returns the factorial of N. The factorials
C    (up to a maximum of MAXFCT) are calculated during the first entry
C    of the routine, thus speeding up subsequent calls.
C
C  Input arguments:
C    N          An integer, whose factorial we seek. Must be a
C               non-negative number.
C
C  Output arguments:
C    F          The factorial of N. F is set to -1 if any error
C               occurs.
C    IERR       Error indicator.
C
C  ---------------------------------------------------------------------
      IMPLICIT $(TYPE) (A-H,O-Z)
C  Common blocks:
      COMMON /BIGCOM/ BIGNUM, BLIM, DWARF, DWALOG
      LOGICAL STRO, WEAK
      COMMON /PRILOG/ STRO, WEAK
      COMMON /UNITS/ IUINP, IUOUT
C  ---------------------------------------------------------------------
C  Local variables:
      SAVE NFACT, FIRST, FCT, ATLIM
      PARAMETER ( MAXFCT = 200 )
      DIMENSION FCT(0:MAXFCT)
      CHARACTER*10 STR, STR2
      LOGICAL FIRST, ATLIM
      DATA FIRST / .TRUE. /
C  ---------------------------------------------------------------------
C
C  Calculate factorials on the first entry.
      IF (FIRST) THEN
         FIRST = .FALSE.
         FCT(0) = 1
         SUMLOG = 0
         BIGLOG = LOG(BIGNUM)
         DO 10,I=1,MAXFCT
            X = I
            SUMLOG = SUMLOG+LOG(X)
            IF (SUMLOG.GE.BIGLOG) THEN
               NFACT = I-1
               ATLIM = .TRUE.
               GOTO 20
            ELSE
               FCT(I) = I*FCT(I-1)
            END IF
10       CONTINUE
         NFACT = MAXFCT
         ATLIM = .FALSE.
20       CONTINUE
      END IF
C
C  Return the requested factorial.
      IF (N.LT.0) THEN
         IF (WEAK) THEN
            CALL I2STR(N,STR,LES)
            WRITE (IUOUT,40) STR(1:LES)
         END IF
         F = -1
         IERR = -2
      ELSE IF (N.LE.NFACT) THEN
         F = FCT(N)
         IERR = 0
      ELSE IF (ATLIM) THEN
         IF (WEAK) THEN
            CALL I2STR(N,STR,LES)
            CALL R2STR(BIGNUM,STR2,LES2)
            WRITE (IUOUT,50) STR(1:LES), STR2(1:LES2)
         END IF
         F = -1
         IERR = -2
      ELSE
         F = FCT(NFACT)
         DO 30,I=NFACT+1,N
30          F = F*I
         IERR = 0
      END IF
C
40    FORMAT (/$(FC)'Error: You have requested the factorial of a ',
     &        'negative number: ',A,'.')
50    FORMAT (/$(FC)'Error: You requested the factorial of ',A,
     &        ', that will result',
     &        /$(FC)'       in a very large real. (Larger than ',A,
     &        ').')
      END
C  ---------------------------------------------------------------------
      SUBROUTINE FIXSTR ( IND, STR )
C  ---------------------------------------------------------------------
C
C  Description:
C    Encodes the fix status of variable IND in a string that may be
C    used in printouts such as SHORTDIS, GRADDIS etc.
C
C  Input arguments:
C    IND        Index of the minimization variable.
C
C  Output arguments:
C    STR        String containing the encoded fix status:
C               STR = '.. * ..'  for a fixed variable.
C               STR = '.......'  for a free variable.
C
C  ---------------------------------------------------------------------
C  Arguments:
      CHARACTER*(*) STR
C  ---------------------------------------------------------------------
C  Common blocks:
      PARAMETER ( MXV = $(MXV) )
      COMMON /DETAIL/ IXAT(MXV), MARG(MXV)
C  ---------------------------------------------------------------------
C
      STR = '.......'
      IF (IXAT(IND).EQ.0) STR(3:5) = ' * '
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE FREEVA ( NFREE )
C  ---------------------------------------------------------------------
C
C  Description:
C    Determines the number of free (non-fixed) variables.
C
C  Output arguments:
C    NFREE      Number of free variables.
C
C  ---------------------------------------------------------------------
C  Common blocks:
      PARAMETER ( MXV = $(MXV) )
      COMMON /DETAIL/ IXAT(MXV), MARG(MXV)
      COMMON /DIMEN/ NOD, NOF
C  ---------------------------------------------------------------------
C
      NFREE = 0
      DO 10,I=1,NOD
10       IF (IXAT(I).NE.0) NFREE = NFREE+1
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE GETDIM ( NOVAR, NOTERM )
C  ---------------------------------------------------------------------
C
C  Description:
C    Glue routine. Returns the dimensionality of the objective function
C    and the number of terms if the function is a sum-of-squares.
C
C  Output arguments:
C    NOVAR      Dimensionality of the objective function.
C    NOTERM     Number of terms if the function is a sum-of-squares;
C               0 otherwise.
C
C  ---------------------------------------------------------------------
      IMPLICIT $(TYPE) (A-H,O-Z)
C  Common blocks:
      COMMON /DIMEN/ NOD, NOF
C  ---------------------------------------------------------------------
C
      NOVAR = NOD
      NOTERM = NOF
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE GETX ( X )
C  ---------------------------------------------------------------------
C
C  Description:
C    Glue routine. Returns the current point.
C
C  Output arguments:
C    X          The current point.
C
C  ---------------------------------------------------------------------
      IMPLICIT $(TYPE) (A-H,O-Z)
C  Arguments:
      DIMENSION X(*)
C  ---------------------------------------------------------------------
C  Common blocks:
      PARAMETER ( MXV = $(MXV) )
      COMMON /PARAMS/ POINT(MXV), VAL
      COMMON /DIMEN/ NOD, NOF
C  ---------------------------------------------------------------------
C
      DO 10,I=1,NOD
10       X(I) = POINT(I)
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE SETX ( X )
C  ---------------------------------------------------------------------
C
C  Description:
C    Glue routine. Sets the current point.
C
C  Input arguments:
C    X          The point to set.
C
C  ---------------------------------------------------------------------
      IMPLICIT $(TYPE) (A-H,O-Z)
C  Arguments:
      DIMENSION X(*)
C  ---------------------------------------------------------------------
C  Common blocks:
      PARAMETER ( MXV = $(MXV) )
      COMMON /PARAMS/ POINT(MXV), VAL
      COMMON /DIMEN/ NOD, NOF
      COMMON /INDEFI/ INDEF(MXV)
C  ---------------------------------------------------------------------
C
      DO 10,I=1,NOD
         POINT(I) = X(I)
         INDEF(I) = 1
10    CONTINUE
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE ARGNO ( NARGS )
C  ---------------------------------------------------------------------
C
C  Description:
C    Glue routine. Returns the number of arguments in the Merlin command
C    line.
C
C  Output arguments:
C    NARGS      Number of arguments in the Merlin command line.
C
C  ---------------------------------------------------------------------
      IMPLICIT $(TYPE) (A-H,O-Z)
C  Common blocks:
      PARAMETER ( LINLEN = 80 )
      INTEGER IS(LINLEN/2), IE(LINLEN/2)
      COMMON /INDAT1/ IS, IE, NIT, LEC
C  ---------------------------------------------------------------------
C
      NARGS = NIT-1
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE CMDARG ( N, ARG, LENARG )
C  ---------------------------------------------------------------------
C
C  Description:
C    Glue routine. Returns the N-th argument from the Merlin command
C    line.
C
C  Input arguments:
C    N          Index of the argument being requested. N=0 corresponds
C               to the command name itself.
C
C  Output arguments:
C    ARG        The N-th argument. If the effective length of the N-th
C               is greater than LEN(ARG), it will truncated, so as to
C               fit in ARG.
C    LENARG     Effective length of ARG.
C
C  ---------------------------------------------------------------------
C  Arguments:
      CHARACTER*(*) ARG
C  ---------------------------------------------------------------------
C  Common blocks:
      PARAMETER ( LINLEN = 80 )
      INTEGER IS(LINLEN/2), IE(LINLEN/2)
      COMMON /INDAT1/ IS, IE, NIT, LEC
      CHARACTER*(LINLEN) INPLIN, QINPL
      COMMON /INDAT2/ INPLIN, QINPL
C  ---------------------------------------------------------------------
C
      IF (N.LT.0) THEN
         CALL MABORT('Incorrect argument number.','CMDARG')
      ELSE IF (N.GT.NIT-1) THEN
         ARG = ' '
         LENARG = 0
      ELSE
         LESTO = LEN(ARG)
         N1 = N+1
         LEACT = IE(N1)-IS(N1)+1
         LENARG = MIN(LESTO,LEACT)
         IF (LENARG.NE.0) ARG(1:LENARG) = QINPL(IS(N1):IE(N1))
         IF (LENARG.LT.LESTO) ARG(LENARG+1:) = ' '
      END IF
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE GETFIX ( IX )
C  ---------------------------------------------------------------------
C
C  Description:
C    Glue routine. Returns the fix-status array.
C
C  Output arguments:
C    IX         The fix status array.
C
C  ---------------------------------------------------------------------
C  Arguments:
      DIMENSION IX(*)
C  ---------------------------------------------------------------------
C  Common blocks:
      PARAMETER ( MXV = $(MXV) )
      COMMON /DIMEN/ NOD, NOF
      COMMON /DETAIL/ IXAT(MXV), MARG(MXV)
C  ---------------------------------------------------------------------
C
      DO 10,I=1,NOD
10       IX(I) = IXAT(I)
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE GETMAR ( MF, XL, XU )
C  ---------------------------------------------------------------------
C
C  Description:
C    Glue routine. Returns the margins.
C
C  Output arguments:
C    MF         Array containing the margin statuses.
C    XL         Array containing the left margins.
C    XU         Array containing the right margins.
C
C  ---------------------------------------------------------------------
      IMPLICIT $(TYPE) (A-H,O-Z)
C  Arguments:
      DIMENSION XL(*), XU(*), MF(*)
C  ---------------------------------------------------------------------
C  Common blocks:
      PARAMETER ( MXV = $(MXV) )
      COMMON /XLIMS/ XLL(MXV), XRL(MXV)
      COMMON /DIMEN/ NOD, NOF
      COMMON /DETAIL/ IXAT(MXV), MARG(MXV)
C  ---------------------------------------------------------------------
C
      DO 10,I=1,NOD
         XL(I) = XLL(I)
         XU(I) = XRL(I)
         MF(I) = MARG(I)
10    CONTINUE
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE GETACC ( ACC )
C  ---------------------------------------------------------------------
C
C  Description:
C    Glue routine. Returns the machine accuracy.
C
C  Output arguments:
C    ACC        Machine accuracy, either as estimated by the Merlin
C               routine MACAC, or as defined in the Merlin configuration
C               file.
C
C  ---------------------------------------------------------------------
      IMPLICIT $(TYPE) (A-H,O-Z)
C  Common blocks:
      COMMON /ACCURA/ COMACC
C  ---------------------------------------------------------------------
C
      ACC = COMACC
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE GETCNT ( NF, NFP, NG, NGP, NH, NHP, NJ, NJP )
C  ---------------------------------------------------------------------
C
C  Description:
C    Glue routine. Returns the total and partial Merlin counters.
C
C  Output arguments:
C    NF         Total number of function evaluations.
C    NFP        Number of function evaluations since the last RESET.
C    NG         Total number of gradient evaluations (calls to the
C               user supplied routine GRANAL).
C    NGP        Number of gradient evaluations since the last RESET.
C    NH         Total number of Hessian evaluations (calls to the
C               user supplied routine HANAL).
C    NHP        Number of Hessian evaluations since the last RESET.
C    NJ         Total number of Jacobian evaluations (calls to the
C               user supplied routine JANAL).
C    NJP        Number of Jacobian evaluations since the last RESET.
C
C  ---------------------------------------------------------------------
C  Common blocks:
      COMMON /TOTCAL/ ITNOCS, IPARTC, ITDER, IPDER, ITHESS, IPHESS,
     &                ITJAC, IPJAC
C  ---------------------------------------------------------------------
C
      NF = ITNOCS
      NFP = IPARTC
      NG = ITDER
      NGP = IPDER
      NH = ITHESS
      NHP = IPHESS
      NJ = ITJAC
      NJP = IPJAC
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE GETFLA ( I, F )
C  ---------------------------------------------------------------------
C
C  Description:
C    Glue routine. Returns the I-th flag.
C
C  Input arguments:
C    I          Index of the requested flag.
C
C  Output arguments:
C    F          The I-th flag.
C
C  ---------------------------------------------------------------------
      IMPLICIT $(TYPE) (A-H,O-Z)
C  Common blocks:
      PARAMETER ( MAXFLA = 20 )
      COMMON /FLAGS/ FLAG(MAXFLA)
C  ---------------------------------------------------------------------
C  Local variables:
      CHARACTER*10 STR
C  ---------------------------------------------------------------------
C
      IF (I.LT.0 .OR. I.GT.MAXFLA) THEN
         CALL I2STR(I,STR,LES)
         CALL MABORT('Incorrect flag number '//STR(1:LES),'GETFLA')
      ELSE
         F = FLAG(I)
      END IF
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE SETFLA ( I, F )
C  ---------------------------------------------------------------------
C
C  Description:
C    Glue routine. Sets the I-th flag.
C
C  Input arguments:
C    I          Index of the requested flag.
C    F          Value of the I-th flag.
C
C  ---------------------------------------------------------------------
      IMPLICIT $(TYPE) (A-H,O-Z)
C  Common blocks:
      PARAMETER ( MAXFLA = 20 )
      COMMON /FLAGS/ FLAG(MAXFLA)
C  ---------------------------------------------------------------------
C  Local variables:
      CHARACTER*10 STR
C  ---------------------------------------------------------------------
C
      IF (I.LT.0 .OR. I.GT.MAXFLA) THEN
         CALL I2STR(I,STR,LES)
         CALL MABORT('Incorrect flag number '//STR(1:LES),'SETFLA')
      ELSE
         FLAG(I) = F
      END IF
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE GETMC ( BIG, SMALL )
C  ---------------------------------------------------------------------
C
C  Description:
C    Glue routine. Returns some machine constants.
C
C  Output arguments:
C    BIG        Largest real this machine can handle.
C    SMALL      Smaller non-zero real this machine can handle.
C
C  ---------------------------------------------------------------------
      IMPLICIT $(TYPE) (A-H,O-Z)
C  Common blocks:
      COMMON /BIGCOM/ BIGNUM, BLIM, DWARF, DWALOG
C  ---------------------------------------------------------------------
C
      BIG = BIGNUM
      SMALL = DWARF
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE SETCOD ( NERR )
C  ---------------------------------------------------------------------
C
C  Description:
C    Sets the error code to be returned to Merlin, upon termination
C    of a plug-in module.
C    Note that the error code, should follow the Merlin conventions.
C    (See subroutine EXECOM.)
C
C  Input arguments:
C    NERR       The error code to be returned.
C
C  ---------------------------------------------------------------------
C  Common blocks:
      COMMON /PLGERR/ IPLGER
C  ---------------------------------------------------------------------
C
      IPLGER = NERR
C
      END
C  ---------------------------------------------------------------------
      FUNCTION ISMCL ( )
C  ---------------------------------------------------------------------
C
C  Description:
C    Glue routine. Returns 1 if we are running MCL; 0 otherwise.
C
C  ---------------------------------------------------------------------
      IMPLICIT $(TYPE) (A-H,O-Z)
C  Common blocks:
      COMMON /MODE3/ MODEV, IFFORM, INMCL
C  ---------------------------------------------------------------------
C
      ISMCL = INMCL
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE GETIOU ( NIN, NOUT )
C  ---------------------------------------------------------------------
C
C  Description:
C    Glue routine. Returns the Merlin input/output units.
C
C  Output arguments:
C    NIN        The Merlin input unit.
C    NOUT       The Merlin output unit.
C
C  ---------------------------------------------------------------------
C  Common blocks:
      COMMON /UNITS/ IUINP, IUOUT
C  ---------------------------------------------------------------------
C
      NIN = IUINP
      NOUT = IUOUT
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE GETPRL ( ST, WE )
C  ---------------------------------------------------------------------
C
C  Description:
C    Glue routine. Returns the logical variables that control Merlin
C    printouts.
C
C  Output arguments:
C    ST        The STRO setting. Normal output is printed
C              only IF (STRO).
C    WE        The WEAK setting. Error messages are printed
C              only IF (WEAK).
C
C  ---------------------------------------------------------------------
C  Arguments:
      LOGICAL ST, WE
C  ---------------------------------------------------------------------
C  Common blocks:
      LOGICAL STRO, WEAK
      COMMON /PRILOG/ STRO, WEAK
C  ---------------------------------------------------------------------
C
      ST = STRO
      WE = WEAK
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE GETX1 ( I, X )
C  ---------------------------------------------------------------------
C
C  Description:
C    Glue routine. Returns the current value  of the I-th minimization
C    variable.
C
C  Input arguments:
C    I          Index of the variable to be returned.
C
C  Output arguments:
C    X          Value of the I-th minimization variable.
C
C  ---------------------------------------------------------------------
      IMPLICIT $(TYPE) (A-H,O-Z)
C  Common blocks:
      PARAMETER ( MXV = $(MXV) )
      COMMON /PARAMS/ POINT(MXV), VAL
C  ---------------------------------------------------------------------
C
      X = POINT(I)
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE SETX1 ( I, X )
C  ---------------------------------------------------------------------
C
C  Description:
C    Glue routine. Sets the current value  of the I-th minimization
C    variable.
C
C  Input arguments:
C    I          Index of the variable to be set.
C    X          the value to set to the I-th minimization variable.
C
C  ---------------------------------------------------------------------
      IMPLICIT $(TYPE) (A-H,O-Z)
C  Common blocks:
      PARAMETER ( MXV = $(MXV) )
      COMMON /PARAMS/ POINT(MXV), VAL
      COMMON /INDEFI/ INDEF(MXV)
C  ---------------------------------------------------------------------
C
      POINT(I) = X
      INDEF(I) = 1
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE GETFX1 ( I, IX )
C  ---------------------------------------------------------------------
C
C  Description:
C    Glue routine. Returns the fix status of the I-th minimization
C    variable.
C
C  Input arguments:
C    I          Index of the variable.
C
C  Output arguments:
C    IX         Fix status of the I-th variable.
C
C  ---------------------------------------------------------------------
C  Common blocks:
      PARAMETER ( MXV = $(MXV) )
      COMMON /DETAIL/ IXAT(MXV), MARG(MXV)
C  ---------------------------------------------------------------------
C
      IX = IXAT(I)
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE GETMR1 ( I, MF, XL, XU )
C  ---------------------------------------------------------------------
C
C  Description:
C    Glue routine. Returns margin information for the I-th minimization
C    variable.
C
C  Input arguments:
C    I          Index of the variable.
C
C  Output arguments:
C    MF         Margins status for the I-th variable.
C    XL         Left margin of the I-th variable.
C    XU         Right margin of the I-th variable.
C
C  ---------------------------------------------------------------------
      IMPLICIT $(TYPE) (A-H,O-Z)
C  Arguments:
      PARAMETER ( MXV = $(MXV) )
      COMMON /XLIMS/ XLL(MXV), XRL(MXV)
C  ---------------------------------------------------------------------
C  Common blocks:
      COMMON /DETAIL/ IXAT(MXV), MARG(MXV)
C  ---------------------------------------------------------------------
C
      XL = XLL(I)
      XU = XRL(I)
      MF = MARG(I)
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE GETCFL ( I, CFL )
C  ---------------------------------------------------------------------
C
C  Description:
C    Glue routine. Returns the value of the I-th character flag.
C
C  Input arguments:
C    I          Index of the flag.
C
C  Output arguments:
C    CFL        The I-th character flag.
C
C  ---------------------------------------------------------------------
C  Arguments:
      CHARACTER*(*) CFL
C  ---------------------------------------------------------------------
C  Common blocks:
      PARAMETER ( MAXCFL = 10, MAXCLE = 30 )
      CHARACTER*(MAXCLE) CFLAG(MAXCFL)
      COMMON /CFLAGS/ CFLAG
C  ---------------------------------------------------------------------
C  Local variables:
      CHARACTER*10 STR
C  ---------------------------------------------------------------------
C
      IF (I.LT.0 .OR. I.GT.MAXCFL) THEN
         CALL I2STR(I,STR,LES)
         CALL MABORT('Incorrect cflag number '//STR(1:LES),'GETCFL')
      ELSE
         CFL = CFLAG(I)
      END IF
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE SETCFL ( I, CFL )
C  ---------------------------------------------------------------------
C
C  Description:
C    Glue routine. Sets the value of the I-th character flag.
C
C  Input arguments:
C    I          Index of the flag.
C    CFL        The value to be set to the I-th character flag.
C
C  ---------------------------------------------------------------------
C  Arguments:
      CHARACTER*(*) CFL
C  ---------------------------------------------------------------------
C  Common blocks:
      PARAMETER ( MAXCFL = 10, MAXCLE = 30 )
      CHARACTER*(MAXCLE) CFLAG(MAXCFL)
      COMMON /CFLAGS/ CFLAG
C  ---------------------------------------------------------------------
C  Local variables:
      CHARACTER*10 STR
C  ---------------------------------------------------------------------
C
      IF (I.LT.0 .OR. I.GT.MAXCFL) THEN
         CALL I2STR(I,STR,LES)
         CALL MABORT('Incorrect cflag number '//STR(1:LES),'SETCFL')
      ELSE
         CFLAG(I) = CFL
      END IF
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE GETVAL ( VALUE )
C  ---------------------------------------------------------------------
C
C  Description:
C    Glue routine. Returns the current value.
C
C  Output arguments:
C    VALUE      The current value of the objective function.
C
C  ---------------------------------------------------------------------
      IMPLICIT $(TYPE) (A-H,O-Z)
C  Common blocks:
      PARAMETER ( MXV = $(MXV) )
      COMMON /PARAMS/ POINT(MXV), VAL
C  ---------------------------------------------------------------------
C
      VALUE = VAL
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE SETVAL ( VALUE )
C  ---------------------------------------------------------------------
C
C  Description:
C    Glue routine. Sets the current value.
C
C  Input arguments:
C    VALUE      The value to set as the current value of the
C               objective function.
C
C  ---------------------------------------------------------------------
      IMPLICIT $(TYPE) (A-H,O-Z)
C  Common blocks:
      PARAMETER ( MXV = $(MXV) )
      COMMON /PARAMS/ POINT(MXV), VAL
C  ---------------------------------------------------------------------
C
      VAL = VALUE
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE GETEVM ( IEV )
C  ---------------------------------------------------------------------
C
C  Description:
C    Glue routine. Returns the current evaluation mode.
C
C  Output arguments:
C    IEV        The current evaluation mode.
C
C  ---------------------------------------------------------------------
C  Common blocks:
      COMMON /MODE3/ MODEV, IFFORM, INMCL
C  ---------------------------------------------------------------------
C
      IEV = MODEV
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE GETFFO ( IFF )
C  ---------------------------------------------------------------------
C
C  Description:
C    Glue routine. Returns the form (GENERAL/SOS) of the objective
C    function.
C
C  Output arguments:
C    IFF        Form of the objective function.
C
C  ---------------------------------------------------------------------
C  Common blocks:
      COMMON /MODE3/ MODEV, IFFORM, INMCL
C  ---------------------------------------------------------------------
C
      IFF = IFFORM
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE GETTRG ( TRG, ISTARG )
C  ---------------------------------------------------------------------
C
C  Description:
C    Glue routine. Returns the current target value.
C
C  Output arguments:
C    TRG        The current target value.
C    ISTARG     ISTARG = 1 if target is on.
C               ISTARG = 0 if target is off.
C
C  ---------------------------------------------------------------------
      IMPLICIT $(TYPE) (A-H,O-Z)
C  Common blocks:
      LOGICAL AIMON
      COMMON /AIM/ TARGET, AIMON
      COMMON /BIGCOM/ BIGNUM, BLIM, DWARF, DWALOG
      COMMON /MODE3/ MODEV, IFFORM, INMCL
C  ---------------------------------------------------------------------
C
      IF (AIMON) THEN
         TRG = TARGET
         ISTARG = 1
      ELSE
         ISTARG = 0
         IF (IFFORM.EQ.0) THEN
            TRG = -BIGNUM
         ELSE
            TRG = 0.
         END IF
      END IF
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE SETADE ( IAUTO )
C  ---------------------------------------------------------------------
C
C  Description:
C    Enables/disables the Merlin automatic derivatives.
C
C  Input arguments:
C    IAUTO      IAUTO = 0 -> Automatic derivatives are disabled.
C               IAUTO = 1 -> Automatic derivatives are enabled.
C
C  ---------------------------------------------------------------------
C  Common blocks:
      PARAMETER ( MXV = $(MXV) )
      CHARACTER*5 GRADI
      COMMON /DERIMO/ GRADI(MXV)
      LOGICAL AUTODE
      COMMON /AUTOMA/ AUTODE
      COMMON /DIMEN/ NOD, NOF
      CHARACTER*10 DERIVA, PRINTO, PROCES, JACOMO
      COMMON /MODEIN/ DERIVA, PRINTO, PROCES, JACOMO
C  ---------------------------------------------------------------------
C
      AUTODE = IAUTO.NE.0
      IF (.NOT.AUTODE) THEN
         DERIVA = GRADI(1)
         DO 10,I=2,NOD
            IF (GRADI(I).NE.DERIVA) THEN
               DERIVA = 'MIXED'
               RETURN
            END IF
10       CONTINUE
      END IF
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE GETNAM ( NAME )
C  ---------------------------------------------------------------------
C
C  Description:
C    Glue routine. Returns the names of the minimization variables.
C
C  Output arguments:
C    NAME       Character array. NAME(I) is set to the name of the
C               I-th minimization variable.
C
C  ---------------------------------------------------------------------
      IMPLICIT $(TYPE) (A-H,O-Z)
C  Arguments:
      CHARACTER*(*) NAME(*)
C  ---------------------------------------------------------------------
C  Common blocks:
      PARAMETER ( MXV = $(MXV) )
      COMMON /DIMEN/ NOD, NOF
      CHARACTER*10 NICKNA
      COMMON /NAMES/ NICKNA(MXV)
C  ---------------------------------------------------------------------
C
      DO 10,I=1,NOD
10       NAME(I) = NICKNA(I)
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE GETNM1 ( I, NAME )
C  ---------------------------------------------------------------------
C
C  Description:
C    Glue routine. Returns the name of the I-th minimization
C    variable.
C
C  Input arguments:
C    I          Index of the variable whose name will be returned.
C
C  Output arguments:
C    NAME       The name of the I-th minimization variable.
C
C  ---------------------------------------------------------------------
      IMPLICIT $(TYPE) (A-H,O-Z)
C  Arguments:
      CHARACTER*(*) NAME
C  ---------------------------------------------------------------------
C  Common blocks:
      PARAMETER ( MXV = $(MXV) )
      CHARACTER*10 NICKNA
      COMMON /NAMES/ NICKNA(MXV)
C  ---------------------------------------------------------------------
C
      NAME = NICKNA(I)
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE GETG ( GRAD )
C  ---------------------------------------------------------------------
C
C  Description:
C    Glue routine. Returns the gradient vector.
C
C  Output arguments:
C    GRAD       The gradient vector.
C
C  ---------------------------------------------------------------------
      IMPLICIT $(TYPE) (A-H,O-Z)
C  Arguments:
      DIMENSION GRAD(*)
C  ---------------------------------------------------------------------
C  Common blocks:
      PARAMETER ( MXV = $(MXV) )
      COMMON /PARAMS/ POINT(MXV), VAL
C  ---------------------------------------------------------------------
C
      MODE = 2
      CALL DERMER(POINT,VAL,GRAD,N1,N2,MODE)
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE GETG1 ( I, GRAD )
C  ---------------------------------------------------------------------
C
C  Description:
C    Glue routine. Returns the I-th component of the gradient vector.
C
C  Input arguments:
C    I          Index of the component to be returned.
C
C  Output arguments:
C    GRAD       The I-th component of the gradient vector.
C
C  ---------------------------------------------------------------------
      IMPLICIT $(TYPE) (A-H,O-Z)
C  Common blocks:
      PARAMETER ( MXV = $(MXV) )
      COMMON /PARAMS/ POINT(MXV), VAL
      COMMON /RWORK/ XT1(MXV), XT2(MXV), XT3(MXV), XT4(MXV), XT5(MXV),
     &               XT6(MXV), XT7(MXV+1)
C  ---------------------------------------------------------------------
C
      MODE = 2
      CALL DERMER(POINT,VAL,XT1,N1,N2,MODE)
      GRAD = XT1(I)
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE GETHES ( H, LD, HOW )
C  ---------------------------------------------------------------------
C
C  Description:
C    Glue routine. Returns the Hessian matrix, evaluated at the
C    current point.
C
C  Input arguments:
C    LD         The leading dimension of matrix H.
C    HOW        Character variable that determines the method of
C               approximation.
C               HOW = 'F' -> Use function values.
C               HOW = 'G' -> Use gradient values. Use it only if the
C                            user supplied routine GRANAL is available.
C               HOW = 'A' -> Use the user supplied subprogram HANAL.
C
C  Output arguments:
C    H          The Hessian matrix. On output the second partial
C               derivatives are stored in the lower triangular part
C               of H.
C
C  ---------------------------------------------------------------------
      IMPLICIT $(TYPE) (A-H,O-Z)
C  Arguments:
      DIMENSION H(LD,*)
      CHARACTER HOW
C  ---------------------------------------------------------------------
C  Common blocks:
      PARAMETER ( MXV = $(MXV) )
      COMMON /PARAMS/ POINT(MXV), VAL
      COMMON /DIMEN/ NOD, NOF
      COMMON /RWORK/ XT1(MXV), XT2(MXV), XT3(MXV), XT4(MXV), XT5(MXV),
     &               XT6(MXV), XT7(MXV+1)
C  ---------------------------------------------------------------------
C
      CALL SECDER(H,LD,NOD,POINT,VAL,HOW,XT1,XT2,NC)
      END
C  ---------------------------------------------------------------------
      SUBROUTINE GETJAC ( FJ, LD )
C  ---------------------------------------------------------------------
C
C  Description:
C    Glue routine. Returns the Jacobian matrix, evaluated at the
C    current point using the Jacobian mode currently in effect.
C
C  Input arguments:
C    LD         The leading dimension of matrix FJ.
C
C  Output arguments:
C    FJ         The Jacobian matrix with FJ(I,J)=df(i)/dx(j).
C
C  ---------------------------------------------------------------------
      IMPLICIT $(TYPE) (A-H,O-Z)
C  Arguments:
      DIMENSION FJ(LD,*)
C  ---------------------------------------------------------------------
C  Common blocks:
      PARAMETER ( MXV = $(MXV) )
      PARAMETER ( MXT = $(MXT) )
      COMMON /PARAMS/ POINT(MXV), VAL
      COMMON /DIMEN/ NOD, NOF
      COMMON /RWORKT/ FT1(MXT), FT2(MXT), FT3(MXT)
C  ---------------------------------------------------------------------
C
      CALL LSQFCN(NOF,NOD,POINT,FT1)
      CALL MJACOB(NOF,NOD,POINT,FT1,FJ,LD)
      END
C  ---------------------------------------------------------------------
      SUBROUTINE INIT ( VERSIM, MAXW )
C  ---------------------------------------------------------------------
C
C  Description:
C    Subroutine INIT assigns initial values to most of the Merlin
C    common blocks, upon program startup.
C
C  Input arguments:
C    VERSIM     The two dimensional Merlin work array.
C    MAXW       Number of available storage in array VERSIM.
C
C  ---------------------------------------------------------------------
      IMPLICIT $(TYPE) (A-H,O-Z)
C  Arguments:
      DIMENSION VERSIM(MAXW)
C  ---------------------------------------------------------------------
C  Common blocks:
      PARAMETER ( MXV = $(MXV) )
      COMMON /DETAIL/ IXAT(MXV), MARG(MXV)
      COMMON /XLIMS/ XLL(MXV), XRL(MXV)
      CHARACTER*5 GRADI
      COMMON /DERIMO/ GRADI(MXV)
      CHARACTER*10 NICKNA
      COMMON /NAMES/ NICKNA(MXV)
      COMMON /INDEFI/ INDEF(MXV)
      COMMON /ROSTEP/ STEP(MXV)
      COMMON /ALICO2/ NALIAS
      PARAMETER ( MAXFLA = 20 )
      COMMON /FLAGS/ FLAG(MAXFLA)
      PARAMETER ( MAXCFL = 10, MAXCLE = 30 )
      CHARACTER*(MAXCLE) CFLAG(MAXCFL)
      COMMON /CFLAGS/ CFLAG
      PARAMETER ( MAXMFI = 20 )
      COMMON /MCLV2/ ISTACK, INFILE, MBASE, INAD, MPC, MCLLNO, LENBUF,
     &               IBUF, MCLUNI(MAXMFI), NUMFI
      PARAMETER ( MAXINL = 20 )
      INTEGER BACKIN(MAXINL), WHATIN(MAXINL), REDIN(MAXINL)
      COMMON /INNER1/ LEVIN, BACKIN, WHATIN, REDIN
      PARAMETER ( MXBUF = $(MCLBUF), MXPTR = MXBUF/5 )
      INTEGER MCLPTR(MXPTR)
      COMMON /MCLBU2/ NPTR, ICPTR, MCLPTR
      COMMON /TOTCAL/ ITNOCS, IPARTC, ITDER, IPDER, ITHESS, IPHESS,
     &                ITJAC, IPJAC
      COMMON /BIGCOM/ BIGNUM, BLIM, DWARF, DWALOG
      COMMON /MODE3/ MODEV, IFFORM, INMCL
      COMMON /UPDATE/ IGRADU, IHESSU
      CHARACTER*10 DERIVA, PRINTO, PROCES, JACOMO
      COMMON /MODEIN/ DERIVA, PRINTO, PROCES, JACOMO
      LOGICAL AUTODE
      COMMON /AUTOMA/ AUTODE
      PARAMETER ( MXFILE = 80 )
      COMMON /MEMO2/ MEMOMO, MEMOKE, MEMREC, IUMEM
      COMMON /MTIT2/ LEMTIT
      COMMON /BACKIT/ IUBA, IUBAMO, IBASA
      COMMON /LOGIT/ IULOG
      COMMON /UNTSDF/ IU31, IU32
      COMMON /DIMEN/ NOD, NOF
C  ---------------------------------------------------------------------
C  Local variables:
      DIMENSION RVEC(1)
C  ---------------------------------------------------------------------
C
      DO 40,I=1,MAXW
40       VERSIM(I) = 0.
C
      ITNOCS = 0
      IPARTC = 0
      ITDER = 0
      IPDER = 0
      ITHESS = 0
      IPHESS = 0
      ITJAC = 0
      IPJAC = 0
C
      DO 10,I=1,MXV
         XLL(I) = -BLIM
         XRL(I) = BLIM
         GRADI(I) = 'QUAD'
         NICKNA(I) = ' '
         INDEF(I) = 0
         IXAT(I) = 1
         MARG(I) = 0
         STEP(I) = 1.
10    CONTINUE
C
      NALIAS = 0
C
      DO 20,I=1,MAXFLA
20       FLAG(I) = 0.
C
      MODEV = 1
      INMCL = 0
C
C  Set default function mode according to the number of terms.
      IF (NOF.EQ.0) THEN
         IFFORM = 0
      ELSE
         IFFORM = 1
      END IF
C
      DO 30,I=1,MAXCFL
30       CFLAG(I) = ' '
C
      IGRADU = 0
      IHESSU = 0
C
      DERIVA = 'QUAD      '
      JACOMO = 'JNUMER    '
      AUTODE = .FALSE.
      INFILE = -1
      LEVIN = 0
      IUMEM = -1
      LEMTIT = 0
      IUBA = -1
      IULOG = -1
      IU31 = 0
      IU32 = 0
      NPTR = 0
C
C  Just a precaution, in case one requests the seeds (RLUXUT) before
C  producing any random numbers. The sequence received by the user,
C  starts always at the second random number.
      CALL RANLUX(RVEC,1)
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE CPINIT
C  ---------------------------------------------------------------------
C
C  Description:
C    Subroutine CPINIT initializes some common blocks, necessary while
C    parsing the Merlin configuration file.
C
C  ---------------------------------------------------------------------
      IMPLICIT $(TYPE) (A-H,O-Z)
C  Common blocks:
      COMMON /UNITS/ IUINP, IUOUT
      LOGICAL HEADER, IPRECO, OPRECO, HASAPP, OUTAPP
      COMMON /CP1/ HEADER, IPRECO, OPRECO, HASAPP, OUTAPP
      PARAMETER ( MXFILE = 80 )
      CHARACTER*(MXFILE) PROLO, EPILO, IN, OUT, HFILE, MFILE, PDFILE
      COMMON /CP2/ PROLO, EPILO, IN, OUT, HFILE, MFILE, PDFILE
      COMMON /CP3/ NEOFA
      COMMON /UALLOC/ MINU, MAXU, LASTU
      COMMON /VSIZES/ NSCHA, NSINT, NSREAL
      CHARACTER*10 DERIVA, PRINTO, PROCES, JACOMO
      COMMON /MODEIN/ DERIVA, PRINTO, PROCES, JACOMO
      LOGICAL STRO, WEAK
      COMMON /PRILOG/ STRO, WEAK
      COMMON /ACCUR1/ NDRTF
      CHARACTER TAB, ESC
      COMMON /SPECHA/ TAB, ESC
      COMMON /BIGCOM/ BIGNUM, BLIM, DWARF, DWALOG
C  ---------------------------------------------------------------------
C
C  Largest real this machine can handle.
      BIGNUM = 1.E+36
C  Default margins.
      BLIM = 1.E+20
C  Smallest real, distinguishable from zero.
      DWARF = 1.E-36
C
      IUINP = 5
      IUOUT = 6
C
      HEADER = .TRUE.
      IPRECO = .TRUE.
      OPRECO = .TRUE.
      HASAPP = .FALSE.
      OUTAPP = .FALSE.
C
      PROLO = ' '
      EPILO = ' '
      IN = ' '
      OUT = ' '
      HFILE = 'HELP'
      MFILE = 'MACROF'
      PDFILE = 'PDESC'
C
C  Return on EOF.
      NEOFA = 3
C
      NSCHA = 1
      NSINT = 4
      NSREAL = 4
C
      MINU = 50
      MAXU = 80
      LASTU = MINU-1
C
      PROCES = 'IAF       '
      PRINTO = 'FULLPRINT '
C
      STRO = .TRUE.
      WEAK = .TRUE.
C
      NDRTF = 0
C
C  Settings for the ASCII character set.
C  Tab character.
      TAB = CHAR(9)
C  Backslash character.
      ESC = CHAR(92)
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE ZINIT
C  ---------------------------------------------------------------------
C
C  Description:
C    Subroutine ZINIT initializes some variables in common, after the
C    Merlin configuration file has been read.
C
C  ---------------------------------------------------------------------
      IMPLICIT $(TYPE) (A-H,O-Z)
C  Common blocks:
      COMMON /BIGCOM/ BIGNUM, BLIM, DWARF, DWALOG
      LOGICAL AIMON
      COMMON /AIM/ TARGET, AIMON
      COMMON /UALLOC/ MINU, MAXU, LASTU
      COMMON /UNITS/ IUINP, IUOUT
      PARAMETER ( MAXONL = 20 )
      INTEGER BACOUT(MAXONL), FROUT(MAXONL)
      COMMON /OUTER/ LEVOUT, BACOUT, FROUT
C  ---------------------------------------------------------------------
C
      DWALOG = LOG(DWARF)
      TARGET  = -BIGNUM
      AIMON = .FALSE.
      LASTU = MINU-1
C
      LEVOUT = 1
      BACOUT(LEVOUT) = IUOUT
      FROUT(LEVOUT) = 0
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE INLINE ( LINE, LCRACK, TOUPP, NEW, IS, IE, NIT, IERR )
C  ---------------------------------------------------------------------
C
C  Description:
C    This is the main input routine of Merlin. It reads a line from the
c    Merlin input file, or from the MCL buffer (when running buffered
C    MCL). Optionally converts the input to upper case and/or cracks
C    the input  line.
C
C  Input arguments:
C    LCRACK     When LCRACK = .TRUE. the input line will be cracked.
C    TOUPP      When TOUPP = .TRUE. the input line will be converted
C               to upper case.
C
C  Output arguments:
C    LINE       The line that was read.
C    NEW        The items of the cracked line. (Only if LCRACK=.TRUE.)
C    IS         IS(I) is the starting character position of item I in
C               the output character NEW. (Only if LCRACK=.TRUE.)
C    IE         IE(I) is the ending character position of item I in the
C               output character NEW. (Only if LCRACK=.TRUE.)
C    NIT        Number of items. (Only if LCRACK=.TRUE.)
C    IERR       Error indicator:
C               IERR =  0 -> Everything ok.
C               IERR = -1 -> End-of-File encountered.
C               IERR = -2 -> Error during cracking. Possibly a quote
C                            is missing.
C               IERR =  1 -> A blank line was input. (Only if LCRACK =
C                            .FALSE.; otherwise a blank line is
C                            signified by NIT = 0)
C
C  ---------------------------------------------------------------------
C  Arguments:
      CHARACTER*(*) LINE, NEW
      LOGICAL LCRACK, TOUPP
      INTEGER IS(*), IE(*)
C  ---------------------------------------------------------------------
C  Common blocks:
      PARAMETER ( MXBUF = $(MCLBUF), MXPTR = MXBUF/5 )
      CHARACTER*(MXBUF) MCLBUF
      INTEGER MCLPTR(MXPTR)
      COMMON /MCLBU1/ MCLBUF
      COMMON /MCLBU2/ NPTR, ICPTR, MCLPTR
      COMMON /UNITS/ IUINP, IUOUT
      COMMON /MODE3/ MODEV, IFFORM, INMCL
C  ---------------------------------------------------------------------
C
      IERR = 0
      IF (INMCL.NE.0 .AND. NPTR.NE.0) THEN
         ICPTR = ICPTR+1
         LINE = MCLBUF(MCLPTR(ICPTR):MCLPTR(ICPTR+1)-1)
         LL = MCLPTR(ICPTR+1)-MCLPTR(ICPTR)
         CALL INCMPC
      ELSE
         READ (IUINP,'(A)',END=100) LINE
         LL = LENGTH(LINE)
      END IF
C
C  Keep the current line in the history file.
      CALL LOGGER(LINE,LL)
C
C  Convert to upper case.
      IF (TOUPP) CALL UPPER(LINE,LL)
C
C  Crack the input line.
      IF (LCRACK) THEN
         CALL QLCR(LINE,LL,NEW,IS,IE,NIT,IERR)
      ELSE
         IF (LL.EQ.0) IERR = 1
      END IF
      RETURN
C
C  EOF in the input file.
100   CONTINUE
      IERR = -1
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE INNOCR ( LINE, TOUPP, IERR )
C  ---------------------------------------------------------------------
C
C  Description:
C    Part of the Merlin input handling routines. It will read a line
C    of input an optionally convert it to upper case.
C
C  Input arguments:
C    TOUPP      The input line is converted to upper case when
C               TOUPP = .TRUE.
C
C  Output arguments:
C    LINE       The line that was read.
C    IERR       Error indicator as described in subroutine INLINE.
C
C  ---------------------------------------------------------------------
C  Arguments:
      CHARACTER*(*) LINE
      LOGICAL TOUPP
C  ---------------------------------------------------------------------
C  Local variables:
      INTEGER IS(1), IE(1)
      CHARACTER NEW
C  ---------------------------------------------------------------------
C
      CALL INLINE(LINE,.FALSE.,TOUPP,NEW,IS,IE,NIT,IERR)
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE INREAL ( X, IERR )
C  ---------------------------------------------------------------------
C
C  Description:
C    Part of the Merlin input handling routines. It will read a line
C    of input and convert it to a real number.
C
C  Output arguments:
C    X          The number that was read.
C    IERR       Error indicator as described in subroutine INLINE.
C
C  ---------------------------------------------------------------------
      IMPLICIT $(TYPE) (A-H,O-Z)
C  Local variables:
      PARAMETER ( LINLEN = 80 )
      CHARACTER*(LINLEN) LINE, NEW
      INTEGER IS(LINLEN/2), IE(LINLEN/2)
C  ---------------------------------------------------------------------
C
      CALL INLINE(LINE,.TRUE.,.TRUE.,NEW,IS,IE,NIT,IERR)
      IF (IERR.EQ.0) THEN
         IF (NIT.EQ.0) THEN
            IERR = 1
         ELSE
            LE = IE(1)-IS(1)+1
            CALL TOREAL(NEW(IS(1):IE(1)),LE,X,IERR)
         END IF
      END IF
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE ININ ( N, IERR )
C  ---------------------------------------------------------------------
C
C  Description:
C    Part of the Merlin input handling routines. It will read a line
C    of input and convert it to an integer number.
C
C  Output arguments:
C    N          The number that was read.
C    IERR       Error indicator as described in subroutine INLINE.
C
C  ---------------------------------------------------------------------
C  Local variables:
      PARAMETER ( LINLEN = 80 )
      CHARACTER*(LINLEN) LINE, NEW
      INTEGER IS(LINLEN/2), IE(LINLEN/2)
C  ---------------------------------------------------------------------
C
      CALL INLINE(LINE,.TRUE.,.TRUE.,NEW,IS,IE,NIT,IERR)
      IF (IERR.EQ.0) THEN
         IF (NIT.EQ.0) THEN
            IERR = 1
         ELSE
            LE = IE(1)-IS(1)+1
            CALL TOINT(NEW(IS(1):IE(1)),LE,N,IERR)
         END IF
      END IF
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE ININ2 ( N1, N2, IERR )
C  ---------------------------------------------------------------------
C
C  Description:
C    Part of the Merlin input handling routines. It will read a line
C    of input and convert it to two integer numbers.
C
C  Input arguments:
C    N1         The first integer.
C    N2         The second integer.
C
C  Output arguments:
C    IERR       Error indicator as described in subroutine INLINE.
C
C  ---------------------------------------------------------------------
C  Local variables:
      PARAMETER ( LINLEN = 80 )
      CHARACTER*(LINLEN) LINE, NEW
      INTEGER IS(LINLEN/2), IE(LINLEN/2)
C  ---------------------------------------------------------------------
C
      CALL INLINE(LINE,.TRUE.,.TRUE.,NEW,IS,IE,NIT,IERR)
      IF (IERR.EQ.0) THEN
         IF (NIT.EQ.0 .OR. NIT.EQ.1) THEN
            IERR = 1
         ELSE
            LE = IE(1)-IS(1)+1
            CALL TOINT(NEW(IS(1):IE(1)),LE,N1,IERR)
            IF (IERR.EQ.0) THEN
               LE = IE(2)-IS(2)+1
               CALL TOINT(NEW(IS(2):IE(2)),LE,N2,IERR)
            END IF
         END IF
      END IF
C
      END
C  ---------------------------------------------------------------------
      FUNCTION ISIAF ( )
C  ---------------------------------------------------------------------
C
C  Description:
C    Returns 0 if the BATCH processing mode is in effect, 1 otherwise.
C
C  ---------------------------------------------------------------------
C  Common blocks:
      CHARACTER*10 DERIVA, PRINTO, PROCES, JACOMO
      COMMON /MODEIN/ DERIVA, PRINTO, PROCES, JACOMO
C  ---------------------------------------------------------------------
C
      IF (PROCES.EQ.'IAF') THEN
         ISIAF = 1
      ELSE
         ISIAF = 0
      END IF
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE LOCC ( COM, IC )
C  ---------------------------------------------------------------------
C
C  Description:
C    Searches the Merlin commands to determine whether the input
C    string COM is a valid Merlin command.
C
C  Input arguments:
C    COM        Command to search for.
C
C  Output arguments:
C    IC         Index of COM in the command array. If COM is not a
C               valid Merlin command, IC is set to 0.
C
C  ---------------------------------------------------------------------
C  Arguments:
      CHARACTER*(*) COM
C  ---------------------------------------------------------------------
C  Common blocks:
      PARAMETER ( NMCOM = 98, MXPLUG = $(MXPLUG), NCOM = NMCOM+MXPLUG )
      CHARACTER*10 COMMAN(NCOM)
      COMMON /COM1/ COMMAN
C  ---------------------------------------------------------------------
C
      DO 20,I=1,NCOM
         IF (ISCOMP(COMMAN(I),COM).NE.0) THEN
            IC = I
            RETURN
         END IF
20    CONTINUE
      IC = 0
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE LOCMAC ( COMM, IUMAC, ISUCC )
C  ---------------------------------------------------------------------
C
C  Description:
C    Searches the macro file (IUMAC) for a macro name (COMM).
C
C  Input arguments:
C    COMM       The macro name to be located.
C    IUMAC      Unit number of the open macro file.
C
C  Output arguments:
C    ISUCC      Error indicator.
C               ISUCC = -1 -> The macro file does not correspond to
C                             the Merlin conventions (a corrupt file).
C               ISUCC =  0 -> Macro not found.
C               ISUCC =  1 -> Macro found.
C
C  ---------------------------------------------------------------------
C  Arguments:
      CHARACTER*(*) COMM
C  ---------------------------------------------------------------------
C  Local variables:
      PARAMETER ( LINLEN = 80 )
      CHARACTER*(LINLEN) MAC, STAR
C  ---------------------------------------------------------------------
C
      REWIND IUMAC
C
C  Read/skip lines until the macro separator (a star) is found.
10    READ (IUMAC,'(A)',END=20) STAR
      IF (STAR.NE.'*') GOTO 10
C
C  Separator found. Read and compare the macro name.
      READ (IUMAC,'(A)',END=30) MAC
      IF (ISCOMP(MAC,COMM).EQ.0) GOTO 10
C
C  The macro name matches.
      ISUCC = 1
      RETURN
C
C  EOF while searching for the macro separator.
20    REWIND IUMAC
      ISUCC = 0
      RETURN
C
C  EOF just after the macro separator.
30    REWIND IUMAC
      ISUCC = -1
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE MABORT ( MESS, SUB )
C  ---------------------------------------------------------------------
C
C  Description:
C    This routine will be called to issue informative messages and stop
C    the program when a situation that normally should never happen,
C    is encountered.
C    Subroutine MABORT is mainly a development aid.
C
C  Input arguments:
C    MESS       An informative message to be displayed explaining the
C               reason for the abort.
C    SUB        The routine name in which the abnormal situation
C               happened.
C
C  ---------------------------------------------------------------------
C  Arguments:
      CHARACTER*(*) MESS, SUB
C  ---------------------------------------------------------------------
C  Common blocks:
      COMMON /UNITS/ IUINP, IUOUT
      COMMON /UNTSDF/ IU31, IU32
C  ---------------------------------------------------------------------
C
C  Is the Merlin output file/unit open ?
      IF (IU32.EQ.0) THEN
         WRITE (*,10) SUB
         WRITE (*,20) MESS
      ELSE
         WRITE (IUOUT,10) SUB
         WRITE (IUOUT,20) MESS
      END IF
      STOP 'ABRT'
C
10    FORMAT (/$(FC)'**** Routine ',A,' aborted.')
20    FORMAT ($(FC)'**** ',A)
      END
C  ---------------------------------------------------------------------
      SUBROUTINE MACAC
C  ---------------------------------------------------------------------
C
C  Description:
C    This routine determines the relative machine accuracy (COMACC)
C    and constructs the run-time formats used by Merlin.
C    Run time formats are used for printing real numbers with
C    as many significant digits as the machine accuracy indicates.
C    The number of significant digits (NDRTF) can be set from the
C    configuration file as well.
C
C  ---------------------------------------------------------------------
      IMPLICIT $(TYPE) (A-H,O-Z)
C  Common blocks:
      COMMON /ACCURA/ COMACC
      COMMON /ACCUR1/ NDRTF
      CHARACTER RTFORM*10, RRFORM*7, REFORM*10
      COMMON /ACCUR2/ RTFORM, RRFORM, REFORM
C  ---------------------------------------------------------------------
C
      TEN = 10.
      IF (NDRTF.EQ.0) THEN
         ONE = 1.
         DO 10,I=1,1000
            COMACC = TEN**(-I)
            X1 = ONE+COMACC
            X2 = ONE-COMACC
            IF (X1.EQ.ONE .OR. X2.EQ.ONE) THEN
               NDRTF = I-1
               GOTO 20
            END IF
10       CONTINUE
         CALL MABORT('Cannot determine accuracy.','MACAC')
      END IF
C
C  Construct run time formats.
20    CONTINUE
      COMACC = TEN**(-NDRTF)
      RTFORM = '(1PG__.__)'
      RRFORM = '(G__.0)'
      REFORM = '(1PE__.__)'
C
C  Impose a maximum of 16 significant digits.
      IF (NDRTF.GT.16) NDRTF = 16
      ND7 = NDRTF+7
      WRITE (RTFORM(5:6),'(I2.2)') ND7
      WRITE (RTFORM(8:9),'(I2.2)') NDRTF
      WRITE (REFORM(5:6),'(I2.2)') ND7
      WRITE (REFORM(8:9),'(I2.2)') NDRTF
      WRITE (RRFORM(3:4),'(I2.2)') ND7
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE MATCHA ( CMD, LE, ACTCOM, IAL )
C  ---------------------------------------------------------------------
C
C  Description:
C    Subroutine MATCHA checks if a given string (CMD) is a valid
C    alias, and returns the actual command (ACTCOM) it corresponds to,
C    as well as its index (IAL) in the alias table.
C
C  Input arguments:
C    CMD        The alias to search for.
C    LE         Effective length of CMD.
C
C  Output arguments:
C    ACTCOM     If CMD is a valid alias, ACTCOM is set to the actual
C               command that CMD stands for.
C    IAL        If CMD is a valid alias, IAL is its index in the alias
C               table. Otherwise IAL is set to 0.
C
C  ---------------------------------------------------------------------
C  Arguments:
      CHARACTER*(*) CMD, ACTCOM
C  ---------------------------------------------------------------------
C  Common blocks:
      PARAMETER ( MAXALI = 20 )
      CHARACTER*5 ALIAS(MAXALI)
      CHARACTER*10 ALICOM(MAXALI)
      COMMON /ALICO1/ ALIAS, ALICOM
      COMMON /ALICO2/ NALIAS
C  ---------------------------------------------------------------------
C
      IAL = 0
      IF (LE.GT.LEN(ALIAS(1))) RETURN
C
      DO 10,I=1,NALIAS
         IF (ISCOMP(CMD,ALIAS(I)).NE.0) THEN
            IAL = I
            ACTCOM = ALICOM(I)
            RETURN
         END IF
10    CONTINUE
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE MATCH ( ARRAY, N, NAME, LE, MINF, IND, IEND )
C  ---------------------------------------------------------------------
C
C  Description:
C    Searches the Merlin command array (ARRAY) for the user entered
C    command NAME, taking into account that NAME may be the
C    abbreviation of a Merlin command. In case of an unambiguous
C    command, its index IND is returned, otherwise the range of commands
C    that would match NAME (IND...IEND).
C
C  Input arguments:
C    ARRAY      The Merlin command array.
C    N          Number of entries in ARRAY.
C    NAME       Command/abbreviation to search for.
C    LE         Effective length of NAME.
C
C  Output arguments:
C    MINF       Error indicator:
C               MINF =  0 -> NAME was unambiguously found in ARRAY. Its
C                            index is returned in IND.
C               MINF = -1 -> NAME is ambiguous. The range of possible
C                            commands that match is IND...IEND.
C               MINF = -2 -> NAME not found.
C    IND        When MINF = 0, the index of NAME in ARRAY.
C               When MINF = -1, the starting index of a range of
C               elements in ARRAY that would match NAME.
C    IEND       When MINF = -1, the ending index of a range of
C               elements in ARRAY that would match NAME.
C
C  ---------------------------------------------------------------------
C  Arguments:
      CHARACTER*(*) ARRAY(N), NAME
C  ---------------------------------------------------------------------
C
      IF (LE.GT.LEN(ARRAY(1))) THEN
         MINF = -2
         RETURN
      END IF
C
      DO 10,I=1,N
         IF (ISCOMP(ARRAY(I)(1:LE),NAME).NE.0) THEN
            IF (I.EQ.N) THEN
               MINF = 0
               IND = I
            ELSE IF (LENGTH(ARRAY(I)).EQ.LE) THEN
               MINF = 0
               IND = I
            ELSE
               IF (ISCOMP(ARRAY(I+1)(1:LE),NAME).NE.0) THEN
                  MINF = -1
                  IND = I
                  DO 20,J=I+2,N
                     IF (ISCOMP(ARRAY(J)(1:LE),NAME).EQ.0) THEN
                        IEND = J-1
                        RETURN
                     END IF
20                CONTINUE
                  IEND = N
                  RETURN
               ELSE
                  MINF = 0
                  IND = I
               END IF
            END IF
            RETURN
         END IF
10    CONTINUE
      MINF = -2
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE MCLEND
C  ---------------------------------------------------------------------
C
C  Description:
C    Subroutine MCLEND is called to terminate an MCL program.
C    It closes all open MCL files and decrements the input level.
C
C  ---------------------------------------------------------------------
      IMPLICIT $(TYPE) (A-H,O-Z)
C  Arguments:
C  ---------------------------------------------------------------------
C  Common blocks:
      COMMON /MODE3/ MODEV, IFFORM, INMCL
      PARAMETER ( MAXMFI = 20 )
      COMMON /MCLV2/ ISTACK, INFILE, MBASE, INAD, MPC, MCLLNO, LENBUF,
     &               IBUF, MCLUNI(MAXMFI), NUMFI
C  ---------------------------------------------------------------------
C
C  Close all files opened for MCL I/O operations.
      DO 10,I=1,NUMFI
10       CLOSE (UNIT=MCLUNI(I))
      NUMFI = 0
C
C  Decrement the input level.
      CALL DECILE(3)
      INFILE = -1
      INMCL = 0
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE USEMCL ( COM, MERR )
C  ---------------------------------------------------------------------
C
C  Description:
C    Routine to interface Merlin with the MCL object code.
C
C  Input arguments:
C    COM        Command to be executed. Does not contain the initial
C               $ sign.
C
C  Output arguments:
C    MERR       Error indicator.
C
C  ---------------------------------------------------------------------
      IMPLICIT $(TYPE) (A-H,O-Z)
C  Arguments:
      CHARACTER*10 COM
C  ---------------------------------------------------------------------
C  Common blocks:
      PARAMETER ( MAXFLA = 20 )
      COMMON /FLAGS/ FLAG(MAXFLA)
      PARAMETER ( MAXCFL = 10, MAXCLE = 30 )
      CHARACTER*10 DERIVA, PRINTO, PROCES, JACOMO
      COMMON /MODEIN/ DERIVA, PRINTO, PROCES, JACOMO
      LOGICAL STRO, WEAK
      COMMON /PRILOG/ STRO, WEAK
      COMMON /UNITS/ IUINP, IUOUT
      PARAMETER ( MXV = $(MXV) )
      PARAMETER ( MXT = $(MXT) )
      COMMON /PARAMS/ POINT(MXV), VAL
      COMMON /DETAIL/ IXAT(MXV), MARG(MXV)
      COMMON /XLIMS/ XLL(MXV), XRL(MXV)
      COMMON /ROSTEP/ STEP(MXV)
      COMMON /INDEFI/ INDEF(MXV)
      COMMON /RWORK/ XT1(MXV), XT2(MXV), XT3(MXV), XT4(MXV), XT5(MXV),
     &               XT6(MXV), XT7(MXV+1)
      COMMON /TERMS/ VECLS(MXT)
      COMMON /TOTCAL/ ITNOCS, IPARTC, ITDER, IPDER, ITHESS, IPHESS,
     &                ITJAC, IPJAC
      COMMON /DIMEN/ NOD, NOF
      COMMON /ACCURA/ COMACC
      COMMON /MODE3/ MODEV, IFFORM, INMCL
      PARAMETER ( LINLEN = 80 )
      CHARACTER*(LINLEN) INPLIN, QINPL
      COMMON /INDAT2/ INPLIN, QINPL
      PARAMETER ( MAXINL = 20 )
      INTEGER BACKIN(MAXINL), WHATIN(MAXINL), REDIN(MAXINL)
      COMMON /INNER1/ LEVIN, BACKIN, WHATIN, REDIN
      PARAMETER ( LSTA = $(MCLMEM) )
      COMMON /MCLV1/ STACK(LSTA)
      PARAMETER ( MAXMFI = 20 )
      COMMON /MCLV2/ ISTACK, INFILE, MBASE, INAD, MPC, MCLLNO, LENBUF,
     &               IBUF, MCLUNI(MAXMFI), NUMFI
      CHARACTER*80 MCLCLI
      COMMON /MCLV3/ MCLCLI
C  ---------------------------------------------------------------------
C  Local variables:
      SAVE
      PARAMETER ( MXFILE = 80 )
      CHARACTER*(MXFILE) FILE
      PARAMETER ( MR = 80, MB = 20, MC = 109 )
      CHARACTER MCLCOM(MC)*10, CHARA*61, COMW*10, KEY*10, TYP*1
      CHARACTER*(MR) BUFFER(MB), TEMPO, CARRY*64
      DIMENSION IRLX(25)
      CHARACTER*80 LINE
      CHARACTER*30 STR
      LOGICAL LO
      LO(I1,I2) = JCOM.GE.I1 .AND. JCOM.LE.I2
      DATA (MCLCOM(I),I=1,40) /
     &     'BACKUP    ',  'DERIVA    ',  'PRINTOUT  ',  'HTCOUNT   ',
     &     'POP       ',  'PUSH      ',  'PUSHC     ',  '>         ',
     &     '<         ',  '=         ',  '#         ',  '>=        ',
     &     '<=        ',  '+         ',  '-         ',  '/         ',
     &     '*         ',  '**        ',  'PAUSE     ',  'CONTINUE  ',
     &     'FINISH    ',  'FACT      ',  'CLOSE     ',  '>>        ',
     &     'STACKSWAP ',  'FLUSH     ',  'NOTE      ',  'OUT       ',
     &     'IN        ',  'GETPP     ',  'HPCOUNT   ',  'CALLSUB   ',
     &     'NORM      ',  'GETSEED   ',  'WSUM      ',  'LOCCHECK  ',
     &     'GETCONT   ',  'SETSEED   ',  'NOT       ',  'OR        ' /
      DATA (MCLCOM(I),I=41,80) /
     &     'AND       ',  'XOR       ',  'ABS       ',  'LOG10     ',
     &     'LOG       ',  'SIN       ',  'COS       ',  'TAN       ',
     &     'ASIN      ',  'ACOS      ',  'ATAN      ',  'SINH      ',
     &     'COSH      ',  'TANH      ',  'ASINH     ',  'ACOSH     ',
     &     'ATANH     ',  'MIN       ',  'MAX       ',  'TRUNC     ',
     &     'ROUND     ',  'MOD       ',  'MEAN      ',  'POPLV     ',
     &     'POPCONT   ',  'X         ',  'R         ',  'L         ',
     &     'STEP      ',  'FIX       ',  'GRAD      ',  'MARG      ',
     &     'TCOUNT    ',  'PCOUNT    ',  'DIM       ',  'VALUE     ',
     &     'PRECISION ',  'SQRT      ',  'GTCOUNT   ',  'PROCESS   ' /
      DATA (MCLCOM(I),I=81,MC) /
     &     'BLOCKCOPY ',  'GRMS      ',  'TESTMOVE  ',  'MOVE      ',
     &     'LINE      ',  'PUSHA     ',  'IOFFSET   ',  'EXP       ',
     &     'RAN       ',  'GRADNORM  ',  'FLAG      ',  'GPCOUNT   ',
     &     'PUSHMB    ',  'POPMB     ',  'IPOP      ',  'ALLOC     ',
     &     'DEALLOC   ',  'PUSHINAD  ',  'POPINAD   ',  'SUBRETURN ',
     &     'IPUSH     ',  'TERMS     ',  'TERM      ',  'FUNMODE   ',
     &     'NFLAGS    ',  'NCFLAGS   ',  'JTCOUNT   ',  'JPCOUNT   ',
     &     'JACOMO    ' /
C  ---------------------------------------------------------------------
C
C  Default value for the error code.
      MERR = 0
C
C  Recognize command.
      JCOM = 0
      DO 101,I=1,MC
         IF (COM(2:).EQ.MCLCOM(I)) THEN
            JCOM = I
            GOTO 102
         END IF
101   CONTINUE
C
C  Not a valid interface command.
      IF (WEAK) THEN
         LE = LENGTH(COM)
         IF (LE.EQ.0) LE = 1
         WRITE (IUOUT,200) COM(1:LE)
      END IF
      MERR = -2
      RETURN
102   CONTINUE
C
C  Send control to the appropriate point.
      IF (LO(1,10)) GOTO (1,2,3,4,5,6,7,8,9,10), JCOM
      IF (LO(11,20)) GOTO (11,12,13,14,15,16,17,18,19,20), JCOM-10
      IF (LO(21,30)) GOTO (21,22,23,24,25,26,27,28,29,30), JCOM-20
      IF (LO(31,40)) GOTO (31,123,38,39,40,41,42,43,44,45), JCOM-30
      IF (LO(41,50)) GOTO (46,47,48,49,50,51,52,53,54,55), JCOM-40
      IF (LO(51,60)) GOTO (56,57,58,59,60,61,62,63,64,65), JCOM-50
      IF (LO(61,70)) GOTO (66,67,68,69,70,71,72,73,74,75), JCOM-60
      IF (LO(71,80)) GOTO (76,77,78,79,80,81,82,83,84,85), JCOM-70
      IF (LO(81,90)) GOTO (122,87,88,89,90,121,120,93,94,104), JCOM-80
      IF (LO(91,95)) GOTO (107,108,109,110,118), JCOM-90
      IF (LO(96,100)) GOTO (112,113,114,115,116), JCOM-95
      IF (LO(101,105)) GOTO(117,119,125,126,128), JCOM-100
      IF (LO(106,MC)) GOTO(129,130,131,132), JCOM-105
      CALL I2STR(JCOM,STR,LES)
      CALL MABORT('Index '//STR(1:LES)//' not valid.','USEMCL')
C  .........................................................  BACKUP
1     CONTINUE
      CALL GETPPS('BACKUP','MODE',STR,LES)
      IF (ISCOMP(STR,'NO').NE.0) THEN
         IBACK = 1
      ELSE IF (ISCOMP(STR,'LAST').NE.0) THEN
         IBACK = 2
      ELSE IF (ISCOMP(STR,'FULL').NE.0) THEN
         IBACK = 3
      ELSE
         CALL MABORT('Cannot recognize backup mode.','USEMCL')
      END IF
      ISTACK = ISTACK + 1
      STACK(ISTACK) = IBACK
      RETURN
C  .........................................................  DERIVA
2     CONTINUE
      IF (DERIVA.EQ.'ANAL') THEN
         IDERV = 1
      ELSE IF (DERIVA.EQ.'NUMER') THEN
         IDERV = 2
      ELSE IF (DERIVA.EQ.'QUAD') THEN
         IDERV = 3
      ELSE IF (DERIVA.EQ.'FAST') THEN
         IDERV = 4
      ELSE IF (DERIVA.EQ.'MIXED') THEN
         IDERV = 5
      ELSE
         CALL MABORT('Cannot recognize derivative mode.','USEMCL')
      END IF
      ISTACK = ISTACK + 1
      STACK(ISTACK) = IDERV
      RETURN
C  .........................................................  PRINTOUT
3     CONTINUE
      IF (PRINTO.EQ.'NOPRINT') THEN
         IPRIN = 1
      ELSE IF (PRINTO.EQ.'HALFPRINT') THEN
         IPRIN = 2
      ELSE IF (PRINTO.EQ.'FULLPRINT') THEN
         IPRIN = 3
      ELSE
         CALL MABORT('Cannot recognize printout mode.','USEMCL')
      END IF
      ISTACK = ISTACK + 1
      STACK(ISTACK) = IPRIN
      RETURN
C  .........................................................  HTCOUNT
4     CONTINUE
      ISTACK = ISTACK + 1
      STACK(ISTACK) = ITHESS
      RETURN
C  .........................................................  POP
5     CONTINUE
      CALL ININ(NUMBER,III)
      IF (III.NE.0) CALL MABORT('Cannot read MOC (POP).','USEMCL')
      STACK(NUMBER+MBASE) = STACK(ISTACK)
      ISTACK = ISTACK-1
      RETURN
C  .........................................................  PUSH
6     CONTINUE
      CALL ININ(NUMBER,III)
      IF (III.NE.0) CALL MABORT('Cannot read MOC (PUSH).','USEMCL')
      ISTACK = ISTACK+1
      IF (NUMBER.LE.0) CALL MABORT('Negative PUSH.','USEMCL')
      STACK(ISTACK) = STACK(NUMBER+MBASE)
      RETURN
C  .........................................................  PUSHC
7     CONTINUE
      CALL INREAL(XNUM,III)
      IF (III.NE.0) CALL MABORT('Cannot read MOC (PUSHC).','USEMCL')
      ISTACK = ISTACK+1
      STACK(ISTACK) = XNUM
      RETURN
C  .........................................................  >
8     CONTINUE
      CALL TOPTWO(UPS,DOS)
      IF (DOS.GT.UPS) STACK(ISTACK) = 1.
      RETURN
C  .........................................................  <
9     CONTINUE
      CALL TOPTWO(UPS,DOS)
      IF (DOS.LT.UPS) STACK(ISTACK) = 1.
      RETURN
C  .........................................................  ==
10    CONTINUE
      CALL TOPTWO(UPS,DOS)
      IF (DOS.EQ.UPS) STACK(ISTACK) = 1.
      RETURN
C  .........................................................  #
11    CONTINUE
      CALL TOPTWO(UPS,DOS)
      IF (DOS.NE.UPS) STACK(ISTACK) = 1.
      RETURN
C  .........................................................  >=
12    CONTINUE
      CALL TOPTWO(UPS,DOS)
      IF (DOS.GE.UPS) STACK(ISTACK) = 1.
      RETURN
C  .........................................................  <=
13    CONTINUE
      CALL TOPTWO(UPS,DOS)
      IF (DOS.LE.UPS) STACK(ISTACK) = 1.
      RETURN
C  .........................................................  +
14    CONTINUE
      CALL TOPTWO(UPS,DOS)
      STACK(ISTACK) = DOS + UPS
      RETURN
C  .........................................................  -
15    CONTINUE
      CALL TOPTWO(UPS,DOS)
      STACK(ISTACK) = DOS - UPS
      RETURN
C  .........................................................  /
16    CONTINUE
      CALL TOPTWO(UPS,DOS)
      STACK(ISTACK) = DOS / UPS
      RETURN
C  .........................................................  *
17    CONTINUE
      CALL TOPTWO(UPS,DOS)
      STACK(ISTACK) = DOS * UPS
      RETURN
C  .........................................................  **
18    CONTINUE
      CALL TOPTWO(UPS,DOS)
      IF (DOS.GT.0.)  THEN
         IF (UPS.NE.0.) THEN
            STACK(ISTACK) = DOS**UPS
         ELSE
            STACK(ISTACK) = 1.
         END IF
      ELSE IF (DOS.EQ.0.) THEN
         IF (UPS.GT.0) THEN
            STACK(ISTACK) = 0.
         ELSE
C  Attempting 0 ** x, x <= 0
            IF (WEAK) WRITE (IUOUT,210)
            MERR = -2
         END IF
      ELSE
         CHECK = ABS(UPS - NINT(UPS))
         IF (CHECK .LT. ABS(UPS)*SQRT(COMACC) )   THEN
            STACK(ISTACK) = DOS**NINT(UPS)
         ELSE
C  Attempting x ** y, x < 0, y not integer.
            IF (WEAK) WRITE (IUOUT,220)
            MERR = -2
         END IF
      END IF
      RETURN
C  .........................................................  PAUSE
19    CONTINUE
      IUPREV = BACKIN(LEVIN)
      IF (STRO) WRITE (IUOUT,230)
      READ (IUPREV,'(A)',END=231) STR
      RETURN
C
231   CONTINUE
      MERR = -2
      RETURN
C  .........................................................  CONTINUE
20    CONTINUE
      RETURN
C  .........................................................  FINISH
21    CONTINUE
      CALL MCLEND
      RETURN
C  .........................................................  FACT
22    CONTINUE
      NFA = NINT(STACK(ISTACK))
      CALL FACT(NFA,FAC,MERR)
      STACK(ISTACK) = FAC
      RETURN
C  .........................................................  CLOSE
23    CONTINUE
      CALL MCLCLO(MERR)
      RETURN
C  .........................................................  >>
24    CONTINUE
      MCLCLI = INPLIN(5:)
      RETURN
C  .........................................................  STACKSWAP
25    CONTINUE
      IS = ISTACK-1
      X1 = STACK(ISTACK)
      X2 = STACK(IS)
      STACK(ISTACK) = X2
      STACK(IS) = X1
      RETURN
C  .........................................................  FLUSH
26    CONTINUE
      TEMPO(LENBUF+1 : ) = ' '
      BUFFER(IBUF) = TEMPO
      TEMPO = ' '
      CALL DETUNT(IUOUT,IOUNIT,MERR)
      IF (MERR.NE.0) RETURN
      CALL FLUSHB(BUFFER,IBUF,IOUNIT)
      LENBUF = 0
      RETURN
C  .........................................................  NOTE
27    CONTINUE
      CHARA = ' '
      CALL ININ(INTE,III)
      IF (III.NE.0) CALL MABORT('Cannot read MOC (NOTE).','USEMCL')
      IF (INTE.GT.0) THEN
         CALL INNOCR(CHARA(1:INTE),.FALSE.,III)
         IF (III.NE.0)
     &          CALL MABORT('Cannot read MOC (NOTE).','USEMCL')
         CALL DETUNT(IUOUT,IOUN,MERR)
         IF (MERR.NE.0) RETURN
         CALL TREBUF(BUFFER,IBUF,LENBUF,CHARA,INTE,TEMPO,IOUN)
      END IF
      RETURN
C  .........................................................  OUT
28    CONTINUE
      XX = STACK(ISTACK)
      ISTACK = ISTACK - 1
      CARRY(1:1) = ' '
      CALL STORIC(XX,0,CARRY(2:),LLL)
      LLL = LLL+1
      CALL DETUNT(IUOUT,IOUT,MERR)
      IF (MERR.NE.0) RETURN
      CALL TREBUF(BUFFER,IBUF,LENBUF,CARRY,LLL,TEMPO,IOUT)
      RETURN
C  .........................................................  IN
29    CONTINUE
C
C  Need to determine previous input unit from the input-stack.
      CALL ININ(INTE,III)
      IF (III.NE.0) CALL MABORT('Cannot read MOC (IN).','USEMCL')
      IUPREV = BACKIN(LEVIN)
      CALL DETUNT(IUPREV,IUT,MERR)
      IF (MERR.NE.0) RETURN
      ILO = ISTACK + 1
      ISTACK = ISTACK + INTE
      READ (IUT,*,END=291,IOSTAT=III) (STACK(KK),KK=ISTACK,ILO,-1)
      IF (III.NE.0) THEN
         IF (WEAK) THEN
            CALL U2NAME(IUT,'input',FILE,LEF)
            IF (LEF.NE.0) THEN
               WRITE (IUOUT,260) FILE(1:LEF)
            ELSE
               WRITE (IUOUT,270)
            END IF
         END IF
         MERR = -2
      END IF
      RETURN
C
291   CONTINUE
      IF (IUT.EQ.IUPREV) THEN
         MERR = -1
      ELSE
         IF (WEAK) THEN
            CALL U2NAME(IUT,'input',FILE,LEF)
            IF (LEF.NE.0) THEN
               WRITE (IUOUT,240) FILE(1:LEF)
            ELSE
               WRITE (IUOUT,250)
            END IF
         END IF
         MERR = -2
      END IF
      RETURN
C  .........................................................  GETPP
30    CONTINUE
      CALL INNOCR(COMW,.FALSE.,III)
      IF (III.NE.0) CALL MABORT('Cannot read MOC (GETPP).','USEMCL')
      CALL INNOCR(KEY,.FALSE.,III)
      IF (III.NE.0) CALL MABORT('Cannot read MOC (GETPP).','USEMCL')
      CALL INNOCR(TYP,.FALSE.,III)
      IF (III.NE.0) CALL MABORT('Cannot read MOC (GETPP).','USEMCL')
      IF (ISCOMP(TYP,'I').NE.0) THEN
         CALL GETPPI(COMW,KEY,IVAL)
         ISTACK = ISTACK+1
         STACK(ISTACK) = IVAL
      ELSE IF (ISCOMP(TYP,'R').NE.0) THEN
         CALL GETPPR(COMW,KEY,XVAL)
         ISTACK = ISTACK+1
         STACK(ISTACK) = XVAL
      ELSE
         CALL MABORT('Cannot return strings (GETPP).','USEMCL')
      END IF
      RETURN
C  .........................................................  HPCOUNT
31    CONTINUE
      ISTACK = ISTACK + 1
      STACK(ISTACK) = IPHESS
      RETURN
C  .........................................................  CALLSUB
123   CONTINUE
      CALL ININ(NUMBER,III)
      IF (III.NE.0) CALL MABORT('Cannot read MOC (CALLSUB).','USEMCL')
      ISTACK = ISTACK+1
      STACK(ISTACK) = MPC+1
      CALL MOVIT(NUMBER)
      RETURN
C  .........................................................  NORM
38    CONTINUE
C
C  SUB NORM ( ARRAY[1:*]; N; L; RESULT )
      IAR = NINT(STACK(INAD+1-1))
      IAD = NINT(STACK(INAD+2-1))
      NN = NINT(STACK(IAD))
      IAD = NINT(STACK(INAD+3-1))
      LL = NINT(STACK(IAD))
      IRES = NINT(STACK(INAD+4-1))
      IF (LL.LE.0 .AND. LL.NE.-1) THEN
         IF (WEAK) THEN
            CALL I2STR(LL,STR,LES)
            WRITE (IUOUT,280) STR(1:LES)
         END IF
         MERR = -2
         RETURN
      END IF
C
      IAR = IAR-1
      IF (LL.EQ.-1) THEN
         XMA = STACK(IAR+1)
         DO 702,I=2,NN
702         XMA = MAX(XMA,ABS(STACK(IAR+I)))
         STACK(IRES) = XMA
      ELSE
         SUM = 0.
         DO 701,I=1,NN
701         SUM = SUM+ABS(STACK(IAR+I))**LL
         ONE = 1.
         STACK(IRES) = SUM**(ONE/LL)
      END IF
      RETURN
C  .........................................................  GETSEED
39    CONTINUE
C
C  SUB GETSEED ( ARRAY[1:25] )
      IAR = NINT(STACK(INAD))-1
      CALL RLUXUT(IRLX)
      DO 703,I=1,25
703      STACK(IAR+I) = IRLX(I)
      RETURN
C  .........................................................  WSUM
40    CONTINUE
C
C   Check if there are enough numbers in the stack.
      CALL ININ(NUMBER,III)
      IF (III.NE.0) CALL MABORT('Cannot read MOC (WSUM).','USEMCL')
      IF (ISTACK.LT.2*NUMBER)
     &    CALL MABORT('Stack underflow (WSUM).','USEMCL')
C
      XSUM = 0.
      DO 201,II=0,NUMBER-1
201      XSUM = XSUM + STACK(ISTACK-II)*STACK(ISTACK-II-NUMBER)
      ISTACK = ISTACK + 1 - 2*NUMBER
      STACK(ISTACK) = XSUM
      RETURN
C  .........................................................  LOCCHECK
41    CONTINUE
      IF ((STACK(ISTACK).GE.STACK(ISTACK-2)) .AND.
     &    (STACK(ISTACK).LE.STACK(ISTACK-1))) THEN
         ISTACK = ISTACK-2
         STACK(ISTACK) = INT(STACK(ISTACK+2))
      ELSE
         IF (WEAK) WRITE (IUOUT,290)
         MERR = -2
      END IF
      RETURN
C  .........................................................  GETCONT
42    CONTINUE
      II = NINT(STACK(ISTACK))
      STACK(ISTACK) = STACK(II)
      RETURN
C  .........................................................  SETSEED
43    CONTINUE
C
C  SUB SETSEED ( ARRAY[1:25] )
      IAR = NINT(STACK(INAD))-1
      DO 704,I=1,25
704      IRLX(I) = NINT(STACK(IAR+I))
      CALL RLUXIN(IRLX)
      IF (STRO) WRITE (IUOUT,300)
      RETURN
C  .........................................................  NOT
44    CONTINUE
      IF (STACK(ISTACK).EQ.0.) THEN
         STACK(ISTACK) = 1.
      ELSE
         STACK(ISTACK) = 0.
      END IF
      RETURN
C  .........................................................  OR
45    CONTINUE
      IF (ISTACK.GE.2) THEN
         IB = NINT(STACK(ISTACK))
         ISTACK = ISTACK-1
         IA = NINT(STACK(ISTACK))
         IF (IB.NE.0) THEN
            STACK(ISTACK) = 1
         ELSE
            STACK(ISTACK) = IA
         END IF
      ELSE
         CALL MABORT('Stack underflow (OR).','USEMCL')
      END IF
      RETURN
C  .........................................................  AND
46    CONTINUE
      IF (ISTACK.GE.2) THEN
         IB = NINT(STACK(ISTACK))
         ISTACK = ISTACK - 1
         IA = NINT(STACK(ISTACK))
         IF (IB.NE.0) THEN
            STACK(ISTACK) = IA
         ELSE
            STACK(ISTACK) = 0
         END IF
      ELSE
         CALL MABORT('Stack underflow (AND).','USEMCL')
      END IF
      RETURN
C  .........................................................  XOR
47    CONTINUE
      IF (ISTACK.GE.2) THEN
         IB = NINT(STACK(ISTACK))
         ISTACK = ISTACK - 1
         IA = NINT(STACK(ISTACK))
         IF (IA.NE.0 .AND. IB.EQ.0) THEN
            STACK(ISTACK) = IA
         ELSE IF (IA.EQ.0 .AND. IB.NE.0) THEN
            STACK(ISTACK) = IB
         ELSE
            STACK(ISTACK) = 0
         END IF
      ELSE
         CALL MABORT('Stack underflow (AND).','USEMCL')
      END IF
      RETURN
C  .........................................................  ABS
48    CONTINUE
      STACK(ISTACK) = ABS(STACK(ISTACK))
      RETURN
C  .........................................................  LOG10
49    CONTINUE
      STACK(ISTACK) = LOG10(STACK(ISTACK))
      RETURN
C  .........................................................  LOG
50    CONTINUE
      STACK(ISTACK) = LOG(STACK(ISTACK))
      RETURN
C  .........................................................  SIN
51    CONTINUE
      STACK(ISTACK) = SIN(STACK(ISTACK))
      RETURN
C  .........................................................  COS
52    CONTINUE
      STACK(ISTACK) = COS(STACK(ISTACK))
      RETURN
C  .........................................................  TAN
53    CONTINUE
      STACK(ISTACK) = TAN(STACK(ISTACK))
      RETURN
C  .........................................................  ASIN
54    CONTINUE
      STACK(ISTACK) = ASIN(STACK(ISTACK))
      RETURN
C  .........................................................  ACOS
55    CONTINUE
      STACK(ISTACK) = ACOS(STACK(ISTACK))
      RETURN
C  .........................................................  ATAN
56    CONTINUE
      STACK(ISTACK) = ATAN(STACK(ISTACK))
      RETURN
C  .........................................................  SINH
57    CONTINUE
      STACK(ISTACK) = SINH(STACK(ISTACK))
      RETURN
C  .........................................................  COSH
58    CONTINUE
      STACK(ISTACK) = COSH(STACK(ISTACK))
      RETURN
C  .........................................................  TANH
59    CONTINUE
      STACK(ISTACK) = TANH(STACK(ISTACK))
      RETURN
C  .........................................................  ASINH
60    CONTINUE
      STACK(ISTACK) = ASINH0(STACK(ISTACK))
      RETURN
C  .........................................................  ACOSH
61    CONTINUE
      XARG = STACK(ISTACK)
      CALL ACOSH0(XARG,RES,MERR)
      IF (MERR.EQ.0) STACK(ISTACK) = RES
      RETURN
C  .........................................................  ATANH
62    CONTINUE
      XARG = STACK(ISTACK)
      CALL ATANH0(XARG,RES,MERR)
      IF (MERR.EQ.0) STACK(ISTACK) = RES
      RETURN
C  .........................................................  MIN
63    CONTINUE
      CALL ININ(NUMBER,III)
      IF (III.NE.0) CALL MABORT('Cannot read MOC (MIN).','USEMCL')
      X1 = STACK(ISTACK)
      DO 203,II=1,NUMBER-1
         ISTACK = ISTACK - 1
         IF (X1.GT.STACK(ISTACK)) X1 = STACK(ISTACK)
203   CONTINUE
      STACK(ISTACK) = X1
      RETURN
C  .........................................................  MAX
64    CONTINUE
      CALL ININ(NUMBER,III)
      IF (III.NE.0) CALL MABORT('Cannot read MOC (MAX).','USEMCL')
      X1 = STACK(ISTACK)
      DO 204,II=1,NUMBER-1
         ISTACK = ISTACK - 1
         IF (X1.LT.STACK(ISTACK)) X1 = STACK(ISTACK)
204   CONTINUE
      STACK(ISTACK) = X1
      RETURN
C  .........................................................  TRUNC
65    CONTINUE
      STACK(ISTACK) = INT(STACK(ISTACK))
      RETURN
C  .........................................................  ROUND
66    CONTINUE
      STACK(ISTACK) = NINT(STACK(ISTACK))
      RETURN
C  .........................................................  MOD
67    CONTINUE
      X2 = STACK(ISTACK)
      ISTACK = ISTACK - 1
      X1 = STACK(ISTACK)
      STACK(ISTACK) = X1-INT(X1/X2)*X2
      RETURN
C  .........................................................  MEAN
68    CONTINUE
      CALL ININ(NUMBER,III)
      IF (III.NE.0) CALL MABORT('Cannot read MOC (MEAN).','USEMCL')
      XX = 0.
      DO 205,I=1,NUMBER
         XX = XX+STACK(ISTACK)
         ISTACK = ISTACK - 1
205   CONTINUE
      ISTACK = ISTACK + 1
      STACK(ISTACK) = XX/NUMBER
      RETURN
C  .........................................................  POPLV
69    CONTINUE
      STACK(NINT(STACK(ISTACK))) = STACK(ISTACK-1)
      ISTACK = ISTACK - 2
      RETURN
C  .........................................................  POPCONT
70    CONTINUE
      STACK(NINT(STACK(ISTACK-1))) = STACK(ISTACK)
      ISTACK = ISTACK - 2
      RETURN
C  .........................................................  X
71    CONTINUE
      IDX = NINT(STACK(ISTACK))
      IF (INDEF(IDX).NE.0) THEN
         STACK(ISTACK) = POINT(IDX)
      ELSE
         IF (WEAK) THEN
            CALL VNAME(IDX,STR,LES)
            WRITE (IUOUT,310) STR(1:LES)
         END IF
         MERR = -2
      END IF
      RETURN
C  .........................................................  R
72    CONTINUE
      STACK(ISTACK) = XRL(NINT(STACK(ISTACK)))
      RETURN
C  .........................................................  L
73    CONTINUE
      STACK(ISTACK) = XLL(NINT(STACK(ISTACK)))
      RETURN
C  .........................................................  STEP
74    CONTINUE
      STACK(ISTACK) = STEP(NINT(STACK(ISTACK)))
      RETURN
C  .........................................................  FIX
75    CONTINUE
      STACK(ISTACK) = IXAT(NINT(STACK(ISTACK)))
      RETURN
C  .........................................................  GRAD
76    CONTINUE
      CALL DERMER(POINT,VAL,XT1,NOGCAL,NGR,0)
      STACK(ISTACK) = XT1(NINT(STACK(ISTACK)))
      RETURN
C  .........................................................  MARG
77    CONTINUE
      STACK(ISTACK) = MARG(NINT(STACK(ISTACK)))
      RETURN
C  .........................................................  TCOUNT
78    CONTINUE
      ISTACK = ISTACK + 1
      STACK(ISTACK) = ITNOCS
      RETURN
C  .........................................................  PCOUNT
79    CONTINUE
      ISTACK = ISTACK + 1
      STACK(ISTACK) = IPARTC
      RETURN
C  .........................................................  DIM
80    CONTINUE
      ISTACK = ISTACK + 1
      STACK(ISTACK) = NOD
      RETURN
C  .........................................................  VALUE
81    CONTINUE
      ISTACK = ISTACK + 1
      STACK(ISTACK) = VAL
      RETURN
C  .........................................................  PRECISION
82    CONTINUE
      ISTACK=ISTACK + 1
      STACK(ISTACK) = COMACC
      RETURN
C  .........................................................  SQRT
83    CONTINUE
      XARG = STACK(ISTACK)
      IF (XARG.LT.0.) THEN
         IF (WEAK) THEN
            CALL R2NRTF(XARG,STR,LES)
            WRITE (IUOUT,320) STR(1:LES)
         END IF
         MERR = -2
      ELSE
         STACK(ISTACK) = SQRT(XARG)
      END IF
      RETURN
C  .........................................................  GTCOUNT
84    CONTINUE
      ISTACK = ISTACK + 1
      STACK(ISTACK) = ITDER
      RETURN
C  .........................................................  PROCESS
85    CONTINUE
      IF (PROCES.EQ.'BATCH') THEN
         IPROC = 1
      ELSE IF (PROCES.EQ.'IAF') THEN
         IPROC = 2
      ELSE
         CALL MABORT('Cannot recognize process mode.','USEMCL')
      END IF
      ISTACK = ISTACK + 1
      STACK(ISTACK) = IPROC
      RETURN
C  .........................................................  BLOCKCOPY
122   CONTINUE
      IST = NINT(STACK(ISTACK))-1
      ISTACK = ISTACK-1
      CALL INNOCR(LINE,.TRUE.,III)
      IF (III.NE.0) CALL MABORT('Cannot read MOC (BLOCKCOPY).','USEMCL')
      IF (LINE.EQ.'X') THEN
         DO 911,I=1,NOD
911         STACK(IST+I) = POINT(I)
      ELSE IF (LINE.EQ.'L') THEN
         DO 912,I=1,NOD
912         STACK(IST+I) = XLL(I)
      ELSE IF (LINE.EQ.'R') THEN
         DO 913,I=1,NOD
913         STACK(IST+I) = XRL(I)
      ELSE IF (LINE.EQ.'FIX') THEN
         DO 914,I=1,NOD
914         STACK(IST+I) = IXAT(I)
      ELSE IF (LINE.EQ.'MARG') THEN
         DO 915,I=1,NOD
915         STACK(IST+I) = MARG(I)
      ELSE IF (LINE.EQ.'GRAD') THEN
         CALL DERMER(POINT,VAL,XT1,NOGCAL,NGR,0)
         DO 916,I=1,NOD
916         STACK(IST+I) = XT1(I)
      ELSE IF (LINE.EQ.'STEP') THEN
         DO 917,I=1,NOD
917         STACK(IST+I) = STEP(I)
      ELSE
         CALL MABORT('Incorrect internal array (BLOCKCOPY).',
     &                  'USEMCL')
      END IF
      RETURN
C  .........................................................  GRMS
87    CONTINUE
      CALL DERMER(POINT,VAL,XT1,NOGCAL,NGR,0)
      GSUM = 0.
      DO 127,I=1,NOD
127      GSUM = GSUM+XT1(I)**2
      STACK(ISTACK) = SQRT(GSUM/NOD)
      RETURN
C  .........................................................  TESTMOVE
88    CONTINUE
      CALL ININ(NUMBER,III)
      IF (III.NE.0) CALL MABORT('Cannot read MOC (TESTMOVE).','USEMCL')
      IF (NINT(STACK(ISTACK)).EQ.1) THEN
         ISTACK = ISTACK-1
         CALL MOVIT(NUMBER)
      ELSE IF (NINT(STACK(ISTACK)).EQ.0) THEN
         ISTACK = ISTACK-1
      ELSE
         CALL MABORT('Incorrect Top-of-Stack (TESTMOVE).','USEMCL')
      END IF
      RETURN
C  .........................................................  MOVE
89    CONTINUE
      CALL ININ(NUMBER,III)
      IF (III.NE.0) CALL MABORT('Cannot read MOC (MOVE).','USEMCL')
      CALL MOVIT(NUMBER)
      RETURN
C  .........................................................  LINE
90    CONTINUE
      CALL ININ(NUMBER,III)
      IF (III.NE.0) CALL MABORT('Cannot read MOC (LINE).','USEMCL')
      MCLLNO = NUMBER
      RETURN
C  .........................................................  PUSHA
121   CONTINUE
      CALL ININ(NUMBER,III)
      IF (III.NE.0) CALL MABORT('Cannot read MOC (PUSHA).','USEMCL')
      ISTACK = ISTACK+1
      STACK(ISTACK) = NUMBER+MBASE
      RETURN
C  .........................................................  IOFFSET
120   CONTINUE
      IARG = NINT(STACK(ISTACK))
      ISTACK = ISTACK-1
      IOFF = NINT(STACK(ISTACK))
      STACK(ISTACK) = IOFF+NINT(STACK(INAD+IARG-1))
      RETURN
C  .........................................................  EXP
93    CONTINUE
      STACK(ISTACK) = EXP(STACK(ISTACK))
      RETURN
C  .........................................................  RAN
94    CONTINUE
      STACK(ISTACK) = RANM()
      RETURN
C  .........................................................  GRADNORM
104   CONTINUE
      INO = NINT(STACK(ISTACK))
      CALL DERMER(POINT,VAL,XT1,NOGCAL,NGR,0)
      IF (INO.EQ.-1) THEN
         GSUM = 0.
         DO 124,I=1,NOD
124         GSUM = MAX(GSUM,ABS(XT1(I)))
      ELSE IF (INO.EQ.1) THEN
         GSUM = 0.
         DO 105,I=1,NOD
105         GSUM = GSUM+ABS(XT1(I))
      ELSE IF (INO.EQ.2) THEN
         GSUM = 0.
         DO 106,I=1,NOD
106         GSUM = GSUM+XT1(I)**2
         GSUM = SQRT(GSUM)
      ELSE
         IF (WEAK) THEN
            CALL I2STR(INO,STR,LES)
            WRITE (IUOUT,330) STR(1:LES)
         END IF
         MERR = -2
         RETURN
      END IF
      STACK(ISTACK) = GSUM
      RETURN
C  .........................................................  FLAG
107   CONTINUE
      STACK(ISTACK) = FLAG(NINT(STACK(ISTACK)))
      RETURN
C  .........................................................  GPCOUNT
108   CONTINUE
      ISTACK = ISTACK + 1
      STACK(ISTACK) = IPDER
      RETURN
C  .........................................................  PUSHMB
109   CONTINUE
      ISTACK = ISTACK+1
      STACK(ISTACK) = MBASE
      RETURN
C  .........................................................  POPMB
110   CONTINUE
      MBASE = NINT(STACK(ISTACK))
      ISTACK = ISTACK-1
      RETURN
C  .........................................................  IPOP
118   CONTINUE
      CALL ININ(NUMBER,III)
      IF (III.NE.0) CALL MABORT('Cannot read MOC (IPOP).','USEMCL')
      STACK(NINT(STACK(INAD+NUMBER-1))) = STACK(ISTACK)
      ISTACK = ISTACK-1
      RETURN
C  .........................................................  ALLOC
112   CONTINUE
      NUMBER = NINT(STACK(ISTACK))
      ISTACK = ISTACK-1
      ITOP = ISTACK
C
C  Allow 500 positions additional stack space. (Should be adequate
C  for most expressions.
      NEED = ISTACK+NUMBER+500
      IF (NEED.GT.LSTA) THEN
         IF (WEAK) THEN
            CALL I2STR(LSTA,STR,LES)
            CALL I2STR(NEED,LINE,LEL)
            WRITE (IUOUT,340) STR(1:LES), LINE(1:LEL)
         END IF
         MERR = -2
      ELSE
         ISTACK = ISTACK+NUMBER
         ISTACK = ISTACK+1
         STACK(ISTACK) = ITOP
      END IF
      RETURN
C  .........................................................  DEALLOC
113   CONTINUE
      NUMBER = NINT(STACK(ISTACK))
      ISTACK = ISTACK-1
      ISTACK = ISTACK-NUMBER
      RETURN
C  .........................................................  PUSHINAD
114   CONTINUE
      ISTACK = ISTACK+1
      STACK(ISTACK) = INAD
      RETURN
C  .........................................................  POPINAD
115   CONTINUE
      INAD = NINT(STACK(ISTACK))
      ISTACK = ISTACK-1
      RETURN
C  .........................................................  SUBRETURN
116   CONTINUE
      MPCOLD = NINT(STACK(ISTACK))
      ISTACK = ISTACK-1
      NUMBER = MPCOLD-MPC
      IF (NUMBER.LT.0) NUMBER = NUMBER+1
      CALL MOVIT(NUMBER)
      RETURN
C  .........................................................  IPUSH
117   CONTINUE
      CALL ININ(NUMBER,III)
      IF (III.NE.0) CALL MABORT('Cannot read MOC (IPUSH).','USEMCL')
      ISTACK = ISTACK+1
      STACK(ISTACK) = STACK(NINT(STACK(INAD+NUMBER-1)))
      RETURN
C  .........................................................  TERMS
119   CONTINUE
      ISTACK = ISTACK+1
      STACK(ISTACK) = NOF
      RETURN
C  .........................................................  TERM
125   CONTINUE
      IF (ITNOCS.EQ.0) THEN
         TR = 0.
      ELSE IF (IFFORM.EQ.1) THEN
         TR = VECLS(NINT(STACK(ISTACK)))
      ELSE
         TR = 0.
      END IF
      STACK(ISTACK) = TR
      RETURN
C  .........................................................  FUNMODE
126   CONTINUE
      ISTACK = ISTACK+1
      STACK(ISTACK) = IFFORM
      RETURN
C  .........................................................  NFLAGS
128   CONTINUE
      ISTACK = ISTACK+1
      STACK(ISTACK) = MAXFLA
      RETURN
C  .........................................................  NCFLAGS
129   CONTINUE
      ISTACK = ISTACK+1
      STACK(ISTACK) = MAXCFL
      RETURN
C  .........................................................  JTCOUNT
130   CONTINUE
      ISTACK = ISTACK+1
      STACK(ISTACK) = ITJAC
      RETURN
C  .........................................................  JPCOUNT
131   CONTINUE
      ISTACK = ISTACK+1
      STACK(ISTACK) = IPJAC
      RETURN
C  .........................................................  JACOMO
132   CONTINUE
      IF (JACOMO.EQ.'JANAL') THEN
         IJA = 1
      ELSE IF (JACOMO.EQ.'JNUMER') THEN
         IJA = 2
      ELSE
         CALL MABORT('Cannot recognize Jacobian mode.','USEMCL')
      END IF
      ISTACK = ISTACK+1
      STACK(ISTACK) = IJA
      RETURN
C  ---------------------------------------------------------------------
C
200   FORMAT (/$(FC)'Error: "',A,'" is not a valid interface command.')
210   FORMAT (/$(FC)'Error: Zero cannot be raised to a non-positive ',
     &        'power.')
220   FORMAT (/$(FC)'Error: Negative numbers cannot be raised to a ',
     &        'non-integer power.')
230   FORMAT (/$(FC)'... Pause ...')
240   FORMAT (/$(FC)'Error: End-of-File encountered in file "',A,'".')
250   FORMAT (/$(FC)'Error: End-of-File encountered.')
260   FORMAT (/$(FC)'Error: Error while reading file "',A,'".')
270   FORMAT (/$(FC)'Error: Error while reading an input file.')
280   FORMAT (/$(FC)'Error: Incorrect norm: ',A,'.')
290   FORMAT (/$(FC)'Error: Array boundary violation.')
300   FORMAT (/$(FC)'The random number generator has been ',
     &       'initialized.')
310   FORMAT (/$(FC)'Error: Variable ',A,' has not been initialized.')
320   FORMAT (/$(FC)'Error: Negative argument in SQRT: ',A)
330   FORMAT (/$(FC)'Error: The gradient norm ',A,' is incorrect.')
340   FORMAT (/$(FC)'Error: There is not enough MCL-memory to run ',
     &        'this MCL program.'
     &        /$(FC)'       Allocated memory: ',A
     &        /$(FC)'       Required memory : ',A)
      END
C  ---------------------------------------------------------------------
      SUBROUTINE FORWAR ( NUMBER )
C  ---------------------------------------------------------------------
C
C  Description:
C    This routine skips NUMBER-2 lines from the Merlin input file.
C
C  Input arguments:
C    NUMBER     Specifies the number of lines to be skipped.
C
C  ---------------------------------------------------------------------
C  Local variables:
      CHARACTER CHA
C  ---------------------------------------------------------------------
C
      DO 10,I=1,NUMBER-1
         CALL INNOCR(CHA,.FALSE.,III)
         IF (III.NE.0) CALL MABORT('Cannot read MOC (FORWAR).',
     &                                'USEMCL')
10    CONTINUE
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE BACWAR ( NUMBER, INFILE )
C  ---------------------------------------------------------------------
C
C  Description:
C    This routine backspaces ABS(NUMBER)+2 lines, the file connected
C    to the unit specified by the INFILE parameter.
C
C  Input arguments:
C    NUMBER     Specifies the number of lines to backspace.
C    INFILE     Unit number of the file to backspace.
C
C  ---------------------------------------------------------------------
C
      DO 10,I=1,ABS(NUMBER)+2
         BACKSPACE INFILE
         CALL DECMPC
10    CONTINUE
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE TREBUF ( BUFFER, IBUF, LENBUF, CARRY, L, TEMPO, IUOUT )
C  ---------------------------------------------------------------------
C
C  Description:
C    This routine treats the output buffer. Adds to the buffer the
C    contents of CARRY(1:L) in the current record number IBUF, only if
C    it fits. If it does not fit, fills the rest of the record with
C    blanks and stores CARRY(1:L) in the next buffer record. If it
C    happens that the buffer is filled completely, the buffer is
C    automatically flushed to unit IUOUT, and CARRY(1:L) is placed at
C    the top of the buffer.
C  Tempo is the current buffer record.
C
C  Input arguments:
C    CARRY      Character variable to be added to the buffer.
C    L          Effective length of CARRY.
C    IUOUT      The output unit.
C
C  Input / Output arguments:
C    BUFFER     Character array. This is the output buffer.
C    IBUF       The current buffer record.
C    LENBUF     Effective length of TEMPO.
C    TEMPO      The current buffer line.
C
C  ---------------------------------------------------------------------
      IMPLICIT $(TYPE) (A-H,O-Z)
C  Arguments:
      PARAMETER ( MR = 80, MB = 20 )
      CHARACTER*(*) BUFFER(MB), TEMPO, CARRY
C  ---------------------------------------------------------------------
C  Local variables:
      LOGICAL CHECKR, CHECKB
C  ---------------------------------------------------------------------
C
C  Check if TEMPO can hold CARRY.
      CHECKR = ( LENBUF + L ) .LE. MR
C
C  If so, add CARRY to TEMPO and update the filled TEMPO positions.
      IF (CHECKR) THEN
         TEMPO(LENBUF + 1 : LENBUF + L ) = CARRY(1:L)
         LENBUF = LENBUF + L
C
C  Since we use LENBUF+1 above, protect the case when the value of
C  LENBUF = MR (the maximum record length), by decrementing LENBUF by 1.
C  This does not create other problems, because L must be .GE. 2
         IF (LENBUF .EQ. MR ) LENBUF = LENBUF-1
         RETURN
      END IF
C
C  Since TEMPO can not hold CARRY(1:L) fill the rest TEMPO-positions
C  with blanks. Fill the current buffer record with TEMPO.
      TEMPO(LENBUF + 1 : MR) = ' '
      BUFFER(IBUF) = TEMPO
      TEMPO(1:) = ' '
C
C  Check if this is the last record of the buffer.
      CHECKB = IBUF .EQ. MB
      IF (CHECKB) THEN
C
C  Flush the whole buffer to unit IUOUT, update the buffer record to 1
C  add CARRY to TEMPO and update its length by setting it equal to L.
         CALL FLUSHB(BUFFER,IBUF,IUOUT)
         TEMPO(1:L) = CARRY(1:L)
         TEMPO(L+1:) = ' '
         IBUF = 1
         LENBUF = L
         RETURN
      END IF
C
C  If not the last record process normally:
      IBUF = IBUF + 1
      TEMPO(1:L) = CARRY(1:L)
      TEMPO(L+1:) = ' '
      LENBUF = L
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE FLUSHB ( BUFFER, IBUF, IUOUT )
C  ---------------------------------------------------------------------
C
C  Description:
C    This routine flushes the contents of the character array BUFFER
C    to the file connected to the unit IUOUT, and resets the buffer
C    length to 1.
C
C  Input arguments:
C    BUFFER     The MCL output buffer.
C    IUOUT      The contents of BUFFER are written to this unit.
C
C  Input / Output arguments:
C    IBUF       Number of records in the buffer.
C
C  ---------------------------------------------------------------------
      IMPLICIT $(TYPE) (A-H,O-Z)
C  Arguments:
      PARAMETER ( MR = 80, MB = 20 )
      CHARACTER*(MR) BUFFER(MB)
C  ---------------------------------------------------------------------
C
      DO 10,I=1,IBUF
         LE = LENGTH(BUFFER(I))
         IF (LE.EQ.0) LE = 1
         WRITE (IUOUT,20) BUFFER(I)(1:LE)
10    CONTINUE
      IBUF = 1
C
20    FORMAT ($(FC)A)
      END
C  ---------------------------------------------------------------------
      SUBROUTINE ACOSH0 ( X, RES, MERR )
C  ---------------------------------------------------------------------
C
C  Description:
C    Calculates the hyperbolic arc-cosine of X. The positive value is
C    always returned.
C
C  Input arguments:
C    X          Argument to ACOSH.
C
C  Output arguments:
C    RES        The hyperbolic arc-cosine of X.
C
C  Input / Output arguments:
C    MERR       Error indicator.
C
C  ---------------------------------------------------------------------
      IMPLICIT $(TYPE) (A-H,O-Z)
C  Common blocks:
      LOGICAL STRO, WEAK
      COMMON /PRILOG/ STRO, WEAK
      COMMON /UNITS/ IUINP, IUOUT
C  ---------------------------------------------------------------------
C  Local variables:
      CHARACTER*30 STR
C  ---------------------------------------------------------------------
C
      IF (X.GE.1) THEN
         RES = LOG( X + SQRT(X*X - 1.) )
      ELSE
         IF (WEAK) THEN
            CALL R2NRTF(X,STR,LES)
            WRITE (IUOUT,10) STR(1:LES)
         END IF
         MERR = -2
      END IF
C
10    FORMAT (/$(FC)'Error: Incorrect argument in ACOSH: ',A)
      END
C  ---------------------------------------------------------------------
      FUNCTION ASINH0 ( X )
C  ---------------------------------------------------------------------
C
C  Description:
C    Calculates the hyperbolic arc-sine of X.
C
C  Input arguments:
C    X          Argument to ASINH.
C
C  ---------------------------------------------------------------------
      IMPLICIT $(TYPE) (A-H,O-Z)
C
      ASINH0 = LOG( X + SQRT( X*X + 1.) )
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE ATANH0 ( X, RES, MERR )
C  ---------------------------------------------------------------------
C
C  Description:
C    Calculates the hyperbolic arc-tangent of X.
C
C  Input arguments:
C    X          Argument to ATANH.
C
C  Output arguments:
C    RES        The hyperbolic arc-tangent of X.
C    MERR       Error indicator.
C
C  ---------------------------------------------------------------------
      IMPLICIT $(TYPE) (A-H,O-Z)
C  Arguments:
      LOGICAL STRO, WEAK
      COMMON /PRILOG/ STRO, WEAK
      COMMON /UNITS/ IUINP, IUOUT
C  ---------------------------------------------------------------------
C  Local variables:
      CHARACTER*30 STR
C  ---------------------------------------------------------------------
C
      IF (ABS(X).LT.1.) THEN
         RES = .5*LOG( (1.+X)/(1.-X) )
      ELSE
         IF (WEAK) THEN
            CALL R2NRTF(X,STR,LES)
            WRITE (IUOUT,10) STR(1:LES)
         END IF
         MERR = -2
      END IF
C
10    FORMAT (/$(FC)'Error: Incorrect argument in ATANH: ',A)
      END
C  ---------------------------------------------------------------------
      SUBROUTINE TOPTWO ( UPS, DOS )
C  ---------------------------------------------------------------------
C
C  Description:
C    Returns the two topmost stack values, decrementing the stack
C    pointer by two.
C
C  Output arguments:
C    UPS        Value at the top of the stack.
C    DOS        Second top of stack.
C
C  ---------------------------------------------------------------------
      IMPLICIT $(TYPE) (A-H,O-Z)
C  Common blocks:
      PARAMETER ( LSTA = $(MCLMEM) )
      COMMON /MCLV1/ STACK(LSTA)
      PARAMETER ( MAXMFI = 20 )
      COMMON /MCLV2/ ISTACK, INFILE, MBASE, INAD, MPC, MCLLNO, LENBUF,
     &               IBUF, MCLUNI(MAXMFI), NUMFI
C  ---------------------------------------------------------------------
C
      UPS = STACK(ISTACK)
      ISTACK = ISTACK-1
      DOS = STACK(ISTACK)
      STACK(ISTACK) = 0.
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE INCMPC
C  ---------------------------------------------------------------------
C
C  Description:
C    Increments the MCL program counter MPC. MPC always contains the
C    current MOC line number.
C
C  ---------------------------------------------------------------------
      IMPLICIT $(TYPE) (A-H,O-Z)
C  Common blocks:
      PARAMETER ( MAXMFI = 20 )
      COMMON /MCLV2/ ISTACK, INFILE, MBASE, INAD, MPC, MCLLNO, LENBUF,
     &               IBUF, MCLUNI(MAXMFI), NUMFI
C  ---------------------------------------------------------------------
C
      MPC = MPC+1
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE DECMPC
C  ---------------------------------------------------------------------
C
C  Description:
C    Decrements the MCL program counter MPC. MPC always contains the
C    current MOC line number.
C
C  ---------------------------------------------------------------------
      IMPLICIT $(TYPE) (A-H,O-Z)
C  Common blocks:
      PARAMETER ( MAXMFI = 20 )
      COMMON /MCLV2/ ISTACK, INFILE, MBASE, INAD, MPC, MCLLNO, LENBUF,
     &               IBUF, MCLUNI(MAXMFI), NUMFI
C  ---------------------------------------------------------------------
C
      MPC = MPC-1
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE MOVIT ( NUMBER )
C  ---------------------------------------------------------------------
C
C  Description:
C    Moves the MOC file forward/backward. If the buffered MOC option
C    is in effect, only the buffer pointers are updated.
C
C  Input arguments:
C    NUMBER     Argument to pass to FORWAR/BACWAR.
C
C  ---------------------------------------------------------------------
      IMPLICIT $(TYPE) (A-H,O-Z)
C  Common blocks:
      PARAMETER ( MAXMFI = 20 )
      COMMON /MCLV2/ ISTACK, INFILE, MBASE, INAD, MPC, MCLLNO, LENBUF,
     &               IBUF, MCLUNI(MAXMFI), NUMFI
      PARAMETER ( MXBUF = $(MCLBUF), MXPTR = MXBUF/5 )
      INTEGER MCLPTR(MXPTR)
      COMMON /MCLBU2/ NPTR, ICPTR, MCLPTR
      COMMON /MODE3/ MODEV, IFFORM, INMCL
C  ---------------------------------------------------------------------
C
      IF (NUMBER.GT.0) THEN
C  Move forward.
         IF (INMCL.NE.0 .AND. NPTR.NE.0) THEN
C  Buffered MOC.
            ICPTR = ICPTR+NUMBER-1
            MPC = MPC+NUMBER-1
         ELSE
C  Unbuffered MOC.
            CALL FORWAR(NUMBER)
         END IF
      ELSE
C  Move backward.
         IF (INMCL.NE.0 .AND. NPTR.NE.0) THEN
C  Buffered MOC.
            ICPTR = ICPTR-ABS(NUMBER)-2
            MPC = MPC-ABS(NUMBER)-2
         ELSE
C  Unbuffered MOC.
            CALL BACWAR(NUMBER, INFILE )
         END IF
      END IF
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE MCLCLO ( MERR )
C  ---------------------------------------------------------------------
C
C  Description:
C    Closes a file that has been opened from within an MCL program.
C
C  Output arguments:
C    MERR       Error indicator.
C
C  ---------------------------------------------------------------------
C  Common blocks:
      PARAMETER ( MAXMFI = 20 )
      COMMON /MCLV2/ ISTACK, INFILE, MBASE, INAD, MPC, MCLLNO, LENBUF,
     &               IBUF, MCLUNI(MAXMFI), NUMFI
      LOGICAL STRO, WEAK
      COMMON /PRILOG/ STRO, WEAK
      COMMON /UNITS/ IUINP, IUOUT
C  ---------------------------------------------------------------------
C  Local variables:
      LOGICAL OPE
      PARAMETER ( MXFILE = 80 )
      CHARACTER*(MXFILE) FILE
C  ---------------------------------------------------------------------
C
C  Get the filename.
      CALL INNOCR(FILE,.FALSE.,III)
      IF (III.NE.0) CALL MABORT ('Cannot read MOC (MCLCLO).','USEMCL')
      LEF = LENGTH(FILE)
C
C  Is it open ?
      INQUIRE (FILE=FILE,OPENED=OPE,NUMBER=IU,IOSTAT=MERR)
      IF (MERR.NE.0) THEN
         CALL ERRINQ(FILE,IERR)
         RETURN
      END IF
      IF (OPE) THEN
         DO 10,I=1,NUMFI
            IF (IU.EQ.MCLUNI(I)) THEN
               CLOSE (IU)
               DO 20,J=I,NUMFI-1
20                MCLUNI(J) = MCLUNI(J+1)
               NUMFI = NUMFI-1
               IF (STRO) WRITE (IUOUT,30) FILE(1:LEF)
               RETURN
            END IF
10       CONTINUE
C
C  Not opened by MCL.
         IF (STRO) WRITE (IUOUT,40) FILE(1:LEF)
      ELSE
C  File not open.
         IF (STRO) WRITE (IUOUT,50) FILE(1:LEF)
      END IF
C
30    FORMAT (/$(FC)'File "',A,'" has been closed.')
40    FORMAT (/$(FC)'Warning: File "',A,'" not opened by MCL. ',
     &        'Not closed.')
50    FORMAT (/$(FC)'Warning: File "',A,'" is not open.')
      END
C  ---------------------------------------------------------------------
      SUBROUTINE MCOPY ( IU1, IU2, LINE, IS, IE, NIT )
C  ---------------------------------------------------------------------
C
C  Description:
C    IU1 is the unit number of a macro file, positioned at the
C    beginning of a macro. Subroutine MCOPY copies the macro (up to a
C    separator *, or up to EOF) to unit IU2 substituting macro
C    parameters.
C
C  Input arguments:
C    IU1        Unit to copy from.
C    IU2        Unit to copy to.
C    LINE       The command line.
C    IS         IS(I) is the starting character position of item I in
C               LINE.
C    IE         IE(I) is the starting character position of item I in
C               LINE.
C    NIT        Number of items in LINE.
C
C  ---------------------------------------------------------------------
C  Arguments:
      CHARACTER*(*) LINE
      INTEGER IS(*), IE(*)
C  ---------------------------------------------------------------------
C  Local variables:
      PARAMETER ( LINLEN = 80 )
      CHARACTER*(LINLEN) LI, NEW, ARG, NUM
      INTEGER BS(LINLEN/2), BE(LINLEN/2)
C  ---------------------------------------------------------------------
C
C  Rewind the output unit (IU2) only. The input unit (IU1)
C  is assumed to be correctly positioned.
      REWIND IU2
C
1000  READ (IU1,10,END=100) LI
C
C  A star indicates we are in a multi-macro file and the end of the
C  macro has been reached.
      IF (LI.EQ.'*') GOTO 100
C
C  Call the bracket-cracker.
      CALL BCR(LI,NBR,BS,BE)
      NN = 0
      NL = 0
C
C  Loop over all pairs of brackets.
      DO 600,I=1,NBR
         ISTART = BS(I)
         IEND = BE(I)
         L = ISTART-NL-1
         IF (L.NE.0) THEN
            NEW(NN+1:NN+L) = LI(NL+1:ISTART-1)
            NN = NN+L
         END IF
C
C  ARG contains the characters in between the brackets.
         ARG = LI(ISTART+1:IEND-1)
C
C  Is there anything in between the brackets ?
C  If not, do not substitute.
         IF (ARG.EQ.' ') GOTO 700
C
C  Do we need the number of arguments in the command line ?
         IF (ARG.EQ.'#') THEN
            CALL I2STR(NIT-1,NUM,LE1)
            NEW(NN+1:NN+L) = NUM(1:LE1)
            NN = NN+LE1
            GOTO 700
         END IF
C
C  Find out which number is in between the brackets.
         LE = LENGTH(ARG)
         CALL TOINT(ARG,LE,N,IERR)
C
C  An error indicates something wrong is in between the
C  brackets. Do not substitute.
         IF (IERR.NE.0) GOTO 700
C
C  A negative number causes no substitution.
         IF (N.LT.0) GOTO 700
C
C  Do we have that many arguments in the command line ?
C  If not, just omit the pattern "[...]" from the output line.
         N1 = N+1
         IF (N1.GT.NIT) GOTO 700
C
C  N = 0 implies the macro name itself.
         IF (N.EQ.0) THEN
            L = IE(1)-IS(1)
            NEW(NN+1:NN+L) = LINE(IS(1)+1:IE(1))
            NN = NN+L
            GOTO 700
         END IF
C
C  Now substitute the command line argument that corresponds to N.
         N1 = N+1
         L = IE(N1)-IS(N1)+1
         NEW(NN+1:NN+L) = LINE(IS(N1):IE(N1))
         NN = NN+L
C
700      CONTINUE
         NL = IEND
600   CONTINUE
C
      L = 80-NL
      IF (L.NE.0) THEN
         NEW(NN+1:NN+L) = LI(NL+1:80)
         NN = NN+L
      END IF
      WRITE (IU2,10) NEW(1:NN)
      GOTO 1000
C
C  End-of-file encountered in the input unit (IU1).
100   CONTINUE
C
C  Write a CLEAR command to ensure proper macro termination.
      WRITE (IU2,10) 'CLEAR'
C
C  Rewind IU2, so as to be ready for execution.
      REWIND IU2
C
10    FORMAT (A)
      END
C  ---------------------------------------------------------------------
      SUBROUTINE BCR ( LINE, NBR, BS, BE )
C  ---------------------------------------------------------------------
C
C  Description:
C    Subroutine BCR scans a given line for pairs of brackets: [...]
C
C  Input arguments:
C    LINE       The input line.
C
C  Output arguments:
C    NBR        Number of bracket pairs.
C    BS         BS(I) is the position of the left bracket [, for
C               the I-th bracket pair.
C    BE         BE(I) is the position of the right bracket ], for
C               the I-th bracket pair.
C
C  ---------------------------------------------------------------------
C  Arguments:
      CHARACTER*(*) LINE
      INTEGER BS(*), BE(*)
C  ---------------------------------------------------------------------
C  Local variables:
      LOGICAL INBRA
      CHARACTER C
C  ---------------------------------------------------------------------
C
      NBR = 0
      INBRA = .FALSE.
      DO 10,I=1,LEN(LINE)
         C = LINE(I:I)
         IF (C.EQ.'[') THEN
            ISTART = I
            INBRA = .TRUE.
         ELSE IF (C.EQ.']') THEN
            IF (INBRA) THEN
               NBR = NBR+1
               BS(NBR) = ISTART
               BE(NBR) = I
               INBRA = .FALSE.
            END IF
         END IF
10    CONTINUE
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE MEREND ( IOPT )
C  ---------------------------------------------------------------------
C
C  Description:
C    Does whatever necessary before Merlin terminates. It exits all
C    macros, closes all files etc.
C    After return of this routine, Merlin must either stop, or return
C    to the calling program.
C
C  Input arguments:
C    IOPT       IOPT = 0 -> Normal Merlin termination (through STOP,
C                           RETURN or QUIT).
C               IOPT = 1 -> Merlin aborts.
C
C  ---------------------------------------------------------------------
C  Common blocks:
      COMMON /MODE3/ MODEV, IFFORM, INMCL
      LOGICAL STRO, WEAK
      COMMON /PRILOG/ STRO, WEAK
      COMMON /UNITS/ IUINP, IUOUT
C  ---------------------------------------------------------------------
C
C  Are we running MCL ?
      IF (INMCL.NE.0) CALL MCLEND
C
C  Exit from all macros as if a "CLEAR" command had been issued.
      CALL TOPLEV(0)
C
C  Close the Merlin files.
      CALL FCLOSE
C
C  Close the output unit/file.
      IF (STRO) THEN
         IF (IOPT.EQ.0) THEN
            WRITE (IUOUT,10)
         ELSE
            WRITE (IUOUT,20)
         END IF
      END IF
      CALL CLOUT
C
10    FORMAT ($(FC)28('-'),' Merlin run has ended ',28('-'))
20    FORMAT ($(FC)31('-'),' Merlin aborted ',31('-'))
      END
C  ---------------------------------------------------------------------
      SUBROUTINE DASH
C  ---------------------------------------------------------------------
C
C  Description:
C    Prints a horizontal line.
C
C  ---------------------------------------------------------------------
C  Common blocks:
      LOGICAL STRO, WEAK
      COMMON /PRILOG/ STRO, WEAK
      COMMON /UNITS/ IUINP, IUOUT
C  ---------------------------------------------------------------------
C
      IF (STRO) WRITE (IUOUT,10)
C
10    FORMAT ($(FC)79('-'))
      END
C  ---------------------------------------------------------------------
      SUBROUTINE MID1 ( MSG )
C  ---------------------------------------------------------------------
C
C  Description:
C    Prints a message (MSG) centered, on an 80 character line.
C    Three dots (...) are printed at the beginning and end of the
C    message.
C
C  Input arguments:
C    MSG        The message to be printed.
C
C  ---------------------------------------------------------------------
C  Arguments:
      CHARACTER*(*) MSG
C  ---------------------------------------------------------------------
C  Common blocks:
      LOGICAL STRO, WEAK
      COMMON /PRILOG/ STRO, WEAK
      COMMON /UNITS/ IUINP, IUOUT
C  ---------------------------------------------------------------------
C  Local variables:
      CHARACTER C
      CHARACTER LINE*80, THREE*3
C  ---------------------------------------------------------------------
C
      C = '.'
      IF (STRO) THEN
         THREE = C//C//C
         LE = LENGTH(MSG)
         IB = (79-LE)/2
         IBLE = IB+LE
         LINE = ' '
         LINE(IB-4:IB-2) = THREE
         LINE(IB+1:IBLE) = MSG(1:LE)
         LINE(IBLE+3:IBLE+5) = THREE
         WRITE (IUOUT,*)
         WRITE (IUOUT,'(A)') LINE(1:IBLE+5)
      END IF
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE CPRINT ( MSG )
C  ---------------------------------------------------------------------
C
C  Description:
C    Prints a message (MSG) centered, on an 80 character line.
C
C  Input arguments:
C    MSG        The message to be printed.
C
C  ---------------------------------------------------------------------
C  Arguments:
      CHARACTER MSG*(*)
C  ---------------------------------------------------------------------
C  Common blocks:
      LOGICAL STRO, WEAK
      COMMON /PRILOG/ STRO, WEAK
      COMMON /UNITS/ IUINP, IUOUT
C  ---------------------------------------------------------------------
C  Local variables:
      CHARACTER LINE*80
C  ---------------------------------------------------------------------
C
      IF (STRO) THEN
         LE = LENGTH(MSG)
         IB = (79-LE)/2
         IBLE = IB+LE
         LINE = ' '
         LINE(IB+1:IBLE) = MSG(1:LE)
         WRITE (IUOUT,'(A)') LINE(1:IBLE+5)
      END IF
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE VSIM ( CNEED, MAXW, NEED, IERR )
C  ---------------------------------------------------------------------
C
C  Description:
C    Subroutine VSIM is called by routines that use the two dimensional
C    Merlin work array. It checks whether there is adequate storage
C    for the calling routine to execute.
C
C  Input arguments:
C    CNEED      Character variable. A short description of the storage
C               in need. Eg. CNEED = 'NxN'
C    MAXW       Storage positions in the two dimensional Merlin work
C               array.
C    NEED       Storage positions needed by the calling routine.
C
C  Output arguments:
C    IERR       IERR =  0 -> Storage is sufficient.
C               IERR = -2 -> Not enough storage.
C
C  ---------------------------------------------------------------------
C  Arguments:
      CHARACTER*(*) CNEED
C  ---------------------------------------------------------------------
C  Common blocks:
      LOGICAL STRO, WEAK
      COMMON /PRILOG/ STRO, WEAK
      COMMON /UNITS/ IUINP, IUOUT
C  ---------------------------------------------------------------------
C  Local variables:
      CHARACTER*10 COM, STR1, STR2
C  ---------------------------------------------------------------------
C
      IF (MAXW.LT.NEED) THEN
         CALL CMDARG(0,COM,LEC)
         IF (WEAK) THEN
            CALL I2STR(MAXW,STR1,LE1)
            CALL I2STR(NEED,STR2,LE2)
            WRITE (IUOUT,10) COM(1:LEC), CNEED, STR1(1:LE1),
     &                             STR2(1:LE2)
         END IF
         IERR = -2
      ELSE
         IERR = 0
      END IF
C
10    FORMAT (/$(FC)'You cannot use the "',A,
     &       '" command because you did not'
     &       /$(FC)'allocate sufficient storage in the "VERSIM" array.'
     &       /$(FC)'The required memory is: ',A
     &       /$(FC)A,' memory words were allocated.'
     &       /$(FC)A,' memory words are in need.')
      END
C  ---------------------------------------------------------------------
      SUBROUTINE POIMAR ( LABEL )
C  ---------------------------------------------------------------------
C
C  Description:
C    Subroutine POIMAR checks if any variable is out of the margin
C    range. In that case it resets the variable to the closest margin
C    and updates the current function value. If however function
C    evaluation is disabled via a NOEVAL command, the objective
C    function is not called.
C    If a fixed variable is out-of-margin range POIMAR looses it
C    and positions it as described above.
C
C  Output arguments:
C    LABEL      LABEL = 0 -> Point is within the margins.
C               LABEL = 1 -> At least one variable was out of margins,
C                            causing the current point to change.
C                            If function evaluation is enabled (ie. a
C                            NOEVAL command is not in effect) the
C                            objective function is evaluated at the
C                            new point.
C
C  ---------------------------------------------------------------------
      IMPLICIT $(TYPE) (A-H,O-Z)
C  Common blocks:
      PARAMETER ( MXV = $(MXV) )
      COMMON /PARAMS/ POINT(MXV), VAL
      COMMON /DETAIL/ IXAT(MXV), MARG(MXV)
      COMMON /XLIMS/ XLL(MXV), XRL(MXV)
      COMMON /DIMEN/ NOD, NOF
      COMMON /MODE3/ MODEV, IFFORM, INMCL
      COMMON /TOTCAL/ ITNOCS, IPARTC, ITDER, IPDER, ITHESS, IPHESS,
     &                ITJAC, IPJAC
      LOGICAL STRO, WEAK
      COMMON /PRILOG/ STRO, WEAK
      COMMON /UNITS/ IUINP, IUOUT
C  ---------------------------------------------------------------------
C  Local variables:
      CHARACTER*20 STR
C  ---------------------------------------------------------------------
C
      LABEL = 0
      CALL UNDEF(IGO,0)
      IF (IGO.EQ.0) RETURN
C
      DO 10,I=1,NOD
         MA = MARG(I)
         IF (MA.EQ.-1) THEN
C  Variable has only a left margin.
            IF (POINT(I).LT.XLL(I)) THEN
               LABEL = 1
               POINT(I) = XLL(I)
               IXAT(I) = 1
               IF (STRO) THEN
                  CALL VNAME(I,STR,LES)
                  WRITE (IUOUT,3) STR(1:LES)
               END IF
            END IF
         ELSE IF (MA.EQ.1) THEN
C  Variable has only a lright margin.
            IF (POINT(I).GT.XRL(I)) THEN
               LABEL = 1
               POINT(I) = XRL(I)
               IXAT(I) = 1
               IF (STRO) THEN
                  CALL VNAME(I,STR,LES)
                  WRITE (IUOUT,3) STR(1:LES)
               END IF
            END IF
         ELSE IF (MA.EQ.2) THEN
C  Variable has both margins.
            IF (POINT(I).LT.XLL(I)) THEN
               LABEL = 1
               POINT(I) = XLL(I)
               IXAT(I) = 1
               IF (STRO) THEN
                  CALL VNAME(I,STR,LES)
                  WRITE (IUOUT,3) STR(1:LES)
               END IF
            ELSE IF (POINT(I).GT.XRL(I)) THEN
               LABEL = 1
               POINT(I) = XRL(I)
               IXAT(I) = 1
               IF (STRO) THEN
                  CALL VNAME(I,STR,LES)
                  WRITE (IUOUT,3) STR(1:LES)
               END IF
            END IF
         END IF
10    CONTINUE
      IF (LABEL.EQ.1 .AND. MODEV.EQ.1) VAL = ACSQ(POINT,NOD)
C
3     FORMAT (/$(FC)'Variable ',A,' was out-of-margin range.')
      END
C  ---------------------------------------------------------------------
      SUBROUTINE PROMPT
C  ---------------------------------------------------------------------
C
C  Description:
C    This subroutine will print:
C      - The merlin prompt.
C      - An informative message for uninitialized variables.
C      - An informative message if function evaluation has been turned
C        off by a NOEVAL command.
C
C    Messages are issued only in the top input level. (Not inside
C    macros or MCL programs.)
C
C  ---------------------------------------------------------------------
C  Common blocks:
      PARAMETER ( MAXINL = 20 )
      INTEGER BACKIN(MAXINL), WHATIN(MAXINL), REDIN(MAXINL)
      COMMON /INNER1/ LEVIN, BACKIN, WHATIN, REDIN
      COMMON /MODE3/ MODEV, IFFORM, INMCL
      COMMON /TOTCAL/ ITNOCS, IPARTC, ITDER, IPDER, ITHESS, IPHESS,
     &                ITJAC, IPJAC
      CHARACTER TAB, ESC
      COMMON /SPECHA/ TAB, ESC
      LOGICAL STRO, WEAK
      COMMON /PRILOG/ STRO, WEAK
      COMMON /UNITS/ IUINP, IUOUT
C  ---------------------------------------------------------------------
C
      IF (LEVIN.EQ.0 .AND. STRO) THEN
         IF (MODEV.EQ.0) THEN
C  Issue message for function evaluation.
            WRITE (IUOUT,10)
         ELSE IF (ITNOCS.EQ.0) THEN
C  Issue message for initialization of variables.
            WRITE (IUOUT,20)
         END IF
C  Unix systems treat the backslash ( ascii char. #92 ) as an escape
C  sequence. We cannot just include it in a format statement.
         WRITE (IUOUT,30) ESC, ESC, ESC, ESC, ESC, ESC
      END IF
C
10    FORMAT (/$(FC)'*** Warning: Function evaluation has been ',
     &        'turned off ***')
20    FORMAT (/$(FC)' ....   W A R N I N G   ....'/
     &              '...  Initialize variables  ...')
30    FORMAT (/$(FC)4X,6('/',A),5X,'Merlin is at your command !!! ')
      END
C  ---------------------------------------------------------------------
      SUBROUTINE RANLUX(RVEC,LENV)
C  ---------------------------------------------------------------------
C         Subtract-and-borrow random number generator proposed by
C         Marsaglia and Zaman, implemented by F. James with the name
C         RCARRY in 1991, and later improved by Martin Luescher
C         in 1993 to produce "Luxury Pseudorandom Numbers".
C     Fortran 77 coded by F. James, 1993
C
C       references:
C  M. Luscher, Computer Physics Communications  79 (1994) 100
C  F. James, Computer Physics Communications 79 (1994) 111
C
C   LUXURY LEVELS.
C   ------ ------      The available luxury levels are:
C
C  level 0  (p=24): equivalent to the original RCARRY of Marsaglia
C           and Zaman, very long period, but fails many tests.
C  level 1  (p=48): considerable improvement in quality over level 0,
C           now passes the gap test, but still fails spectral test.
C  level 2  (p=97): passes all known tests, but theoretically still
C           defective.
C  level 3  (p=223): DEFAULT VALUE.  Any theoretically possible
C           correlations have very small chance of being observed.
C  level 4  (p=389): highest possible luxury, all 24 bits chaotic.
C
C!!! ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
C!!!  Calling sequences for RANLUX:                                  ++
C!!!      CALL RANLUX (RVEC, LEN)   returns a vector RVEC of LEN     ++
C!!!                   32-bit random floating point numbers between  ++
C!!!                   zero (not included) and one (also not incl.). ++
C!!!      CALL RLUXGO(LUX,INT,K1,K2) initializes the generator from  ++
C!!!               one 32-bit integer INT and sets Luxury Level LUX  ++
C!!!               which is integer between zero and MAXLEV, or if   ++
C!!!               LUX .GT. 24, it sets p=LUX directly.  K1 and K2   ++
C!!!               should be set to zero unless restarting at a break++
C!!!               point given by output of RLUXAT (see RLUXAT).     ++
C!!!      CALL RLUXAT(LUX,INT,K1,K2) gets the values of four integers++
C!!!               which can be used to restart the RANLUX generator ++
C!!!               at the current point by calling RLUXGO.  K1 and K2++
C!!!               specify how many numbers were generated since the ++
C!!!               initialization with LUX and INT.  The restarting  ++
C!!!               skips over  K1+K2*E9   numbers, so it can be long.++
C!!!   A more efficient but less convenient way of restarting is by: ++
C!!!      CALL RLUXIN(ISVEC)    restarts the generator from vector   ++
C!!!                   ISVEC of 25 32-bit integers (see RLUXUT)      ++
C!!!      CALL RLUXUT(ISVEC)    outputs the current values of the 25 ++
C!!!                 32-bit integer seeds, to be used for restarting ++
C!!!      ISVEC must be dimensioned 25 in the calling program        ++
C!!! ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
      IMPLICIT $(TYPE) (A-H,O-Z)
      DIMENSION RVEC(LENV)
      DIMENSION SEEDS(24), ISEEDS(24), ISDEXT(25)
      PARAMETER (MAXLEV=4, LXDFLT=3)
      DIMENSION NDSKIP(0:MAXLEV)
      DIMENSION NEXT(24)
      PARAMETER (TWOP12=4096., IGIGA=1000000000,JSDFLT=314159265)
      PARAMETER (ITWO24=2**24, ICONS=2147483563)
      SAVE NOTYET, I24, J24, CARRY, SEEDS, TWOM24, TWOM12, LUXLEV
      SAVE NSKIP, NDSKIP, IN24, NEXT, KOUNT, MKOUNT, INSEED
      INTEGER LUXLEV
      LOGICAL NOTYET
      DATA NOTYET, LUXLEV, IN24, KOUNT, MKOUNT /.TRUE., LXDFLT, 0,0,0/
      DATA I24,J24,CARRY/24,10,0./
C                               default
C  Luxury Level   0     1     2   *3*    4
      DATA NDSKIP/0,   24,   73,  199,  365 /
Corresponds to p=24    48    97   223   389
C     time factor 1     2     3     6    10   on slow workstation
C                 1    1.5    2     3     5   on fast mainframe
C
C  NOTYET is .TRUE. if no initialization has been performed yet.
C              Default Initialization by Multiplicative Congruential
      IF (NOTYET) THEN
         NOTYET = .FALSE.
         JSEED = JSDFLT
         INSEED = JSEED
CC         WRITE(6,'(A,I12)') ' RANLUX DEFAULT INITIALIZATION: ',JSEED
         LUXLEV = LXDFLT
         NSKIP = NDSKIP(LUXLEV)
         IN24 = 0
         KOUNT = 0
         MKOUNT = 0
CC         LP = NSKIP + 24
CC         WRITE(6,'(A,I2,A,I4)')  ' RANLUX DEFAULT LUXURY LEVEL =  ',
CC     +        LUXLEV,'      p =',LP
            TWOM24 = 1.
         DO 25 I= 1, 24
            TWOM24 = TWOM24 * 0.5
         K = JSEED/53668
         JSEED = 40014*(JSEED-K*53668) -K*12211
         IF (JSEED .LT. 0)  JSEED = JSEED+ICONS
         ISEEDS(I) = MOD(JSEED,ITWO24)
   25    CONTINUE
         TWOM12 = TWOM24 * 4096.
         DO 50 I= 1,24
         SEEDS(I) = REAL(ISEEDS(I))*TWOM24
         NEXT(I) = I-1
   50    CONTINUE
         NEXT(1) = 24
         I24 = 24
         J24 = 10
         CARRY = 0.
         IF (SEEDS(24) .EQ. 0.) CARRY = TWOM24
      ENDIF
C
C          The Generator proper: "Subtract-with-borrow",
C          as proposed by Marsaglia and Zaman,
C          Florida State University, March, 1989
C
      DO 100 IVEC= 1, LENV
      UNI = SEEDS(J24) - SEEDS(I24) - CARRY
      IF (UNI .LT. 0.)  THEN
         UNI = UNI + 1.0
         CARRY = TWOM24
      ELSE
         CARRY = 0.
      ENDIF
      SEEDS(I24) = UNI
      I24 = NEXT(I24)
      J24 = NEXT(J24)
      RVEC(IVEC) = UNI
C  small numbers (with less than 12 "significant" bits) are "padded".
      IF (UNI .LT. TWOM12)  THEN
         RVEC(IVEC) = RVEC(IVEC) + TWOM24*SEEDS(J24)
C        and zero is forbidden in case someone takes a logarithm
         IF (RVEC(IVEC) .EQ. 0.)  RVEC(IVEC) = TWOM24*TWOM24
      ENDIF
C        Skipping to luxury.  As proposed by Martin Luscher.
      IN24 = IN24 + 1
      IF (IN24 .EQ. 24)  THEN
         IN24 = 0
         KOUNT = KOUNT + NSKIP
         DO 90 ISK= 1, NSKIP
         UNI = SEEDS(J24) - SEEDS(I24) - CARRY
         IF (UNI .LT. 0.)  THEN
            UNI = UNI + 1.0
            CARRY = TWOM24
         ELSE
            CARRY = 0.
         ENDIF
         SEEDS(I24) = UNI
         I24 = NEXT(I24)
         J24 = NEXT(J24)
   90    CONTINUE
      ENDIF
  100 CONTINUE
      KOUNT = KOUNT + LENV
      IF (KOUNT .GE. IGIGA)  THEN
         MKOUNT = MKOUNT + 1
         KOUNT = KOUNT - IGIGA
      ENDIF
      RETURN
C
C           Entry to input and float integer seeds from previous run
      ENTRY RLUXIN(ISDEXT)
*     IF block added by Phillip Helbig after correpondence with James
      IF (NOTYET) THEN
CC        WRITE(6,'(A)')  ' PROPER RESULTS ONLY WITH INITIALISATION FROM
CC    $25 INTEGERS OBTAINED WITH RLUXUT'
         NOTYET = .FALSE.
      ENDIF
         TWOM24 = 1.
         DO 195 I= 1, 24
         NEXT(I) = I-1
  195    TWOM24 = TWOM24 * 0.5
         NEXT(1) = 24
         TWOM12 = TWOM24 * 4096.
CC    WRITE(6,'(A)') ' FULL INITIALIZATION OF RANLUX WITH 25 INTEGERS:'
CC    WRITE(6,'(5X,5I12)') ISDEXT
      DO 200 I= 1, 24
      SEEDS(I) = REAL(ISDEXT(I))*TWOM24
  200 CONTINUE
      CARRY = 0.
      IF (ISDEXT(25) .LT. 0)  CARRY = TWOM24
      ISD = IABS(ISDEXT(25))
      I24 = MOD(ISD,100)
      ISD = ISD/100
      J24 = MOD(ISD,100)
      ISD = ISD/100
      IN24 = MOD(ISD,100)
      ISD = ISD/100
      LUXLEV = ISD
        IF (LUXLEV .LE. MAXLEV) THEN
          NSKIP = NDSKIP(LUXLEV)
CC        WRITE (6,'(A,I2)') ' RANLUX LUXURY LEVEL SET BY RLUXIN TO: ',
CC   +                         LUXLEV
        ELSE  IF (LUXLEV .GE. 24) THEN
          NSKIP = LUXLEV - 24
CC        WRITE (6,'(A,I5)') ' RANLUX P-VALUE SET BY RLUXIN TO:',LUXLEV
        ELSE
          NSKIP = NDSKIP(MAXLEV)
CC        WRITE (6,'(A,I5)') ' RANLUX ILLEGAL LUXURY RLUXIN: ',LUXLEV
          LUXLEV = MAXLEV
        ENDIF
      INSEED = -1
      RETURN
C
C                    Entry to output seeds as integers
      ENTRY RLUXUT(ISDEXT)
      DO 300 I= 1, 24
         ISDEXT(I) = INT(SEEDS(I)*TWOP12*TWOP12)
  300 CONTINUE
      ISDEXT(25) = I24 + 100*J24 + 10000*IN24 + 1000000*LUXLEV
      IF (CARRY .GT. 0.)  ISDEXT(25) = -ISDEXT(25)
      RETURN
C
C                    Entry to output the "convenient" restart point
      ENTRY RLUXAT(LOUT,INOUT,K1,K2)
      LOUT = LUXLEV
      INOUT = INSEED
      K1 = KOUNT
      K2 = MKOUNT
      RETURN
C
C                    Entry to initialize from one or three integers
      ENTRY RLUXGO(LUX,INS,K1,K2)
         IF (LUX .LT. 0) THEN
            LUXLEV = LXDFLT
         ELSE IF (LUX .LE. MAXLEV) THEN
            LUXLEV = LUX
         ELSE IF (LUX .LT. 24 .OR. LUX .GT. 2000) THEN
            LUXLEV = MAXLEV
CC          WRITE (6,'(A,I7)') ' RANLUX ILLEGAL LUXURY RLUXGO: ',LUX
         ELSE
            LUXLEV = LUX
            DO 310 ILX= 0, MAXLEV
              IF (LUX .EQ. NDSKIP(ILX)+24)  LUXLEV = ILX
  310       CONTINUE
         ENDIF
      IF (LUXLEV .LE. MAXLEV)  THEN
         NSKIP = NDSKIP(LUXLEV)
CC       WRITE(6,'(A,I2,A,I4)') ' RANLUX LUXURY LEVEL SET BY RLUXGO :',
CC   +        LUXLEV,'     P=', NSKIP+24
      ELSE
          NSKIP = LUXLEV - 24
CC        WRITE (6,'(A,I5)') ' RANLUX P-VALUE SET BY RLUXGO TO:',LUXLEV
      ENDIF
      IN24 = 0
CC    IF (INS .LT. 0)  WRITE (6,'(A)')
CC   +   ' Illegal initialization by RLUXGO, negative input seed'
      IF (INS .GT. 0)  THEN
        JSEED = INS
CC      WRITE(6,'(A,3I12)') ' RANLUX INITIALIZED BY RLUXGO FROM SEEDS',
CC   +      JSEED, K1,K2
      ELSE
        JSEED = JSDFLT
CC      WRITE(6,'(A)')' RANLUX INITIALIZED BY RLUXGO FROM DEFAULT SEED'
      ENDIF
      INSEED = JSEED
      NOTYET = .FALSE.
      TWOM24 = 1.
         DO 325 I= 1, 24
           TWOM24 = TWOM24 * 0.5
         K = JSEED/53668
         JSEED = 40014*(JSEED-K*53668) -K*12211
         IF (JSEED .LT. 0)  JSEED = JSEED+ICONS
         ISEEDS(I) = MOD(JSEED,ITWO24)
  325    CONTINUE
      TWOM12 = TWOM24 * 4096.
         DO 350 I= 1,24
         SEEDS(I) = REAL(ISEEDS(I))*TWOM24
         NEXT(I) = I-1
  350    CONTINUE
      NEXT(1) = 24
      I24 = 24
      J24 = 10
      CARRY = 0.
      IF (SEEDS(24) .EQ. 0.) CARRY = TWOM24
C        If restarting at a break point, skip K1 + IGIGA*K2
C        Note that this is the number of numbers delivered to
C        the user PLUS the number skipped (if luxury .GT. 0).
      KOUNT = K1
      MKOUNT = K2
      IF (K1+K2 .NE. 0)  THEN
        DO 500 IOUTER= 1, K2+1
          INNER = IGIGA
          IF (IOUTER .EQ. K2+1)  INNER = K1
          DO 450 ISK= 1, INNER
            UNI = SEEDS(J24) - SEEDS(I24) - CARRY
            IF (UNI .LT. 0.)  THEN
               UNI = UNI + 1.0
               CARRY = TWOM24
            ELSE
               CARRY = 0.
            ENDIF
            SEEDS(I24) = UNI
            I24 = NEXT(I24)
            J24 = NEXT(J24)
  450     CONTINUE
  500   CONTINUE
C         Get the right value of IN24 by direct calculation
        IN24 = MOD(KOUNT, NSKIP+24)
        IF (MKOUNT .GT. 0)  THEN
           IZIP = MOD(IGIGA, NSKIP+24)
           IZIP2 = MKOUNT*IZIP + IN24
           IN24 = MOD(IZIP2, NSKIP+24)
        ENDIF
C       Now IN24 had better be between zero and 23 inclusive
        IF (IN24 .GT. 23) THEN
CC         WRITE (6,'(A/A,3I11,A,I5)')
CC   +    '  Error in RESTARTING with RLUXGO:','  The values', INS,
CC   +     K1, K2, ' cannot occur at luxury level', LUXLEV
           IN24 = 0
        ENDIF
      ENDIF
      RETURN
      END
C  ---------------------------------------------------------------------
      FUNCTION RANM ( )
C  ---------------------------------------------------------------------
C
C  Description:
C    Function RANM returns a random number, uniformly distributed
C    in the interval (0,1). It uses the RANLUX random number generator
C    due to  Luscher and James. For details see:
C      Luscher M., Comput. Phys. Commun. 79 (1994) 100.
C      James F., Comput. Phys. Commun. 79 (1994) 111.
C
C  ---------------------------------------------------------------------
      IMPLICIT $(TYPE) (A-H,O-Z)
C  Local variables:
      PARAMETER ( LVEC = 1 )
      DIMENSION RVEC(LVEC)
C  ---------------------------------------------------------------------
C
      CALL RANLUX(RVEC,LVEC)
      RANM = RVEC(1)
C
      END
C  ---------------------------------------------------------------------
      FUNCTION RANDI ( A, B )
C  ---------------------------------------------------------------------
C
C  Description:
C   Function RANDI returns a random number between A and B.
C
C  ---------------------------------------------------------------------
      IMPLICIT $(TYPE) (A-H,O-Z)
C
      R = RANM()
      RANDI = A+(B-A)*R
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE READCO ( CMD, MERR, IOPT, FORCE )
C  ---------------------------------------------------------------------
C
C  Description:
C    This is the Merlin command reader. It reads the next command from
C    the Merlin input file and calls the line cracker.
C    Macro (.macro) and MCL program (-moc) invocations, are handled
C    here as well.
C
C  Input arguments:
C    IOPT       IOPT = 0 -> Go ahead and read a line from the
C                           Merlin input file.
C               IOPT = 1 -> Use the contents of the character variable
C                           FORCE, instead of reading a line.
C    FORCE      If IOPT = 1, FORCE will be used in place of the
C               input line.
C
C  Output arguments:
C    CMD        The command to be executed.
C    MERR       Error indicator. Merlin must act according to the
C               returned value:
C               MERR =  0 -> Command read, execute it.
C               MERR =  1 -> Control needs to be transfered to the
C                            Merlin prompt.
C               MERR = -1 -> EOF in the input file.
C               MERR = -2 -> Some other error.
C
C  ---------------------------------------------------------------------
C  Arguments:
      CHARACTER*(*) CMD, FORCE
C  ---------------------------------------------------------------------
C  Common blocks:
      PARAMETER ( LINLEN = 80 )
      INTEGER IS(LINLEN/2), IE(LINLEN/2)
      COMMON /INDAT1/ IS, IE, NIT, LEC
      CHARACTER*(LINLEN) INPLIN, QINPL
      COMMON /INDAT2/ INPLIN, QINPL
      PARAMETER ( MAXINL = 20 )
      INTEGER BACKIN(MAXINL), WHATIN(MAXINL), REDIN(MAXINL)
      COMMON /INNER1/ LEVIN, BACKIN, WHATIN, REDIN
      CHARACTER*10 MCPREV(MAXINL)
      COMMON /INNER2/ MCPREV
      COMMON /CNTRL/ IDISK, MACECH, LIMCLK, IECOMA, IPROEC, MCLUBU
      PARAMETER ( MXFILE = 80 )
      CHARACTER*(MXFILE) PROLO, EPILO, IN, OUT, HFILE, MFILE, PDFILE
      COMMON /CP2/ PROLO, EPILO, IN, OUT, HFILE, MFILE, PDFILE
      LOGICAL STRO, WEAK
      COMMON /PRILOG/ STRO, WEAK
      COMMON /UNITS/ IUINP, IUOUT
C  ---------------------------------------------------------------------
C  Local variables:
      CHARACTER*(LINLEN) MAC, COM
      CHARACTER*(LINLEN) LINE
      CHARACTER*10 STR
      LOGICAL EXI, OPE
C  ---------------------------------------------------------------------
C
C  Read and crack the next command line. Check for EOF.
      KOPT = IOPT
100   CONTINUE
      IF (KOPT.EQ.0) THEN
         CALL INLINE(INPLIN,.TRUE.,.FALSE.,QINPL,IS,IE,NIT,IERR)
         IF (IERR.EQ.-1) GOTO 2
      ELSE IF (KOPT.EQ.1) THEN
         INPLIN = FORCE
         LEFOR = LENGTH(FORCE)
         CALL QLCR(INPLIN,LEFOR,QINPL,IS,IE,NIT,IERR)
      ELSE IF (KOPT.EQ.2) THEN
         INPLIN = LINE
         LL = LENGTH(LINE)
         CALL QLCR(INPLIN,LL,QINPL,IS,IE,NIT,IERR)
      END IF
C
C  Any errors during input or line-cracking ?
      IF (IERR.NE.0) THEN
         IF (WEAK) WRITE (IUOUT,20)
         MERR = -2
         RETURN
      END IF
C
C  Do we have a blank line ?
      IF (NIT.EQ.0) THEN
         MERR = 1
         RETURN
      END IF
C
C  Is this a comment line ?
      IF (QINPL(IS(1):IS(1)).EQ.'%') THEN
         MERR = 1
         RETURN
      END IF
      CMD = QINPL(IS(1):IE(1))
      LEC = IE(1)-IS(1)+1
      COM = CMD
C
C  Did the user specify an MCL program ?
      IF (COM(1:1).EQ.'-') THEN
C
C  Is it just the MCL dash ?
         IF (LEC.EQ.1) THEN
            IF (WEAK) WRITE (IUOUT,30)
            MERR = -2
            RETURN
         END IF
         IDI = INDEX(INPLIN,'-')
         IF (IDI.NE.LEN(INPLIN)) THEN
            LINE = 'RUNMCL '//INPLIN(IDI+1:)
         ELSE
            LINE = 'RUNMCL'
         END IF
         KOPT = 2
         GOTO 100
      END IF
C
C  Did the user specify a macro name ?
      IF (CMD(1:1).NE.'.') THEN
         MERR = 0
         RETURN
      END IF
      COM = CMD
C
C  Is it just the macro dot ?
      IF (LEC.EQ.1) THEN
         IF (WEAK) WRITE (IUOUT,40)
         MERR = -2
         RETURN
      END IF
C
C  Allocate a scratch unit. The macro with parameters substituted
C  is copied in this unit.
      ISC = NUNIT()
      IF (ISC.EQ.-1) THEN
         CALL ERRUAL('<SCRATCH>')
         MERR = -2
         RETURN
      END IF
C
C  Check for recursive macro calls.
      DO 10,I=1,LEVIN
         IF (ISCOMP(COM,MCPREV(I)).NE.0) THEN
            IF (WEAK) WRITE (IUOUT,50)
            MERR = -2
            RETURN
         END IF
10    CONTINUE
C
C  Open the macro file.
      IUMAC = -1
      CALL OPMAC(0,INFO,IUMAC)
      IF (INFO.NE.0) THEN
         ISUC = 0
      ELSE
C
C  Try to locate the macro in the macro file.
         CALL LOCMAC(COM,IUMAC,ISUC)
      END IF
C
C  The macro was not found in the macro file.
      IF (ISUC.EQ.0) THEN
         IF (IUMAC.NE.-1) CLOSE (IUMAC)
C
C  Are we allowed to look on disk files for the macro ?
         IF (IDISK.EQ.0) THEN
            IF (WEAK) WRITE (IUOUT,60) COM(1:LEC)
            MERR = -2
            RETURN
         END IF
         MAC = COM(2:)
         INQUIRE (FILE=MAC,EXIST=EXI,OPENED=OPE,NUMBER=NUMA,IOSTAT=IOE)
         IF (IOE.NE.0) THEN
            CALL ERRINQ(MAC,IOE)
            MERR = -2
            RETURN
         END IF
C
C  Does the file exist ?
         IF (.NOT.EXI) THEN
            IF (WEAK) WRITE (IUOUT,60) COM(1:LEC)
            MERR = -2
            RETURN
         END IF
C  Is the file already open ?
         IF (OPE) THEN
            IF (WEAK) THEN
               CALL I2STR(NUMA,STR,LES)
               WRITE (IUOUT,70) COM(1:LEC), STR(1:LES)
            END IF
            MERR = -2
            RETURN
         END IF
C
C  Allocate unit for the macro file itself.
         IM = NUNIT()
         IF (IM.EQ.-1) THEN
            CALL ERRUAL(COM)
            MERR = -2
            RETURN
         END IF
C
C  Copy the macro to the scratch unit substituting any parameters.
         OPEN (UNIT=IM,FILE=MAC,IOSTAT=IOE)
         IF (IOE.NE.0) THEN
            CALL ERROPE(MAC,IOE)
            MERR = -2
            RETURN
         END IF
         OPEN (ISC,STATUS='SCRATCH',IOSTAT=IOE)
         IF (IOE.NE.0) THEN
            CALL ERROPE('<SCRATCH>',IOE)
            CLOSE (IM)
            MERR = -2
            RETURN
         END IF
         REWIND IM
         CALL MCOPY(IM,ISC,QINPL,IS,IE,NIT)
         CLOSE (IM)
C
C  Update the nesting level.
         CALL INCILE(ISC,COM,LEC,IERR)
         IF (IERR.NE.0) THEN
            CLOSE (ISC,STATUS='DELETE')
            MERR = -2
            RETURN
         END IF
         MERR = 1
C
C  The specified macro has been found in the macro file.
      ELSE IF (ISUC.EQ.1) THEN
C
         OPEN (ISC,STATUS='SCRATCH',IOSTAT=IOE)
         IF (IOE.NE.0) THEN
            CALL ERROPE('<SCRATCH>',IOE)
            CLOSE (IUMAC)
            MERR = -2
            RETURN
         END IF
         CALL MCOPY(IUMAC,ISC,QINPL,IS,IE,NIT)
         CLOSE (IUMAC)
C
C  Update the nesting level.
         CALL INCILE(ISC,COM,LEC,IERR)
         IF (IERR.NE.0) THEN
            CLOSE (ISC,STATUS='DELETE')
            MERR = -2
            RETURN
         END IF
         MERR = 1
C
C  The macro file is corrupt.
      ELSE
         CLOSE (IUMAC)
         IF (WEAK) THEN
            LEM = LENGTH(MFILE)
            WRITE (IUOUT,80) MFILE(1:LEM)
         END IF
         MERR = -2
      END IF
      RETURN
C
C  EOF encountered in input.
2     CONTINUE
      IF (LEVIN.GT.0) THEN
C  EOF in a macro file should never happen, since a CLEAR command is
C  always added at the end. Some braindead commands (GRADCHECK, MACRO
C  etc. can cause EOF in a macro file).
C  EOF in an MCL program would never happen since a FINISH command is
C  always added at the end.
         CALL TOPLEV(1)
         MERR = 1
      ELSE
C  We are reading directly from the merlin input file.
         MERR = -1
      END IF
C
3     FORMAT (A)
20    FORMAT (/$(FC)'Error: Syntax error (probably a missing quote).')
30    FORMAT (/$(FC)'Error: No name follows the initial MCL dash.')
40    FORMAT (/$(FC)'Error: No name follows the initial macro dot.')
50    FORMAT (/$(FC)'Error: Macro recursion is not allowed.')
60    FORMAT (/$(FC)'Error: Macro ',A,' does not exist.')
70    FORMAT (/$(FC)'Error: Macro ',A,' is on a busy file. ',
     &        'UNIT = ',A,'.')
80    FORMAT (/$(FC)'The Merlin macro file "',A,'" is corrupt.')
      END
C  ---------------------------------------------------------------------
      SUBROUTINE TOPLEV ( IABORT )
C  ---------------------------------------------------------------------
C
C  Description:
C    Exits all active macros, returning control to the top level input
C    file.
C
C  Input arguments:
C    IABORT     IABORT = 0 -> Normally exit all macros.
C               IABORT = 1 -> Issue abort messages, according to MACECH.
C
C  ---------------------------------------------------------------------
C  Common blocks:
      PARAMETER ( MAXINL = 20 )
      INTEGER BACKIN(MAXINL), WHATIN(MAXINL), REDIN(MAXINL)
      COMMON /INNER1/ LEVIN, BACKIN, WHATIN, REDIN
      COMMON /CNTRL/ IDISK, MACECH, LIMCLK, IECOMA, IPROEC, MCLUBU
      CHARACTER*10 MCPREV(MAXINL)
      COMMON /INNER2/ MCPREV
C  ---------------------------------------------------------------------
C
      LEVCUR = LEVIN
      DO 10,I=LEVCUR,1,-1
         IF (IABORT.NE.0 .AND. MACECH.NE.0) THEN
C  LEVIN is decremented by SUBROUTINE DECILE.
            LE = LENGTH(MCPREV(LEVIN))
            CALL MID1('Execution of macro '//
     &                     MCPREV(LEVIN)(1:LE)//' has been aborted')
         END IF
         CALL DECILE(2)
10    CONTINUE
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE INCILE ( IIN, COM, LEC, IERR )
C  ---------------------------------------------------------------------
C
C  Description:
C    Increments the input level.
C
C  Input arguments:
C    IIN        The unit number of the new input file.
C    COM        Name of the macro to be executed.
C    LEC        Effective length of COM.
C
C  Output arguments:
C    IERR       Error indicator:
C               IERR = 0  -> Everything ok. The input level has been
C                            incremented.
C               IERR = -1 -> To deep a nesting level.
C
C  ---------------------------------------------------------------------
C  Arguments:
      CHARACTER*(*) COM
C  ---------------------------------------------------------------------
C  Common blocks:
      PARAMETER ( MAXINL = 20 )
      INTEGER BACKIN(MAXINL), WHATIN(MAXINL), REDIN(MAXINL)
      COMMON /INNER1/ LEVIN, BACKIN, WHATIN, REDIN
      CHARACTER*10 MCPREV(MAXINL)
      COMMON /CNTRL/ IDISK, MACECH, LIMCLK, IECOMA, IPROEC, MCLUBU
      LOGICAL STRO, WEAK
      COMMON /PRILOG/ STRO, WEAK
      COMMON /UNITS/ IUINP, IUOUT
      COMMON /INNER2/ MCPREV
C  ---------------------------------------------------------------------
C  Local variables:
      CHARACTER*10 COMW
C  ---------------------------------------------------------------------
C
      LEVIN = LEVIN+1
      IF (LEVIN.GT.MAXINL) THEN
         IF (WEAK) WRITE (IUOUT,10)
         LEVIN = LEVIN-1
         IERR = -1
         RETURN
      END IF
C
      BACKIN(LEVIN) = IUINP
      IF (LEC.NE.0) THEN
         WHATIN(LEVIN) = 0
      ELSE
         WHATIN(LEVIN) = 1
      END IF
      MCPREV(LEVIN) = COM
      IF (LEC.NE.0) THEN
         COMW = COM
         IF (MACECH.NE.0) CALL MID1('Execution of macro '//
     &                COMW(1:LEC)//' has been initiated')
      END IF
      IUINP = IIN
C
C  Handle macro redirection.
      CALL REDIR(IURE,IERR)
      IF (IURE.NE.-1) THEN
         CALL INCOLE(IURE,2,IERR)
         IF (IERR.NE.0) THEN
            CLOSE (IURE)
            LEVIN = LEVIN-1
            IERR = -1
         END IF
         REDIN(LEVIN) = 1
      ELSE
         REDIN(LEVIN) = 0
      END IF
      IERR = 0
C
10    FORMAT (/$(FC)'Error: Maximum nesting level exceeded.')
      END
C  ---------------------------------------------------------------------
      SUBROUTINE DECILE ( IWHAT )
C  ---------------------------------------------------------------------
C
C  Description:
C    Decrements the input level.
C
C  Input arguments:
C    IWHAT      Passed to subroutine DECOLE.
C
C  ---------------------------------------------------------------------
C  Common blocks:
      PARAMETER ( MAXINL = 20 )
      INTEGER BACKIN(MAXINL), WHATIN(MAXINL), REDIN(MAXINL)
      COMMON /INNER1/ LEVIN, BACKIN, WHATIN, REDIN
      COMMON /UNITS/ IUINP, IUOUT
C  ---------------------------------------------------------------------
C
      IF (LEVIN.GT.0) THEN
         IF (WHATIN(LEVIN).EQ.0) THEN
            CLOSE (IUINP,STATUS='DELETE')
         ELSE
            CLOSE (IUINP,STATUS='KEEP')
         END IF
         IUINP = BACKIN(LEVIN)
         IF (REDIN(LEVIN).NE.0) CALL DECOLE(IWHAT)
         LEVIN = LEVIN-1
      ELSE
         CALL MABORT('"levin" not positive.','DECILE')
      END IF
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE INCOLE ( IOUT, IFRO, IERR )
C  ---------------------------------------------------------------------
C
C  Description:
C    Increments the output level.
C
C  Input arguments:
C    IOUT       The unit to receive all subsequent output.
C    IFRO       Reason for incrementing the output level:
C               IFRO = 1 -> Redirection of a command using > or >>
C               IFRO = 2 -> Redirection of a macro using > or >>
C               IFRO = 3 -> Redirection of an MCL program using > or >>
C               IFRO = 4 -> Redirection due to the HIDEOUT command.
C
C  Output arguments:
C    IERR       Error indicator:
C               IERR =  0 -> Output level has been incremented.
C               IERR = -1 -> Too deep a nesting level.
C
C  ---------------------------------------------------------------------
C  Common blocks:
      PARAMETER ( MAXONL = 20 )
      INTEGER BACOUT(MAXONL), FROUT(MAXONL)
      COMMON /OUTER/ LEVOUT, BACOUT, FROUT
      LOGICAL STRO, WEAK
      COMMON /PRILOG/ STRO, WEAK
      COMMON /UNITS/ IUINP, IUOUT
C  ---------------------------------------------------------------------
C
      LEVOUT = LEVOUT+1
      IF (LEVOUT.GT.MAXONL) THEN
         LEVOUT = LEVOUT-1
         IF (WEAK) WRITE (IUOUT,10)
         IERR = -1
         RETURN
      END IF
C
      BACOUT(LEVOUT) = IOUT
      FROUT(LEVOUT) = IFRO
      IUOUT = IOUT
      IERR = 0
C
10    FORMAT (/$(FC)'Maximum nesting level exceeded.')
      END
C  ---------------------------------------------------------------------
      SUBROUTINE DECOLE ( IFRO )
C  ---------------------------------------------------------------------
C
C  Description:
C    Decrements the output level.
C
C  Input arguments:
C    IFRO       Reason for decrementing the output level.
C               See subroutine INCOLE.
C
C  ---------------------------------------------------------------------
C  Common blocks:
      PARAMETER ( MAXONL = 20 )
      INTEGER BACOUT(MAXONL), FROUT(MAXONL)
      COMMON /OUTER/ LEVOUT, BACOUT, FROUT
      COMMON /UNITS/ IUINP, IUOUT
C  ---------------------------------------------------------------------
C
      DO 10,I=LEVOUT,1,-1
         IF (FROUT(I).EQ.IFRO) GOTO 20
10    CONTINUE
      CALL MABORT('Output level description not found.','DECOLE')
C
20    CONTINUE
      CLOSE (BACOUT(I),STATUS='KEEP')
      IF (I.EQ.LEVOUT) THEN
         IUOUT = BACOUT(LEVOUT-1)
      ELSE
         DO 30,K=I,LEVOUT-1
            K1 = K+1
            BACOUT(K) = BACOUT(K1)
            FROUT(K) = FROUT(K1)
30       CONTINUE
      END IF
      LEVOUT = LEVOUT-1
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE REDIR ( NU, IERR )
C  ---------------------------------------------------------------------
C
C  Description:
C    Handles output redirection. Specifically, redirection occurs when
C    one of the symbols > or >> followed by a file name, are present
C    at the end of the command line. This subroutine will recognize
C    the redirection symbols, open and position appropriately the
C    output file.
C
C  Input arguments:
C    NU         Unit number of the file to receive redirected output.
C               It is set to -1 if no redirection takes place.
C
C  Output arguments:
C    IERR       Error indicator.
C
C  ---------------------------------------------------------------------
C  Common blocks:
      PARAMETER ( LINLEN = 80 )
      INTEGER IS(LINLEN/2), IE(LINLEN/2)
      COMMON /INDAT1/ IS, IE, NIT, LEC
      CHARACTER*(LINLEN) INPLIN, QINPL
      COMMON /INDAT2/ INPLIN, QINPL
      LOGICAL STRO, WEAK
      COMMON /PRILOG/ STRO, WEAK
      COMMON /UNITS/ IUINP, IUOUT
C  ---------------------------------------------------------------------
C  Local variables:
      PARAMETER ( MXFILE = 80 )
      CHARACTER*(MXFILE) FILE
      LOGICAL EXI, OPE
C  ---------------------------------------------------------------------
C
C  At least 3 items should be present in the command line for a proper
C  redirection:  "command > file"  or  "command >> file".
      IF (NIT.LT.3) THEN
         NU = -1
         IERR = 0
         RETURN
      END IF
C
C  Check for the redirection symbols.
      N1 = NIT-1
      IF (QINPL(IS(N1):IE(N1)).EQ.'>') THEN
         IRED = 1
      ELSE IF (QINPL(IS(N1):IE(N1)).EQ.'>>') THEN
         IRED = 2
      ELSE
         NU = -1
         IERR = 0
         RETURN
      END IF
C
C  FILE is the file to receive the redirected output.
      FILE = QINPL(IS(NIT):IE(NIT))
      LEF = LENGTH(FILE)
C
C  Remove ">" or ">>" from the command line. Merlin commands cannot tell
C  if redirection takes place.
      NIT = NIT-2
C
C  LEF is the length of the file name.
      IF (LEF.EQ.0) THEN
         IF (WEAK) WRITE (IUOUT,10)
         IERR = -2
         RETURN
      END IF
C
C  Get info for the file.
      INQUIRE (FILE=FILE,EXIST=EXI,OPENED=OPE,NUMBER=IU,IOSTAT=IOE)
      IF (IOE.NE.0) THEN
         CALL ERRINQ(FILE,IOE)
         IERR = -2
         RETURN
      END IF
C
C  Files that are already open cannot receive redirected output.
      IF (OPE) THEN
         CALL ERRBUS(FILE,IU)
         IERR = -2
         RETURN
      END IF
C
C  Allocate unit number for the file.
      NU = NUNIT()
      IF (NU.EQ.-1) THEN
         CALL ERRUAL(FILE)
         IERR = -2
         RETURN
      END IF
C
C  If ">" was specified, safely delete the file.
      IF (EXI .AND. IRED.EQ.1) THEN
         OPEN (NU,FILE=FILE,IOSTAT=IOE)
         IF (IOE.NE.0) THEN
            CALL ERROPE(FILE,IOE)
            IERR = -2
            RETURN
         END IF
         CLOSE (NU,STATUS='DELETE',IOSTAT=IOE)
      END IF
C
C  Open the file.
      IF (IRED.EQ.1) THEN
         OPEN (NU,FILE=FILE,IOSTAT=IOE)
         IF (IOE.EQ.0) REWIND NU
      ELSE
         CALL OPAPPE(NU,FILE,LEF,IOE)
      END IF
      IF (IOE.NE.0) THEN
         CALL ERROPE(FILE,IOE)
         IERR = -2
         RETURN
      END IF
C
10    FORMAT (/$(FC)'Missing file name for output redirection.')
      END
C  ---------------------------------------------------------------------
      SUBROUTINE RTF2R ( STR, X, IERR )
C  ---------------------------------------------------------------------
C
C  Description:
C    This subroutine reads a real number (X) from a string (STR), using
C    the run-time format, constructed by subroutine MACAC.
C
C  Input arguments:
C    STR        The string to read from.
C
C  Output arguments:
C    X          The real.
C    IERR       Error indicator.
C
C  ---------------------------------------------------------------------
      IMPLICIT $(TYPE) (A-H,O-Z)
C  Arguments:
      CHARACTER*(*) STR
C  ---------------------------------------------------------------------
C  Common blocks:
      CHARACTER RTFORM*10, RRFORM*7, REFORM*10
      COMMON /ACCUR2/ RTFORM, RRFORM, REFORM
C  ---------------------------------------------------------------------
C
      READ (STR,RRFORM,ERR=10) X
      IERR = 0
      RETURN
C
10    CONTINUE
      IERR = 1
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE SETLMA ( IND, XM )
C  ---------------------------------------------------------------------
C
C  Description:
C    Sets the the value of the left margin to XM for the variable with
C    index IND.
C    If the left and right margin coincide, the variable is fixed.
C
C  Input arguments:
C    IND        Index of the variable.
C    XM         Value to set as the left margin.
C
C  ---------------------------------------------------------------------
      IMPLICIT $(TYPE) (A-H,O-Z)
C  Common blocks:
      PARAMETER ( MXV = $(MXV) )
      COMMON /PARAMS/ POINT(MXV), VAL
      COMMON /DETAIL/ IXAT(MXV), MARG(MXV)
      COMMON /XLIMS/ XLL(MXV), XRL(MXV)
      COMMON /BIGCOM/ BIGNUM, BLIM, DWARF, DWALOG
      LOGICAL STRO, WEAK
      COMMON /PRILOG/ STRO, WEAK
      COMMON /UNITS/ IUINP, IUOUT
C  ---------------------------------------------------------------------
C  Local variables:
      CHARACTER*20 STR
C  ---------------------------------------------------------------------
C
      XLL(IND) = XM
      IF (MARG(IND).EQ.0) MARG(IND) = -1
      IF (MARG(IND).EQ.1) MARG(IND) = 2
      IF (XLL(IND).GT.XRL(IND)) THEN
C
C  The left margin is above the right margin. The user is probably
C  attempting to set a new left margin, so the right margin is reset.
C  (Although the possibility of a user error exists as well.)
         MARG(IND) = -1
         XRL(IND) = BLIM
         IF (STRO) THEN
            CALL VNAME(IND,STR,LES)
            WRITE (IUOUT,10) STR(1:LES)
         END IF
      ELSE IF (XLL(IND).EQ.XRL(IND)) THEN
C
C  Left and right margins coincide. Fix the variable.
         POINT(IND) = XM
         IXAT(IND) = 0
         IF (STRO) THEN
            CALL VNAME(IND,STR,LES)
            WRITE (IUOUT,20) STR(1:LES)
         END IF
      END IF
C
10    FORMAT (/$(FC)'Illegal margins in variable ',A,'. ',
     &        'Right margin is reset.')
20    FORMAT (/$(FC)'Left and right margins coincide. Variable ',A,
     &        ' effectively fixed.')
      END
C  ---------------------------------------------------------------------
      SUBROUTINE SETNAM ( IND, NAME, ICODE )
C  ---------------------------------------------------------------------
C
C  Description:
C    Subroutine SETNAM assigns names to the Merlin minimization
C    variables. It makes sure that no two variables have the same
C    name.
C
C  Input arguments:
C    IND        The index of the variable whose name we wish to set.
C    NAME       The name to set.
C
C  Output arguments:
C    ICODE      Error indicator.
C               ICODE =  0 -> The name was set.
C               ICODE = -2 -> The name exists already and was not set.
C
C  ---------------------------------------------------------------------
C  Arguments:
      CHARACTER*(*) NAME
C  ---------------------------------------------------------------------
C  Common blocks:
      PARAMETER ( MXV = $(MXV) )
      CHARACTER*10 NICKNA
      COMMON /NAMES/ NICKNA(MXV)
      COMMON /DIMEN/ NOD, NOF
      LOGICAL STRO, WEAK
      COMMON /PRILOG/ STRO, WEAK
      COMMON /UNITS/ IUINP, IUOUT
C  ---------------------------------------------------------------------
C  Local variables:
      CHARACTER*10 STR
C  ---------------------------------------------------------------------
C
C  This should never happen.
      IF (IND.LT.1 .OR. IND.GT.NOD) THEN
         CALL I2STR(IND,STR,LES)
         CALL MABORT('Incorrect index: '//STR(1:LES),'SETNAM')
      END IF
C
C  Check for duplicate names.
      DO 10,I=1,NOD
         IF (IND.EQ.I .OR. NICKNA(I).EQ.' ') GOTO 10
         IF (ISCOMP(NAME,NICKNA(I)).NE.0) THEN
            LE1 = LENGTH(NAME)
            CALL I2STR(IND,STR,LES)
            IF (WEAK) WRITE (IUOUT,20) NAME(1:LE1), STR(1:LES)
            ICODE = -2
            RETURN
         END IF
10    CONTINUE
C
C  This is a unique name. Set it.
      NICKNA(IND) = NAME
      ICODE = 0
C
20    FORMAT (/$(FC)'Warning: Duplicate name "',A,'". ',
     &        'Naming of variable ',a,' failed.')
      END
C  ---------------------------------------------------------------------
      SUBROUTINE SETRMA ( IND, XM )
C  ---------------------------------------------------------------------
C
C  Description:
C    Sets the the value of the right margin to XM for the variable with
C    index IND.
C    If the left and right margin coincide, the variable is fixed.
C
C  Input arguments:
C    IND        Index of the variable.
C    XM         Value to set as the right margin.
C
C  ---------------------------------------------------------------------
      IMPLICIT $(TYPE) (A-H,O-Z)
C  Common blocks:
      PARAMETER ( MXV = $(MXV) )
      COMMON /PARAMS/ POINT(MXV), VAL
      COMMON /DETAIL/ IXAT(MXV), MARG(MXV)
      COMMON /XLIMS/ XLL(MXV), XRL(MXV)
      COMMON /BIGCOM/ BIGNUM, BLIM, DWARF, DWALOG
      LOGICAL STRO, WEAK
      COMMON /PRILOG/ STRO, WEAK
      COMMON /UNITS/ IUINP, IUOUT
C  ---------------------------------------------------------------------
C  Local variables:
      CHARACTER*20 STR
C  ---------------------------------------------------------------------
C
      XRL(IND) = XM
      IF (MARG(IND).EQ.0) MARG(IND) = 1
      IF (MARG(IND).EQ.-1) MARG(IND) = 2
      IF (XLL(IND).GT.XRL(IND)) THEN
C
C  The left margin is above the right margin. The user is probably
C  attempting to set a new right margin, so the left margin is reset.
C  (Although the possibility of a user error exists as well.)
         MARG(IND) = 1
         XLL(IND) = -BLIM
         IF (STRO) THEN
            CALL VNAME(IND,STR,LES)
            WRITE (IUOUT,10) STR(1:LES)
         END IF
      ELSE IF (XLL(IND).EQ.XRL(IND)) THEN
C
C  Left and right margins coincide. Fix the variable.
         POINT(IND) = XM
         IXAT(IND) = 0
         IF (STRO) THEN
            CALL VNAME(IND,STR,LES)
            WRITE (IUOUT,20) STR(1:LES)
         END IF
      END IF
C
10    FORMAT (/$(FC)'Illegal margins in variable ',A,'. ',
     &        'Left margin is reset.')
20    FORMAT (/$(FC)'Left and right margins coincide. Variable ',A,
     &        ' effectively fixed.')
      END
C  ---------------------------------------------------------------------
      SUBROUTINE SHRINK ( H, LD, NFREE )
C  ---------------------------------------------------------------------
C
C  Description:
C    Shrinks the Hessian matrix H, removing the rows and columns that
C    correspond to fixed variables.
C
C  Input arguments:
C    LD         Leading dimension of matrix H.
C
C  Output arguments:
C    NFREE      Number of free variables.
C
C  Input / Output arguments:
C    H          On input an NODxNOD matrix.
C               On output an NFREExNFREE matrix, with rows and columns
C               corresponding to fixed variables, removed.
C
C  ---------------------------------------------------------------------
      IMPLICIT $(TYPE) (A-H,O-Z)
C  Arguments:
      DIMENSION H(LD,*)
C  ---------------------------------------------------------------------
C  Common blocks:
      PARAMETER ( MXV = $(MXV) )
      COMMON /DETAIL/ IXAT(MXV), MARG(MXV)
      COMMON /DIMEN/ NOD, NOF
C  ---------------------------------------------------------------------
C
      NFREE = 0
      DO 56,I=1,NOD
         IF (IXAT(I).NE.0) THEN
            NFREE = NFREE+1
            IF (NFREE.NE.I) THEN
               DO 57,J=1,NOD
57                H(J,NFREE) = H(J,I)
            END IF
            NCF = 0
            DO 58,J=1,NOD
               IF (IXAT(J).NE.0) THEN
                  NCF = NCF+1
                  H(NCF,NFREE) = H(J,NFREE)
               END IF
58          CONTINUE
         END IF
56    CONTINUE
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE STBALO
C  ---------------------------------------------------------------------
C
C  Description:
C    Starts the BACKUP and HISTORY mechanisms. In addition it reads
C    some of the default panel settings that need to be in common
C    blocks.
C
C  ---------------------------------------------------------------------
      IMPLICIT $(TYPE) (A-H,O-Z)
C  Common blocks:
      COMMON /BIGCOM/ BIGNUM, BLIM, DWARF, DWALOG
      COMMON /CNTRL/ IDISK, MACECH, LIMCLK, IECOMA, IPROEC, MCLUBU
C  ---------------------------------------------------------------------
C  Local variables:
      CHARACTER*5 CMODE
C  ---------------------------------------------------------------------
C
      CALL GETPPS('BACKUP','MODE',CMODE,LEC)
      CALL BACK(CMODE,MERR)
C
      CALL HISTOR(1,MERR)
C
      CALL GETPPI('CONTROL','DISKMAC',IDISK)
      CALL GETPPI('CONTROL','ECHOMAC',MACECH)
      CALL GETPPI('CONTROL','ECHOPRO',IPROEC)
      CALL GETPPI('CONTROL','MCLKEYS',LIMCLK)
      CALL GETPPI('CONTROL','ECHOCOM',IECOMA)
      CALL GETPPI('CONTROL','MCLBUF',MCLUBU)
C
      CALL SETPPR('AUTO','TARGET',-BIGNUM,IERR)
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE STOREN ( NOD, NOF, IERR )
C  ---------------------------------------------------------------------
C
C  Description:
C    Checks whether the one-dimensional arrays in Merlin provide
C    adequate storage for a minimization problem with NOD variables
C    and NOF terms.
C    This subroutine does not use the Merlin output unit, since it
C    must be called before Merlin takes any action.
C
C  Input arguments:
C    NOD        Number of variables of the objective function.
C    NOF        Number of terms in the objective function.
C
C  Output arguments:
C    IERR       IERR = 0 -> There is enough storage for this
C                           minimization problem.
C               IERR = 1 -> There is not enough storage for this
C                           minimization problem.
C
C  ---------------------------------------------------------------------
C  Common blocks:
      PARAMETER ( MXV = $(MXV) )
      PARAMETER ( MXT = $(MXT) )
C  ---------------------------------------------------------------------
C  Local variables:
      CHARACTER*10 STR1, STR2
C  ---------------------------------------------------------------------
C
      IERR = 0
      IF (NOD.GT.MXV) THEN
         IERR = 1
         CALL I2STR(NOD,STR1,LE1)
         CALL I2STR(MXV,STR2,LE2)
         WRITE (*,10) STR1(1:LE1), STR2(1:LE2)
      END IF
      IF (NOF.GT.MXT) THEN
         IERR = 1
         CALL I2STR(NOF,STR1,LE1)
         CALL I2STR(MXT,STR2,LE2)
         WRITE (*,20) STR1(1:LE1), STR2(1:LE2)
      END IF
C
10    FORMAT (//$(FC)5X,A,' variables are too many for this ',
     &        'object code.'
     &        /$(FC)5X,'The maximum number of variables is: ',A
     &        /$(FC)5X,'You have to change the appropriate ',
     &        'parameters and recompile Merlin.')
20    FORMAT (//$(FC)5X,A,' terms are too many for this ',
     &        'object code.'
     &        /$(FC)5X,'The maximum number of terms is: ',A
     &        /$(FC)5X,'You have to change the appropriate ',
     &        'parameters and recompile Merlin.')
      END
C  ---------------------------------------------------------------------
      SUBROUTINE SUBST ( ACTCOM )
C  ---------------------------------------------------------------------
C
C  Description:
C    Subroutine SUBST is called by the Merlin operating system when
C    the current user command is an alias. It will substitute the
C    actual command (ACTCOM) in the command line.
C
C  Input arguments:
C    ACTCOM     The actual command to substitute in place of the alias.
C
C  ---------------------------------------------------------------------
C  Arguments:
      CHARACTER*(*) ACTCOM
C  ---------------------------------------------------------------------
C  Common blocks:
      PARAMETER ( LINLEN = 80 )
      INTEGER IS(LINLEN/2), IE(LINLEN/2)
      COMMON /INDAT1/ IS, IE, NIT, LEC
      CHARACTER*(LINLEN) INPLIN, QINPL
      COMMON /INDAT2/ INPLIN, QINPL
C  ---------------------------------------------------------------------
C  Local variables:
      CHARACTER*(LINLEN) LINE
C  ---------------------------------------------------------------------
C
      LEC1 = LENGTH(ACTCOM)
C
      IF (LEC1.EQ.LEC) THEN
         QINPL(IS(1):IE(1)) = ACTCOM
      ELSE
         LINE = ' '
         IDIFF = LEC1-LEC
         LINE(IS(1):IE(1)+IDIFF) = ACTCOM
         IE(1) = IE(1)+IDIFF
         LELI = LEN(LINE)
         DO 37,I=2,NIT
            IS1 = IS(I)+IDIFF
            IE1 = IE(I)+IDIFF
            IF (IS1.GT.LELI) THEN
               NIT = I-1
               GOTO 38
            ELSE IF (IE1.GT.LELI) THEN
               IE1 = LELI
               LINE(IS1:IE1) = QINPL(IS(I):IE(I))
               IS(I) = IS1
               IE(I) = IE1
               NIT = I
               GOTO 38
            ELSE
               LINE(IS1:IE1) = QINPL(IS(I):IE(I))
               IS(I) = IS1
               IE(I) = IE1
            END IF
37       CONTINUE
38       CONTINUE
         LEC = LEC1
         QINPL = LINE
      END IF
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE TITDIS
C  ---------------------------------------------------------------------
C
C  Description:
C    Displays the title (if any) for the current minimization session.
C    A title can be set or removed using the TITLE command.
C
C  ---------------------------------------------------------------------
C  Common blocks:
      CHARACTER*80 MTITLE
      COMMON /MTIT1/ MTITLE
      COMMON /MTIT2/ LEMTIT
      LOGICAL STRO, WEAK
      COMMON /PRILOG/ STRO, WEAK
      COMMON /UNITS/ IUINP, IUOUT
C  ---------------------------------------------------------------------
C
      IF (STRO .AND. LEMTIT.NE.0) THEN
         IF (LEMTIT.LE.72) THEN
            WRITE (IUOUT,10) MTITLE(1:MIN(LEMTIT,72))
         ELSE
            WRITE (IUOUT,10) MTITLE(1:69)//'...'
         END IF
      END IF
C
10    FORMAT ($(FC)'Title: ',A)
      END
C  ---------------------------------------------------------------------
      SUBROUTINE UNDEF ( IGO, IPRINT )
C  ---------------------------------------------------------------------
C
C  Description:
C    This routine checks if any of the point-variables is indefinite.
C    If INDEF(I) = 0, then the variable POINT(I) is indefinite.
C    An informative message is issued for every indefinite variable.
C
C  Input arguments:
C    IPRINT     Indicates whether a message should be printed for
C               each indefinite variable.
C               IPRINT = 0    -> Do not print any messages. Just set IGO
C                                to a proper value.
C               IPRINT .NE. 0 -> Print informative messages for each
C                                parameter that has not been assigned a
C                                value.
C
C  Output arguments:
C    IGO        Indicates whether there are indefinite variables:
C               IGO = 0 -> At least one variable has not been assigned
C                          an initial value.
C               IGO = 1 -> All variables have been assigned values.
C
C  ---------------------------------------------------------------------
C  Common blocks:
      PARAMETER ( MXV = $(MXV) )
      COMMON /INDEFI/ INDEF(MXV)
      COMMON /DIMEN/ NOD, NOF
      LOGICAL STRO, WEAK
      COMMON /PRILOG/ STRO, WEAK
      COMMON /UNITS/ IUINP, IUOUT
C  ---------------------------------------------------------------------
C  Local variables:
      CHARACTER*20 STR
C  ---------------------------------------------------------------------
C
      IGO = 1
      DO 10,I=1,NOD
         J = INDEF(I)
         IF (J.EQ.0 .AND. WEAK) THEN
            CALL VNAME(I,STR,LES)
            IF (IPRINT.NE.0) WRITE (IUOUT,20) STR(1:LES)
            IGO = 0
         END IF
10    CONTINUE
C
20    FORMAT ($(FC)'Undefined variable: ',A)
      END
C  ---------------------------------------------------------------------
      SUBROUTINE AUTO ( V, MAXW, MERR )
C  ---------------------------------------------------------------------
C
C  Description:
C    Implements the AUTO command.
C    Syntax: AUTO (has panel)
C
C  Input arguments:
C    V          Work array. Should have at least NOD*(NOD+1) elements.
C    MAXW       Number of available storage positions in array V.
C
C  Output arguments:
C    MERR       Error indicator.
C
C  ---------------------------------------------------------------------
      IMPLICIT $(TYPE) (A-H,O-Z)
C  Arguments:
      DIMENSION V(MAXW)
C  ---------------------------------------------------------------------
C  Common blocks:
      COMMON /TOTCAL/ ITNOCS, IPARTC, ITDER, IPDER, ITHESS, IPHESS,
     &                ITJAC, IPJAC
      COMMON /MODE3/ MODEV, IFFORM, INMCL
      COMMON /DIMEN/ NOD, NOF
C  ---------------------------------------------------------------------
C  Local variables:
      CHARACTER*80 MESS
      CHARACTER*10 STR
C  ---------------------------------------------------------------------
C
C  Did the user supply enough storage ?
      NEED = NOD**2+NOD
      CALL VSIM('N**2+N',MAXW,NEED,MERR)
      IF (MERR.NE.0) RETURN
C
      CALL SETPI('CANCEL',1,IERR)
      CALL CHANGE(MERR)
      IF (MERR.NE.0) RETURN
      CALL GETPI('CANCEL',ICANC)
      IF (ICANC.NE.1 .OR. ITNOCS.EQ.0 .OR. MODEV.EQ.0) RETURN
C
C  Get panel parameters.
      CALL GETPI('NOC',NTOT)
      CALL GETPR('TARGET',VALOW)
      CALL GETPS('MAD',STR,LES)
      IF (ISCOMP(STR,'ON').NE.0) THEN
         MADA = 1
      ELSE
         MADA = 0
      END IF
C
C  Set automatic derivatives.
      CALL SETADE(MADA)
C
C  Call the minimizer.
      IFST = ITNOCS
      IGST = ITDER
      CALL MAGIC(NTOT,VALOW,V,INFO)
C
C  Construct a message describing the reason for termination.
      CALL AUTOTM(INFO,MESS)
C
C  Report the reason for termination.
      NOCEX = ITNOCS-IFST
      NOGEX = ITDER-IGST
      ITER = -1
      NOJAC = -1
      CALL MINREP(MESS,NOCEX,NOGEX,NOJAC,ITER)
C
C  Disable automatic derivatives.
      CALL SETADE(0)
C
C  Add backup record.
      CALL BACKUP('M')
C
C  Assign values to MCL-specific keywords.
      CALL SETPI('FCALLS',NOCEX,IERR)
      CALL SETPI('GCALLS',NOGEX,IERR)
      CALL SETPI('INFO',INFO,IERR)
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE MAGIC ( NTOT, VALOW, VERSIM, INFO )
C  ---------------------------------------------------------------------
C
C  Description:
C    This routine is an implementation of an automatic route to
C    multidimensional function minimization. Since it is quite general,
C    it may not be very efficient for all function classes. The user is
C    advised then to learn to operate merlin manually.
C    This is the main entry point to the MERLIN optimization package.
C
C  Input arguments:
C    NTOT       Maximum allowed number of function evaluations.
C    VALOW      Target value.
C    VERSIM     Work array. Should have at lest NOD*(NOD+1) elements.
C
C  Output arguments:
C    INFO       Reason for termination:
C               INFO = 1 -> All calls exhausted.
C               INFO = 2 -> Further progress is not possible.
C               INFO = 3 -> Target value has been reached.
C
C  ---------------------------------------------------------------------
      IMPLICIT $(TYPE) (A-H,O-Z)
C  Arguments:
      DIMENSION VERSIM(*)
C  ---------------------------------------------------------------------
C  Common blocks:
      PARAMETER ( MXV = $(MXV) )
      COMMON /PARAMS/ POINT(MXV), VAL
      COMMON /DETAIL/ IXAT(MXV), MARG(MXV)
      COMMON /XLIMS/ XLL(MXV), XRL(MXV)
      COMMON /TOTCAL/ ITNOCS, IPARTC, ITDER, IPDER, ITHESS, IPHESS,
     &                ITJAC, IPJAC
      COMMON /DIMEN/ NOD, NOF
      COMMON /ACCURA/ COMACC
      CHARACTER*10 DERIVA, PRINTO, PROCES, JACOMO
      COMMON /MODEIN/ DERIVA, PRINTO, PROCES, JACOMO
      COMMON /UPDATE/ IGRADU, IHESSU
      COMMON /RWORK/ XT1(MXV), XT2(MXV), XT3(MXV), XT4(MXV), XT5(MXV),
     &               XT6(MXV), XT7(MXV+1)
C  ---------------------------------------------------------------------
C  Local variables:
      DIMENSION RATE(4), RATES(4), IMET(4)
      LOGICAL EXAU(4)
      INTEGER SCALLS, FCALLS
      CHARACTER*10 METH(4), METHOD
      DATA METH / 'BFGS' , 'TRUST' , 'SIMPLEX' , 'ROLL' /
      DATA ITMAX, IPRINT /
     &        -1,      0 /
      DATA LSMETH,   RHO, SIGMA, LSITER, FTOLB /
     &          1, 1.E-4,   0.9,     30,    0. /
      DATA  RHOT, FTOLT /
     &     1.E-4,    0. /
      DATA ALPHA, BETA, GAMMA, FTOLS, INISI /
     &        1.,  0.5,    2.,    0.,     1 /
      DATA LSNOCR, TOLR, ENH, IFM, FTOLR /
     &         30, .001,  3.,   4,    0. /
C  ---------------------------------------------------------------------
C
C  Initialize parameters
      NREM = NTOT
      XTOLSI = COMACC
      GTOLB = COMACC
      XTOLB = COMACC
      GTOLT = COMACC
      XTOLT = COMACC
      IUVS = 0
      XCONST = 0.1
      LSNOCS = 50
      TOLSS = 0.01
      DO 1,I=1,4
         EXAU(I) = .TRUE.
         RATE(I) = 1.
1     CONTINUE
C
C  Adjust initial steps.
      ISTECL = ITNOCS
      IF (DERIVA.EQ.'ANAL') THEN
         CALL STEPAL
      ELSE
         CALL PRADJ
      END IF
      NREM = NREM-ITNOCS+ISTECL
C
C  Decision session starts here.
3     CONTINUE
C  Form list of candidate methods.
      DO 4,I=1,4
         IMET(I) = I
         RATES(I) = -RATE(I)
4     CONTINUE
C
C  Arrange them in descending rate order.
      CALL REORDE(RATES,IMET,4)
C
C  Decide which method to choose. Check if highest rate is zero.
C  If so proceed with special handling.
      IF (RATE(IMET(1)).EQ.0.) GOTO 6
      DO 5,I=1,4
         J = IMET(I)
C  Pick a candidate.
         METHOD = METH(J)
C  Choose it, if EXAU is .TRUE.
         IF (EXAU(J)) GOTO 8
5     CONTINUE
C
C  If all candidates are rejected, proceed with the following route.
6     CONTINUE
      NOC = MAX(NREM/4,600)
C
      CALL HYPER(NOC,TOLR,ENH,IFM,IPRINT,NOCEX,FTOLR,LSNOCR,INFO,ITER)
      CALL REPORT('ROLL',ITNOCS-ISTECL,NTOT)
      IF (VAL.LE.VALOW) GOTO 7
C
      CALL TRUSMV(VERSIM,POINT,XT1,NOD,VAL,NOC,COMACC,GTOLT,
     &            ITMAX,XTOLT,RHOT,FTOLT,INFO,ITER,NOCEX,
     &            NOGEX,IPRINT,IXAT,IGRADU,IHESSU,XT2,XT3,
     &            XT4,XT5,XT6,0,DELTA,MARG)
      CALL REPORT('TRUST',ITNOCS-ISTECL,NTOT)
      IF (VAL.LE.VALOW) GOTO 7
C
      CALL SIMPMV(VERSIM,NOD,NOD,POINT,VAL,ALPHA,BETA,GAMMA,FTOLS,
     &            NOCEX,ITER,ITMAX,INFO,NOC,INISI,XTOLSI,IXAT,MARG,
     &            XLL,XRL,IPRINT,XT1,XT2,XT3,XT4,XT5,XT7,COMACC,
     &            IUVS,XCONST,LSNOCS,TOLSS)
      CALL REPORT('SIMPLEX',ITNOCS-ISTECL,NTOT)
      IF (VAL.LE.VALOW) GOTO 7
C
      CALL BFGSMV(VERSIM,POINT,XT1,NOD,VAL,NOC,COMACC,GTOLB,
     &            ITMAX,XTOLB,LSMETH,RHO,SIGMA,LSITER,FTOLB,INFO,ITER,
     &            NOCEX,NOGEX,IPRINT,IXAT,MARG,XLL,XRL,IGRADU,IHESSU,
     &            XT2,XT3,XT4,XT5)
      CALL REPORT('BFGS',ITNOCS-ISTECL,NTOT)
      IF (VAL.LE.VALOW) GOTO 7
C
      INFO = 2
      RETURN
C
7     CONTINUE
      INFO = 3
      RETURN
C
C   Set parameters for the chosen method.
8     CONTINUE
      NOC = MAX(NREM/10,500)
      NOC = MIN(NOC,2000)
      DO 9,I=1,4
         IF (METHOD.EQ.METH(I) .AND. EXAU(I)) GOTO 11
9     CONTINUE
C
C   Revise calling parameters if all EXAU's are .FALSE.
      DO 10,I=1,4
         IF (METHOD.EQ.METH(I)) THEN
            IF (I.EQ.1 .OR. I.EQ.2) THEN
               IF (DERIVA.NE.'NUMER') NOC = NINT(1.3*NOC)
            ELSE IF (I.EQ.4) THEN
               IFM = IFM + 2
            END IF
            GOTO 11
         END IF
10    CONTINUE
C
C  Execution session starts here.
11    CONTINUE
C
C  Check if there are any calls left. If not quit.
      JMET = I
      IF (NREM.LE.0) THEN
         INFO = 1
         RETURN
      END IF
C
C  Update starting values.
      SVAL = VAL
      SCALLS = ITNOCS
C
C  Begin execution.
      IF (JMET.EQ.1) THEN
         CALL BFGSMV(VERSIM,POINT,XT1,NOD,VAL,NOC,COMACC,GTOLB,
     &                  ITMAX,XTOLB,LSMETH,RHO,SIGMA,LSITER,FTOLB,
     &                  INFO,ITER,NOCEX,NOGEX,IPRINT,IXAT,MARG,XLL,
     &                  XRL,IGRADU,IHESSU,XT2,XT3,XT4,XT5)
      ELSE IF (JMET.EQ.2) THEN
         CALL TRUSMV(VERSIM,POINT,XT1,NOD,VAL,NOC,COMACC,GTOLT,
     &                  ITMAX,XTOLT,RHOT,FTOLT,INFO,ITER,NOCEX,
     &                  NOGEX,IPRINT,IXAT,IGRADU,IHESSU,
     &                  XT2,XT3,XT4,XT5,XT6,0,DELTA,MARG)
      ELSE IF (JMET.EQ.3) THEN
         CALL SIMPMV(VERSIM,NOD,NOD,POINT,VAL,ALPHA,BETA,GAMMA,FTOLS,
     &                  NOCEX,ITER,ITSI,INFO,NOC,INISI,XTOLSI,IXAT,MARG,
     &                  XLL,XRL,IPRINT,XT1,XT2,XT3,XT4,XT5,XT7,COMACC,
     &                  IUVS,XCONST,LSNOCS,TOLSS)
      ELSE IF (JMET.EQ.4) THEN
         CALL HYPER(NOC,TOLR,ENH,IFM,IPRINT,NOCEX,FTOLR,LSNOCR,
     &                 INFO,ITER)
      END IF
C
C  Update post execution values.
      FCALLS = ITNOCS
      FVAL = VAL
C
C  Calculate rate, remaining calls, EXAU.
      NOCM = FCALLS-SCALLS
      DROP = SVAL-FVAL
      IF (DROP.EQ.0.) THEN
         RATE(JMET) = 0.
      ELSE
         RATE(JMET) = ABS(DROP)/(1.+NOCM)/MAX(ABS(SVAL),ABS(FVAL))
      END IF
      NREM = NREM-NOCM
      EXAU(JMET) = NOCM.GE.NOC
C
C  Report post execution status.
      CALL REPORT(METHOD,ITNOCS-ISTECL,NTOT)
C
C  Check if target value is reached and if so quit.
      IF (VAL.LE.VALOW) THEN
         INFO = 3
         RETURN
      END IF
C
C  Start a new cycle. Transfer control to decision session.
      GOTO 3
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE REPORT ( METHOD, NOCEX, NTOT )
C  ---------------------------------------------------------------------
C
C  Description:
C    Called by the automatic minimization routine MAGIC. It reports the
C    minimization status, during the automatic minimization session.
C
C  Input arguments:
C    METHOD     The method in use.
C    NOCEX      Number of function calls so far.
C    NTOT       Total number of allowed function calls.
C
C  ---------------------------------------------------------------------
      IMPLICIT $(TYPE) (A-H,O-Z)
C  Arguments:
      CHARACTER*(*) METHOD
C  ---------------------------------------------------------------------
C  Common blocks:
      PARAMETER ( MXV = $(MXV) )
      COMMON /PARAMS/ POINT(MXV), VAL
      LOGICAL STRO, WEAK
      COMMON /PRILOG/ STRO, WEAK
      COMMON /UNITS/ IUINP, IUOUT
C  ---------------------------------------------------------------------
C  Local variables:
      CHARACTER*8 METH, STR1, STR2
      CHARACTER*30 CVAL
C  ---------------------------------------------------------------------
C
      IF (STRO) THEN
         METH = METHOD
         CALL I2STR(NOCEX,STR1,LE1)
         CALL I2STR(NTOT,STR2,LE2)
         CALL R2RTF(VAL,CVAL,LEV)
         WRITE (IUOUT,200) METH, CVAL(1:LEV), STR1(1:LE1),
     &                        STR2(1:LE2)
         CALL DASH
      END IF
      CALL BACKUP('M')
C
200   FORMAT ($(FC)1X,'Method: ',A,1X,'Lower value: ',A,1X,'Calls: ',
     &        A,' of ',A)
      END
C  ---------------------------------------------------------------------
      SUBROUTINE REORDE ( RATE, IND, N )
C  ---------------------------------------------------------------------
C
C  Description:
C    Called by the automatic minimization routine MAGIC. It arranges the
C    drop rates in descending order, rearranging the indices IND. It
C    uses a bubble sort. See:
C      N. Wirth, Algorithms & Data Structures,
C      Prentice-Hall, 1986.
C      Knuth D.E., The art of computer programming, vol 3,
C      Addison-Wesley, Reading, Mass., 1973.
C
C  Input arguments:
C    N          Number of elements in array RATE.
C
C  Input / Output arguments:
C    RATE       The drop rates. On output they are sorted in descending
C               order.
C    IND        Associated indices.
C
C  ---------------------------------------------------------------------
      IMPLICIT $(TYPE) (A-H,O-Z)
C  Arguments:
      DIMENSION RATE(N)
      INTEGER IND(N)
C  ---------------------------------------------------------------------
C  Local variables:
      LOGICAL SORTED
C  ---------------------------------------------------------------------
C
      DO 10,K=1,N-1
         SORTED = .TRUE.
         DO 20,I=N,K+1,-1
            IM1 = I-1
            IF (RATE(I).LT.RATE(IM1)) THEN
               SORTED = .FALSE.
               TEMP = RATE(I)
               RATE(I) = RATE(IM1)
               RATE(IM1) = TEMP
               ITEMP = IND(I)
               IND(I) = IND(IM1)
               IND(IM1) = ITEMP
            END IF
20       CONTINUE
         IF (SORTED) RETURN
10    CONTINUE
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE AUTOTM ( INFO, MESS )
C  ---------------------------------------------------------------------
C
C  Description:
C    Constructs a message, describing the reason for termination of the
C    AUTO method.
C
C  Input arguments:
C    INFO       Reason for termination of the AUTO method, as returned
C               by subroutine MAGIC.
C
C  Output arguments:
C    MESS       Message, describing the reason for termination.
C
C  ---------------------------------------------------------------------
C  Arguments:
      CHARACTER*(*) MESS
C  ---------------------------------------------------------------------
C
      IF (INFO.EQ.1) THEN
         MESS = 'All function evaluations have been used'
      ELSE IF (INFO.EQ.2) THEN
         MESS = 'Further progress is not possible'
      ELSE IF (INFO.EQ.3) THEN
         MESS = 'Target value has been reached'
      ELSE
         CALL MABORT('Incorrect "info".','AUTOTM')
      END IF
C
      END
C  ---------------------------------------------------------------------
      FUNCTION AXP ( Z )
C  ---------------------------------------------------------------------
C
C  Description:
C    Calculates EXP(Z) avoiding underflows. If EXP(Z) would cause an
C    underflow, AXP is set to 0.
C
C  Input arguments:
C    Z          Argument for EXP(Z)
C
C  ---------------------------------------------------------------------
      IMPLICIT $(TYPE) (A-H,O-Z)
C  Common blocks:
      COMMON /BIGCOM/ BIGNUM, BLIM, DWARF, DWALOG
C  ---------------------------------------------------------------------
C
      IF (Z.LE.DWALOG) THEN
         AXP = 0.
      ELSE
         AXP = EXP(Z)
      END IF
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE BFGS ( V, MAXW, MERR )
C  ---------------------------------------------------------------------
C
C  Description:
C    Implements the BFGS command.
C    Syntax: BFGS (Has panel)
C
C  Input arguments:
C    MAXW       Number of available storage positions in array V.
C
C  Output arguments:
C    MERR       Error indicator.
C
C  Input / Output arguments:
C    V          On input the upper triangular part of V may contain an
C               initial approximation to the Choleski L[t] factor of
C               the Hessian matrix. On output contains the upper
C               triangular part contains the Choleski L[t] of the
C               BFGS approximation to the Hessian.
C
C  ---------------------------------------------------------------------
      IMPLICIT $(TYPE) (A-H,O-Z)
C  Arguments:
      DIMENSION V(MAXW)
C  ---------------------------------------------------------------------
C  Common blocks:
      PARAMETER ( MXV = $(MXV) )
      COMMON /PARAMS/ POINT(MXV), VAL
      COMMON /DETAIL/ IXAT(MXV), MARG(MXV)
      COMMON /XLIMS/ XLL(MXV), XRL(MXV)
      COMMON /DIMEN/ NOD, NOF
      COMMON /TOTCAL/ ITNOCS, IPARTC, ITDER, IPDER, ITHESS, IPHESS,
     &                ITJAC, IPJAC
      COMMON /MODE3/ MODEV, IFFORM, INMCL
      COMMON /UPDATE/ IGRADU, IHESSU
      COMMON /ACCURA/ COMACC
      COMMON /RWORK/ XT1(MXV), XT2(MXV), XT3(MXV), XT4(MXV), XT5(MXV),
     &               XT6(MXV), XT7(MXV+1)
C  ---------------------------------------------------------------------
C  Local variables:
      CHARACTER*80 MESS
      CHARACTER*10 STR
      SAVE IFIRST
      DATA IFIRST / 0 /
C  ---------------------------------------------------------------------
C
C  Initialize panel parameters that are machine/configuration dependent
C  and cannot be set in the panel description file.
      IF (IFIRST.EQ.0) THEN
         CALL SETPR('XTOL',COMACC,IERR)
         CALL SETPR('GTOL',COMACC,IERR)
         IFIRST = 1
      END IF
C
C  Did the user supply enough storage ?
      NEED = NOD**2
      CALL VSIM('N**2',MAXW,NEED,MERR)
      IF (MERR.NE.0) RETURN
C
C  Set default values for some panel parameters and process any
C  changes the user enters.
      IG = IGRADU
      IH = IHESSU
      CALL SETPI('USEG',IG,IERR)
      CALL SETPI('USEH',IH,IERR)
      CALL SETPI('CANCEL',1,IERR)
      CALL CHANGE(MERR)
      IF (MERR.NE.0) RETURN
      CALL GETPI('CANCEL',ICANC)
      IF (ICANC.NE.1 .OR. ITNOCS.EQ.0 .OR. MODEV.EQ.0) RETURN
C
C  Get parameters from the panel.
      CALL GETPI('NOC',NOC)
      CALL GETPI('USEG',IG)
      CALL GETPI('USEH',IH)
      CALL GETPI('PRINT',IPRINT)
      CALL GETPR('FTOL',FTOL)
      CALL GETPR('XTOL',XTOL)
      CALL GETPR('GTOL',GTOL)
      CALL GETPS('LS',STR,LES)
      IF (ISCOMP(STR,'WEAK').NE.0) THEN
         LSMETH = 1
      ELSE
         LSMETH = 2
      END IF
      CALL GETPI('LSITER',LSITER)
      CALL GETPR('RHO',RHO)
      CALL GETPR('SIGMA',SIGMA)
      CALL GETPI('ITER',ITMAX)
      CALL GETPS('MAD',STR,LES)
      IF (ISCOMP(STR,'ON').NE.0) THEN
         MAD = 1
      ELSE
         MAD = 0
      END IF
C
C  Set automatic derivatives.
      CALL SETADE(MAD)
C
C  Call the minimizer.
      CALL BFGSMV(V,POINT,XT5,NOD,VAL,NOC,COMACC,GTOL,ITMAX,XTOL,
     &            LSMETH,RHO,SIGMA,LSITER,FTOL,INFO,ITER,NOCEX,NOGEX,
     &            IPRINT,IXAT,MARG,XLL,XRL,IG,IH,XT1,XT2,XT3,XT4)
C
C  Construct a message describing the reason for termination.
      CALL BFGTM(INFO,MESS)
C
C  Report the reason for termination.
      NOJAC = -1
      CALL MINREP(MESS,NOCEX,NOGEX,NOJAC,ITER)
C
C  Disable automatic derivatives.
      CALL SETADE(0)
C
C  Add backup record.
      CALL BACKUP('M')
C
C  Assign values to MCL-specific keywords.
      CALL SETPI('FCALLS',NOCEX,IERR)
      CALL SETPI('GCALLS',NOGEX,IERR)
      CALL SETPI('ITERDONE',ITER,IERR)
      CALL SETPI('INFO',INFO,IERR)
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE BFGSMV ( H, POINT, GRAD, NOD, VAL, NOMAXC, COMACC,
     &                    GTOL, ITMAX, XTOL, LSMETH, RHO, SIGMA, LSITER,
     &                    FTOL, INFO, ITER, NOCEX, NOGEX, IPRINT, IXAT,
     &                    MARG, XLL, XRL, IGRADU, IHESSU, XS, GN, Y,
     &                    GS )
C  ---------------------------------------------------------------------
C
C  Description:
C    Implementation of the BFGS method. Uses Choleski factorization
C    of the Hessian. The Choleski factor L[t] is stored in the upper
C    triangular part of matrix H. Fixed variables are removed, reducing
C    thus the dimensionality of the problem. Margins are treated
C    explicitly (without transformations) by the line search routine.
C    For more information see:
C      Dennis J.E. & Schnabel R.B.
C      Numerical methods for unconstrained optimization and
C      nonlinear equations
C      Prentice Hall, Englewood Cliffs, New Jersey, 1983.
C
C  Input arguments:
C    GRAD       Work array of length NOD.
C    NOD        Dimensionality of the objective function.
C    NOMAXC     Maximum allowed number of calls to the objective
C               function.
C    COMACC     Machine accuracy.
C    GTOL       G-convergence criterion.
C    ITMAX      Maximum allowed number of BFGS iterations.
C               Set ITMAX = -1 for an unlimited number.
C    XTOL       X-convergence criterion.
C    LSMETH     Line search method. See subroutine LNS.
C    RHO        The line search "rho". See subroutine LNS.
C    SIGMA      The line search "sigma". See subroutine LNS.
C    LSITER     Maximum allowed number of line search iterations. See
C               subroutine LSITER.
C    FTOL       F-convergence criterion.
C    IPRINT     Printout level:
C               IPRINT = 0 -> Nothing is printed.
C               IPRINT = 1 -> Simple printout: Prints ITER, VAL,
C                             NOC, MAXNOC
C               IPRINT = 2 -> Full printout. Prints in addition the
C                             minimization parameters, contained in
C                             array POINT.
C    IXAT       Fix statuses.
C    MARG       Margin statuses.
C    XLL        Left margins.
C    XRL        Right margins.
C    IGRADU     IGRADU = 0 -> Ask subroutine DERMER to recalculate
C                             the gradient at the current point. If
C                             a valid cached gradient exists, DERMER
C                             will return it.
C               IGRADU = 1 -> Ask subroutine DERMER to return its
C                             cached gradient. No recalculation
C                             takes place, unless of course the
C                             gradient cache is empty.
C    IHESSU     IHESSU = 0 -> The Hessian will be initialized to the
C                             identity matrix.
C               IHESSU = 1 -> The upper triangular part of H contains
C                             the Choleski factor L[t] of a matrix,
C                             to be used as an initial approximation
C                             to the Hessian.
C    XS         Work array of length NOD.
C    GN         Work array of length NOD.
C    Y          Work array of length NOD.
C    GS         Work array of length NOD.
C
C  Output arguments:
C    INFO       Reason for termination:
C               INFO = 1 -> Target value has been reached.
C               INFO = 2 -> The gradient criterion is satisfied.
C               INFO = 3 -> All function calls have been exhausted.
C               INFO = 4 -> X-convergence has been achieved.
C               INFO = 5 -> F-convergence has occurred.
C               INFO = 6 -> Division by zero while updating the
C                           Hessian matrix.
C               INFO = 7 -> Iterations exhausted.
C               INFO = 8 -> All variables are fixed.
C               INFO = 9 -> The line search cannot proceed any further.
C    ITER       Number of BFGS iterations that were performed.
C    NOCEX      Number of function evaluations that were performed.
C    NOGEX      Number of gradient evaluations that were performed.
C
C  Input / Output arguments:
C    H          On input the upper triangular part of H may contain an
C               initial approximation to the Choleski L[t] factor of
C               the Hessian matrix. On output contains the upper
C               triangular part contains the Choleski L[t] of the
C               BFGS approximation to the Hessian.
C    POINT      On input, the current point. On output, the new point
C               after the minimization.
C    VAL        Value of the objective function.
C
C  ---------------------------------------------------------------------
      IMPLICIT $(TYPE) (A-H,O-Z)
C  Arguments:
      DIMENSION H(NOD,NOD), POINT(NOD), GRAD(NOD)
      DIMENSION IXAT(NOD), MARG(NOD), XLL(NOD), XRL(NOD)
      DIMENSION XS(NOD), GN(NOD), Y(NOD), GS(NOD)
C  ---------------------------------------------------------------------
C
C  Number of iterations performed so far.
      ITER = 0
C
C  Number of function calls so far.
      NOCEX = 0
C
C  Number of (analytic) gradient calls so far.
      NOGEX = 0
C
C  Did we reach our target value ?
      IF (ITCONV(VAL).NE.0) THEN
         INFO = 1
         RETURN
      END IF
C
C  Count non-fixed variables. M will be used as the dimensionality
C  of the objective function in all subsequent matrix/vector operations.
      CALL FREEVA(M)
      IF (M.EQ.0) THEN
         INFO = 8
         RETURN
      END IF
C
C  Calculate derivatives at the starting point.
      IF (IGRADU.EQ.0) THEN
         MODE = 0
      ELSE
         MODE = 2
      END IF
      CALL DERMER(POINT,VAL,GRAD,N1,N2,MODE)
      NOCEX = NOCEX+N1
      NOGEX = NOGEX+N2
C
C  Copy derivatives to an array of M elements.
      MM = 0
      DO 17,I=1,NOD
         IF (IXAT(I).NE.0) THEN
            MM = MM+1
            GS(MM) = GRAD(I)
         END IF
17    CONTINUE
C
C  Check the gradient criterion for the initial point.
C  The numerical differentiation routine DERMER is expected to
C  return GRAD(I) = 0 for all fixed variables. Otherwise the
C  following test will fail, since the whole array GRAD is examined.
      IF (IGCONV(GRAD,POINT,NOD,VAL,GTOL).NE.0) THEN
         INFO = 2
         RETURN
      END IF
C
C  Initialize Hessian. (Upper triangular part only.)
      IF (IHESSU.EQ.0) THEN
         DO 10,J=1,M
            H(J,J) = 1.
            DO 20,I=1,J-1
20             H(I,J) = 0.
10       CONTINUE
      END IF
C  ---------------------------------------------------------------------
C  The following loop performs BFGS iterations.
C  ---------------------------------------------------------------------
1000  CONTINUE
      IF (ITMAX.NE.-1 .AND. ITER.GE.ITMAX) THEN
         INFO = 7
         RETURN
      END IF
      IF (NOCEX.GE.NOMAXC) THEN
         INFO = 3
         RETURN
      END IF
      ITER = ITER+1
C
C  Solve H * s = -g for the search direction XS. The minus sign will
C  be added later.
      CALL LLSOLV(H,XS,GS,M,NOD)
C
C  Save current point and value for the convergence tests.
      OLDVAL = VAL
      DO 99,I=1,NOD
99       GN(I) = POINT(I)
C
C  Prepare a vector of length NOD with the proper line search
C  direction.
      MM = 0
      DO 71,I=1,NOD
         IF (IXAT(I).NE.0) THEN
            MM = MM+1
            GS(I) = -XS(MM)
         ELSE
            GS(I) = 0.
         END IF
71    CONTINUE
C
C  Initial estimate for the line search.
      ALPHA = 1.
C
C  Perform the line search. Here, XS and Y are used as work arrays.
      CALL QNLS(GS,GRAD,N1,N2,ALPHA,COMACC,LSMETH,RHO,SIGMA,LSITER,
     &          POINT,VAL,NOD,MINFO,IXAT,MARG,XLL,XRL,XS,Y)
      NOCEX = NOCEX+N1
      NOGEX = NOGEX+N2
C
C  Any progress during the line search ?
      IF (ALPHA.EQ.0.) THEN
         INFO = 9
         RETURN
      END IF
      IF (VAL.LT.OLDVAL)
     &    CALL MDIS(POINT,VAL,ITER,NOCEX,NOMAXC,IPRINT,0)
C
C  Did we reach our target value ?
      IF (ITCONV(VAL).NE.0) THEN
         INFO = 1
         RETURN
      END IF
C
C  Check for X-convergence.
      IF (IXCONV(POINT,GN,NOD,XTOL).NE.0) THEN
         INFO = 4
         RETURN
      END IF
C
C  Check for F-convergence.
      IF (IFCONV(VAL,OLDVAL,FTOL).NE.0) THEN
         INFO = 5
         RETURN
      END IF
C
C  Calculate derivatives at the point the line search stopped.
      CALL DERMER(POINT,VAL,GN,N1,N2,0)
      NOCEX = NOCEX+N1
      NOGEX = NOGEX+N2
C
C  Check the gradient criterion.
      IF (IGCONV(GN,POINT,NOD,VAL,GTOL).NE.0) THEN
         INFO = 2
         RETURN
      END IF
C
C  Calculate the difference in the gradient: Y = g[k+1]-g[k]
C  and in the variables: XS = x[k+1] - x[k]
C  Update derivatives in GRAD.
      MM = 0
      DO 13,I=1,NOD
         IF (IXAT(I).NE.0) THEN
            MM = MM+1
            XS(MM) = ALPHA*GS(I)
            Y(MM) = GN(I)-GRAD(I)
            GS(MM) = GN(I)
         END IF
         GRAD(I) = GN(I)
13    CONTINUE
C
C  Calculate vector inner products needed to determine whether the
C  Hessian can be updated.
      CALL VVMULT(Y,Y,YTY,M)
      CALL VVMULT(XS,XS,DTD,M)
      CALL VVMULT(Y,XS,YTS,M)
C
C  Can we update the Hessian ?
      IF (YTS.GT.SQRT(COMACC*DTD*YTY)) THEN
C
C  XS = L[t] * s
         CALL LTMULT(H,XS,XS,M,NOD)
C
C  TT = s[t] * H * s
         CALL VVMULT(XS,XS,TT,M)
         IF (TT.EQ.0.) THEN
            INFO = 6
            RETURN
         END IF
C
C  GN = H * s
         CALL LMULT(H,XS,GN,M,NOD)
         AA = SQRT(YTS/TT)
         DO 57,I=1,M
57          GN(I) = AA*(Y(I)-AA*GN(I))/YTS
C
C  Update Choleski factors.
         CALL LLUPD(H,GN,XS,M,NOD)
      END IF
      GOTO 1000
C  ---------------------------------------------------------------------
C  End of BFGS iterations.
C  ---------------------------------------------------------------------
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE BFGTM ( INFO, MESS )
C  ---------------------------------------------------------------------
C
C  Description:
C    Constructs a message, describing the reason for termination of the
C    BFGS method.
C
C  Input arguments:
C    INFO       Reason for termination of the BFGS method, as returned
C               by subroutine BFGSMV.
C
C  Output arguments:
C    MESS       Message, describing the reason for termination.
C
C  ---------------------------------------------------------------------
C  Arguments:
      CHARACTER*(*) MESS
C  ---------------------------------------------------------------------
C
      IF (INFO.EQ.1) THEN
         MESS = 'Target value has been reached'
      ELSE IF (INFO.EQ.2) THEN
         MESS = 'The gradient criterion is satisfied'
      ELSE IF (INFO.EQ.3) THEN
         MESS = 'All function evaluations have been used'
      ELSE IF (INFO.EQ.4) THEN
         MESS = 'The X-convergence criterion is satisfied'
      ELSE IF (INFO.EQ.5) THEN
         MESS = 'The function value has converged'
      ELSE IF (INFO.EQ.6) THEN
         MESS = 'The Hessian matrix cannot be updated'
      ELSE IF (INFO.EQ.7) THEN
         MESS = 'The specified number of iterations has been reached'
      ELSE IF (INFO.EQ.8) THEN
         MESS = 'All variables are fixed'
      ELSE IF (INFO.EQ.9) THEN
         MESS = 'Further progress is not possible'
      ELSE
         CALL MABORT('Incorrect "info".','BFGTM')
      END IF
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE BRAK ( POINT, DIR, NOD, A, B, C, FA, FB, FC, ITER,
     &                  NOCEX, XALIM, XBLIM, LIM, INFO, FTARG, MAXNOC,
     &                  MARG, XLL, XRL, WORK, IXAT )
C  ---------------------------------------------------------------------
C
C  Description:
C    Given two distinct initial points A, B, and a search direction DIR,
C    subroutine BRAK will calculate the triplet (A,FA), (B,FB), (C,FC)
C    that brackets a minimum: A < B < C
C                             F(B) < F(A)
C                             F(B) < F(C)
C    The size of the interval B-A is not at issue here. Intervals of
C    any size that satisfy the bracketing criterion, are accepted.
C
C  Input arguments:
C    POINT      The current point.
C    DIR        Vector. The direction of search.
C    NOD        Dimensionality of the objective function.
C    XALIM      Left margin along the direction DIR.
C    XBLIM      Right margin along the direction DIR.
C    LIM        LIM =  0 -> No margins along DIR.
C               LIM = -1 -> A left margin exists along DIR.
C               LIM =  1 -> A right margin exists along DIR.
C               LIM =  2 -> Both margins exist along the direction DIR.
C    FTARG      Target value.
C    MAXNOC     Maximum allowed number of function evaluations.
C    MARG       Margin statuses.
C    XLL        Left margins.
C    XRL        Right margins.
C    WORK       Work array of length NOD.
C    IXAT       Fix statuses.
C
C  Output arguments:
C    C          Point C of the bracketing triplet.
C    FB         Function value at B.
C    FC         Function value at C.
C    ITER       Number of iterations that were performed.
C    NOCEX      Number of function calls.
C    INFO       INFO = 1 -> Bracket successfully located.
C               INFO = 2 -> The minimum probably lies outside the
C                           margins. The best point so far is returned
C                           in B, FB.
C               INFO = 3 -> The target value (FTARG) has been reached.
C                           The best point so far is returned in B, FB.
C               INFO = 4 -> The allowed number (MAXNOC) of function
C                           evaluations has been reached. The best
C                           point so far is returned in B, FB.
C               INFO = 5 -> Some error has occurred during
C                           interpolation. The best point so far
C                           is returned in B, FB.
C
C  Input / Output arguments:
C    A          On input, (A,B) is an initial estimation of the
C               bracketing interval.
C               On output, point A of the bracketing triplet.
C    B          On input, (A,B) is an initial estimation of the
C               bracketing interval.
C    FA         On input the function value at A. On output the
C               function value at the new A.
C
C  ---------------------------------------------------------------------
      IMPLICIT $(TYPE) (A-H,O-Z)
C  Arguments:
      DIMENSION POINT(NOD), DIR(NOD), MARG(NOD), XLL(NOD), XRL(NOD)
      DIMENSION IXAT(NOD), WORK(NOD)
C  ---------------------------------------------------------------------
C  Local variables:
      PARAMETER ( CONST = 2., RLIM = 1000. )
      LOGICAL FORCEA, FORCEB, FORCEX, TL
C  ---------------------------------------------------------------------
C
      XAA = XALIM
      XBB = XBLIM
      LL = LIM
      FORCEA = (LL.EQ.-1 .OR. LL.EQ.2) .AND. A.LE.XAA
      FORCEB = (LL.EQ. 1 .OR. LL.EQ.2) .AND. B.GE.XBB
      IF (FORCEB) B = XBLIM
      CALL MEVAL(POINT,DIR,B,NOD,IXAT,MARG,XLL,XRL,WORK,1,FB)
C
C  Arrange values so that the direction from A to B is
C  a descent direction.
      IF (FB.GT.FA) THEN
         T = A
         A = B
         B = T
         T = FA
         FA = FB
         FB = T
         T = XAA
         XAA = XBB
         XBB = T
         TL = FORCEA
         FORCEA = FORCEB
         FORCEB = TL
         IF (ABS(LL).EQ.1) LL = -LL
      END IF
C
C  NOCEX is the number of function evaluations.
      NOCEX = 1
C
C  ITER is the number of iterations (interpolations) performed.
      ITER = 0
C
C  Check target value.
      IF (FB.LE.FTARG) THEN
         INFO = 3
         RETURN
      END IF
C
C  Initial estimate for the third point of the bracketing triplet.
      C = B+CONST*(B-A)
C
C  Is C in between the margins ?
      IF ((LL.EQ. 1 .OR. LL.EQ.2) .AND. (C-XBB)*(XBB-B).GE.0.) THEN
         IF (FORCEB) THEN
            INFO = 2
            RETURN
         END IF
         C = XBB
         FORCEB = .TRUE.
      END IF
      CALL MEVAL(POINT,DIR,C,NOD,IXAT,MARG,XLL,XRL,WORK,1,FC)
      NOCEX = NOCEX+1
C  ---------------------------------------------------------------------
C  The following loop performs bracketing iterations.
C  ---------------------------------------------------------------------
1000  CONTINUE
C
C  Is this a bracketing triplet ?
      IF (FC.GT.FB) THEN
         INFO = 1
         RETURN
      END IF
C
C  Check target value.
      IF (FC.LE.FTARG) THEN
         INFO = 3
         B = C
         FB = FC
         RETURN
      END IF
C
C  Check function evaluations.
      IF (NOCEX.GE.MAXNOC) THEN
         INFO = 4
         B = C
         FB = FC
         RETURN
      END IF
C
C  There's still descent from B to C. We need to extrapolate for
C  our next guess.
      CALL QMIN2(A,FA,B,FB,C,FC,NR,XMIN)
      ITER = ITER+1
C
C  Did the extrapolation succeed in finding a minimum ?
      IF (NR.NE.0) THEN
C
C  Establish a right limit for the extrapolated minimum.
         XLIM = B+RLIM*(C-B)
C
C  This limit should lie inside the margins, however.
         FORCEX = (LL.EQ.1 .OR. LL.EQ.2) .AND. (XLIM-XBB)*(XBB-B).GT.0.
         IF (FORCEX) XLIM = XBB
C
C  Where is the newly calculated XMIN ?
C  Is XMIN between B and C ?
         IF ((B-XMIN)*(XMIN-C).GT.0.) THEN
            CALL MEVAL(POINT,DIR,XMIN,NOD,IXAT,MARG,XLL,XRL,WORK,1,FX)
            NOCEX = NOCEX+1
            IF (FX.LT.FC) THEN
               A = B
               B = XMIN
               FA = FB
               FB = FX
               INFO = 1
               RETURN
            ELSE IF (FX.GT.FB) THEN
               C = XMIN
               FC = FX
               INFO = 1
               RETURN
            END IF
            XMIN = C+CONST*(C-B)
            IF ((LL.EQ. 1 .OR. LL.EQ.2) .AND.
     &          (XMIN-XBB)*(XBB-B).GE.0.) THEN
               IF (FORCEB) THEN
                  B = C
                  FB = FC
                  INFO = 2
                  RETURN
               END IF
               FORCEB = .TRUE.
               XMIN = XBB
            END IF
            CALL MEVAL(POINT,DIR,XMIN,NOD,IXAT,MARG,XLL,XRL,WORK,1,FX)
            NOCEX = NOCEX+1
C
C  Is XMIN between C and XLIM ?
         ELSE IF ((C-XMIN)*(XMIN-XLIM).GT.0.) THEN
            CALL MEVAL(POINT,DIR,XMIN,NOD,IXAT,MARG,XLL,XRL,WORK,1,FX)
            NOCEX = NOCEX+1
C
C  No need to check whether FX > FC (successful bracketing). It
C  will be discovered early in the next iteration.
            IF (FX.LE.FC) THEN
               B = C
               C = XMIN
               XMIN = C+CONST*(C-B)
               IF ((LL.EQ. 1 .OR. LL.EQ.2) .AND.
     &                      (XMIN-XBB)*(XBB-B).GE.0.) THEN
                  IF (FORCEB) THEN
                     B = C
                     FB = FC
                     INFO = 2
                     RETURN
                  END IF
                  FORCEB = .TRUE.
                  XMIN = XBB
               END IF
               FB = FC
               FC = FX
               CALL MEVAL(POINT,DIR,XMIN,NOD,IXAT,MARG,XLL,XRL,
     &                             WORK,1,FX)
               NOCEX = NOCEX+1
            END IF
C
C  Is XMIN greater than XLIM ?
         ELSE IF ((XMIN-XLIM)*(XLIM-C).GE.0.) THEN
C
C  Force XMIN to the limit.
            IF (FORCEB .AND. FORCEX) THEN
               B = C
               FB = FC
               INFO = 2
               RETURN
            END IF
            XMIN = XLIM
            FORCEB = FORCEX
            CALL MEVAL(POINT,DIR,XMIN,NOD,IXAT,MARG,XLL,XRL,WORK,1,FX)
            NOCEX = NOCEX+1
         ELSE
C
C  The extrapolated minimum is between A and B or less than A.
C  This should never happen. Attempt to recover.
            CALL MABORT('Min. between A and B or less than A','BRAK')
         END IF
      ELSE
C
C  The extrapolation failed to establish a minimum.
         XMIN = C+CONST*(C-B)
         IF ((LL.EQ. 1 .OR. LL.EQ.2) .AND.
     &       (XMIN-XBB)*(XBB-B).GE.0.) THEN
            IF (FORCEB) THEN
               B = C
               FB = FC
               INFO = 2
               RETURN
            END IF
            FORCEB = .TRUE.
            XMIN = XBB
         END IF
         CALL MEVAL(POINT,DIR,XMIN,NOD,IXAT,MARG,XLL,XRL,WORK,1,FX)
         NOCEX = NOCEX+1
      END IF
C
C  Discard point A for the next iteration.
      A = B
      B = C
      C = XMIN
      FA = FB
      FB = FC
      FC = FX
      GOTO 1000
C  ---------------------------------------------------------------------
C  End of bracketing iterations.
C  ---------------------------------------------------------------------
      END
C  ---------------------------------------------------------------------
      SUBROUTINE BRENT ( POINT, DIR, NOD, AA, BB, CC, FB, TOL, XM, FM,
     &                   MAXNOC, NOCEX, INFO, FTARG, MARG, XLL, XRL,
     &                   WORK, IXAT )
C  ---------------------------------------------------------------------
C
C  Description:
C    Given a bracketing triplet AA, BB, CC as calculated by subroutine
C    BRAK, this routine will locate the minimum XM, along the direction
C    DIR using successive interpolation and bracketing. The minimum
C    will be located with a relative tolerance
C      (B-A)/MAX(ONE,ABS(XM) < TOL
C    where A, B are the endpoints of the bracketing interval.
C    TOL should be at least SQRT(COMACC). For details see:
C      Brent R.P.
C      Algorithms for minimization without derivatives.
C      Prentice-Hall, 1973, pp. 72-80.
C
C  Input arguments:
C    POINT      The current point.
C    DIR        Vector. The direction of search.
C    NOD        Dimensionality of the objective function.
C    AA         First point of the bracketing triplet.
C    BB         Second point of the bracketing triplet.
C    CC         Third point of the bracketing triplet.
C    FB         Function value at BB.
C    TOL        Relative tolerance.
C    MAXNOC     Maximum allowed number of function evaluations.
C    FTARG      Target value.
C    MARG       Margin statuses.
C    XLL        Left margins.
C    XRL        Right margins.
C    WORK       Work array of length NOD.
C    IXAT       Fix statuses.
C
C  Output arguments:
C    XM         The minimizer along the direction DIR.
C    FM         Function value at XM.
C    NOCEX      Number of function calls.
C    INFO       INFO = 1 -> The convrgence criterion has been satisfied.
C               INFO = 2 -> All function calls have been used.
C               INFO = 3 -> Target value has been reached.
C
C  ---------------------------------------------------------------------
      IMPLICIT $(TYPE) (A-H,O-Z)
C  Arguments:
      DIMENSION POINT(NOD), DIR(NOD), MARG(NOD), XLL(NOD), XRL(NOD)
      DIMENSION IXAT(NOD), WORK(NOD)
C  ---------------------------------------------------------------------
C
C  Some constants.
      ONE = 1.
C
C  Arrange bracketing points so that A < B.
      A = MIN(AA,CC)
      B = MAX(AA,CC)
C
C  Interpolation is always performed on points X, V, W.
      X = BB
      V = BB
      W = BB
      FX = FB
      FV = FB
      FW = FB
      STP = 0.
C
C  XLOW, FLOW need not satisfy the convergence criterion, and
C  are returned whenever all calls are exhausted.
      XLOW = BB
      FLOW = FB
C
C  NOCEX is the number of function evaluations.
      NOCEX = 0
C
C  ITER is the number of iterations performed.
      ITER = 0
C  ---------------------------------------------------------------------
C  The following loop performs brent iterations.
C  ---------------------------------------------------------------------
1000  CONTINUE
C
C  Keep lowest point so far, in case we run out of function calls.
      IF (X.LT.XLOW) THEN
         XLOW = X
         FLOW = FX
      END IF
C
C  Test for convergence.
      IF ((B-A)/MAX(ONE,ABS(X)).LE.TOL) THEN
         XM = X
         FM = FX
         INFO = 1
         RETURN
      END IF
C
C  Check target value.
      IF (FX.LE.FTARG) THEN
         XM = X
         FM = FX
         INFO = 3
         RETURN
      END IF
C
C  Did we perform too many function calls ?
      IF (NOCEX.GE.MAXNOC) THEN
         XM = XLOW
         FM = FLOW
         INFO = 2
         RETURN
      END IF
      ITER = ITER+1
C
C  Select step STP. On the first iteration use the golden section step.
      IF (ITER.EQ.1) THEN
         CALL GSTEP(A,B,X,STP)
      ELSE
C
C  Do the interpolation points coincide ?
         IF ((X-W)*(W-V)*(X-V).EQ.0.) THEN
            CALL GSTEP(A,B,X,STP)
         ELSE
            CALL QMIN2(X,FX,W,FW,V,FV,NR,XMIN)
            IF (NR.EQ.0) THEN
C
C  The interpolation failed to produce a minimum.
C  Proceed with golden section.
               CALL GSTEP(A,B,X,STP)
            ELSE
C
C  Is the interpolated minimum within the bracketing interval ?
               IF (XMIN.GT.A .AND. XMIN.LT.B) THEN
                  STP = XMIN-X
               ELSE
C
C  The interpolated minimum lies outside the bracket.
                  CALL GSTEP(A,B,X,STP)
               END IF
            END IF
         END IF
      END IF
C
C  Evaluate the function at the trial point U.
      U = X+STP
      CALL MEVAL(POINT,DIR,U,NOD,IXAT,MARG,XLL,XRL,WORK,1,FU)
      NOCEX = NOCEX+1
C
C  Update points according to Brent's algorithm.
      IF (FU.LE.FX) THEN
         IF (U.GE.X) THEN
            A = X
         ELSE
            B = X
         END IF
         V = W
         W = X
         X = U
         FV = FW
         FW = FX
         FX = FU
      ELSE
         IF (U.LT.X) THEN
            A = U
         ELSE
            B = U
         END IF
         IF (FU.LE.FW .OR. W.EQ.X) THEN
            V = W
            W = U
            FV = FW
            FW = FU
         ELSE IF (FU.LE.FV .OR. V.EQ.X .OR. V.EQ.W) THEN
            V = U
            FV = FU
         END IF
      END IF
      GOTO 1000
C  ---------------------------------------------------------------------
C  End of brent iterations.
C  ---------------------------------------------------------------------
      END
C  ---------------------------------------------------------------------
      SUBROUTINE GSTEP ( A, B, X, STP )
C  ---------------------------------------------------------------------
C
C  Description:
C    Called by subroutine BRENT. Selects step STP according to the
C    golden section rules.
C
C  Input arguments:
C    A          Left limit of the bracketing interval.
C    B          Right limit of the bracketing interval.
C    X          Current minimum in between (A,B).
C
C  Output arguments:
C    STP        The golden section step.
C
C  ---------------------------------------------------------------------
      IMPLICIT $(TYPE) (A-H,O-Z)
C  Arguments:
      PARAMETER ( GOLD = 0.618034, CONST = 1.-GOLD )
C  ---------------------------------------------------------------------
C
      AB2 = (A+B)/2.
      IF (X.GE.AB2) THEN
         STP = CONST*(A-X)
      ELSE
         STP = CONST*(B-X)
      END IF
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE CONGRA ( MERR )
C  ---------------------------------------------------------------------
C
C  Description:
C    Implements the CONGRA command.
C    Syntax: CONGRA (Has panel)
C
C  Output arguments:
C    MERR       Error indicator.
C
C  ---------------------------------------------------------------------
      IMPLICIT $(TYPE) (A-H,O-Z)
C  Common blocks:
      PARAMETER ( MXV = $(MXV) )
      COMMON /PARAMS/ POINT(MXV), VAL
      COMMON /DETAIL/ IXAT(MXV), MARG(MXV)
      COMMON /XLIMS/ XLL(MXV), XRL(MXV)
      COMMON /DIMEN/ NOD, NOF
      COMMON /MODE3/ MODEV, IFFORM, INMCL
      COMMON /ACCURA/ COMACC
      COMMON /UPDATE/ IGRADU, IHESSU
      COMMON /TOTCAL/ ITNOCS, IPARTC, ITDER, IPDER, ITHESS, IPHESS,
     &                ITJAC, IPJAC
      COMMON /RWORK/ XT1(MXV), XT2(MXV), XT3(MXV), XT4(MXV), XT5(MXV),
     &               XT6(MXV), XT7(MXV+1)
C  ---------------------------------------------------------------------
C  Local variables:
      CHARACTER*80 MESS
      CHARACTER*10 STR
      SAVE IFIRST
      DATA IFIRST / 0 /
C  ---------------------------------------------------------------------
C
C  Initialize panel parameters that are machine/configuration dependent
C  and cannot be set in the panel description file.
      IF (IFIRST.EQ.0) THEN
         CALL SETPR('XTOL',COMACC,IERR)
         CALL SETPR('GTOL',COMACC,IERR)
         IFIRST = 1
      END IF
C
C  Set default values for some panel parameters and process any
C  changes the user enters.
      IG = IGRADU
      CALL SETPI('USEG',IG,IERR)
      CALL SETPI('CANCEL',1,IERR)
      CALL CHANGE(MERR)
      IF (MERR.NE.0) RETURN
      CALL GETPI('CANCEL',ICANC)
      IF (ICANC.NE.1 .OR. ITNOCS.EQ.0 .OR. MODEV.EQ.0) RETURN
C
C  Get parameters from the panel.
      CALL GETPI('NOC',NOC)
      CALL GETPI('USEG',IG)
      CALL GETPI('PRINT',IPRINT)
      CALL GETPR('FTOL',FTOL)
      CALL GETPS('METHOD',STR,LES)
      IF (ISCOMP(STR,'PR').NE.0) THEN
         METHOD = 1
      ELSE IF (ISCOMP(STR,'FR').NE.0) THEN
         METHOD = 2
      ELSE IF (ISCOMP(STR,'GPR').NE.0) THEN
         METHOD = 3
      ELSE
         CALL MABORT('Incorrect CONGRA method.','CONGRA')
      END IF
      CALL GETPR('XTOL',XTOL)
      CALL GETPR('GTOL',GTOL)
      CALL GETPI('LSITER',LSITER)
      CALL GETPR('RHO',RHO)
      CALL GETPR('SIGMA',SIGMA)
      CALL GETPI('ITER',ITMAX)
      CALL GETPI('RESTART',IRES)
      CALL GETPS('MAD',STR,LES)
      IF (ISCOMP(STR,'ON').NE.0) THEN
         MAD = 1
      ELSE
         MAD = 0
      END IF
C
C  Set automatic derivatives.
      CALL SETADE(MAD)
C
C  Call the minimizer.
      CALL COGRMV(NOC,NOCEX,IPRINT,FTOL,METHOD,XT1,XTOL,GTOL,ITER,
     &            RHO,SIGMA,LSITER,ITMAX,IRES,COMACC,INFO,NOGEX,IXAT,
     &            MARG,IG,XLL,XRL,NOD,POINT,VAL,XT2,XT3,XT4,XT5,IFFORM)
C
C  Construct a message describing the reason for termination.
      CALL CGRTM(INFO,MESS)
C
C  Report the reason for termination.
      NOJAC = -1
      CALL MINREP(MESS,NOCEX,NOGEX,NOJAC,ITER)
C
C  Disable automatic derivatives.
      CALL SETADE(0)
C
C  Add backup record.
      CALL BACKUP('M')
C
C  Assign values to MCL-specific keywords.
      CALL SETPI('FCALLS',NOCEX,IERR)
      CALL SETPI('GCALLS',NOGEX,IERR)
      CALL SETPI('ITERDONE',ITER,IERR)
      CALL SETPI('INFO',INFO,IERR)
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE COGRMV ( NOMAXC, NOCEX, IPRINT, FTOL, METHOD, GRAD,
     &                    XTOL, GTOL, ITER, RHO, SIGMA, LSITER, ITMAX,
     &                    IRES, COMACC, INFO, NOGEX, IXAT, MARG,
     &                    IGRADU, XLL, XRL, NOD, POINT, VAL, XSTEP,
     &                    W, GNEW, XTEST, IFFORM )
C  ---------------------------------------------------------------------
C
C  Description:
C    Implements the the Polak-Ribiere, Fletcher-Reeves and generalized
C    Polak-Ribiere conjugate gradient formulae.
C    For more information see:
C      [1] Fletcher R.
C          Practical methods of optimization
C          John Wiley, New York, 1980. pp. 63-76 (chapter 4).
C      [2] Khoda K.M, Liu Y., Storey C.
C          Generalized Polak-Ribiere Algorithm.
C          JOTA, 75 (1992) 345-354.
C
C  Input arguments:
C    NOMAXC     Maximum allowed number of calls to the objective
C               function.
C    IPRINT     Printout level:
C               IPRINT = 0 -> Nothing is printed.
C               IPRINT = 1 -> Simple printout: Prints ITER, VAL,
C                             NOC, MAXNOC
C               IPRINT = 2 -> Full printout. Prints in addition the
C                             minimization parameters, contained in
C                             array POINT.
C    FTOL       F-convergence criterion.
C    METHOD     METHOD = 1 -> Use the Polak-Ribiere formula.
C               METHOD = 2 -> Use the Fletcher-Reeves formula.
C               METHOD = 3 -> Use the Generalized Polak-Ribiere
C                             algorithm.
C    GRAD       Work array of length NOD.
C    XTOL       X-convergence criterion.
C    GTOL       G-convergence criterion.
C    RHO        The line search "rho". See subroutine LNS.
C    SIGMA      The line search "sigma". See subroutine LNS.
C    LSITER     Maximum allowed number of line search iterations. See
C               subroutine LSITER.
C    ITMAX      Maximum allowed number of BFGS iterations.
C               Set ITMAX = -1 for an unlimited number of iterations.
C    IRES       IRES = 0 -> Use the N-step restart.
C               IRES = 1 -> Do not use the N-step restart.
C    COMACC     Machine accuracy.
C    IXAT       Fix statuses.
C    MARG       Margin statuses.
C    IGRADU     IGRADU = 0 -> Ask subroutine DERMER to recalculate
C                             the gradient at the current point. If
C                             a valid cached gradient exists, DERMER
C                             will return it.
C               IGRADU = 1 -> Ask subroutine DERMER to return its
C                             cached gradient. No recalculation
C                             takes place, unless of course the
C                             gradient cache is empty.
C    XLL        Left margins.
C    XRL        Right margins.
C    NOD        Dimensionality of the objective function.
C    XSTEP      Work array of length NOD.
C    W          Work array of length NOD.
C    GNEW       Work array of length NOD.
C    XTEST      Work array of length NOD.
C    IFFORM     Determines the functional form of the objective
C               function:
C               IFFORM = 0 -> A general function.
C               IFFORM = 1 -> A sum of squares.
C               Subsequently, this information is used to compute
C               initial estimates for the line search ALPHA.
C
C  Output arguments:
C    NOCEX      Number of function evaluations that were performed.
C    ITER       Number of BFGS iterations that were performed.
C    INFO       Reason for termination:
C               INFO = 1 -> Target value has been reached.
C               INFO = 2 -> The gradient criterion is satisfied.
C               INFO = 3 -> All function calls have been exhausted.
C               INFO = 4 -> X-convergence has been achieved.
C               INFO = 5 -> F-convergence has occurred.
C               INFO = 7 -> Iterations exhausted.
C               INFO = 8 -> The line search cannot proceed any further.
C               INFO = 9 -> All variables are fixed.
C    NOGEX      Number of gradient evaluations that were performed.
C
C  Input / Output arguments:
C    POINT      On input, the current point. On output, the new point
C               after the minimization.
C    VAL        Value of the objective function.
C
C  ---------------------------------------------------------------------
      IMPLICIT $(TYPE) (A-H,O-Z)
C  Arguments:
      DIMENSION GRAD(NOD), IXAT(NOD), MARG(NOD), XLL(NOD), XRL(NOD)
      DIMENSION POINT(NOD), XSTEP(NOD), W(NOD), GNEW(NOD), XTEST(NOD)
C  ---------------------------------------------------------------------
C
C  Some constants.
      ONE = 1.
      TWO = 2.
C
C  Number of iterations performed so far.
      ITER = 0
C
C  Number of function calls so far.
      NOCEX = 0
C
C  Number of (analytic) gradient calls so far.
      NOGEX = 0
C
C  Did we reach our target value ?
      IF (ITCONV(VAL).NE.0) THEN
         INFO = 1
         RETURN
      END IF
C
C  Count non-fixed variables.
      CALL FREEVA(M)
      IF (M.EQ.0) THEN
         INFO = 9
         RETURN
      END IF
C
C  Calculate derivatives at the starting point.
      IF (IGRADU.EQ.0) THEN
         MODE = 0
      ELSE
         MODE = 2
      END IF
      CALL DERMER(POINT,VAL,GRAD,NOCS1,NGR,MODE)
      NOCEX = NOCEX+NOCS1
      NOGEX = NOGEX+NGR
C
C  Calculate the vector product: GRAD * GRAD.
      CALL VVMULT(GRAD,GRAD,GG,NOD)
C
C  Check the gradient criterion for the initial point.
      IF (IGCONV(GRAD,POINT,NOD,VAL,GTOL).NE.0) THEN
         INFO = 2
         RETURN
      END IF
C
C  Set initial step.
      DO 3,I=1,NOD
3        XSTEP(I) = -GRAD(I)
C
C  NSTP is the number of iterations from the previous steepest
C  descent restart.
      NSTP = 0
C
C  ---------------------------------------------------------------------
C  The following loop performs conjugate gradient iterations.
C  ---------------------------------------------------------------------
1000  CONTINUE
      IF (ITMAX.NE.-1 .AND. ITER.GE.ITMAX) THEN
         INFO = 7
         RETURN
      END IF
      IF (NOCEX.GE.NOMAXC) THEN
         INFO = 3
         RETURN
      END IF
      ITER = ITER+1
C
C  Calculate derivative along the XSTEP direction.
      CALL VVMULT(GRAD,XSTEP,G,NOD)
C
C  Predict the value of ALPHA for the line search.
      IF (IFFORM.EQ.0) THEN
C  A general function.
         IF (ITER.EQ.1) THEN
            PRED = 0.1
         ELSE
            PRED = MIN(TWO,MAX(ABS(2.*(VAL-OLDVAL)/G),100.*COMACC))
         END IF
      ELSE
C  Sum of squares.
         PRED = MIN(TWO,ABS(2.*VAL/G))
      END IF
C
C  Check if this is a descent direction. If not, restart.
      IF (G.GT.0.) THEN
         DO 123,I=1,NOD
123         XSTEP(I) = -GRAD(I)
         NSTP = 0
      END IF
C
C  Save current point and value.
      OLDVAL = VAL
      DO 99,I=1,NOD
99       W(I) = POINT(I)
C
C  Perform a line search along the step direction.
C  Here, GNEW and XTEST are used as work arrays.
      ALPHA = PRED
      LSMETH = 2
      CALL QNLS(XSTEP,GRAD,NOCS2,NOG,ALPHA,COMACC,LSMETH,RHO,SIGMA,
     &          LSITER,POINT,VAL,NOD,MINFO,IXAT,MARG,XLL,XRL,GNEW,XTEST)
C
C  Any progress during the line search ?
      IF (ALPHA.EQ.0.) THEN
         INFO = 8
         RETURN
      END IF
      NOCEX = NOCEX+NOCS2
      NOGEX = NOGEX+NOG
      IF (VAL.LT.OLDVAL)
     &    CALL MDIS(POINT,VAL,ITER,NOCEX,NOMAXC,IPRINT,0)
C
C  Did we reach our target value ?
      IF (ITCONV(VAL).NE.0) THEN
         INFO = 1
         RETURN
      END IF
C
C  Check for X-convergence.
      IF (IXCONV(POINT,W,NOD,XTOL).NE.0) THEN
         INFO = 4
         RETURN
      END IF
C
C  Check for F-convergence.
      IF (IFCONV(VAL,OLDVAL,FTOL).NE.0) THEN
         INFO = 5
         RETURN
      END IF
C
C  Calculate derivatives at the point the line search stopped.
      CALL DERMER(POINT,VAL,GNEW,NOCS3,NGR,0)
      NOCEX = NOCEX+NOCS3
      NOGEX = NOGEX+NGR
C
C  Calculate the vector product: GNEW * GNEW.
      CALL VVMULT(GNEW,GNEW,GGNEW,NOD)
C
C  Check the gradient criterion.
      IF (IGCONV(GNEW,POINT,NOD,VAL,GTOL).NE.0) THEN
         INFO = 2
         RETURN
      END IF
C
C  Calculate conjugate gradient updating factors.
      IF (METHOD.EQ.1) THEN
C  ---------------------------------------------------------------------
C  Proceed with the Polak-Ribiere formula.
         BETA = 0.
         DO 7,I=1,NOD
7           BETA = BETA+(GNEW(I)-GRAD(I))*GNEW(I)
         BETA = BETA/GG
         IF (BETA.LE.0.) THEN
            BETA = 0.
            NSTP = 0
         END IF
      ELSE IF (METHOD.EQ.2) THEN
C  ---------------------------------------------------------------------
C  Proceed with the Fletcher-Reeves formula.
         BETA = GGNEW/GG
         IF (BETA.LE.0.) THEN
            BETA = 0.
            NSTP = 0
         END IF
      ELSE
C  ---------------------------------------------------------------------
C  Proceed with the generalized Polak-Ribiere algorithm.
C
C  Compute W = s-bar.
         CALL VVMULT(GNEW,XSTEP,B1,NOD)
         BB = B1/GGNEW
C
         DO 90,I=1,NOD
90          W(I) = XSTEP(I)-BB*GNEW(I)
         CALL VVMULT(W,W,SS,NOD)
         DELTA = MIN(ONE,SQRT(COMACC/SS))
C
         DO 92,I=1,NOD
92          XTEST(I) = POINT(I)-DELTA*W(I)
         FTEST = ACSQ(XTEST,NOD)
         CALL DERMER(XTEST,FTEST,GRAD,NOCS3,NGR,1)
         NOCEX = NOCEX+NOCS3+1
         NOGEX = NOGEX+NGR
C
         A1 = 0.
         A2 = 0.
         DO 94,I=1,NOD
94          A1 = A1+GNEW(I)*(GNEW(I)-GRAD(I))
         CALL VVMULT(GRAD,W,A2,NOD)
         BETA = -A1/A2
C  Should be accept BETA <= 0 ?
         IF (BETA.LE.0.) THEN
            BETA = 0.
            NSTP = 0
         END IF
      END IF
C
C  Calculate the new direction of search.
      IF (IRES.NE.0 .AND. NSTP.GE.NOD-1) THEN
C  Restart every NOD updates.
         DO 88,I=1,NOD
88          XSTEP(I) = -GNEW(I)
         NSTP = 0
      ELSE
C  No N-step restart.
         DO 8,I=1,NOD
8           XSTEP(I) = -GNEW(I)+BETA*XSTEP(I)
         NSTP = NSTP+1
      END IF
C
C  Restore gradient and vector products.
      DO 74,I=1,NOD
74       GRAD(I) = GNEW(I)
      GG = GGNEW
      GOTO 1000
C  ---------------------------------------------------------------------
C  End of conjugate gradient iterations.
C  ---------------------------------------------------------------------
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE CGRTM ( INFO, MESS )
C  ---------------------------------------------------------------------
C
C  Description:
C    Constructs a message, describing the reason for termination of the
C    CONGRA method.
C
C  Input arguments:
C    INFO       Reason for termination of the CONGRA method, as returned
C               by subroutine CGRMV.
C
C  Output arguments:
C    MESS       Message, describing the reason for termination.
C
C  ---------------------------------------------------------------------
C  Arguments:
      CHARACTER*(*) MESS
C  ---------------------------------------------------------------------
C
      IF (INFO.EQ.1) THEN
         MESS = 'Target value has been reached'
      ELSE IF (INFO.EQ.2) THEN
         MESS = 'The gradient criterion is satisfied'
      ELSE IF (INFO.EQ.3) THEN
         MESS = 'All function evaluations have been used'
      ELSE IF (INFO.EQ.4) THEN
         MESS = 'The X-convergence criterion is satisfied'
      ELSE IF (INFO.EQ.5) THEN
         MESS = 'The function value has converged'
      ELSE IF (INFO.EQ.7) THEN
         MESS = 'The specified number of iterations has been reached'
      ELSE IF (INFO.EQ.8) THEN
         MESS = 'Further progress is not possible'
      ELSE IF (INFO.EQ.9) THEN
         MESS = 'All variables are fixed'
      ELSE
         CALL MABORT('Incorrect "info".','CGRTM')
      END IF
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE DFP ( V, MAXW, MERR )
C  ---------------------------------------------------------------------
C
C  Description:
C    Implements the DFP command.
C    Syntax: DFP (Has panel)
C
C  Input arguments:
C    MAXW       Number of available storage positions in array V.
C
C  Output arguments:
C    MERR       Error indicator.
C
C  Input / Output arguments:
C    V          On input the upper triangular part of V may contain an
C               initial approximation to the Choleski L[t] factor of
C               the Hessian matrix. On output contains the upper
C               triangular part contains the Choleski L[t] of the
C               DFP approximation to the Hessian.
C
C  ---------------------------------------------------------------------
      IMPLICIT $(TYPE) (A-H,O-Z)
C  Arguments:
      DIMENSION V(MAXW)
C  ---------------------------------------------------------------------
C  Common blocks:
      PARAMETER ( MXV = $(MXV) )
      COMMON /PARAMS/ POINT(MXV), VAL
      COMMON /DETAIL/ IXAT(MXV), MARG(MXV)
      COMMON /XLIMS/ XLL(MXV), XRL(MXV)
      COMMON /DIMEN/ NOD, NOF
      COMMON /TOTCAL/ ITNOCS, IPARTC, ITDER, IPDER, ITHESS, IPHESS,
     &                ITJAC, IPJAC
      COMMON /MODE3/ MODEV, IFFORM, INMCL
      COMMON /UPDATE/ IGRADU, IHESSU
      COMMON /ACCURA/ COMACC
      COMMON /RWORK/ XT1(MXV), XT2(MXV), XT3(MXV), XT4(MXV), XT5(MXV),
     &               XT6(MXV), XT7(MXV+1)
C  ---------------------------------------------------------------------
C  Local variables:
      CHARACTER*80 MESS
      CHARACTER*10 STR
      SAVE IFIRST
      DATA IFIRST / 0 /
C  ---------------------------------------------------------------------
C
C  Initialize panel parameters that are machine/configuration dependent
C  and cannot be set in the panel description file.
      IF (IFIRST.EQ.0) THEN
         CALL SETPR('XTOL',COMACC,IERR)
         CALL SETPR('GTOL',COMACC,IERR)
         IFIRST = 1
      END IF
C
C  Did the user supply enough storage ?
      NEED = NOD**2
      CALL VSIM('N**2',MAXW,NEED,MERR)
      IF (MERR.NE.0) RETURN
C
C  Set default values for some panel parameters and process any
C  changes the user enters.
      IG = IGRADU
      IH = IHESSU
      CALL SETPI('USEG',IG,IERR)
      CALL SETPI('USEH',IH,IERR)
      CALL SETPI('CANCEL',1,IERR)
      CALL CHANGE(MERR)
      IF (MERR.NE.0) RETURN
      CALL GETPI('CANCEL',ICANC)
      IF (ICANC.NE.1 .OR. ITNOCS.EQ.0 .OR. MODEV.EQ.0) RETURN
C
C  Get parameters from the panel.
      CALL GETPI('NOC',NOC)
      CALL GETPI('USEG',IG)
      CALL GETPI('USEH',IH)
      CALL GETPI('PRINT',IPRINT)
      CALL GETPR('FTOL',FTOL)
      CALL GETPR('XTOL',XTOL)
      CALL GETPR('GTOL',GTOL)
      CALL GETPS('LS',STR,LES)
      IF (ISCOMP(STR,'WEAK').NE.0) THEN
         LSMETH = 1
      ELSE
         LSMETH = 2
      END IF
      CALL GETPI('LSITER',LSITER)
      CALL GETPR('RHO',RHO)
      CALL GETPR('SIGMA',SIGMA)
      CALL GETPI('ITER',ITMAX)
      CALL GETPS('MAD',STR,LES)
      IF (ISCOMP(STR,'ON').NE.0) THEN
         MAD = 1
      ELSE
         MAD = 0
      END IF
C
C  Set automatic derivatives.
      CALL SETADE(MAD)
C
C  Call the minimizer.
      CALL DFPMV(V,POINT,XT5,NOD,VAL,NOC,COMACC,GTOL,ITMAX,XTOL,
     &           LSMETH,RHO,SIGMA,LSITER,FTOL,INFO,ITER,NOCEX,NOGEX,
     &           IPRINT,IXAT,MARG,XLL,XRL,IG,IH,XT1,XT2,XT3,XT4)
C
C  Construct a message describing the reason for termination.
      CALL DFPTM(INFO,MESS)
C
C  Report the reason for termination.
      NOJAC = -1
      CALL MINREP(MESS,NOCEX,NOGEX,NOJAC,ITER)
C
C  Disable automatic derivatives.
      CALL SETADE(0)
C
C  Add backup record.
      CALL BACKUP('M')
C
C  Assign values to MCL-specific keywords.
      CALL SETPI('FCALLS',NOCEX,IERR)
      CALL SETPI('GCALLS',NOGEX,IERR)
      CALL SETPI('ITERDONE',ITER,IERR)
      CALL SETPI('INFO',INFO,IERR)
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE DFPMV ( H, POINT, GRAD, NOD, VAL, NOMAXC, COMACC, GTOL,
     &                  ITMAX, XTOL, LSMETH, RHO, SIGMA, LSITER, FTOL,
     &                  INFO, ITER, NOCEX, NOGEX, IPRINT, IXAT, MARG,
     &                  XLL, XRL, IGRADU, IHESSU, XS, GN, Y, GS )
C  ---------------------------------------------------------------------
C
C  Description:
C    Implementation of the DFP method. Uses Choleski factorization
C    of the Hessian. The Choleski factor L[t] is stored in the upper
C    triangular part of matrix H. Fixed variables are removed, reducing
C    thus the dimensionality of the problem. Margins are treated
C    explicitly (without transformations) by the line search routine.
C    For more information see:
C      Dennis J.E. & Schnabel R.B.
C      Numerical methods for unconstrained optimization and
C      nonlinear equations
C      Prentice Hall, Englewood Cliffs, New Jersey, 1983.
C
C  Input arguments:
C    GRAD       Work array of length NOD.
C    NOD        Dimensionality of the objective function.
C    NOMAXC     Maximum allowed number of calls to the objective
C               function.
C    COMACC     Machine accuracy.
C    GTOL       G-convergence criterion.
C    ITMAX      Maximum allowed number of DFP iterations.
C               Set ITMAX = -1 for an unlimited number of iterations.
C    XTOL       X-convergence criterion.
C    LSMETH     Line search method. See subroutine LNS.
C    RHO        The line search "rho". See subroutine LNS.
C    SIGMA      The line search "sigma". See subroutine LNS.
C    LSITER     Maximum allowed number of line search iterations. See
C               subroutine LSITER.
C    FTOL       F-convergence criterion.
C    IPRINT     Printout level:
C               IPRINT = 0 -> Nothing is printed.
C               IPRINT = 1 -> Simple printout: Prints ITER, VAL,
C                             NOC, MAXNOC
C               IPRINT = 2 -> Full printout. Prints in addition the
C                             minimization parameters, contained in
C                             array POINT.
C    IXAT       Fix statuses.
C    MARG       Margin statuses.
C    XLL        Left margins.
C    XRL        Right margins.
C    IGRADU     IGRADU = 0 -> Ask subroutine DERMER to recalculate
C                             the gradient at the current point. If
C                             a valid cached gradient exists, DERMER
C                             will return it.
C               IGRADU = 1 -> Ask subroutine DERMER to return its
C                             cached gradient. No recalculation
C                             takes place, unless of course the
C                             gradient cache is empty.
C    IHESSU     IHESSU = 0 -> The Hessian will be initialized to the
C                             identity matrix.
C               IHESSU = 1 -> The upper triangular part of H contains
C                             the Choleski factor L[t] of a matrix,
C                             to be used as an initial approximation
C                             to the Hessian.
C    XS         Work array of length NOD.
C    GN         Work array of length NOD.
C    Y          Work array of length NOD.
C    GS         Work array of length NOD.
C
C  Output arguments:
C    INFO       Reason for termination:
C               INFO = 1 -> Target value has been reached.
C               INFO = 2 -> The gradient criterion is satisfied.
C               INFO = 3 -> All function calls have been exhausted.
C               INFO = 4 -> X-convergence has been achieved.
C               INFO = 5 -> F-convergence has occurred.
C               INFO = 6 -> Division by zero while updating the
C                           Hessian matrix.
C               INFO = 7 -> Iterations exhausted.
C               INFO = 8 -> All variables are fixed.
C               INFO = 9 -> The line search cannot proceed any further.
C    ITER       Number of DFP iterations that were performed.
C    NOCEX      Number of function evaluations that were performed.
C    NOGEX      Number of gradient evaluations that were performed.
C
C  Input / Output arguments:
C    H          On input the upper triangular part of H may contain an
C               initial approximation to the Choleski L[t] factor of
C               the Hessian matrix. On output contains the upper
C               triangular part contains the Choleski L[t] of the
C               DFP approximation to the Hessian.
C    POINT      On input, the current point. On output, the new point
C               after the minimization.
C    VAL        Value of the objective function.
C
C  ---------------------------------------------------------------------
      IMPLICIT $(TYPE) (A-H,O-Z)
C  Arguments:
      DIMENSION H(NOD,NOD), POINT(NOD), GRAD(NOD)
      DIMENSION IXAT(NOD), MARG(NOD), XLL(NOD), XRL(NOD)
      DIMENSION XS(NOD), GN(NOD), Y(NOD), GS(NOD)
C  ---------------------------------------------------------------------
C
C  Number of iterations performed so far.
      ITER = 0
C
C  Number of function calls so far.
      NOCEX = 0
C
C  Number of (analytic) gradient calls so far.
      NOGEX = 0
C
C  Did we reach our target value ?
      IF (ITCONV(VAL).NE.0) THEN
         INFO = 1
         RETURN
      END IF
C
C  Count non-fixed variables. M will be used as the dimensionality
C  of the objective function in all subsequent matrix/vector operations.
      CALL FREEVA(M)
      IF (M.EQ.0) THEN
         INFO = 8
         RETURN
      END IF
C
C  Calculate derivatives at the starting point.
      IF (IGRADU.EQ.0) THEN
         MODE = 0
      ELSE
         MODE = 2
      END IF
      CALL DERMER(POINT,VAL,GRAD,N1,N2,MODE)
      NOCEX = NOCEX+N1
      NOGEX = NOGEX+N2
C
C  Copy derivatives to an array of M elements.
      MM = 0
      DO 17,I=1,NOD
         IF (IXAT(I).NE.0) THEN
            MM = MM+1
            GS(MM) = GRAD(I)
         END IF
17    CONTINUE
C
C  Check the gradient criterion for the initial point.
C  The numerical differentiation routine DERMER is expected to
C  return GRAD(I) = 0 for all fixed variables. Otherwise the
C  following test will fail, since the whole array GRAD is examined.
      IF (IGCONV(GRAD,POINT,NOD,VAL,GTOL).NE.0) THEN
         INFO = 2
         RETURN
      END IF
C
C  Initialize Hessian. (Upper triangular part only.)
      IF (IHESSU.EQ.0) THEN
         DO 10,J=1,M
            H(J,J) = 1.
            DO 20,I=1,J-1
20             H(I,J) = 0.
10       CONTINUE
      END IF
C  ---------------------------------------------------------------------
C  The following loop performs DFP iterations.
C  ---------------------------------------------------------------------
1000  CONTINUE
      IF (ITMAX.NE.-1 .AND. ITER.GE.ITMAX) THEN
         INFO = 7
         RETURN
      END IF
      IF (NOCEX.GE.NOMAXC) THEN
         INFO = 3
         RETURN
      END IF
      ITER = ITER+1
C
C  Solve H * s = -g for the search direction XS. The minus sign will
C  be added later.
      CALL LLSOLV(H,XS,GS,M,NOD)
C
C  Save current point and value for the convergence tests.
      OLDVAL = VAL
      DO 99,I=1,NOD
99       GN(I) = POINT(I)
C
C  Prepare a vector of length NOD with the proper line search
C  direction.
      MM = 0
      DO 71,I=1,NOD
         IF (IXAT(I).NE.0) THEN
            MM = MM+1
            GS(I) = -XS(MM)
         ELSE
            GS(I) = 0.
         END IF
71    CONTINUE
C
C  Initial estimate for the line search.
      ALPHA = 1.
C
C  Perform the line search. Here, XS and Y are used as work arrays.
      CALL QNLS(GS,GRAD,N1,N2,ALPHA,COMACC,LSMETH,RHO,SIGMA,LSITER,
     &          POINT,VAL,NOD,MINFO,IXAT,MARG,XLL,XRL,XS,Y)
      NOCEX = NOCEX+N1
      NOGEX = NOGEX+N2
C
C  Any progress during the line search ?
      IF (ALPHA.EQ.0.) THEN
         INFO = 9
         RETURN
      END IF
      IF (VAL.LT.OLDVAL)
     &    CALL MDIS(POINT,VAL,ITER,NOCEX,NOMAXC,IPRINT,0)
C
C  Did we reach our target value ?
      IF (ITCONV(VAL).NE.0) THEN
         INFO = 1
         RETURN
      END IF
C
C  Check for X-convergence.
      IF (IXCONV(POINT,GN,NOD,XTOL).NE.0) THEN
         INFO = 4
         RETURN
      END IF
C
C  Check for F-convergence.
      IF (IFCONV(VAL,OLDVAL,FTOL).NE.0) THEN
         INFO = 5
         RETURN
      END IF
C
C  Calculate derivatives at the point the line search stopped.
      CALL DERMER(POINT,VAL,GN,N1,N2,0)
      NOCEX = NOCEX+N1
      NOGEX = NOGEX+N2
C
C  Check the gradient criterion.
      IF (IGCONV(GN,POINT,NOD,VAL,GTOL).NE.0) THEN
         INFO = 2
         RETURN
      END IF
C
C  Calculate the difference in the gradient: Y = g[k+1]-g[k]
C  and in the variables: XS = x[k+1] - x[k]
C  Update derivatives in GRAD.
      MM = 0
      DO 13,I=1,NOD
         IF (IXAT(I).NE.0) THEN
            MM = MM+1
            XS(MM) = ALPHA*GS(I)
            Y(MM) = GN(I)-GRAD(I)
            GS(MM) = GN(I)
         END IF
         GRAD(I) = GN(I)
13    CONTINUE
C
C  Calculate vector inner products needed to determine whether the
C  Hessian can be updated.
      CALL VVMULT(Y,Y,YTY,M)
      CALL VVMULT(XS,XS,DTD,M)
      CALL VVMULT(Y,XS,YTS,M)
C
C  Can we update the Hessian ?
      IF (YTS.GT.SQRT(COMACC*DTD*YTY)) THEN
C
C  GN = L**(-1) * y
C  TT = y[t] * H**(-1) * y
         CALL LMSOLV(H,GN,Y,M,NOD)
         CALL VVMULT(GN,GN,TT,M)
         IF (TT.EQ.0.) THEN
            INFO = 6
            RETURN
         END IF
C
C  XS = L[t] * s
         CALL LTMULT(H,XS,XS,M,NOD)
         AA = SQRT(YTS/TT)
         DO 57,I=1,M
57          GN(I) = (AA*GN(I)-XS(I))/YTS
C
C  Update Choleski factors.
         CALL LLUPD(H,Y,GN,M,NOD)
      END IF
      GOTO 1000
C  ---------------------------------------------------------------------
C  End of DFP iterations.
C  ---------------------------------------------------------------------
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE DFPTM ( INFO, MESS )
C  ---------------------------------------------------------------------
C
C  Description:
C    Constructs a message, describing the reason for termination of the
C    DFPS method.
C
C  Input arguments:
C    INFO       Reason for termination of the DFP method, as returned
C               by subroutine DFPMV.
C
C  Output arguments:
C    MESS       Message, describing the reason for termination.
C
C  ---------------------------------------------------------------------
C  Arguments:
      CHARACTER*(*) MESS
C  ---------------------------------------------------------------------
C
      IF (INFO.EQ.1) THEN
         MESS = 'Target value has been reached'
      ELSE IF (INFO.EQ.2) THEN
         MESS = 'The gradient criterion is satisfied'
      ELSE IF (INFO.EQ.3) THEN
         MESS = 'All function evaluations have been used'
      ELSE IF (INFO.EQ.4) THEN
         MESS = 'The X-convergence criterion is satisfied'
      ELSE IF (INFO.EQ.5) THEN
         MESS = 'The function value has converged'
      ELSE IF (INFO.EQ.6) THEN
         MESS = 'The Hessian matrix cannot be updated'
      ELSE IF (INFO.EQ.7) THEN
         MESS = 'The specified number of iterations has been reached'
      ELSE IF (INFO.EQ.8) THEN
         MESS = 'All variables are fixed'
      ELSE IF (INFO.EQ.9) THEN
         MESS = 'Further progress is not possible'
      ELSE
         CALL MABORT('Incorrect "info".','DFPTM')
      END IF
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE DIRLIM ( POINT, DIR, MARG, XLL, XRL, NOD, LIM, XALIM,
     &                    XBLIM )
C  ---------------------------------------------------------------------
C
C  Description:
C    Determines margins along the search direction DIR.
C
C  Input arguments:
C    POINT      The current point.
C    DIR        The seearch direction.
C    MARG       Margin statuses.
C    XLL        Left margins.
C    XRL        Right margins.
C    NOD        Dimensionality of the objective function.
C
C  Output arguments:
C    LIM        LIM =  0 -> No margins along DIR.
C               LIM = -1 -> A left margin exists along DIR.
C               LIM =  1 -> A right margin exists along DIR.
C               LIM =  2 -> Both margins exist along the direction DIR.
C    XALIM      Left margin along the direction DIR.
C    XBLIM      Right margin along the direction DIR.
C
C  ---------------------------------------------------------------------
      IMPLICIT $(TYPE) (A-H,O-Z)
C  Arguments:
      DIMENSION POINT(NOD), DIR(NOD), MARG(NOD), XLL (NOD), XRL(NOD)
C  ---------------------------------------------------------------------
C
C  Get machine constants.
      CALL GETMC(BIGNUM,DWARF)
C
C  Find left limit along the search direction.
      NL = 0
      XALIM = -BIGNUM
      DO 10,I=1,NOD
         MA = MARG(I)
         IF (DIR(I).EQ.0. .OR. MA.EQ.0) GOTO 10
         IF (DIR(I).GT.0) THEN
            IF (MA.EQ.-1 .OR. MA.EQ.2) THEN
               XTLIM = (XLL(I)-POINT(I))/DIR(I)
               NL = NL+1
            ELSE
               XTLIM = -BIGNUM
            END IF
         ELSE
            IF (MA.EQ.1 .OR. MA.EQ.2) THEN
               XTLIM = (XRL(I)-POINT(I))/DIR(I)
               NL = NL+1
            ELSE
               XTLIM = -BIGNUM
            END IF
         END IF
         IF (XTLIM.GT.XALIM) XALIM = XTLIM
10    CONTINUE
C
C  Find right limit along the search direction.
      NR = 0
      XBLIM = BIGNUM
      DO 20,I=1,NOD
         MA = MARG(I)
         IF (DIR(I).EQ.0. .OR. MA.EQ.0) GOTO 20
         IF (DIR(I).GT.0.) THEN
            IF (MA.EQ.1 .OR. MA.EQ.2) THEN
               XTLIM = (XRL(I)-POINT(I))/DIR(I)
               NR = NR+1
            ELSE
               XTLIM = BIGNUM
            END IF
         ELSE
            IF (MA.EQ.-1 .OR. MA.EQ.2) THEN
               XTLIM = (XLL(I)-POINT(I))/DIR(I)
               NR = NR+1
            ELSE
               XTLIM = BIGNUM
            END IF
         END IF
         IF (XTLIM.LT.XBLIM) XBLIM = XTLIM
20    CONTINUE
C
C  LIM indicates whether margins exist along the search direction.
      IF (NL.EQ.0 .AND. NR.EQ.0) THEN
         LIM = 0
      ELSE IF (NL.EQ.0) THEN
         LIM = 1
      ELSE IF (NR.EQ.0) THEN
         LIM = -1
      ELSE
         LIM = 2
      END IF
C
C  Left margin should always be negative, right margin positive.
C  This is not always true due to small round off errors in the
C  calculation of XALIM and XBLIM.
      IF ((LIM.EQ.-1 .OR. LIM.EQ.2) .AND. XALIM.GT.0.) XALIM = 0.
      IF ((LIM.EQ. 1 .OR. LIM.EQ.2) .AND. XBLIM.LT.0.) XBLIM = 0.
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE TTDRI ( DELTA, M, GS, H, LD, RHO, VAL, XTOL, NOCS,
     &                   POINT, NOD, IXAT, STEP, SN, XNEW, IRET, FUN )
C  ---------------------------------------------------------------------
C
C  Description:
C    Implements the trust region approach. It calculates the Cauchy
C    and Newton points, takes the appropriate step, and updates the
C    trust region radius accordingly.
C    For more information see:
C      Dennis J.E. & Schnabel R.B.
C      Numerical methods for unconstrained optimization and
C      nonlinear equations
C      Prentice Hall, Englewood Cliffs, New Jersey, 1983. pp 139-147.
C
C  Input arguments:
C    M          Number of non-fixed variables. This is used as the
C               dimensionality of the problem.
C    GS         The gradient vector for the M non-fixed variables.
C    H          Contains the Choleski factor L[t] on the upper
C               triangular part.
C    LD         Leading dimension of matrix H.
C    RHO        The rho-line parameter.
C    XTOL       X-convergence criterion.
C    NOD        Dimensionality of the objective function.
C    IXAT       Fix statuses.
C    STEP       Work array of length NOD.
C    SN         Work array of length NOD.
C    XNEW       Work array of length NOD.
C
C  Output arguments:
C    NOCS       Number of function calls.
C    IRET       Reason for return:
C               IRET = 0 -> A new point has been successfully found.
C               IRET = 1 -> The step to be taken is too small. (It will
C                           not result in a new point.)
C
C  Input / Output arguments:
C    DELTA      On input the trust region radius. On output, the new,
C               updated radius.
C    VAL        Value of the objective function.
C    POINT      On input the current point. On output, the new, updated
C               point.
C
C  ---------------------------------------------------------------------
      IMPLICIT $(TYPE) (A-H,O-Z)
C  Arguments:
      DIMENSION POINT(NOD), IXAT(NOD), GS(M), H(LD,M)
      DIMENSION STEP(M), SN(M), XNEW(NOD)
      $(TYPE) FUN
      EXTERNAL FUN
C  ---------------------------------------------------------------------
C  Local variables:
      LOGICAL FIRST
C  ---------------------------------------------------------------------
C
C  Number of function calls.
      NOCS = 0
C
C  Solve H * s = -g for the Newton step SN. The minus sign will be
C  added later.
      CALL LLSOLV(H,SN,GS,M,LD)
C
C  SNLEN is the Newton step length.
      CALL VVMULT(SN,SN,SNLEN,M)
      SNLEN = SQRT(SNLEN)
C
      FIRST = .TRUE.
1000  CONTINUE
      IF (SNLEN.LE.DELTA) THEN
         DO 20,I=1,M
20          STEP(I) = -SN(I)
         IDEL = 0
      ELSE
         IF (FIRST) THEN
            FIRST = .FALSE.
C
C  GHG = g[t] * H * g
C  GG = g[t] * g
            CALL LLMULT(H,GS,GHG,M,LD)
            CALL VVMULT(GS,GS,GG,M)
C
C  Cauchy step: SC = -al * g,  al = (g*g)/(g*H*g)
            ALC = GG/GHG
C
C  SCLEN is the Cauchy step length.
            SCLEN = ALC*SQRT(GG)
            ALC = -ALC
            CALL VVMULT(GS,SN,GSN,M)
            AN = 0.2+0.8*GG**2/(GHG*GSN)
         END IF
         CALL DOGLEG(DELTA,GS,STEP,M,SN,SNLEN,ALC,SCLEN,AN,GSN,GG)
         IDEL = 1
      END IF
      CALL TRUPD(STEP,M,RHO,VAL,H,LD,DELTA,XTOL,IRET,FNEW,GS,POINT,
     &           NOD,IXAT,XNEW,IDEL,FUN)
      NOCS = NOCS+1
      IF (IRET.EQ.2) GOTO 1000
C
      IF (IRET.EQ.0 .OR. FNEW.LT.VAL) THEN
         DO 10,I=1,NOD
10          POINT(I) = XNEW(I)
         VAL = FNEW
      END IF
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE DOGLEG ( DELTA, GS, STEP, M, SN, SNLEN, ALC, SCLEN, AN,
     &                    GSN, GG )
C  ---------------------------------------------------------------------
C
C  Description:
C    Calculates the double dogleg step.
C
C  Input arguments:
C    DELTA      Number of variables of the objective function.
C    GS         The gradient vector for the M non-fixed variables.
C    M          Number of non-fixed variables.
C    SN         The Newton step.
C    SNLEN      The Newton step length.
C    ALC        The Cauchy step is ALC * GS.
C    SCLEN      The Cauchy step length.
C    AN         Distance along the Newton direction that defines
C               the last part of the double dogleg.
C    GSN        The vector product GS * SN.
C    GG         The vector product GS * GS.
C
C  Output arguments:
C    STEP       The double dogleg step.
C
C  ---------------------------------------------------------------------
      IMPLICIT $(TYPE) (A-H,O-Z)
C  Arguments:
      DIMENSION GS(M), STEP(M), SN(M)
C  ---------------------------------------------------------------------
C
      IF (AN*SNLEN.LE.DELTA) THEN
C  Take only a partial Newton step. (Point SN is inside the radius)
         C = -DELTA/SNLEN
         DO 50,I=1,M
50          STEP(I) = C*SN(I)
      ELSE IF (SCLEN.GE.DELTA) THEN
C  Take only a partial steepest descent step. (CP is out of the radius)
         C = -DELTA/SQRT(GG)
         DO 60,I=1,M
60          STEP(I) = C*GS(I)
      ELSE
C  Take the double dogleg step.
         USC = -ALC*(AN*GSN+ALC*GG)
         UU = 2*AN*ALC*GSN+(AN*SNLEN)**2+ALC**2*GG
         C = (-USC+SQRT(USC**2-UU*(SCLEN**2-DELTA**2)))/UU
         C1 = ALC*(1.-C)
         C2 = C*AN
         DO 80,I=1,M
80          STEP(I) = C1*GS(I)-C2*SN(I)
      END IF
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE TRUPD ( STEP, M, RHO, VAL, H, LD, DELTA, XTOL, IRET,
     &                   FNEW, GS, POINT, NOD, IXAT, XNEW, IDEL, FUN )
C  ---------------------------------------------------------------------
C
C  Description:
C    Updates the trust region radius.
C
C  Input arguments:
C    STEP       The step to be taken.
C    M          Number of non-fixed variables.
C    RHO        The rho-line parameter.
C    VAL        Current value of the objective function.
C    H          Choleski factor L[t] of the Hessian matrix.
C    LD         Leading dimension of matrix H.
C    XTOL       X-convergence criterion. Defines the minimum
C               acceptable step length.
C    GS         The gradient vector for the M non-fixed variables.
C    POINT      The current point.
C    NOD        Dimensionality of the objective function.
C    IXAT       Fix statuses.
C    IDEL       IDEL = 0 -> STEP is the full Newton step.
C               IDEL = 1 -> STEP is the double dogleg step.
C
C  Output arguments:
C    IRET       Reason for return:
C               IRET = 0 -> A new, lower function value has been found,
C                           and the trust region radius has been
C                           updated.
C               IRET = 1 -> The step to be taken is too small.
C               IRET = 2 -> Failed to produce a lower function value.
C                           The trust region radius has been updated
C                           however. The calling routine must now
C                           calculate a new STEP and call this routine
C                           once more.
C    FNEW       The new function value. (Not necessarily lower than
C               VAL.)
C    XNEW       The new point.
C
C  Input / Output arguments:
C    DELTA      On input the current trust region radius. On output, the
C               new, updated radius.
C
C  ---------------------------------------------------------------------
      IMPLICIT $(TYPE) (A-H,O-Z)
C  Arguments:
      DIMENSION POINT(NOD), IXAT(NOD), XNEW(NOD)
      DIMENSION STEP(M), GS(M), H(LD,M)
C  ---------------------------------------------------------------------
C
      CALL VVMULT(STEP,STEP,STPLEN,M)
      STPLEN = SQRT(STPLEN)
      CALL VVMULT(GS,STEP,G0,M)
C
C  XNEW is the new point.
      MM = 0
      DO 10,I=1,NOD
         IF (IXAT(I).EQ.0) THEN
            XNEW(I) = POINT(I)
         ELSE
            MM = MM+1
            XNEW(I) = POINT(I)+STEP(MM)
         END IF
10    CONTINUE
C
C  This is normally neccesary only if FNEW is above the rho-line.
C  (A backtrack is in need).
      ONE = 1.
      XMAX = 0.
      MM = 0
      DO 40,I=1,NOD
         IF (IXAT(I).NE.0) THEN
            MM = MM+1
            XMAX = MAX(XMAX,ABS(STEP(MM))/MAX(ABS(XNEW(I)),ONE))
         END IF
40    CONTINUE
      IF (XMAX.LE.XTOL) THEN
         FNEW = VAL
         IRET = 1
         RETURN
      END IF
C
C  Evaluation of the objective function occurs only in the following
C  statement.
      FNEW = FUN(XNEW,NOD)
      DF = FNEW-VAL
      IF (DF.GE.RHO*G0) THEN
C
C  Interpolate for the new DELTA.
         DD = -G0*STPLEN/(2.*(DF-G0))
         IF (DD.LT.0.1*DELTA) THEN
            DELTA = 0.1*DELTA
         ELSE IF (DD.GT.0.5*DELTA) THEN
            DELTA = 0.5*DELTA
         ELSE
            DELTA = DD
         END IF
         IRET = 2
      ELSE
C
C  Update the trust region radius according to the predicted and
C  actual reductions in the function value.
C  SHS = s[t] * H * s
         CALL LLMULT(H,STEP,SHS,M,LD)
         DFPRED = G0+SHS/2.
         IF (DF.GT.0.25*DFPRED) THEN
            DELTA = DELTA/2.
         ELSE IF (DF.LT.0.75*DFPRED) THEN
            IF (IDEL.NE.0) DELTA = 2.*DELTA
         END IF
         IRET = 0
      END IF
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE GTRANS ( G, X )
C  ---------------------------------------------------------------------
C
C  Description:
C    This subroutine returns the gradient vector with respect to
C    a set of transformed variables.
C    This is necessary for minimization methods that handle bounds via
C    transformations.
C    See also subroutines XTOY and YTOX.
C
C  Input arguments:
C    X          The transformed variables.
C
C  Input / Output arguments:
C    G          On input the gradient as calculated by
C               SUBROUTINE DERMER.
C               On output the gradient with respect to the transformed
C               variables.
C
C  ---------------------------------------------------------------------
      IMPLICIT $(TYPE) (A-H,O-Z)
C  Arguments:
      DIMENSION G(*), X(*)
C  ---------------------------------------------------------------------
C  Common blocks:
      PARAMETER ( MXV = $(MXV) )
      COMMON /DETAIL/ IXAT(MXV), MARG(MXV)
      COMMON /XLIMS/ XLL(MXV), XRL(MXV)
      COMMON /DIMEN/ NOD, NOF
C  ---------------------------------------------------------------------
C
      TWO = 2.
      DO 10,I=1,NOD
         IF (IXAT(I).NE.0) THEN
            MA = MARG(I)
            IF (MA.EQ.-1) THEN
               G(I) = TWO*X(I)*G(I)
            ELSE IF (MA.EQ.1) THEN
               G(I) = -TWO*X(I)*G(I)
            ELSE IF (MA.EQ.2) THEN
               XI = X(I)
               G(I) = -TWO*XI*(XRL(I)-XLL(I))*EXP(-XI**2)*G(I)
            END IF
         END IF
10    CONTINUE
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE QMIN ( X1, F1, G1, X2, F2, NR, XMIN )
C  ---------------------------------------------------------------------
C
C  Description:
C    Returns the minimizing point of the parabola that interpolates
C    (X1,F1,G1) and (X2,F2).
C
C  Input arguments:
C    X1         Coodrinate of the first point.
C    F1         Function value at X1.
C    G1         Gradient at X1.
C    X2         Coordinate of the second point.
C    F2         Function value at X2.
C
C  Output arguments:
C    NR         NR = 1 if a minimizer was found; NR = 0 otherwise.
C    XMIN       Minimizer of the parabola.
C
C  ---------------------------------------------------------------------
      IMPLICIT $(TYPE) (A-H,O-Z)
C
      DX = X2-X1
      DF = F2-F1
      B = (DF/DX-G1)/DX
C
      IF (B.GT.0) THEN
         XMIN = X1 - G1/(2*B)
         NR = 1
      ELSE
         NR = 0
      END IF
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE CMIN ( X1, F1, G1, X2, F2, G2, NR, XMIN )
C  ---------------------------------------------------------------------
C
C  Description:
C    Returns the minimizing point of the cubic that interpolates
C    (X1,F1,G1) and (X2,F2,G2).
C
C  Input arguments:
C    X1         Coodrinate of the first point.
C    F1         Function value at X1.
C    G1         Gradient at X1.
C    X2         Coordinate of the second point.
C    F2         Function value at X2.
C    G2         Gradient at X2.
C
C  Output arguments:
C    NR         NR = 1 if a minimizer was found; NR = 0 otherwise.
C    XMIN       Minimizer of the cubic polynomial.
C
C  ---------------------------------------------------------------------
      IMPLICIT $(TYPE) (A-H,O-Z)
C
      H = X2-X1
      A = (G2+G1)/H**2 - 2*(F2-F1)/H**3
      B = (3*(F2-F1) - H*(G2+2*G1))/H**2
      D = B**2-3*A*G1
      IF (D.GE.0) THEN
         IF (A.NE.0) THEN
            XMIN = X1+(-B+SQRT(D))/(3*A)
            NR = 1
         ELSE
            IF (B.GT.0) THEN
               XMIN = X1 - G1/(2*B)
               NR = 1
            ELSE
               NR = 0
            END IF
         END IF
      ELSE
         NR = 0
      END IF
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE QMIN2 ( X1, F1, X2, F2, X3, F3, NR, XMIN )
C  ---------------------------------------------------------------------
C
C  Description:
C    Returns the minimizing point of the parabola that interpolates
C    (X1,F1), (X2,F2) and (X3,F3).
C
C  Input arguments:
C    X1         Coodrinate of the first point.
C    F1         Function value at X1.
C    X2         Coordinate of the second point.
C    F2         Function value at X2.
C    X3         Coordinate of the third point.
C    F3         Function value at X3.
C
C  Output arguments:
C    NR         NR = 1 if a minimizer was found; NR = 0 otherwise.
C    XMIN       Minimizer of the parabola.
C
C  ---------------------------------------------------------------------
      IMPLICIT $(TYPE) (A-H,O-Z)
C
      D1 = X2-X1
      D2 = X1-X3
      D3 = X3-X2
C
      DF1 = F2-F1
      DF2 = F3-F2
C
      A=(DF1/D1-DF2/D3)/D2
      IF (A.GT.0.) THEN
         ANU = DF1*D3**2 + DF2*D1**2
         DEN = DF1*D3 - DF2*D1
         XMIN = X2 + ANU/(2.*DEN)
         NR = 1
      ELSE
         NR = 0
      END IF
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE CMIN2 ( X1, F1, G1, X2, F2, X3, F3, NR, XMIN )
C  ---------------------------------------------------------------------
C
C  Description:
C    Returns the minimizing point of the cubic that interpolates
C    (X1,F1,G1), (X2,F2) and (X3,F3).
C
C  Input arguments:
C    X1         Coodrinate of the first point.
C    F1         Function value at X1.
C    G1         Gradient at X1.
C    X2         Coordinate of the second point.
C    F2         Function value at X2.
C    X3         Coordinate of the third point.
C    F3         Function value at X3.
C
C  Output arguments:
C    NR         NR = 1 if a minimizer was found; NR = 0 otherwise.
C    XMIN       Minimizer of the cubic polynomial.
C
C  ---------------------------------------------------------------------
      IMPLICIT $(TYPE) (A-H,O-Z)
C
      W1 = (F2-F1-(X2-X1)*G1)/(X2-X1)**2
      W2 = (F3-F1-(X3-X1)*G1)/(X3-X1)**2
      A = (W2-W1)/(X3-X2)
      B = (X2-X1)*W2/(X2-X3)-(X3-X1)*W1/(X2-X3)
      D = B**2-3*A*G1
      IF (D.GE.0) THEN
         IF (A.NE.0) THEN
            XMIN = X1+(-B+SQRT(D))/(3*A)
            NR = 1
         ELSE
            IF (B.GT.0) THEN
               XMIN = X1-G1/(2*B)
               NR = 1
            ELSE
               NR = 0
            END IF
         END IF
      ELSE
         NR = 0
      END IF
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE JTRANS ( FJ, X, LD )
C  ---------------------------------------------------------------------
C
C  Description:
C    This subroutine returns the Jacobian matrix with respect to
C    a set of transformed variables.
C    This is necessary for minimization methods that handle bounds via
C    transformations.
C    See also subroutines XTOY and YTOX.
C
C  Input arguments:
C    LD         The leading dimension of matrix FJ.
C    X          The transformed variables.
C
C  Input / Output arguments:
C    FJ         On input the Jacobian matrix as calculated by
C               SUBROUTINE MJACOB.
C               On output the Jacobian matrix with respect to the
C               transformed variables.
C
C  ---------------------------------------------------------------------
      IMPLICIT $(TYPE) (A-H,O-Z)
C  Arguments:
      DIMENSION FJ(LD,*), X(*)
C  ---------------------------------------------------------------------
C  Common blocks:
      PARAMETER ( MXV = $(MXV) )
      COMMON /DETAIL/ IXAT(MXV), MARG(MXV)
      COMMON /XLIMS/ XLL(MXV), XRL(MXV)
      COMMON /DIMEN/ NOD, NOF
C  ---------------------------------------------------------------------
C
      TWO = 2.
      DO 10,I=1,NOD
         IF (IXAT(I).NE.0) THEN
            MA = MARG(I)
            IF (MA.EQ.-1) THEN
               CON = TWO*X(I)
               DO 20,J=1,NOF
20                FJ(J,I) = CON*FJ(J,I)
            ELSE IF (MA.EQ.1) THEN
               CON = -TWO*X(I)
               DO 30,J=1,NOF
30                FJ(J,I) = CON*FJ(J,I)
            ELSE IF (MA.EQ.2) THEN
               XI = X(I)
               CON = -TWO*XI*(XRL(I)-XLL(I))*EXP(-XI**2)
               DO 40,J=1,NOF
40                FJ(J,I) = CON*FJ(J,I)
            END IF
         END IF
10    CONTINUE
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE LEVE ( V, MAXW, MERR )
C  ---------------------------------------------------------------------
C
C  Description:
C    Implements the LEVE command.
C    Syntax: LEVE (Has panel)
C
C  Input arguments:
C    V          Work array. Should have at least NOD*NOF elements.
C    MAXW       Number of available storage positions in array V.
C
C  Output arguments:
C    MERR       Error indicator.
C
C  ---------------------------------------------------------------------
      IMPLICIT $(TYPE) (A-H,O-Z)
C  Arguments:
      DIMENSION V(MAXW)
C  ---------------------------------------------------------------------
C  Common blocks:
      COMMON /TOTCAL/ ITNOCS, IPARTC, ITDER, IPDER, ITHESS, IPHESS,
     &                ITJAC, IPJAC
      COMMON /MODE3/ MODEV, IFFORM, INMCL
      COMMON /DIMEN/ NOD, NOF
      LOGICAL STRO, WEAK
      COMMON /PRILOG/ STRO, WEAK
      COMMON /UNITS/ IUINP, IUOUT
C  ---------------------------------------------------------------------
C  Local variables:
      CHARACTER*80 MESS
C  ---------------------------------------------------------------------
C
C  Did the user supply enough storage ?
      NEED = NOD*NOF
      CALL VSIM('N*M',MAXW,NEED,MERR)
      IF (MERR.NE.0) RETURN
C
      CALL SETPI('CANCEL',1,IERR)
      CALL CHANGE(MERR)
      IF (MERR.NE.0) RETURN
      CALL GETPI('CANCEL',ICANC)
      IF (ICANC.NE.1 .OR. ITNOCS.EQ.0 .OR. MODEV.EQ.0) RETURN
C
C  Get panel parameters.
      CALL GETPI('NOC',NOCL)
      CALL GETPR('FTOL',FTOLL)
      CALL GETPR('XTOL',XTOLL)
      CALL GETPR('GTOL',GTOLL)
      CALL GETPI('PRINT',IPRINT)
C
C  Do we have a sum-of-squares function ?
      IF (IFFORM.EQ.0) THEN
         IF (WEAK) WRITE (IUOUT,20)
         MERR = -2
         RETURN
      END IF
C
C  Call the minimizer.
      IFST = ITNOCS
      IJST = ITJAC
      CALL LEVE1(V,NOCL,IPRINT,GTOLL,FTOLL,XTOLL,NOCX,INFO,ITER)
C
C  Construct a message describing the reason for termination.
      CALL LEVETM(INFO,MESS)
C
C  Report the reason for termination.
      NOCEX = ITNOCS-IFST
      NOGEX = -1
      NOJAC = ITJAC-IJST
      CALL MINREP(MESS,NOCEX,NOGEX,NOJAC,ITER)
C
C  Add backup record.
      CALL BACKUP('M')
C
C  Assign values to MCL-specific keywords.
      CALL SETPI('FCALLS',NOCEX,IERR)
      CALL SETPI('JCALLS',NOJAC,IERR)
      CALL SETPI('ITERDONE',ITER,IERR)
      CALL SETPI('INFO',INFO,IERR)
C
20    FORMAT (/$(FC)'You can use this method only on a ',
     &        'Sum-of-Squares type of function.')
      END
C  ---------------------------------------------------------------------
      SUBROUTINE LEVE1 ( W, MAXFEV, IPRINT, GTOL, FTOL, XTOL, NFEV,
     &                   INFO, ITER )
C  ---------------------------------------------------------------------
C
C  Description:
C    Front end to the MINPACK-1 routine LMDIF. It uses transformations
C    to handle margins and fixed variables.
C    For information on the MINPACK-1 package see:
C      Jorge J. More, Burton S. Garbow and  Kenneth E. Hillstrom
C      User Guide for MINPACK-1
C      ANL-80-74, Argonne National Laboratory, Argonne, Illinois, 1980.
C
C  Input arguments:
C    W          Work array. It should be have at least NOD * NOF
C               elements.
C
C  The following arguments are passed directly to subroutine LMDIF:
C    MAXFEV, IPRINT, GTOL, FTOL, XTOL.
C
C  Output arguments:
C    NFEV       Number of function evaluations that were performed by
C               subroutine LMDIF.
C    INFO       Reason for termination, as described in subroutine
C               LMDIF.
C    ITER       Number of iterations that were performed by subroutine
C               LMDIF.
C
C  ---------------------------------------------------------------------
      IMPLICIT $(TYPE) (A-H,O-Z)
C  Arguments:
      DIMENSION W(*)
C  ---------------------------------------------------------------------
C  Common blocks:
      PARAMETER ( MXT = $(MXT) )
      PARAMETER ( MXV = $(MXV) )
      COMMON /PARAMS/ POINT(MXV), VAL
      COMMON /DIMEN/ NOD, NOF
      COMMON /IWORK/ INDE(MXV)
      COMMON /RWORK/ XT1(MXV), XT2(MXV), XT3(MXV), XT4(MXV), XT5(MXV),
     &               XT6(MXV), XT7(MXV+1)
      COMMON /RWORKT/ FT1(MXT), FT2(MXT), FT3(MXT)
C  ---------------------------------------------------------------------
C  Local variables:
      EXTERNAL FCN
C  ---------------------------------------------------------------------
C
C  Subroutine LMDIF operates on a set of transformed variables that
C  take care of margins and fix status.
      DO 1,I=1,NOD
         XT6(I) = POINT(I)
1     CONTINUE
      CALL XTOY(XT6,NOD)
C
      MODE = 1
      FACTOR = 100.
      LDFJAC = NOF
      ITER = 0
      CALL LMDIF(FCN,NOF,NOD,XT6,FT1,FTOL,XTOL,GTOL,MAXFEV,XT4,
     &           MODE,FACTOR,IPRINT,INFO,NFEV,W,LDFJAC,INDE,XT5,XT1,
     &           XT2,XT3,FT2,ITER)
C
      CALL YTOX(XT6,NOD)
      DO 2,I=1,NOD
         POINT(I) = XT6(I)
2     CONTINUE
      VAL = ACSQ(POINT,NOD)
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE FCN ( M, N, X, FVEC, IFLAG )
C  ---------------------------------------------------------------------
C
C  Description:
C    This is the objective function that subroutine LMDIF minimizes.
C
C  Input arguments:
C    M          Number of functions.
C    N          Number of variables.
C    X          The minimization variables.
C    IFLAG      Non-zero IFLAGs request printouts and are ignored.
C
C  Output arguments:
C    FVEC       Function values at X.
C
C  ---------------------------------------------------------------------
      IMPLICIT $(TYPE) (A-H,O-Z)
C  Arguments:
      DIMENSION X(N), FVEC(M)
C  ---------------------------------------------------------------------
C  Common blocks:
      PARAMETER ( MXV = $(MXV) )
      COMMON /RWORK/ XT1(MXV), XT2(MXV), XT3(MXV), XT4(MXV), XT5(MXV),
     &               XT6(MXV), XT7(MXV+1)
C  ---------------------------------------------------------------------
C
C  A zero IFLAG indicates a request for printout; ignore.
      IF (IFLAG.EQ.0) RETURN
C
C  Back-transform the variables.
      DO 10,I=1,N
10       XT7(I) = X(I)
      CALL YTOX(XT7,N)
C
C  Call the Merlin Sum-of-Squares function.
      CALL LSQFCN(M,N,XT7,FVEC)
C
      END
C  ---------------------------------------------------------------------
      FUNCTION SPMPAR ( I )
C  ---------------------------------------------------------------------
C
C  Description:
C    Provides machine constants to to the MINPACK routines.
C
C  Input arguments:
C    I          I = 1 -> Return the machine accuracy.
C               I = 2 -> Return the smallest positive number
C                        distinguishable from zero.
C               I = 3 -> Return the largest real the machine can handle.
C
C  ---------------------------------------------------------------------
      IMPLICIT $(TYPE) (A-H,O-Z)
C  Arguments:
C  ---------------------------------------------------------------------
C  Common blocks:
      COMMON /ACCURA/ COMACC
      COMMON /BIGCOM/ BIGNUM, BLIM, DWARF, DWALOG
C  ---------------------------------------------------------------------
C  Local variables:
      DIMENSION RMACH(3)
C  ---------------------------------------------------------------------
C
      RMACH(1) = COMACC
      RMACH(2) = DWARF
      RMACH(3) = BIGNUM
      SPMPAR = RMACH(I)
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE LEVETM ( INFO, MESS )
C  ---------------------------------------------------------------------
C
C  Description:
C    Constructs a message, describing the reason for termination of the
C    LEVE method.
C
C  Input arguments:
C    INFO       Reason for termination of the LEVE method, as returned
C               by subroutine LMDIF.
C
C  Output arguments:
C    MESS       Message, describing the reason for termination.
C
C  ---------------------------------------------------------------------
C  Arguments:
      CHARACTER*(*) MESS
C  ---------------------------------------------------------------------
C
      IF (INFO.EQ.0) THEN
C  This should never happen.
         MESS = 'Some input parameters were incorrect'
      ELSE IF (INFO.EQ.1) THEN
         MESS = 'The function value has converged'
      ELSE IF (INFO.EQ.2) THEN
         MESS = 'The X-convergence criterion is satisfied'
      ELSE IF (INFO.EQ.3) THEN
         MESS = 'F and X-convergence has been achieved'
      ELSE IF (INFO.EQ.4) THEN
         MESS = 'The gradient criterion is satisfied'
      ELSE IF (INFO.EQ.5) THEN
         MESS = 'All function evaluations have been used'
      ELSE IF (INFO.EQ.6) THEN
         MESS = 'Further progress is not possible'
      ELSE IF (INFO.EQ.7) THEN
         MESS = 'Further progress is not possible'
      ELSE IF (INFO.EQ.8) THEN
         MESS = 'Further progress is not possible'
      ELSE IF (INFO.EQ.9) THEN
         MESS = 'Target value has been reached'
      ELSE
         CALL MABORT('Incorrect "info".','LEVETM')
      END IF
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE LMDIF(FCN,M,N,X,FVEC,FTOL,XTOL,GTOL,MAXFEV,
     *                 DIAG,MODE,FACTOR,NPRINT,INFO,NFEV,FJAC,LDFJAC,
     *                 IPVT,QTF,WA1,WA2,WA3,WA4,ITER)
C
      IMPLICIT $(TYPE) (A-H,O-Z)
      PARAMETER ( MXV = $(MXV) )
      DIMENSION XTMP(MXV)
      INTEGER M,N,MAXFEV,MODE,NPRINT,INFO,NFEV,LDFJAC
      INTEGER IPVT(N)
      DIMENSION X(N),FVEC(M),DIAG(N),FJAC(LDFJAC,N),QTF(N),WA1(N),
     *     WA2(N),WA3(N),WA4(M)
      EXTERNAL FCN
C     **********
C
C     SUBROUTINE LMDIF
C
C     THE PURPOSE OF LMDIF IS TO MINIMIZE THE SUM OF THE SQUARES OF
C     M NONLINEAR FUNCTIONS IN N VARIABLES BY A MODIFICATION OF
C     THE LEVENBERG-MARQUARDT ALGORITHM. THE USER MUST PROVIDE A
C     SUBROUTINE WHICH CALCULATES THE FUNCTIONS. THE JACOBIAN IS
C     THEN CALCULATED BY A FORWARD-DIFFERENCE APPROXIMATION.
C
C     THE SUBROUTINE STATEMENT IS
C
C       SUBROUTINE LMDIF(FCN,M,N,X,FVEC,FTOL,XTOL,GTOL,MAXFEV,EPSFCN,
C                        DIAG,MODE,FACTOR,NPRINT,INFO,NFEV,FJAC,
C                        LDFJAC,IPVT,QTF,WA1,WA2,WA3,WA4)
C
C     WHERE
C
C       FCN IS THE NAME OF THE USER-SUPPLIED SUBROUTINE WHICH
C         CALCULATES THE FUNCTIONS. FCN MUST BE DECLARED
C         IN AN EXTERNAL STATEMENT IN THE USER CALLING
C         PROGRAM, AND SHOULD BE WRITTEN AS FOLLOWS.
C
C         SUBROUTINE FCN(M,N,X,FVEC,IFLAG)
C         INTEGER M,N,IFLAG
C         REAL X(N),FVEC(M)
C         ----------
C         CALCULATE THE FUNCTIONS AT X AND
C         RETURN THIS VECTOR IN FVEC.
C         ----------
C         RETURN
C         END
C
C         THE VALUE OF IFLAG SHOULD NOT BE CHANGED BY FCN UNLESS
C         THE USER WANTS TO TERMINATE EXECUTION OF LMDIF.
C         IN THIS CASE SET IFLAG TO A NEGATIVE INTEGER.
C
C       M IS A POSITIVE INTEGER INPUT VARIABLE SET TO THE NUMBER
C         OF FUNCTIONS.
C
C       N IS A POSITIVE INTEGER INPUT VARIABLE SET TO THE NUMBER
C         OF VARIABLES. N MUST NOT EXCEED M.
C
C       X IS AN ARRAY OF LENGTH N. ON INPUT X MUST CONTAIN
C         AN INITIAL ESTIMATE OF THE SOLUTION VECTOR. ON OUTPUT X
C         CONTAINS THE FINAL ESTIMATE OF THE SOLUTION VECTOR.
C
C       FVEC IS AN OUTPUT ARRAY OF LENGTH M WHICH CONTAINS
C         THE FUNCTIONS EVALUATED AT THE OUTPUT X.
C
C       FTOL IS A NONNEGATIVE INPUT VARIABLE. TERMINATION
C         OCCURS WHEN BOTH THE ACTUAL AND PREDICTED RELATIVE
C         REDUCTIONS IN THE SUM OF SQUARES ARE AT MOST FTOL.
C         THEREFORE, FTOL MEASURES THE RELATIVE ERROR DESIRED
C         IN THE SUM OF SQUARES.
C
C       XTOL IS A NONNEGATIVE INPUT VARIABLE. TERMINATION
C         OCCURS WHEN THE RELATIVE ERROR BETWEEN TWO CONSECUTIVE
C         ITERATES IS AT MOST XTOL. THEREFORE, XTOL MEASURES THE
C         RELATIVE ERROR DESIRED IN THE APPROXIMATE SOLUTION.
C
C       GTOL IS A NONNEGATIVE INPUT VARIABLE. TERMINATION
C         OCCURS WHEN THE COSINE OF THE ANGLE BETWEEN FVEC AND
C         ANY COLUMN OF THE JACOBIAN IS AT MOST GTOL IN ABSOLUTE
C         VALUE. THEREFORE, GTOL MEASURES THE ORTHOGONALITY
C         DESIRED BETWEEN THE FUNCTION VECTOR AND THE COLUMNS
C         OF THE JACOBIAN.
C
C       MAXFEV IS A POSITIVE INTEGER INPUT VARIABLE. TERMINATION
C         OCCURS WHEN THE NUMBER OF CALLS TO FCN IS AT LEAST
C         MAXFEV BY THE END OF AN ITERATION.
C
C       DIAG IS AN ARRAY OF LENGTH N. IF MODE = 1 (SEE
C         BELOW), DIAG IS INTERNALLY SET. IF MODE = 2, DIAG
C         MUST CONTAIN POSITIVE ENTRIES THAT SERVE AS
C         MULTIPLICATIVE SCALE FACTORS FOR THE VARIABLES.
C
C       MODE IS AN INTEGER INPUT VARIABLE. IF MODE = 1, THE
C         VARIABLES WILL BE SCALED INTERNALLY. IF MODE = 2,
C         THE SCALING IS SPECIFIED BY THE INPUT DIAG. OTHER
C         VALUES OF MODE ARE EQUIVALENT TO MODE = 1.
C
C       FACTOR IS A POSITIVE INPUT VARIABLE USED IN DETERMINING THE
C         INITIAL STEP BOUND. THIS BOUND IS SET TO THE PRODUCT OF
C         FACTOR AND THE EUCLIDEAN NORM OF DIAG*X IF NONZERO, OR ELSE
C         TO FACTOR ITSELF. IN MOST CASES FACTOR SHOULD LIE IN THE
C         INTERVAL (.1,100.). 100. IS A GENERALLY RECOMMENDED VALUE.
C
C       NPRINT IS AN INTEGER INPUT VARIABLE THAT ENABLES CONTROLLED
C         PRINTING OF ITERATES IF IT IS POSITIVE. IN THIS CASE,
C         FCN IS CALLED WITH IFLAG = 0 AT THE BEGINNING OF THE FIRST
C         ITERATION AND EVERY NPRINT ITERATIONS THEREAFTER AND
C         IMMEDIATELY PRIOR TO RETURN, WITH X AND FVEC AVAILABLE
C         FOR PRINTING. IF NPRINT IS NOT POSITIVE, NO SPECIAL CALLS
C         OF FCN WITH IFLAG = 0 ARE MADE.
C
C       INFO IS AN INTEGER OUTPUT VARIABLE. IF THE USER HAS
C         TERMINATED EXECUTION, INFO IS SET TO THE (NEGATIVE)
C         VALUE OF IFLAG. SEE DESCRIPTION OF FCN. OTHERWISE,
C         INFO IS SET AS FOLLOWS.
C
C         INFO = 0  IMPROPER INPUT PARAMETERS.
C
C         INFO = 1  BOTH ACTUAL AND PREDICTED RELATIVE REDUCTIONS
C                   IN THE SUM OF SQUARES ARE AT MOST FTOL.
C
C         INFO = 2  RELATIVE ERROR BETWEEN TWO CONSECUTIVE ITERATES
C                   IS AT MOST XTOL.
C
C         INFO = 3  CONDITIONS FOR INFO = 1 AND INFO = 2 BOTH HOLD.
C
C         INFO = 4  THE COSINE OF THE ANGLE BETWEEN FVEC AND ANY
C                   COLUMN OF THE JACOBIAN IS AT MOST GTOL IN
C                   ABSOLUTE VALUE.
C
C         INFO = 5  NUMBER OF CALLS TO FCN HAS REACHED OR
C                   EXCEEDED MAXFEV.
C
C         INFO = 6  FTOL IS TOO SMALL. NO FURTHER REDUCTION IN
C                   THE SUM OF SQUARES IS POSSIBLE.
C
C         INFO = 7  XTOL IS TOO SMALL. NO FURTHER IMPROVEMENT IN
C                   THE APPROXIMATE SOLUTION X IS POSSIBLE.
C
C         INFO = 8  GTOL IS TOO SMALL. FVEC IS ORTHOGONAL TO THE
C                   COLUMNS OF THE JACOBIAN TO MACHINE PRECISION.
C
C       NFEV IS AN INTEGER OUTPUT VARIABLE SET TO THE NUMBER OF
C         CALLS TO FCN.
C
C       FJAC IS AN OUTPUT M BY N ARRAY. THE UPPER N BY N SUBMATRIX
C         OF FJAC CONTAINS AN UPPER TRIANGULAR MATRIX R WITH
C         DIAGONAL ELEMENTS OF NONINCREASING MAGNITUDE SUCH THAT
C
C                T     T           T
C               P *(JAC *JAC)*P = R *R,
C
C         WHERE P IS A PERMUTATION MATRIX AND JAC IS THE FINAL
C         CALCULATED JACOBIAN. COLUMN J OF P IS COLUMN IPVT(J)
C         (SEE BELOW) OF THE IDENTITY MATRIX. THE LOWER TRAPEZOIDAL
C         PART OF FJAC CONTAINS INFORMATION GENERATED DURING
C         THE COMPUTATION OF R.
C
C       LDFJAC IS A POSITIVE INTEGER INPUT VARIABLE NOT LESS THAN M
C         WHICH SPECIFIES THE LEADING DIMENSION OF THE ARRAY FJAC.
C
C       IPVT IS AN INTEGER OUTPUT ARRAY OF LENGTH N. IPVT
C         DEFINES A PERMUTATION MATRIX P SUCH THAT JAC*P = Q*R,
C         WHERE JAC IS THE FINAL CALCULATED JACOBIAN, Q IS
C         ORTHOGONAL (NOT STORED), AND R IS UPPER TRIANGULAR
C         WITH DIAGONAL ELEMENTS OF NONINCREASING MAGNITUDE.
C         COLUMN J OF P IS COLUMN IPVT(J) OF THE IDENTITY MATRIX.
C
C       QTF IS AN OUTPUT ARRAY OF LENGTH N WHICH CONTAINS
C         THE FIRST N ELEMENTS OF THE VECTOR (Q TRANSPOSE)*FVEC.
C
C       WA1, WA2, AND WA3 ARE WORK ARRAYS OF LENGTH N.
C
C       WA4 IS A WORK ARRAY OF LENGTH M.
C
C     SUBPROGRAMS CALLED
C
C       USER-SUPPLIED ...... FCN
C
C       MINPACK-SUPPLIED ... SPMPAR,ENORM,FDJAC2,LMPAR,QRFAC
C
C       FORTRAN-SUPPLIED ... ABS,AMAX1,AMIN1,SQRT,MOD
C
C     ARGONNE NATIONAL LABORATORY. MINPACK PROJECT. MARCH 1980.
C     BURTON S. GARBOW, KENNETH E. HILLSTROM, JORGE J. MORE
C
C     **********
      COMMON /TOTCAL/ ITNOCS, IPARTC, ITDER, IPDER, ITHESS, IPHESS,
     &                ITJAC, IPJAC
      INTEGER I,IFLAG,ITER,J,L
      DATA ONE,P1,P5,P25,P75,P0001,ZERO
     *     /1.0E0,1.0E-1,5.0E-1,2.5E-1,7.5E-1,1.0E-4,0.0E0/
C
C     EPSMCH IS THE MACHINE PRECISION.
C
      EPSMCH = SPMPAR(1)
C
      INFO = 0
      IFLAG = 0
      NFEV = 0
C
C     CHECK THE INPUT PARAMETERS FOR ERRORS.
C
      IF (N .LE. 0 .OR. M .LT. N .OR. LDFJAC .LT. M
     *    .OR. FTOL .LT. ZERO .OR. XTOL .LT. ZERO .OR. GTOL .LT. ZERO
     *    .OR. MAXFEV .LE. 0 .OR. FACTOR .LE. ZERO) GO TO 300
      IF (MODE .NE. 2) GO TO 20
      DO 10 J = 1, N
         IF (DIAG(J) .LE. ZERO) GO TO 300
   10    CONTINUE
   20 CONTINUE
C
C     EVALUATE THE FUNCTION AT THE STARTING POINT
C     AND CALCULATE ITS NORM.
C
      IFLAG = 1
      CALL FCN(M,N,X,FVEC,IFLAG)
      NFEV = 1
      IF (IFLAG .LT. 0) GO TO 300
      FNORM = ENORM(M,FVEC)
C
C     INITIALIZE LEVENBERG-MARQUARDT PARAMETER AND ITERATION COUNTER.
C
      PAR = ZERO
      ITER = 1
C
C     BEGINNING OF THE OUTER LOOP.
C
   30 CONTINUE
         CALL MDIS(X,FNORM**2,ITER,NFEV,MAXFEV,NPRINT,1)
C
C  Did we reach our target value ?
         IF (ITCONV(FNORM**2).NE.0) THEN
            INFO = 9
            RETURN
         END IF
C
C        CALCULATE THE JACOBIAN MATRIX.
C
         IFLAG = 2
CC       CALL FDJAC2(FCN,M,N,X,FVEC,FJAC,LDFJAC,IFLAG,EPSFCN,WA4)
CC       NFEV = NFEV + N
         DO 400,I=1,N
400         XTMP(I) = X(I)
         CALL YTOX(XTMP,N)
         NC1 = ITNOCS
         CALL MJACOB(M,N,XTMP,FVEC,FJAC,LDFJAC)
         NC2 = ITNOCS
         NFEV = NFEV + (NC2-NC1)
         CALL JTRANS(FJAC,X,LDFJAC)
         IF (IFLAG .LT. 0) GO TO 300
C
C        IF REQUESTED, CALL FCN TO ENABLE PRINTING OF ITERATES.
C
         IF (NPRINT .LE. 0) GO TO 40
         IFLAG = 0
CC         IF (MOD(ITER-1,NPRINT) .EQ. 0) CALL FCN(M,N,X,FVEC,IFLAG)
C
         IF (IFLAG .LT. 0) GO TO 300
   40    CONTINUE
C
C        COMPUTE THE QR FACTORIZATION OF THE JACOBIAN.
C
         CALL QRFAC(M,N,FJAC,LDFJAC,.TRUE.,IPVT,N,WA1,WA2,WA3)
C
C        ON THE FIRST ITERATION AND IF MODE IS 1, SCALE ACCORDING
C        TO THE NORMS OF THE COLUMNS OF THE INITIAL JACOBIAN.
C
         IF (ITER .NE. 1) GO TO 80
         IF (MODE .EQ. 2) GO TO 60
         DO 50 J = 1, N
            DIAG(J) = WA2(J)
            IF (WA2(J) .EQ. ZERO) DIAG(J) = ONE
   50       CONTINUE
   60    CONTINUE
C
C        ON THE FIRST ITERATION, CALCULATE THE NORM OF THE SCALED X
C        AND INITIALIZE THE STEP BOUND DELTA.
C
         DO 70 J = 1, N
            WA3(J) = DIAG(J)*X(J)
   70       CONTINUE
         XNORM = ENORM(N,WA3)
         DELTA = FACTOR*XNORM
         IF (DELTA .EQ. ZERO) DELTA = FACTOR
   80    CONTINUE
C
C        FORM (Q TRANSPOSE)*FVEC AND STORE THE FIRST N COMPONENTS IN
C        QTF.
C
         DO 90 I = 1, M
            WA4(I) = FVEC(I)
   90       CONTINUE
         DO 130 J = 1, N
            IF (FJAC(J,J) .EQ. ZERO) GO TO 120
            SUM = ZERO
            DO 100 I = J, M
               SUM = SUM + FJAC(I,J)*WA4(I)
  100          CONTINUE
            TEMP = -SUM/FJAC(J,J)
            DO 110 I = J, M
               WA4(I) = WA4(I) + FJAC(I,J)*TEMP
  110          CONTINUE
  120       CONTINUE
            FJAC(J,J) = WA1(J)
            QTF(J) = WA4(J)
  130       CONTINUE
C
C        COMPUTE THE NORM OF THE SCALED GRADIENT.
C
         GNORM = ZERO
         IF (FNORM .EQ. ZERO) GO TO 170
         DO 160 J = 1, N
            L = IPVT(J)
            IF (WA2(L) .EQ. ZERO) GO TO 150
            SUM = ZERO
            DO 140 I = 1, J
               SUM = SUM + FJAC(I,J)*(QTF(I)/FNORM)
  140          CONTINUE
            GNORM = MAX(GNORM,ABS(SUM/WA2(L)))
  150       CONTINUE
  160       CONTINUE
  170    CONTINUE
C
C        TEST FOR CONVERGENCE OF THE GRADIENT NORM.
C
         IF (GNORM .LE. GTOL) INFO = 4
         IF (INFO .NE. 0) GO TO 300
C
C        RESCALE IF NECESSARY.
C
         IF (MODE .EQ. 2) GO TO 190
         DO 180 J = 1, N
            DIAG(J) = MAX(DIAG(J),WA2(J))
  180       CONTINUE
  190    CONTINUE
C
C        BEGINNING OF THE INNER LOOP.
C
  200    CONTINUE
C
C           DETERMINE THE LEVENBERG-MARQUARDT PARAMETER.
C
            CALL LMPAR(N,FJAC,LDFJAC,IPVT,DIAG,QTF,DELTA,PAR,WA1,WA2,
     *                 WA3,WA4)
C
C           STORE THE DIRECTION P AND X + P. CALCULATE THE NORM OF P.
C
            DO 210 J = 1, N
               WA1(J) = -WA1(J)
               WA2(J) = X(J) + WA1(J)
               WA3(J) = DIAG(J)*WA1(J)
  210          CONTINUE
            PNORM = ENORM(N,WA3)
C
C           ON THE FIRST ITERATION, ADJUST THE INITIAL STEP BOUND.
C
            IF (ITER .EQ. 1) DELTA = MIN(DELTA,PNORM)
C
C           EVALUATE THE FUNCTION AT X + P AND CALCULATE ITS NORM.
C
            IFLAG = 1
            CALL FCN(M,N,WA2,WA4,IFLAG)
            NFEV = NFEV + 1
            IF (IFLAG .LT. 0) GO TO 300
            FNORM1 = ENORM(M,WA4)
C
C           COMPUTE THE SCALED ACTUAL REDUCTION.
C
            ACTRED = -ONE
            IF (P1*FNORM1 .LT. FNORM) ACTRED = ONE - (FNORM1/FNORM)**2
C
C           COMPUTE THE SCALED PREDICTED REDUCTION AND
C           THE SCALED DIRECTIONAL DERIVATIVE.
C
            DO 230 J = 1, N
               WA3(J) = ZERO
               L = IPVT(J)
               TEMP = WA1(L)
               DO 220 I = 1, J
                  WA3(I) = WA3(I) + FJAC(I,J)*TEMP
  220             CONTINUE
  230          CONTINUE
            TEMP1 = ENORM(N,WA3)/FNORM
            TEMP2 = (SQRT(PAR)*PNORM)/FNORM
            PRERED = TEMP1**2 + TEMP2**2/P5
            DIRDER = -(TEMP1**2 + TEMP2**2)
C
C           COMPUTE THE RATIO OF THE ACTUAL TO THE PREDICTED
C           REDUCTION.
C
            RATIO = ZERO
            IF (PRERED .NE. ZERO) RATIO = ACTRED/PRERED
C
C           UPDATE THE STEP BOUND.
C
            IF (RATIO .GT. P25) GO TO 240
               IF (ACTRED .GE. ZERO) TEMP = P5
               IF (ACTRED .LT. ZERO)
     *            TEMP = P5*DIRDER/(DIRDER + P5*ACTRED)
               IF (P1*FNORM1 .GE. FNORM .OR. TEMP .LT. P1) TEMP = P1
               DELTA = TEMP*MIN(DELTA,PNORM/P1)
               PAR = PAR/TEMP
               GO TO 260
  240       CONTINUE
               IF (PAR .NE. ZERO .AND. RATIO .LT. P75) GO TO 250
               DELTA = PNORM/P5
               PAR = P5*PAR
  250          CONTINUE
  260       CONTINUE
C
C           TEST FOR SUCCESSFUL ITERATION.
C
            IF (RATIO .LT. P0001) GO TO 290
C
C           SUCCESSFUL ITERATION. UPDATE X, FVEC, AND THEIR NORMS.
C
            DO 270 J = 1, N
               X(J) = WA2(J)
               WA2(J) = DIAG(J)*X(J)
  270          CONTINUE
            DO 280 I = 1, M
               FVEC(I) = WA4(I)
  280          CONTINUE
            XNORM = ENORM(N,WA2)
            FNORM = FNORM1
            ITER = ITER + 1
  290       CONTINUE
C
C           TESTS FOR CONVERGENCE.
C
            IF (ABS(ACTRED) .LE. FTOL .AND. PRERED .LE. FTOL
     *          .AND. P5*RATIO .LE. ONE) INFO = 1
            IF (DELTA .LE. XTOL*XNORM) INFO = 2
            IF (ABS(ACTRED) .LE. FTOL .AND. PRERED .LE. FTOL
     *          .AND. P5*RATIO .LE. ONE .AND. INFO .EQ. 2) INFO = 3
            IF (INFO .NE. 0) GO TO 300
C
C           TESTS FOR TERMINATION AND STRINGENT TOLERANCES.
C
            IF (NFEV .GE. MAXFEV) INFO = 5
            IF (ABS(ACTRED) .LE. EPSMCH .AND. PRERED .LE. EPSMCH
     *          .AND. P5*RATIO .LE. ONE) INFO = 6
            IF (DELTA .LE. EPSMCH*XNORM) INFO = 7
            IF (GNORM .LE. EPSMCH) INFO = 8
            IF (INFO .NE. 0) GO TO 300
C
C           END OF THE INNER LOOP. REPEAT IF ITERATION UNSUCCESSFUL.
C
            IF (RATIO .LT. P0001) GO TO 200
C
C        END OF THE OUTER LOOP.
C
         GO TO 30
  300 CONTINUE
C
C     TERMINATION, EITHER NORMAL OR USER IMPOSED.
C
      IF (IFLAG .LT. 0) INFO = IFLAG
      IFLAG = 0
CC      IF (NPRINT .GT. 0) CALL FCN(M,N,X,FVEC,IFLAG)
      RETURN
C
C     LAST CARD OF SUBROUTINE LMDIF.
C
      END
      FUNCTION ENORM(N,X)
C
      IMPLICIT $(TYPE) (A-H,O-Z)
      INTEGER N
      DIMENSION X(N)
C     **********
C
C     FUNCTION ENORM
C
C     GIVEN AN N-VECTOR X, THIS FUNCTION CALCULATES THE
C     EUCLIDEAN NORM OF X.
C
C     THE EUCLIDEAN NORM IS COMPUTED BY ACCUMULATING THE SUM OF
C     SQUARES IN THREE DIFFERENT SUMS. THE SUMS OF SQUARES FOR THE
C     SMALL AND LARGE COMPONENTS ARE SCALED SO THAT NO OVERFLOWS
C     OCCUR. NON-DESTRUCTIVE UNDERFLOWS ARE PERMITTED. UNDERFLOWS
C     AND OVERFLOWS DO NOT OCCUR IN THE COMPUTATION OF THE UNSCALED
C     SUM OF SQUARES FOR THE INTERMEDIATE COMPONENTS.
C     THE DEFINITIONS OF SMALL, INTERMEDIATE AND LARGE COMPONENTS
C     DEPEND ON TWO CONSTANTS, RDWARF AND RGIANT. THE MAIN
C     RESTRICTIONS ON THESE CONSTANTS ARE THAT RDWARF**2 NOT
C     UNDERFLOW AND RGIANT**2 NOT OVERFLOW. THE CONSTANTS
C     GIVEN HERE ARE SUITABLE FOR EVERY KNOWN COMPUTER.
C
C     THE FUNCTION STATEMENT IS
C
C       REAL FUNCTION ENORM(N,X)
C
C     WHERE
C
C       N IS A POSITIVE INTEGER INPUT VARIABLE.
C
C       X IS AN INPUT ARRAY OF LENGTH N.
C
C     SUBPROGRAMS CALLED
C
C       FORTRAN-SUPPLIED ... ABS,SQRT
C
C     ARGONNE NATIONAL LABORATORY. MINPACK PROJECT. MARCH 1980.
C     BURTON S. GARBOW, KENNETH E. HILLSTROM, JORGE J. MORE
C
C     **********
      INTEGER I
      DATA ONE,ZERO,RDWARF,RGIANT /1.0E0,0.0E0,3.834E-20,1.304E19/
      S1 = ZERO
      S2 = ZERO
      S3 = ZERO
      X1MAX = ZERO
      X3MAX = ZERO
      FLOATN = N
      AGIANT = RGIANT/FLOATN
      DO 90 I = 1, N
         XABS = ABS(X(I))
         IF (XABS .GT. RDWARF .AND. XABS .LT. AGIANT) GO TO 70
            IF (XABS .LE. RDWARF) GO TO 30
C
C              SUM FOR LARGE COMPONENTS.
C
               IF (XABS .LE. X1MAX) GO TO 10
                  S1 = ONE + S1*(X1MAX/XABS)**2
                  X1MAX = XABS
                  GO TO 20
   10          CONTINUE
                  S1 = S1 + (XABS/X1MAX)**2
   20          CONTINUE
               GO TO 60
   30       CONTINUE
C
C              SUM FOR SMALL COMPONENTS.
C
               IF (XABS .LE. X3MAX) GO TO 40
                  S3 = ONE + S3*(X3MAX/XABS)**2
                  X3MAX = XABS
                  GO TO 50
   40          CONTINUE
                  IF (XABS .NE. ZERO) S3 = S3 + (XABS/X3MAX)**2
   50          CONTINUE
   60       CONTINUE
            GO TO 80
   70    CONTINUE
C
C           SUM FOR INTERMEDIATE COMPONENTS.
C
            S2 = S2 + XABS**2
   80    CONTINUE
   90    CONTINUE
C
C     CALCULATION OF NORM.
C
      IF (S1 .EQ. ZERO) GO TO 100
         ENORM = X1MAX*SQRT(S1+(S2/X1MAX)/X1MAX)
         GO TO 130
  100 CONTINUE
         IF (S2 .EQ. ZERO) GO TO 110
            IF (S2 .GE. X3MAX)
     *         ENORM = SQRT(S2*(ONE+(X3MAX/S2)*(X3MAX*S3)))
            IF (S2 .LT. X3MAX)
     *         ENORM = SQRT(X3MAX*((S2/X3MAX)+(X3MAX*S3)))
            GO TO 120
  110    CONTINUE
            ENORM = X3MAX*SQRT(S3)
  120    CONTINUE
  130 CONTINUE
      RETURN
C
C     LAST CARD OF FUNCTION ENORM.
C
      END
      SUBROUTINE FDJAC2(M,N,X,FVEC,FJAC,LDFJAC,IFLAG,EPSFCN,WA)
C
      IMPLICIT $(TYPE) (A-H,O-Z)
      INTEGER M,N,LDFJAC,IFLAG
      DIMENSION X(N),FVEC(M),FJAC(LDFJAC,N),WA(M)
C     **********
C
C     SUBROUTINE FDJAC2
C
C     THIS SUBROUTINE COMPUTES A FORWARD-DIFFERENCE APPROXIMATION
C     TO THE M BY N JACOBIAN MATRIX ASSOCIATED WITH A SPECIFIED
C     PROBLEM OF M FUNCTIONS IN N VARIABLES.
C
C     THE SUBROUTINE STATEMENT IS
C
C       SUBROUTINE FDJAC2(M,N,X,FVEC,FJAC,LDFJAC,IFLAG,EPSFCN,WA)
C
C     WHERE
C
C       FCN IS THE NAME OF THE USER-SUPPLIED SUBROUTINE WHICH
C         CALCULATES THE FUNCTIONS. FCN MUST BE DECLARED
C         IN AN EXTERNAL STATEMENT IN THE USER CALLING
C         PROGRAM, AND SHOULD BE WRITTEN AS FOLLOWS.
C
C         SUBROUTINE FCN(M,N,X,FVEC,IFLAG)
C         INTEGER M,N,IFLAG
C         REAL X(N),FVEC(M)
C         ----------
C         CALCULATE THE FUNCTIONS AT X AND
C         RETURN THIS VECTOR IN FVEC.
C         ----------
C         RETURN
C         END
C
C         THE VALUE OF IFLAG SHOULD NOT BE CHANGED BY FCN UNLESS
C         THE USER WANTS TO TERMINATE EXECUTION OF FDJAC2.
C         IN THIS CASE SET IFLAG TO A NEGATIVE INTEGER.
C
C       M IS A POSITIVE INTEGER INPUT VARIABLE SET TO THE NUMBER
C         OF FUNCTIONS.
C
C       N IS A POSITIVE INTEGER INPUT VARIABLE SET TO THE NUMBER
C         OF VARIABLES. N MUST NOT EXCEED M.
C
C       X IS AN INPUT ARRAY OF LENGTH N.
C
C       FVEC IS AN INPUT ARRAY OF LENGTH M WHICH MUST CONTAIN THE
C         FUNCTIONS EVALUATED AT X.
C
C       FJAC IS AN OUTPUT M BY N ARRAY WHICH CONTAINS THE
C         APPROXIMATION TO THE JACOBIAN MATRIX EVALUATED AT X.
C
C       LDFJAC IS A POSITIVE INTEGER INPUT VARIABLE NOT LESS THAN M
C         WHICH SPECIFIES THE LEADING DIMENSION OF THE ARRAY FJAC.
C
C       IFLAG IS AN INTEGER VARIABLE WHICH CAN BE USED TO TERMINATE
C         THE EXECUTION OF FDJAC2. SEE DESCRIPTION OF FCN.
C
C       EPSFCN IS AN INPUT VARIABLE USED IN DETERMINING A SUITABLE
C         STEP LENGTH FOR THE FORWARD-DIFFERENCE APPROXIMATION. THIS
C         APPROXIMATION ASSUMES THAT THE RELATIVE ERRORS IN THE
C         FUNCTIONS ARE OF THE ORDER OF EPSFCN. IF EPSFCN IS LESS
C         THAN THE MACHINE PRECISION, IT IS ASSUMED THAT THE RELATIVE
C         ERRORS IN THE FUNCTIONS ARE OF THE ORDER OF THE MACHINE
C         PRECISION.
C
C       WA IS A WORK ARRAY OF LENGTH M.
C
C     SUBPROGRAMS CALLED
C
C       USER-SUPPLIED ...... FCN
C
C       MINPACK-SUPPLIED ... SPMPAR
C
C       FORTRAN-SUPPLIED ... ABS,AMAX1,SQRT
C
C     ARGONNE NATIONAL LABORATORY. MINPACK PROJECT. MARCH 1980.
C     BURTON S. GARBOW, KENNETH E. HILLSTROM, JORGE J. MORE
C
C     **********
      INTEGER I,J
      DATA ZERO /0.0E0/
C
C     EPSMCH IS THE MACHINE PRECISION.
C
      EPSMCH = SPMPAR(1)
C
      EPS = SQRT(MAX(EPSFCN,EPSMCH))
      DO 20 J = 1, N
         TEMP = X(J)
         H = EPS*ABS(TEMP)
         IF (H .EQ. ZERO) H = EPS
         X(J) = TEMP + H
         CALL LSQFCN(M,N,X,WA)
         IF (IFLAG .LT. 0) GO TO 30
         X(J) = TEMP
         DO 10 I = 1, M
            FJAC(I,J) = (WA(I) - FVEC(I))/H
   10       CONTINUE
   20    CONTINUE
   30 CONTINUE
      RETURN
C
C     LAST CARD OF SUBROUTINE FDJAC2.
C
      END
      SUBROUTINE LMPAR(N,R,LDR,IPVT,DIAG,QTB,DELTA,PAR,X,SDIAG,WA1,
     *                 WA2)
C
      IMPLICIT $(TYPE) (A-H,O-Z)
      INTEGER N,LDR
      INTEGER IPVT(N)
      DIMENSION R(LDR,N),DIAG(N),QTB(N),X(N),SDIAG(N),WA1(N),WA2(N)
C     **********
C
C     SUBROUTINE LMPAR
C
C     GIVEN AN M BY N MATRIX A, AN N BY N NONSINGULAR DIAGONAL
C     MATRIX D, AN M-VECTOR B, AND A POSITIVE NUMBER DELTA,
C     THE PROBLEM IS TO DETERMINE A VALUE FOR THE PARAMETER
C     PAR SUCH THAT IF X SOLVES THE SYSTEM
C
C           A*X = B ,     SQRT(PAR)*D*X = 0 ,
C
C     IN THE LEAST SQUARES SENSE, AND DXNORM IS THE EUCLIDEAN
C     NORM OF D*X, THEN EITHER PAR IS ZERO AND
C
C           (DXNORM-DELTA) .LE. 0.1*DELTA ,
C
C     OR PAR IS POSITIVE AND
C
C           ABS(DXNORM-DELTA) .LE. 0.1*DELTA .
C
C     THIS SUBROUTINE COMPLETES THE SOLUTION OF THE PROBLEM
C     IF IT IS PROVIDED WITH THE NECESSARY INFORMATION FROM THE
C     QR FACTORIZATION, WITH COLUMN PIVOTING, OF A. THAT IS, IF
C     A*P = Q*R, WHERE P IS A PERMUTATION MATRIX, Q HAS ORTHOGONAL
C     COLUMNS, AND R IS AN UPPER TRIANGULAR MATRIX WITH DIAGONAL
C     ELEMENTS OF NONINCREASING MAGNITUDE, THEN LMPAR EXPECTS
C     THE FULL UPPER TRIANGLE OF R, THE PERMUTATION MATRIX P,
C     AND THE FIRST N COMPONENTS OF (Q TRANSPOSE)*B. ON OUTPUT
C     LMPAR ALSO PROVIDES AN UPPER TRIANGULAR MATRIX S SUCH THAT
C
C            T   T                   T
C           P *(A *A + PAR*D*D)*P = S *S .
C
C     S IS EMPLOYED WITHIN LMPAR AND MAY BE OF SEPARATE INTEREST.
C
C     ONLY A FEW ITERATIONS ARE GENERALLY NEEDED FOR CONVERGENCE
C     OF THE ALGORITHM. IF, HOWEVER, THE LIMIT OF 10 ITERATIONS
C     IS REACHED, THEN THE OUTPUT PAR WILL CONTAIN THE BEST
C     VALUE OBTAINED SO FAR.
C
C     THE SUBROUTINE STATEMENT IS
C
C       SUBROUTINE LMPAR(N,R,LDR,IPVT,DIAG,QTB,DELTA,PAR,X,SDIAG,
C                        WA1,WA2)
C
C     WHERE
C
C       N IS A POSITIVE INTEGER INPUT VARIABLE SET TO THE ORDER OF R.
C
C       R IS AN N BY N ARRAY. ON INPUT THE FULL UPPER TRIANGLE
C         MUST CONTAIN THE FULL UPPER TRIANGLE OF THE MATRIX R.
C         ON OUTPUT THE FULL UPPER TRIANGLE IS UNALTERED, AND THE
C         STRICT LOWER TRIANGLE CONTAINS THE STRICT UPPER TRIANGLE
C         (TRANSPOSED) OF THE UPPER TRIANGULAR MATRIX S.
C
C       LDR IS A POSITIVE INTEGER INPUT VARIABLE NOT LESS THAN N
C         WHICH SPECIFIES THE LEADING DIMENSION OF THE ARRAY R.
C
C       IPVT IS AN INTEGER INPUT ARRAY OF LENGTH N WHICH DEFINES THE
C         PERMUTATION MATRIX P SUCH THAT A*P = Q*R. COLUMN J OF P
C         IS COLUMN IPVT(J) OF THE IDENTITY MATRIX.
C
C       DIAG IS AN INPUT ARRAY OF LENGTH N WHICH MUST CONTAIN THE
C         DIAGONAL ELEMENTS OF THE MATRIX D.
C
C       QTB IS AN INPUT ARRAY OF LENGTH N WHICH MUST CONTAIN THE FIRST
C         N ELEMENTS OF THE VECTOR (Q TRANSPOSE)*B.
C
C       DELTA IS A POSITIVE INPUT VARIABLE WHICH SPECIFIES AN UPPER
C         BOUND ON THE EUCLIDEAN NORM OF D*X.
C
C       PAR IS A NONNEGATIVE VARIABLE. ON INPUT PAR CONTAINS AN
C         INITIAL ESTIMATE OF THE LEVENBERG-MARQUARDT PARAMETER.
C         ON OUTPUT PAR CONTAINS THE FINAL ESTIMATE.
C
C       X IS AN OUTPUT ARRAY OF LENGTH N WHICH CONTAINS THE LEAST
C         SQUARES SOLUTION OF THE SYSTEM A*X = B, SQRT(PAR)*D*X = 0,
C         FOR THE OUTPUT PAR.
C
C       SDIAG IS AN OUTPUT ARRAY OF LENGTH N WHICH CONTAINS THE
C         DIAGONAL ELEMENTS OF THE UPPER TRIANGULAR MATRIX S.
C
C       WA1 AND WA2 ARE WORK ARRAYS OF LENGTH N.
C
C     SUBPROGRAMS CALLED
C
C       MINPACK-SUPPLIED ... SPMPAR,ENORM,QRSOLV
C
C       FORTRAN-SUPPLIED ... ABS,AMAX1,AMIN1,SQRT
C
C     ARGONNE NATIONAL LABORATORY. MINPACK PROJECT. MARCH 1980.
C     BURTON S. GARBOW, KENNETH E. HILLSTROM, JORGE J. MORE
C
C     **********
      INTEGER I,ITER,J,JM1,JP1,K,L,NSING
      DATA P1,P001,ZERO /1.0E-1,1.0E-3,0.0E0/
C
C     DWARF IS THE SMALLEST POSITIVE MAGNITUDE.
C
      DWARF = SPMPAR(2)
C
C     COMPUTE AND STORE IN X THE GAUSS-NEWTON DIRECTION. IF THE
C     JACOBIAN IS RANK-DEFICIENT, OBTAIN A LEAST SQUARES SOLUTION.
C
      NSING = N
      DO 10 J = 1, N
         WA1(J) = QTB(J)
         IF (R(J,J) .EQ. ZERO .AND. NSING .EQ. N) NSING = J - 1
         IF (NSING .LT. N) WA1(J) = ZERO
   10    CONTINUE
      IF (NSING .LT. 1) GO TO 50
      DO 40 K = 1, NSING
         J = NSING - K + 1
         WA1(J) = WA1(J)/R(J,J)
         TEMP = WA1(J)
         JM1 = J - 1
         IF (JM1 .LT. 1) GO TO 30
         DO 20 I = 1, JM1
            WA1(I) = WA1(I) - R(I,J)*TEMP
   20       CONTINUE
   30    CONTINUE
   40    CONTINUE
   50 CONTINUE
      DO 60 J = 1, N
         L = IPVT(J)
         X(L) = WA1(J)
   60    CONTINUE
C
C     INITIALIZE THE ITERATION COUNTER.
C     EVALUATE THE FUNCTION AT THE ORIGIN, AND TEST
C     FOR ACCEPTANCE OF THE GAUSS-NEWTON DIRECTION.
C
      ITER = 0
      DO 70 J = 1, N
         WA2(J) = DIAG(J)*X(J)
   70    CONTINUE
      DXNORM = ENORM(N,WA2)
      FP = DXNORM - DELTA
      IF (FP .LE. P1*DELTA) GO TO 220
C
C     IF THE JACOBIAN IS NOT RANK DEFICIENT, THE NEWTON
C     STEP PROVIDES A LOWER BOUND, PARL, FOR THE ZERO OF
C     THE FUNCTION. OTHERWISE SET THIS BOUND TO ZERO.
C
      PARL = ZERO
      IF (NSING .LT. N) GO TO 120
      DO 80 J = 1, N
         L = IPVT(J)
         WA1(J) = DIAG(L)*(WA2(L)/DXNORM)
   80    CONTINUE
      DO 110 J = 1, N
         SUM = ZERO
         JM1 = J - 1
         IF (JM1 .LT. 1) GO TO 100
         DO 90 I = 1, JM1
            SUM = SUM + R(I,J)*WA1(I)
   90       CONTINUE
  100    CONTINUE
         WA1(J) = (WA1(J) - SUM)/R(J,J)
  110    CONTINUE
      TEMP = ENORM(N,WA1)
      PARL = ((FP/DELTA)/TEMP)/TEMP
  120 CONTINUE
C
C     CALCULATE AN UPPER BOUND, PARU, FOR THE ZERO OF THE FUNCTION.
C
      DO 140 J = 1, N
         SUM = ZERO
         DO 130 I = 1, J
            SUM = SUM + R(I,J)*QTB(I)
  130       CONTINUE
         L = IPVT(J)
         WA1(J) = SUM/DIAG(L)
  140    CONTINUE
      GNORM = ENORM(N,WA1)
      PARU = GNORM/DELTA
      IF (PARU .EQ. ZERO) PARU = DWARF/MIN(DELTA,P1)
C
C     IF THE INPUT PAR LIES OUTSIDE OF THE INTERVAL (PARL,PARU),
C     SET PAR TO THE CLOSER ENDPOINT.
C
      PAR = MAX(PAR,PARL)
      PAR = MIN(PAR,PARU)
      IF (PAR .EQ. ZERO) PAR = GNORM/DXNORM
C
C     BEGINNING OF AN ITERATION.
C
  150 CONTINUE
         ITER = ITER + 1
C
C        EVALUATE THE FUNCTION AT THE CURRENT VALUE OF PAR.
C
         IF (PAR .EQ. ZERO) PAR = MAX(DWARF,P001*PARU)
         TEMP = SQRT(PAR)
         DO 160 J = 1, N
            WA1(J) = TEMP*DIAG(J)
  160       CONTINUE
         CALL QRSOLV(N,R,LDR,IPVT,WA1,QTB,X,SDIAG,WA2)
         DO 170 J = 1, N
            WA2(J) = DIAG(J)*X(J)
  170       CONTINUE
         DXNORM = ENORM(N,WA2)
         TEMP = FP
         FP = DXNORM - DELTA
C
C        IF THE FUNCTION IS SMALL ENOUGH, ACCEPT THE CURRENT VALUE
C        OF PAR. ALSO TEST FOR THE EXCEPTIONAL CASES WHERE PARL
C        IS ZERO OR THE NUMBER OF ITERATIONS HAS REACHED 10.
C
         IF (ABS(FP) .LE. P1*DELTA
     *       .OR. PARL .EQ. ZERO .AND. FP .LE. TEMP
     *            .AND. TEMP .LT. ZERO .OR. ITER .EQ. 10) GO TO 220
C
C        COMPUTE THE NEWTON CORRECTION.
C
         DO 180 J = 1, N
            L = IPVT(J)
            WA1(J) = DIAG(L)*(WA2(L)/DXNORM)
  180       CONTINUE
         DO 210 J = 1, N
            WA1(J) = WA1(J)/SDIAG(J)
            TEMP = WA1(J)
            JP1 = J + 1
            IF (N .LT. JP1) GO TO 200
            DO 190 I = JP1, N
               WA1(I) = WA1(I) - R(I,J)*TEMP
  190          CONTINUE
  200       CONTINUE
  210       CONTINUE
         TEMP = ENORM(N,WA1)
         PARC = ((FP/DELTA)/TEMP)/TEMP
C
C        DEPENDING ON THE SIGN OF THE FUNCTION, UPDATE PARL OR PARU.
C
         IF (FP .GT. ZERO) PARL = MAX(PARL,PAR)
         IF (FP .LT. ZERO) PARU = MIN(PARU,PAR)
C
C        COMPUTE AN IMPROVED ESTIMATE FOR PAR.
C
         PAR = MAX(PARL,PAR+PARC)
C
C        END OF AN ITERATION.
C
         GO TO 150
  220 CONTINUE
C
C     TERMINATION.
C
      IF (ITER .EQ. 0) PAR = ZERO
      RETURN
C
C     LAST CARD OF SUBROUTINE LMPAR.
C
      END
      SUBROUTINE QRFAC(M,N,A,LDA,PIVOT,IPVT,LIPVT,RDIAG,ACNORM,WA)
C
      IMPLICIT $(TYPE) (A-H,O-Z)
      INTEGER M,N,LDA,LIPVT
      INTEGER IPVT(LIPVT)
      LOGICAL PIVOT
      DIMENSION A(LDA,N),RDIAG(N),ACNORM(N),WA(N)
C     **********
C
C     SUBROUTINE QRFAC
C
C     THIS SUBROUTINE USES HOUSEHOLDER TRANSFORMATIONS WITH COLUMN
C     PIVOTING (OPTIONAL) TO COMPUTE A QR FACTORIZATION OF THE
C     M BY N MATRIX A. THAT IS, QRFAC DETERMINES AN ORTHOGONAL
C     MATRIX Q, A PERMUTATION MATRIX P, AND AN UPPER TRAPEZOIDAL
C     MATRIX R WITH DIAGONAL ELEMENTS OF NONINCREASING MAGNITUDE,
C     SUCH THAT A*P = Q*R. THE HOUSEHOLDER TRANSFORMATION FOR
C     COLUMN K, K = 1,2,...,MIN(M,N), IS OF THE FORM
C
C                           T
C           I - (1/U(K))*U*U
C
C     WHERE U HAS ZEROS IN THE FIRST K-1 POSITIONS. THE FORM OF
C     THIS TRANSFORMATION AND THE METHOD OF PIVOTING FIRST
C     APPEARED IN THE CORRESPONDING LINPACK SUBROUTINE.
C
C     THE SUBROUTINE STATEMENT IS
C
C       SUBROUTINE QRFAC(M,N,A,LDA,PIVOT,IPVT,LIPVT,RDIAG,ACNORM,WA)
C
C     WHERE
C
C       M IS A POSITIVE INTEGER INPUT VARIABLE SET TO THE NUMBER
C         OF ROWS OF A.
C
C       N IS A POSITIVE INTEGER INPUT VARIABLE SET TO THE NUMBER
C         OF COLUMNS OF A.
C
C       A IS AN M BY N ARRAY. ON INPUT A CONTAINS THE MATRIX FOR
C         WHICH THE QR FACTORIZATION IS TO BE COMPUTED. ON OUTPUT
C         THE STRICT UPPER TRAPEZOIDAL PART OF A CONTAINS THE STRICT
C         UPPER TRAPEZOIDAL PART OF R, AND THE LOWER TRAPEZOIDAL
C         PART OF A CONTAINS A FACTORED FORM OF Q (THE NON-TRIVIAL
C         ELEMENTS OF THE U VECTORS DESCRIBED ABOVE).
C
C       LDA IS A POSITIVE INTEGER INPUT VARIABLE NOT LESS THAN M
C         WHICH SPECIFIES THE LEADING DIMENSION OF THE ARRAY A.
C
C       PIVOT IS A LOGICAL INPUT VARIABLE. IF PIVOT IS SET TRUE,
C         THEN COLUMN PIVOTING IS ENFORCED. IF PIVOT IS SET FALSE,
C         THEN NO COLUMN PIVOTING IS DONE.
C
C       IPVT IS AN INTEGER OUTPUT ARRAY OF LENGTH LIPVT. IPVT
C         DEFINES THE PERMUTATION MATRIX P SUCH THAT A*P = Q*R.
C         COLUMN J OF P IS COLUMN IPVT(J) OF THE IDENTITY MATRIX.
C         IF PIVOT IS FALSE, IPVT IS NOT REFERENCED.
C
C       LIPVT IS A POSITIVE INTEGER INPUT VARIABLE. IF PIVOT IS FALSE,
C         THEN LIPVT MAY BE AS SMALL AS 1. IF PIVOT IS TRUE, THEN
C         LIPVT MUST BE AT LEAST N.
C
C       RDIAG IS AN OUTPUT ARRAY OF LENGTH N WHICH CONTAINS THE
C         DIAGONAL ELEMENTS OF R.
C
C       ACNORM IS AN OUTPUT ARRAY OF LENGTH N WHICH CONTAINS THE
C         NORMS OF THE CORRESPONDING COLUMNS OF THE INPUT MATRIX A.
C         IF THIS INFORMATION IS NOT NEEDED, THEN ACNORM CAN COINCIDE
C         WITH RDIAG.
C
C       WA IS A WORK ARRAY OF LENGTH N. IF PIVOT IS FALSE, THEN WA
C         CAN COINCIDE WITH RDIAG.
C
C     SUBPROGRAMS CALLED
C
C       MINPACK-SUPPLIED ... SPMPAR,ENORM
C
C       FORTRAN-SUPPLIED ... AMAX1,SQRT,MIN0
C
C     ARGONNE NATIONAL LABORATORY. MINPACK PROJECT. MARCH 1980.
C     BURTON S. GARBOW, KENNETH E. HILLSTROM, JORGE J. MORE
C
C     **********
      INTEGER I,J,JP1,K,KMAX,MINMN
      DATA ONE,P05,ZERO /1.0E0,5.0E-2,0.0E0/
C
C     EPSMCH IS THE MACHINE PRECISION.
C
      EPSMCH = SPMPAR(1)
C
C     COMPUTE THE INITIAL COLUMN NORMS AND INITIALIZE SEVERAL ARRAYS.
C
      DO 10 J = 1, N
         ACNORM(J) = ENORM(M,A(1,J))
         RDIAG(J) = ACNORM(J)
         WA(J) = RDIAG(J)
         IF (PIVOT) IPVT(J) = J
   10    CONTINUE
C
C     REDUCE A TO R WITH HOUSEHOLDER TRANSFORMATIONS.
C
      MINMN = MIN0(M,N)
      DO 110 J = 1, MINMN
         IF (.NOT.PIVOT) GO TO 40
C
C        BRING THE COLUMN OF LARGEST NORM INTO THE PIVOT POSITION.
C
         KMAX = J
         DO 20 K = J, N
            IF (RDIAG(K) .GT. RDIAG(KMAX)) KMAX = K
   20       CONTINUE
         IF (KMAX .EQ. J) GO TO 40
         DO 30 I = 1, M
            TEMP = A(I,J)
            A(I,J) = A(I,KMAX)
            A(I,KMAX) = TEMP
   30       CONTINUE
         RDIAG(KMAX) = RDIAG(J)
         WA(KMAX) = WA(J)
         K = IPVT(J)
         IPVT(J) = IPVT(KMAX)
         IPVT(KMAX) = K
   40    CONTINUE
C
C        COMPUTE THE HOUSEHOLDER TRANSFORMATION TO REDUCE THE
C        J-TH COLUMN OF A TO A MULTIPLE OF THE J-TH UNIT VECTOR.
C
         AJNORM = ENORM(M-J+1,A(J,J))
         IF (AJNORM .EQ. ZERO) GO TO 100
         IF (A(J,J) .LT. ZERO) AJNORM = -AJNORM
         DO 50 I = J, M
            A(I,J) = A(I,J)/AJNORM
   50       CONTINUE
         A(J,J) = A(J,J) + ONE
C
C        APPLY THE TRANSFORMATION TO THE REMAINING COLUMNS
C        AND UPDATE THE NORMS.
C
         JP1 = J + 1
         IF (N .LT. JP1) GO TO 100
         DO 90 K = JP1, N
            SUM = ZERO
            DO 60 I = J, M
               SUM = SUM + A(I,J)*A(I,K)
   60          CONTINUE
            TEMP = SUM/A(J,J)
            DO 70 I = J, M
               A(I,K) = A(I,K) - TEMP*A(I,J)
   70          CONTINUE
            IF (.NOT.PIVOT .OR. RDIAG(K) .EQ. ZERO) GO TO 80
            TEMP = A(J,K)/RDIAG(K)
            RDIAG(K) = RDIAG(K)*SQRT(MAX(ZERO,ONE-TEMP**2))
            IF (P05*(RDIAG(K)/WA(K))**2 .GT. EPSMCH) GO TO 80
            RDIAG(K) = ENORM(M-J,A(JP1,K))
            WA(K) = RDIAG(K)
   80       CONTINUE
   90       CONTINUE
  100    CONTINUE
         RDIAG(J) = -AJNORM
  110    CONTINUE
      RETURN
C
C     LAST CARD OF SUBROUTINE QRFAC.
C
      END
      SUBROUTINE QRSOLV(N,R,LDR,IPVT,DIAG,QTB,X,SDIAG,WA)
C
      IMPLICIT $(TYPE) (A-H,O-Z)
      INTEGER N,LDR
      INTEGER IPVT(N)
      DIMENSION R(LDR,N),DIAG(N),QTB(N),X(N),SDIAG(N),WA(N)
C     **********
C
C     SUBROUTINE QRSOLV
C
C     GIVEN AN M BY N MATRIX A, AN N BY N DIAGONAL MATRIX D,
C     AND AN M-VECTOR B, THE PROBLEM IS TO DETERMINE AN X WHICH
C     SOLVES THE SYSTEM
C
C           A*X = B ,     D*X = 0 ,
C
C     IN THE LEAST SQUARES SENSE.
C
C     THIS SUBROUTINE COMPLETES THE SOLUTION OF THE PROBLEM
C     IF IT IS PROVIDED WITH THE NECESSARY INFORMATION FROM THE
C     QR FACTORIZATION, WITH COLUMN PIVOTING, OF A. THAT IS, IF
C     A*P = Q*R, WHERE P IS A PERMUTATION MATRIX, Q HAS ORTHOGONAL
C     COLUMNS, AND R IS AN UPPER TRIANGULAR MATRIX WITH DIAGONAL
C     ELEMENTS OF NONINCREASING MAGNITUDE, THEN QRSOLV EXPECTS
C     THE FULL UPPER TRIANGLE OF R, THE PERMUTATION MATRIX P,
C     AND THE FIRST N COMPONENTS OF (Q TRANSPOSE)*B. THE SYSTEM
C     A*X = B, D*X = 0, IS THEN EQUIVALENT TO
C
C                  T       T
C           R*Z = Q *B ,  P *D*P*Z = 0 ,
C
C     WHERE X = P*Z. IF THIS SYSTEM DOES NOT HAVE FULL RANK,
C     THEN A LEAST SQUARES SOLUTION IS OBTAINED. ON OUTPUT QRSOLV
C     ALSO PROVIDES AN UPPER TRIANGULAR MATRIX S SUCH THAT
C
C            T   T               T
C           P *(A *A + D*D)*P = S *S .
C
C     S IS COMPUTED WITHIN QRSOLV AND MAY BE OF SEPARATE INTEREST.
C
C     THE SUBROUTINE STATEMENT IS
C
C       SUBROUTINE QRSOLV(N,R,LDR,IPVT,DIAG,QTB,X,SDIAG,WA)
C
C     WHERE
C
C       N IS A POSITIVE INTEGER INPUT VARIABLE SET TO THE ORDER OF R.
C
C       R IS AN N BY N ARRAY. ON INPUT THE FULL UPPER TRIANGLE
C         MUST CONTAIN THE FULL UPPER TRIANGLE OF THE MATRIX R.
C         ON OUTPUT THE FULL UPPER TRIANGLE IS UNALTERED, AND THE
C         STRICT LOWER TRIANGLE CONTAINS THE STRICT UPPER TRIANGLE
C         (TRANSPOSED) OF THE UPPER TRIANGULAR MATRIX S.
C
C       LDR IS A POSITIVE INTEGER INPUT VARIABLE NOT LESS THAN N
C         WHICH SPECIFIES THE LEADING DIMENSION OF THE ARRAY R.
C
C       IPVT IS AN INTEGER INPUT ARRAY OF LENGTH N WHICH DEFINES THE
C         PERMUTATION MATRIX P SUCH THAT A*P = Q*R. COLUMN J OF P
C         IS COLUMN IPVT(J) OF THE IDENTITY MATRIX.
C
C       DIAG IS AN INPUT ARRAY OF LENGTH N WHICH MUST CONTAIN THE
C         DIAGONAL ELEMENTS OF THE MATRIX D.
C
C       QTB IS AN INPUT ARRAY OF LENGTH N WHICH MUST CONTAIN THE FIRST
C         N ELEMENTS OF THE VECTOR (Q TRANSPOSE)*B.
C
C       X IS AN OUTPUT ARRAY OF LENGTH N WHICH CONTAINS THE LEAST
C         SQUARES SOLUTION OF THE SYSTEM A*X = B, D*X = 0.
C
C       SDIAG IS AN OUTPUT ARRAY OF LENGTH N WHICH CONTAINS THE
C         DIAGONAL ELEMENTS OF THE UPPER TRIANGULAR MATRIX S.
C
C       WA IS A WORK ARRAY OF LENGTH N.
C
C     SUBPROGRAMS CALLED
C
C       FORTRAN-SUPPLIED ... ABS,SQRT
C
C     ARGONNE NATIONAL LABORATORY. MINPACK PROJECT. MARCH 1980.
C     BURTON S. GARBOW, KENNETH E. HILLSTROM, JORGE J. MORE
C
C     **********
      INTEGER I,J,JP1,K,KP1,L,NSING
      DATA P5,P25,ZERO /5.0E-1,2.5E-1,0.0E0/
C
C     COPY R AND (Q TRANSPOSE)*B TO PRESERVE INPUT AND INITIALIZE S.
C     IN PARTICULAR, SAVE THE DIAGONAL ELEMENTS OF R IN X.
C
      DO 20 J = 1, N
         DO 10 I = J, N
            R(I,J) = R(J,I)
   10       CONTINUE
         X(J) = R(J,J)
         WA(J) = QTB(J)
   20    CONTINUE
C
C     ELIMINATE THE DIAGONAL MATRIX D USING A GIVENS ROTATION.
C
      DO 100 J = 1, N
C
C        PREPARE THE ROW OF D TO BE ELIMINATED, LOCATING THE
C        DIAGONAL ELEMENT USING P FROM THE QR FACTORIZATION.
C
         L = IPVT(J)
         IF (DIAG(L) .EQ. ZERO) GO TO 90
         DO 30 K = J, N
            SDIAG(K) = ZERO
   30       CONTINUE
         SDIAG(J) = DIAG(L)
C
C        THE TRANSFORMATIONS TO ELIMINATE THE ROW OF D
C        MODIFY ONLY A SINGLE ELEMENT OF (Q TRANSPOSE)*B
C        BEYOND THE FIRST N, WHICH IS INITIALLY ZERO.
C
         QTBPJ = ZERO
         DO 80 K = J, N
C
C           DETERMINE A GIVENS ROTATION WHICH ELIMINATES THE
C           APPROPRIATE ELEMENT IN THE CURRENT ROW OF D.
C
            IF (SDIAG(K) .EQ. ZERO) GO TO 70
            IF (ABS(R(K,K)) .GE. ABS(SDIAG(K))) GO TO 40
               COTAN = R(K,K)/SDIAG(K)
               SIN = P5/SQRT(P25+P25*COTAN**2)
               COS = SIN*COTAN
               GO TO 50
   40       CONTINUE
               TAN = SDIAG(K)/R(K,K)
               COS = P5/SQRT(P25+P25*TAN**2)
               SIN = COS*TAN
   50       CONTINUE
C
C           COMPUTE THE MODIFIED DIAGONAL ELEMENT OF R AND
C           THE MODIFIED ELEMENT OF ((Q TRANSPOSE)*B,0).
C
            R(K,K) = COS*R(K,K) + SIN*SDIAG(K)
            TEMP = COS*WA(K) + SIN*QTBPJ
            QTBPJ = -SIN*WA(K) + COS*QTBPJ
            WA(K) = TEMP
C
C           ACCUMULATE THE TRANFORMATION IN THE ROW OF S.
C
            KP1 = K + 1
            IF (N .LT. KP1) GO TO 70
            DO 60 I = KP1, N
               TEMP = COS*R(I,K) + SIN*SDIAG(I)
               SDIAG(I) = -SIN*R(I,K) + COS*SDIAG(I)
               R(I,K) = TEMP
   60          CONTINUE
   70       CONTINUE
   80       CONTINUE
   90    CONTINUE
C
C        STORE THE DIAGONAL ELEMENT OF S AND RESTORE
C        THE CORRESPONDING DIAGONAL ELEMENT OF R.
C
         SDIAG(J) = R(J,J)
         R(J,J) = X(J)
  100    CONTINUE
C
C     SOLVE THE TRIANGULAR SYSTEM FOR Z. IF THE SYSTEM IS
C     SINGULAR, THEN OBTAIN A LEAST SQUARES SOLUTION.
C
      NSING = N
      DO 110 J = 1, N
         IF (SDIAG(J) .EQ. ZERO .AND. NSING .EQ. N) NSING = J - 1
         IF (NSING .LT. N) WA(J) = ZERO
  110    CONTINUE
      IF (NSING .LT. 1) GO TO 150
      DO 140 K = 1, NSING
         J = NSING - K + 1
         SUM = ZERO
         JP1 = J + 1
         IF (NSING .LT. JP1) GO TO 130
         DO 120 I = JP1, NSING
            SUM = SUM + R(I,J)*WA(I)
  120       CONTINUE
  130    CONTINUE
         WA(J) = (WA(J) - SUM)/SDIAG(J)
  140    CONTINUE
  150 CONTINUE
C
C     PERMUTE THE COMPONENTS OF Z BACK TO COMPONENTS OF X.
C
      DO 160 J = 1, N
         L = IPVT(J)
         X(L) = WA(J)
  160    CONTINUE
      RETURN
C
C     LAST CARD OF SUBROUTINE QRSOLV.
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE LNS ( POINT, GRAD, DIR, NOD, F0, RHO, SIGMA, MAXIT,
     &                 FTARG, AL, FAL, COMACC, W, XTG, METHOD, NFC,
     &                 NGC, INFO, LIM, XLIM, MARG, XLL, XRL, IXAT )
C  ---------------------------------------------------------------------
C
C  Description:
C    Line search for the quasi-newton methods as described in:
C      [1] Al-Baali M. & Fletcher R.
C          JOTA vol. 48, No 3, pp. 259-377, 1986.
C          (algorithms S1 and S2)
C    Several modifications as suggested in:
C      [2] Numerical methods for unconstrained optimization and
C          nonlinear equations
C          Dennis J.E. & Schnabel R.B.
C          Prentice Hall, Englewood Cliffs, New Jersey, 1983.
C    This algorithm is not a general one-dimensional minimization
C    procedure. It is meant for the location of "acceptable points"
C    that obey the Wolfe conditions:
C     1) F(AL) <= F(0) + AL * RHO * F'(0)
C    and one of:
C     2) F'(AL) >= SIGMA * F'(0)
C     3) ABS(F'(AL)) <= -SIGMA * F'(0)
C
C  Input arguments:
C    POINT      The starting point.
C    GRAD       The gradient vector evaluated at POINT.
C    DIR        The direction along which the line search will be
C               performed.
C    NOD        Dimensionality of the objective funcction.
C    F0         Function value at the starting point POINT.
C    RHO        The "rho" parameter (see ref. [1]).
C    SIGMA      The "sigma" parameter (see ref. [1]).
C    MAXIT      Maximum allowed number of line search iterations.
C    FTARG      Target value.
C    COMMAC     Machine accuracy.
C    W          Work array.
C    XTG        Work array.
C    METHOD     METHOD = 1 or 2 for algorithms S1 or S2, respectively.
C               S1 uses conditions 1 and 2 while S2 uses the strong
C               Wolfe conditions 1 and 3 (see the Description section).
C               Note that S1 does not reduce to an exact line search
C               for limiting values of the "rho" and "sigma"
C               parameters. Hence it is not appropriate for conjugate
C               gradient methods.
C    LIM        LIM = 0 if a right margin does not exist;
C               non zero otherwise.
C    XLIM       Right margin for AL if LIM.NE.0; Ignored otherwise.
C    MARG       Margin statuses.
C    XLL        Left margins.
C    XRL        Right margins.
C    IXAT       Fix statuses.
C
C  Output arguments:
C    AL         The minimizing point along the direction DIR.
C    FAL        The value of the objective function at POINT + AL * DIR.
C    NFC        Number of function evaluations that were performed.
C    NFG        Number of gradient evaluations that were performed.
C    INFO       The reason for termination:
C               INFO = 0 -> Both line-search criteria are satisfied.
C               INFO = 1 -> All line search iterations have been
C                           exhausted.
C               INFO = 2 -> Not a descent direction. The directional
C                           derivative along dir is >= 0.
C               INFO = 3 -> Resulting AL is too small.
C               INFO = 4 -> Interpolating schemes have failed.
C               INFO = 5 -> Target value has been reached.
C               INFO = 6 -> Termination due to round off.
C               INFO = 7 -> Minimum seems to be out of margin.
C
C  Notes:
C    Regardless of the value of INFO, the calling routine may check
C    if subroutine LNS achieved any progress by testing if AL.NE.0
C
C  ---------------------------------------------------------------------
      IMPLICIT $(TYPE) (A-H,O-Z)
C  Arguments:
      DIMENSION POINT(NOD), GRAD(NOD), DIR(NOD)
      DIMENSION W(NOD), XTG(NOD)
      DIMENSION IXAT(NOD), MARG(NOD), XLL(NOD), XRL(NOD)
C  ---------------------------------------------------------------------
C  Local variables:
      LOGICAL NOB, FORCEB
C  ---------------------------------------------------------------------
C
C  Some constants.
      ONE = 1.
      TEN = 10.
C
C  Truncation constants.
      T1 = 0.1
      T2 = 0.5
      T3 = 1+T1
      T4 = 10.
      T5 = 0.1
      T6 = 0.5
C
C  FLOW is the lowest function value so far. It need not satisfy
C  the line-search conditions and is returned whenever a normal
C  return is not possible.
      FLOW = F0
      ALLOW = 0.
C
C  Zero counters (function calls, gradient calls).
      NFC = 0
      NGC = 0
C
C  Calculate directional derivative at the starting point.
      CALL VVMULT(GRAD,DIR,G0,NOD)
      IF (G0.GE.0.) THEN
C  This is not a descent direction.
         AL = ALLOW
         FAL = FLOW
         INFO = 2
         RETURN
      END IF
C
C  Obtain a lower bound on AL.
      XMAX = 0.
      DO 40,I=1,NOD
40       XMAX = MAX(XMAX,ABS(DIR(I))/MAX(ABS(POINT(I)),ONE))
      ALAMIN = MIN(COMACC/XMAX,TEN)
C
C  Initial values for the left bracket.
      A = 0.
      FA = F0
      GA = G0
C
C  NOB = .TRUE. indicates that a right bracket has not been found.
      NOB = .TRUE.
C
C  IBK is the number of backtracks towards the left bracket.
C  Note that the left bracket always satisfies the rho-condition.
      IBK = 0
C
C  ITER is number of iterations of the method.
      ITER = 0
C
      FORCEB = .FALSE.
100   CONTINUE
      ITER = ITER+1
C
C  If AL is too small, terminate the line search, since convergence
C  would be detected anyway by the calling routine. (X-convergence)
      IF (AL.LT.ALAMIN) THEN
         AL = ALLOW
         FAL = FLOW
         INFO = 3
         RETURN
      END IF
C
C  Is AL getting too close on the A-bracket ?
C  This is probably an indication of severe round off error near
C  the solution.
      IF (ABS(A-AL)/MAX(A,ONE).LT.ALAMIN) THEN
         AL = ALLOW
         FAL = FLOW
         INFO = 3
         RETURN
      END IF
C
C  Are we getting out of the right margin ?
      IF (LIM.NE.0 .AND. AL.GT.XLIM) THEN
         IF (FORCEB) THEN
            AL = ALLOW
            FAL = FLOW
            INFO = 7
            RETURN
         ELSE
            AL = XLIM
            FORCEB = .TRUE.
         END IF
      ELSE
         FORCEB = .FALSE.
      END IF
C
C  Evaluate the objective function. When margins are present, small
C  round off errors may force the point out of the margins. Take care.
      CALL MEVAL(POINT,DIR,AL,NOD,IXAT,MARG,XLL,XRL,W,1,FAL)
      NFC = NFC+1
      IF (FAL.LT.FLOW) THEN
         FLOW = FAL
         ALLOW = AL
      END IF
      IF (FAL.LE.FTARG) THEN
C  Function value less than target.
         INFO = 5
         RETURN
      END IF
C
      IF (FAL.GT.F0+AL*RHO*G0 .OR. FAL.GE.FA) THEN
C  Above the rho-line, or greater than the A-bracket.
         IBK = IBK+1
         IF (IBK.EQ.1) THEN
            IF (NOB) THEN
               ISIDE = 0
            ELSE
               ISIDE = 0
               IF ((B-A).LT.0.) ISIDE = 1
            END IF
C
C  This is the first backtrack. Use a parabola.
            CALL QMIN(A,FA,GA,AL,FAL,NR,XMIN)
            IF (NR.EQ.0) THEN
C  Interpolation has failed.
               AL = ALLOW
               FAL = FLOW
               INFO = 4
               RETURN
            END IF
            AAL = ABS(A-AL)
            IF (ISIDE.EQ.0) THEN
               IF (XMIN.LT.A+T1*AAL) XMIN = A+T1*AAL
            ELSE
               IF (XMIN.GT.A-T1*AAL) XMIN = A-T1*AAL
            END IF
         ELSE
C
C  On the second and subsequent backtracks, use a cubic polynomial.
            CALL CMIN2(A,FA,GA,AL,FAL,B,FB,NR,XMIN)
            IF (NR.EQ.0) THEN
               CALL QMIN(A,FA,GA,AL,FAL,NR,XMIN)
               IF (NR.EQ.0) THEN
C  Interpolation has failed.
                  AL = ALLOW
                  FAL = FLOW
                  INFO = 4
                  RETURN
               END IF
            END IF
            CALL KEEPT(A,AL,T1,T2,XMIN)
         END IF
         B = AL
         FB = FAL
         NOB = .FALSE.
      ELSE
C  Below rho-line, less than A-bracket.
         IBK = 0
C  Evaluate the directional derivative.
         DO 30,I=1,NOD
30          W(I) = POINT(I)+AL*DIR(I)
         CALL DERMER(W,FAL,XTG,NOCD,NGR,0)
         CALL VVMULT(DIR,XTG,GAL,NOD)
         NFC = NFC+NOCD
         NGC = NGC+NGR
         IF (METHOD.EQ.1) THEN
C  Algorithm S1.
            IF (GAL.GE.SIGMA*G0) THEN
C  Both line search conditions are satisfied.
               INFO = 0
               RETURN
            END IF
         ELSE
C  Algorithm S2.
            IF (ABS(GAL).LE.-SIGMA*G0) THEN
C  Both line search conditions are satisfied.
               INFO = 0
               RETURN
            END IF
         END IF
C
C  The gradient-criterion is not satisfied.
         IF (NOB) THEN
            ISIDE = 0
            IF (GAL.GT.0) ISIDE = 1
         ELSE
            ISIDE = 1
            IF ((B-A)*GAL.LT.0.) ISIDE = 0
         END IF
C
C  ISIDE should always be 0 when METHOD = 1.
         IF (METHOD.EQ.1 .AND. ISIDE.EQ.1)
     &          CALL MABORT('METHOD = 1, ISIDE = 1','LNS')
         IF (ISIDE.EQ.0) THEN
            IF (GAL.GT.GA) THEN
C  The function is convex.
               CALL CMIN(A,FA,GA,AL,FAL,GAL,NR,XMIN)
               IF (NR.EQ.0) THEN
C  Quadratic extrapolation. Use the slopes at A, AL.
                  AA = (GA-GAL)/(2.*(A-AL))
                  BB = GA-2*AA*A
                  XMIN = -BB/(2*AA)
               END IF
               CALL KEEPE(A,AL,XMIN,B,NOB,T3,T4,T5,T6)
            ELSE
C  The function is concave.
C  Interpolating schemes are no good if the slopes indicate that this
C  is a concave part of the function. (A maximum would actualy be
C  found.)
               IF (NOB) THEN
C  A right bracket has not been found. Try to establish a right
C  bracket; use the right side of the allowed interval.
                  XMIN = T4*AL
               ELSE
C  A right bracket is available.
                  CALL CMIN2(AL,FAL,GAL,A,FA,B,FB,NR,XMIN)
                  IF (NR.EQ.0) THEN
                     CALL QMIN(AL,FAL,GAL,B,FB,NR,XMIN)
                     IF (NR.EQ.0) THEN
C  Interpolation has failed.
                        AL = ALLOW
                        FAL = FLOW
                        INFO = 4
                        RETURN
                     END IF
                  END IF
                  CALL KEEPE(A,AL,XMIN,B,NOB,T3,T4,T5,T6)
               END IF
            END IF
         ELSE
C  Other side.
            CALL CMIN(A,FA,GA,AL,FAL,GAL,NR,XMIN)
            IF (NR.EQ.0) THEN
               CALL QMIN(AL,FAL,GAL,A,FA,NR,XMIN)
               IF (NR.EQ.0) THEN
C  Interpolation has failed.
                  AL = ALLOW
                  FAL = FLOW
                  INFO = 4
                  RETURN
               END IF
            END IF
            CALL KEEPT(A,AL,T1,T2,XMIN)
            B = A
            FB = FA
            NOB = .FALSE.
         END IF
         A = AL
         FA = FAL
         GA = GAL
      END IF
      AL = XMIN
C
C  Perform another iteration.
      IF (ITER.LT.MAXIT) GOTO 100
C
C  The allowed number of line search iterations has been exhausted.
      AL = ALLOW
      FAL = FLOW
      INFO = 1
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE KEEPT ( A, AL, T1, T2, XMIN )
C  ---------------------------------------------------------------------
C
C  Description:
C    Called by subroutine LNS. Keeps interpolated values inside
C    the allowed interval. See ref. [1] in the description of
C    subroutine LNS.
C
C  Input arguments:
C    A          Left end of the interpolating interval.
C    AL         Right end of the interpolating interval.
C    T1         T1 constant. See ref. [1].
C    T2         T2 constant. See ref. [1].
C
C  Output arguments:
C    XMIN       On input, the result of an intepolation. On output it
C               is forced to lie in between A, B as described
C               in ref. [1].
C
C  ---------------------------------------------------------------------
      IMPLICIT $(TYPE) (A-H,O-Z)
C
      AALEN = ABS(A-AL)
C
      IF (A.LT.AL) THEN
         XLEFT = A+T1*AALEN
         XRIGHT = AL-T2*AALEN
      ELSE
         XLEFT = AL+T2*AALEN
         XRIGHT = A-T1*AALEN
      END IF
C
      IF (XMIN.LT.XLEFT) THEN
         XMIN = XLEFT
      ELSE IF (XMIN.GT.XRIGHT) THEN
         XMIN = XRIGHT
      END IF
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE KEEPE ( A, AL, XMIN, B, NOB, T3, T4, T5, T6 )
C  ---------------------------------------------------------------------
C
C  Description:
C    Called by subroutine LNS. Keeps extrapolated values inside
C    the allowed interval. See ref. [1] in the description of
C    subroutine LNS.
C
C  Input arguments:
C    A          Left bracket.
C    AL         Current value of AL.
C    B          Right bracket.
C    NOB        .TRUE. if a right bracket has not been established.
C    T3         T3 constant. See ref. [1].
C    T4         T4 constant. See ref. [1].
C    T5         T5 constant. See ref. [1].
C    T6         T6 constant. See ref. [1].
C
C  Output arguments:
C    XMIN       On input, the result of an extrapolation. On output it
C               is forced to lie in between AL, B as described
C               in ref. [1].
C
C  ---------------------------------------------------------------------
      IMPLICIT $(TYPE) (A-H,O-Z)
C  Arguments:
      LOGICAL NOB
C  ---------------------------------------------------------------------
C
      ALBLEN = ABS(AL-B)
      IF (A.LT.AL) THEN
         IF (NOB) THEN
            XLEFT = T3*AL
            XRIGHT = T4*AL
         ELSE
            XLEFT = AL+T5*ALBLEN
            XRIGHT = B-T6*ALBLEN
         END IF
      ELSE
         XLEFT = B+T6*ALBLEN
         XRIGHT = AL-T5*ALBLEN
      END IF
C
      IF (XMIN.LT.XLEFT) THEN
         XMIN = XLEFT
      ELSE IF (XMIN.GT.XRIGHT) THEN
         XMIN = XRIGHT
      END IF
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE LTRANS ( H, LD, X )
C  ---------------------------------------------------------------------
C
C  Description:
C    This subroutine returns the Choleski factor L[t] (of the Hessian
C    matrix) with respect to a set of transformed variables.
C    This is necessary for minimization methods that handle bounds via
C    transformations.
C    See also subroutines XTOY and YTOX.
C
C  Input arguments:
C    LD         The leading dimension of matrix H.
C    X          The transformed variables.
C
C  Input / Output arguments:
C    H          On input the upper triangular part contains a
C               Choleski factor L[t].
C               On output the upper triangular part contains the
C               Choleski factor L[t] with respect to the transformed
C               variables.
C
C  ---------------------------------------------------------------------
      IMPLICIT $(TYPE) (A-H,O-Z)
C  Arguments:
      DIMENSION H(LD,*), X(*)
C  ---------------------------------------------------------------------
C  Common blocks:
      PARAMETER ( MXV = $(MXV) )
      COMMON /DETAIL/ IXAT(MXV), MARG(MXV)
      COMMON /XLIMS/ XLL(MXV), XRL(MXV)
      COMMON /DIMEN/ NOD, NOF
C  ---------------------------------------------------------------------
C
      TWO = 2.
      MM = 0
      DO 10,I=1,NOD
         IF (IXAT(I).NE.0) THEN
            MM = MM+1
            MA = MARG(I)
            IF (MA.EQ.0) THEN
               CON = 1.
            ELSE IF (MA.EQ.-1) THEN
               CON = TWO*X(I)
            ELSE IF (MA.EQ.1) THEN
               CON = -TWO*X(I)
            ELSE
               YY = X(I)
               CON = -TWO*YY*(XLL(I)-XRL(I))*EXP(-YY**2)
            END IF
            H(MM,MM) = H(MM,MM)*CON
         END IF
10    CONTINUE
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE JROT ( A, I, AA, BB, N, LD )
C  ---------------------------------------------------------------------
C
C  Description:
C    Performs a Jacobi rotation on rows I and I+1 of the NxN matrix A.
C    The parameters of the rotation are:
C      COS(phi) = AA/SQRT(AA**2+BB**2)  SIN(phi) = BB/SQRT(AA**2+BB**2)
C
C    For more information see:
C
C    Dennis J.E. and Schnabel R.B.
C    Numerical Methods for Unconstrained Optimization and Nonlinear
C    Equations
C    Prentice-Hall, 1983. p. 55-58.
C
C  Input arguments:
C    I          Rotation is performed on rows I, I+1.
C    AA         First parameter of the rotation.
C    BB         Second parameter of the rotation.
C    N          Dimension of matrix A.
C    LD         First (leading) dimension of matrix A.
C
C  Input / Output arguments:
C    A          NxN matrix to be rotated.
C
C  ---------------------------------------------------------------------
      IMPLICIT $(TYPE) (A-H,O-Z)
C  Arguments:
      DIMENSION A(LD,N)
C  ---------------------------------------------------------------------
C
      ONE = 1.
      IF (AA.EQ.0.) THEN
         C = 0.
         S = SIGN(ONE,BB)
      ELSE
         R = SQRT(AA**2+BB**2)
         C = AA/R
         S = BB/R
      END IF
C
      I1 = I+1
      DO 10,J=I,N
         Y = A(I,J)
         W = A(I1,J)
         A(I,J) = C*Y-S*W
         A(I1,J) = S*Y+C*W
10    CONTINUE
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE VVMULT ( X, Y, XY, N )
C  ---------------------------------------------------------------------
C
C  Description:
C    Computes the vector inner product X * Y.
C
C  Input arguments:
C    X          First vecror.
C    Y          Second vector.
C    N          Dimensionality of the vectors.
C
C  Output arguments:
C    XY         The inner product X*Y.
C
C  ---------------------------------------------------------------------
      IMPLICIT $(TYPE) (A-H,O-Z)
C  Arguments:
      DIMENSION X(N), Y(N)
C  ---------------------------------------------------------------------
C
      XY = 0.
      DO 10,I=1,N
10       XY = XY+X(I)*Y(I)
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE LLSOLV ( A, X, B, N, LD )
C  ---------------------------------------------------------------------
C
C  Description:
C    Solves the set of linear equations A * X = B by direct forward and
C    backward substitution. A is input as its upper triangular Choleski
C    factor L[t].
C
C  Input arguments:
C    A          Upper triangular (Choleski factor L[t]).
C    B          The right hand vector.
C    N          Dimensionality of the problem.
C    LD         Leading (first) dimension of matrix A.
C
C  Output arguments:
C    X          The solution vector.
C
C  Notes:
C    X and B normally are two different vectors. If however the calling
C    routine uses the same vector for X and B, the algorithm will work
C    and B will be overwritten.
C
C  ---------------------------------------------------------------------
      IMPLICIT $(TYPE) (A-H,O-Z)
C  Arguments:
      DIMENSION A(LD,N), B(N), X(N)
C  ---------------------------------------------------------------------
C
      DO 10,I=1,N
         SUM = 0.
         DO 20,J=1,I-1
20          SUM = SUM+A(J,I)*X(J)
         X(I) = (B(I)-SUM)/A(I,I)
10    CONTINUE
C
      DO 30,I=N,1,-1
         SUM = 0.
         DO 40,J=I+1,N
40          SUM = SUM+A(I,J)*X(J)
         X(I) = (X(I)-SUM)/A(I,I)
30    CONTINUE
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE LMSOLV ( A, X, B, N, LD )
C  ---------------------------------------------------------------------
C
C  Description:
C    Solves L * X = B.
C    L is input as the Choleski factor L[t].
C
C  Input arguments:
C    A          Upper triangular (Choleski factor L[t]).
C    B          The right hand vector.
C    N          Dimensionality of the problem.
C    LD         Leading (first) dimension of matrix A.
C
C  Output arguments:
C    X          The solution vector.
C
C  Notes:
C    X and B normally are two different vectors. If however the calling
C    routine uses the same vector for X and B, the algorithm will work
C    and B will be overwritten.
C
C  ---------------------------------------------------------------------
      IMPLICIT $(TYPE) (A-H,O-Z)
C  Arguments:
      DIMENSION A(LD,N), B(N), X(N)
C  ---------------------------------------------------------------------
C
      DO 10,I=1,N
         SUM = 0.
         DO 20,J=1,I-1
20          SUM = SUM+A(J,I)*X(J)
         X(I) = (B(I)-SUM)/A(I,I)
10    CONTINUE
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE LTMULT ( A, X, B, N, LD )
C  ---------------------------------------------------------------------
C
C  Description:
C    Performs the multiplication B = L[t] * X
C    where L[t] is an upper triangular Cloleski factor.
C
C  Input arguments:
C    A          Upper triangular (Choleski factor L[t]).
C    X          Vector.
C    N          Dimensionality of the problem.
C    LD         Leading (first) dimension of matrix A.
C
C  Output arguments:
C    B          The resulting left hand vector.
C
C  Notes:
C    X and B normally are two different vectors. If however the calling
C    routine uses the same vector for X and B, the algorithm will work
C    and X will be overwritten.
C
C  ---------------------------------------------------------------------
      IMPLICIT $(TYPE) (A-H,O-Z)
C  Arguments:
      DIMENSION A(LD,N), X(N), B(N)
C  ---------------------------------------------------------------------
C
      DO 10,I=1,N
         SUM = 0.
         DO 20,J=I,N
20          SUM = SUM+A(I,J)*X(J)
         B(I) = SUM
10    CONTINUE
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE LMULT ( A, X, B, N, LD )
C  ---------------------------------------------------------------------
C
C  Description:
C    Performs the multiplication B = L * X.
C    where L is a lower triangular Choleski factor.
C    L is input as its traspose L[t].
C
C  Input arguments:
C    A          Upper triangular (Choleski factor L[t]).
C    X          Vector.
C    N          Dimensionality of the problem.
C    LD         Leading (first) dimension of matrix A.
C
C  Output arguments:
C    B          The resulting left hand vector.
C
C  Notes:
C    X and B normally are two different vectors. If however the calling
C    routine uses the same vector for X and B, the algorithm will work
C    and X will be overwritten.
C
C  ---------------------------------------------------------------------
      IMPLICIT $(TYPE) (A-H,O-Z)
C  Arguments:
      DIMENSION A(LD,N), X(N), B(N)
C  ---------------------------------------------------------------------
C
      DO 10,I=N,1,-1
         SUM = 0.
         DO 20,J=1,I
20          SUM = SUM+A(J,I)*X(J)
         B(I) = SUM
10    CONTINUE
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE LLMULT ( A, X, B, N, LD )
C  ---------------------------------------------------------------------
C
C  Description:
C    Performs the multiplication B = X[t] * L * L[t] * X
C    where X is a vector and L is a lower triangular Choleski factor.
C    L is input as its traspose L[t].
C
C  Input arguments:
C    A          Upper triangular (Choleski factor L[t]).
C    X          Vector.
C    N          Dimensionality of the problem.
C    LD         Leading (first) dimension of matrix A.
C
C  Output arguments:
C    B          Result of the multiplication (scalar).
C
C  ---------------------------------------------------------------------
      IMPLICIT $(TYPE) (A-H,O-Z)
C  Arguments:
      DIMENSION A(LD,N), X(N)
C  ---------------------------------------------------------------------
C
      B = 0.
      DO 10,I=1,N
         SUM = 0.
         DO 20,J=I,N
20          SUM = SUM+A(I,J)*X(J)
         B = B+SUM**2
10    CONTINUE
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE LLUPD ( A, V, U, N, LD )
C  ---------------------------------------------------------------------
C
C  Description:
C    Updates the Choleski factors of a matrix with the rank 2 update:
C    L[new] = L + V * U[t]
C    The matrix is input as the upper triangular Choleski factor L[t].
C
C    For more information see:
C
C    Dennis J.E. and Schnabel R.B.
C    Numerical Methods for Unconstrained Optimization and Nonlinear
C    Equations
C    Prentice-Hall, 1983. p. 55-58.
C
C  Input arguments:
C    V          Vector.
C    U          Vector.
C    N          Dimensionality of the problem.
C    LD         Leading (first) dimension of matrix A.
C
C  Input / Output arguments:
C    A          Upper triangular (Choleski factor L[t]).
C               On output contains the updated factor.
C
C  ---------------------------------------------------------------------
      IMPLICIT $(TYPE) (A-H,O-Z)
C  Arguments:
      DIMENSION A(LD,N), U(N), V(N)
C  ---------------------------------------------------------------------
C
C  Zero first lower diagonal.
      DO 60,I=2,N
60       A(I,I-1) = 0.
C
C  Search for non-zero U(I).
      DO 10,K=N,1,-1
10       IF (U(K).NE.0.) GOTO 20
      K = 1
C
20    CONTINUE
      DO 30,I=K-1,1,-1
         AA = U(I)
         BB = -U(I+1)
         CALL JROT(A,I,AA,BB,N,LD)
         IF (AA.EQ.0.) THEN
            U(I) = ABS(BB)
         ELSE
            U(I) = SQRT(AA**2+BB**2)
         END IF
30    CONTINUE
C
      DO 40,J=1,N
40       A(1,J) = A(1,J)+U(1)*V(J)
C
      DO 50,I=1,K-1
         AA = A(I,I)
         BB = -A(I+1,I)
         CALL JROT(A,I,AA,BB,N,LD)
50    CONTINUE
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE LLCOMP ( A, N, LD )
C  ---------------------------------------------------------------------
C
C  Description:
C    Performs the multiplication A = L * L[t]
C    where L is a lower triangular Choleski factor.
C    L is input as its traspose L[t].
C
C  Input arguments:
C    N          Dimensionality of the problem.
C    LD         Leading (first) dimension of matrix A.
C
C  Input / Output arguments:
C    A          On input upper triangular (Choleski factor L[t]).
C               On output the product L * L[t]
C
C  ---------------------------------------------------------------------
      IMPLICIT $(TYPE) (A-H,O-Z)
C  Arguments:
      DIMENSION A(LD,N)
C  ---------------------------------------------------------------------
C
      DO 10,I=N,1,-1
         DO 20,J=N,I,-1
            S = 0.
            DO 30,K=1,I
30             S = S+A(K,I)*A(K,J)
            A(I,J) = S
20       CONTINUE
10    CONTINUE
C
C  Fill in lower part.
      DO 40,I=2,N
         DO 50,J=1,I-1
50          A(I,J) = A(J,I)
40    CONTINUE
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE MDIS ( POINT, VAL, ITER, NOC, MAXNOC, IPRINT, ITRA )
C  ---------------------------------------------------------------------
C
C  Description:
C    Issues informative messages during a minimization session.
C    Normally, it is called when a lower value is discovered.
C
C  Input arguments:
C    POINT      The current point.
C    VAL        Value of the objective function at POINT.
C    ITER       Number of iterations that have been performed by the
C               calling routine.
C    NOC        Number of function evaluations that have been performed
C               by the calling routine.
C    MAXNOC     The calling routine is allowed MAXNOC calls at most.
C    IPRINT     Printout level:
C               IPRINT = 0 -> Nothing is printed.
C               IPRINT = 1 -> Simple printout: Prints ITER, VAL,
C                             NOC, MAXNOC
C               IPRINT = 2 -> Full printout. Prints in addition the
C                             minimization parameters, contained in
C                             array POINT.
C    ITRA       Important only when IPRINT=2.
C               ITRA = 0  -> The calling method does not use
C                            transofrmations.
C               ITRA.NE.0 -> The calling uses tranformations. When
C                            IPRINT=2, we should apply the inverse
C                            transformation before printing the
C                            parameters.
C
C  ---------------------------------------------------------------------
      IMPLICIT $(TYPE) (A-H,O-Z)
C  Arguments:
      DIMENSION POINT(*)
C  ---------------------------------------------------------------------
C  Common blocks:
      LOGICAL STRO, WEAK
      COMMON /PRILOG/ STRO, WEAK
      COMMON /UNITS/ IUINP, IUOUT
      COMMON /DIMEN/ NOD, NOF
C  ---------------------------------------------------------------------
C  Local variables:
      PARAMETER ( MXV = $(MXV) )
      DIMENSION XTMP(MXV)
      CHARACTER*10 C1, C2, C3
      CHARACTER*23 CVAL
C  ---------------------------------------------------------------------
C
      CALL BACKUP('L')
      IF (.NOT.STRO .OR. IPRINT.EQ.0) RETURN
C
      IF (IPRINT.EQ.1) THEN
         CALL I2STR(ITER,C1,LE1)
         CALL I2STR(NOC,C2,LE2)
         CALL R2RTF(VAL,CVAL,LEV)
         IF (NOC.LE.MAXNOC) THEN
            CALL I2STR(MAXNOC,C3,LE3)
            WRITE (IUOUT,10) C1(1:LE1), CVAL,
     &                             C2(1:LE2), C3(1:LE3)
         ELSE
            WRITE (IUOUT,10) C1(1:LE1), CVAL, C2(1:LE2)
         END IF
         CALL DASH
      ELSE IF (IPRINT.EQ.2) THEN
         DO 30,I=1,NOD
30          XTMP(I) = POINT(I)
         IF (ITRA.NE.0) CALL YTOX(XTMP,NOD)
         CALL WRITER(XTMP,VAL,IUOUT,1,0)
         CALL I2STR(ITER,C1,LE1)
         CALL I2STR(NOC,C2,LE2)
         IF (NOC.LE.MAXNOC) THEN
            CALL I2STR(MAXNOC,C3,LE3)
            WRITE (IUOUT,20) C1(1:LE1), C2(1:LE2), C3(1:LE3)
         ELSE
            WRITE (IUOUT,20) C1(1:LE1), C2(1:LE2)
         END IF
         CALL DASH
      ELSE
         CALL I2STR(IPRINT,C1,LE1)
         CALL MABORT('Incorrect "iprint" value: '//C1(1:LE1),'MDIS')
      END IF
C
10    FORMAT ($(FC)3X,'Iter: ',A,3X,'Lower value:',3X,A,
     &        3X,'Calls: ',A:,' of ',A)
20    FORMAT ($(FC)3X,'Iter: ',A,3X,' Calls: ',A:,' of ',A)
      END
C  ---------------------------------------------------------------------
      SUBROUTINE MEVAL ( POINT, DIR, AL, NOD, IXAT, MARG, XLL, XRL,
     &                   WORK, IOPT, FAL )
C  ---------------------------------------------------------------------
C
C  Description:
C    Evaluates the objective function at WORK = POINT + AL * DIR.
C    AL is assumed to be in between the margins along the direction DIR.
C    Round off errors however, may force the new point slightly out of
C    the allowed range. This situation is corrected here. When IOPT = 0
C    the objective function is not evaluated.
C
C  Input arguments:
C    POINT      The current point.
C    DIR        The direction vector.
C    AL         Distance along the direction DIR.
C    NOD        Dimensionality of the objective function.
C    IXAT       Fix statuses.
C    MARG       Margin statuses.
C    XLL        Left margins.
C    XRL        Right margins.
C    IOPT       Determines the action to be taken:
C               IOPT = 0     -> The objective function is not evaluated.
C               IOPT .NE. 0  -> The value of the objective function
C                               at the new point WORK is returned
C                               in FAL.
C
C  Output arguments:
C    WORK       On output contains the new point: POINT + AL * DIR.
C    FAL        If IOPT.NE.0, FAL is the function value at the new
C               point WORK.
C
C  Notes:
C    POINT and WORK are normally two different arrays. The calling
C    routine however may use the save vector, in which case the contents
C    of POINT are overwritten.
C
C  ---------------------------------------------------------------------
      IMPLICIT $(TYPE) (A-H,O-Z)
C  Arguments:
      DIMENSION POINT(NOD), DIR(NOD), MARG(NOD), XLL(NOD), XRL(NOD)
      DIMENSION IXAT(NOD), WORK(NOD)
C  ---------------------------------------------------------------------
C
      DO 10,I=1,NOD
         IF (IXAT(I).NE.0) THEN
            WORK(I) = POINT(I)+AL*DIR(I)
            MA = MARG(I)
            IF (MA.EQ.-1 .OR. MA.EQ.2) WORK(I) = MAX(WORK(I),XLL(I))
            IF (MA.EQ. 1 .OR. MA.EQ.2) WORK(I) = MIN(WORK(I),XRL(I))
         ELSE
            WORK(I) = POINT(I)
         END IF
10    CONTINUE
      IF (IOPT.NE.0) FAL = ACSQ(WORK,NOD)
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE MINDIR ( DIR, NOCALL, ALPHA, XTOL, POINT, VAL, NOD,
     &                    INFO, MARG, XLL, XRL, IXAT, WORK, FTARG,
     &                    LSNOC )
C  ---------------------------------------------------------------------
C
C  Description:
C    Subroutine MINDIR performs one dimensional minimization on the
C    direction DIR without using any derivative information.
C    Initialy it brackets the minimum (subroutine BRAK) and then
C    uses the method of Brent (subroutine BRENT) to isolate the minimum.
C    It returns the new point: POINT + ALPHA * DIR.
C
C  Input arguments:
C    NOD        Dimensionality of the objective function.
C    XTOL       Convergence criterion. See subroutine BRENT.
C    MARG       Margin statuses.
C    XLL        Left margins.
C    XRL        Right margins.
C    IXAT       Fix statuses.
C    WORK       Work array of length NOD.
C    FTARG      Target value.
C    LSNOC      Maximum allowed number of function evaluations.
C
C  Output arguments:
C    NOCALL     Number of function evaluations that were performed.
C    ALPHA      The line search ALPHA, or 0 if no progress was possible.
C    INFO       Reason for termination of this routine:
C               INFO = 1 -> Convergence criterion is satisfied.
C               INFO = 2 -> The minimum probably lies outside the
C                           margins. The best point so far is returned.
C               INFO = 3 -> Target value has been reached.
C               INFO = 4 -> We are out of function calls.
C                           The best point so far is returned.
C               INFO = 5 -> On entry the direction of search was 0.
C               INFO = 6 -> All variables are fixed, or cannot move due
C                           to margins.
C               INFO = 7 -> Progress is not possible because the margins
C                           along the direction of search, coincide.
C                           Probably a variable has XLL(I) = XRL(I)
C                           and has not been fixed. Normally this should
C                           never happen.
C
C  Input / Output arguments:
C    DIR        On input the direction of search. It should be a
C               descent direction. On output, if any DIR(I) takes us
C               outside the magins, it is set to 0.
C               Warning: This may produce a non-descent direction.
C    POINT      On input, the current point. On output, the new point
C               after the line search.
C    VAL        Value of the objective function.
C
C  Notes:
C    Regardless of the value of INFO, the calling routine may check
C    if any progress was achieved, by testing if ALPHA.NE.0
C
C  ---------------------------------------------------------------------
      IMPLICIT $(TYPE) (A-H,O-Z)
C  Arguments:
      DIMENSION DIR(NOD), POINT(NOD), MARG(NOD), XLL(NOD), XRL(NOD)
      DIMENSION IXAT(NOD), WORK(NOD)
C  ---------------------------------------------------------------------
C
C  Some constants.
      ONE = 1.
      TENTH = 0.1
C
C  Initialize output variables.
      ALPHA = 0.
      NOCALL = 0
C
C  Calculate minimum acceptable ALPHA. Too small ALPHA values will
C  cause X-convergence to be detected.
      XMAX = 0.
      DO 40,I=1,NOD
40       XMAX = MAX(XMAX,ABS(DIR(I))/MAX(ABS(POINT(I)),ONE))
      IF (XMAX.EQ.0.) THEN
         INFO = 5
         RETURN
      END IF
      ALAMIN = MIN(XTOL/XMAX,TENTH)
C
C  Consider fixed the parameters that result in very small ALPHA.
      ALA = 10.*ALAMIN
      PP0 = 10.*ALA
      DO 11,I=1,NOD
         IF (DIR(I).EQ.0) GOTO 11
         MA = MARG(I)
         PP = PP0
         IF (DIR(I).GT.0.) THEN
            IF (MA.EQ. 1 .OR. MA.EQ.2) PP = (XRL(I)-POINT(I))/DIR(I)
         ELSE
            IF (MA.EQ.-1 .OR. MA.EQ.2) PP = (XLL(I)-POINT(I))/DIR(I)
         END IF
         IF (PP.LT.ALA) DIR(I) = 0.
11    CONTINUE
C
C  If fixed parameters, set direction vector to zero.
      NOFIX = 0
      DO 1,I=1,NOD
         IF (IXAT(I).EQ.0 .OR. DIR(I).EQ.0.) THEN
            DIR(I) = 0.
            NOFIX = NOFIX+1
         END IF
1     CONTINUE
C
C  If all parameters are fixed then return.
C  Note that at this point all parameters can be fixed due to FIX
C  statements, or due to the presence of margins.
      IF (NOFIX.EQ.NOD) THEN
         INFO = 6
         RETURN
      END IF
C
C  Calculate margins XALIM, XBLIM along the direction of search.
      CALL DIRLIM(POINT,DIR,MARG,XLL,XRL,NOD,LIM,XALIM,XBLIM)
      IF (LIM.EQ.2 .AND. XALIM-XBLIM.EQ.0.) THEN
         INFO = 7
         RETURN
      END IF
C
C  Calculate the bracketing triplet A, B, C.
      A = 0.
      FA = VAL
      B = 1.
      CALL BRAK(POINT,DIR,NOD,A,B,C,FA,FB,FC,ITER,NOC1,XALIM,XBLIM,
     &          LIM,INFO,FTARG,LSNOC,MARG,XLL,XRL,WORK,IXAT)
C
C  If bracketing was successful, proceed with Brent's method.
      IF (INFO.EQ.1) THEN
         NCB = LSNOC-NOC1
         IF (NCB.LE.0) THEN
            INFO = 4
            NOC2 = 0
            ALPHA = B
            FXMIN = FB
         ELSE
            CALL BRENT(POINT,DIR,NOD,A,B,C,FB,XTOL,ALPHA,FXMIN,
     &                       NCB,NOC2,INFO,FTARG,MARG,XLL,XRL,WORK,IXAT)
            IF (INFO.EQ.2) INFO = 4
         END IF
      ELSE
         NOC2 = 0
         ALPHA = B
         FXMIN = FB
      END IF
      NOCALL = NOC1+NOC2
C
C  Calculate the new point and make sure it lies inside the margins.
      CALL MEVAL(POINT,DIR,ALPHA,NOD,IXAT,MARG,XLL,XRL,POINT,0,FTMP)
      VAL = FXMIN
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE MINREP ( TMMESS, NOCEX, NOGEX, NOJAC, ITER )
C  ---------------------------------------------------------------------
C
C  Description:
C    Prints a standard message, describing the reason for termination
C    of a minimization method.
C
C  Input arguments:
C    TMMESS     The message to be displayed. It must be constructed
C               by the calling routine.
C    NOCEX      Number of function evaluations performed by the calling
C               routine.
C    NOGEX      Number of gradient evaluations (calls to the user
C               supplied routine GRANAL) performed by the calling
C               routine.
C    NOJAC      Number of Jacobian evaluations (calls to the user
C               supplied routine JANAL) performed by the calling
C               routine.
C               Not applicable if NOJAC < 0.
C    ITER       Number of iterations performed by the calling routine.
C               Not applicable if ITER < 0.
C
C  ---------------------------------------------------------------------
C  Arguments:
      CHARACTER*(*) TMMESS
C  ---------------------------------------------------------------------
C  Common blocks:
      LOGICAL STRO, WEAK
      COMMON /PRILOG/ STRO, WEAK
      COMMON /UNITS/ IUINP, IUOUT
C  ---------------------------------------------------------------------
C  Local variables:
      CHARACTER*80 MESS
      CHARACTER*10 NAME, CF, CG, CI, CJ
C  ---------------------------------------------------------------------
C
      IF (.NOT.STRO) RETURN
C
      MESS = TMMESS
      LE1 = LENGTH(MESS)
      CALL CMDARG(0,NAME,LE2)
      CALL I2STR(NOCEX,CF,LEF)
      CALL CPRINT(NAME(1:LE2)//': '//MESS(1:LE1))
      CALL DASH
      WRITE (IUOUT,10) 'Function evaluations:', CF(1:LEF)
      IF (NOGEX.GE.0) THEN
         CALL I2STR(NOGEX,CG,LEG)
         WRITE (IUOUT,10) 'Gradient evaluations:', CG(1:LEG)
      END IF
      IF (NOJAC.GE.0) THEN
         IF (NOJAC.GE.0) CALL I2STR(NOJAC,CJ,LEJ)
         WRITE (IUOUT,10) 'Jacobian evaluations:', CJ(1:LEJ)
      END IF
      IF (ITER.GE.0) THEN
         CALL I2STR(ITER,CI,LEI)
         WRITE (IUOUT,10) 'Iterations:          ', CI(1:LEI)
      END IF
      CALL DASH
10    FORMAT ($(FC)A,3X,A)
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE ADDCON (N,M,A,IA,IACT,NACT,Z,U,RELACC,INDXBD,ZTC,
     1  CGRAD)
      IMPLICIT $(TYPE) (A-H,O-Z)
      DIMENSION A(IA,*),IACT(*),Z(*),U(*),ZTC(*),CGRAD(*)
      NP=NACT+1
      ICON=IACT(INDXBD)
      IACT(INDXBD)=IACT(NP)
      IACT(NP)=ICON
C
C     Form ZTC when the new constraint is a bound.
C
      IF (ICON .GT. M) THEN
          INEWBD=ICON-M
          IF (INEWBD .LE. N) THEN
              TEMP=-1.0
          ELSE
              INEWBD=INEWBD-N
              TEMP=1.0
          END IF
          IZNBD=INEWBD*N-N
          DO 10 J=1,N
   10     ZTC(J)=TEMP*Z(IZNBD+J)
C
C     Else form ZTC for an ordinary constraint.
C
      ELSE
          DO 20 I=1,N
   20     CGRAD(I)=A(I,ICON)
          DO 30 J=1,N
          ZTC(J)=0.0
          IZ=J
          DO 30 I=1,N
          ZTC(J)=ZTC(J)+Z(IZ)*CGRAD(I)
   30     IZ=IZ+N
      END IF
C
C     Find any Givens rotations to apply to the last columns of Z.
C
      J=N
   40 JP=J
      J=J-1
      IF (J .GT. NACT) THEN
          IF (ZTC(JP) .EQ. 0.0) GOTO 40
          IF (ABS(ZTC(JP)) .LE. RELACC*ABS(ZTC(J))) THEN
              TEMP=ABS(ZTC(J))
          ELSE IF (ABS(ZTC(J)) .LE. RELACC*ABS(ZTC(JP))) THEN
              TEMP=ABS(ZTC(JP))
          ELSE
              TEMP=ABS(ZTC(JP))*SQRT(1.0+(ZTC(J)/ZTC(JP))**2)
          END IF
          WCOS=ZTC(J)/TEMP
          WSIN=ZTC(JP)/TEMP
          ZTC(J)=TEMP
C
C     Apply the rotation when the new constraint is a bound.
C
          IZ=J
          IF (ICON .GT. M) THEN
              DO 50 I=1,N
              TEMP=WCOS*Z(IZ+1)-WSIN*Z(IZ)
              Z(IZ)=WCOS*Z(IZ)+WSIN*Z(IZ+1)
              Z(IZ+1)=TEMP
   50         IZ=IZ+N
              Z(IZNBD+JP)=0.0
C
C     Else apply the rotation for an ordinary constraint.
C
          ELSE
              WPIV=0.0
              DO 60 I=1,N
              TEMPA=WCOS*Z(IZ+1)
              TEMPB=WSIN*Z(IZ)
              TEMP=ABS(CGRAD(I))*(ABS(TEMPA)+ABS(TEMPB))
              IF (TEMP .GT. WPIV) THEN
                  WPIV=TEMP
                  IPIV=I
              END IF
              Z(IZ)=WCOS*Z(IZ)+WSIN*Z(IZ+1)
              Z(IZ+1)=TEMPA-TEMPB
   60         IZ=IZ+N
C
C     Ensure orthogonality of Z(.,JP) to CGRAD.
C
              SUM=0.0
              IZ=JP
              DO 70 I=1,N
              SUM=SUM+Z(IZ)*CGRAD(I)
   70         IZ=IZ+N
              IF (SUM .NE. 0.0) THEN
                  IZ=IPIV*N-N+JP
                  Z(IZ)=Z(IZ)-SUM/CGRAD(IPIV)
              END IF
          END IF
          GO TO 40
      END IF
C
C     Test for linear independence in the proposed new active set.
C
      IF (ZTC(NP) .EQ. 0.0) GOTO 90
      IF (ICON .LE. M) THEN
          SUM=0.0
          SUMABS=0.0
          IZ=NP
          DO 80 I=1,N
          TEMP=Z(IZ)*CGRAD(I)
          SUM=SUM+TEMP
          SUMABS=SUMABS+ABS(TEMP)
   80     IZ=IZ+N
          IF (ABS(SUM) .LE. RELACC*SUMABS) GOTO 90
      END IF
C
C     Set the new diagonal element of U and return.
C
      U(NP)=1.0/ZTC(NP)
      NACT=NP
   90 RETURN
      END
C
C  ---------------------------------------------------------------------
C
      SUBROUTINE ADJTOL (N,M,A,IA,B,XL,XU,X,IACT,NACT,XBIG,RELACC,TOL,
     1  MEQL)
      IMPLICIT $(TYPE) (A-H,O-Z)
      DIMENSION A(IA,*),B(*),XL(*),XU(*),X(*),IACT(*),XBIG(*)
C
C     Set VIOL to the greatest relative constraint residual of the first
C       NACT constraints.
C
      VIOL=0.0
      IF (NACT .GT. MEQL) THEN
          KL=MEQL+1
          DO 20 K=KL,NACT
          J=IACT(K)
          IF (J .LE. M) THEN
              RES=B(J)
              RESABS=ABS(B(J))
              DO 10 I=1,N
              RES=RES-A(I,J)*X(I)
   10         RESABS=RESABS+ABS(A(I,J)*XBIG(I))
          ELSE
              JM=J-M
              IF (JM .LE. N) THEN
                  RES=X(JM)-XL(JM)
                  RESABS=XBIG(JM)+ABS(XL(JM))
              ELSE
                  JM=JM-N
                  RES=XU(JM)-X(JM)
                  RESABS=XBIG(JM)+ABS(XU(JM))
              END IF
          END IF
          IF (RES .GT. 0.0) VIOL=MAX(VIOL,RES/RESABS)
   20     CONTINUE
      END IF
C
C     Adjust TOL.
C
      TOL=0.1*MIN(TOL,VIOL)
      IF (TOL .LE. RELACC+RELACC) THEN
          TOL=RELACC
          DO 30 I=1,N
   30     XBIG(I)=ABS(X(I))
      END IF
      RETURN
      END
C
C  ---------------------------------------------------------------------
C
      SUBROUTINE CONRES (N,M,A,IA,B,XL,XU,X,IACT,NACT,PAR,G,Z,U,XBIG,
     1  BRES,D,ZTG,RELACC,TOL,STEPCB,SUMRES,MEQL,MSAT,MTOT,INDXBD,
     2  GM,GMNEW,PARNEW,CGRAD)
      IMPLICIT $(TYPE) (A-H,O-Z)
      DIMENSION A(IA,*),B(*),XL(*),XU(*),X(*),IACT(*),PAR(*),G(*),
     1  Z(*),U(*),XBIG(*),BRES(*),D(*),ZTG(*),GM(*),GMNEW(*),PARNEW(*),
     2  CGRAD(*)
      IDIFF=MTOT-MSAT
C
C     Calculate and partition the residuals of the inactive constraints,
C       and set the gradient vector when seeking feasibility.
C
      IF (IDIFF .GT. 0.0) THEN
          DO 10 I=1,N
   10     G(I)=0.0
          SUMRES=0.0
      END IF
      MSATK=MSAT
      MDEG=NACT
      MSAT=NACT
      KL=MEQL+1
      DO 50 K=KL,MTOT
      J=IACT(K)
C
C     Calculate the residual of the current constraint.
C
      IF (J .LE. M) THEN
          RES=B(J)
          RESABS=ABS(B(J))
          DO 20 I=1,N
          RES=RES-X(I)*A(I,J)
   20     RESABS=RESABS+ABS(XBIG(I)*A(I,J))
      ELSE
          JM=J-M
          IF (JM .LE. N) THEN
              RES=X(JM)-XL(JM)
              RESABS=ABS(XBIG(JM))+ABS(XL(JM))
          ELSE
              JM=JM-N
              RES=XU(JM)-X(JM)
              RESABS=ABS(XBIG(JM))+ABS(XU(JM))
          END IF
      END IF
      BRES(J)=RES
C
C     Set TEMP to the relative residual.
C
      TEMP=0.0
      IF (RESABS .NE. 0.0) TEMP=RES/RESABS
      IF (K .GT. MSATK .AND. TEMP .LT. 0.0) THEN
          IF (TEMP+RELACC .GE. 0.0) THEN
              IF (J .LE. M) THEN
                  SUM=ABS(B(J))
                  DO 30 I=1,N
   30             SUM=SUM+ABS(X(I)*A(I,J))
              ELSE
                  JM=J-M
                  IF (JM .LE. N) THEN
                      SUM=ABS(X(JM))+ABS(XL(JM))
                  ELSE
                      SUM=ABS(X(JM-N))+ABS(XU(JM-N))
                  END IF
              END IF
              IF (ABS(RES) .LE. SUM*RELACC) TEMP=0.0
          END IF
      END IF
C
C     Place the residual in the appropriate position.
C
      IF (K .LE. NACT) GOTO 50
      IF (K .LE. MSATK .OR. TEMP .GE. 0.0) THEN
          MSAT=MSAT+1
          IF (MSAT .LT. K) THEN
              IACT(K)=IACT(MSAT)
          END IF
          IF (TEMP .GT. TOL) THEN
              IACT(MSAT)=J
          ELSE
              MDEG=MDEG+1
              IACT(MSAT)=IACT(MDEG)
              IACT(MDEG)=J
          END IF
C
C     Update the gradient and SUMRES if the constraint is violated when
C       seeking feasibility.
C
      ELSE
          IF (J .LE. M) THEN
              DO 40 I=1,N
   40         G(I)=G(I)+A(I,J)
          ELSE
              J=J-M
              IF (J .LE. N) THEN
                  G(J)=G(J)-1.0
              ELSE
                  G(J-N)=G(J-N)+1.0
              END IF
          END IF
          SUMRES=SUMRES+ABS(RES)
      END IF
   50 CONTINUE
C
C   Seek the next search direction unless CONRES was called from GETFES
C   and feasibility has been achieved.
C
      STEPCB=0.0
      IF (IDIFF .GT. 0 .AND. MSAT .EQ. MTOT) GOTO 60
      CALL GETD (N,M,A,IA,IACT,NACT,PAR,G,Z,U,D,ZTG,RELACC,DDOTG,MEQL,
     1  MDEG,GM,GMNEW,PARNEW,CGRAD)
C
C     Calculate the (bound on the) step-length due to the constraints.
C
      IF (DDOTG .LT. 0.0) THEN
          CALL STEPBD (N,M,A,IA,IACT,BRES,D,STEPCB,DDOTG,MDEG,MSAT,
     1      MTOT,INDXBD)
      END IF
      IF (IDIFF .EQ. 0) SUMRES=DDOTG
   60 RETURN
      END
C
C  ---------------------------------------------------------------------
C
      SUBROUTINE DELCON (N,M,A,IA,IACT,NACT,Z,U,RELACC,IDROP)
      IMPLICIT $(TYPE) (A-H,O-Z)
      DIMENSION A(IA,*),IACT(*),Z(*),U(*)
      NM=NACT-1
      IF (IDROP .EQ. NACT) GOTO 60
      ISAVE=IACT(IDROP)
C
C     Cycle through the constraint exchanges that are needed.
C
      DO 50 J=IDROP,NM
      JP=J+1
      ICON=IACT(JP)
      IACT(J)=ICON
C
C     Calculate the (J,JP) element of R.
C
      IF (ICON .LE. M) THEN
          RJJP=0.0
          IZ=J
          DO 10 I=1,N
          RJJP=RJJP+Z(IZ)*A(I,ICON)
   10     IZ=IZ+N
      ELSE
          IBD=ICON-M
          IF (IBD .LE. N) THEN
              IZBD=IBD*N-N
              RJJP=-Z(IZBD+J)
          ELSE
              IBD=IBD-N
              IZBD=IBD*N-N
              RJJP=Z(IZBD+J)
          END IF
      END IF
C
C     Calculate the parameters of the next rotation.
C
      UJP=U(JP)
      TEMP=RJJP*UJP
      DENOM=ABS(TEMP)
      IF (DENOM*RELACC .LT. 1.0) DENOM=SQRT(1.0+DENOM*DENOM)
      WCOS=TEMP/DENOM
      WSIN=1.0/DENOM
C
C     Rotate Z when a bound constraint is promoted.
C
      IZ=J
      IF (ICON .GT. M) THEN
          DO 20 I=1,N
          TEMP=WCOS*Z(IZ+1)-WSIN*Z(IZ)
          Z(IZ)=WCOS*Z(IZ)+WSIN*Z(IZ+1)
          Z(IZ+1)=TEMP
   20     IZ=IZ+N
          Z(IZBD+JP)=0.0
C
C     Rotate Z when an ordinary constraint is promoted.
C
      ELSE
          WPIV=0.0
          DO 30 I=1,N
          TEMPA=WCOS*Z(IZ+1)
          TEMPB=WSIN*Z(IZ)
          TEMP=ABS(A(I,ICON))*(ABS(TEMPA)+ABS(TEMPB))
          IF (TEMP .GT. WPIV) THEN
              WPIV=TEMP
              IPIV=I
          END IF
          Z(IZ)=WCOS*Z(IZ)+WSIN*Z(IZ+1)
          Z(IZ+1)=TEMPA-TEMPB
   30     IZ=IZ+N
C
C     Ensure orthogonality to promoted constraint.
C
          SUM=0.0
          IZ=JP
          DO 40 I=1,N
          SUM=SUM+Z(IZ)*A(I,ICON)
   40     IZ=IZ+N
          IF (SUM .NE. 0.0) THEN
              IZ=IPIV*N-N+JP
              Z(IZ)=Z(IZ)-SUM/A(IPIV,ICON)
          END IF
      END IF
C
C     Set the new diagonal elements of U.
C
      U(JP)=-DENOM*U(J)
      U(J)=UJP/DENOM
   50 CONTINUE
C
C     Return.
C
      IACT(NACT)=ISAVE
   60 NACT=NM
      RETURN
      END
C
C  ---------------------------------------------------------------------
C
      SUBROUTINE EQCONS (N,M,MEQ,A,IA,B,XU,IACT,MEQL,INFO,Z,U,RELACC,
     1  AM,CGRAD)
      IMPLICIT $(TYPE) (A-H,O-Z)
      DIMENSION A(IA,*),B(*),XU(*),IACT(*),Z(*),U(*),AM(*),CGRAD(*)
C
C     Try to add the next equality constraint to the active set.
C
      DO 50 KEQ=1,MEQ
      IF (MEQL .LT. N) THEN
          NP=MEQL+1
          IACT(NP)=KEQ
          CALL ADDCON (N,M,A,IA,IACT,MEQL,Z,U,RELACC,NP,AM,CGRAD)
          IF (MEQL .EQ. NP) GOTO 50
      END IF
C
C     If linear dependence occurs then find the multipliers of the
C       dependence relation and apply them to the right hand sides.
C
      SUM=B(KEQ)
      SUMABS=ABS(B(KEQ))
      IF (MEQL .GT. 0) THEN
          DO 10 I=1,N
   10     AM(I)=A(I,KEQ)
          K=MEQL
   20     VMULT=0.0
          IZ=K
          DO 30 I=1,N
          VMULT=VMULT+Z(IZ)*AM(I)
   30     IZ=IZ+N
          VMULT=VMULT*U(K)
          J=IACT(K)
          IF (J .LE. M) THEN
              DO 40 I=1,N
   40         AM(I)=AM(I)-VMULT*A(I,J)
              RHS=B(J)
          ELSE
              JM=J-M-N
              AM(JM)=AM(JM)-VMULT
              RHS=XU(JM)
          END IF
          SUM=SUM-RHS*VMULT
          SUMABS=SUMABS+ABS(RHS*VMULT)
          K=K-1
          IF (K .GE. 1) GOTO 20
      END IF
C
C     Error return if the constraints are inconsistent.
C
      IF (ABS(SUM) .GT. RELACC*SUMABS) THEN
          INFO=5
          GOTO 60
      END IF
   50 CONTINUE
   60 RETURN
      END
C
C  ---------------------------------------------------------------------
C
      SUBROUTINE GETD (N,M,A,IA,IACT,NACT,PAR,G,Z,U,D,ZTG,RELACC,
     1  DDOTG,MEQL,MDEG,GM,GMNEW,PARNEW,CGRAD)
      IMPLICIT $(TYPE) (A-H,O-Z)
      DIMENSION A(IA,*),IACT(*),PAR(*),G(*),Z(*),U(*),D(*),ZTG(*),
     1  GM(*),GMNEW(*),PARNEW(*),CGRAD(*)
C
C     Initialize GM and cycle backwards through the active set.
C
CC dgp
      ABCD = 1.
10    CONTINUE
      DO 20 II=1,N
 20   GM(II)=G(II)
      K=NACT
 30   CONTINUE
      IF (K .GT. 0) THEN
C
C     Set TEMP to the next multiplier, but reduce the active set if
C       TEMP has an unacceptable sign.
C
          TEMP=0.0
          IZ=K
          DO 40 I=1,N
          TEMP=TEMP+Z(IZ)*GM(I)
   40     IZ=IZ+N
          TEMP=TEMP*U(K)
          IF (K .GT. MEQL .AND. TEMP .GT. 0.0) THEN
              CALL DELCON (N,M,A,IA,IACT,NACT,Z,U,RELACC,K)
              GOTO 10
          END IF
C
C     Update GM using the multiplier that has just been calculated.
C
          J=IACT(K)
          IF (J .LE. M) THEN
              DO 50 I=1,N
   50         GM(I)=GM(I)-TEMP*A(I,J)
          ELSE
              JM=J-M
              IF (JM .LE. N) THEN
                  GM(JM)=GM(JM)+TEMP
              ELSE
                  GM(JM-N)=GM(JM-N)-TEMP
              END IF
          END IF
          PAR(K)=TEMP
          K=K-1
          GOTO 30
      END IF
C
C     Calculate the search direction and DDOTG.
C
      DDOTG=0.0
      IF (NACT .LT. N) THEN
          CALL SDEGEN (N,M,A,IA,IACT,NACT,PAR,Z,U,D,ZTG,GM,RELACC,
     1      DDOTGM,MEQL,MDEG,GMNEW,PARNEW,CGRAD)
          IF (DDOTGM .LT. 0.0) THEN
              DO 60 I=1,N
   60         DDOTG=DDOTG+D(I)*G(I)
          END IF
      END IF
      RETURN
      END
C
C  ---------------------------------------------------------------------
C
      SUBROUTINE GETFES (N,M,A,IA,B,XL,XU,X,IACT,NACT,PAR,INFO,G,Z,
     1  U,XBIG,RELACC,TOL,MEQL,MSAT,MTOT,BRES,D,ZTG,GM,GMNEW,PARNEW,
     2  CGRAD)
      IMPLICIT $(TYPE) (A-H,O-Z)
      DIMENSION A(IA,*),B(*),XL(*),XU(*),X(*),IACT(*),PAR(*),G(*),Z(*),
     1  U(*),XBIG(*),BRES(*),D(*),ZTG(*),GM(*),GMNEW(*),PARNEW(*),
     2  CGRAD(*)
C
C     Make the correction to X for the active constraints.
C
      INFO=0
   10 CALL SATACT (N,M,A,IA,B,XL,XU,X,IACT,NACT,INFO,Z,U,XBIG,RELACC,
     1  TOL,MEQL)
      IF (INFO .GT. 0) MSAT=NACT
      IF (MSAT .EQ. MTOT) GOTO 60
C
C     Try to correct the infeasibility.
C
   20 MSATK=MSAT
      SUMRSK=0.0
   30 CALL CONRES (N,M,A,IA,B,XL,XU,X,IACT,NACT,PAR,G,Z,U,XBIG,BRES,
     1  D,ZTG,RELACC,TOL,STEPCB,SUMRES,MEQL,MSAT,MTOT,INDXBD,GM,GMNEW,
     2  PARNEW,CGRAD)
C
C     Include the new constraint in the active set.
C
      IF (STEPCB .GT. 0.0) THEN
          DO 40 I=1,N
          X(I)=X(I)+STEPCB*D(I)
   40     XBIG(I)=MAX(XBIG(I),ABS(X(I)))
          CALL ADDCON (N,M,A,IA,IACT,NACT,Z,U,RELACC,INDXBD,GMNEW,CGRAD)
      END IF
C
C     Test whether to continue the search for feasibility.
C
      IF (MSAT .LT. MTOT) THEN
          IF (STEPCB .EQ. 0.0) GOTO 50
          IF (MSATK .LT. MSAT) GOTO 20
          IF (SUMRSK .EQ. 0.0 .OR. SUMRES .LT. SUMRSK) THEN
              SUMRSK=SUMRES
              ITEST=0
          END IF
          ITEST=ITEST+1
          IF (ITEST .LE. 2) GOTO 30
C
C     Reduce TOL if it may be too large to allow feasibility.
C
   50     IF (TOL .GT. RELACC) THEN
              CALL ADJTOL (N,M,A,IA,B,XL,XU,X,IACT,NACT,XBIG,RELACC,
     1          TOL,MEQL)
              GOTO 10
          END IF
      END IF
   60 RETURN
      END
C
C  ---------------------------------------------------------------------
C
      SUBROUTINE GETMIN (N,M,MEQ,A,IA,B,XL,XU,X,ACC,IACT,NACT,PAR,
     1  IPRINT,INFO,W)
      IMPLICIT $(TYPE) (A-H,O-Z)
      DIMENSION A(IA,*),B(*),XL(*),XU(*),X(*),IACT(*),PAR(*),W(*)
C
C  This is the entry point to a package of subroutines that calculate
C     the least value of a differentiable function of several variables
C     subject to linear constraints on the values of the variables.
C  N is the number of variables and must be set by the user.
C  M is the number of linear constraints (excluding simple bounds) and
C     must be set by the user.
C  MEQ is the number of constraints that are equalities and must be set
C     by the user.
C  A(.,.) is a 2-dimensional array whose columns are the gradients of
C     the M constraint functions. Its entries must be set by the user
C     and its dimensions must be at least N by M.
C  IA is the actual first dimension of the array A that is supplied by
C     the user, so its value may not be less than N.
C  B(.) is a vector of constraint right hand sides that must also be set
C     by the user.  Specifically the constraints on the variables X(I)
C     I=1(1)N are
C          A(1,K)*X(1)+...+A(N,K)*X(N) .EQ. B(K)  K=1,...,MEQ
C          A(1,K)*X(1)+...+A(N,K)*X(N) .LE. B(K)  K=MEQ+1,...,M  .
C     Note that the data that define the equality constraints come
C     before the data of the inequalities.
C  XL(.) and XU(.) are vectors whose components must be set to lower and
C     upper bounds on the variables.  Choose very large negative and
C     positive entries if a component should be unconstrained, or set
C     XL(I)=XU(I) to freeze the I-th variable.  Specifically these
C     simple bounds are
C          XL(I) .LE. X(I) and X(I) .LE. XU(I)  I=1,...,N  .
C  X(.) is the vector of variables of the optimization calculation. Its
C     initial elements must be set by the user to an estimate of the
C     required solution.  The subroutines can usually cope with poor
C     estimates, and there is no need for X(.) to be feasible initially.
C     These variables are adjusted automatically and the values that
C     give the least feasible calculated value of the objective function
C     are available in X(.) on the return from GETMIN.
C  ACC is a tolerance on the first order conditions at the calculated
C     solution of the optimization problem.  These first order
C     conditions state that, if X(.) is a solution, then there is a set
C     of active constraints with indices IACT(K) K=1(1)NACT, say,
C     such that X(.) is on the boundaries of these constraints,
C     and the gradient of the objective function can be expressed
C     in the form
C           GRAD(F)=PAR(1)*GRAD(C(IACT(1)))+...
C                        ...+PAR(NACT)*GRAD(C(IACT(NACT)))  .
C     Here PAR(K) K=1(1)NACT are Lagrange multipliers that are
C     nonpositive for inequality constraints, and GRAD(C(IACT(K)))
C     is the gradient of the IACT(K)-th constraint function, so it is
C     A(.,IACT(K)) if IACT(K) .LE. M, and it is minus or plus the J-th
C     coordinate vector if the constraint is the lower or upper bound on
C     X(J) respectively. The normal return from the calculation occurs
C     when X(.) is feasible and the sum of squares of components of the
C     vector RESKT(.) is at most ACC**2, where RESKT(.) is the
C     N-component vector of residuals of the first order condition that
C     is displayed above.
C     Sometimes the package cannot satisfy this condition, because noise
C     in the function values can prevent a change to the variables,
C     no line search being allowed to increase the objective function.
C  IACT(.) is a working space array of integer variables that must be
C     provided by the user.  Its length must be at least (M+2*N).  Its
C     leading entries on the return from the subroutine are the indices
C     IACT(K) K=1(1)NACT that are mentioned in the previous paragraph:
C     in other words they are the indices of the final active
C     constraints.
C     Here the indices M+1,...,M+N and M+N+1,...,M+2*N denote the lower
C     and upper bounds respectively.
C  NACT is set automatically to the integer variable of this ilk that
C     has been introduced already.
C  PAR is a one-dimensional array that will hold the Lagrange
C     multipliers PAR(K) K=1(1)NACT on the return from GETMIN, these
C     parameters being defined in the above paragraph on ACC.
C     The length of PAR should be at least N.
C  IPRINT must be set by the user to specify the frequency of printing
C     during the execution of the optimization package.  There is no
C     printed output if IPRINT=0.  Otherwise, after ensuring
C     feasibility, information is given every IABS(IPRINT) iterations
C     and whenever a parameter called TOL is reduced. The printing
C     provides the values of X(.), F(.) and G(.)=GRAD(F) if IPRINT
C     is positive, while if IPRINT is negative this information is
C     augmented by the current values of IACT(K) K=1(1)NACT,
C     PAR(K) K=1(1)NACT and RESKT(I) I=1(1)N. The reason for returning
C     to the calling program is also displayed when IPRINT is nonzero.
C  INFO is an integer variable that should be set to zero initially,
C     unless the user wishes to impose an upper bound on the number of
C     evaluations of the objective function and its gradient, in which
C     case INFO should be set to the value of this bound.  On the exit
C     from GETMIN it will have one of the following integer values to
C     indicate the reason for leaving the optimization package:
C          INFO=1   X(.) is feasible and the condition that depends on
C     ACC is satisfied.
C          INFO=2   X(.) is feasible and rounding errors are preventing
C     further progress.
C          INFO=3   X(.) is feasible but the objective function fails to
C     decrease although a decrease is predicted by the current gradient
C     vector.  If this return occurs and KTRES(.) has large components
C     then the user's calculation of the gradient of the objective
C     function may be incorrect.  One should also question the coding of
C     the gradient when the final rate of convergence is slow.
C          INFO=4   In this case the calculation cannot begin because IA
C     is less than N or because the lower bound on a variable is greater
C     than the upper bound.
C          INFO=5   This value indicates that the equality constraints
C     are inconsistent.   These constraints include any components of
C     X(.) that are frozen by setting XL(I)=XU(I).
C          INFO=6   In this case there is an error return because the
C     equality constraints and the bounds on the variables are found to
C     be inconsistent.
C          INFO=7   This value indicates that there is no vector of
C     variables that satisfies all of the constraints.  Specifically,
C     when this return or an INFO=6 return occurs, the current active
C     constraints (whose indices are IACT(K) K=1(1)NACT) prevent any
C     change in X(.) that reduces the sum of constraint violations,
C     where only bounds are included in this sum if INFO=6.
C          INFO=8   In this case the limit on the number of calls of
C     subroutine FGCALC (see below) has been reached, and there would
C     have been further calculation otherwise.
C  W(.) is a working space array of real variables that must be provided
C     by the user.  Its length must be at least (M+11*N+N**2).  On exit
C     from the package one can find the final components of GRAD(F) and
C     RESKT(.) in W(1),...,W(N) and W(N+1),...,W(2*N) respectively.
C  Note 1.   The variables N, M, MEQ, IA, ACC and IPRINT and the
C     elements of the arrays A(,.,), B(.), XL(.) and XU(.) are not
C     altered by the optimization procedure. Their values, the value of
C     INFO and the initial components of X(.) must be set on entry
C     to GETMIN.
C  Note 2.   Of course the package needs the objective function and its
C     gradient.  Therefore the user must provide a subroutine called
C     FGCALC, its first two lines being
C          SUBROUTINE FGCALC (N,X,F,G)
C          DIMENSION X(*),G(*)   .
C     It is called automatically with N set as above and with X(.) set
C     to a feasible vector of variables.  It should calculate the value
C     of the objective function and its gradient for this X(.) and
C     should set them in F and G(I) I=1(1)N respectively, without
C     disturbing N or any of the components of X(.).
C  Note 3.   A paper on the method of calculation and a report on the
C     main features of the computer code are available from the author
C     M.J.D.Powell (D.A.M.T.P., University of Cambridge, Silver Street,
C     Cambridge CB3 9EW, England).
C
C     Partition the workspace array.
C
      IG=1
      IRESKT=IG+N
      IZ=IRESKT+N
      IU=IZ+N*N
      IXBIG=IU+N
      IBRES=IXBIG+N
      ID=IBRES+M+N+N
      IZTG=ID+N
      IGM=IZTG+N
      IXS=IGM+N
      IGS=IXS+N
C
C     Call the optimization package.
C
      CALL MINFLC (N,M,MEQ,A,IA,B,XL,XU,X,ACC,IACT,NACT,PAR,IPRINT,
     1  INFO,W(IG),W(IZ),W(IU),W(IXBIG),W(IRESKT),W(IBRES),W(ID),
     2  W(IZTG),W(IGM),W(IXS),W(IGS))
      RETURN
      END
C
C  ---------------------------------------------------------------------
C
      SUBROUTINE INITZU (N,M,XL,XU,X,IACT,MEQL,INFO,Z,U,XBIG,RELACC)
      IMPLICIT $(TYPE) (A-H,O-Z)
      DIMENSION XL(*),XU(*),X(*),IACT(*),Z(*),U(*),XBIG(*)
C
C     Set RELACC.
C
      ZTPAR=100.0
      RELACC=1.0
   10 RELACC=0.5*RELACC
      TEMPA=ZTPAR+0.5*RELACC
      TEMPB=ZTPAR+RELACC
      IF (ZTPAR .LT. TEMPA .AND. TEMPA .LT. TEMPB) GOTO 10
C
C     Seek bound inconsistencies and bound equality constraints.
C
      MEQL=0
      DO 20 I=1,N
      IF (XL(I) .GT. XU(I)) GOTO 50
      IF (XL(I) .EQ. XU(I)) MEQL=MEQL+1
   20 CONTINUE
C
C     Initialize U, Z and XBIG.
C
      JACT=0
      NN=N*N
      DO 30 I=1,NN
   30 Z(I)=0.0
      IZ=0
      DO 40 I=1,N
      IF (XL(I) .EQ. XU(I)) THEN
          X(I)=XU(I)
          JACT=JACT+1
          U(JACT)=1.0
          IACT(JACT)=I+M+N
          J=JACT
      ELSE
          J=I+MEQL-JACT
      END IF
      Z(IZ+J)=1.0
      IZ=IZ+N
   40 XBIG(I)=ABS(X(I))
      INFO=1
   50 RETURN
      END
C
C  ---------------------------------------------------------------------
C
      SUBROUTINE KTVEC (N,M,A,IA,IACT,NACT,PAR,G,RESKT,Z,U,BRES,RELAXF,
     1  MEQL,SSQKT,PARW,RESKTW)
      IMPLICIT $(TYPE) (A-H,O-Z)
      DIMENSION A(IA,*),IACT(*),PAR(*),G(*),RESKT(*),Z(*),U(*),
     1  BRES(*),PARW(*),RESKTW(*)
C
C     Calculate the Lagrange parameters and the residual vector.
C
      DO 10 I=1,N
   10 RESKT(I)=G(I)
      IF (NACT .GT. 0) THEN
          ICASE=0
   20     DO 50 KK=1,NACT
          K=NACT+1-KK
          J=IACT(K)
          TEMP=0.0
          IZ=K
          DO 30 I=1,N
          TEMP=TEMP+Z(IZ)*RESKT(I)
   30     IZ=IZ+N
          TEMP=TEMP*U(K)
          IF (ICASE .EQ. 0) PAR(K)=0.0
          IF (K .LE. MEQL .OR. PAR(K)+TEMP .LT. 0.0) THEN
              PAR(K)=PAR(K)+TEMP
          ELSE
              TEMP=-PAR(K)
              PAR(K)=0.0
          END IF
          IF (TEMP .NE. 0.0) THEN
              IF (J .LE. M) THEN
                  DO 40 I=1,N
   40             RESKT(I)=RESKT(I)-TEMP*A(I,J)
              ELSE
                  JM=J-M
                  IF (JM .LE. N) THEN
                      RESKT(JM)=RESKT(JM)+TEMP
                  ELSE
                      RESKT(JM-N)=RESKT(JM-N)-TEMP
                  END IF
              END IF
          END IF
   50     CONTINUE
C
C     Calculate the sum of squares of the KT residual vector.
C
          SSQKT=0.0
          IF (NACT .EQ. N) GOTO 130
          DO 60 I=1,N
   60     SSQKT=SSQKT+RESKT(I)**2
C
C     Apply iterative refinement to the residual vector.
C
          IF (ICASE .EQ. 0) THEN
              ICASE=1
              DO 70 K=1,NACT
   70         PARW(K)=PAR(K)
              DO 80 I=1,N
   80         RESKTW(I)=RESKT(I)
              SSQKTW=SSQKT
              GOTO 20
          END IF
C
C     Undo the iterative refinement if it does not reduce SSQKT.
C
          IF (SSQKTW .LT. SSQKT) THEN
              DO 90 K=1,NACT
   90         PAR(K)=PARW(K)
              DO 100 I=1,N
  100         RESKT(I)=RESKTW(I)
              SSQKT=SSQKTW
          END IF
C
C     Calculate SSQKT when there are no active constraints.
C
      ELSE
          SSQKT=0.0
          DO 110 I=1,N
  110     SSQKT=SSQKT+G(I)**2
      END IF
C
C     Predict the reduction in F if one corrects any positive residuals
C       of active inequality constraints.
C
      RELAXF=0.0
      IF (MEQL .LT. NACT) THEN
          KL=MEQL+1
          DO 120 K=KL,NACT
          J=IACT(K)
          IF (BRES(J) .GT. 0.0) THEN
              RELAXF=RELAXF-PAR(K)*BRES(J)
          END IF
  120     CONTINUE
      END IF
  130 RETURN
      END
C
C  ---------------------------------------------------------------------
C
      SUBROUTINE LSRCH (N,X,G,D,XS,GS,RELACC,STEPCB,DDOTG,F,STEP,
     1  NFVALS,NFMAX,GOPT)
      IMPLICIT $(TYPE) (A-H,O-Z)
      DIMENSION X(*),G(*),D(*),XS(*),GS(*),GOPT(*)
      COMMON /TOTCAL/ ITNOCS, IPARTC, ITDER, IPDER, ITHESS, IPHESS,
     &                ITJAC, IPJAC
C
C     Initialization.
C
      RELINT=0.9
      ICOUNT=0
      RATIO=-1.0
      DO 10 I=1,N
      XS(I)=X(I)
      GS(I)=G(I)
      GOPT(I)=G(I)
      IF (D(I) .NE. 0.0) THEN
          TEMP=ABS(X(I)/D(I))
          IF (RATIO .LT. 0.0 .OR. TEMP .LT. RATIO) RATIO=TEMP
      END IF
   10 CONTINUE
      AMX = 1.0
      STEP=MIN(AMX,STEPCB)
      STPMIN=MAX(RELACC*RATIO,1.0E-12*STEP)
      STEP=MAX(STPMIN,STEP)
      SBASE=0.0
      FBASE=F
      DDOTGB=DDOTG
      STPLOW=0.0
      FLOW=F
      DGLOW=DDOTG
      STPHGH=0.0
      STPOPT=0.0
      FOPT=F
      DGOPT=ABS(DDOTG)
C
C     Calculate another function and gradient value.
C
   20 DO 30 I=1,N
   30 X(I)=XS(I)+STEP*D(I)
      ISOFAR = ITNOCS
      CALL FGCALC (N,X,F,G)
      ICOUNT=ICOUNT+(ITNOCS-ISOFAR)
      DGMID=0.0
      DO 40 I=1,N
   40 DGMID=DGMID+D(I)*G(I)
      IF (F .LE. FOPT) THEN
          IF (F .LT. FOPT .OR. ABS(DGMID) .LT. DGOPT) THEN
              STPOPT=STEP
              FOPT=F
              DO 50 I=1,N
   50         GOPT(I)=G(I)
              DGOPT=ABS(DGMID)
          END IF
      END IF
      IF (NFVALS+ICOUNT .EQ. NFMAX) GOTO 70
C
C      Modify the bounds on the steplength or convergence.
C
      IF (F .GE. FBASE+0.1*(STEP-SBASE)*DDOTGB) THEN
          IF (STPHGH .GT. 0.0 .OR. F .GT. FBASE .OR. DGMID .GT.
     1      0.5*DDOTG) THEN
              STPHGH=STEP
              FHGH=F
              DGHGH=DGMID
              GOTO 60
          END IF
          SBASE=STEP
          FBASE=F
          DDOTGB=DGMID
      END IF
      IF (DGMID .GE. 0.7*DDOTGB) GOTO 70
      STPLOW=STEP
      FLOW=F
      DGLOW=DGMID
   60 IF (STPHGH .GT. 0.0 .AND. STPLOW .GE. RELINT*STPHGH) GOTO 70
C
C     Calculate the next step length or end the iterations.
C
      IF (STPHGH .EQ. 0.0) THEN
          IF (STEP .EQ. STEPCB) GOTO 70
          TEMP=10.0
          IF (DGMID .GT. 0.9*DDOTG) TEMP=DDOTG/(DDOTG-DGMID)
          STEP=MIN(TEMP*STEP,STEPCB)
          GOTO 20
      ELSE IF (ICOUNT .EQ. 1 .OR. STPLOW .GT. 0.0) THEN
          DGKNOT=2.0*(FHGH-FLOW)/(STPHGH-STPLOW)-0.5*(DGLOW+DGHGH)
          IF (DGKNOT .GE. 0.0) THEN
              AMX = 0.1
              RATIO=MAX(AMX,0.5*DGLOW/(DGLOW-DGKNOT))
          ELSE
              RATIO=(0.5*DGHGH-DGKNOT)/(DGHGH-DGKNOT)
          END IF
          STEP=STPLOW+RATIO*(STPHGH-STPLOW)
          GOTO 20
      ELSE
          STEP=0.1*STEP
          IF (STEP .GE. STPMIN) GOTO 20
      END IF
C
C     Return from subroutine.
C
   70 IF (STEP .NE. STPOPT) THEN
          STEP=STPOPT
          F=FOPT
          DO 80 I=1,N
          X(I)=XS(I)+STEP*D(I)
   80     G(I)=GOPT(I)
      END IF
      NFVALS=NFVALS+ICOUNT
      RETURN
      END
C
C  ---------------------------------------------------------------------
C
      SUBROUTINE MINFLC (N,M,MEQ,A,IA,B,XL,XU,X,ACC,IACT,NACT,PAR,
     1  IPRINT,INFO,G,Z,U,XBIG,RESKT,BRES,D,ZTG,GM,XS,GS)
      IMPLICIT $(TYPE) (A-H,O-Z)
      DIMENSION A(IA,*),B(*),XL(*),XU(*),X(*),IACT(*),PAR(*),G(*),
     1  Z(*),U(*),XBIG(*),RESKT(*),BRES(*),D(*),ZTG(*),GM(*),XS(*),
     2  GS(*)
      COMMON /UNITS/ IUINP, IUOUT
C
C     Initialize ZZNORM, ITERC, NFVALS and NFMAX.
C
      ZZNORM=-1.0
      ITERC=0
      NFVALS=0
      NFMAX=0
      IF (INFO .GT. 0) NFMAX=INFO
C
C     Check the bounds on N, M and MEQ.
C
      INFO=4
      IF (MAX0(1-N,MEQ*(MEQ-M)) .GT. 0) THEN
          IF (IPRINT .NE. 0) WRITE (IUOUT,1010)
 1010     FORMAT (/5X,'ERROR RETURN FROM GETMIN BECAUSE A CONDITION',
     1      ' ON N, M OR MEQ IS VIOLATED')
          GOTO 40
      END IF
C
C     Initialize RELACC, Z, U and TOL.
C
      CALL INITZU (N,M,XL,XU,X,IACT,MEQL,INFO,Z,U,XBIG,RELACC)
      AMX = 0.01
      TOL=MAX(AMX,10.0*RELACC)
      IF (INFO .EQ. 4) THEN
          IF (IPRINT .NE. 0) WRITE (IUOUT,1020)
 1020     FORMAT (/5X,'ERROR RETURN FROM GETMIN BECAUSE A LOWER',
     1      ' BOUND EXCEEDS AN UPPER BOUND')
          GOTO 40
      END IF
C
C     Add any equality constraints to the active set.
C
      IF (MEQ .GT. 0) THEN
          CALL EQCONS (N,M,MEQ,A,IA,B,XU,IACT,MEQL,INFO,Z,U,RELACC,XS,
     1      GS)
          IF (INFO .EQ. 5) THEN
              IF (IPRINT .NE. 0) WRITE (IUOUT,1030)
 1030         FORMAT (/5X,'ERROR RETURN FROM GETMIN BECAUSE THE',
     1          ' EQUALITY CONSTRAINTS ARE INCONSISTENT')
              GOTO 40
          END IF
      END IF
      NACT=MEQL
      MSAT=MEQL
C
C     Add the bounds to the list of constraints.
C
      MTOT=NACT
      DO 10 I=1,N
      IF (XL(I) .LT. XU(I)) THEN
          MTOT=MTOT+2
          IACT(MTOT-1)=M+I
          IACT(MTOT)=M+N+I
      END IF
   10 CONTINUE
C
C     Try to satisfy the bound constraints.
C
      CALL GETFES (N,M,A,IA,B,XL,XU,X,IACT,NACT,PAR,INFO,G,Z,U,XBIG,
     1  RELACC,TOL,MEQL,MSAT,MTOT,BRES,D,ZTG,GM,RESKT,XS,GS)
      IF (MSAT .LT. MTOT) THEN
          IF (IPRINT .NE. 0) WRITE (IUOUT,1040)
 1040     FORMAT (/5X,'ERROR RETURN FROM GETMIN BECAUSE THE',
     1      ' EQUALITIES AND BOUNDS ARE INCONSISTENT')
          INFO=6
          GOTO 40
      END IF
C
C     Add the ordinary inequalities to the list of constraints.
C
      IF (M .GT. MEQ) THEN
          MP=MEQ+1
          DO 20 K=MP,M
          MTOT=MTOT+1
   20     IACT(MTOT)=K
      END IF
C
C     Correct any constraint violations.
C
   30 CALL GETFES (N,M,A,IA,B,XL,XU,X,IACT,NACT,PAR,INFO,G,Z,U,XBIG,
     1  RELACC,TOL,MEQL,MSAT,MTOT,BRES,D,ZTG,GM,RESKT,XS,GS)
      IF (MSAT .LT. MTOT) THEN
          IF (IPRINT .NE. 0) WRITE (IUOUT,1050)
 1050     FORMAT (/5X,'ERROR RETURN FROM GETMIN BECAUSE THE',
     1      ' CONSTRAINTS ARE INCONSISTENT')
          INFO=7
          GOTO 40
      ELSE IF (MEQL .EQ. N) THEN
          IF (IPRINT .NE. 0) WRITE (IUOUT,1060)
 1060     FORMAT (/5X,'GETMIN FINDS THAT THE VARIABLES ARE',
     1      ' DETERMINED BY THE EQUALITY CONSTRAINTS')
          GOTO 40
      END IF
C
C     Minimize the objective function in the case when constraints are
C       treated as degenerate if their residuals are less than TOL.
C
      IEXAU = 0
      CALL MINFUN (N,M,A,IA,B,XL,XU,X,ACC,IACT,NACT,PAR,IPRINT,INFO,G,Z,
     1  U,XBIG,RELACC,ZZNORM,TOL,MEQL,MTOT,ITERC,NFVALS,NFMAX,RESKT,
     2  BRES,D,ZTG,GM,XS,GS,F,IEXAU)
      IF (IEXAU.NE.0) RETURN
C
C     Reduce TOL if necessary.
C
      IF (TOL .GT. RELACC .AND. NACT .GT. 0) THEN
          IF (NFVALS .LT. NFMAX) THEN
              CALL ADJTOL (N,M,A,IA,B,XL,XU,X,IACT,NACT,XBIG,RELACC,TOL,
     1          MEQL)
              GOTO 30
          ELSE
              INFO=8
          END IF
      END IF
C      IF (IPRINT .NE. 0) THEN
C          IF (INFO .EQ. 1) WRITE (IUOUT,1070)
C 1070     FORMAT (/5X,'GETMIN HAS ACHIEVED THE REQUIRED ACCURACY')
C          IF (INFO .EQ. 2) WRITE (IUOUT,1080)
C 1080     FORMAT (/5X,'GETMIN CAN MAKE NO FURTHER PROGRESS BECAUSE',
C     1      ' OF ROUNDING ERRORS')
C          IF (INFO .EQ. 3) WRITE (IUOUT,1090)
C 1090     FORMAT (/5X,'GETMIN CAN MAKE NO FURTHER PROGRESS BECAUSE',
C     1      ' F WILL NOT DECREASE ANY MORE')
C          IF (INFO .EQ. 8) WRITE (IUOUT,1100)
C 1100     FORMAT (/5X,'GETMIN HAS REACHED THE GIVEN LIMIT ON THE',
C     1      ' NUMBER OF CALLS OF FGCALC')
C      END IF
   40 RETURN
      END
C
C  ---------------------------------------------------------------------
C
      SUBROUTINE MINFUN (N,M,A,IA,B,XL,XU,X,ACC,IACT,NACT,PAR,IPRINT,
     1  INFO,G,Z,U,XBIG,RELACC,ZZNORM,TOL,MEQL,MTOT,ITERC,NFVALS,
     2  NFMAX,RESKT,BRES,D,ZTG,GM,XS,GS,F,IEXAU)
      IMPLICIT $(TYPE) (A-H,O-Z)
      DIMENSION A(IA,*),B(*),XL(*),XU(*),X(*),IACT(*),PAR(*),G(*),Z(*),
     1  U(*),XBIG(*),RESKT(*),BRES(*),D(*),ZTG(*),GM(*),XS(*),GS(*)
      COMMON /UNITS/ IUINP, IUOUT
      COMMON /TOTCAL/ ITNOCS, IPARTC, ITDER, IPDER, ITHESS, IPHESS,
     &                ITJAC, IPJAC
C
C     Initialize the minimization calculation.
C
      MSAT=MTOT
      ITERK=ITERC
      NFVALK=NFVALS
      IF (NFVALS .EQ. 0 .OR. INFO .EQ. 1) THEN
          ISOFAR = ITNOCS
          CALL FGCALC (N,X,F,G)
          NFVALS=NFVALS+(ITNOCS-ISOFAR)
      END IF
      FPREV=ABS(F+F+1.0)
      ITERP=-1
      IF (IPRINT .NE. 0) THEN
C          WRITE (IUOUT,1000) TOL
C 1000     FORMAT (/5X,'NEW VALUE OF TOL =',1PE13.5)
          ITERP=ITERC
      END IF
C
C     Calculate the next search direction.
C
   10 CALL CONRES (N,M,A,IA,B,XL,XU,X,IACT,NACT,PAR,G,Z,U,XBIG,BRES,D,
     1  ZTG,RELACC,TOL,STEPCB,DDOTG,MEQL,MSAT,MTOT,INDXBD,GM,RESKT,XS,
     2  GS)
C
C     Calculate the Kuhn Tucker residual vector.
C
      CALL KTVEC (N,M,A,IA,IACT,NACT,PAR,G,RESKT,Z,U,BRES,RELAXF,MEQL,
     1  SSQKT,XS,GS)
C
C     Test for convergence.
C
      IF (SSQKT .LE. ACC*ACC) THEN
          INFO=1
          GOTO 70
      END IF
      IF (DDOTG .GE. 0.0) THEN
          INFO=2
          GOTO 70
      END IF
C
C     Test for termination due to no decrease in F.
C
      IF (F .GE. FPREV) THEN
          IF (TOL .EQ. RELACC .OR. NACT .EQ. 0) THEN
              IF (DIFF .GT. 0.0) GOTO 20
          END IF
          INFO=3
          GOTO 70
      END IF
   20 DIFF=FPREV-F
      FPREV=F
C
C     Test that more calls of FGCALC are allowed.
C
      IF (NFVALS .GE. NFMAX) THEN
          INFO=8
          GOTO 70
      END IF
C
C     Test whether to reduce TOL and to provide printing.
C
      IF (TOL .GT. RELACC .AND. ITERC .GT. ITERK .AND.
     1  0.1*RELAXF .GE. MAX(DIFF,-0.5*DDOTG)) GOTO 70
      IF (ITERP .EQ. ITERC) THEN
          ITERP=ITERC+IABS(IPRINT)
          GOTO 80
      END IF
C
C     Calculate the step along the search direction.
C
   40 ITERC=ITERC+1
      CALL LSRCH (N,X,G,D,XS,GS,RELACC,STEPCB,DDOTG,F,STEP,NFVALS,
     1  NFMAX,BRES)
      IF (STEP .EQ. 0.0) THEN
          INFO=3
          SUM=0.0
          DO 50 I=1,N
   50     SUM=SUM+ABS(D(I)*GS(I))
          IF (DDOTG+RELACC*SUM .GE. 0.0) INFO=2
          GOTO 70
      END IF
C
C     Revise XBIG.
C
      DO 60 I=1,N
   60 XBIG(I)=MAX(XBIG(I),ABS(X(I)))
C
C     Revise the second derivative approximation.
C
      CALL ZBFGS (N,X,NACT,G,Z,ZTG,XS,GS,ZZNORM)
C
C     Add a constraint to the active set if it restricts the step.
C
      IF (STEP .EQ. STEPCB) THEN
          K=IACT(INDXBD)
          IF (K .GT. M) THEN
              K=K-M
              IF (K .LE. N) THEN
                  X(K)=XL(K)
              ELSE
                  X(K-N)=XU(K-N)
              END IF
          END IF
          CALL ADDCON (N,M,A,IA,IACT,NACT,Z,U,RELACC,INDXBD,XS,GS)
      END IF
      GOTO 10
C
C     Printing from the subroutine.
C
   70 IF (IPRINT .EQ. 0) GOTO 90
      ITERP=-1
80    CALL MDIS(X,F,ITERC,NFVALS,NFMAX,IPRINT,0)
C
C  Did we reach our target value ?
      IF (ITCONV(F).NE.0) THEN
          INFO = 9
          IEXAU = 1
          RETURN
      END IF
 1010 FORMAT (/5X,'ITERS =',I4,5X,'F.VALS =',I4,5X,'F =',1PE15.7)
C      IF (NFVALS .GT. NFVALK) THEN
C          WRITE (*,1020) (X(I),I=1,N)
C 1020     FORMAT ('  X =',(1P5E14.5))
C          WRITE (*,1030) (G(I),I=1,N)
C 1030     FORMAT ('  G =',(1P5E14.5))
C      ELSE
C          WRITE (*,1040)
C 1040     FORMAT (5X,'NO CHANGE TO X AND G SINCE PREVIOUS OUTPUT')
C      END IF
C      IF (IPRINT .LT. 0) THEN
C          IF (NACT .EQ. 0) THEN
C              WRITE (*,1050)
C 1050         FORMAT (5X,'NO ACTIVE CONSTRAINTS')
C          ELSE
C              WRITE (*,1060) (IACT(I),I=1,NACT)
C 1060         FORMAT (' IA =',(14I5))
C              WRITE (*,1070) (PAR(I),I=1,NACT)
C 1070         FORMAT (' LP =',(1P5E14.5))
C          END IF
C          IF (NACT .EQ. N) THEN
C              WRITE (*,1080)
C 1080         FORMAT (5X,'KT RESIDUAL VECTOR IS ZERO')
C          ELSE
C              WRITE (*,1090) (RESKT(I),I=1,N)
C 1090         FORMAT (' KT =',(1P5E14.5))
C          END IF
C      END IF
      IF (ITERP .GE. 0) GOTO 40
   90 RETURN
      END
C
C  ---------------------------------------------------------------------
C
      SUBROUTINE NEWCON (N,M,A,IA,IACT,NACT,Z,U,D,RELACC,MDEG,ZZDIAG,
     1  GMNEW,CGRAD)
      IMPLICIT $(TYPE) (A-H,O-Z)
      DIMENSION A(IA,*),IACT(*),Z(*),U(*),D(*),ZZDIAG(*),GMNEW(*),
     1  CGRAD(*)
C
C     Initialization.
C
      NP=NACT+1
      KHIGH=MDEG
      IZ=0
      DO 20 I=1,N
      ZZDIAG(I)=0.0
      DO 10 J=NP,N
   10 ZZDIAG(I)=ZZDIAG(I)+Z(IZ+J)**2
   20 IZ=IZ+N
C
C     Calculate the scalar products of D with its constraints.
C
   30 CVMAX=0.0
      DO 50 K=NP,KHIGH
      J=IACT(K)
      IF (J .LE. M) THEN
          SUM=0.0
          SUMABS=0.0
          SUMD=0.0
          DO 40 I=1,N
          TEMP=D(I)*A(I,J)
          SUM=SUM+TEMP
          SUMABS=SUMABS+ABS(TEMP)
   40     SUMD=SUMD+ZZDIAG(I)*A(I,J)**2
      ELSE
          JM=J-M
          IF (JM .LE. N) THEN
              SUM=-D(JM)
          ELSE
              JM=JM-N
              SUM=D(JM)
          END IF
          SUMABS=ABS(SUM)
          SUMD=ZZDIAG(JM)
      END IF
C
C     Pick out the most violated constraint, or return if the
C       violation is negligible.
C
      IF (SUM .GT. RELACC*SUMABS) THEN
          CVIOL=SUM*SUM/SUMD
          IF (CVIOL .GT. CVMAX) THEN
              CVMAX=CVIOL
              IADD=K
              SAVSUM=SUM
              SAVABS=SUMABS
          END IF
      END IF
   50 CONTINUE
      IF (CVMAX .LE. 0.0) GOTO 140
      IF (NACT .EQ. 0) GOTO 120
C
C     Set GMNEW to the gradient of the most violated constraint.
C
      J=IACT(IADD)
      IF (J .LE. M) THEN
          JMV=0
          DO 60 I=1,N
   60     GMNEW(I)=A(I,J)
      ELSE
          JMV=J-M
          DO 70 I=1,N
   70     GMNEW(I)=0.0
          IF (JMV .LE. N) THEN
              GMNEW(JMV)=-1.0
          ELSE
              JMV=JMV-N
              GMNEW(JMV)=1.0
          END IF
      END IF
C
C     Modify GMNEW for the next active constraint.
C
      K=NACT
   80 TEMP=0.0
      IZ=K
      DO 90 I=1,N
      TEMP=TEMP+Z(IZ)*GMNEW(I)
   90 IZ=IZ+N
      TEMP=TEMP*U(K)
      J=IACT(K)
      IF (J .LE. M) THEN
          DO 100 I=1,N
  100     GMNEW(I)=GMNEW(I)-TEMP*A(I,J)
      ELSE
          JM=J-M
          IF (JM .LE. N) THEN
              GMNEW(JM)=GMNEW(JM)+TEMP
          ELSE
              GMNEW(JM-N)=GMNEW(JM-N)-TEMP
          END IF
      END IF
C
C     Revise the values of SAVSUM and SAVABS.
C
      SUM=0.0
      SUMABS=0.0
      DO 110 I=1,N
      TEMP=D(I)*GMNEW(I)
      SUM=SUM+TEMP
  110 SUMABS=SUMABS+ABS(TEMP)
      SAVSUM=MIN(SAVSUM,SUM)
      SAVABS=MAX(SAVABS,SUMABS)
      K=K-1
      IF (K .GE. 1) GOTO 80
C
C     Add the new constraint to the active set if the constraint
C       violation is still significant.
C
      IF (JMV .GT. 0) D(JMV)=0.0
      IF (SAVSUM .LE. RELACC*SAVABS) GOTO 130
  120 K=NACT
      CALL ADDCON (N,M,A,IA,IACT,NACT,Z,U,RELACC,IADD,GMNEW,CGRAD)
      IF (NACT .GT. K) GOTO 140
C
C     Seek another constraint violation.
C
      IADD=NP
  130 IF (NP .LT. KHIGH) THEN
          K=IACT(KHIGH)
          IACT(KHIGH)=IACT(IADD)
          IACT(IADD)=K
          KHIGH=KHIGH-1
          GOTO 30
      END IF
  140 RETURN
      END
C
C  ---------------------------------------------------------------------
C
      SUBROUTINE SATACT (N,M,A,IA,B,XL,XU,X,IACT,NACT,INFO,Z,U,XBIG,
     1  RELACC,TOL,MEQL)
      IMPLICIT $(TYPE) (A-H,O-Z)
      DIMENSION A(IA,*),B(*),XL(*),XU(*),X(*),IACT(*),Z(*),U(*),
     1  XBIG(*)
      IF (NACT .EQ. 0) GOTO 50
      DO 30 K=1,NACT
C
C     Calculate the next constraint residual.
C
      J=IACT(K)
      IF (J .LE. M) THEN
          RES=B(J)
          RESABS=ABS(B(J))
          RESBIG=RESABS
          DO 10 I=1,N
          TEMPA=A(I,J)
          TEMP=TEMPA*X(I)
          RES=RES-TEMP
          RESABS=RESABS+ABS(TEMP)
   10     RESBIG=RESBIG+ABS(TEMPA)*XBIG(I)
      ELSE
          JX=J-M
          IF (JX .LE. N) THEN
              RES=X(JX)-XL(JX)
              RESABS=ABS(X(JX))+ABS(XL(JX))
              RESBIG=XBIG(JX)+ABS(XL(JX))
              SAVEX=XL(JX)
          ELSE
              JX=JX-N
              RES=XU(JX)-X(JX)
              RESABS=ABS(X(JX))+ABS(XU(JX))
              RESBIG=XBIG(JX)+ABS(XU(JX))
              SAVEX=XU(JX)
          END IF
      END IF
C
C     Shift X if necessary.
C
      IF (RES .NE. 0.0) THEN
          TEMP=RES/RESABS
          IF (K .LE. MEQL) TEMP=-ABS(TEMP)
          IF (TOL .EQ. RELACC .OR. TEMP+RELACC .LT. 0.0) THEN
              INFO=1
              SCALE=RES*U(K)
              IZ=K
              DO 20 I=1,N
              X(I)=X(I)+SCALE*Z(IZ)
              IZ=IZ+N
   20         XBIG(I)=MAX(XBIG(I),ABS(X(I)))
              IF (J .GT. M) X(JX)=SAVEX
C
C     Else flag a constraint deletion if necessary.
C
          ELSE IF (RES/RESBIG .GT. TOL) THEN
              IACT(K)=-IACT(K)
          END IF
      END IF
   30 CONTINUE
C
C     Delete any flagged constraints and then return.
C
      IDROP=NACT
  40  IF (IACT(IDROP) .LT. 0) THEN
          IACT(IDROP)=-IACT(IDROP)
          CALL DELCON (N,M,A,IA,IACT,NACT,Z,U,RELACC,IDROP)
      END IF
      IDROP=IDROP-1
      IF (IDROP .GT. MEQL) GOTO 40
   50 RETURN
      END
C
C  ---------------------------------------------------------------------
C
      SUBROUTINE SDEGEN (N,M,A,IA,IACT,NACT,PAR,Z,U,D,ZTG,GM,RELACC,
     1  DDOTGM,MEQL,MDEG,GMNEW,PARNEW,CGRAD)
      IMPLICIT $(TYPE) (A-H,O-Z)
      DIMENSION A(IA,*),IACT(*),PAR(*),Z(*),U(*),D(*),ZTG(*),GM(*),
     1  GMNEW(*),PARNEW(*),CGRAD(*)
      MP=MEQL+1
      DTEST=0.0
C
C     Calculate the search direction and branch if it is not downhill.
C
   10 CALL SDIRN (N,NACT,Z,D,ZTG,GM,RELACC,DDOTGM)
      IF (DDOTGM .EQ. 0.0) GOTO 120
C
C     Branch if there is no need to consider any degenerate constraints.
C     The test gives termination if two consecutive additions to the
C       active set fail to increase the predicted new value of F.
C
      IF (NACT .EQ. MDEG) GOTO 120
      NP=NACT+1
      SUM=0.0
      DO 20 J=NP,N
   20 SUM=SUM+ZTG(J)**2
      IF (DTEST .GT. 0.0 .AND. SUM .GE. DTEST) THEN
          IF (ITEST .EQ. 1) GOTO 120
          ITEST=1
      ELSE
          DTEST=SUM
          ITEST=0
      END IF
C
C     Add a constraint to the active set if there are any significant
C       violations of degenerate constraints.
C
      K=NACT
      CALL NEWCON (N,M,A,IA,IACT,NACT,Z,U,D,RELACC,MDEG,GMNEW,PARNEW,
     1  CGRAD)
      IF (NACT .EQ. K) GOTO 120
      PAR(NACT)=0.0
C
C     Calculate the new reduced gradient and Lagrange parameters.
C
   30 DO 40 I=1,N
   40 GMNEW(I)=GM(I)
      K=NACT
   50 TEMP=0.0
      IZ=K
      DO 60 I=1,N
      TEMP=TEMP+Z(IZ)*GMNEW(I)
   60 IZ=IZ+N
      TEMP=TEMP*U(K)
      PARNEW(K)=PAR(K)+TEMP
      AMX = 0.
      IF (K .EQ. NACT) PARNEW(K)=MIN(PARNEW(K),AMX)
      J=IACT(K)
      IF (J .LE. M) THEN
          DO 70 I=1,N
   70     GMNEW(I)=GMNEW(I)-TEMP*A(I,J)
      ELSE
          JM=J-M
          IF (JM .LE. N) THEN
              GMNEW(JM)=GMNEW(JM)+TEMP
          ELSE
              GMNEW(JM-N)=GMNEW(JM-N)-TEMP
          END IF
      END IF
      K=K-1
      IF (K .GT. MEQL) GOTO 50
C
C     Set RATIO for linear interpolation between PAR and PARNEW.
C
      RATIO=0.0
      IF (MP .LT. NACT) THEN
          KU=NACT-1
          DO 80 K=MP,KU
          IF (PARNEW(K) .GT. 0.0) THEN
              RATIO=PARNEW(K)/(PARNEW(K)-PAR(K))
              IDROP=K
          END IF
   80     CONTINUE
      END IF
C
C     Apply the linear interpolation.
C
      THETA=1.0-RATIO
      AMX = 0.
      DO 90 K=MP,NACT
   90 PAR(K)=MIN(THETA*PARNEW(K)+RATIO*PAR(K),AMX)
      DO 100 I=1,N
  100 GM(I)=THETA*GMNEW(I)+RATIO*GM(I)
C
C     Drop a constraint if RATIO is positive.
C
      IF (RATIO .GT. 0.0) THEN
          CALL DELCON (N,M,A,IA,IACT,NACT,Z,U,RELACC,IDROP)
          DO 110 K=IDROP,NACT
  110     PAR(K)=PAR(K+1)
          GOTO 30
      END IF
C
C     Return if there is no freedom for a new search direction.
C
      IF (NACT .LT. N) GOTO 10
      DDOTGM=0.0
  120 RETURN
      END
C
C  ---------------------------------------------------------------------
C
      SUBROUTINE SDIRN (N,NACT,Z,D,ZTG,GM,RELACC,DDOTGM)
      IMPLICIT $(TYPE) (A-H,O-Z)
      DIMENSION Z(*),D(*),ZTG(*),GM(*)
      DDOTGM=0.0
      IF (NACT .GE. N) GOTO 60
C
C     Premultiply GM by the transpose of Z.
C
      NP=NACT+1
      DO 20 J=NP,N
      SUM=0.0
      SUMABS=0.0
      IZ=J
      DO 10 I=1,N
      TEMP=Z(IZ)*GM(I)
      SUM=SUM+TEMP
      SUMABS=SUMABS+ABS(TEMP)
   10 IZ=IZ+N
      IF (ABS(SUM) .LE. RELACC*SUMABS) SUM=0.0
   20 ZTG(J)=SUM
C
C     Form D by premultiplying ZTG by -Z.
C
      IZ=0
      DO 40 I=1,N
      SUM=0.0
      SUMABS=0.0
      DO 30 J=NP,N
      TEMP=Z(IZ+J)*ZTG(J)
      SUM=SUM-TEMP
   30 SUMABS=SUMABS+ABS(TEMP)
      IF (ABS(SUM) .LE. RELACC*SUMABS) SUM=0.0
      D(I)=SUM
   40 IZ=IZ+N
C
C     Test that the search direction is downhill.
C
      SUMABS=0.0
      DO 50 I=1,N
      TEMP=D(I)*GM(I)
      DDOTGM=DDOTGM+TEMP
   50 SUMABS=SUMABS+ABS(TEMP)
      IF (DDOTGM+RELACC*SUMABS .GE. 0.0) DDOTGM=0.0
   60 RETURN
      END
C
C  ---------------------------------------------------------------------
C
      SUBROUTINE STEPBD (N,M,A,IA,IACT,BRES,D,STEPCB,DDOTG,MDEG,MSAT,
     1  MTOT,INDXBD)
      IMPLICIT $(TYPE) (A-H,O-Z)
      DIMENSION A(IA,*),IACT(*),BRES(*),D(*)
C
C     Set steps to constraint boundaries and find the least
C     positive one.
C
      IFLAG=0
      STEPCB=0.0
      INDXBD=0
      K=MDEG
   10 K=K+1
      IF (K .GT. MTOT) GOTO 700
C
C     Form the scalar product of D with the current constraint normal.
C
   20     J=IACT(K)
          IF (J .LE. M) THEN
              SP=0.0
              DO 30 I=1,N
   30         SP=SP+D(I)*A(I,J)
          ELSE
              JM=J-M
              IF (JM .LE. N) THEN
                  SP=-D(JM)
              ELSE
                  SP=D(JM-N)
              END IF
          END IF
C
C     The next branch is taken if label 20 was reached via label 40.
C
          IF (IFLAG .EQ. 1) GOTO 50
C
C     Set BRES(J) to indicate the status of the j-th constraint.
C
          IF (SP*BRES(J) .LE. 0.0) THEN
              BRES(J)=0.0
          ELSE
              BRES(J)=BRES(J)/SP
              IF (STEPCB .EQ. 0.0 .OR. BRES(J) .LT. STEPCB) THEN
                  STEPCB=BRES(J)
                  INDXBD=K
              END IF
          END IF
          GO TO 10
700   CONTINUE
C
C     Try to pass through the boundary of a violated constraint.
C
   40 IF (INDXBD .LE. MSAT) GOTO 800
          IFLAG=1
          K=INDXBD
          GOTO 20
   50     MSAT=MSAT+1
          IACT(INDXBD)=IACT(MSAT)
          IACT(MSAT)=J
          BRES(J)=0.0
          INDXBD=MSAT
          DDOTG=DDOTG-SP
          IF (DDOTG .LT. 0.0 .AND. MSAT .LT. MTOT) THEN
C
C     Seek the next constraint boundary along the search direction.
C
              TEMP=0.0
              KL=MDEG+1
              DO 60 K=KL,MTOT
              J=IACT(K)
              IF (BRES(J) .GT. 0.0) THEN
                  IF (TEMP .EQ. 0.0 .OR. BRES(J) .LT. TEMP) THEN
                      TEMP=BRES(J)
                      INDXBD=K
                  END IF
              END IF
   60         CONTINUE
              IF (TEMP .GT. 0.0) THEN
                  STEPCB=TEMP
                  GOTO 40
              END IF
          END IF
800   CONTINUE
      RETURN
      END
C
C  ---------------------------------------------------------------------
C
      SUBROUTINE ZBFGS (N,X,NACT,G,Z,ZTG,XS,GS,ZZNORM)
      IMPLICIT $(TYPE) (A-H,O-Z)
      DIMENSION X(*),G(*),Z(*),ZTG(*),XS(*),GS(*)
C
C     Test if there is sufficient convexity for the update.
C
      DD=0.0
      DG=0.0
      TEMP=0.0
      DO 10 I=1,N
      XS(I)=X(I)-XS(I)
      DD=DD+XS(I)**2
      TEMP=TEMP+GS(I)*XS(I)
      GS(I)=G(I)-GS(I)
   10 DG=DG+GS(I)*XS(I)
      IF (DG .LT. 0.1*ABS(TEMP)) GOTO 90
C
C     Transform the Z matrix.
C
      K=N
   20 KP=K
      K=K-1
      IF (K .GT. NACT) THEN
          IF (ZTG(KP) .EQ. 0.0) GOTO 20
          TEMP=ABS(ZTG(KP))*SQRT(1.0+(ZTG(K)/ZTG(KP))**2)
          WCOS=ZTG(K)/TEMP
          WSIN=ZTG(KP)/TEMP
          ZTG(K)=TEMP
          IZ=K
          DO 30 I=1,N
          TEMP=WCOS*Z(IZ+1)-WSIN*Z(IZ)
          Z(IZ)=WCOS*Z(IZ)+WSIN*Z(IZ+1)
          Z(IZ+1)=TEMP
   30     IZ=IZ+N
          GOTO 20
      END IF
C
C     Update the value of ZZNORM.
C
      IF (ZZNORM .LT. 0.0) THEN
          ZZNORM=DD/DG
      ELSE
          TEMP=SQRT(ZZNORM*DD/DG)
          ZZNORM=MIN(ZZNORM,TEMP)
          ZZNORM=MAX(ZZNORM,0.1*TEMP)
      END IF
C
C     Complete the updating of Z.
C
      NP=NACT+1
      TEMP=SQRT(DG)
      IZ=NP
      DO 40 I=1,N
      Z(IZ)=XS(I)/TEMP
   40 IZ=IZ+N
      IF (NP .LT. N) THEN
          KM=NP+1
          DO 80 K=KM,N
          TEMP=0.0
          IZ=K
          DO 50 I=1,N
          TEMP=TEMP+GS(I)*Z(IZ)
   50     IZ=IZ+N
          TEMP=TEMP/DG
          SUM=0.0
          IZ=K
          DO 60 I=1,N
          Z(IZ)=Z(IZ)-TEMP*XS(I)
          SUM=SUM+Z(IZ)**2
   60     IZ=IZ+N
          IF (SUM .LT. ZZNORM) THEN
              TEMP=SQRT(ZZNORM/SUM)
              IZ=K
              DO 70 I=1,N
              Z(IZ)=TEMP*Z(IZ)
   70         IZ=IZ+N
          END IF
   80     CONTINUE
      END IF
   90 RETURN
      END
C  ---------------------------------------------------------------------
      SUBROUTINE PRADJ
C  ---------------------------------------------------------------------
C
C  Description:
C   PRADJ is a step-size procedure. It does dot require the
C   values of the gradient components. It also performs a minor
C   minimization. It is invoked by the ADJUST command.
C
C  ---------------------------------------------------------------------
      IMPLICIT $(TYPE) (A-H,O-Z)
C  Common blocks:
      PARAMETER ( MXV = $(MXV) )
      COMMON /PARAMS/ POINT(MXV), VAL
      COMMON /DETAIL/ IXAT(MXV), MARG(MXV)
      COMMON /XLIMS/ XLL(MXV), XRL(MXV)
      COMMON /ROSTEP/ STEP(MXV)
      COMMON /ACCURA/ COMACC
      COMMON /DIMEN/ NOD, NOF
      COMMON /TOTCAL/ ITNOCS, IPARTC, ITDER, IPDER, ITHESS, IPHESS,
     &                ITJAC, IPJAC
      COMMON /MODE3/ MODEV, IFFORM, INMCL
      COMMON /RWORK/ XT1(MXV), XT2(MXV), XT3(MXV), XT4(MXV), XT5(MXV),
     &               XT6(MXV), XT7(MXV+1)
C  ---------------------------------------------------------------------
C  Local variables:
      LOGICAL LM, RM
C  ---------------------------------------------------------------------
C
      IF (ITNOCS.EQ.0 .OR. MODEV.EQ.0) RETURN
      COMTOL = SQRT(COMACC)
C
C  Add a backup record.
      CALL BACKUP('M')
C
      DO 3,I=1,NOD
         IF (IXAT(I).EQ.0) GOTO 3
         LM = MARG(I).EQ.-1 .OR. MARG(I).EQ.2
         RM = MARG(I).EQ.1  .OR. MARG(I).EQ.2
         ISTORY = 0
         CALL POIMAR(IFEV)
         DO 1,II=1,NOD
            IF (POINT(II).EQ.(POINT(II)+STEP(II)))
     &                STEP(II) = (1.+ABS(POINT(II)))*COMTOL
            XT1(II) = POINT(II)
1        CONTINUE
         JJJ = 0
2        CONTINUE
         JJJ = JJJ+1
         IF (JJJ.GT.20) GOTO 3
         XT1(I) = POINT(I)+STEP(I)
         IF (LM .AND. XT1(I).LT.XLL(I)) XT1(I) = XLL(I)
         IF (RM .AND. XT1(I).GT.XRL(I)) XT1(I) = XRL(I)
         TRY = ACSQ(XT1,NOD)
         H1 = XT1(I)-POINT(I)
         FP1 = 0.
         IF (H1.NE.0.) FP1 = (TRY-VAL)/H1
         IF (TRY.LT.VAL) THEN
            VAL = TRY
            POINT(I) = XT1(I)
            STEP(I) = 3.0*STEP(I)
            ISTORY = 1
            GOTO 2
         ELSE
            IF (ISTORY.EQ.1) THEN
               STEP(I) = STEP(I)/2.
               GOTO 3
            ELSE
               XT1(I) = POINT(I)-STEP(I)
               IF (LM .AND. XT1(I).LT.XLL(I)) XT1(I) = XLL(I)
               IF (RM .AND. XT1(I).GT.XRL(I)) XT1(I) = XRL(I)
               TRYS = ACSQ(XT1,NOD)
               H2 = XT1(I)-POINT(I)
               FP2 = 0.
               IF (H2.NE.0.) FP2 = (TRYS-VAL)/H2
               IF (TRYS.LT.VAL) THEN
                  VAL = TRYS
                  STEP(I) = -STEP(I)*3.0
                  POINT(I) = XT1(I)
                  ISTORY = 1
                  GOTO 2
               ELSE
                  IF (ABS(FP1-FP2).GT.1.E-30) THEN
                     STEP(I) = -.5*(H1*FP2-H2*FP1)/(FP1-FP2)
                     GOTO 3
                  ELSE
                     STEP(I) = STEP(I)/2.
                     IF (TRY.GT.TRYS) STEP(I) = -STEP(I)
                     GOTO 2
                  END IF
               END IF
            END IF
         END IF
3     CONTINUE
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE QNLS ( DIR, GRAD, NOCALL, NOG, ALPHA, COMACC, LSMETH,
     &                  RHO, SIGMA, LSITER, POINT, VAL, NOD, INFO, IXAT,
     &                  MARG, XLL, XRL, WORK, WORK2 )
C  ---------------------------------------------------------------------
C
C  Description:
C    This subroutine is called whenever a minimization method needs to
C    perform a line search on a descent direction. It acts as a front
C    end to the actual line search routine LNS and returns the new
C    point: POINT + ALPHA * DIR.
C    Components of DIR that take us outside the margins are set to 0,
C    and may produce a non-descent direction. If margins are present,
C    ALPHA may not be positive.
C
C  Input arguments:
C    GRAD       Gradient at the current point.
C    COMACC     Machine accuracy.
C    LSMETH     Line search method. See subroutine LNS.
C    RHO        The line search "rho". See subroutine LNS.
C    SIGMA      The line search "sigma". See subroutine LNS.
C    LSITER     Maximum allowed number of line search iterations. See
C               subroutine LSITER.
C    NOD        Dimensionality of the objective function.
C    IXAT       Fix statuses.
C    MARG       Margin statuses.
C    XLL        Left margins.
C    XRL        Right margins.
C    WORK       Work array of length NOD.
C    WORK2      Work array of length NOD.
C
C  Output arguments:
C    NOCALL     Number of function evaluations that were performed.
C    NOG        Number of gradient evaluations that were performed.
C    INFO       The INFO parameter, as returned by subroutine LNS.
C               In addition, INFO = -1 is returned by this routine,
C               indicating that all the components of the search
C               direction DIR, drive us outside the margins.
C
C  Input / Output arguments:
C    DIR        On input the direction of search. It should be a
C               descent direction. On output, if any DIR(I) takes us
C               outside the magins, it is set to 0.
C               Warning: This may produce a non-descent direction.
C    ALPHA      On input an initial estimate. On output, the line
C               search ALPHA, or 0 if no progress was possible.
C               If margins are present, ALPHA is not guaranteed to
C               be positive.
C    POINT      On input, the current point. On output, the new point
C               after the line search.
C    VAL        Value of the objective function.
C
C  ---------------------------------------------------------------------
      IMPLICIT $(TYPE) (A-H,O-Z)
C  Arguments:
      DIMENSION DIR(NOD), GRAD(NOD), POINT(NOD), IXAT(NOD)
      DIMENSION MARG(NOD), XLL(NOD), XRL(NOD), WORK(NOD), WORK2(NOD)
C  ---------------------------------------------------------------------
C  Common blocks:
      COMMON /BIGCOM/ BIGNUM, BLIM, DWARF, DWALOG
      LOGICAL AIMON
      COMMON /AIM/ TARGET, AIMON
C  ---------------------------------------------------------------------
C
C  Prepare target value. Take into account the TARGET command.
      IF (AIMON) THEN
         FTARG = TARGET
      ELSE
         FTARG = -BIGNUM
      END IF
C
C  Some constants.
      ONE = 1.
      TEN = 10.
C
C  Calculate minimum acceptable ALPHA. Too small ALPHA values would
C  cause X-convergence to be detected in the calling routine.
      XMAX = 0.
      DO 40,I=1,NOD
40       XMAX = MAX(XMAX,ABS(DIR(I))/MAX(ABS(POINT(I)),ONE))
C
C  XMAX cannot be zero, since convergence would have been detected
C  by the calling routine. The following division is safe.
      ALAMIN = MIN(COMACC/XMAX,TEN)
C
C  Consider fixed the parameters that result in very small ALPHA.
      ALA = 10.*ALAMIN
      PP0 = 10.*ALA
      DO 11,I=1,NOD
         DD = DIR(I)
         IF (DD.NE.0) THEN
            MA = MARG(I)
            PP = PP0
            CP = POINT(I)
            IF (DD.GT.0.) THEN
               IF (MA.EQ. 1 .OR. MA.EQ.2) PP = (XRL(I)-CP)/DD
            ELSE
               IF (MA.EQ.-1 .OR. MA.EQ.2) PP = (XLL(I)-CP)/DD
            END IF
            IF (PP.LT.ALA) DIR(I) = 0.
         END IF
11    CONTINUE
C
C  Count zero directions.
      NFIXED = 0
      DO 1,I=1,NOD
1        IF (DIR(I).EQ.0.) NFIXED = NFIXED+1
C
C  If all parameters are fixed then return.
C  Note that at this point all parameters cannot be fixed due to FIX
C  statements, since this would have been detected by the calling
C  routine. If the direction vector DIR is zero, this is due to the
C  margins. The calling routine may want to attempt a steepest descent
C  (or some other) restart.
      IF (NFIXED.EQ.NOD) THEN
         ALPHA = 0
         INFO = -1
         RETURN
      END IF
C
C  Calculate margins XALIM, XBLIM along the direction of search.
      CALL DIRLIM(POINT,DIR,MARG,XLL,XRL,NOD,LIM,XALIM,XBLIM)
C
C  Setting DIR(I) = 0 for some parameters (due to the existence of
C  margins) may result in a non-descent direction.
C  Warning: The calling method should be prepared to handle negative
C  values for ALPHA.
      ISW = 0
      CALL VVMULT(DIR,GRAD,G,NOD)
      IF (G.GT.0. .AND. LIM.NE.0) THEN
         DO 88,I=1,NOD
88          DIR(I) = -DIR(I)
         XBLIM = ABS(XALIM)
         ISW = 1
      END IF
C
      IF (XBLIM.EQ.0. .AND. LIM.NE.0) THEN
         ALPHA = 0.
         NOCALL = 0
         NOG = 0
      ELSE
         IF (ALPHA.LT.ALAMIN) ALPHA = 1.2*ALAMIN
         CALL LNS(POINT,GRAD,DIR,NOD,VAL,RHO,SIGMA,LSITER,FTARG,
     &               ALPHA,FXMIN,COMACC,WORK,WORK2,LSMETH,NOCALL,NOG,
     &               INFO,LIM,XBLIM,MARG,XLL,XRL,IXAT)
      END IF
C
      IF (ISW.EQ.1) THEN
         DO 77,I=1,NOD
77          DIR(I) = -DIR(I)
         ALPHA = -ALPHA
      END IF
C
C  Calculate the new point and make sure it lies inside the margins.
      CALL MEVAL(POINT,DIR,ALPHA,NOD,IXAT,MARG,XLL,XRL,POINT,0,FTMP)
      VAL = FXMIN
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE ROLL ( MERR )
C  ---------------------------------------------------------------------
C
C  Description:
C    Imlements the ROLL command.
C    Syntax: ROLL (Has panel)
C
C  Output arguments:
C    MERR       Error indicator.
C
C  ---------------------------------------------------------------------
      IMPLICIT $(TYPE) (A-H,O-Z)
C  Common blocks:
      COMMON /MODE3/ MODEV, IFFORM, INMCL
      COMMON /TOTCAL/ ITNOCS, IPARTC, ITDER, IPDER, ITHESS, IPHESS,
     &                ITJAC, IPJAC
C  ---------------------------------------------------------------------
C  Local variables:
      CHARACTER*80 MESS
C  ---------------------------------------------------------------------
C
      CALL SETPI('CANCEL',1,IERR)
      CALL CHANGE(MERR)
      IF (MERR.NE.0) RETURN
      CALL GETPI('CANCEL',ICANC)
      IF (ICANC.NE.1 .OR. ITNOCS.EQ.0 .OR. MODEV.EQ.0) RETURN
C
C  Get parameters from the panel.
      CALL GETPI('NOC',NOCH)
      CALL GETPR('LSTOL',ACCH)
      CALL GETPR('SFACTOR',ENHH)
      CALL GETPI('FAIL',IFMH)
      CALL GETPI('PRINT',IPRINT)
      CALL GETPR('FTOL',FTOL)
      CALL GETPI('LSNOC',LSNOC)
C
C  Call the minimizer.
      CALL HYPER(NOCH,ACCH,ENHH,IFMH,IPRINT,ICOUNT,FTOL,LSNOC,INFO,ITER)
C
C  Construct a message describing the reason for termination.
      CALL ROLLTM(INFO,MESS)
C
C  Report the reason for termination.
      NOGEX = -1
      NOJAC = -1
      CALL MINREP(MESS,ICOUNT,NOGEX,NOJAC,ITER)
C
C  Add backup record.
      CALL BACKUP('M')
C
C  Assign values to MCL-specific keywords.
      CALL SETPI('FCALLS',ICOUNT,IERR)
      CALL SETPI('ITERDONE',ITER,IERR)
      CALL SETPI('INFO',INFO,IERR)
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE HYPER ( NOCAL, ACCU, ENH, IFM, IPRINT, ICOUNT, FTOL,
     &                   LSNOC, INFO, ITER )
C  ---------------------------------------------------------------------
C
C  Description:
C    Implementation of the ROLL algorithm. It was first described in:
C    Evangelakis G.A et al.
C    Merlin - A portable system for multidimensional minimization.
C    Comput. Phys. Commun., 46 (1987) 401-415.
C
C  Input arguments:
C    NOCAL      Approximate maximum number of ACSQ evaluations allowed.
C    ACCU       Line search tolerance.
C    ENH        Enhancement step-factor.
C    IFM        Maximum number of consecutive failure-cycles allowed.
C    IPRINT     Printout level:
C               IPRINT = 0 -> Nothing is printed.
C               IPRINT = 1 -> Simple printout: Prints iterations,
C                             function value, number of calls, maximum
C                             number of calls.
C               IPRINT = 2 -> Full printout. Prints in addition the
C                             minimization parameters, contained in
C                             array POINT.
C    FTOL       Minimum relative drop accepted.
C    LSNOC      Maximum number of ACSQ calls for each line search.
C
C  Output arguments:
C    ICOUNT     Number of ACSQ evaluations.
C    INFO       Reason for termination:
C               INFO = 1 -> Target value has been reached.
C               INFO = 2 -> All calls exhausted.
C               INFO = 3 -> Too many cycle failures.
C    ITER       Number of ROLL iterations.
C
C  ---------------------------------------------------------------------
      IMPLICIT $(TYPE) (A-H,O-Z)
C  Arguments:
C  ---------------------------------------------------------------------
C  Common blocks:
      PARAMETER ( MXV = $(MXV) )
      COMMON /PARAMS/ POINT(MXV), VAL
      COMMON /DETAIL/ IXAT(MXV), MARG(MXV)
      COMMON /XLIMS/ XLL(MXV), XRL(MXV)
      COMMON /ROSTEP/ STEP(MXV)
      COMMON /DIMEN/ NOD, NOF
      COMMON /ACCURA/ COMACC
      LOGICAL AIMON
      COMMON /AIM/ TARGET, AIMON
      COMMON /BIGCOM/ BIGNUM, BLIM, DWARF, DWALOG
      COMMON /RWORK/ XT1(MXV), XT2(MXV), XT3(MXV), XT4(MXV), XPO(MXV),
     &               XT6(MXV), XT7(MXV+1)
      COMMON /IWORK/ INDE(MXV)
      CHARACTER TAB, ESC
      COMMON /SPECHA/ TAB, ESC
      LOGICAL STRO, WEAK
      COMMON /PRILOG/ STRO, WEAK
      COMMON /UNITS/ IUINP, IUOUT
C  ---------------------------------------------------------------------
C
C  Initialize target value.
      IF (AIMON) THEN
         FTARG = TARGET
      ELSE
         FTARG = -BIGNUM
      END IF
      COMTOL = SQRT(COMACC)
C
C  Initialize call, failure and iteration (cycle) counters.
      ICOUNT = 0
      IFAIL = 0
      ITER = 0
C
C  Initialize progress status.
      DO 10,I=1,NOD
10       INDE(I) = 0
C
C  Did we reach our target value ?
      IF (ITCONV(VAL).NE.0) THEN
         INFO = 1
         RETURN
      END IF
      IF (ICOUNT.GE.NOCAL) THEN
         INFO = 2
         RETURN
      END IF
C
C  Begining of the cycle-loop
C
C  ---------------------------------------------------------------------
C  The following loop performs ROLL iterations (cycles).
C  ---------------------------------------------------------------------
1000  CONTINUE
      ITER = ITER+1
C
C  Loop to check if steps are meaningfull. If not adjust them.
      DO 1,K=1,NOD
         XXX = POINT(K)+STEP(K)
         IF (XXX.EQ.POINT(K)) STEP(K) = (1.+ABS(XXX))*COMTOL
1     CONTINUE
C
      IMIN = 0
      OLDVAL = VAL
C
C  Begining of loop over all variables.
      DO 5,I=1,NOD
C
C  Set the trial point XPO equal to current point.
         DO 2,J=1,NOD
2           XPO(J) = POINT(J)
C
C  Skip to the next variable if this one is fixed.
         IF (IXAT(I).EQ.0) GOTO 5
C
C  Modify the I-th component of the trial point XPO.
         XPO(I) = POINT(I)+STEP(I)
C
C  Check if this component complies with the margins.
C  If it is inside an infeasible region, reset it appropriately.
         IF (MARG(I).EQ.-1) THEN
            IF (XPO(I).LT.XLL(I)) XPO(I) = (XLL(I)+POINT(I))/2.
         ELSE IF (MARG(I).EQ.1) THEN
            IF (XPO(I).GT.XRL(I)) XPO(I) = (XRL(I)+POINT(I))/2.
         ELSE IF (MARG(I).EQ.2) THEN
            IF (XPO(I).LT.XLL(I)) XPO(I) = (XLL(I)+POINT(I))/2.
            IF (XPO(I).GT.XRL(I)) XPO(I) = (XRL(I)+POINT(I))/2.
         END IF
C
         STEP(I) = XPO(I)-POINT(I)
         H1 = STEP(I)
C
C  Check if POINT is any different than XPO.
C  If so, loop to the next variable.
         IF (POINT(I).EQ.XPO(I)) GOTO 5
C
C  Evaluate the function at the first trial point.
         TRIAL1 = ACSQ(XPO,NOD)
         ICOUNT = ICOUNT+1
C
C  Calculate an estimate for the derivative.
         FP1 = (TRIAL1-VAL)/H1
         IF (TRIAL1.LT.VAL) THEN
C
C  Success at the first try.
            INDE(I) = 1
            IMIN = I
C
C  Update value and point.
            VAL = TRIAL1
            DO 3,K=1,NOD
3              POINT(K) = XPO(K)
            CALL MDIS(POINT,VAL,ITER,ICOUNT,NOCAL,IPRINT,0)
C
C  Modify step and check target and calls criteria.
            STEP(I) = ENH*STEP(I)
C
C  Did we reach our target value ?
            IF (ITCONV(VAL).NE.0) THEN
               INFO = 1
               RETURN
            END IF
            IF (ICOUNT.GE.NOCAL) THEN
               INFO = 2
               RETURN
            END IF
         ELSE
C
C  Failure at the first try. Modify the I-th component of the
C  current point, in the opposite direction.
            XPO(I) = POINT(I)-STEP(I)
C
C  Check if this component complies with the margins.
C  If it is inside an infeasible region, reset it appropriately.
            IF (MARG(I).EQ.-1) THEN
               IF (XPO(I).LT.XLL(I)) XPO(I) = (XLL(I)+POINT(I))/2.
            ELSE IF (MARG(I).EQ.1) THEN
               IF (XPO(I).GT.XRL(I)) XPO(I) = (XRL(I)+POINT(I))/2.
            ELSE IF (MARG(I).EQ.2) THEN
               IF (XPO(I).LT.XLL(I)) XPO(I) = (XLL(I)+POINT(I))/2.
               IF (XPO(I).GT.XRL(I)) XPO(I) = (XRL(I)+POINT(I))/2.
            END IF
C
            STEP(I) = POINT(I)-XPO(I)
            H2 = STEP(I)
C
C  Check if POINT is any different than XPO.
C  If so, loop to the next variable.
            IF (POINT(I).EQ.XPO(I)) GOTO 5
C
C  Evaluate the function at the second trial point.
            TRIAL2=ACSQ(XPO,NOD)
            ICOUNT = ICOUNT+1
C
C  Calculate an estimate for the derivative.
            FP2 = -(TRIAL2-VAL)/H2
            IF (TRIAL2.LT.VAL) THEN
C
C  Success at the second try.
               INDE(I) = 1
               IMIN = I
C
C  Update value and point.
               VAL = TRIAL2
               DO 4,K=1,NOD
4                 POINT(K) = XPO(K)
               CALL MDIS(POINT,VAL,ITER,ICOUNT,NOCAL,IPRINT,0)
C
               STEP(I) = -STEP(I)*ENH
C
C  Check if the second derivative is positive.
C  In that case form the newton step xxx = -F'/F''
               IF ((FP1-FP2)*(H1+H2).GT.0.) THEN
                  XXX = -.5*(H2*FP1+H1*FP2)/(FP1-FP2)
C
C  Pick as step the longest one.
                  IF (ABS(XXX).GT.ABS(STEP(I))) STEP(I) = XXX
               END IF
C
C  Did we reach our target value ?
               IF (ITCONV(VAL).NE.0) THEN
                  INFO = 1
                  RETURN
               END IF
               IF (ICOUNT.GE.NOCAL) THEN
                  INFO = 2
                  RETURN
               END IF
C
            ELSE
C
C  Failure at the second try.
               INDE(I) = 0
C
C  The following expresion approximates the ratio of the
C  first to second partial derivatives.
               IF ((FP1-FP2).NE.0.) THEN
                  STEP(I)=-.5*(H2*FP1+H1*FP2)/(FP1-FP2)
               ELSE
                  SIGN = 1
                  IF ((H2*FP1+H1*FP2)*(H1+H2).LT.0.) SIGN = -1
                  STEP(I) = -SIGN*ABS(STEP(I))/2.
               END IF
            END IF
         END IF
5     CONTINUE
C
C  End of the loop over all variables.
C
C  Did we reach our target value ?
      IF (ITCONV(VAL).NE.0) THEN
         INFO = 1
         RETURN
      END IF
      IF (ICOUNT.GE.NOCAL) THEN
         INFO = 2
         RETURN
      END IF
C
      DO 7,I=1,NOD
         XPO(I) = POINT(I)
         IF (IMIN.EQ.0) THEN
C
C  No progress has been achieved.
            XT2(I) = STEP(I)
         ELSE
C
C  Progress has been achieved in the variable loop.
            XT2(I) = STEP(I)*INDE(I)
         END IF
7     CONTINUE
C
C  Perform a line search along the direction stored in XT2.
      OLDV = VAL
      CALL MINDIR(XT2,NOCMIN,ALPHA,ACCU,POINT,VAL,NOD,INFO,MARG,
     &            XLL,XRL,IXAT,XT1,FTARG,LSNOC)
      ICOUNT = ICOUNT+NOCMIN
C
      IF (VAL.LT.OLDV) THEN
         IF (IPRINT.EQ.1 .AND. STRO) WRITE (IUOUT,9) ESC, ESC, ESC
         IF (IPRINT.EQ.2 .AND. STRO) WRITE (IUOUT,9) ESC, ESC, ESC
         CALL MDIS(POINT,VAL,ITER,ICOUNT,NOCAL,IPRINT,0)
      END IF
C
C  Did we reach our target value ?
      IF (ITCONV(VAL).NE.0) THEN
         INFO = 1
         RETURN
      END IF
      IF (ICOUNT.GE.NOCAL) THEN
         INFO = 2
         RETURN
      END IF
C
C  Check convergence criteria for cycle failures.
C  Check for F-convergence.
      IF (IFCONV(VAL,OLDVAL,FTOL).NE.0) THEN
         IFAIL = IFAIL+1
      ELSE
         IFAIL = 0
      END IF
C
      IF (IFAIL.GT.IFM) THEN
         INFO = 3
         RETURN
      END IF
      OLDVAL = VAL
      GOTO 1000
C  ---------------------------------------------------------------------
C  End of ROLL iterations.
C  ---------------------------------------------------------------------
C
9     FORMAT (30X,A,'/ -- ',A,'/ -- ',A,'/  Line-search')
      END
C  ---------------------------------------------------------------------
      SUBROUTINE ROLLTM ( INFO, MESS )
C  ---------------------------------------------------------------------
C
C  Description:
C    Constructs a message, describing the reason for termination of the
C    ROLL method.
C
C  Input arguments:
C    INFO       Reason for termination of the ROLL method, as returned
C               by subroutine HYPER.
C
C  Output arguments:
C    MESS       Message, describing the reason for termination.
C
C  ---------------------------------------------------------------------
C  Arguments:
      CHARACTER*(*) MESS
C  ---------------------------------------------------------------------
C
      IF (INFO.EQ.1) THEN
         MESS = 'Target value has been reached'
      ELSE IF (INFO.EQ.2) THEN
         MESS = 'All function evaluations have been used'
      ELSE IF (INFO.EQ.3) THEN
         MESS = 'Too many cycle failures'
      ELSE
         CALL MABORT('Incorrect "info".','ROLLTM')
      END IF
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE SIMPLX ( V, MAXW, MERR )
C  ---------------------------------------------------------------------
C
C  Description:
C    Implements the SIMPLEX command.
C    Syntax: SIMPLEX (Has panel)
C
C  Input arguments:
C    MAXW       Number of available storage positions in array V.
C
C  Output arguments:
C    MERR       Error indicator.
C
C  Input / Output arguments:
C    V          Storage for the simplex vertices. On input may contain
C               the vertices from a previous run. On output contains
C               the current simplex vertices.
C
C  ---------------------------------------------------------------------
      IMPLICIT $(TYPE) (A-H,O-Z)
C  Arguments:
      DIMENSION V(MAXW)
C  ---------------------------------------------------------------------
C  Common blocks:
      PARAMETER ( MXV = $(MXV) )
      COMMON /PARAMS/ POINT(MXV), VAL
      COMMON /DETAIL/ IXAT(MXV), MARG(MXV)
      COMMON /XLIMS/ XLL(MXV), XRL(MXV)
      COMMON /DIMEN/ NOD, NOF
      COMMON /TOTCAL/ ITNOCS, IPARTC, ITDER, IPDER, ITHESS, IPHESS,
     &                ITJAC, IPJAC
      COMMON /MODE3/ MODEV, IFFORM, INMCL
      COMMON /ACCURA/ COMACC
      COMMON /RWORK/ XT1(MXV), XT2(MXV), XT3(MXV), XT4(MXV), XT5(MXV),
     &               XT6(MXV), XT7(MXV+1)
C  ---------------------------------------------------------------------
C  Local variables:
      CHARACTER*80 MESS
      SAVE IFIRST
      DATA IFIRST / 0 /
C  ---------------------------------------------------------------------
C
C  Initialize panel parameters that are machine/configuration dependent
C  and cannot be set in the panel description file.
      IF (IFIRST.EQ.0) THEN
         CALL SETPR('XTOL',COMACC,IERR)
         IFIRST = 1
      END IF
C
C  Did the user supply enough storage ?
      NEED = (NOD*(NOD+1))
      CALL VSIM('N**2',MAXW,NEED,MERR)
      IF (MERR.NE.0) RETURN
C
C  Set default values for some panel parameters and process any
C  changes the user enters.
      IUV = 0
      CALL SETPI('USEV',IUV,IERR)
      CALL SETPI('CANCEL',1,IERR)
      CALL CHANGE(MERR)
      IF (MERR.NE.0) RETURN
      CALL GETPI('CANCEL',ICANC)
      IF (ICANC.NE.1 .OR. ITNOCS.EQ.0 .OR. MODEV.EQ.0) RETURN
C
C  Get parameters from the panel.
      CALL GETPI('NOC',NOC)
      CALL GETPI('INIT',INI)
      CALL GETPI('ITER',ITMAX)
      CALL GETPR('FTOL',FTOL)
      CALL GETPR('ALPHA',ALPHA)
      CALL GETPR('BETA',BETA)
      CALL GETPR('GAMMA',GAMMA)
      CALL GETPI('PRINT',IPRINT)
      CALL GETPR('XTOL',XTOL)
      CALL GETPI('USEV',IUV)
      CALL GETPR('DISP',XCONST)
      CALL GETPI('LSNOC',LSNOC)
      CALL GETPR('LSTOL',TOLS)
C
C  Call the minimizer.
      CALL SIMPMV(V,MXV,NOD,POINT,VAL,ALPHA,BETA,GAMMA,FTOL,NOCEX,ITER,
     &            ITMAX,INFO,NOC,INI,XTOL,IXAT,MARG,XLL,XRL,IPRINT,XT1,
     &            XT2,XT3,XT4,XT5,XT7,COMACC,IUV,XCONST,LSNOC,TOLS)
C
C  Construct a message describing the reason for termination.
      CALL SIMPTM(INFO,MESS)
C
C  Report the reason for termination.
      NOGEX = -1
      NOJAC = -1
      CALL MINREP(MESS,NOCEX,NOGEX,NOJAC,ITER)
C
C  Add backup record.
      CALL BACKUP('M')
C
C  Assign values to MCL-specific keywords.
      CALL SETPI('FCALLS',NOCEX,IERR)
      CALL SETPI('ITERDONE',ITER,IERR)
      CALL SETPI('INFO',INFO,IERR)
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE SIMPMV ( V, LD, NOD, POINT, VAL, ALPHA, BETA, GAMMA,
     &                    FTOL, NOCEX, ITER, ITMAX, INFO, NOMAXC, INI,
     &                    XTOL, IXAT, MARG, XLL, XRL, IPRINT, OLDPOI,
     &                    C, VR, VE, WORK, F, COMACC, IUV, XCONST,
     &                    LSNOC, TOLS )
C  ---------------------------------------------------------------------
C
C  Description:
C    Implements the simplex method. All fixed variables are removed,
C    reducing thus the dimensionality of the problem. Margins are
C    handled explicitly during Reflection/Contraction/Expansion.
C    For more information see:
C      Nedler J.A., Mead R.
C      A simplex method for function minimization
C      Computer J., 7 (1965) 308-313.
C    A minor (and to an extent, obvious) modification is described in:
C      Mika K., Chaves Th.
C      A detailed survey of numerical methods for unconstrained
C      minimization
C      Kernforschungsanlage, Julich, GmbH, 1980.
C
C  Input arguments:
C    LD         Leading dimension of matrix V.
C    NOD        Dimensionality of the objective function.
C    ALPHA      The reflection factor.
C    BETA       The contraction factor.
C    GAMMA      The expansion factor.
C    FTOL       F-convergence criterion.
C    ITMAX      Maximum allowed number of simplex iterations.
C               Set ITMAX = -1 for an unlimited number of iterations.
C    NOMAXC     Maximum allowed number of calls to the objective
C               function.
C    INI        Determines the construction of the initial simplex:
C               INI = 1 -> Use the constant displacement method.
C               INI = 2 -> Perform a line search along each direction.
C    XTOL       X-convergence criterion.
C    IXAT       Fix statuses.
C    MARG       Margin statuses.
C    XLL        Left margins.
C    XRL        Right margins.
C    IPRINT     Printout level:
C               IPRINT = 0 -> Nothing is printed.
C               IPRINT = 1 -> Simple printout: Prints ITER, VAL,
C                             NOC, MAXNOC
C               IPRINT = 2 -> Full printout. Prints in addition the
C                             minimization parameters, contained in
C                             array POINT.
C    OLDPOI     Work array of length NOD.
C    C          Work array of length NOD.
C    VR         Work array of length NOD.
C    VE         Work array of length NOD.
C    WORK       Work array of length NOD.
C    F          Work array of length NOD+1.
C    COMACC     Machine accuracy.
C    IUV        IUV = 0 -> Reaclculate the vertices according to INI.
C               IUV = 1 -> Use the vertices already present in array V.
C    XCONST     The displacement factor, used for calculating the
C               vertices when INI=1
C    LSNOC      Number of calls for the line search, in case INI=2.
C    TOLS       Line search termination tolerance, in case INI=2.
C
C  Output arguments:
C    NOCEX      Number of function evaluations that were performed.
C    ITER       Number of SIMPLEX iterations that were performed.
C    INFO       Reason for termination:
C               INFO =  1 -> The function convergence criterion
C                            FTOL, is satisfied.
C                       2 -> Iterations exhausted.
C                       3 -> Function calls exhausted.
C                       4 -> Target value has been reached.
C                       5 -> X-convergence has been achieved.
C                       6 -> The simplex has become too small.
C                       7 -> All variables are fixed.
C                       8 -> Further progress is not possible.
C
C  Input / Output arguments:
C    V          Simplex vertices.
C    POINT      On input, the current point. On output, the new point
C               after the minimization.
C    VAL        Value of the objective function.
C
C  ---------------------------------------------------------------------
      IMPLICIT $(TYPE) (A-H,O-Z)
C  Arguments:
      DIMENSION V(LD,NOD+1), F(NOD+1)
      DIMENSION POINT(NOD), IXAT(NOD), MARG(NOD), XLL(NOD), XRL(NOD)
      DIMENSION OLDPOI(NOD), C(NOD), VR(NOD), VE(NOD), WORK(NOD)
C  ---------------------------------------------------------------------
C
C  Number of function calls so far.
      NOCEX = 0
C
C  Number of iterations performed so far.
      ITER = 0
C
C  Did we reach our target value ?
      IF (ITCONV(VAL).NE.0) THEN
         INFO = 4
         RETURN
      END IF
C
C  Check for non-fixed variables.
      CALL FREEVA(NFREE)
      IF (NFREE.EQ.0) THEN
         INFO = 7
         RETURN
      END IF
C
C  M is the number of vertices: One more than the free variables.
      M = NFREE+1
C
C  Initialize the simplex. (VR, VE are used as work arrays)
      CALL VINI(V,LD,NOD,M,POINT,VAL,F,IUV,NOCI,IXAT,MARG,XLL,XRL,INI,
     &          XCONST,LSNOC,TOLS,VR,VE,WORK)
      NOCEX = NOCEX+NOCI
C
C  Save current point and value.
      DO 99,I=1,NOD
99       OLDPOI(I) = POINT(I)
      OLDVAL = VAL
C  ---------------------------------------------------------------------
C  The following loop performs SIMPLEX iterations.
C  ---------------------------------------------------------------------
1000  CONTINUE
C
C  Calculate lowest/highest/second highest function values.
      CALL VHILO(F,M,ILO,IHI,ISHI,FLO,FHI,FSHI)
      IF (FLO.LT.OLDVAL)
     &    CALL MDIS(V(1,ILO),FLO,ITER,NOCEX,NOMAXC,IPRINT,0)
C
C  Test for termination.
      IF (ITMAX.NE.-1 .AND. ITER.GE.ITMAX) THEN
         CALL VTERM(V,LD,NOD,M,F,ILO,POINT,VAL)
         INFO = 2
         RETURN
      END IF
      IF (NOCEX.GE.NOMAXC) THEN
         CALL VTERM(V,LD,NOD,M,F,ILO,POINT,VAL)
         INFO = 3
         RETURN
      END IF
C
C  Did we reach our target value ?
      IF (ITCONV(FLO).NE.0) THEN
         CALL VTERM(V,LD,NOD,M,F,ILO,POINT,VAL)
         INFO = 4
         RETURN
      END IF
C
C  Check whether the simplex has become too small.
      CALL VXCONV(V,LD,NOD,M,XTOL,IXAT,IOK)
      IF (IOK.EQ.1) THEN
         CALL VTERM(V,LD,NOD,M,F,ILO,POINT,VAL)
         INFO = 6
         RETURN
      END IF
C
C  Check for X-convergence.
      IF (FLO.LT.OLDVAL) THEN
         IF (IXCONV(V(1,ILO),OLDPOI,NOD,XTOL).NE.0) THEN
            CALL VTERM(V,LD,NOD,M,F,ILO,POINT,VAL)
            IF (XTOL.GT.COMACC) THEN
               INFO = 5
            ELSE
               INFO = 8
            END IF
            RETURN
         END IF
C
C  Keep current minimum for further tests.
         DO 98,I=1,NOD
98          OLDPOI(I) = V(I,ILO)
         OLDVAL = FLO
      END IF
C
C  Check the simplex termination criterion.
      CALL VFCONV(F,M,FTOL,ITERM)
      IF (ITERM.NE.0) THEN
         CALL VTERM(V,LD,NOD,M,F,ILO,POINT,VAL)
         INFO = 1
         RETURN
      END IF
C
      ITER = ITER+1
C
C  Calculate the centroid C of the vertex with highest function value.
      CALL VCENT(V,LD,NOD,M,IHI,IXAT,C)
C
C  Reflect the point with highest function value through the centroid C.
      CALL VNEW(NOD,C,V(1,IHI),1,ALPHA,BETA,GAMMA,VR,MC,
     &          WORK,IXAT,MARG,XLL,XRL)
C
C  Skip the rest of the algorithm if reflection was not possible due
C  to margins; go directly to contraction.
      IF (MC.EQ.1) GOTO 700
      FR = ACSQ(VR,NOD)
      NOCEX = NOCEX+1
C
      IF (FR.LT.FLO) THEN
C  The reflection has produced a new minimum. Expand further.
         CALL VNEW(NOD,C,VR,2,ALPHA,BETA,GAMMA,VE,MC,
     &                WORK,IXAT,MARG,XLL,XRL)
         IF (MC.EQ.0) THEN
            FE = ACSQ(VE,NOD)
            NOCEX = NOCEX+1
         ELSE
C  Fake value to force a failed expansion in the following IF.
            FE = FR
         END IF
         IF (FE.LT.FR) THEN
C  The expansion was successful.
            DO 20,I=1,NOD
20             V(I,IHI) = VE(I)
            F(IHI) = FE
         ELSE
C  The expansion failed.
            DO 30,I=1,NOD
30             V(I,IHI) = VR(I)
            F(IHI) = FR
         END IF
         GOTO 1000
      END IF
C
      IF (FR.GE.FLO .AND. FR.LT.FSHI) THEN
         DO 10,I=1,NOD
10          V(I,IHI) = VR(I)
         F(IHI) = FR
         GOTO 1000
      END IF
C
      IF (FR.LT.FHI) THEN
C  Replace the "highest" vertex with the new point, before contraction.
         DO 50,I=1,NOD
50          V(I,IHI) = VR(I)
         F(IHI) = FR
         FHI = FR
      END IF
C
C  Contraction. Reuse the VR array.
700   CONTINUE
      CALL VNEW(NOD,C,V(1,IHI),3,ALPHA,BETA,GAMMA,VR,MC,
     &          WORK,IXAT,MARG,XLL,XRL)
      FC = ACSQ(VR,NOD)
      NOCEX = NOCEX+1
      IF (FC.LT.FHI) THEN
C  The contraction succeded.
         DO 60,I=1,NOD
60          V(I,IHI) = VR(I)
         F(IHI) = FC
      ELSE
C  The contraction failed. Make the simplex smaller.
         DO 70,I=1,M
            IF (I.NE.ILO) THEN
               DO 80,J=1,NOD
80                IF (IXAT(J).EQ.1) V(J,I) = (V(J,I)+V(J,ILO))/2.
               F(I) = ACSQ(V(1,I),NOD)
               NOCEX = NOCEX+1
            END IF
70       CONTINUE
      END IF
C  ---------------------------------------------------------------------
C  End of SIMPLEX iterations.
C  ---------------------------------------------------------------------
      GOTO 1000
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE VINI ( V, LD, NOD, M, POINT, VAL, F, IUV, NOC, IXAT,
     &                  MARG, XLL, XRL, INI, XCONST, LSNOC, TOLS, DIR,
     &                  WORK1, WORK2 )
C  ---------------------------------------------------------------------
C
C  Description:
C    Part of the SIMPLEX minimization method. Initializes the simplex
C    making sure that the initial vertices lie inside the margins.
C
C  Input arguments:
C    LD         Leading dimension of matrix V.
C    NOD        Dimensionality of the objective function.
C    M          Number of simplex vertices.
C    POINT      The current point.
C    VAL        Value of the objective function at POINT.
C    IUV        Determines the construction of the initial simplex:
C               IUV = 0 -> Recalculate the simplex vertices.
C               IUV = 1 -> Use the simplex vertices already in V.
C    IXAT       Fix statuses.
C    MARG       Margin statuses.
C    XLL        Left margins.
C    XRL        Right margins.
C    INI        INI = 1 -> Take a constant step along each direction
C                          and use the resulting step as a vertex.
C               INI = 2 -> Perform a line search along each direction,
C                          and use the resulting minimum as a vertex.
C    XCONST     The displacement factor, used for calculating the
C               vertices when INI=1.
C    LSNOC      Number of calls for the line search, in case INI=2.
C    TOLS       Line search termination tolerance, in case INI=2.
C
C  Output arguments:
C    F          Function values at the simplex vertices.
C    NOC        Number of evaluations of the objective function that
C               were performed by this routine.
C
C  Input / Output arguments:
C    V          Simplex vertices.
C
C  ---------------------------------------------------------------------
      IMPLICIT $(TYPE) (A-H,O-Z)
C  Arguments:
      DIMENSION V(LD,M), POINT(NOD), F(M), DIR(NOD), WORK1(NOD)
      DIMENSION IXAT(NOD), MARG(NOD), XLL(NOD), XRL(NOD), WORK2(NOD)
C  ---------------------------------------------------------------------
C  Common blocks:
      LOGICAL STRO, WEAK
      COMMON /PRILOG/ STRO, WEAK
      COMMON /UNITS/ IUINP, IUOUT
C  ---------------------------------------------------------------------
C
      ONE = 1.
      NOC = 0
C
      IF (IUV.EQ.1) THEN
C  The user supplied the initial simplex vertices.
C  What if fix status or margins have changed ?
         DO 40,I=1,M
            DO 50,J=1,NOD
50             IF (IXAT(J).EQ.0) V(J,I) = POINT(J)
            F(I) = ACSQ(V(1,I),NOD)
40       CONTINUE
         NOC = M
         RETURN
      END IF
C
C  We need to recalculate the simplex vertices.
C  First vertex is the current point.
      DO 10,I=1,NOD
10       V(I,1) = POINT(I)
      F(1) = VAL
C
C  Other vertices.
      CALL GETMC(BIGNUM,DWARF)
      FTARG = -BIGNUM
      MM = 1
      DO 100,I=1,NOD
         IF (IXAT(I).EQ.0) GOTO 100
         MM = MM+1
         DO 20,J=1,NOD
            V(J,MM) = POINT(J)
20       CONTINUE
         IF (INI.EQ.1) THEN
C  Constant displacement.
            PP = POINT(I)
            MA = MARG(I)
            CONST = XCONST*MAX(ABS(PP),ONE)
            IF (MA.EQ.0 .OR. MA.EQ.-1) THEN
               VV = PP+CONST
            ELSE IF (MA.EQ.1) THEN
               VV = PP-CONST
            ELSE
               A = XLL(I)
               B = XRL(I)
               VVP = PP+CONST
               VVM = PP-CONST
               IF (VVP.GE.A .AND. VVP.LE.B) THEN
                  VV = VVP
               ELSE IF (VVM.GE.A .AND. VVM.LE.B) THEN
                  VV = VVM
               ELSE
                  D1 = A-PP
                  D2 = B-PP
                  IF (ABS(D1).GT.ABS(D2)) THEN
                     VV = (PP+A)/2.
                  ELSE
                     VV = (PP+B)/2.
                  END IF
               END IF
            END IF
            V(I,MM) = VV
            F(MM) = ACSQ(V(1,MM),NOD)
         ELSE
C  Line search across each direction.
            DO 60,L=1,NOD
               WORK1(L) = POINT(L)
               DIR(L) = 0.
60          CONTINUE
            DIR(I) = 1.
            VAL1 = VAL
            CALL MINDIR(DIR,NOCALL,ALPHA,TOLS,WORK1,VAL1,NOD,
     &                     INFO,MARG,XLL,XRL,IXAT,WORK2,FTARG,LSNOC)
            NOC = NOC+NOCALL
            V(I,MM) = POINT(I)+ALPHA
            F(MM) = VAL1
         END IF
100   CONTINUE
      IF (INI.EQ.1) NOC = NOC+M-1
      IF (INI.EQ.2 .AND. STRO) WRITE (IUOUT,80) NOC
C
80    FORMAT ($(FC)5X,I5,' calls were spent for the initial vertices.')
      END
C  ---------------------------------------------------------------------
      SUBROUTINE VHILO ( F, M, ILO, IHI, ISHI, FLO, FHI, FSHI )
C  ---------------------------------------------------------------------
C
C  Description:
C    Part of the SIMPLEX minimization method. Determines the highest,
C    lowest and second highest function values of the simplex.
C
C  Input arguments:
C    F          Function values at the simplex vertices.
C    M          Number of simplex vertices.
C
C  Output arguments:
C    ILO        Index of the vertex with the lowest function value.
C    IHI        Index of the vertex with the highest function value.
C    ISHI       Index of the vertex with the second highest function
C               value.
C    FLO        The lowest function value. FLO = F(ILO)
C    FHI        The highest function value. FHI = F(IHI)
C    FSHI       The second highest function value. FSHI = F(ISHI)
C
C  ---------------------------------------------------------------------
      IMPLICIT $(TYPE) (A-H,O-Z)
C  Arguments:
      DIMENSION F(M)
C  ---------------------------------------------------------------------
C
      ILO = 1
      IF (F(1).GT.F(2)) THEN
         IHI = 1
         ISHI = 2
      ELSE
         IHI = 2
         ISHI = 1
      END IF
      FLO = F(ILO)
      FHI = F(IHI)
      FSHI = F(ISHI)
C
      DO 10,I=2,M
         Y = F(I)
         IF (Y.LT.FLO) THEN
            ILO = I
            FLO = Y
         ELSE IF (Y.GT.FHI) THEN
            ISHI = IHI
            FSHI = FHI
            IHI = I
            FHI = Y
         ELSE IF (Y.GT.FSHI .AND. Y.LT.FHI) THEN
            ISHI = I
            FSHI = Y
         END IF
10    CONTINUE
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE VCENT ( V, LD, NOD, M, K, IXAT, C )
C  ---------------------------------------------------------------------
C
C  Description:
C    Part of the SIMPLEX minimization method. Calculates the centroid
C    of the simplex with respect to vertex K. Note that since all
C    simplex vertices lie inside the margins, the centroid will too.
C
C  Input arguments:
C    V          Simplex vertices.
C    LD         Leading dimension of matrix V.
C    NOD        Dimensionality of the objective function.
C    M          Number of simplex vertices.
C    K          The centroid is calculated with respect to this
C               vertex. Normally this is the vertex with the highest
C               function value.
C    IXAT       Fix statuses.
C
C  Output arguments:
C    C          The centroid.
C
C  ---------------------------------------------------------------------
      IMPLICIT $(TYPE) (A-H,O-Z)
C  Arguments:
      DIMENSION V(LD,M), C(NOD), IXAT(NOD)
C  ---------------------------------------------------------------------
C
      DO 10,I=1,NOD
10       C(I) = 0.
C
      DO 20,I=1,M
         IF (I.NE.K) THEN
            DO 30,J=1,NOD
30             C(J) = C(J)+V(J,I)
         END IF
20    CONTINUE
C
      DO 40,I=1,NOD
         IF (IXAT(I).EQ.1) THEN
            C(I) = C(I)/(M-1)
         ELSE
            C(I) = V(I,K)
         END IF
40    CONTINUE
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE VNEW ( NOD, C, VO, IOP, ALPHA, BETA, GAMMA, VN, MC,
     &                  WORK, IXAT, MARG, XLL, XRL )
C  ---------------------------------------------------------------------
C
C  Description:
C    Part of the SIMPLEX minimization method. Calculates new vertex VN,
C    through "generalized" reflection of the vertex VO in the
C    direction VO ---> C. C is the centroid of the simplex with respect
C    to vertex K.
C
C  Input arguments:
C    NOD        Dimensionality of the objective function.
C    C          Centroid of the simplex.
C    VO         The vertex to be reflected.
C    IOP        Determines the operation to be performed:
C               IOP = 1 -> Reflection.
C               IOP = 2 -> Expansion.
C               IOP = 3 -> Contraction.
C    ALPHA      The reflection factor.
C    BETA       The contraction factor.
C    GAMMA      The expansion factor.
C    WORK       Work array of length NOD.
C    IXAT       Fix statuses.
C    MARG       Margin statuses.
C    XLL        Left margins.
C    XRL        Right margins.
C
C  Output arguments:
C    VN         The new, "reflected", vertex.
C    MC         Indicates what the calling routine should do:
C               MC = 0 -> The calling routine must proceed normally.
C               MC = 1 -> If a reflection was requested (IOP=1) the
C                         calling routine must immediately contract,
C                         since there is no room for reflection (due
C                         to margins).
C                         If an expansion was requested (IOP=2) the
C                         calling routine must consider this as a
C                         failed expansion, since there was no room
C                         for expanding.
C
C  ---------------------------------------------------------------------
      IMPLICIT $(TYPE) (A-H,O-Z)
C  Arguments:
      DIMENSION C(NOD), VO(NOD), VN(NOD), WORK(NOD)
      DIMENSION IXAT(NOD), MARG(NOD), XLL(NOD), XRL(NOD)
C  ---------------------------------------------------------------------
C  Local variables:
      CHARACTER*10 STR
C  ---------------------------------------------------------------------
C
C  Calculate direction along which reflection/expansion/contraction will
C  take place.
      DO 88,I=1,NOD
88       WORK(I) = C(I)-VO(I)
C
C  Calculate margins along the direction in WORK.
      CALL DIRLIM(VO,WORK,MARG,XLL,XRL,NOD,LIM,XALIM,XBLIM)
C
      MC = 0
      IF (IOP.EQ.1) THEN
C
C  This is a reflection.
         F1 = 1.+ALPHA
         IF (F1.GT.XBLIM) THEN
            F1 = XBLIM
            IF (F1.LE.1.) THEN
               MC = 1
               RETURN
            END IF
         END IF
      ELSE IF (IOP.EQ.2) THEN
C
C  This is an expansion.
         F1 = 1.-GAMMA
         IF (F1.LT.XALIM) THEN
            F1 = XALIM
            IF (F1.GE.0.) THEN
               MC = 1
               RETURN
            END IF
         END IF
      ELSE IF (IOP.EQ.3) THEN
C
C  This is a contraction.
         F1 = 1.-BETA
      ELSE
         CALL I2STR(IOP,STR,LE1)
         CALL MABORT('Incorrect operation: '//STR(1:LE1),'VNEW')
      END IF
C
C  Calculate generalized reflection: VN = VO + F1 * WORK
      CALL MEVAL(VO,WORK,F1,NOD,IXAT,MARG,XLL,XRL,VN,0,FTMP)
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE VFCONV ( F, M, FTOL, ITERM )
C  ---------------------------------------------------------------------
C
C  Description:
C    Part of the SIMPLEX minimization method. Tests for termination of
C    the method, using the standard simplex criterion.
C
C  Input arguments:
C    F          Function values at the simplex vertices.
C    M          Number of simplex vertices.
C    FTOL       The simplex (F-convergence) termination criterion.
C
C  Output arguments:
C    ITERM      ITERM = 0 -> The simplex criterion is not satisfied.
C               ITERM = 1 -> The simplex criterion is satisfied and
C                            the algorithm must terminate.
C
C  ---------------------------------------------------------------------
      IMPLICIT $(TYPE) (A-H,O-Z)
C  Arguments:
      DIMENSION F(M)
C  ---------------------------------------------------------------------
C
      ONE = 1.
      S = 0.
      DO 10,I=1,M
10       S = S+F(I)
      FAVG = S/M
C
      S = 0.
      DO 20,I=1,M
20       S = S+(F(I)-FAVG)**2
      S = SQRT(S/M)/MAX(ONE,ABS(FAVG))
C
      IF (S.LE.FTOL) THEN
         ITERM = 1
      ELSE
         ITERM = 0
      END IF
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE VXCONV ( V, LD, NOD, M, XTOL, IXAT, IOK )
C  ---------------------------------------------------------------------
C
C  Description:
C    Part of the SIMPLEX minimization method. Checks whether the
C    simplex has become too small.
C
C  Input arguments:
C    V          Simplex vertices.
C    LD         Leading dimesnsion of matrix V.
C    NOD        Dimensionality of the objective function.
C    M          Number of simplex vertices.
C    XTOL       The X-convergence criterion.
C    IXAT       Fix statuses.
C
C  Output arguments:
C    IOK        IOK = 0 -> The simplex is still large enough.
C               IOK = 1 -> The simplex has become too small.
C
C  ---------------------------------------------------------------------
      IMPLICIT $(TYPE) (A-H,O-Z)
C  Arguments:
      DIMENSION V(LD,M), IXAT(NOD)
C  ---------------------------------------------------------------------
C
      ONE = 1.
      DO 10,I=1,NOD
         IF (IXAT(I).NE.0) THEN
            S = 0.
            DO 20,J=1,M
20             S = S+V(I,J)
            XAVG = S/M
C
            S = 0.
            DO 30,J=1,M
30             S = S+(V(I,J)-XAVG)**2
            S = SQRT(S/M)/MAX(ONE,ABS(XAVG))
            IF (S.GT.XTOL) THEN
               IOK = 0
               RETURN
            END IF
         END IF
10    CONTINUE
      IOK = 1
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE VTERM ( V, LD, NOD, M, F, ILO, POINT, VAL )
C  ---------------------------------------------------------------------
C
C  Description:
C    Part of the SIMPLEX minimization method. Prepares the values to be
C    returned upon termination of the simplex method.
C
C  Input arguments:
C    V          Simplex vertices.
C    LD         Leading dimension of array V.
C    NOD        Dimensionality of the objective function.
C    M          Number of simplex vertices.
C    F          Function values at the simplex vertices.
C    ILO        Index of the vertex with the lowest function value.
C
C  Output arguments:
C    POINT      The best point found by the simplex method.
C    VAL        Value of the objective function at POINT.
C
C  ---------------------------------------------------------------------
      IMPLICIT $(TYPE) (A-H,O-Z)
C  Arguments:
      DIMENSION V(LD,M), F(M), POINT(NOD)
C  ---------------------------------------------------------------------
C
      DO 10,I=1,NOD
10       POINT(I) = V(I,ILO)
      VAL = F(ILO)
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE SIMPTM ( INFO, MESS )
C  ---------------------------------------------------------------------
C
C  Description:
C    Constructs a message, describing the reason for termination of the
C    SIMPLEX method.
C
C  Input arguments:
C    INFO       Reason for termination of the simplex method, as
C               returned by subroutine SIMPMV.
C
C  Output arguments:
C    MESS       Message, describing the reason for termination.
C
C  ---------------------------------------------------------------------
C  Arguments:
      CHARACTER*(*) MESS
C  ---------------------------------------------------------------------
C
      IF (INFO.EQ.1) THEN
         MESS = 'The specified function accuracy has been achieved'
      ELSE IF (INFO.EQ.2) THEN
         MESS = 'The specified number of iterations has been reached'
      ELSE IF (INFO.EQ.3) THEN
         MESS = 'All function evaluations have been used'
      ELSE IF (INFO.EQ.4) THEN
         MESS = 'Target value has been reached'
      ELSE IF (INFO.EQ.5) THEN
         MESS = 'The X-convergence criterion is satisfied'
      ELSE IF (INFO.EQ.6) THEN
         MESS = 'The simplex has become too small'
      ELSE IF (INFO.EQ.7) THEN
         MESS = 'All variables are fixed'
      ELSE IF (INFO.EQ.8) THEN
         MESS = 'Further progress is not possible'
      ELSE
         CALL MABORT('Incorrect "info".','SIMPTM')
      END IF
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE STEPAL
C  ---------------------------------------------------------------------
C
C  Description:
C    Implements the STEPALL command.
C    Syntax: STEPALL
C
C    STEPALL performs a line search along the direction of the gradient.
C
C  ---------------------------------------------------------------------
      IMPLICIT $(TYPE) (A-H,O-Z)
C  Common blocks:
      PARAMETER ( MXV = $(MXV) )
      COMMON /PARAMS/ POINT(MXV), VAL
      COMMON /DETAIL/ IXAT(MXV), MARG(MXV)
      COMMON /XLIMS/ XLL(MXV), XRL(MXV)
      COMMON /ROSTEP/ STEP(MXV)
      COMMON /TOTCAL/ ITNOCS, IPARTC, ITDER, IPDER, ITHESS, IPHESS,
     &                ITJAC, IPJAC
      COMMON /MODE3/ MODEV, IFFORM, INMCL
      COMMON /DIMEN/ NOD, NOF
      LOGICAL AIMON
      COMMON /AIM/ TARGET, AIMON
      COMMON /BIGCOM/ BIGNUM, BLIM, DWARF, DWALOG
      COMMON /RWORK/ XT1(MXV), XT2(MXV), XT3(MXV), XT4(MXV), XT5(MXV),
     &               XT6(MXV), XT7(MXV+1)
C  ---------------------------------------------------------------------
C
      IF (ITNOCS.EQ.0 .OR. MODEV.EQ.0) RETURN
C
C  Add backup record.
      CALL BACKUP('M')
C
C  Calculate the gradient.
      CALL DERMER(POINT,VAL,XT2,NODC,NODG,0)
C
C  Make a line search along the direction of the gradient.
      XTOL = 0.0001
      LSNOC = 50
      IF (AIMON) THEN
         FTARG = TARGET
      ELSE
         FTARG = -BIGNUM
      END IF
      CALL MINDIR(XT2,NOCALL,ALPHA,XTOL,POINT,VAL,NOD,INFO,MARG,
     &            XLL,XRL,IXAT,XT1,FTARG,LSNOC)
C
C  Update the ROLL steps.
      DO 10,I=1,NOD
10       STEP(I) = ALPHA*XT2(I)
C
      END
C  ---------------------------------------------------------------------
      FUNCTION ITCONV ( VAL )
C  ---------------------------------------------------------------------
C
C  Description:
C    Checks whether the target value has been reached.
C
C  Input arguments:
C    VAL        Current value of the objective function.
C
C  Returned value:
C    ITCONV = 0 -> The target value has not been reached.
C    ITCONV = 1 -> Target value has been reached.
C
C  ---------------------------------------------------------------------
      IMPLICIT $(TYPE) (A-H,O-Z)
C  Common blocks:
      LOGICAL AIMON
      COMMON /AIM/ TARGET, AIMON
C  ---------------------------------------------------------------------
C
      ITCONV = 0
      IF (AIMON .AND. VAL.LE.TARGET) ITCONV = 1
C
      END
C  ---------------------------------------------------------------------
      FUNCTION IFCONV ( VAL, OLDVAL, EPS )
C  ---------------------------------------------------------------------
C
C  Description:
C    Checks whether the relative drop of the function value, in
C    two successive iterations of a minimization method is less than
C    EPS.
C
C  Input arguments:
C    VAL        Current value of the objective function.
C    OLDVAL     Value of the objective function at the previous
C               iteration.
C    EPS        Minimum relative drop.
C
C  Returned value:
C    IFCONV = 0 -> Relative drop greater than EPS.
C    IFCONV = 1 -> Relative drop less than or equal to EPS.
C
C  ---------------------------------------------------------------------
      IMPLICIT $(TYPE) (A-H,O-Z)
C
      ONE = 1.
      T1 = ABS(VAL-OLDVAL)
      T2 = EPS*MAX(ONE,ABS(VAL),ABS(OLDVAL))
      IFCONV = 0
      IF (T1.LE.T2) IFCONV = 1
C
      END
C  ---------------------------------------------------------------------
      FUNCTION IGCONV ( GRAD, POINT, NOD, VAL, GTOL )
C  ---------------------------------------------------------------------
C
C  Description:
C    Checks whether any component of the (scaled) gradient vector is
C    larger than GTOL. For details see:
C      Dennis J.E. & Schnabel R.B., Numerical methods for unconstrained
C      optimization and nonlinear equations.
C      Prentice-Hall, 1983. pp. 159-161, 278-280.
C
C  Input arguments:
C    GRAD       The current gradient vector.
C    POINT      The current point.
C    NOD        Dimensionality of the objective function.
C    VAL        Current value of the objective function.
C    GTOL       Tolerance for the scaled gradient.
C
C  Returned value:
C    IGCONV = 0 -> At least one scaled gradient component is larger
C                  than GTOL.
C    IGCONV = 1 -> No scaled gradient component is larger than GTOL.
C
C  ---------------------------------------------------------------------
      IMPLICIT $(TYPE) (A-H,O-Z)
C  Arguments:
      DIMENSION GRAD(NOD), POINT(NOD)
C  ---------------------------------------------------------------------
C
      ONE = 1.
      GMAX = 0.
      DO 88,I=1,NOD
         GC = ABS(MAX(ABS(POINT(I)),ONE)*GRAD(I)/MAX(ABS(VAL),ONE))
         GMAX = MAX(GMAX,GC)
88    CONTINUE
      IGCONV = 0
      IF (GMAX.LE.GTOL) IGCONV = 1
C
      END
C  ---------------------------------------------------------------------
      FUNCTION IXCONV ( POINT, OLDPOI, NOD, XTOL )
C  ---------------------------------------------------------------------
C
C  Description:
C    Checks whether the scaled distance between two successive iterates
C    of a minimization routine are considered close enough (less than
C    XTOL). For details see:
C      Dennis J.E. & Schnabel R.B., Numerical methods for unconstrained
C      optimization and nonlinear equations.
C      Prentice-Hall, 1983. pp. 159-161, 278-280.
C
C  Input arguments:
C    POINT      The current POINT.
C    OLDPOI     The point at the previous iteration of the calling
C               routine.
C    NOD        Dimensionality of the objective function.
C    XTOL       X-tolerance.
C
C  Returned value:
C    IXCONV = 0 -> At least one minimization parameter has a scaled
C                  distance > XTOL, from the previous iterate.
C    IXCONV = 1 -> Scaled distance less than XTOL for all minimization
C                  parameters.
C
C  ---------------------------------------------------------------------
      IMPLICIT $(TYPE) (A-H,O-Z)
C  Arguments:
      DIMENSION POINT(NOD), OLDPOI(NOD)
C  ---------------------------------------------------------------------
C
      ONE = 1.
      XMAX = 0.
      DO 98,I=1,NOD
         XC = ABS((POINT(I)-OLDPOI(I))/MAX(ABS(OLDPOI(I)),ONE))
         XMAX = MAX(XMAX,XC)
98    CONTINUE
      IXCONV = 0
      IF (XMAX.LE.XTOL) IXCONV = 1
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE TOLMIN ( V, MAXW, MERR )
C  ---------------------------------------------------------------------
C
C  Description:
C    Implements the TOLMIN command.
C    Syntax: TOLMIN (Has panel)
C
C  Input arguments:
C    V          Work array. Should have at least NOD**2+11*NOD elements.
C    MAXW       Number of available storage positions in array V.
C
C  Output arguments:
C    MERR       Error indicator.
C
C  ---------------------------------------------------------------------
      IMPLICIT $(TYPE) (A-H,O-Z)
C  Arguments:
      DIMENSION V(MAXW)
C  ---------------------------------------------------------------------
C  Common blocks:
      COMMON /TOTCAL/ ITNOCS, IPARTC, ITDER, IPDER, ITHESS, IPHESS,
     &                ITJAC, IPJAC
      COMMON /MODE3/ MODEV, IFFORM, INMCL
      COMMON /DIMEN/ NOD, NOF
C  ---------------------------------------------------------------------
C  Local variables:
      CHARACTER*80 MESS
      CHARACTER*10 STR
C  ---------------------------------------------------------------------
C
C  Did the user supply enough storage ?
      NEED = NOD**2+11*NOD
      CALL VSIM('N**2+11*N',MAXW,NEED,MERR)
      IF (MERR.NE.0) RETURN
C
      CALL SETPI('CANCEL',1,IERR)
      CALL CHANGE(MERR)
      IF (MERR.NE.0) RETURN
      CALL GETPI('CANCEL',ICANC)
      IF (ICANC.NE.1 .OR. ITNOCS.EQ.0 .OR. MODEV.EQ.0) RETURN
C
C  Get parameters from the panel.
      CALL GETPI('NOC',NOCT)
      CALL GETPR('ACC',ACCT)
      CALL GETPI('PRINT',IPRINT)
      CALL GETPS('MAD',STR,LES)
      IF (ISCOMP(STR,'ON').NE.0) THEN
         MAD = 1
      ELSE
         MAD = 0
      END IF
C
C  Set automatic derivatives.
      CALL SETADE(MAD)
C
C  Call the minimizer.
      IGS = ITDER
      CALL TOLI(V,NOCT,ACCT,IPRINT,NOCEX,INFO)
C
C  Construct a message describing the reason for termination.
      CALL TOLITM(INFO,MESS)
C
C  Report the reason for termination.
      ITER = -1
      NOGEX = ITDER-IGS
      NOJAC = -1
      CALL MINREP(MESS,NOCEX,NOGEX,NOJAC,ITER)
C
C  Disable automatic derivatives.
      CALL SETADE(0)
C
C  Add backup record.
      CALL BACKUP('M')
C
C  Assign values to MCL-specific keywords.
      CALL SETPI('FCALLS',NOCEX,IERR)
      CALL SETPI('GCALLS',NOGEX,IERR)
      CALL SETPI('INFO',INFO,IERR)
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE TOLI ( W, NOCMAX, ACC, IPRINT, NOCEXC, INFO )
C  ---------------------------------------------------------------------
C
C  Description:
C    Front end to the TOLMIN minimization package.
C
C    For more information see:
C
C    1) Powell M.J.D.,
C       TOLMIN: A Fortran package for linearly constrained
C       optimization calculations,
C       DAMTP 1989/NA2, Dept. of Applied Mathematics and Theoretical
C       Physics, Univ. of Cambridge, 1989.
C    2) Powell M.J.D.,
C       A tolerant algorithm for linearly constrained optimization
C       calculations,
C       Math. Prog., 45 (1989) 547-566.
C
C  Input arguments:
C    W          Work array
C    NOCMAX     Maximum allowed number of function evaluations.
C    ACC        Required accuracy of the solution.
C    IPRINT     Printout level.
C
C  Output arguments:
C    NOCEXC     Number of function evaluations that were performed.
C    INFO       Reason for termination, as returned by subroutine
C               GETMIN.
C
C  ---------------------------------------------------------------------
      IMPLICIT $(TYPE) (A-H,O-Z)
C  Arguments:
      DIMENSION W(*)
C  ---------------------------------------------------------------------
C  Common blocks:
      PARAMETER ( MXV = $(MXV) )
      COMMON /PARAMS/ POINT(MXV), VAL
      COMMON /DETAIL/ IXAT(MXV), MARG(MXV)
      COMMON /XLIMS/ XLL(MXV), XRL(MXV)
      COMMON /DIMEN/ NOD, NOF
      COMMON /BIGCOM/ BIGNUM, BLIM, DWARF, DWALOG
      COMMON /TOTCAL/ ITNOCS, IPARTC, ITDER, IPDER, ITHESS, IPHESS,
     &                ITJAC, IPJAC
      COMMON /RWORK/ XT1(MXV), XT2(MXV), XT3(MXV), XT4(MXV), XT5(MXV),
     &               XT6(MXV), XT7(MXV+1)
C  ---------------------------------------------------------------------
C  Local variables:
      DIMENSION B(1)
      DIMENSION IACT(2*MXV)
C  ---------------------------------------------------------------------
C
C  Prepare upper-lower bounds.
      DO 10,I=1,NOD
         IF (IXAT(I).EQ.0) THEN
            XT1(I) = POINT(I)
            XT2(I) = POINT(I)
         ELSE
            MA = MARG(I)
            IF (MA.EQ.-1) THEN
               XT1(I) = XLL(I)
               XT2(I) = BLIM
            ELSE IF (MA.EQ.1) THEN
               XT1(I) = -BLIM
               XT2(I) = XRL(I)
            ELSE IF (MA.EQ.2) THEN
               XT1(I) = XLL(I)
               XT2(I) = XRL(I)
            ELSE IF (MA.EQ.0) THEN
               XT1(I) = -BLIM
               XT2(I) = BLIM
            END IF
         END IF
10    CONTINUE
C
      INFO = NOCMAX
      M = 0
      MEQ = 0
      IA = MXV
      INICAL = ITNOCS
      CALL GETMIN(NOD,M,MEQ,XT3,IA,B,XT1,XT2,POINT,ACC,IACT,NACT,XT4,
     &            IPRINT,INFO,W)
      VAL = ACSQ(POINT,NOD)
      NOCEXC = ITNOCS-INICAL
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE FGCALC ( N, X, F, G )
C  ---------------------------------------------------------------------
C
C  Description:
C    This is the function that the TOLMIN package minimizes.
C
C  Input arguments:
C    N          Dimensionality of the objective function.
C    X          Point at which the objective function must be evaluated.
C
C  Output arguments:
C    F          Value of the objective function at X.
C    G          The gradient vector at X.
C
C  ---------------------------------------------------------------------
      IMPLICIT $(TYPE) (A-H,O-Z)
C  Arguments:
      DIMENSION X(N), G(N)
C  ---------------------------------------------------------------------
C
      F = ACSQ(X,N)
      CALL DERMER(X,F,G,NOC,NGR,0)
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE TOLITM ( INFO, MESS )
C  ---------------------------------------------------------------------
C
C  Description:
C    Constructs a message, describing the reason for termination of the
C    TOLMIN method.
C
C  Input arguments:
C    INFO       Reason for termination of the TOLMIN method, as returned
C               by subroutine GETMIN.
C
C  Output arguments:
C    MESS       Message, describing the reason for termination.
C
C  ---------------------------------------------------------------------
C  Arguments:
      CHARACTER*(*) MESS
C  ---------------------------------------------------------------------
C
      IF (INFO.EQ.1) THEN
         MESS = 'Accuracy satisfied'
      ELSE IF (INFO.EQ.2) THEN
         MESS = 'Further progress is impossible, '//
     &             'due to rounding errors'
      ELSE IF (INFO.EQ.3) THEN
         MESS = 'Further progress is impossible'
      ELSE IF (INFO.EQ.4) THEN
C  This should never happen.
         MESS = 'Some input parameters are incorrect'
      ELSE IF (INFO.EQ.5) THEN
C  This should never happen.
         MESS = 'Incosistemt equality constraints'
      ELSE IF (INFO.EQ.6) THEN
C  This should never happen.
         MESS = 'Inconsistent constrains/bounds'
      ELSE IF (INFO.EQ.7) THEN
C  This should never happen.
         MESS = 'All of the constrains cannot be satified'
      ELSE IF (INFO.EQ.8) THEN
         MESS = 'All function evaluations have been used'
      ELSE IF (INFO.EQ.9) THEN
         MESS = 'Target value has been reached'
      ELSE
         CALL MABORT('Incorrect "info".','TOLITM')
      END IF
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE TRUST ( V, MAXW, MERR )
C  ---------------------------------------------------------------------
C
C  Description:
C    Implements the TRUST command.
C    Syntax: TRUST (Has panel)
C
C  Input arguments:
C    MAXW       Number of available storage positions in array V.
C
C  Output arguments:
C    MERR       Error indicator.
C
C  Input / Output arguments:
C    V          On input the upper triangular part of V may contain an
C               initial approximation to the Choleski L[t] factor of
C               the Hessian matrix. On output contains the upper
C               triangular part contains the Choleski L[t] of the
C               BFGS approximation to the Hessian.
C
C  ---------------------------------------------------------------------
      IMPLICIT $(TYPE) (A-H,O-Z)
C  Arguments:
      DIMENSION V(*)
C  ---------------------------------------------------------------------
C  Common blocks:
      PARAMETER ( MXV = $(MXV) )
      COMMON /PARAMS/ POINT(MXV), VAL
      COMMON /DETAIL/ IXAT(MXV), MARG(MXV)
      COMMON /DIMEN/ NOD, NOF
      COMMON /TOTCAL/ ITNOCS, IPARTC, ITDER, IPDER, ITHESS, IPHESS,
     &                ITJAC, IPJAC
      COMMON /MODE3/ MODEV, IFFORM, INMCL
      COMMON /UPDATE/ IGRADU, IHESSU
      COMMON /ACCURA/ COMACC
      COMMON /RWORK/ XT1(MXV), XT2(MXV), XT3(MXV), XT4(MXV), XT5(MXV),
     &               XT6(MXV), XT7(MXV+1)
C  ---------------------------------------------------------------------
C  Local variables:
      DIMENSION XTMP(MXV)
      CHARACTER*80 MESS
      CHARACTER*10 STR
      SAVE IFIRST
      DATA IFIRST / 0 /
C  ---------------------------------------------------------------------
C
C  Initialize panel parameters that are machine/configuration dependent
C  and cannot be set in the panel description file.
      IF (IFIRST.EQ.0) THEN
         CALL SETPR('XTOL',COMACC,IERR)
         CALL SETPR('GTOL',COMACC,IERR)
         IFIRST = 1
      END IF
C
C  Did the user supply enough storage ?
      NEED = NOD**2
      CALL VSIM('N**2',MAXW,NEED,MERR)
      IF (MERR.NE.0) RETURN
C
C  Set default values for some panel parameters and process any
C  changes the user enters.
      IG = IGRADU
      IH = IHESSU
      CALL SETPI('USEG',IG,IERR)
      CALL SETPI('USEH',IH,IERR)
      CALL SETPI('USER',0,IERR)
      CALL SETPI('CANCEL',1,IERR)
      CALL CHANGE(MERR)
      IF (MERR.NE.0) RETURN
      CALL GETPI('CANCEL',ICANC)
      IF (ICANC.NE.1 .OR. ITNOCS.EQ.0 .OR. MODEV.EQ.0) RETURN
C
C  Get parameters from the panel.
      CALL GETPI('NOC',NOC)
      CALL GETPI('USEG',IGRADU)
      CALL GETPI('USEH',IHESSU)
      CALL GETPI('USER',IRADU)
      CALL GETPI('PRINT',IPRINT)
      CALL GETPR('FTOL',FTOL)
      CALL GETPR('XTOL',XTOL)
      CALL GETPR('GTOL',GTOL)
      CALL GETPR('RHO',RHO)
      CALL GETPI('ITER',ITMAX)
      CALL GETPR('RADIUS',DELTA)
      IF (DELTA.EQ.0) IRADU = 0
      CALL GETPS('MAD',STR,LES)
      IF (ISCOMP(STR,'ON').NE.0) THEN
         MAD = 1
      ELSE
         MAD = 0
      END IF
C
C  Set automatic derivatives.
      CALL SETADE(MAD)
C
C  Transform the variables.
      DO 10,I=1,NOD
         XTMP(I) = POINT(I)
10    CONTINUE
      CALL XTOY(XTMP,NOD)
C
C  Call the minimizer.
      CALL TRUSMV(V,XTMP,XT1,NOD,VAL,NOC,COMACC,GTOL,ITMAX,XTOL,
     &            RHO,FTOL,INFO,ITER,NOCEX,NOGEX,IPRINT,IXAT,IG,
     &            IH,XT2,XT3,XT4,XT5,XT6,IRADU,DELTA,MARG)
C
C  Inverse transformation.
      CALL YTOX(XTMP,NOD)
      DO 20,I=1,NOD
         POINT(I) = XTMP(I)
20    CONTINUE
      VAL = ACSQ(POINT,NOD)
C
C  Construct a message describing the reason for termination.
      CALL TRTM(INFO,MESS)
C
C  Report the reason for termination.
      NOJAC = -1
      CALL MINREP(MESS,NOCEX,NOGEX,NOJAC,ITER)
C
C  Disable automatic derivatives.
      CALL SETADE(0)
C
C  Add backup record.
      CALL BACKUP('M')
C
C  Assign values to MCL-specific keywords.
      CALL SETPI('FCALLS',NOCEX,IERR)
      CALL SETPI('GCALLS',NOGEX,IERR)
      CALL SETPI('ITERDONE',ITER,IERR)
      CALL SETPI('INFO',INFO,IERR)
      CALL SETPR('RADIUS',DELTA,IERR)
C
      END
C  ---------------------------------------------------------------------
      FUNCTION FCNT ( X, N )
C  ---------------------------------------------------------------------
C
C  Description:
C    This is the objective function that subroutine TRUSMV minimizes.
C
C  Input arguments:
C    X          The minimization variables (transformed).
C    N          Number of variables.
C
C  ---------------------------------------------------------------------
      IMPLICIT $(TYPE) (A-H,O-Z)
C  Arguments:
      DIMENSION X(N)
C  ---------------------------------------------------------------------
C  Local variables:
      PARAMETER ( MXV = $(MXV) )
      DIMENSION XTMP(MXV)
C  ---------------------------------------------------------------------
C
C  Back-transform the variables.
      DO 10,I=1,N
10       XTMP(I) = X(I)
      CALL YTOX(XTMP,N)
      FCNT = ACSQ(XTMP,N)
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE TRUSMV ( H, POINT, GRAD, NOD, VAL, NOMAXC, COMACC,
     &                    GTOL, ITMAX, XTOL, RHO, FTOL, INFO, ITER,
     &                    NOCEX, NOGEX, IPRINT, IXAT, IGRADU, IHESSU,
     &                    XS, GN, Y, GS, W, IRADU, DELTA, MARG )
C  ---------------------------------------------------------------------
C
C  Description:
C    Implements the BFGS method, taking the trust region approach,
C    in place of the line search. Uses Choleski factorization
C    of the Hessian. The Choleski factor L[t] is stored in the upper
C    triangular part of matrix H. Fixed variables are removed, thus
C    reducing the dimensionality of the problem. Margins are treated
C    by appropriate transformations.
C    For more information see:
C      Dennis J.E. & Schnabel R.B.
C      Numerical methods for unconstrained optimization and
C      nonlinear equations
C      Prentice Hall, Englewood Cliffs, New Jersey, 1983.
C
C  Input arguments:
C    GRAD       Work array of length NOD.
C    NOD        Dimensionality of the objective function.
C    NOMAXC     Maximum allowed number of calls to the objective
C               function.
C    COMACC     Machine accuracy.
C    GTOL       G-convergence criterion.
C    ITMAX      Maximum allowed number of BFGS iterations.
C               Set ITMAX = -1 for an unlimited number.
C    XTOL       X-convergence criterion.
C    RHO        Parameter that defines the rho-line.
C    FTOL       F-convergence criterion.
C    IPRINT     Printout level:
C               IPRINT = 0 -> Nothing is printed.
C               IPRINT = 1 -> Simple printout: Prints ITER, VAL,
C                             NOC, MAXNOC
C               IPRINT = 2 -> Full printout. Prints in addition the
C                             minimization parameters, contained in
C                             array POINT.
C    IXAT       Fix statuses.
C    IGRADU     IGRADU = 0 -> Ask subroutine DERMER to recalculate
C                             the gradient at the current point. If
C                             a valid cached gradient exists, DERMER
C                             will return it.
C               IGRADU = 1 -> Ask subroutine DERMER to return its
C                             cached gradient. No recalculation
C                             takes place, unless of course the
C                             gradient cache is empty.
C    XS         Work array of length NOD.
C    GN         Work array of length NOD.
C    Y          Work array of length NOD.
C    GS         Work array of length NOD.
C    W          Work array of length NOD.
C    IRADU      IRADU = 0 -> Initialize the trust region radius.
C               IRADU = 1 -> Use the radius from the previous run.
C    MARG       Margin statuses.
C
C  Output arguments:
C    INFO       Reason for termination:
C               INFO = 1 -> Target value has been reached.
C               INFO = 2 -> The gradient criterion is satisfied.
C               INFO = 3 -> All function calls have been exhausted.
C               INFO = 4 -> X-convergence has been achieved.
C               INFO = 5 -> F-convergence has occurred.
C               INFO = 6 -> Division by zero while updating H.
C               INFO = 7 -> Iterations exhausted.
C               INFO = 8 -> All variables are fixed.
C               INFO = 9 -> Further progress not possible.
C    ITER       Number of BFGS iterations that were performed.
C    NOCEX      Number of function evaluations that were performed.
C    NOGEX      Number of gradient evaluations that were performed.
C    DELTA      The trust region radius.
C
C  Input / Output arguments:
C    H          On input the upper triangular part of H may contain an
C               initial approximation to the Choleski L[t] factor of
C               the Hessian matrix. On output contains the upper
C               triangular part contains the Choleski L[t] of the
C               BFGS approximation to the Hessian.
C    POINT      On input, the current point. On output, the new point
C               after the minimization.
C    VAL        Value of the objective function.
C
C  ---------------------------------------------------------------------
      IMPLICIT $(TYPE) (A-H,O-Z)
C  Arguments:
      DIMENSION H(NOD,NOD), POINT(NOD), GRAD(NOD)
      DIMENSION IXAT(NOD), MARG(NOD)
      DIMENSION XS(NOD), GN(NOD), Y(NOD), GS(NOD), W(NOD)
      EXTERNAL FCNT
C  ---------------------------------------------------------------------
C  Local variables:
      PARAMETER ( MXV = $(MXV) )
      DIMENSION XTMP(MXV)
C  ---------------------------------------------------------------------
C
C  Number of iterations performed so far.
      ITER = 0
C
C  Number of function calls so far.
      NOCEX = 0
C
C  Number of (analytic) gradient calls so far.
      NOGEX = 0
C
C  Did we reach our target value ?
      IF (ITCONV(VAL).NE.0) THEN
         INFO = 1
         RETURN
      END IF
C
C  Count non-fixed variables. M will be used as the dimensionality
C  of the objective function in all subsequent matrix/vector operations.
      CALL FREEVA(M)
      IF (M.EQ.0) THEN
         INFO = 8
         RETURN
      END IF
C
C  Calculate derivatives at the starting point.
      IF (IGRADU.EQ.0) THEN
         MODE = 0
      ELSE
         MODE = 2
      END IF
      DO 30,I=1,NOD
30       XTMP(I) = POINT(I)
      CALL YTOX(XTMP,NOD)
      CALL DERMER(XTMP,VAL,GRAD,N1,N2,MODE)
      CALL GTRANS(GRAD,POINT)
      NOCEX = NOCEX+N1
      NOGEX = NOGEX+N2
C
C  Copy derivatives to an array of M elements.
      MM = 0
      DO 17,I=1,NOD
         IF (IXAT(I).NE.0) THEN
            MM = MM+1
            GS(MM) = GRAD(I)
         END IF
17    CONTINUE
C
C  Check the gradient criterion for the initial point.
C  The numerical differentiation routine DERMER is expected to
C  return GRAD(I) = 0 for all fixed variables. Otherwise the
C  following test will fail, since the whole array GRAD is examined.
      IF (IGCONV(GRAD,POINT,NOD,VAL,GTOL).NE.0) THEN
         INFO = 2
         RETURN
      END IF
C
C  Initialize Hessian. (Upper triangular part only.)
      IF (IHESSU.EQ.0) THEN
         DO 10,J=1,M
            H(J,J) = 1.
            DO 20,I=1,J-1
20             H(I,J) = 0.
10       CONTINUE
      END IF
C
C  Check if margins are present.
      NMA = 0
      DO 60,I=1,NOD
         IF (MARG(I).NE.0) NMA = NMA+1
60    CONTINUE
      IF (NMA.NE.0) THEN
C  Transform the Hessian to new coordinates.
         CALL LTRANS(H,NOD,POINT)
C  Compose the Hessian from its Choleski factors.
         CALL LLCOMP(H,NOD,NOD)
C  Decompose the Hessian forcing possitive definiteness.
         CALL HPDD(H,XTMP,NOD,NOD,COMACC,ISPD)
      END IF
C
C  Prepare an initial trust region radius.
      IF (IRADU.EQ.0) THEN
         CALL VVMULT(GS,GS,DELTA,NOD)
         DELTA = SQRT(DELTA)/10.
      END IF
C  ---------------------------------------------------------------------
C  The following loop performs TRUST iterations.
C  ---------------------------------------------------------------------
1000  CONTINUE
      IF (ITMAX.NE.-1 .AND. ITER.GE.ITMAX) THEN
         INFO = 7
         RETURN
      END IF
      IF (NOCEX.GE.NOMAXC) THEN
         INFO = 3
         RETURN
      END IF
      ITER = ITER+1
C
C  Save current point and value for the convergence tests.
      OLDVAL = VAL
      DO 99,I=1,NOD
99       XS(I) = POINT(I)
C
C  Calculate the next point and update the trust region.
      CALL TTDRI(DELTA,M,GS,H,NOD,RHO,VAL,XTOL,N1,POINT,NOD,IXAT,GN,
     &           Y,W,IRET,FCNT)
      NOCEX = NOCEX+N1
C
C  Any progress ? (IRET = 1, means the step taken was too small.)
      IF (IRET.EQ.1 .AND. XTOL.LE.COMACC) THEN
         INFO = 9
         RETURN
      END IF
      IF (VAL.LT.OLDVAL)
     &    CALL MDIS(POINT,VAL,ITER,NOCEX,NOMAXC,IPRINT,1)
C
C  Did we reach our target value ?
      IF (ITCONV(VAL).NE.0) THEN
         INFO = 1
         RETURN
      END IF
C
C  Check for X-convergence.
      IF (IXCONV(POINT,XS,NOD,XTOL).NE.0) THEN
         INFO = 4
         RETURN
      END IF
C
C  Check for F-convergence.
      IF (IFCONV(VAL,OLDVAL,FTOL).NE.0) THEN
         INFO = 5
         RETURN
      END IF
C
C  Calculate derivatives at the new point.
      DO 50,I=1,NOD
50       XTMP(I) = POINT(I)
      CALL YTOX(XTMP,NOD)
      CALL DERMER(XTMP,VAL,GN,N1,N2,0)
      CALL GTRANS(GN,POINT)
      NOCEX = NOCEX+N1
      NOGEX = NOGEX+N2
C
C  Check the gradient criterion.
      IF (IGCONV(GN,POINT,NOD,VAL,GTOL).NE.0) THEN
         INFO = 2
         RETURN
      END IF
C
C  Calculate the difference in the gradient: Y = g[k+1]-g[k]
C  and in the variables: XS = x[k+1] - x[k]
C  Update derivatives in GRAD.
      MM = 0
      DO 13,I=1,NOD
         IF (IXAT(I).NE.0) THEN
            MM = MM+1
            XS(MM) = POINT(I)-XS(I)
            Y(MM) = GN(I)-GRAD(I)
            GS(MM) = GN(I)
         END IF
         GRAD(I) = GN(I)
13    CONTINUE
C
C  Calculate vector inner products needed to determine whether the
C  Hessian can be updated.
      CALL VVMULT(Y,Y,YTY,M)
      CALL VVMULT(XS,XS,DTD,M)
      CALL VVMULT(Y,XS,YTS,M)
C
C  Can we update the Hessian ?
      IF (YTS.GT.SQRT(COMACC*DTD*YTY)) THEN
C
C  XS = L[t] * s
         CALL LTMULT(H,XS,XS,M,NOD)
C
C  TT = s[t] * H * s
         CALL VVMULT(XS,XS,TT,M)
         IF (TT.EQ.0.) THEN
            INFO = 6
            RETURN
         END IF
C
C  GN = H * s
         CALL LMULT(H,XS,GN,M,NOD)
         AA = SQRT(YTS/TT)
         DO 57,I=1,M
57          GN(I) = AA*(Y(I)-AA*GN(I))/YTS
C
C  Update Choleski factors.
         CALL LLUPD(H,GN,XS,M,NOD)
      END IF
      GOTO 1000
C  ---------------------------------------------------------------------
C  End of TRUST iterations.
C  ---------------------------------------------------------------------
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE TRTM ( INFO, MESS )
C  ---------------------------------------------------------------------
C
C  Description:
C    Constructs a message, describing the reason for termination of the
C    TRUST method.
C
C  Input arguments:
C    INFO       Reason for termination of the TRUST method, as returned
C               by subroutine TRUSMV.
C
C  Output arguments:
C    MESS       Message, describing the reason for termination.
C
C  ---------------------------------------------------------------------
C  Arguments:
      CHARACTER*(*) MESS
C  ---------------------------------------------------------------------
C
      IF (INFO.EQ.1) THEN
         MESS = 'Target value has been reached'
      ELSE IF (INFO.EQ.2) THEN
         MESS = 'The gradient criterion is satisfied'
      ELSE IF (INFO.EQ.3) THEN
         MESS = 'All function evaluations have been used'
      ELSE IF (INFO.EQ.4) THEN
         MESS = 'The minimization variables have converged'
      ELSE IF (INFO.EQ.5) THEN
         MESS = 'The function value has converged'
      ELSE IF (INFO.EQ.6) THEN
         MESS = 'The Hessian matrix cannot be updated'
      ELSE IF (INFO.EQ.7) THEN
         MESS = 'The specified number of iterations has been reached'
      ELSE IF (INFO.EQ.8) THEN
         MESS = 'All variables are fixed'
      ELSE IF (INFO.EQ.9) THEN
         MESS = 'Further progress is not possible'
      ELSE
         CALL MABORT('Incorrect "info".','TRTM')
      END IF
      END
C  ---------------------------------------------------------------------
      SUBROUTINE XTOY ( X, N )
C  ---------------------------------------------------------------------
C
C  Description:
C    Subroutine XTOY eliminates margins and fix statuses by transforming
C    the minimization parameters. This is necessary for routines that
C    cannot handle margins and fixed variables directly.
C    The tranformations used are:
C      For a parameter with a left margin (XL) only:
C        XNEW = SQRT(XOLD-XL)
C      For a parameter with a right margin (XR) only:
C        XNEW = SQRT(XR-XOLD)
C      For a parameter with both margins:
C        XNEW = SQRT(LOG((XR-XL)/(XOLD-XL)
C    Subroutine YTOX performs the reverse transformation. For a general
C    discussion on tranformations and some of the pitfalls, see:
C      Sisser F.S., Elimination of bounds in optimization problems by
C      transforming variables.
C      Math. Prog. 20 (1981) 110-121.
C
C  Input arguments:
C    N          Dimensionality of the objective function.
C
C  Input / Output arguments:
C    X          On input an array of Merlin parameters. On output
C               the transformed parameters, as described earlier.
C
C  Notes:
C    Minimization routines using the XTOY/YTOX transformations should
C    change the contents of array POINT in common block PARAMS, only
C    upon termination.
C
C  ---------------------------------------------------------------------
      IMPLICIT $(TYPE) (A-H,O-Z)
C  Arguments:
      DIMENSION X(N)
C  ---------------------------------------------------------------------
C  Common blocks:
      PARAMETER ( MXV = $(MXV) )
      COMMON /DETAIL/ IXAT(MXV), MARG(MXV)
      COMMON /XLIMS/ XLL(MXV), XRL(MXV)
      COMMON /ACCURA/ COMACC
C  ---------------------------------------------------------------------
C
      DO 10,I=1,N
         MA = MARG(I)
         XL = XLL(I)
         XR = XRL(I)
         IF (IXAT(I).EQ.0) THEN
            MA = 2
            XL = X(I)
            XR = X(I)
         END IF
         IF (MA.EQ.-1) THEN
            X(I) = SQRT(X(I)-XL)
         ELSE IF (MA.EQ.1) THEN
            X(I) = SQRT(XR-X(I))
         ELSE IF (MA.EQ.2) THEN
            BMA = XR-XL
            IF (BMA.NE.0) THEN
               D = 0
               DXL = X(I)-XL
               IF (DXL.EQ.0) D = COMACC
               X(I) = SQRT(LOG(BMA/(DXL+D)))
            ELSE
               X(I) = 1
            END IF
         END IF
10    CONTINUE
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE YTOX ( X, N )
C  ---------------------------------------------------------------------
C
C  Description:
C    Subroutine YTOX cancels the effect of subroutine XTOY, performing
C    the reverse transformations:
C      For a parameter with a left margin (XL) only:
C        XNEW = XL + XOLD**2
C      For a parameter with a right margin (XR) only:
C        XNEW = XR - XOLD**2
C      For a parameter with both margins:
C        XNEW = (XR-XL)*EXP(-X(I)**2) + XL
C    See also the discussion in subroutine XTOY.
C
C  Input arguments:
C    N          Dimensionality of the objective function.
C
C  Input / Output arguments:
C    X          On input an array of transformed parameters. On output
C               an array of normal Merlin parameters.
C
C  Notes:
C    Minimization routines using the XTOY/YTOX transformations should
C    change the contents of array POINT in common block PARAMS, only
C    upon termination.
C
C  ---------------------------------------------------------------------
      IMPLICIT $(TYPE) (A-H,O-Z)
C  Arguments:
      DIMENSION X(N)
C  ---------------------------------------------------------------------
C  Common blocks:
      PARAMETER ( MXV = $(MXV) )
      COMMON /DETAIL/ IXAT(MXV), MARG(MXV)
      COMMON /XLIMS/ XLL(MXV), XRL(MXV)
      COMMON /PARAMS/ POINT(MXV), VAL
C  ---------------------------------------------------------------------
C
      DO 10,I=1,N
         MA = MARG(I)
         XL = XLL(I)
         XR = XRL(I)
         IF (IXAT(I).EQ.0) THEN
            MA = 2
            XL = POINT(I)
            XR = POINT(I)
         END IF
         IF (MA.EQ.-1) THEN
            X(I) = XL + X(I)**2
         ELSE IF (MA.EQ.1) THEN
            X(I) = XR - X(I)**2
         ELSE IF (MA.EQ.2) THEN
            X(I) = (XR-XL)*AXP(-X(I)**2) + XL
         END IF
10    CONTINUE
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE ACCUDE ( POINT, II, VAL, GRAD, NOC, ERB, RE, IERR )
C  ---------------------------------------------------------------------
C
C  Description:
C    Subroutine ACCUDE computes an approximation (DER) to the first
C    partial derivative of the objective function, with respect to
C    variable II. Normally it uses a 6 point central difference
C    formula, although sometimes more terms in the Taylor series
C    have to be added in order to achive the required accuracy.
C
C    For a discussion of numerical differentiation techniques see:
C
C    1) Dennis J.E. and Schnabel R.B.
C       Numerical Methods for Unconstrained Optimization and Nonlinear
C       Equations
C       Prentice-Hall, 1983. pp. 103-105, 322-323.
C    2) Press W.H., Teukolsky S.A., Vetterling W.T., Flannery B.P
C       Numerical recipes, second edition
C       Cambridge University Press, 1992. pp. 180-184.
C
C  Input arguments:
C    POINT      The current point.
C    II         Differentiation is performed with respect to variable
C               II.
C    VAL        Current value of the objective function.
C    ERB        Requested accuracy.
C
C  Output arguments:
C    GRAD       The computed approximation to the derivative.
C    NOC        Number of evaluations of the objective function
C               that were performed by this routine.
C    RE         An estimation of the relative error due to the
C               truncation of the Taylor series.
C    IERR       IERR =  0 -> The derivative was computed within the
C                            requested accuracy.
C               IERR = -1 ->  Accuracy requirement not satisfied,
C                             inspite higher order terms are calculated
C                             in the taylor expansion.
C
C  ---------------------------------------------------------------------
      IMPLICIT $(TYPE) (A-H,O-Z)
C  Arguments:
      DIMENSION POINT(*)
C  ---------------------------------------------------------------------
C  Common blocks:
      COMMON /ACCURA/ COMACC
C  ---------------------------------------------------------------------
C  Local variables:
C  ---------------------------------------------------------------------
C
      COMTOL = SQRT(COMACC)
      X = POINT(II)
      NOC = 0
      ONE = 1.
      XDIV = 511*63*45
      A19 = 2**19/XDIV
      A15 = (2**15+2**13+2**10)/XDIV
      A9 = (2**9+2**6+2**4)/XDIV
      ICHEK = 0
      IERR = 0
      H = 0.001
      IF (ABS(ABS(X)-ABS(H)).LT.ABS(X/10.)) H = H*0.01
      IF ((X+H).EQ.X) H = (1.+ABS(X))*COMTOL
      ICOUNT = 0
2     CONTINUE
      ICOUNT = ICOUNT+1
      XX = X+H
      FP1 = ACSQ1(POINT,II,XX)
      CHEDER = ABS(FP1-VAL)/(ABS(VAL)+COMACC)
      IF (CHEDER.LT.COMACC) THEN
         ICHEK = ICHEK+1
         ICOUNT = ICOUNT-1
         NOC = NOC+1
         H = 10.*H
         IF (ICHEK.LE.5) GOTO 2
      END IF
      XX = X-H
      FM1 = ACSQ1(POINT,II,XX)
      XX = X+2.*H
      FP2 = ACSQ1(POINT,II,XX)
      XX = X-2.*H
      FM2 = ACSQ1(POINT,II,XX)
      XX = X+4.*H
      FP4 = ACSQ1(POINT,II,XX)
      XX = X-4.*H
      FM4 = ACSQ1(POINT,II,XX)
      NOC = NOC+6
      DH1 = 2.*(FP1-FM1)/H-(FP2-FM2)/(4.*H)
      DH2 = (FP2-FM2)/H-(FP4-FM4)/(8.*H)
      DF1 = DH1/3.
      E1 = (DH1-DH2)/45.
      RE = ABS(E1)/MAX(ONE,ABS(DF1+E1))
      IF (RE.LT.ERB) THEN
C  Derivative calculated within desired accuracy
         GRAD = DF1+E1
         GOTO 3
      END IF
      FMX = ABS(VAL)+COMACC
      XSTEP = (ERB*FMX*4.*H**4/(ABS(E1)+COMACC))**0.2
      IF (H.GT.XSTEP) THEN
         IF (ICOUNT.LT.2) THEN
            H = XSTEP/5.
            GOTO 2
         ELSE
C  Failure to calculate derivative within the desired accuracy
C  inspite step reduction and recalculation of fp1, fp2,... etc.
            IERR = -1
            GRAD = DF1+E1
            RE = 1.
            GOTO 3
         END IF
      ELSE
         XX = X+H/2.
         FPH = ACSQ1(POINT,II,XX)
         XX = X-H/2.
         FMH = ACSQ1(POINT,II,XX)
         DHH = 4.*(FPH-FMH)/H-(FP1-FM1)/(2.*H)
         NOC = NOC+2
         GRAD = (1024.*DHH-80.*DH1+DH2)/2835.
         E2 = GRAD-DF1-E1
         RE = ABS(E2)/MAX(ONE,ABS(GRAD))
         IF (RE.LT.ERB) GOTO 3
         XX = X+H/4.
         FPQ = ACSQ1(POINT,II,XX)
         XX = X-H/4.
         FMQ = ACSQ1(POINT,II,XX)
         NOC = NOC+2
         DHQ = 8.*(FPQ-FMQ)/H-(FPH-FMH)/H
         DERN = A19*DHQ-A15*DHH+A9*DH1-DH2/XDIV
         E3 = DERN-GRAD
         RE = ABS(E3)/MAX(ONE,ABS(DERN))
         GRAD = DERN
         IF (RE.GT.ERB) IERR = -1
C  Accuracy requirement not satisfied, inspite higher order
C  terms are calculated in the Taylor expansion.
         GOTO 3
      END IF
3     CONTINUE
C
      END
C  ---------------------------------------------------------------------
      FUNCTION ACSQ1 ( XPOINT, II, X )
C  ---------------------------------------------------------------------
C
C  Description:
C    This is the function that the numerical differentiation routines
C    use. It simply sets the II-th component of XPOINT equal to X, and
C    calls the objective function ACSQ.
C
C  Input arguments:
C    XPOINT     The point at which differentiation occurs.
C    II         Minimization variable whose value must be changed.
C    X          New value for variable II.
C
C  ---------------------------------------------------------------------
      IMPLICIT $(TYPE) (A-H,O-Z)
C  Arguments:
      DIMENSION XPOINT(*)
C  ---------------------------------------------------------------------
C  Common blocks:
      COMMON /DIMEN/ NOD, NOF
C  ---------------------------------------------------------------------
C
      XSAVE = XPOINT(II)
      XPOINT(II) = X
      ACSQ1 = ACSQ(XPOINT,NOD)
      XPOINT(II) = XSAVE
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE CDERI ( POINT, II , F0, STPCON, DER, NOC )
C  ---------------------------------------------------------------------
C
C  Description:
C    Subroutine CDERI computes an approximation (DER) to the first
C    partial derivative of the objective function, with respect to
C    variable II. Normally it uses the central difference formula
C    DER = (F(X+H)-F(X-H))/(2*H). If any one of X+H, X-H extend
C    beyond the margins, the 3 point forward or backward difference
C    formulae DER = (4*F(X+H)-3*F(X)-F(X+2*H))/(2*H) and
C    DER = -(4*F(X-H)-3*F(X)-F(X-2*H))/(2*H) are used.
C    If these fail too (X+2*H, X-2*H are outside the margins) DER = 0
C    is returned (the problem is probably very badly scaled).
C
C    For a discussion of numerical differentiation techniques see:
C
C    1) Dennis J.E. and Schnabel R.B.
C       Numerical Methods for Unconstrained Optimization and Nonlinear
C       Equations
C       Prentice-Hall, 1983. pp. 103-105, 322-323.
C    2) Press W.H., Teukolsky S.A., Vetterling W.T., Flannery B.P
C       Numerical recipes, second edition
C       Cambridge University Press, 1992. pp. 180-184.
C
C  Input arguments:
C    POINT      The current point.
C    II         Differentiation is performed with respect to variable
C               II.
C    F0         Current value of the objective function.
C    STPCON     Constant that determines the step size. The cited
C               references discuss the selection of the step size.
C
C  Output arguments:
C    DER        The computed derivative.
C    NOC        Number of calls to the objective function.
C
C  ---------------------------------------------------------------------
      IMPLICIT $(TYPE) (A-H,O-Z)
C  Arguments:
      DIMENSION POINT(*)
C  ---------------------------------------------------------------------
C  Common blocks:
      PARAMETER ( MXV = $(MXV) )
      COMMON /DETAIL/ IXAT(MXV), MARG(MXV)
      COMMON /XLIMS/ XLL(MXV), XRL(MXV)
C  ---------------------------------------------------------------------
C
      ONE = 1.
      TMP = POINT(II)
      H = STPCON*MAX(ABS(TMP),ONE)
      MA = MARG(II)
      NOC = 0
C
      XP = TMP+H
      H = XP-TMP
      XM = TMP-H
C
C  Make sure the new "X" lies inside the margins.
      IF (MA.EQ.-1 .OR. MA.EQ.2) THEN
         IF (XM.LT.XLL(II)) THEN
            XPP = XP+H
            IF (MA.EQ.1 .OR. MA.EQ.2) THEN
               IF (XPP.GT.XRL(II)) THEN
                  DER = 0.
                  RETURN
               END IF
            END IF
C  Use the 3 point forward difference formula.
            FPP = ACSQ1(POINT,II,XPP)
            FP = ACSQ1(POINT,II,XP)
            DER = (4.*FP-3.*F0-FPP)/(2.*H)
            NOC = 2
            RETURN
         END IF
      END IF
      IF (MA.EQ.1 .OR. MA.EQ.2) THEN
         IF (XP.GT.XRL(II)) THEN
            XMM = XM-H
            IF (MA.EQ.-1 .OR. MA.EQ.2) THEN
               IF (XMM.LT.XLL(II)) THEN
                  DER = 0.
                  RETURN
               END IF
            END IF
C  Use the 3 point backward difference formula.
            FMM = ACSQ1(POINT,II,XMM)
            FM = ACSQ1(POINT,II,XM)
            DER = -(4.*FM-3.*F0-FMM)/(2.*H)
            NOC = 2
            RETURN
         END IF
      END IF
C
      FP = ACSQ1(POINT,II,XP)
      FM = ACSQ1(POINT,II,XM)
      DER = (FP-FM)/(2.*H)
      NOC = 2
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE DERMER ( POINT, VAL, GRAD, NOC, NGR, MODE )
C  ---------------------------------------------------------------------
C
C  Description:
C    This is the main Merlin routine that computes approximations
C    to the gradient vector, according to the curent derivative mode.
C    Subroutine DERMER uses an internal cache (GRALOC) to store the
C    computed derivatives. Depending on the setting of IGRADU and the
C    requested operation (MODE), DERMER can return the cached gradient,
C    saving thus unecessary computations.
C
C  Input arguments:
C    POINT      The point where derivatives are requested.
C    VAL        The value of the objective function calculated at POINT.
C    MODE       The requested operation:
C               MODE = 0 -> Normal operation. If the gradient is
C                           available in the cache, it is retured.
C                           Otherwise the derivatives are evaluated
C                           using the current derivative mode, stored
C                           in the gradient cache, and returned to
C                           the calling routine.
C               MODE = 1 -> The gradient vector is calculated. The
C                           gradient cache is not consulted beforehand,
C                           nor is the computed vector stored in the
C                           cache.
C               MODE = 2 -> Returns the cached gradient (does not
C                           evaluate gradient, unless the gradient
C                           cache is empty).
C
C  Output arguments:
C    GRAD       The gradient vector.
C    NOC        Number of function evaluations.
C    NGR        Number of calls to the user supplied subroutine GRANAL.
C
C  ---------------------------------------------------------------------
      IMPLICIT $(TYPE) (A-H,O-Z)
C  Arguments:
      DIMENSION POINT(*), GRAD(*)
C  ---------------------------------------------------------------------
C  Common blocks:
      PARAMETER ( MXV = $(MXV) )
      COMMON /DETAIL/ IXAT(MXV), MARG(MXV)
      CHARACTER*5 GRADI
      COMMON /DERIMO/ GRADI(MXV)
      COMMON /ACCURA/ COMACC
      COMMON /UPDATE/ IGRADU, IHESSU
      COMMON /TOTCAL/ ITNOCS, IPARTC, ITDER, IPDER, ITHESS, IPHESS,
     &                ITJAC, IPJAC
      LOGICAL AUTODE
      COMMON /AUTOMA/ AUTODE
      COMMON /DIMEN/ NOD, NOF
      LOGICAL STRO, WEAK
      COMMON /PRILOG/ STRO, WEAK
      COMMON /UNITS/ IUINP, IUOUT
C  ---------------------------------------------------------------------
C  Local variables:
      DIMENSION WORK(MXV)
      DIMENSION GRALOC(MXV)
      SAVE GRALOC
      CHARACTER*5 GRMODE
      LOGICAL LPRI
      CHARACTER*15 STR
      DATA ICACHE / 0 /
C  ---------------------------------------------------------------------
C
C  Get parameters from the MAD panel.
      IF (AUTODE) THEN
         CALL GETPPR('MAD','FQ',FQTHRE)
         CALL GETPPR('MAD','QN',QNTHRE)
         CALL GETPPI('MAD','RECALC',IRECAL)
         CALL GETPPI('MAD','PRINT',IPRIM)
      END IF
      N = NOD
C
C  Total number of function calls.
      NOC = 0
C  Total number of GRANAL calls.
      NGR = 0
C
      IF (MODE.EQ.0 .AND. IGRADU.EQ.1 .OR.
     &    MODE.EQ.2 .AND. ICACHE.NE.0) THEN
         DO 40,I=1,N
40          GRAD(I) = GRALOC(I)*IXAT(I)
         RETURN
      END IF
C
C  Number of analytic gradient components.
      NANAL = 0
      LPRI = AUTODE .AND. IPRIM.EQ.1 .AND. STRO
C
C  FEPS should be initialized to something more appropriate.
C  It is supposed to be the accuracy to which the objective function
C  is calculated.
      FEPS = COMACC
      ONE = 1.
      THREE = 3.
      COMTOL = SQRT(COMACC)
      ERB = COMTOL
      CON1 = COMTOL
      CON2 = FEPS**(ONE/THREE)
C
C  Loop over all variables.
      DO 20,II=1,N
120      CONTINUE
         GRMODE = GRADI(II)
         IF (IXAT(II).EQ.0) THEN
            GRAD(II) = 0.
         ELSE IF (GRMODE.EQ.'FAST') THEN
            CALL FDERI(POINT,II,VAL,CON1,DER,NOCF)
            GRAD(II) = DER
            NOC = NOC+NOCF
            IF (MODE.EQ.0) THEN
               IF (AUTODE .AND. ABS(DER).LT.FQTHRE) THEN
                  GRADI(II) = 'QUAD'
                  IF (LPRI) THEN
                     CALL VNAME(II,STR,LES)
                     WRITE (IUOUT,100) STR(1:LES), 'QUAD'
                  END IF
                  IF (IRECAL.NE.0) GOTO 120
               END IF
            END IF
         ELSE IF (GRMODE.EQ.'QUAD') THEN
            CALL CDERI(POINT,II,VAL,CON2,DER,NOCC)
            GRAD(II) = DER
            NOC = NOC+NOCC
            IF (MODE.EQ.0) THEN
               IF (AUTODE .AND. ABS(DER).LT.QNTHRE) THEN
                  GRADI(II) = 'NUMER'
                  IF (LPRI) THEN
                     CALL VNAME(II,STR,LES)
                     WRITE (IUOUT,100) STR(1:LES), 'NUMER'
                  END IF
                  IF (IRECAL.NE.0) GOTO 120
               END IF
            END IF
         ELSE IF (GRMODE.EQ.'NUMER') THEN
            CALL ACCUDE(POINT,II,VAL,DER,NOCA,ERB,TE,IE)
            GRAD(II) = DER
            NOC = NOC+NOCA
         ELSE
            NANAL = NANAL+1
         END IF
20    CONTINUE
C
      IF (NANAL.NE.0) THEN
         ITDER = ITDER+1
         IPDER = IPDER+1
         CALL GRANAL(N,POINT,WORK)
         NGR = 1
         DO 50,I=1,N
            IF (GRADI(I).EQ.'ANAL') GRAD(I) = WORK(I)*IXAT(I)
50       CONTINUE
      END IF
C
C  Store in the cache.
      IF (MODE.EQ.0) THEN
         DO 30,I=1,N
30          GRALOC(I) = GRAD(I)
         IGRADU = 1
         ICACHE = 1
      END IF
C
100   FORMAT ($(FC)'Automatic derivatives: Variable ',A,
     &        ' set to ',A,'.')
200   FORMAT ($(FC)'Automatic derivatives: Variable ',A,
     &        ' reset to ',A,'.')
      END
C  ---------------------------------------------------------------------
      SUBROUTINE FDERI ( POINT, II, F0, STPCON, DER, NOC )
C  ---------------------------------------------------------------------
C
C  Description:
C    Subroutine FDERI computes an approximation (DER) to the first
C    partial derivative of the objective function, with respect to
C    variable II. Normally it uses the forward difference formula
C    DER = (F(X+H)-F(X))/H. If however X+H extends beyond the margins
C    the backward difference formula DER = (F(X)-F(X-H))/H
C    is used. If X-H lies outside of the margins as well, DER = 0 is
C    returned (the problem is probably very badly scaled).
C
C    For a discussion of numerical differentiation techniques see:
C
C    1) Dennis J.E. and Schnabel R.B.
C       Numerical Methods for Unconstrained Optimization and Nonlinear
C       Equations
C       Prentice-Hall, 1983. pp. 103-105, 322-323.
C    2) Press W.H., Teukolsky S.A., Vetterling W.T., Flannery B.P
C       Numerical recipes, second edition
C       Cambridge University Press, 1992. pp. 180-184.
C
C  Input arguments:
C    POINT      The current point.
C    II         Differentiation is performed with respect to variable
C               II.
C    F0         Current value of the objective function.
C    STPCON     Constant that determines the step size. The cited
C               references discuss the selection of the step size.
C
C  Output arguments:
C    DER        The computed derivative.
C    NOC        Number of calls to the objective function.
C
C  ---------------------------------------------------------------------
      IMPLICIT $(TYPE) (A-H,O-Z)
C  Arguments:
      DIMENSION POINT(*)
C  ---------------------------------------------------------------------
C  Common blocks:
      PARAMETER ( MXV = $(MXV) )
      COMMON /DETAIL/ IXAT(MXV), MARG(MXV)
      COMMON /XLIMS/ XLL(MXV), XRL(MXV)
C  ---------------------------------------------------------------------
C
      ONE = 1.
      TMP = POINT(II)
      H = STPCON*MAX(ABS(TMP),ONE)*SIGN(ONE,TMP)
      MA = MARG(II)
      NOC = 0
C
      X = TMP+H
C
C  Make sure the new "X" lies inside the margins.
      IF (H.LT.0. .AND. (MA.EQ.-1 .OR. MA.EQ.2)) THEN
         IF (X.LT.XLL(II)) THEN
            X = TMP-H
            IF (MA.EQ.1 .OR. MA.EQ.2) THEN
               IF (X.GT.XRL(II)) THEN
                  DER = 0.
                  RETURN
               END IF
            END IF
         END IF
      ELSE IF (H.GE.0. .AND. (MA.EQ.1 .OR. MA.EQ.2)) THEN
         IF (X.GT.XRL(II)) THEN
            X = TMP-H
            IF (MA.EQ.-1 .OR. MA.EQ.2) THEN
               IF (X.LT.XLL(II)) THEN
                  DER = 0.
                  RETURN
               END IF
            END IF
         END IF
      END IF
C
      H = X-TMP
      FP = ACSQ1(POINT,II,X)
      DER = (FP-F0)/H
      NOC = 1
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE MJACOB ( M, N, X, FVEC, FJ, LD )
C  ---------------------------------------------------------------------
C
C  Description:
C    Returns the Jacobian matrix, according to the jacobian mode.
C    It either calls the user supplied routine JANAL, or uses forward
C    differences to approximate the Jacobian.
C
C  Input arguments:
C    M          Number of terms in the sum-of-squares.
C    N          The dimensionality of the objective function.
C    X          The point at which the JAcobina is requested.
C    FVEC       FVEC(I) contains the value of term I.
C    LD         The leading dimension of matrix FJ.
C
C  Output arguments:
C    FJ         The Jacobian matrxix. FJ(I,J) corresponds to df(i)/dx(j)
C
C  ---------------------------------------------------------------------
      IMPLICIT $(TYPE) (A-H,O-Z)
C  Arguments:
      DIMENSION X(N), FJ(LD,N), FVEC(M)
C  ---------------------------------------------------------------------
C  Common blocks:
      PARAMETER ( MXT = $(MXT) )
      CHARACTER*10 DERIVA, PRINTO, PROCES, JACOMO
      COMMON /MODEIN/ DERIVA, PRINTO, PROCES, JACOMO
      COMMON /TOTCAL/ ITNOCS, IPARTC, ITDER, IPDER, ITHESS, IPHESS,
     &                ITJAC, IPJAC
      COMMON /ACCURA/ COMACC
      COMMON /RWORKT/ FT1(MXT), FT2(MXT), FT3(MXT)
C  ---------------------------------------------------------------------
C
      IF (JACOMO.EQ.'JANAL') THEN
         CALL JANAL(M,N,X,FJ,LD)
         ITJAC = ITJAC+1
         IPJAC = IPJAC+1
      ELSE IF (JACOMO.EQ.'JNUMER') THEN
C  We should set EPSFCN to something more reasonable.
         EPSFCN = COMACC
         IFLAG = 2
         CALL FDJAC2(M,N,X,FVEC,FJ,LD,IFLAG,EPSFCN,FT3)
      ELSE
         CALL MABORT('No such Jacobian mode: '//JACOMO,'MJACOB')
      END IF
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE SECDER ( H, LD, N, X, VAL, HOW, W1, W2, NC )
C  ---------------------------------------------------------------------
C
C  Description:
C    Subroutine SECDER is the main entry point to a set of routines
C    that compute approximations to the second derivative matrix.
C
C  Input arguments:
C    LD         The first (leading) dimension of H.
C    N          Dimensionality of the objective function.
C    X          The current point.
C    VAL        Current value of the objective function.
C    HOW        Selects the method of approximation:
C               HOW = 'F' -> Use function values only.
C               HOW = 'G' -> Use gradient values. Use only if the
C                            user supplied routine GRANAL is available.
C               HOW = 'A' -> Call the user supplied routine HANAL
C                            that computes the second derivative
C                            matrix.
C
C  Output arguments:
C    H          The second derivative matrix. On output the lower
C               triangular part of H contains the partial second
C               derivatives.
C    W1         Work array.
C    W2         Work array.
C    NC         If HOW = 'F', NC is the number evaluations of the
C                             objective function that were performed.
C               If HOW = 'G', NC is the number of calls that were
C                             performed to the user supplied subroutine
C                             GRANAL.
C               If HOW = 'A', NC is set to 1, since one call is
C                             performed to the user supplied subroutine
C                             HANAL.
C
C  ---------------------------------------------------------------------
      IMPLICIT $(TYPE) (A-H,O-Z)
C  Arguments:
      DIMENSION H(LD,N), X(N), W1(N), W2(N)
      CHARACTER HOW
C  ---------------------------------------------------------------------
C  Common blocks:
      PARAMETER ( MXV = $(MXV) )
      COMMON /DETAIL/ IXAT(MXV), MARG(MXV)
      COMMON /TOTCAL/ ITNOCS, IPARTC, ITDER, IPDER, ITHESS, IPHESS,
     &                ITJAC, IPJAC
      COMMON /ACCURA/ COMACC
C  ---------------------------------------------------------------------
C
      IF (ISCOMP(HOW,'F').NE.0) THEN
         CALL HESSF(H,LD,N,X,VAL,IXAT,W1,W2,COMACC,NC)
      ELSE IF (ISCOMP(HOW,'G').NE.0) THEN
         CALL HESSG(H,LD,N,X,IXAT,W1,W2,COMACC,NC)
         ITDER = ITDER+NC
         IPDER = IPDER+NC
      ELSE IF (ISCOMP(HOW,'A').NE.0) THEN
         CALL HANAL(H,LD,N,X)
         ITHESS = ITHESS+1
         IPHESS = IPHESS+1
         NC = 1
         DO 10,I=1,N
            DO 20,J=1,I
20             IF (IXAT(I)*IXAT(J).EQ.0) H(I,J) = 0.
10       CONTINUE
      END IF
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE HESSF ( H, LD, N, X, VAL, IXAT, STP, FUN, COMACC, NF )
C  ---------------------------------------------------------------------
C
C  Description:
C    Subroutine HESSF computes an approximation to the second derivative
C    matrix. It uses only function values and employs a forward
C    difference formula. Only the lower triangular part of the matrix
C    is referenced.
C
C    For a discussion on numerical differentiation techniques see:
C
C    1) Dennis J.E. and Schnabel R.B.
C       Numerical Methods for Unconstrained Optimization and Nonlinear
C       Equations
C       Prentice-Hall, 1983. pp. 103-105, 322-323.
C    2) Press W.H., Teukolsky S.A., Vetterling W.T., Flannery B.P
C       Numerical recipes, second edition
C       Cambridge University Press, 1992. pp. 180-184.
C
C  Input arguments:
C    LD         The first (leading) dimension of H.
C    N          Dimensionality of the objective function.
C    X          The current point.
C    VAL        Current value of the objective function.
C    IXAT       Fix statuses.
C    COMACC     Relative machine accuracy.
C
C  Output arguments:
C    H          The second derivative matrix. On output the lower
C               triangular part of H contains the partial second
C               derivatives. The upper part is not referenced.
C    STP        Work array. On output contains the derivative steps.
C    FUN        Work array.
C    NF         Number evaluations of the objective function that
C               were performed.
C
C  ---------------------------------------------------------------------
      IMPLICIT $(TYPE) (A-H,O-Z)
C  Arguments:
      DIMENSION H(LD,N), X(N), STP(N), FUN(N)
      DIMENSION IXAT(N)
C  ---------------------------------------------------------------------
C
      ONE = 1.
      THREE = 3.
      CON = COMACC**(ONE/THREE)
C
      NF = 0
      DO 10,I=1,N
         IF (IXAT(I).NE.0) THEN
            XI = X(I)
            XP = XI+CON*MAX(ONE,ABS(XI))*SIGN(ONE,XI)
            STP(I) = XP-XI
            X(I) = XP
            FUN(I) = ACSQ(X,N)
            NF = NF+1
            X(I) = XI
         END IF
10    CONTINUE
C
      DO 20,I=1,N
         IF (IXAT(I).NE.0) THEN
            FUNI = FUN(I)
            STPI = STP(I)
            XI = X(I)
            X(I) = XI+2*STPI
            FII = ACSQ(X,N)
            NF = NF+1
            H(I,I) = (VAL-2*FUNI+FII)/STPI**2
            X(I) = XI+STP(I)
C
            DO 30,J=1,I-1
               IF (IXAT(J).NE.0) THEN
                  XJ = X(J)
                  X(J) = XJ+STP(J)
                  FIJ = ACSQ(X,N)
                  NF = NF+1
                  X(J) = XJ
                  H(I,J) = (VAL-FUNI+FIJ-FUN(J))/(STPI*STP(J))
               ELSE
                  H(I,J) = 0.
               END IF
30          CONTINUE
            X(I) = XI
         ELSE
            DO 40,J=1,I
40             H(I,J) = 0.
         END IF
20    CONTINUE
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE HESSG ( H, LD, N, X, IXAT, GC, G, COMACC, NG )
C  ---------------------------------------------------------------------
C
C  Description:
C    Subroutine HESSG computes an approximation to the second derivative
C    matrix. It uses only gradient values and employs a forward
C    difference formula. The whole H matrix is referenced.
C
C    For a discussion on numerical differentiation techniques see:
C
C    1) Dennis J.E. and Schnabel R.B.
C       Numerical Methods for Unconstrained Optimization and Nonlinear
C       Equations
C       Prentice-Hall, 1983. pp. 103-105, 322-323.
C    2) Press W.H., Teukolsky S.A., Vetterling W.T., Flannery B.P
C       Numerical recipes, second edition
C       Cambridge University Press, 1992. pp. 180-184.
C
C  Input arguments:
C    LD         The first (leading) dimension of H.
C    N          Dimensionality of the objective function.
C    X          The current point.
C    IXAT       Fix statuses.
C    COMACC     Relative machine accuracy.
C
C  Output arguments:
C    H          The second derivative matrix. On output the lower
C               triangular part of H contains the partial second
C               derivatives.
C    GC         Work array. On output contains the first derivatives
C               evauated at the current point.
C    G          Work array.
C    NG         Number of calls to the user supplied subroutine GRANAL.
C
C  ---------------------------------------------------------------------
      IMPLICIT $(TYPE) (A-H,O-Z)
C  Arguments:
      DIMENSION H(LD,N), X(N), GC(N), G(N)
      DIMENSION IXAT(N)
C  ---------------------------------------------------------------------
C
      ONE = 1.
      CON = SQRT(COMACC)
C
      NG = 1
      CALL GRANAL(N,X,GC)
      DO 60,I=1,N
60       GC(I) = IXAT(I)*GC(I)
C
      DO 10,I=1,N
         IF (IXAT(I).NE.0) THEN
            XI = X(I)
            X(I) = XI+CON*MAX(ABS(XI),ONE)*SIGN(ONE,XI)
            STP = X(I)-XI
            CALL GRANAL(N,X,G)
            DO 70,J=1,N
70             G(J) = IXAT(J)*G(J)
            NG = NG+1
            DO 20,J=1,N
20             H(I,J) = (G(J)-GC(J))/STP
            X(I) = XI
         ELSE
            DO 50,J=1,N
50             H(I,J) = 0.
         END IF
10    CONTINUE
C
      DO 40,I=1,N
         DO 30,J=1,I-1
            HH = (H(I,J)+H(J,I))/2.
            H(I,J) = HH
            H(J,I) = HH
30       CONTINUE
40    CONTINUE
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE ISAMEF ( NU, DESC )
C  ---------------------------------------------------------------------
C
C  Description:
C    Checks whether the input unit number NU is one of the open
C    Merlin files. (input, output, backup, memo or history files)
C    If NU is indeed the unit number of an open Merlin file, a
C    short description is returned in the character variable DESC.
C
C  Input arguments:
C    NU         Unit number to check. NU should be the unit number
C               of an open file, obtained via INQUIRE ( NUMBER = NU...
C               If NU = -1, the unit is considered unallocated and
C               this routine just returns.
C
C  Output arguments:
C    DESC       Short description of the unit number NU. It should
C               be at least 7 characters long. DESC is set as follows:
C               DESC = ' '       if NU does not belong to one of the
C                                open Merlin files.
C               DESC = 'backup'  if NU belongs to the currently active
C                                backup file.
C               DESC = 'history' if NU belongs to the currently active
C                                history file.
C               DESC = 'memo'    if NU belongs to the currently open
C                                memo file.
C               DESC = 'input'   if NU belongs to the currently open
C                                Merlin input file.
C               DESC = 'output'  if NU belongs to the currently open
C                                output file.
C
C  ---------------------------------------------------------------------
C  Arguments:
      CHARACTER*(*) DESC
C  ---------------------------------------------------------------------
C  Common blocks:
      COMMON /BACKIT/ IUBA, IUBAMO, IBASA
      COMMON /MEMO2/ MEMOMO, MEMOKE, MEMREC, IUMEM
      COMMON /LOGIT/ IULOG
      COMMON /UNTSDF/ IU31, IU32
C  ---------------------------------------------------------------------
C
      DESC = ' '
      IF (NU.EQ.-1) RETURN
C
      IF (NU.EQ.IUBA) THEN
         DESC = 'backup'
      ELSE IF (NU.EQ.IUMEM) THEN
         DESC = 'memo'
      ELSE IF (NU.EQ.IULOG) THEN
         DESC = 'history'
      ELSE IF (NU.EQ.IU31) THEN
         DESC = 'input'
      ELSE IF (NU.EQ.IU32) THEN
         DESC = 'output'
      END IF
C
      END
C  ---------------------------------------------------------------------
      FUNCTION LENDIR ( )
C  ---------------------------------------------------------------------
C
C  Description:
C    Returns the appropriate record length for an unformatted, direct
C    access Merlin file. (Usualy refered to as a binary file)
C    The returned value is used in an open statement of the form:
C      OPEN (FILE=..., ACCESS='DIRECT', FORM='UNFORMATTED', RECL=... )
C    Function LENDIR takes into account the configuration directives
C    that set the sizes for characters, reals and integers.
C
C  ---------------------------------------------------------------------
C  Common blocks:
      PARAMETER ( MXV = $(MXV) )
      CHARACTER*10 NICKNA
      COMMON /NAMES/ NICKNA(MXV)
      COMMON /VSIZES/ NSCHA, NSINT, NSREAL
C  ---------------------------------------------------------------------
C
      LENDIR = LEN(NICKNA(1))*NSCHA+2*NSINT+3*NSREAL
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE OPAPPE ( NU, FILE, LEF, IERR )
C  ---------------------------------------------------------------------
C
C  Description:
C    Opens a sequential, formatted file (usualy refered to as text file)
C    and positions it at End-of-File, so that a subsequent WRITE
C    statement will append data to it. If the common extension
C    ACCESS='APPEND' is present in the system (you have to declare
C    this in the configuration file) it is used, since it is expected
C    to speed up things (especialy when large files are opened).
C
C  Input arguments:
C    NU         Unit number to be used when opening the file.
C    FILE       Character variable. The file to open.
C    LEF        Number of useful characters in FILE.
C
C  Output arguments:
C    IERR       Error indicator.
C
C  ---------------------------------------------------------------------
C  Arguments:
      CHARACTER*(*) FILE
C  ---------------------------------------------------------------------
C  Common blocks:
      LOGICAL HEADER, IPRECO, OPRECO, HASAPP, OUTAPP
      COMMON /CP1/ HEADER, IPRECO, OPRECO, HASAPP, OUTAPP
C  ---------------------------------------------------------------------
C  Local variables:
      CHARACTER*6 APPND
C  ---------------------------------------------------------------------
C
      APPND = 'APPEND'
      IF (HASAPP) THEN
C
C  The configuration file assures us that this system recognizes the
C  APPEND extension.
         OPEN (NU,FILE=FILE(1:LEF),ACCESS=APPND,IOSTAT=IERR)
      ELSE
C
C  The APPEND extension is not present in this system. Use the
C  traditional way: Open the file and position it at End-of-File.
         OPEN (NU,FILE=FILE(1:LEF),IOSTAT=IERR)
         IF (IERR.EQ.0) CALL POSIT(NU)
      END IF
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE OPOUT ( ERROR )
C  ---------------------------------------------------------------------
C
C  Description:
C    Opens the Merlin output unit/file. Several diagnostics are issued
C    in case of errors. Since the output unit is not yet open, all
C    messages of this routine are printed in the standard output
C    unit *.
C
C  Output arguments:
C    ERROR      If an error occurs, ERROR is set to .true.; .false.
C               otherwise.
C
C  ---------------------------------------------------------------------
      IMPLICIT $(TYPE) (A-H,O-Z)
C  Arguments:
      LOGICAL ERROR
C  ---------------------------------------------------------------------
C  Common blocks:
      COMMON /UNITS/ IUINP, IUOUT
      COMMON /UNTSDF/ IU31, IU32
      LOGICAL HEADER, IPRECO, OPRECO, HASAPP, OUTAPP
      COMMON /CP1/ HEADER, IPRECO, OPRECO, HASAPP, OUTAPP
      PARAMETER ( MXFILE = 80 )
      CHARACTER*(MXFILE) PROLO, EPILO, IN, OUT, HFILE, MFILE, PDFILE
      COMMON /CP2/ PROLO, EPILO, IN, OUT, HFILE, MFILE, PDFILE
C  ---------------------------------------------------------------------
C  Local variables:
      LOGICAL OPE
      CHARACTER*10 STR1, STR2
C  ---------------------------------------------------------------------
C
      ERROR = .TRUE.
C
      LEOUT = LENGTH(OUT)
      IF (LEOUT.EQ.0) THEN
C  Merlin writes to a unit.
         IF (.NOT.OPRECO) THEN
C
C  The output unit is not preconnected. Open it.
            OPEN (IUOUT,IOSTAT=IERR)
            IF (IERR.NE.0) THEN
               CALL I2STR(IUOUT,STR1,LE1)
               CALL I2STR(IERR,STR2,LE2)
               WRITE (*,10) STR1(1:LE1), STR2(1:LE2)
               RETURN
            END IF
         END IF
      ELSE
C  Merlin writes to a file.
         INQUIRE (FILE=OUT,OPENED=OPE,NUMBER=IU,IOSTAT=IERR)
         IF (IERR.NE.0) THEN
            CALL I2STR(IERR,STR1,LE1)
            WRITE (*,20) OUT(1:LEOUT), STR1(1:LE1)
            RETURN
         END IF
C
C  The output file is already open ?
         IF (OPE) THEN
            CALL I2STR(IU,STR1,LE1)
            WRITE (*,30) OUT(1:LEOUT), STR1(1:LE1)
            RETURN
         END IF
C
C  Allocate unit for the output file.
         IUOUT = NUNIT()
         IF (IUOUT.EQ.-1) THEN
            WRITE (*,40) OUT(1:LEOUT)
            RETURN
         END IF
C
C  Open the output file.
         IF (OUTAPP) THEN
            CALL OPAPPE(IUOUT,OUT,LEOUT,IERR)
         ELSE
            OPEN (IUOUT,FILE=OUT,IOSTAT=IERR)
            IF (IERR.EQ.0) REWIND IUOUT
         END IF
         IF (IERR.NE.0) THEN
            CALL I2STR(IERR,STR1,LE1)
            WRITE (*,50) OUT(1:LEOUT), STR1(1:LE1)
            RETURN
         END IF
      END IF
      ERROR = .FALSE.
      IU32 = IUOUT
C
10    FORMAT (/$(FC)'The output unit ',A,' cannot be opened. ',
     &        'IOSTAT = ',A,'.')
20    FORMAT (/$(FC)'Cannot inquire the output file "',A,'". ',
     &        'IOSTAT = ',A,'.')
30    FORMAT (/$(FC)'The output file "',A,'" is already open. ',
     &        'UNIT = ',A,'.')
40    FORMAT (/$(FC)'Cannot allocate unit number for the output ',
     &        'file "',A,'".')
50    FORMAT (/$(FC)'Cannot open the output file "',A,'". ',
     &        'IOSTAT = ',A,'.')
      END
C  ---------------------------------------------------------------------
      SUBROUTINE CLOUT
C  ---------------------------------------------------------------------
C
C  Description:
C    Closes the Merlin output unit/file.
C
C  ---------------------------------------------------------------------
      IMPLICIT $(TYPE) (A-H,O-Z)
C  Common blocks:
      COMMON /UNITS/ IUINP, IUOUT
      LOGICAL HEADER, IPRECO, OPRECO, HASAPP, OUTAPP
      COMMON /CP1/ HEADER, IPRECO, OPRECO, HASAPP, OUTAPP
      PARAMETER ( MXFILE = 80 )
      CHARACTER*(MXFILE) PROLO, EPILO, IN, OUT, HFILE, MFILE, PDFILE
      COMMON /CP2/ PROLO, EPILO, IN, OUT, HFILE, MFILE, PDFILE
C  ---------------------------------------------------------------------
C  Local variables:
C  ---------------------------------------------------------------------
      CHARACTER*10 STR1, STR2
C  ---------------------------------------------------------------------
      IF (OUT.EQ.' ') THEN
C  Output was written to a unit.
         IF (.NOT.OPRECO) THEN
            CLOSE (IUOUT,IOSTAT=IERR)
            IF (IERR.NE.0) THEN
               CALL I2STR(IERR,STR1,LE1)
               CALL I2STR(IUOUT,STR2,LE2)
               WRITE (*,20) STR2(1:LE2), STR1(1:LE1)
            END IF
         END IF
      ELSE
C  Output was written to a file.
C  Should we delete the output file ?
         CALL GETPPI('CONTROL','KEEPOUT',KEEP)
         IF (KEEP.EQ.0) THEN
            CLOSE (IUOUT,IOSTAT=IERR,STATUS='DELETE')
         ELSE
            CLOSE (IUOUT,IOSTAT=IERR,STATUS='KEEP')
         END IF
         IF (IERR.NE.0) THEN
            CALL I2STR(IERR,STR1,LE1)
            WRITE (*,20) OUT(1:LENGTH(OUT)), STR1(1:LE1)
         END IF
      END IF
C
10    FORMAT (/$(FC)'The Merlin output unit ',A,' could not be ',
     &        'closed properly. IOSTAT = ',A,'.')
20    FORMAT (/$(FC)'The Merlin output file "',A,'" could not be ',
     &        'closed properly. IOSTAT = ',A,'.')
      END
C  ---------------------------------------------------------------------
      SUBROUTINE OPIN ( ERROR )
C  ---------------------------------------------------------------------
C
C  Description:
C    Opens the Merlin input unit/file.
C
C  Output arguments:
C    ERROR      If an error occurs, ERROR is set to .true.; .false.
C               otherwise.
C
C  ---------------------------------------------------------------------
      IMPLICIT $(TYPE) (A-H,O-Z)
C  Arguments:
      LOGICAL ERROR
C  ---------------------------------------------------------------------
C  Common blocks:
      COMMON /UNITS/ IUINP, IUOUT
      COMMON /UNTSDF/ IU31, IU32
      LOGICAL HEADER, IPRECO, OPRECO, HASAPP, OUTAPP
      COMMON /CP1/ HEADER, IPRECO, OPRECO, HASAPP, OUTAPP
      PARAMETER ( MXFILE = 80 )
      CHARACTER*(MXFILE) PROLO, EPILO, IN, OUT, HFILE, MFILE, PDFILE
      COMMON /CP2/ PROLO, EPILO, IN, OUT, HFILE, MFILE, PDFILE
C  ---------------------------------------------------------------------
C  Local variables:
      LOGICAL EXI, OPE
      CHARACTER*10 STR1, STR2
C  ---------------------------------------------------------------------
      ERROR = .TRUE.
C
      IF (IN.EQ.' ') THEN
C  Input is read from a unit.
         IF (.NOT.IPRECO) THEN
            OPEN (IUINP,IOSTAT=IERR)
            IF (IERR.NE.0) THEN
               CALL I2STR(IUINP,STR1,LE1)
               CALL I2STR(IERR,STR2,LE2)
               WRITE (IUOUT,10) STR1(1:LE1), STR2(1:LE2)
               RETURN
            END IF
         END IF
      ELSE
C  Input is read from a file.
         INQUIRE (FILE=IN,EXIST=EXI,OPENED=OPE,NUMBER=IU,IOSTAT=IERR)
         IF (IERR.NE.0) THEN
            CALL I2STR(IERR,STR1,LE1)
            WRITE (IUOUT,20) IN(1:LENGTH(IN)), STR1(1:LE1)
            RETURN
         END IF
C
C  Does the input file exist ?
         IF (.NOT.EXI) THEN
            WRITE (IUOUT,60) IN(1:LENGTH(IN))
            RETURN
         END IF
C
C  Is the input file already open ?
         IF (OPE) THEN
            CALL I2STR(IU,STR1,LE1)
            WRITE (IUOUT,30) IN(1:LENGTH(IN)), STR1(1:LE1)
            RETURN
         END IF
C
C  Allocate unit for the input file.
         IUINP = NUNIT()
         IF (IUINP.EQ.-1) THEN
            WRITE (IUOUT,40) IN(1:LENGTH(IN))
            RETURN
         END IF
C
C  Open the input file.
         OPEN (IUINP,FILE=IN,IOSTAT=IERR)
         IF (IERR.NE.0) THEN
            CALL I2STR(IERR,STR1,LE1)
            WRITE (IUOUT,50) IN(1:LENGTH(IN)), STR1(1:LE1)
            RETURN
         END IF
         REWIND IUINP
      END IF
      ERROR = .FALSE.
      IU31 = IUINP
C
10    FORMAT (/$(FC)'The input unit ',A,' cannot be opened. ',
     &        'IOSTAT = ',A,'.')
20    FORMAT (/$(FC)'Cannot inquire the input file "',A,'". ',
     &        'IOSTAT = ',A,'.')
30    FORMAT (/$(FC)'The input file "',A,'" is already open. ',
     &        'UNIT = ',A,'.')
40    FORMAT (/$(FC)'Cannot allocate unit number for the input ',
     &        'file "',A,'".')
50    FORMAT (/$(FC)'Cannot open the input file "',A,'". ',
     &        'IOSTAT = ',A,'.')
60    FORMAT (/$(FC)'The input file "',A,'" does not exist. ')
      END
C  ---------------------------------------------------------------------
      SUBROUTINE FCLOSE
C  ---------------------------------------------------------------------
C
C  Description:
C    Closes the Merlin input unit/file. If any of the backup, history
C    and memo files are open, they are closed as well.
C
C  ---------------------------------------------------------------------
      IMPLICIT $(TYPE) (A-H,O-Z)
C  Common blocks:
      COMMON /UNITS/ IUINP, IUOUT
      COMMON /MEMO2/ MEMOMO, MEMOKE, MEMREC, IUMEM
      COMMON /BACKIT/ IUBA, IUBAMO, IBASA
      COMMON /LOGIT/ IULOG
      LOGICAL HEADER, IPRECO, OPRECO, HASAPP, OUTAPP
      COMMON /CP1/ HEADER, IPRECO, OPRECO, HASAPP, OUTAPP
      PARAMETER ( MXFILE = 80 )
      CHARACTER*(MXFILE) PROLO, EPILO, IN, OUT, HFILE, MFILE, PDFILE
      COMMON /CP2/ PROLO, EPILO, IN, OUT, HFILE, MFILE, PDFILE
C  ---------------------------------------------------------------------
C  Local variables:
      CHARACTER*10 STR1, STR2
      CHARACTER*(MXFILE) FILE
C  ---------------------------------------------------------------------
C
      IF (IN.EQ.' ') THEN
C  Input was read from a unit.
         IF (.NOT.IPRECO) THEN
            CLOSE (IUINP,IOSTAT=IERR)
            IF (IERR.NE.0) THEN
               CALL I2STR(IERR,STR1,LE1)
               CALL I2STR(IUINP,STR2,LE2)
               WRITE (IUOUT,20) STR2(1:LE2), STR1(1:LE1)
            END IF
         END IF
      ELSE
C  Input was read from a file.
C  Should we delete the input file ?
         CALL GETPPI('CONTROL','KEEPIN',KEEP)
         IF (KEEP.EQ.0) THEN
            CLOSE (IUINP,STATUS='DELETE',IOSTAT=IERR)
         ELSE
            CLOSE (IUINP,STATUS='KEEP',IOSTAT=IERR)
         END IF
         CLOSE (IUINP,IOSTAT=IERR)
         IF (IERR.NE.0) THEN
            CALL I2STR(IERR,STR1,LE1)
            WRITE (IUOUT,20) IN(1:LENGTH(IN)), STR1(1:LE1)
         END IF
      END IF
C
C  Close the memo file.
      IF (IUMEM.NE.-1) THEN
         CLOSE (IUMEM,IOSTAT=IERR)
         IF (IERR.NE.0) THEN
            CALL U2NAME(IUMEM,'memo',FILE,LEF)
            CALL I2STR(IERR,STR1,LE1)
            WRITE (IUOUT,30) 'memo', FILE(1:LEF), STR1(1:LE1)
         END IF
      END IF
C
C  Close the backup file.
      IF (IUBA.NE.-1) THEN
         CLOSE (IUBA,IOSTAT=IERR)
         IF (IERR.NE.0) THEN
            CALL U2NAME(IUBA,'backup',FILE,LEF)
            CALL I2STR(IERR,STR1,LE1)
            WRITE (IUOUT,30) 'backup', FILE(1:LEF), STR1(1:LE1)
         END IF
      END IF
C
C  Close the history file.
      IF (IULOG.NE.-1) THEN
         CLOSE (IULOG,IOSTAT=IERR)
         IF (IERR.NE.0) THEN
            CALL U2NAME(IULOG,'history',FILE,LEF)
            CALL I2STR(IERR,STR1,LE1)
            WRITE (IUOUT,30) 'history', FILE(1:LEF), STR1(1:LE1)
         END IF
      END IF
C
10    FORMAT (/$(FC)'The Merlin input unit ',A,' could not be ',
     &        'closed properly. IOSTAT = ',A,'.')
20    FORMAT (/$(FC)'The Merlin input file "',A,'" could not be ',
     &        'closed properly. IOSTAT = ',A,'.')
30    FORMAT (/$(FC)'The ',A,' file "',A,'" could not be closed ',
     &        'properly. IOSTAT = ',A,'.')
      END
C  ---------------------------------------------------------------------
      SUBROUTINE OPENGF ( FILE, IERREX, NU )
C  ---------------------------------------------------------------------
C
C  Description:
C    Opens a file and returns the correspondig unit number.
C
C  Input arguments:
C    FILE       The file to open.
C    IERREX     IERREX = 0 -> No error if the file does not exist.
C               IERREX = 1 -> The file must exist, otherwise an error
C                             is issued.
C
C  Output arguments:
C    NU         Unit number of the open file. NU is set to -1 if any
C               error occurs.
C
C  ---------------------------------------------------------------------
C  Arguments:
      CHARACTER*(*) FILE
C  ---------------------------------------------------------------------
C  Common blocks:
      LOGICAL STRO, WEAK
      COMMON /PRILOG/ STRO, WEAK
      COMMON /UNITS/ IUINP, IUOUT
C  ---------------------------------------------------------------------
C  Local variables:
      LOGICAL OPE, EXI
C  ---------------------------------------------------------------------
C
      NU = -1
      LEF = LENGTH(FILE)
      IF (LEF.EQ.0) THEN
         IF (WEAK) WRITE (IUOUT,80)
         RETURN
      END IF
C
C  Get info.
      INQUIRE (FILE=FILE,EXIST=EXI,OPENED=OPE,NUMBER=NUM,IOSTAT=IOE)
      IF (IOE.NE.0) THEN
         CALL ERRINQ(FILE,IOE)
         RETURN
      END IF
C
C  Does it exist ?
      IF (IERREX.NE.0) THEN
         IF (.NOT.EXI) THEN
            CALL ERREXI(FILE)
            RETURN
         END IF
      END IF
C
C  Is it open ?
      IF (OPE) THEN
         CALL ERRBUS(FILE,NUM)
         RETURN
      END IF
C
C  Allocate unit number.
      NU = NUNIT()
      IF (NU.EQ.-1) THEN
         CALL ERRUAL(FILE)
         RETURN
      END IF
C
C  Open the file.
      OPEN (NU,FILE=FILE,IOSTAT=IOE)
      IF (IOE.NE.0) THEN
         CALL ERROPE(FILE,NU)
         NU = -1
         RETURN
      END IF
      REWIND NU
C
80    FORMAT (/$(FC)'You must specify a file name.')
      END
C  ---------------------------------------------------------------------
      SUBROUTINE OPENIT ( FILE, TYPE, IPOS, ICREA, NU, NREC )
C  ---------------------------------------------------------------------
C
C  Description:
C    Opens a Merlin file for disposition of Merlin records.
C    Merlin uses two types of files:
C
C      1) Formatted, sequential files. These are usualy refered to as
C         TEXT files. Storing information in such a file is slow and
C         may result in loss of accuraccy when reading the information
C         back in. It is however the only portable way of transfering
C         data accross different computer platforms.
C      2) Unformatted, direct access files. These are usualy refered
C         to as BINARY files. Storing information in such a file is
C         faster and accuraccy is not lost. Data cannot be transfered
C         however, accross different computer platforms.
C
C  Input arguments:
C    FILE       Character variable. The file to open.
C    TYPE       Determines the type of the file:
C               TYPE = 'TEXT' for a text file.
C               TYPE = 'BIN'  for a binary file.
C    IPOS       Determines how the file will be positioned after it
C               is opened (significant only for text files):
C               IPOS = 0  -> rewind the file.
C               IPOS = 0  -> position at End-of-File.
C    ICREA      Determines what action is to be taken if the file
C               does not exist:
C               ICREA = 0 -> Do not create the file and return an
C                            error condition.
C               ICREA = 1 -> Create the file.
C
C  Output arguments:
C    NU         The unit number of the open file. If any error occurs,
C               NU is set to -1.
C    NREC       Number of Merlin records present in the file.
C
C  ---------------------------------------------------------------------
C  Arguments:
      CHARACTER*(*) FILE, TYPE
C  ---------------------------------------------------------------------
C  Common blocks:
      LOGICAL STRO, WEAK
      COMMON /PRILOG/ STRO, WEAK
      COMMON /UNITS/ IUINP, IUOUT
      COMMON /DIMEN/ NOD, NOF
C  ---------------------------------------------------------------------
C  Local variables:
      LOGICAL OPE, EXI
      CHARACTER*10 STR
C  ---------------------------------------------------------------------
C
C  Cannot use an empty string.
      LENF = LENGTH(FILE)
      IF (LENF.EQ.0) THEN
         IF (STRO) WRITE (IUOUT,40)
         NU = -1
         RETURN
      END IF
C
C  Get info about the file.
      INQUIRE (FILE=FILE(1:LENF),EXIST=EXI,OPENED=OPE,NUMBER=NUM,
     &         IOSTAT=IERR)
      IF (IERR.NE.0) THEN
         CALL ERRINQ(FILE,IERR)
         NU = -1
         RETURN
      END IF
C
C  Does it exist ?
      IF (.NOT.EXI .AND. ICREA.EQ.0) THEN
         CALL ERREXI(FILE)
         NU = -1
         RETURN
      END IF
C
C  Is it already open ?
      IF (OPE) THEN
         CALL ERRBUS(FILE,NUM)
         NU = -1
         RETURN
      END IF
C
C  Allocate a unit number for the file.
      NU = NUNIT()
      IF (NU.EQ.-1) THEN
         CALL ERRUAL(FILE)
         RETURN
      END IF
C
C  Open the file according to the specified mode.
      IF (ISCOMP(TYPE,'TEXT').NE.0) THEN
         MODE = 1
      ELSE
         MODE = 2
      END IF
C
      IF (MODE.EQ.1) THEN
         OPEN (NU,FILE=FILE(1:LENF),IOSTAT=IERR)
      ELSE
         LENREC = LENDIR()
         OPEN (NU,FILE=FILE(1:LENF),ACCESS='DIRECT',FORM='UNFORMATTED',
     &            RECL=LENREC,IOSTAT=IERR)
      END IF
C
C  Any errors during OPEN ?
      IF (IERR.NE.0) THEN
         CALL ERROPE(FILE,IERR)
         NU = -1
         RETURN
      END IF
C
C  If the file exists read in number of variables.
      IF (EXI) THEN
         IF (MODE.EQ.1) THEN
            REWIND NU
            READ (NU,*,END=10,IOSTAT=IERR) NVAR
            NREC = 0
         ELSE
            READ (NU,REC=1,IOSTAT=IERR) NVAR, NREC
         END IF
C
C  Any error during READ ?
         IF (IERR.NE.0) THEN
            CALL ERRRD(FILE,IERR)
            CLOSE (NU)
            NU = -1
            RETURN
         END IF
C
C  The number of variables in the file should match the dimenionality
C  of the current minimization problem.
         IF (NVAR.NE.NOD) THEN
            CALL I2STR(NVAR,STR,LES)
            IF (STRO) WRITE (IUOUT,30) FILE(1:LENF), STR(1:LES)
            CLOSE (NU)
            NU = -1
            RETURN
         END IF
C
C  Everything ok. Position the file if necessary.
         IF (MODE.EQ.1 .AND. IPOS.NE.0) CALL POSIT(NU)
      ELSE
C
C  The file does not exist. Create it.
         GOTO 10
      END IF
      RETURN
C
10    CONTINUE
C
C  Possibly an empty or completely new file.
      NREC = 0
      IF (MODE.EQ.1) THEN
         WRITE (NU,'(i6)') NOD
      ELSE
         WRITE (NU,REC=1) NOD, NREC
      END IF
C
40    FORMAT (/$(FC)'You must specify a file name.')
30    FORMAT (/$(FC)'File "',A,'" seems to contain points from a ',
     &        'problem with ',A,' variables.'
     &        /$(FC)'Use a discard command to ovewrite this file.')
      END
C  ---------------------------------------------------------------------
      SUBROUTINE OPMAC ( ICREA, INFO, IUMAC )
C  ---------------------------------------------------------------------
C
C  Description:
C    Subroutine OPMAC opens the Merlin macro file.
C
C  Input arguments:
C    ICREA      Determines the action to be taken for a non existent
C               macro file:
C               ICREA = 0 -> Do not create.
C               ICREA = 1 -> Create the file if it doesnt exist.
C
C  Output arguments:
C    INFO       Error indicator:
C               INFO = 0 -> Macro file opened.
C               INFO = 1 -> A macro file is not defined in the
C                           configuration file (MFILE = ' ') or if
C                           ICREA = 0, the macro file does not exist.
C               INFO = 2 -> Some error occurred during the file
C                           operations.
C    IUMAC      The unit number of the open macro file.
C
C  ---------------------------------------------------------------------
C  Common blocks:
      PARAMETER ( MXFILE = 80 )
      CHARACTER*(MXFILE) PROLO, EPILO, IN, OUT, HFILE, MFILE, PDFILE
      COMMON /CP2/ PROLO, EPILO, IN, OUT, HFILE, MFILE, PDFILE
C  ---------------------------------------------------------------------
C  Local variables:
      LOGICAL EXI, OPE
C  ---------------------------------------------------------------------
C
C  Is a macro file defined ?
      LEM = LENGTH(MFILE)
      IF (LEM.EQ.0) THEN
         INFO = 1
         RETURN
      END IF
C
C  Allocate a unit number for the macro file.
      INFO = 2
      IUMAC = NUNIT()
      IF (IUMAC.EQ.-1) THEN
         CALL ERRUAL(MFILE)
         RETURN
      END IF
C
C  Get info about the macro file.
      INQUIRE (FILE=MFILE,EXIST=EXI,OPENED=OPE,NUMBER=NUM,IOSTAT=IOE)
      IF (IOE.NE.0) THEN
         CALL ERRINQ(MFILE,IOE)
         RETURN
      END IF
C
      IF (EXI) THEN
C  The macro file exists.
         IF (OPE) THEN
C  The macro file is already open.
            CALL ERRBUS(MFILE,NUM)
         ELSE
C  Open the file.
            OPEN (IUMAC,FILE=MFILE,IOSTAT=IOE)
            IF (IOE.NE.0) THEN
               CALL ERROPE(MFILE,IOE)
            ELSE
               INFO = 0
            END IF
         END IF
      ELSE
C  The macro file does not exist.
         IF (ICREA.NE.0) THEN
C  Create it.
            OPEN (IUMAC,FILE=MFILE,IOSTAT=IOE)
            IF (IOE.NE.0) THEN
               CALL ERROPE(MFILE,IOE)
            ELSE
               INFO = 0
            END IF
         ELSE
            INFO = 1
         END IF
      END IF
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE POSIT ( IUNIT )
C  ---------------------------------------------------------------------
C
C  Description:
C    Positions the unit specified by IUNIT at the end of information.
C    Any subsequent WRITE statement will append to the file, while
C    any subsequent READ will cause an End-of-File condition.
C
C  Input arguments:
C    IUNIT      The unit number of the file to position. It is assumed
C               to belong to an open text file.
C
C  ---------------------------------------------------------------------
C  Local variables:
      CHARACTER TRIC
C  ---------------------------------------------------------------------
C
      REWIND IUNIT
C
C  NL counts the number of lines in the file.
      NL = 0
10    READ (IUNIT,'(A)',END=20) TRIC
      NL = NL+1
      GOTO 10
C
20    CONTINUE
      REWIND IUNIT
      DO 30,I=1,NL
30       READ (IUNIT,'(A)') TRIC
      END
C  ---------------------------------------------------------------------
      SUBROUTINE POSITC ( IUNIT, NREC, IEOF )
C  ---------------------------------------------------------------------
C
C  Description:
C    Positions the unit specified by IUNIT at the end of information.
C    Any subsequent WRITE statement will append to the file, while
C    any subsequent READ will cause an End-of-File condition.
C    In addition it returns the number of full Merlin records (NREC)
C    that are present in the file.
C    A Merlin record consists of NOD+1 lines. The first NOD lines
C    contain information about the minimization variables (point,
C    margins, fix status, etc.) and the last line contains the
C    functions value.
C
C  Input arguments:
C    IUNIT      The unit number of the file to position. It is assumed
C               to belong to an open text file.
C
C  Output arguments:
C    NREC       Full Merlin records contained in the file.
C    IEOF       IEOF = 0 when the file containes a complete set of
C                        Merlin records.
C               IEOF > 0 when the last Merlin record is incomplete.
C                        In this case IEOF is set to the number of
C                        lines of the last (incomlete) record.
C
C  ---------------------------------------------------------------------
C  Common blocks:
      COMMON /DIMEN/ NOD, NOF
C  ---------------------------------------------------------------------
C  Local variables:
      CHARACTER TRIC
C  ---------------------------------------------------------------------
C
C
C     Positions the unit specified by iunit at the end of information
C     returns number of records. Ieof=0 (all records complete)
C     ieof /= 0   (last record incomplete, ieof = lines present
C     in last record)
C
C
      REWIND IUNIT
C
C  NL counts the number of lines in the file.
      NL = 0
10    READ (IUNIT,'(A)',END=20) TRIC
      NL = NL+1
      GOTO 10
C
20    CONTINUE
      REWIND IUNIT
      DO 30,I=1,NL
30       READ (IUNIT,'(A)') TRIC
C
C  Calculate the number of full Merlin records NREC and the
C  remainder IEOF.
      NOD1 = NOD+1
      NL1 = NL-1
      NREC = NL1/NOD1
      IEOF = MOD(NL1,NOD1)
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE READER ( IUNIT, MODE, IREC, IEOF )
C  ---------------------------------------------------------------------
C
C  Description:
C    Subroutine READER reads a Merlin record from unit IUNIT. Text
C    and binary files can be read according to MODE. In case of a
C    text file it assumes that the file is correctly positioned.
C    Parameters read are becoming the current parameters (IXAT,
C    NICKNA, POINT, XLL, XRL) and the function is evaluated at the
C    new point.
C
C  Input arguments:
C    IUNIT      Unit to read from.
C    MODE       Determines file type:
C               MODE = 1 -> Text file.
C               MODE = 2 -> Binary file.
C    IREC       In case of a binary file (MODE=2) this is the Merlin
C               record to read. (Not the actual record in the
C               corresponding direct access file.)
C               IREC is ignored for a text file (MODE=1).
C
C  Output arguments:
C    IEOF       In case of a text file (MODE=1) IEOF signals an End-of-
C               File condition:
C               IEOF = 0 -> A Merlin record has been successfully read.
C               IEOF = 1 -> An End-of-File condition was encountered
C                           while reading from the file. Maybe there
C                           are no more Merlin records in the file, or
C                           the file is corrupt.
C               IEOF = 2 -> Some error in conversion. The file is
C                           corrupt.
C               IEOF is always 0 for a binary file.
C
C  ---------------------------------------------------------------------
      IMPLICIT $(TYPE) (A-H,O-Z)
C  Common blocks:
      PARAMETER ( MXV = $(MXV) )
      COMMON /PARAMS/ POINT(MXV), VAL
      COMMON /DETAIL/ IXAT(MXV), MARG(MXV)
      COMMON /XLIMS/ XLL(MXV), XRL(MXV)
      COMMON /INDEFI/ INDEF(MXV)
      COMMON /DIMEN/ NOD, NOF
      COMMON /MODE3/ MODEV, IFFORM, INMCL
C  ---------------------------------------------------------------------
C  Local variables:
      CHARACTER NAME*10
C  ---------------------------------------------------------------------
C
      IF (MODE.EQ.1) THEN
C  Read from a text file.
         DO 10,I=1,NOD
            CALL ROLDAT(IUNIT,NAME,IXAT(I),POINT(I),MARG(I),
     &                        XLL(I),XRL(I),IEOF)
            IF (IEOF.NE.0) RETURN
            CALL SETNAM(I,NAME,IERR)
            INDEF(I) = 1
10       CONTINUE
         IF (MODEV.EQ.1) THEN
            VAL = ACSQ(POINT,NOD)
         ELSE
            CALL RVSDAT(IUNIT,VAL,IEOF)
         END IF
      ELSE
C  Read from a binary file.
         IEOF = 0
         DO 20,I=1,NOD
            CALL RDIDAT(IUNIT,IREC,I,NAME,IXAT(I),POINT(I),
     &                        MARG(I),XLL(I),XRL(I))
            CALL SETNAM(I,NAME,IERR)
            INDEF(I) = 1
20       CONTINUE
         IF (MODEV.EQ.1) THEN
            VAL = ACSQ(POINT,NOD)
         ELSE
            CALL RVDDAT(IUNIT,IREC,VAL)
         END IF
      END IF
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE ROLDAT ( IU, NAME, IXA, POI, MA, XL, XR, IEOF )
C  ---------------------------------------------------------------------
C
C  Description:
C    Subroutine ROLDAT reads one line from a Merlin text file and
C    returns the data for a Merlin variable (value, fix status, name,
C    etc.)
C
C  Input arguments:
C    IU         Unit to read from.
C
C  Output arguments:
C    NAME       Name of the variable.
C    IXA        Fix status.
C    POI        Value of the variable.
C    MA         Margin indicator (-1,0,1)
C    XL         Left margin.
C    XR         Right margin.
C    IEOF       Error indicator:
C               IEOF = 0 -> End-of-file was not encountered.
C               IEOF = 1 -> End-of-file was encountered or the file
C                           contains incomplete Merlin records.
C               IEOF = 2 -> Error in conversion. The file most probably
C                           is corrupt.
C
C  ---------------------------------------------------------------------
      IMPLICIT $(TYPE) (A-H,O-Z)
C  Arguments:
      CHARACTER*(*) NAME
C  ---------------------------------------------------------------------
C  Common blocks:
      LOGICAL STRO, WEAK
      COMMON /PRILOG/ STRO, WEAK
      COMMON /UNITS/ IUINP, IUOUT
C  ---------------------------------------------------------------------
C  Local variables:
      PARAMETER ( MXFILE = 80 )
      CHARACTER*(MXFILE) FILE
      CHARACTER*11 MARL, MARR
      CHARACTER*3 CFIX
      CHARACTER*10 NICK
      CHARACTER*23 CPOI
C  ---------------------------------------------------------------------
C
C  Read one line.
      READ (IU,30,END=10) NICK, CFIX, CPOI, MARL, MARR
C
C  Decode point.
      CALL RTF2R(CPOI,POI,IPERR)
      IF (IPERR.NE.0) GOTO 100
C
C  Set name to return.
      NAME = NICK
C
C  Decode fix status.
      IF (CFIX.EQ.' * ') THEN
         IXA = 0
      ELSE IF (CFIX.EQ.'...') THEN
         IXA = 1
      ELSE
         GOTO 100
      END IF
C
C  Decode margins.
      IF (MARL.EQ.' ') THEN
         IF (MARR.EQ.' ') THEN
            MA = 0
         ELSE
            MA = 1
            READ (MARR,20,ERR=100) XR
         END IF
      ELSE
         IF (MARR.EQ.' ') THEN
            MA = -1
            READ (MARL,20,ERR=100) XL
         ELSE
            MA = 2
            READ (MARL,20,ERR=100) XL
            READ (MARR,20,ERR=100) XR
         END IF
      END IF
      IEOF = 0
      RETURN
C
C  End-of-File while reading.
10    CONTINUE
      IEOF = 1
      RETURN
C
C  Some error in conversion. The file is corrupt.
100   CONTINUE
      IF (WEAK) THEN
         CALL U2NAME(IU,' ',FILE,LEF)
         IF (LEF.EQ.0) THEN
            WRITE (IUOUT,50)
         ELSE
            WRITE (IUOUT,40) FILE(1:LEF)
         END IF
      END IF
      IEOF = 2
      RETURN
C
20    FORMAT (G11.0)
30    FORMAT (8X,A10,3X,A3,3X,A,2(2X,A11))
40    FORMAT (/$(FC)'*** Caution: File "',A,'" contains illegal data.')
50    FORMAT (/$(FC)'*** Caution: Illegal data in file.')
      END
C  ---------------------------------------------------------------------
      SUBROUTINE RDIDAT ( IU, IREC, IDX, NAME, IXA, POI, MA, XL, XR )
C  ---------------------------------------------------------------------
C
C  Description:
C    Subroutine RDIDAT reads one Merlin record from a binary file and
C    returns the data for a Merlin variable (value, fix status, name,
C    etc.)
C
C  Input arguments:
C    IU         Unit to read from.
C    IREC       The Merlin record to read.
C    IDX        The index of the variable whose data we want to read.
C
C  Output arguments:
C    NAME       Name of the variable.
C    IXA        Fix status.
C    POI        Value of the variable.
C    MA         Margin indicator (-1,0,1)
C    XL         Left margin.
C    XR         Right margin.
C
C  ---------------------------------------------------------------------
      IMPLICIT $(TYPE) (A-H,O-Z)
C  Arguments:
      CHARACTER*(*) NAME
C  ---------------------------------------------------------------------
C  Common blocks:
      COMMON /DIMEN/ NOD, NOF
C  ---------------------------------------------------------------------
C
      NR = 1+(IREC-1)*(NOD+1)+IDX
      READ (IU,REC=NR) NAME, IXA, POI, MA, XL, XR
      END
C  ---------------------------------------------------------------------
      SUBROUTINE RVSDAT ( IU, VAL, IEOF )
C  ---------------------------------------------------------------------
C
C  Description:
C    Subroutine RVSDAT will read the value of the objective function
C    from the Merlin text file with unit IU. The file is assumed to
C    be correctly positioned.
C
C  Input arguments:
C    IU         Unit to read from.
C
C  Output arguments:
C    VAL        The value just read.
C    IEOF       Error indicator:
C               IEOF = 0 -> End-of-file was not encountered.
C               IEOF = 1 -> End-of-file was encountered or the file
C                           contains incomplete Merlin records.
C               IEOF = 2 -> Error in conversion. The file most probably
C                           is corrupt.
C
C  ---------------------------------------------------------------------
      IMPLICIT $(TYPE) (A-H,O-Z)
C  Common blocks:
      LOGICAL STRO, WEAK
      COMMON /PRILOG/ STRO, WEAK
      COMMON /UNITS/ IUINP, IUOUT
C  ---------------------------------------------------------------------
C  Local variables:
      PARAMETER ( MXFILE = 80 )
      CHARACTER*(MXFILE) FILE
      CHARACTER*23 CVAL
C  ---------------------------------------------------------------------
C
      IEOF = 0
      READ (IU,20,END=10) CVAL
      CALL RTF2R(CVAL,VAL,IERR)
      IF (IERR.NE.0) THEN
         IF (WEAK) THEN
            CALL U2NAME(IU,' ',FILE,LEF)
            IF (LEF.EQ.0) THEN
               WRITE (IUOUT,50)
            ELSE
               WRITE (IUOUT,40) FILE(1:LEF)
            END IF
         END IF
         IEOF = 2
      END IF
      RETURN
C
C  End-of-file while reading.
10    CONTINUE
      IEOF = 1
C
20    FORMAT (24X,A)
40    FORMAT (/$(FC)'*** Caution: File "',A,'" contains illegal data.')
50    FORMAT (/$(FC)'*** Caution: Illegal data in file.')
      END
C  ---------------------------------------------------------------------
      SUBROUTINE RVDDAT ( IU, IREC, VAL )
C  ---------------------------------------------------------------------
C
C  Description:
C    Subroutine RVDDAT will read the value of the objective function
C    from the Merlin binary file with unit IU.
C
C  Input arguments:
C    IU         Unit to read from.
C    IREC       The Merlin record to read.
C
C  Output arguments:
C    VAL        The value just read.
C
C  ---------------------------------------------------------------------
      IMPLICIT $(TYPE) (A-H,O-Z)
C  Common blocks:
      COMMON /DIMEN/ NOD, NOF
C  ---------------------------------------------------------------------
C
      NR = 1+IREC*(NOD+1)
      READ (IU,REC=NR) VAL
      END
C  ---------------------------------------------------------------------
      SUBROUTINE U2NAME ( NU, DESC, FILE, LEF )
C  ---------------------------------------------------------------------
C
C  Description:
C    Ginen a unit number (NU), this routine will determine the name
C    of the corresponding open file.
C
C  Input arguments:
C    NU         Unit number of an open file.
C    DESC       Short (one word) description of the file, to be used
C               in error messages. Eg. memo, history input, etc.
C
C  Output arguments:
C    FILE       The name of the file attached to unit NU.
C    LEF        Number of useful characters in FILE. LEF is set to 0
C               if an error occurs, or if the file is not open.
C
C  ---------------------------------------------------------------------
C  Arguments:
      CHARACTER*(*) FILE, DESC
C  ---------------------------------------------------------------------
C  Common blocks:
      LOGICAL STRO, WEAK
      COMMON /PRILOG/ STRO, WEAK
      COMMON /UNITS/ IUINP, IUOUT
C  ---------------------------------------------------------------------
C  Local variables:
      CHARACTER*10 STR
      LOGICAL OPE
C  ---------------------------------------------------------------------
C
      INQUIRE (UNIT=NU,OPENED=OPE,NAME=FILE,IOSTAT=IOE)
      IF (IOE.NE.0) THEN
         IF (WEAK) THEN
            CALL I2STR(IOE,STR,LES)
            LED = LENGTH(DESC)
            IF (LED.EQ.0) THEN
               WRITE (IUOUT,10) STR(1:LES)
            ELSE
               WRITE (IUOUT,20) DESC(1:LED), STR(1:LES)
            END IF
         END IF
         LEF = 0
      ELSE
         IF (OPE) THEN
            LEF = LENGTH(FILE)
         ELSE
            LEF = 0
         END IF
      END IF
C
10    FORMAT (/$(FC)'Error: Cannot inquire file. IOSTAT = ',A,'.')
20    FORMAT (/$(FC)'Error: Cannot inquire the ',A,' file. ',
     &        'IOSTAT = ',A,'.')
      END
C  ---------------------------------------------------------------------
      SUBROUTINE WRITER ( X, VALUE, IFILE, MODE, IREC )
C  ---------------------------------------------------------------------
C
C  Description:
C    Subroutine WRITER writes to the file specified by IFILE, a Merlin
C    record using X as the values of the minimization variables and
C    VALUE as the corresponding function value. Subroutine WRITER
C    can write text anf binary files according to MODE.
C
C  Input arguments:
C    X          Values of the minimization variables.
C    VALUE      Value of the objective function.
C    IFILE      Unit to write to.
C    MODE       Determines file mode:
C               MODE = 1 -> Text file.
C               MODE = 2 -> Binary file.
C    IREC       For a binary file (MODE=2) this is the Merlin record
C               to write. IREC is ignored for a text file (MODE=1).
C
C  ---------------------------------------------------------------------
      IMPLICIT $(TYPE) (A-H,O-Z)
C  Arguments:
      DIMENSION X(*)
C  ---------------------------------------------------------------------
C  Common blocks:
      PARAMETER ( MXV = $(MXV) )
      COMMON /DETAIL/ IXAT(MXV), MARG(MXV)
      COMMON /XLIMS/ XLL(MXV), XRL(MXV)
      COMMON /DIMEN/ NOD, NOF
      CHARACTER*10 NICKNA
      COMMON /NAMES/ NICKNA(MXV)
      COMMON /TOTCAL/ ITNOCS, IPARTC, ITDER, IPDER, ITHESS, IPHESS,
     &                ITJAC, IPJAC
C  ---------------------------------------------------------------------
C
      IF (ITNOCS.EQ.0) RETURN
C
      IF (MODE.EQ.1) THEN
C  Write a text file.
         DO 10,I=1,NOD
            CALL WOLDAT(IFILE,I,NICKNA(I),X(I),MARG(I),
     &                        XLL(I),XRL(I))
10       CONTINUE
         CALL WVSDAT(IFILE,VALUE)
      ELSE
C  Write a binary file.
         DO 20,I=1,NOD
            CALL WDIDAT(IFILE,IREC,I,NICKNA(I),IXAT(I),X(I),
     &                        MARG(I),XLL(I),XRL(I))
20       CONTINUE
         CALL WVDDAT(IFILE,IREC,VALUE)
      END IF
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE WRITE2 ( INDE, NCH )
C  ---------------------------------------------------------------------
C
C  Description:
C    Subroutine WRITE2 will print on the output unit (IUOUT) a Merlin
C    record containing the current minimization variables, fix
C    statuses, names, etc. However, information will not be printed for
C    all variables. The first NCH positions of array INDE contain the
C    indices of the variables that are eligible for printing.
C
C  Input arguments:
C    INDE       Integer array. Holds the indices of the variables that
C               will be printed.
C    NCH        Number of indices stored in array INDE.
C
C  ---------------------------------------------------------------------
      IMPLICIT $(TYPE) (A-H,O-Z)
C  Arguments:
      INTEGER INDE(*)
C  ---------------------------------------------------------------------
C  Common blocks:
      PARAMETER ( MXV = $(MXV) )
      COMMON /DETAIL/ IXAT(MXV), MARG(MXV)
      COMMON /XLIMS/ XLL(MXV), XRL(MXV)
      COMMON /PARAMS/ POINT(MXV), VAL
      COMMON /UNITS/ IUINP, IUOUT
      CHARACTER*10 NICKNA
      COMMON /NAMES/ NICKNA(MXV)
      COMMON /TOTCAL/ ITNOCS, IPARTC, ITDER, IPDER, ITHESS, IPHESS,
     &                ITJAC, IPJAC
C  ---------------------------------------------------------------------
C
      IF (ITNOCS.EQ.0) RETURN
C
      DO 10,II=1,NCH
         I = INDE(II)
         CALL WOLDAT(IUOUT,I,NICKNA(I),POINT(I),MARG(I),
     &                  XLL(I),XRL(I))
10    CONTINUE
      CALL WVSDAT(IUOUT,VAL)
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE WOLDAT ( IU, IDX, NAME, POI, MA, XL, XR )
C  ---------------------------------------------------------------------
C
C  Description:
C    Subroutine WOLDAT will write on the text file specified by IU, a
C    line containing data (value, name, fix status etc.) for the
C    minimization variable with index IDX.
C
C  Input arguments:
C    IU         Unit to write to.
C    IDX        Index of the minimization variable.
C    NAME       Name of the variable.
C    POI        Value of the variable.
C    MA         Margin indicator (-1,0,1)
C    XL         Left margin.
C    XR         Right margin.
C
C  ---------------------------------------------------------------------
      IMPLICIT $(TYPE) (A-H,O-Z)
C  Arguments:
      CHARACTER*(*) NAME
C  ---------------------------------------------------------------------
C  Common blocks:
C  ---------------------------------------------------------------------
C  Local variables:
      CHARACTER*11 MARL, MARR
      CHARACTER*7 CFIX
      CHARACTER*10 NICK
      CHARACTER*23 CPOI
C  ---------------------------------------------------------------------
C
C  Prepare name.
      NICK = NAME
C
C  Prepare fix status.
      CALL FIXSTR(IDX,CFIX)
C
C  Prepare point.
      CALL R2RTF(POI,CPOI,LEP)
      MARL = ' '
      MARR = ' '
C
C  Write according to margins. Empty MARL and MARR stings are not
C  written thus saving disk space. (On unix/mac at least)
      IF (MA.EQ.-1) THEN
         WRITE (MARL,10) XL
         WRITE (IU,20) IDX, NICK, CFIX, CPOI, MARL
      ELSE IF (MA.EQ.1) THEN
         WRITE (MARR,10) XR
         WRITE (IU,20) IDX, NICK, CFIX, CPOI, MARL, MARR
      ELSE IF (MA.EQ.2) THEN
         WRITE (MARL,10) XL
         WRITE (MARR,10) XR
         WRITE (IU,20) IDX, NICK, CFIX, CPOI, MARL, MARR
      ELSE
         WRITE (IU,20) IDX, NICK, CFIX, CPOI
      END IF
C
10    FORMAT (1PG11.4)
20    FORMAT (1X,I5,') ',A10,1X,A,1X,A:,2X,A11:,2X,A11)
      END
C  ---------------------------------------------------------------------
      SUBROUTINE WDIDAT ( IU, IREC, IDX, NAME, IXA, POI, MA, XL, XR )
C  ---------------------------------------------------------------------
C
C  Description:
C    Subroutine WDIDAT will write on the binary file specified by IU,
C    a record containing data (value, name, fix status etc.) for the
C    minimization variable with index IDX.
C
C  Input arguments:
C    IU         Unit to write to.
C    IREC       Merlin record to write.
C    IDX        Index of the minimization variable.
C    NAME       Name of the variable.
C    IXA        Fix status.
C    POI        Value of the variable.
C    MA         Margin indicator (-1,0,1)
C    XL         Left margin.
C    XR         Right margin.
C
C  ---------------------------------------------------------------------
      IMPLICIT $(TYPE) (A-H,O-Z)
C  Arguments:
      CHARACTER*(*) NAME
C  ---------------------------------------------------------------------
C  Common blocks:
      COMMON /DIMEN/ NOD, NOF
C
C
      NR = 1+(IREC-1)*(NOD+1)+IDX
      WRITE (IU,REC=NR) NAME, IXA, POI, MA, XL, XR
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE WVSDAT ( IU, VALUE )
C  ---------------------------------------------------------------------
C
C  Description:
C    Subroutine WVSDAT will write on the text file specified by IU the
C    value of the objective function.
C
C  Input arguments:
C    IU         Unit to write to.
C    VALUE      Value of the objective function.
C
C  ---------------------------------------------------------------------
      IMPLICIT $(TYPE) (A-H,O-Z)
C  Local variables:
      CHARACTER*23 CVAL
C  ---------------------------------------------------------------------
C
      CALL R2RTF(VALUE,CVAL,LEV)
      WRITE (IU,10) 'Value ', CVAL(1:LEV)
C
10    FORMAT (4X,A10,7('.'),3X,A)
      END
C  ---------------------------------------------------------------------
      SUBROUTINE WVDDAT ( IU, IREC, VALUE )
C  ---------------------------------------------------------------------
C
C  Description:
C    Subroutine WVSDAT will write on the binary file specified by IU the
C    value of the objective function.
C
C  Input arguments:
C    IU         Unit to write to.
C    IREC       Merlin record to write to.
C    VALUE      Value of the objective function.
C
C  ---------------------------------------------------------------------
      IMPLICIT $(TYPE) (A-H,O-Z)
C  Common blocks:
      COMMON /DIMEN/ NOD, NOF
C  ---------------------------------------------------------------------
C
      NR = 1+IREC*(NOD+1)
      WRITE (IU,REC=NR) VALUE
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE CHANGE ( ICODE )
C  ---------------------------------------------------------------------
C
C  Description:
C    Subroutine CHANGE handles panel I/O for all panel commands. It
C    presents the panel, accepts any user changes to the panel
C    parameters, and stores the new values in the appropriate data
C    structures. Command line arguments are handled as well.
C
C    The expected input is:
C      KEY1 VAL1 KEY2 VAL2 ...
C      Keywords may be specified by their name, or by their index in
C      the panel.
C      Help for any keyword is issued if the corresponding value is a
C      a question mark ?
C      Help for all keywords is issued if the input consists of only
C      one item; a question mark ?
C
C    The following special constructs are recognized:
C      1) A semicolon ; at the end of the command line causes the panel
C         to be presented after all command line arguments have been
C         processed. Normally when command line arguments are entered,
C         the panel is not presented.
C      2) An exclamation mark ! at the end of the user input, causes
C         all user changes to be assigned to the appropriate keywords;
C         execution of the command however will be prohibited, by
C         setting ICODE to a non-zero value.
C
C  Output arguments:
C    ICODE      Error indicator:
C               ICODE =  0 -> Everything ok.
C               ICODE = -1 -> End-of-File on input.
C               Any other value: Some error.
C
C  ---------------------------------------------------------------------
      IMPLICIT $(TYPE) (A-H,O-Z)
C  Common blocks:
C  Maximum length of a keyword.
      PARAMETER ( KEYLEN = 10 )
C  Maximum number of panel commands.
      PARAMETER ( MXPCOM = 80 )
C  Maximum number of keywords in all panel commands.
      PARAMETER ( MXPKEY = 1000 )
C  Maximum storage for integer/real keywords.
      PARAMETER ( MXPIR = 5000 )
C  Maximum storage for string keywords and descriptions.
      PARAMETER ( MXPSTR = 20000 )
C  Maximum number of pointers in the above string storage.
      PARAMETER ( MXPPST = 5000 )
C
      CHARACTER*(KEYLEN) PCOM(MXPCOM)
      INTEGER            PONF(MXPCOM)
      INTEGER            KPOINT(MXPCOM)
      INTEGER            NCHELP(2*MXPCOM)
      CHARACTER*(KEYLEN) PKEY(MXPKEY)
      INTEGER            NKHELP(2*MXPKEY)
      CHARACTER          KTYPE(MXPKEY)
      INTEGER            TPOINT(MXPKEY)
      CHARACTER          KIO(MXPKEY)
      $(TYPE)            PIR(MXPIR)
      INTEGER            SPOINT(MXPPST)
      CHARACTER*(MXPSTR) PANSTR
      COMMON /PCOMS1/ KPOINT, NCHELP, NKHELP, TPOINT, SPOINT, PONF
      COMMON /PCOMS2/ PIR
      COMMON /PCOMS3/ PKEY, KTYPE, PANSTR, PCOM, KIO
C
C  NPCOM: Number of panel commands in the panel description file.
C  NPKEY: Number of keywords in all panel commands.
C  NPIR:  Number of positions filled in array PIR.
C  NPSP:  Number of positions filled in the char-pointer array SPOINT.
C  NPCH:  Number of character positions filled in PANSTR.
      COMMON /PCOMS4/ NPCOM, NPKEY, NPIR, NPSP, NPCH
      CHARACTER*10 CURCOM
      COMMON /CURC/ CURCOM
      COMMON /MODE3/ MODEV, IFFORM, INMCL
      LOGICAL STRO, WEAK
      COMMON /PRILOG/ STRO, WEAK
      COMMON /UNITS/ IUINP, IUOUT
C  ---------------------------------------------------------------------
C  Local variables:
      PARAMETER ( LINLEN = 80 )
      CHARACTER*(LINLEN) LINE, NEW
      DIMENSION IS(LINLEN/2), IE(LINLEN/2)
      CHARACTER*(KEYLEN) PNAME
C  ---------------------------------------------------------------------
C
      PNAME = CURCOM
C
C  Search for the panel.
      CALL PLOCP(PNAME,IC)
      IF (IC.EQ.0) THEN
C  The panel has not been found.
         IF (WEAK) WRITE (IUOUT,258) PNAME(1:LENGTH(PNAME))
         ICODE = -2
         RETURN
      END IF
C
C  Are we running MCL ?
      IF (INMCL.NE.0) THEN
         CALL MCLCHA(ICODE)
         RETURN
      END IF
C
120   CONTINUE
C  Did the user enter any command line arguments ?
      IPRESP = 1
      NOEXEC = 0
      CALL ARGNO(NARGS)
      CALL CMDARG(NARGS,LINE,LE1)
      IF (LINE(1:LE1).EQ.'!') THEN
         NOEXEC = 1
         NARGS = NARGS-1
      END IF
      IF (NARGS.GT.0) THEN
         NPRO = NARGS
         IPRESP = 0
         ITRIMS = 0
         CALL CMDARG(NARGS,LINE,LE1)
         IF (LINE(1:LE1).EQ.';') THEN
            NPRO = NARGS-1
            IPRESP = 1
         ELSE IF (LINE(LE1:LE1).EQ.';') THEN
            ITRIMS = 1
            IPRESP = 1
         END IF
C  Process command line arguments.
         KEYHLP = 0
         NTERR = 0
         IF (NPRO.EQ.1) THEN
            CALL CMDARG(1,LINE,LEK)
            IF (LINE.EQ.'?') THEN
               CALL KHELPA(IC)
               KEYHLP = KEYHLP+1
               NPRO = 0
            END IF
         END IF
         I = 0
510      CONTINUE
         I = I+1
         IF (I.LE.NPRO) THEN
            CALL CMDARG(I,LINE,LEK)
            IF (I+1.LE.NPRO) THEN
               CALL CMDARG(I+1,NEW,LEV)
               IF (I+1.EQ.NPRO .AND. ITRIMS.EQ.1) LEV = LEV-1
               CALL CHAN1(IC,LINE(1:LEK),LEK,NEW(1:LEV),LEV,
     &                             IERR,KEYHLP)
               IF (IERR.NE.0) NTERR = NTERR+1
            ELSE
               IF (WEAK) WRITE (IUOUT,253) LINE(1:LEK)
               NTERR = NTERR+1
            END IF
C  Process next keyword-value pair.
            I = I+1
            GOTO 510
         END IF
C  Did any errors occur during the above processing ?
         IF (IPRESP.EQ.0) THEN
            IF (NTERR.NE.0) THEN
               ICODE = -2
               RETURN
            ELSE IF (KEYHLP.NE.0) THEN
               ICODE = 1
               CALL DASH
               RETURN
            END IF
         END IF
      END IF
C
C  Noexec ! Entered ?
      IF (NOEXEC.NE.0) THEN
         ICODE = 2
         RETURN
      END IF
C
C  Do we need to present the panel ? was there a semicolon at the
C  end of the command line ?
      IF (IPRESP.EQ.0) THEN
         ICODE = 0
         RETURN
      END IF
C
C  Take in account the panel-mode.
      IF (PONF(IC).EQ.0) THEN
         ICODE = 0
         RETURN
      END IF
C
C  Present the panel.
      IF (STRO) THEN
         CALL PPR(IC)
         WRITE (IUOUT,'($(FC)''Enter changes:'')')
      END IF
C
C  Get and crack the changes.
      CALL INNOCR(LINE,.FALSE.,IERR)
C  An empty line ?
      IF (IERR.EQ.1) THEN
         ICODE = 0
         RETURN
C  EOF encountered ?
      ELSE IF (IERR.EQ.-1) THEN
         ICODE = -1
         RETURN
      END IF
      LL = LENGTH(LINE)
      CALL QLCR(LINE,LL,NEW,IS,IE,NIT,IERR)
      IF (IERR.NE.0) THEN
         IF (WEAK) WRITE (IUOUT,250)
         ICODE = -2
         RETURN
      END IF
      IF (NIT.EQ.0) THEN
         ICODE = 0
         RETURN
      END IF
C
C  Process the changes from the panel.
      KEYHLP = 0
      NTERR = 0
      NOEXEC = 0
      IF (NEW(IS(NIT):IE(NIT)).EQ.'!') THEN
         NOEXEC = 1
         NIT = NIT-1
      END IF
      IF (NIT.EQ.1) THEN
         IF (NEW(IS(1):IE(1)).EQ.'?') THEN
            CALL KHELPA(IC)
            KEYHLP = KEYHLP+1
            NIT = 0
         END IF
      END IF
      I = 0
500   CONTINUE
      I = I+1
      IF (I.LE.NIT) THEN
         IF (I+1.LE.NIT) THEN
            LEK = IE(I)-IS(I)+1
            LEV = IE(I+1)-IS(I+1)+1
            CALL CHAN1(IC,NEW(IS(I):IE(I)),LEK,
     &                       NEW(IS(I+1):IE(I+1)),LEV,IERR,KEYHLP)
            IF (IERR.NE.0) NTERR = NTERR+1
         ELSE
            IF (WEAK) WRITE (IUOUT,253) NEW(IS(I):IE(I))
            NTERR = NTERR+1
         END IF
C  Process next keyword-value pair.
         I = I+1
         GOTO 500
      END IF
C
C  Did any errors occur during the above processing ?
      IF (NTERR.NE.0) THEN
         ICODE = -2
      ELSE IF (KEYHLP.NE.0) THEN
         ICODE = 1
         CALL DASH
      ELSE IF (NOEXEC.NE.0) THEN
         ICODE = 2
      ELSE
         ICODE = 0
      END IF
C
250   FORMAT (/$(FC)'A closing quote is missing. Changes canceled.')
253   FORMAT (/$(FC)'Last item is obsolete: ',A)
258   FORMAT (/$(FC)'The ',A,' panel has not been defined in the panel'
     &        /$(FC)'description file.')
      END
C  ---------------------------------------------------------------------
      SUBROUTINE CHAN1 ( IC, KEY, LEK, VAL, LEV, IERR, KEYHLP )
C  ---------------------------------------------------------------------
C
C  Description:
C    Called by subroutine CHANGE. Assigns the value VAL to keyword KEY
C    in the panel with index IC.
C
C  Input arguments:
C    IC         Index of the current panel in the panel table.
C    KEY        The keyword.
C    LEK        Effective length of KEY.
C    VAL        The value to be assigned to KEY.
C    LEV        Effective length of VAL.
C
C  Input / Output arguments:
C    IERR       Error indicator.
C    KEYHLP     Help counter. Counts the keywords that help was issued
C               for.
C
C  ---------------------------------------------------------------------
      IMPLICIT $(TYPE) (A-H,O-Z)
C  Arguments:
      CHARACTER*(*) KEY, VAL
C  ---------------------------------------------------------------------
C  Common blocks:
C  Maximum length of a keyword.
      PARAMETER ( KEYLEN = 10 )
C  Maximum number of panel commands.
      PARAMETER ( MXPCOM = 80 )
C  Maximum number of keywords in all panel commands.
      PARAMETER ( MXPKEY = 1000 )
C  Maximum storage for integer/real keywords.
      PARAMETER ( MXPIR = 5000 )
C  Maximum storage for string keywords and descriptions.
      PARAMETER ( MXPSTR = 20000 )
C  Maximum number of pointers in the above string storage.
      PARAMETER ( MXPPST = 5000 )
C
      CHARACTER*(KEYLEN) PCOM(MXPCOM)
      INTEGER            PONF(MXPCOM)
      INTEGER            KPOINT(MXPCOM)
      INTEGER            NCHELP(2*MXPCOM)
      CHARACTER*(KEYLEN) PKEY(MXPKEY)
      INTEGER            NKHELP(2*MXPKEY)
      CHARACTER          KTYPE(MXPKEY)
      INTEGER            TPOINT(MXPKEY)
      CHARACTER          KIO(MXPKEY)
      $(TYPE)            PIR(MXPIR)
      INTEGER            SPOINT(MXPPST)
      CHARACTER*(MXPSTR) PANSTR
      COMMON /PCOMS1/ KPOINT, NCHELP, NKHELP, TPOINT, SPOINT, PONF
      COMMON /PCOMS2/ PIR
      COMMON /PCOMS3/ PKEY, KTYPE, PANSTR, PCOM, KIO
C
C  NPCOM: Number of panel commands in the panel description file.
C  NPKEY: Number of keywords in all panel commands.
C  NPIR:  Number of positions filled in array PIR.
C  NPSP:  Number of positions filled in the char-pointer array SPOINT.
C  NPCH:  Number of character positions filled in PANSTR.
      COMMON /PCOMS4/ NPCOM, NPKEY, NPIR, NPSP, NPCH
      LOGICAL STRO, WEAK
      COMMON /PRILOG/ STRO, WEAK
      COMMON /UNITS/ IUINP, IUOUT
C  ---------------------------------------------------------------------
C  Local variables:
      CHARACTER*10 STR
C  ---------------------------------------------------------------------
C
      IERR = 0
      KP1 = KPOINT(IC)
      NPK = KPOINT(IC+1)-KP1
C
      CALL TOINT(KEY,LEK,NN,NERR)
      IF (NERR.NE.0) THEN
         CALL PLOCK(IC,KEY(1:LEK),JKEY)
         IF (JKEY.EQ.0) THEN
            LE2 = LENGTH(PCOM(IC))
            IF (WEAK) WRITE (IUOUT,251)
     &                             KEY(1:LEK), PCOM(IC)(1:LE2)
            IERR = -1
            RETURN
         END IF
      ELSE IF (NN.LE.0) THEN
         LE2 = LENGTH(PCOM(IC))
         IF (WEAK) WRITE (IUOUT,251) KEY(1:LEK), PCOM(IC)(1:LE2)
         IERR = -1
         RETURN
      ELSE
         JJ = 0
         DO 10,I=KP1,KP1+NPK-1
            IF (KIO(I).EQ.' ') JJ = JJ+1
            IF (JJ.EQ.NN) THEN
               JKEY = I
               GOTO 11
            END IF
10       CONTINUE
C
         CALL I2STR(NPK,STR,LES)
         LE2 = LENGTH(PCOM(IC))
         IF (WEAK) WRITE (IUOUT,251) KEY(1:LEK), PCOM(IC)(1:LE2)
         IERR = -1
         RETURN
11       CONTINUE
      END IF
C
88    CONTINUE
      IF (KIO(JKEY).NE.' ') THEN
         LE2 = LENGTH(PCOM(IC))
         IF (WEAK) WRITE (IUOUT,251) KEY(1:LEK), PCOM(IC)(1:LE2)
         IERR = -1
         RETURN
      END IF
C
C  Does the user need help for this keyword ?
      IF (VAL.EQ.'?') THEN
         IF (KEYHLP.EQ.0) CALL DASH
         CALL KHELP(JKEY)
         KEYHLP = KEYHLP+1
         RETURN
      END IF
C
      IF (ISCOMP(KTYPE(JKEY),'I').NE.0) THEN
C  Keyword with an integer value.
         CALL TOINT(VAL,LEV,IVAL,NERR)
         IF (NERR.NE.0) THEN
            LE2 = LENGTH(PKEY(JKEY))
            IF (WEAK) WRITE (IUOUT,253)
     &                             VAL(1:LEV), PKEY(JKEY)(1:LE2)
            IERR = -1
            RETURN
         END IF
         CALL SETPI(PKEY(JKEY),IVAL,IERR)
      ELSE IF (ISCOMP(KTYPE(JKEY),'R').NE.0) THEN
C  Keyword with a real value.
         CALL TOREAL(VAL,LEV,XVAL,NERR)
         IF (NERR.NE.0) THEN
            LE2 = LENGTH(PKEY(JKEY))
            IF (WEAK) WRITE (IUOUT,254)
     &                             VAL(1:LEV), PKEY(JKEY)(1:LE2)
            IERR = -1
            RETURN
         END IF
         CALL SETPR(PKEY(JKEY),XVAL,IERR)
      ELSE IF (ISCOMP(KTYPE(JKEY),'S').NE.0 .OR.
     &         ISCOMP(KTYPE(JKEY),'C').NE.0) THEN
C  Keyword with a string value.
         CALL SETPS(PKEY(JKEY),VAL,LEV,IERR)
      END IF
C
251   FORMAT (/$(FC)'No such keyword: ',A,' in the ',A,' panel.')
253   FORMAT (/$(FC)'Incorrect integer value: ',A,' for the ',
     &        A,' keyword.')
254   FORMAT (/$(FC)'Incorrect real value: ',A,' for the ',
     &        A,' keyword.')
      END
C  ---------------------------------------------------------------------
      SUBROUTINE MCLCHA ( ICODE )
C  ---------------------------------------------------------------------
C
C  Description:
C    Called by subroutine CHANGE. Handles panel changes when an MCL
C    program is running.
C    The MOC file should have the form:
C      <Panel_Command>
C      <Number_of_Changes>    -> An integer >= 0.
C      <Keyword_1>
C      <Type_of_Keyword_1>    -> I, R, S or C.
C      <Keyword_2>
C      <Type_of_Keyword_2>    -> I, R, S or C.
C       ...
C      For an integer or real keyword the corresponding values are
C      popped from the stack.
C      For a string (type S or C) keyword the MOC file contains:
C      <Keyword>
C      <Type_of_Keyword>      -> S or C.
C      <Length_of_Value>      -> An integer > 0.
C      <String_Value>
C
C  Input arguments:
C    IC         Index of the panel in the panel table.
C
C  Output arguments:
C    ICODE      Error indicator.
C
C  ---------------------------------------------------------------------
      IMPLICIT $(TYPE) (A-H,O-Z)
C  Common blocks:
C  Maximum length of a keyword.
      PARAMETER ( KEYLEN = 10 )
C  Maximum number of panel commands.
      PARAMETER ( MXPCOM = 80 )
C  Maximum number of keywords in all panel commands.
      PARAMETER ( MXPKEY = 1000 )
C  Maximum storage for integer/real keywords.
      PARAMETER ( MXPIR = 5000 )
C  Maximum storage for string keywords and descriptions.
      PARAMETER ( MXPSTR = 20000 )
C  Maximum number of pointers in the above string storage.
      PARAMETER ( MXPPST = 5000 )
C
      CHARACTER*(KEYLEN) PCOM(MXPCOM)
      INTEGER            PONF(MXPCOM)
      INTEGER            KPOINT(MXPCOM)
      INTEGER            NCHELP(2*MXPCOM)
      CHARACTER*(KEYLEN) PKEY(MXPKEY)
      INTEGER            NKHELP(2*MXPKEY)
      CHARACTER          KTYPE(MXPKEY)
      INTEGER            TPOINT(MXPKEY)
      CHARACTER          KIO(MXPKEY)
      $(TYPE)            PIR(MXPIR)
      INTEGER            SPOINT(MXPPST)
      CHARACTER*(MXPSTR) PANSTR
      COMMON /PCOMS1/ KPOINT, NCHELP, NKHELP, TPOINT, SPOINT, PONF
      COMMON /PCOMS2/ PIR
      COMMON /PCOMS3/ PKEY, KTYPE, PANSTR, PCOM, KIO
C
C  NPCOM: Number of panel commands in the panel description file.
C  NPKEY: Number of keywords in all panel commands.
C  NPIR:  Number of positions filled in array PIR.
C  NPSP:  Number of positions filled in the char-pointer array SPOINT.
C  NPCH:  Number of character positions filled in PANSTR.
      COMMON /PCOMS4/ NPCOM, NPKEY, NPIR, NPSP, NPCH
      PARAMETER ( LSTA = $(MCLMEM) )
      COMMON /MCLV1/ STACK(LSTA)
      PARAMETER ( MAXMFI = 20 )
      COMMON /MCLV2/ ISTACK, INFILE, MBASE, INAD, MPC, MCLLNO, LENBUF,
     &               IBUF, MCLUNI(MAXMFI), NUMFI
C  ---------------------------------------------------------------------
C  Local variables:
      CHARACTER*(KEYLEN) KEY
      CHARACTER*80 LINE
      CHARACTER CKTYPE
C  ---------------------------------------------------------------------
C
C  Get number of changes.
      CALL ININ2(NC,NEX,ICODE)
      IF (ICODE.NE.0) RETURN
C
C  Loop over all changes.
      IR = 0
      DO 10,I=1,NC
C
C  Get keyword.
         CALL INNOCR(KEY,.FALSE.,ICODE)
         IF (ICODE.NE.0) RETURN
C
C  Get keyword type.
         CALL INNOCR(CKTYPE,.FALSE.,ICODE)
         IF (ICODE.NE.0) RETURN
C
C  Proceed according to argument type.
         IF (ISCOMP(CKTYPE,'I').NE.0) THEN
C  An integer keyword.
            IR = IR+1
            IVAL = NINT(STACK(ISTACK-NEX+IR))
            CALL SETPI(KEY,IVAL,ICODE)
         ELSE IF (ISCOMP(CKTYPE,'R').NE.0) THEN
C  A real keyword.
            IR = IR+1
            XVAL = STACK(ISTACK-NEX+IR)
            CALL SETPR(KEY,XVAL,ICODE)
         ELSE IF (ISCOMP(CKTYPE,'S').NE.0 .OR.
     &               ISCOMP(CKTYPE,'C').NE.0) THEN
C  A string keyword.
            CALL ININ(LL,ICODE)
            IF (ICODE.NE.0) RETURN
            CALL INNOCR(LINE,.FALSE.,ICODE)
            IF (ICODE.NE.0) RETURN
            CALL SETPS(KEY,LINE,LL,IERR)
         ELSE
            ICODE = -2
            RETURN
         END IF
10    CONTINUE
      ISTACK = ISTACK-NEX
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE KHELP ( IK )
C  ---------------------------------------------------------------------
C
C  Description:
C    Issues help for the keyword with index IK in the PKEY table.
C
C  Input arguments:
C    IK         Index of the keyword in the PKEY table.
C
C  ---------------------------------------------------------------------
      IMPLICIT $(TYPE) (A-H,O-Z)
C  Arguments:
C  ---------------------------------------------------------------------
C  Common blocks:
C  Maximum length of a keyword.
      PARAMETER ( KEYLEN = 10 )
C  Maximum number of panel commands.
      PARAMETER ( MXPCOM = 80 )
C  Maximum number of keywords in all panel commands.
      PARAMETER ( MXPKEY = 1000 )
C  Maximum storage for integer/real keywords.
      PARAMETER ( MXPIR = 5000 )
C  Maximum storage for string keywords and descriptions.
      PARAMETER ( MXPSTR = 20000 )
C  Maximum number of pointers in the above string storage.
      PARAMETER ( MXPPST = 5000 )
C
      CHARACTER*(KEYLEN) PCOM(MXPCOM)
      INTEGER            PONF(MXPCOM)
      INTEGER            KPOINT(MXPCOM)
      INTEGER            NCHELP(2*MXPCOM)
      CHARACTER*(KEYLEN) PKEY(MXPKEY)
      INTEGER            NKHELP(2*MXPKEY)
      CHARACTER          KTYPE(MXPKEY)
      INTEGER            TPOINT(MXPKEY)
      CHARACTER          KIO(MXPKEY)
      $(TYPE)            PIR(MXPIR)
      INTEGER            SPOINT(MXPPST)
      CHARACTER*(MXPSTR) PANSTR
      COMMON /PCOMS1/ KPOINT, NCHELP, NKHELP, TPOINT, SPOINT, PONF
      COMMON /PCOMS2/ PIR
      COMMON /PCOMS3/ PKEY, KTYPE, PANSTR, PCOM, KIO
C
C  NPCOM: Number of panel commands in the panel description file.
C  NPKEY: Number of keywords in all panel commands.
C  NPIR:  Number of positions filled in array PIR.
C  NPSP:  Number of positions filled in the char-pointer array SPOINT.
C  NPCH:  Number of character positions filled in PANSTR.
      COMMON /PCOMS4/ NPCOM, NPKEY, NPIR, NPSP, NPCH
      PARAMETER ( MXFILE = 80 )
      CHARACTER*(MXFILE) PROLO, EPILO, IN, OUT, HFILE, MFILE, PDFILE
      COMMON /CP2/ PROLO, EPILO, IN, OUT, HFILE, MFILE, PDFILE
      LOGICAL STRO, WEAK
      COMMON /PRILOG/ STRO, WEAK
      COMMON /UNITS/ IUINP, IUOUT
C  ---------------------------------------------------------------------
C
C  Open the help file.
      CALL OPHELP(PDFILE,NU,2)
      IF (NU.EQ.-1) RETURN
C
      LISTA = NKHELP(2*IK-1)
      LIEND = NKHELP(2*IK)
      IF (LISTA.EQ.0) THEN
         LEK = LENGTH(PKEY(IK))
         IF (STRO) WRITE (IUOUT,40) PKEY(IK)(1:LEK)
      ELSE
         IF (STRO) WRITE (IUOUT,*)
         CALL PCOPY(NU,LISTA,LIEND)
      END IF
      CLOSE (NU)
C
40    FORMAT (/$(FC)'Sorry, there is no help available for the "',A,
     &        '" keyword.')
      END
C  ---------------------------------------------------------------------
      SUBROUTINE KHELPA ( IC )
C  ---------------------------------------------------------------------
C
C  Description:
C    Issues help for all keywords for the panel with index IC.
C
C  Input arguments:
C    IC         Index of the panel in the panel table.
C
C  ---------------------------------------------------------------------
      IMPLICIT $(TYPE) (A-H,O-Z)
C  Arguments:
C  ---------------------------------------------------------------------
C  Common blocks:
C  Maximum length of a keyword.
      PARAMETER ( KEYLEN = 10 )
C  Maximum number of panel commands.
      PARAMETER ( MXPCOM = 80 )
C  Maximum number of keywords in all panel commands.
      PARAMETER ( MXPKEY = 1000 )
C  Maximum storage for integer/real keywords.
      PARAMETER ( MXPIR = 5000 )
C  Maximum storage for string keywords and descriptions.
      PARAMETER ( MXPSTR = 20000 )
C  Maximum number of pointers in the above string storage.
      PARAMETER ( MXPPST = 5000 )
C
      CHARACTER*(KEYLEN) PCOM(MXPCOM)
      INTEGER            PONF(MXPCOM)
      INTEGER            KPOINT(MXPCOM)
      INTEGER            NCHELP(2*MXPCOM)
      CHARACTER*(KEYLEN) PKEY(MXPKEY)
      INTEGER            NKHELP(2*MXPKEY)
      CHARACTER          KTYPE(MXPKEY)
      INTEGER            TPOINT(MXPKEY)
      CHARACTER          KIO(MXPKEY)
      $(TYPE)            PIR(MXPIR)
      INTEGER            SPOINT(MXPPST)
      CHARACTER*(MXPSTR) PANSTR
      COMMON /PCOMS1/ KPOINT, NCHELP, NKHELP, TPOINT, SPOINT, PONF
      COMMON /PCOMS2/ PIR
      COMMON /PCOMS3/ PKEY, KTYPE, PANSTR, PCOM, KIO
C
C  NPCOM: Number of panel commands in the panel description file.
C  NPKEY: Number of keywords in all panel commands.
C  NPIR:  Number of positions filled in array PIR.
C  NPSP:  Number of positions filled in the char-pointer array SPOINT.
C  NPCH:  Number of character positions filled in PANSTR.
      COMMON /PCOMS4/ NPCOM, NPKEY, NPIR, NPSP, NPCH
      PARAMETER ( MXFILE = 80 )
      CHARACTER*(MXFILE) PROLO, EPILO, IN, OUT, HFILE, MFILE, PDFILE
      COMMON /CP2/ PROLO, EPILO, IN, OUT, HFILE, MFILE, PDFILE
      LOGICAL STRO, WEAK
      COMMON /PRILOG/ STRO, WEAK
      COMMON /UNITS/ IUINP, IUOUT
C  ---------------------------------------------------------------------
C
C  Open the help file.
      CALL OPHELP(PDFILE,NU,2)
      IF (NU.EQ.-1) RETURN
      IKST = KPOINT(IC)
      IKEND = KPOINT(IC+1)-1
C
C  Loop over all keywords.
      CALL DASH
      DO 10,IK=IKST,IKEND
         IF (KIO(IK).NE.' ') GOTO 10
         LISTA = NKHELP(2*IK-1)
         LIEND = NKHELP(2*IK)
         IF (LISTA.EQ.0) THEN
            LEK = LENGTH(PKEY(IK))
            IF (STRO) WRITE (IUOUT,40) PKEY(IK)(1:LEK)
         ELSE
            IF (STRO) WRITE (IUOUT,*)
            CALL PCOPY(NU,LISTA,LIEND)
         END IF
10    CONTINUE
      CLOSE (NU)
C
40    FORMAT (/$(FC)'Sorry, there is no help available for the "',A,
     &        '" keyword.')
      END
C  ---------------------------------------------------------------------
      SUBROUTINE GETPI ( KEY, IVAL )
C  ---------------------------------------------------------------------
C
C  Description:
C    Subroutine GETPI is part of a set of subroutines that manipulate
C    the values assigned to the keywords of the various panels. It
C    returns the integer value IVAL assigned to keyword KEY in the
C    current panel. See the description of subroutine PARSP for a
C    discussion of the complex internal structure storing panel related
C    information.
C
C    This routine will abort if:
C       1) The panel name (CURCOM) is not defined in the panel
C          description file.
C       2) KEY is not a valid keyword for this panel.
C       3) KEY is not an integer keyword.
C    Normally none of these should happen.
C
C  Input arguments:
C    KEY        The keyword.
C
C  Output arguments:
C    IVAL       The value assigned to keyword KEY.
C
C  ---------------------------------------------------------------------
      IMPLICIT $(TYPE) (A-H,O-Z)
C  Arguments:
      CHARACTER*(*) KEY
C  ---------------------------------------------------------------------
C  Common blocks:
C  Maximum length of a keyword.
      PARAMETER ( KEYLEN = 10 )
C  Maximum number of panel commands.
      PARAMETER ( MXPCOM = 80 )
C  Maximum number of keywords in all panel commands.
      PARAMETER ( MXPKEY = 1000 )
C  Maximum storage for integer/real keywords.
      PARAMETER ( MXPIR = 5000 )
C  Maximum storage for string keywords and descriptions.
      PARAMETER ( MXPSTR = 20000 )
C  Maximum number of pointers in the above string storage.
      PARAMETER ( MXPPST = 5000 )
C
      CHARACTER*(KEYLEN) PCOM(MXPCOM)
      INTEGER            PONF(MXPCOM)
      INTEGER            KPOINT(MXPCOM)
      INTEGER            NCHELP(2*MXPCOM)
      CHARACTER*(KEYLEN) PKEY(MXPKEY)
      INTEGER            NKHELP(2*MXPKEY)
      CHARACTER          KTYPE(MXPKEY)
      INTEGER            TPOINT(MXPKEY)
      CHARACTER          KIO(MXPKEY)
      $(TYPE)            PIR(MXPIR)
      INTEGER            SPOINT(MXPPST)
      CHARACTER*(MXPSTR) PANSTR
      COMMON /PCOMS1/ KPOINT, NCHELP, NKHELP, TPOINT, SPOINT, PONF
      COMMON /PCOMS2/ PIR
      COMMON /PCOMS3/ PKEY, KTYPE, PANSTR, PCOM, KIO
C
C  NPCOM: Number of panel commands in the panel description file.
C  NPKEY: Number of keywords in all panel commands.
C  NPIR:  Number of positions filled in array PIR.
C  NPSP:  Number of positions filled in the char-pointer array SPOINT.
C  NPCH:  Number of character positions filled in PANSTR.
      COMMON /PCOMS4/ NPCOM, NPKEY, NPIR, NPSP, NPCH
      CHARACTER*10 CURCOM
      COMMON /CURC/ CURCOM
      CHARACTER*(KEYLEN) PREVPA
      COMMON /PRPA1/ PREVPA
      COMMON /PRPA2/ IPREVP
C  ---------------------------------------------------------------------
C  Local variables:
      CHARACTER*(KEYLEN) PNAME, KNAME
C  ---------------------------------------------------------------------
C
      PNAME = CURCOM
      KNAME = KEY
C
C  Look at the cache first.
      IF (ISCOMP(PREVPA,PNAME).NE.0) THEN
         I = IPREVP
      ELSE
C  Search for the panel.
         LE1 = LENGTH(PNAME)
         IF (LE1.EQ.0) CALL MABORT('Zero panel length.','GETPI')
         CALL PLOCP(PNAME(1:LE1),I)
         IF (I.EQ.0)
     &          CALL MABORT('Panel not found: '//PNAME(1:LE1),'GETPI')
C  Store in the cache.
         PREVPA = PNAME
         IPREVP = I
      END IF
C
C  Search for the keyword.
      LE1 = LENGTH(KNAME)
      IF (LE1.EQ.0) CALL MABORT('Zero keyword length.','GETPI')
      CALL PLOCK(I,KNAME(1:LE1),I)
      IF (I.EQ.0)
     &    CALL MABORT('Keyword not found: '//KNAME(1:LE1),'GETPI')
C
C  Get keyword type.
      IF (ISCOMP(KTYPE(I),'I').EQ.0)
     &    CALL MABORT('Not an integer keyword '//KNAME(1:LE1),'GETPI')
C
C  Get keyword value.
      IP = TPOINT(I)
      IRL = NINT(PIR(IP))
      IF (IRL.EQ.1) THEN
         IP = IP+4
      ELSE IF (IRL.EQ.2) THEN
         IP = IP+1
         NITE = NINT(PIR(IP))
         IP = IP+NITE
      END IF
      IP = IP+1
      IVAL = NINT(PIR(IP))
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE GETPR ( KEY, XVAL )
C  ---------------------------------------------------------------------
C
C  Description:
C    Subroutine GETPR is part of a set of subroutines that manipulate
C    the values assigned to the keywords of the various panels. It
C    returns the real value XVAL assigned to keyword KEY in the
C    current panel. See the description of subroutine PARSP for a
C    discussion of the complex internal structure storing panel related
C    information.
C
C    This routine will abort if:
C       1) The panel name (CURCOM) is not defined in the panel
C          description file.
C       2) KEY is not a valid keyword for this panel.
C       3) KEY is not a real keyword.
C    Normally none of these should happen.
C
C  Input arguments:
C    KEY        The keyword.
C
C  Output arguments:
C    XVAL       The value assigned to keyword KEY.
C
C  ---------------------------------------------------------------------
      IMPLICIT $(TYPE) (A-H,O-Z)
C  Arguments:
      CHARACTER*(*) KEY
C  ---------------------------------------------------------------------
C  Common blocks:
C  Maximum length of a keyword.
      PARAMETER ( KEYLEN = 10 )
C  Maximum number of panel commands.
      PARAMETER ( MXPCOM = 80 )
C  Maximum number of keywords in all panel commands.
      PARAMETER ( MXPKEY = 1000 )
C  Maximum storage for integer/real keywords.
      PARAMETER ( MXPIR = 5000 )
C  Maximum storage for string keywords and descriptions.
      PARAMETER ( MXPSTR = 20000 )
C  Maximum number of pointers in the above string storage.
      PARAMETER ( MXPPST = 5000 )
C
      CHARACTER*(KEYLEN) PCOM(MXPCOM)
      INTEGER            PONF(MXPCOM)
      INTEGER            KPOINT(MXPCOM)
      INTEGER            NCHELP(2*MXPCOM)
      CHARACTER*(KEYLEN) PKEY(MXPKEY)
      INTEGER            NKHELP(2*MXPKEY)
      CHARACTER          KTYPE(MXPKEY)
      INTEGER            TPOINT(MXPKEY)
      CHARACTER          KIO(MXPKEY)
      $(TYPE)            PIR(MXPIR)
      INTEGER            SPOINT(MXPPST)
      CHARACTER*(MXPSTR) PANSTR
      COMMON /PCOMS1/ KPOINT, NCHELP, NKHELP, TPOINT, SPOINT, PONF
      COMMON /PCOMS2/ PIR
      COMMON /PCOMS3/ PKEY, KTYPE, PANSTR, PCOM, KIO
C
C  NPCOM: Number of panel commands in the panel description file.
C  NPKEY: Number of keywords in all panel commands.
C  NPIR:  Number of positions filled in array PIR.
C  NPSP:  Number of positions filled in the char-pointer array SPOINT.
C  NPCH:  Number of character positions filled in PANSTR.
      COMMON /PCOMS4/ NPCOM, NPKEY, NPIR, NPSP, NPCH
      CHARACTER*10 CURCOM
      COMMON /CURC/ CURCOM
      CHARACTER*(KEYLEN) PREVPA
      COMMON /PRPA1/ PREVPA
      COMMON /PRPA2/ IPREVP
C  ---------------------------------------------------------------------
C  Local variables:
      CHARACTER*(KEYLEN) PNAME, KNAME
C  ---------------------------------------------------------------------
C
      PNAME = CURCOM
      KNAME = KEY
C
C  Look at the cache first.
      IF (ISCOMP(PREVPA,PNAME).NE.0) THEN
         I = IPREVP
      ELSE
C  Search for the panel.
         LE1 = LENGTH(PNAME)
         IF (LE1.EQ.0) CALL MABORT('Zero panel length.','GETPR')
         CALL PLOCP(PNAME(1:LE1),I)
         IF (I.EQ.0)
     &          CALL MABORT('Panel not found: '//PNAME(1:LE1),'GETPR')
C  Store in the cache.
         PREVPA = PNAME
         IPREVP = I
      END IF
C
C  Search for the keyword.
      LE1 = LENGTH(KNAME)
      IF (LE1.EQ.0) CALL MABORT('Zero keyword length.','GETPR')
      CALL PLOCK(I,KNAME(1:LE1),I)
      IF (I.EQ.0)
     &    CALL MABORT('Keyword not found: '//KNAME(1:LE1),'GETPR')
C
C  Get keyword type.
      IF (ISCOMP(KTYPE(I),'R').EQ.0)
     &    CALL MABORT('Not a real keyword '//KNAME(1:LE1),'GETPR')
C
C  Get keyword value.
      IP = TPOINT(I)
      IRL = NINT(PIR(IP))
      IF (IRL.EQ.1) THEN
         IP = IP+4
      ELSE IF (IRL.EQ.2) THEN
         IP = IP+1
         NITE = NINT(PIR(IP))
         IP = IP+NITE
      END IF
      IP = IP+1
      XVAL = PIR(IP)
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE GETPS ( KEY, SVAL, LES )
C  ---------------------------------------------------------------------
C
C  Description:
C    Subroutine GETPS is part of a set of subroutines that manipulate
C    the values assigned to the keywords of the various panels. It
C    returns the string value SVAL assigned to keyword KEY in the
C    current panel. See the description of subroutine PARSP for a
C    discussion of the complex internal structure storing panel related
C    information.
C
C    This routine will abort if:
C       1) The panel name (CURCOM) is not defined in the panel
C          description file.
C       2) KEY is not a valid keyword for this panel.
C       3) KEY is not a string keyword.
C    Normally none of these should happen.
C
C  Input arguments:
C    KEY        The keyword.
C
C  Output arguments:
C    SVAL       The value assigned to keyword KEY.
C    LES        Number of useful characters in SVAL.
C
C  ---------------------------------------------------------------------
      IMPLICIT $(TYPE) (A-H,O-Z)
C  Arguments:
      CHARACTER*(*) KEY, SVAL
C  ---------------------------------------------------------------------
C  Common blocks:
C  Maximum length of a keyword.
      PARAMETER ( KEYLEN = 10 )
C  Maximum number of panel commands.
      PARAMETER ( MXPCOM = 80 )
C  Maximum number of keywords in all panel commands.
      PARAMETER ( MXPKEY = 1000 )
C  Maximum storage for integer/real keywords.
      PARAMETER ( MXPIR = 5000 )
C  Maximum storage for string keywords and descriptions.
      PARAMETER ( MXPSTR = 20000 )
C  Maximum number of pointers in the above string storage.
      PARAMETER ( MXPPST = 5000 )
C
      CHARACTER*(KEYLEN) PCOM(MXPCOM)
      INTEGER            PONF(MXPCOM)
      INTEGER            KPOINT(MXPCOM)
      INTEGER            NCHELP(2*MXPCOM)
      CHARACTER*(KEYLEN) PKEY(MXPKEY)
      INTEGER            NKHELP(2*MXPKEY)
      CHARACTER          KTYPE(MXPKEY)
      INTEGER            TPOINT(MXPKEY)
      CHARACTER          KIO(MXPKEY)
      $(TYPE)            PIR(MXPIR)
      INTEGER            SPOINT(MXPPST)
      CHARACTER*(MXPSTR) PANSTR
      COMMON /PCOMS1/ KPOINT, NCHELP, NKHELP, TPOINT, SPOINT, PONF
      COMMON /PCOMS2/ PIR
      COMMON /PCOMS3/ PKEY, KTYPE, PANSTR, PCOM, KIO
C
C  NPCOM: Number of panel commands in the panel description file.
C  NPKEY: Number of keywords in all panel commands.
C  NPIR:  Number of positions filled in array PIR.
C  NPSP:  Number of positions filled in the char-pointer array SPOINT.
C  NPCH:  Number of character positions filled in PANSTR.
      COMMON /PCOMS4/ NPCOM, NPKEY, NPIR, NPSP, NPCH
      CHARACTER*10 CURCOM
      COMMON /CURC/ CURCOM
      CHARACTER*(KEYLEN) PREVPA
      COMMON /PRPA1/ PREVPA
      COMMON /PRPA2/ IPREVP
C  ---------------------------------------------------------------------
C  Local variables:
      CHARACTER*(KEYLEN) PNAME, KNAME
C  ---------------------------------------------------------------------
C
      PNAME = CURCOM
      KNAME = KEY
C
C  Look at the cache first.
      IF (ISCOMP(PREVPA,PNAME).NE.0) THEN
         I = IPREVP
      ELSE
C  Search for the panel.
         LE1 = LENGTH(PNAME)
         IF (LE1.EQ.0) CALL MABORT('Zero panel length.','GETPS')
         CALL PLOCP(PNAME(1:LE1),I)
         IF (I.EQ.0)
     &          CALL MABORT('Panel not found: '//PNAME(1:LE1),'GETPS')
C  Store in the cache.
         PREVPA = PNAME
         IPREVP = I
      END IF
C
C  Search for the keyword.
      LE1 = LENGTH(KNAME)
      IF (LE1.EQ.0) CALL MABORT('Zero keyword length.','GETPS')
      CALL PLOCK(I,KNAME(1:LE1),I)
      IF (I.EQ.0)
     &    CALL MABORT('Keyword not found: '//KNAME(1:LE1),'GETPS')
C
C  Get keyword type.
      IF (ISCOMP(KTYPE(I),'S').EQ.0 .AND. ISCOMP(KTYPE(I),'C').EQ.0)
     &    CALL MABORT('Not a string keyword '//KNAME(1:LE1),'GETPS')
C
C  Get the value.
      IP = TPOINT(I)
      IAL = NINT(PIR(IP))
      IF (IAL.EQ.1) THEN
      ELSE IF (IAL.EQ.2) THEN
         IP = IP+1
         NITE = NINT(PIR(IP))
         IP = IP+NITE
      END IF
      IP = IP+1
      IDEF = NINT(PIR(IP))
      IST = SPOINT(IDEF)
      IEND = SPOINT(IDEF+1)
      LES = IEND-IST+1
      LEC = LEN(SVAL)
      IF (LES.LE.LEC) THEN
         SVAL = PANSTR(IST:IEND)
      ELSE
C  Truncate, if string too long.
         SVAL = PANSTR(IST:IST+LEC-1)
         LES = LEC
      END IF
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE GETPPI ( PANEL, KEY, IVAL )
C  ---------------------------------------------------------------------
C
C  Description:
C    Subroutine GETPPI is part of a set of subroutines that manipulate
C    the values assigned to the keywords of the various panels. It
C    returns the integer value IVAL assigned to keyword KEY in the
C    panel specified by PANEL. See the description of subroutine PARSP
C    for a discussion of the complex internal structure storing panel
C    related information.
C
C    This routine will abort if:
C       1) The panel name PANEL is not defined in the panel
C          description file.
C       2) KEY is not a valid keyword for this panel.
C       3) KEY is not an integer keyword.
C    Normally none of these should happen.
C
C  Input arguments:
C    PANEL      The panel.
C    KEY        The keyword.
C
C  Output arguments:
C    IVAL       The value assigned to keyword KEY.
C
C  ---------------------------------------------------------------------
      IMPLICIT $(TYPE) (A-H,O-Z)
C  Arguments:
      CHARACTER*(*) PANEL, KEY
C  ---------------------------------------------------------------------
C  Common blocks:
C  Maximum length of a keyword.
      PARAMETER ( KEYLEN = 10 )
C  Maximum number of panel commands.
      PARAMETER ( MXPCOM = 80 )
C  Maximum number of keywords in all panel commands.
      PARAMETER ( MXPKEY = 1000 )
C  Maximum storage for integer/real keywords.
      PARAMETER ( MXPIR = 5000 )
C  Maximum storage for string keywords and descriptions.
      PARAMETER ( MXPSTR = 20000 )
C  Maximum number of pointers in the above string storage.
      PARAMETER ( MXPPST = 5000 )
C
      CHARACTER*(KEYLEN) PCOM(MXPCOM)
      INTEGER            PONF(MXPCOM)
      INTEGER            KPOINT(MXPCOM)
      INTEGER            NCHELP(2*MXPCOM)
      CHARACTER*(KEYLEN) PKEY(MXPKEY)
      INTEGER            NKHELP(2*MXPKEY)
      CHARACTER          KTYPE(MXPKEY)
      INTEGER            TPOINT(MXPKEY)
      CHARACTER          KIO(MXPKEY)
      $(TYPE)            PIR(MXPIR)
      INTEGER            SPOINT(MXPPST)
      CHARACTER*(MXPSTR) PANSTR
      COMMON /PCOMS1/ KPOINT, NCHELP, NKHELP, TPOINT, SPOINT, PONF
      COMMON /PCOMS2/ PIR
      COMMON /PCOMS3/ PKEY, KTYPE, PANSTR, PCOM, KIO
C
C  NPCOM: Number of panel commands in the panel description file.
C  NPKEY: Number of keywords in all panel commands.
C  NPIR:  Number of positions filled in array PIR.
C  NPSP:  Number of positions filled in the char-pointer array SPOINT.
C  NPCH:  Number of character positions filled in PANSTR.
      COMMON /PCOMS4/ NPCOM, NPKEY, NPIR, NPSP, NPCH
      CHARACTER*(KEYLEN) PREVPA
      COMMON /PRPA1/ PREVPA
      COMMON /PRPA2/ IPREVP
C  ---------------------------------------------------------------------
C  Local variables:
      CHARACTER*(KEYLEN) PNAME, KNAME
C  ---------------------------------------------------------------------
C
      PNAME = PANEL
      KNAME = KEY
C
C  Look at the cache first.
      IF (ISCOMP(PREVPA,PNAME).NE.0) THEN
         I = IPREVP
      ELSE
C  Search for the panel.
         LE1 = LENGTH(PNAME)
         IF (LE1.EQ.0) CALL MABORT('Zero panel length.','GETPPI')
         CALL PLOCP(PNAME(1:LE1),I)
         IF (I.EQ.0)
     &          CALL MABORT('Panel not found: '//PNAME(1:LE1),'GETPPI')
C  Store in the cache.
         PREVPA = PNAME
         IPREVP = I
      END IF
C
C  Search for the keyword.
      LE1 = LENGTH(KNAME)
      IF (LE1.EQ.0) CALL MABORT('Zero panel length.','GETPPI')
      CALL PLOCK(I,KNAME(1:LE1),I)
      IF (I.EQ.0)
     &    CALL MABORT('Keyword not found: '//KNAME(1:LE1),'GETPPI')
C
C  Get keyword type.
      IF (ISCOMP(KTYPE(I),'I').EQ.0)
     &    CALL MABORT('Not an integer keyword '//KNAME(1:LE1),'GETPPI')
C
C  Get keyword value.
      IP = TPOINT(I)
      IRL = NINT(PIR(IP))
      IF (IRL.EQ.1) THEN
         IP = IP+4
      ELSE IF (IRL.EQ.2) THEN
         IP = IP+1
         NITE = NINT(PIR(IP))
         IP = IP+NITE
      END IF
      IP = IP+1
      IVAL = NINT(PIR(IP))
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE GETPPR ( PANEL, KEY, XVAL )
C  ---------------------------------------------------------------------
C
C  Description:
C    Subroutine GETPPR is part of a set of subroutines that manipulate
C    the values assigned to the keywords of the various panels. It
C    returns the real value XVAL assigned to keyword KEY in the
C    panel specified by PANEL. See the description of subroutine PARSP
C    for a discussion of the complex internal structure storing panel
C    related information.
C
C    This routine will abort if:
C       1) The panel name PANEL is not defined in the panel
C          description file.
C       2) KEY is not a valid keyword for this panel.
C       3) KEY is not a real keyword.
C    Normally none of these should happen.
C
C  Input arguments:
C    PANEL      The panel.
C    KEY        The keyword.
C
C  Output arguments:
C    IVAL       The value assigned to keyword KEY.
C
C  ---------------------------------------------------------------------
      IMPLICIT $(TYPE) (A-H,O-Z)
C  Arguments:
      CHARACTER*(*) PANEL, KEY
C  ---------------------------------------------------------------------
C  Common blocks:
C  Maximum length of a keyword.
      PARAMETER ( KEYLEN = 10 )
C  Maximum number of panel commands.
      PARAMETER ( MXPCOM = 80 )
C  Maximum number of keywords in all panel commands.
      PARAMETER ( MXPKEY = 1000 )
C  Maximum storage for integer/real keywords.
      PARAMETER ( MXPIR = 5000 )
C  Maximum storage for string keywords and descriptions.
      PARAMETER ( MXPSTR = 20000 )
C  Maximum number of pointers in the above string storage.
      PARAMETER ( MXPPST = 5000 )
C
      CHARACTER*(KEYLEN) PCOM(MXPCOM)
      INTEGER            PONF(MXPCOM)
      INTEGER            KPOINT(MXPCOM)
      INTEGER            NCHELP(2*MXPCOM)
      CHARACTER*(KEYLEN) PKEY(MXPKEY)
      INTEGER            NKHELP(2*MXPKEY)
      CHARACTER          KTYPE(MXPKEY)
      INTEGER            TPOINT(MXPKEY)
      CHARACTER          KIO(MXPKEY)
      $(TYPE)            PIR(MXPIR)
      INTEGER            SPOINT(MXPPST)
      CHARACTER*(MXPSTR) PANSTR
      COMMON /PCOMS1/ KPOINT, NCHELP, NKHELP, TPOINT, SPOINT, PONF
      COMMON /PCOMS2/ PIR
      COMMON /PCOMS3/ PKEY, KTYPE, PANSTR, PCOM, KIO
C
C  NPCOM: Number of panel commands in the panel description file.
C  NPKEY: Number of keywords in all panel commands.
C  NPIR:  Number of positions filled in array PIR.
C  NPSP:  Number of positions filled in the char-pointer array SPOINT.
C  NPCH:  Number of character positions filled in PANSTR.
      COMMON /PCOMS4/ NPCOM, NPKEY, NPIR, NPSP, NPCH
      CHARACTER*(KEYLEN) PREVPA
      COMMON /PRPA1/ PREVPA
      COMMON /PRPA2/ IPREVP
C  ---------------------------------------------------------------------
C  Local variables:
      CHARACTER*(KEYLEN) PNAME, KNAME
C  ---------------------------------------------------------------------
C
      PNAME = PANEL
      KNAME = KEY
C
C  Look at the cache first.
      IF (ISCOMP(PREVPA,PNAME).NE.0) THEN
         I = IPREVP
      ELSE
C  Search for the panel.
         LE1 = LENGTH(PNAME)
         IF (LE1.EQ.0) CALL MABORT('Zero panel length.','GETPPR')
         CALL PLOCP(PNAME(1:LE1),I)
         IF (I.EQ.0)
     &          CALL MABORT('Panel not found: '//PNAME(1:LE1),'GETPPR')
C  Store in the cache.
         PREVPA = PNAME
         IPREVP = I
      END IF
C
C  Search for the keyword.
      LE1 = LENGTH(KNAME)
      CALL PLOCK(I,KNAME(1:LE1),I)
      IF (LE1.EQ.0) CALL MABORT('Zero keyword length.','GETPPR')
      IF (I.EQ.0)
     &    CALL MABORT('Keyword not found: '//KNAME(1:LE1),'GETPPR')
C
C  Get keyword type.
      IF (ISCOMP(KTYPE(I),'R').EQ.0)
     &    CALL MABORT('Not a real keyword '//KNAME(1:LE1),'GETPPR')
C
C  Get keyword value.
      IP = TPOINT(I)
      IRL = NINT(PIR(IP))
      IF (IRL.EQ.1) THEN
         IP = IP+4
      ELSE IF (IRL.EQ.2) THEN
         IP = IP+1
         NITE = NINT(PIR(IP))
         IP = IP+NITE
      END IF
      IP = IP+1
      XVAL = PIR(IP)
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE GETPPS ( PANEL, KEY, SVAL, LES )
C  ---------------------------------------------------------------------
C
C  Description:
C    Subroutine GETPPS is part of a set of subroutines that manipulate
C    the values assigned to the keywords of the various panels. It
C    returns the string value SVAL assigned to keyword KEY in the
C    panel specified by PANEL. See the description of subroutine PARSP
C    for a discussion of the complex internal structure storing panel
C    related information.
C
C    This routine will abort if:
C       1) The panel name PANEL is not defined in the panel
C          description file.
C       2) KEY is not a valid keyword for this panel.
C       3) KEY is not a string keyword.
C    Normally none of these should happen.
C
C  Input arguments:
C    PANEL      The panel.
C    KEY        The keyword.
C
C  Output arguments:
C    SVAL       The value assigned to keyword KEY.
C    LES        Number of useful characters in SVAL.
C
C  ---------------------------------------------------------------------
      IMPLICIT $(TYPE) (A-H,O-Z)
C  Arguments:
      CHARACTER*(*) PANEL, KEY, SVAL
C  ---------------------------------------------------------------------
C  Common blocks:
C  Maximum length of a keyword.
      PARAMETER ( KEYLEN = 10 )
C  Maximum number of panel commands.
      PARAMETER ( MXPCOM = 80 )
C  Maximum number of keywords in all panel commands.
      PARAMETER ( MXPKEY = 1000 )
C  Maximum storage for integer/real keywords.
      PARAMETER ( MXPIR = 5000 )
C  Maximum storage for string keywords and descriptions.
      PARAMETER ( MXPSTR = 20000 )
C  Maximum number of pointers in the above string storage.
      PARAMETER ( MXPPST = 5000 )
C
      CHARACTER*(KEYLEN) PCOM(MXPCOM)
      INTEGER            PONF(MXPCOM)
      INTEGER            KPOINT(MXPCOM)
      INTEGER            NCHELP(2*MXPCOM)
      CHARACTER*(KEYLEN) PKEY(MXPKEY)
      INTEGER            NKHELP(2*MXPKEY)
      CHARACTER          KTYPE(MXPKEY)
      INTEGER            TPOINT(MXPKEY)
      CHARACTER          KIO(MXPKEY)
      $(TYPE)            PIR(MXPIR)
      INTEGER            SPOINT(MXPPST)
      CHARACTER*(MXPSTR) PANSTR
      COMMON /PCOMS1/ KPOINT, NCHELP, NKHELP, TPOINT, SPOINT, PONF
      COMMON /PCOMS2/ PIR
      COMMON /PCOMS3/ PKEY, KTYPE, PANSTR, PCOM, KIO
C
C  NPCOM: Number of panel commands in the panel description file.
C  NPKEY: Number of keywords in all panel commands.
C  NPIR:  Number of positions filled in array PIR.
C  NPSP:  Number of positions filled in the char-pointer array SPOINT.
C  NPCH:  Number of character positions filled in PANSTR.
      COMMON /PCOMS4/ NPCOM, NPKEY, NPIR, NPSP, NPCH
      CHARACTER*(KEYLEN) PREVPA
      COMMON /PRPA1/ PREVPA
      COMMON /PRPA2/ IPREVP
C  ---------------------------------------------------------------------
C  Local variables:
      CHARACTER*(KEYLEN) PNAME, KNAME
C  ---------------------------------------------------------------------
C
      PNAME = PANEL
      KNAME = KEY
C
C  Look at the cache first.
      IF (ISCOMP(PREVPA,PNAME).NE.0) THEN
         I = IPREVP
      ELSE
C  Search for the panel.
         LE1 = LENGTH(PNAME)
         IF (LE1.EQ.0) CALL MABORT('Zero panel length.','GETPPS')
         CALL PLOCP(PNAME(1:LE1),I)
         IF (I.EQ.0)
     &          CALL MABORT('Panel not found: '//PNAME(1:LE1),'GETPPS')
C  Store in the cache.
         PREVPA = PNAME
         IPREVP = I
      END IF
C
C  Search for the keyword.
      LE1 = LENGTH(KNAME)
      IF (LE1.EQ.0) CALL MABORT('Zero keyword length.','GETPPS')
      CALL PLOCK(I,KNAME(1:LE1),I)
      IF (I.EQ.0)
     &    CALL MABORT('Keyword not found: '//KNAME(1:LE1),'GETPPS')
C
C  Get keyword type.
      IF (ISCOMP(KTYPE(I),'S').EQ.0 .AND. ISCOMP(KTYPE(I),'C').EQ.0)
     &    CALL MABORT('Not a string keyword '//KNAME(1:LE1),'GETPPS')
C
C  Get keyword value.
      IP = TPOINT(I)
      IAL = NINT(PIR(IP))
      IF (IAL.EQ.1) THEN
      ELSE IF (IAL.EQ.2) THEN
         IP = IP+1
         NITE = NINT(PIR(IP))
         IP = IP+NITE
      END IF
      IP = IP+1
      IDEF = NINT(PIR(IP))
      IST = SPOINT(IDEF)
      IEND = SPOINT(IDEF+1)
      LES = IEND-IST+1
      LEC = LEN(SVAL)
      IF (LES.LE.LEC) THEN
         SVAL = PANSTR(IST:IEND)
      ELSE
C  Truncate, if string too long.
         SVAL = PANSTR(IST:IST+LEC-1)
         LES = LEC
      END IF
C
      END
C  ---------------------------------------------------------------------
      BLOCK DATA GSINI
C  ---------------------------------------------------------------------
C
C  Description:
C    Block data GSINI assigns initial values to common blocks PRPA1 and
C    PRPA2. They are used as cache for the GETP_/SETP_ set of
C    subroutines, storing the name and index of the panel used in
C    the most recent call of these subroutines. When one of GETP_/SETP_
C    is called, the cache is examined first, and if it matches the
C    current command, the subsequent lookup in the panel table is
C    skipped. Significant speedup occurs when a number of GETP_/SETP_
C    operations are performed on the same panel.
C
C  ---------------------------------------------------------------------
      IMPLICIT $(TYPE) (A-H,O-Z)
C  Common blocks:
C  Maximum length of a keyword.
      PARAMETER ( KEYLEN = 10 )
C  Maximum number of panel commands.
      PARAMETER ( MXPCOM = 80 )
C  Maximum number of keywords in all panel commands.
      PARAMETER ( MXPKEY = 1000 )
C  Maximum storage for integer/real keywords.
      PARAMETER ( MXPIR = 5000 )
C  Maximum storage for string keywords and descriptions.
      PARAMETER ( MXPSTR = 20000 )
C  Maximum number of pointers in the above string storage.
      PARAMETER ( MXPPST = 5000 )
C
      CHARACTER*(KEYLEN) PCOM(MXPCOM)
      INTEGER            PONF(MXPCOM)
      INTEGER            KPOINT(MXPCOM)
      INTEGER            NCHELP(2*MXPCOM)
      CHARACTER*(KEYLEN) PKEY(MXPKEY)
      INTEGER            NKHELP(2*MXPKEY)
      CHARACTER          KTYPE(MXPKEY)
      INTEGER            TPOINT(MXPKEY)
      CHARACTER          KIO(MXPKEY)
      $(TYPE)            PIR(MXPIR)
      INTEGER            SPOINT(MXPPST)
      CHARACTER*(MXPSTR) PANSTR
      COMMON /PCOMS1/ KPOINT, NCHELP, NKHELP, TPOINT, SPOINT, PONF
      COMMON /PCOMS2/ PIR
      COMMON /PCOMS3/ PKEY, KTYPE, PANSTR, PCOM, KIO
C
C  NPCOM: Number of panel commands in the panel description file.
C  NPKEY: Number of keywords in all panel commands.
C  NPIR:  Number of positions filled in array PIR.
C  NPSP:  Number of positions filled in the char-pointer array SPOINT.
C  NPCH:  Number of character positions filled in PANSTR.
      COMMON /PCOMS4/ NPCOM, NPKEY, NPIR, NPSP, NPCH
      CHARACTER*(KEYLEN) PREVPA
      COMMON /PRPA1/ PREVPA
      COMMON /PRPA2/ IPREVP
C  ---------------------------------------------------------------------
C
      DATA PREVPA / ' ' /
      DATA IPREVP / 0 /
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE OVER ( TABLE, ARRAY, SUB, MAXIM, PARAM )
C  ---------------------------------------------------------------------
C
C  Description:
C    When any of the Merlin tables overflows (especialy a panel-related
C    table), this routine will be called to issue informative messages
C    and stop execution of the program.
C
C  Input arguments:
C    TABLE      Character variable. The name of the table.
C    ARRAY      Character variable. This is the name of the array
C               implementing the table.
C    SUB        Character variable. This is the name of the routine
C               in which the overflow occurred.
C    MAXIM      The maximum number of entries allowed in the table.
C    PARAM      Character variable. This is the name of the parameter
C               that dimensions the table.
C
C  ---------------------------------------------------------------------
      IMPLICIT INTEGER (A-Z)
C  Arguments:
      CHARACTER*(*) TABLE, ARRAY, SUB, PARAM
C  ---------------------------------------------------------------------
C  Local variables:
      CHARACTER*6 STR
C  ---------------------------------------------------------------------
C
      CALL I2STR(MAXIM,STR,LES)
      WRITE (*,10) TABLE, SUB
      WRITE (*,20) ARRAY, STR(1:LES)
      WRITE (*,30) PARAM
C
      STOP 'OVRF'
C
10    FORMAT (/$(FC)'**** The ',A,' table has overflown in ',
     &        'routine ',A,'.')
20    FORMAT ($(FC)'**** Array ',A,' was assigned only ',A,
     &        ' storage locations.')
30    FORMAT ($(FC)'**** This array is dimensioned by the ',A,
     &        ' parameter.')
      END
C  ---------------------------------------------------------------------
      SUBROUTINE PPR ( IPC )
C  ---------------------------------------------------------------------
C
C  Description:
C    Subroutine PPR presents a panel. It prints all panel keywords,
C    their description, current and allowed values.
C
C  Input arguments:
C    IPC        Index of the panel in the panel table PCOM.
C
C  ---------------------------------------------------------------------
      IMPLICIT $(TYPE) (A-H,O-Z)
C  Common blocks:
C  Maximum length of a keyword.
      PARAMETER ( KEYLEN = 10 )
C  Maximum number of panel commands.
      PARAMETER ( MXPCOM = 80 )
C  Maximum number of keywords in all panel commands.
      PARAMETER ( MXPKEY = 1000 )
C  Maximum storage for integer/real keywords.
      PARAMETER ( MXPIR = 5000 )
C  Maximum storage for string keywords and descriptions.
      PARAMETER ( MXPSTR = 20000 )
C  Maximum number of pointers in the above string storage.
      PARAMETER ( MXPPST = 5000 )
C
      CHARACTER*(KEYLEN) PCOM(MXPCOM)
      INTEGER            PONF(MXPCOM)
      INTEGER            KPOINT(MXPCOM)
      INTEGER            NCHELP(2*MXPCOM)
      CHARACTER*(KEYLEN) PKEY(MXPKEY)
      INTEGER            NKHELP(2*MXPKEY)
      CHARACTER          KTYPE(MXPKEY)
      INTEGER            TPOINT(MXPKEY)
      CHARACTER          KIO(MXPKEY)
      $(TYPE)            PIR(MXPIR)
      INTEGER            SPOINT(MXPPST)
      CHARACTER*(MXPSTR) PANSTR
      COMMON /PCOMS1/ KPOINT, NCHELP, NKHELP, TPOINT, SPOINT, PONF
      COMMON /PCOMS2/ PIR
      COMMON /PCOMS3/ PKEY, KTYPE, PANSTR, PCOM, KIO
C
C  NPCOM: Number of panel commands in the panel description file.
C  NPKEY: Number of keywords in all panel commands.
C  NPIR:  Number of positions filled in array PIR.
C  NPSP:  Number of positions filled in the char-pointer array SPOINT.
C  NPCH:  Number of character positions filled in PANSTR.
      COMMON /PCOMS4/ NPCOM, NPKEY, NPIR, NPSP, NPCH
      COMMON /UNITS/ IUINP, IUOUT
C  ---------------------------------------------------------------------
C  Local variables:
      CHARACTER CIDX*3, CKEY*10, CDESC*26, CVAL*11, CMENU*23
      CHARACTER*79 LINE
C  ---------------------------------------------------------------------
C
C  Print the panel header.
      LE1 = LENGTH(PCOM(IPC))
      CALL DASH
      CALL CPRINT(PCOM(IPC)(1:LE1)//' Panel')
      CALL DASH
      KP1 = KPOINT(IPC)
      NKEY = 0
      CIDX = 'Ind'
      CKEY = 'Keyword'
      CDESC = 'Description'
      CVAL = 'Value'
      LES = 5
      CALL SCENT(CVAL,LES)
      CMENU = 'Allowed values'
      WRITE (LINE,50) CIDX, CKEY, CDESC, CVAL, CMENU
      WRITE (IUOUT,'($(FC)A)') LINE
C
C  Loop over all panel keywords.
      DO 30,I=KP1,KPOINT(IPC+1)-1
         IF (KIO(I).NE.' ') GOTO 30
         NKEY = NKEY+1
         WRITE (CIDX,'(I2,'')'')') NKEY
         CKEY = PKEY(I)
C
         IP = TPOINT(I)
         IF (ISCOMP(KTYPE(I),'I').NE.0) THEN
C  An integer keyword.
            IRL = NINT(PIR(IP))
            IF (IRL.EQ.1) THEN
               IL = NINT(PIR(IP+1))
               NLE = NINT(PIR(IP+2))
               IR = NINT(PIR(IP+3))
               NRI = NINT(PIR(IP+4))
               IP = IP+4
               CALL IPMENU(IL,NLE,IR,NRI,CMENU)
            ELSE IF (IRL.EQ.2) THEN
               IP = IP+1
               NITE = NINT(PIR(IP))
               CALL IPLIST(PIR(IP+1),NITE,CMENU)
               IP = IP+NITE
            END IF
            IP = IP+1
            IDEF = NINT(PIR(IP))
            CALL I2STR(IDEF,CVAL,LES)
            CALL SCENT(CVAL,LES)
            IP = IP+1
            IDESC = NINT(PIR(IP))
            IST = SPOINT(IDESC)
            IEND = SPOINT(IDESC+1)
            LES = IEND-IST+1
            LEC = LEN(CDESC)
            CDESC = PANSTR(IST:IEND)
            IF (LES.GT.LEC) CDESC(LEC-2:LEC) = '...'
         ELSE IF (ISCOMP(KTYPE(I),'R').NE.0) THEN
C  A real keyword.
            IRL = NINT(PIR(IP))
            IF (IRL.EQ.1) THEN
               IL = NINT(PIR(IP+1))
               XLE = PIR(IP+2)
               IR = NINT(PIR(IP+3))
               XRI = PIR(IP+4)
               IP = IP+4
               CALL RPMENU(IL,XLE,IR,XRI,CMENU)
            ELSE IF (IRL.EQ.2) THEN
               IP = IP+1
               NITE = NINT(PIR(IP))
               CALL RPLIST(PIR(IP+1),NITE,CMENU)
               IP = IP+NITE
            END IF
            IP = IP+1
            XDEF = PIR(IP)
            CALL R2STR(XDEF,CVAL,LES)
            CALL SCENT(CVAL,LES)
            IP = IP+1
            IDESC = NINT(PIR(IP))
            IST = SPOINT(IDESC)
            IEND = SPOINT(IDESC+1)
            LES = IEND-IST+1
            LEC = LEN(CDESC)
            CDESC = PANSTR(IST:IEND)
            IF (LES.GT.LEC) CDESC(LEC-2:LEC) = '...'
         ELSE IF (ISCOMP(KTYPE(I),'S').NE.0 .OR.
     &               ISCOMP(KTYPE(I),'C').NE.0) THEN
C  A string keyword.
            IAL = NINT(PIR(IP))
            IF (IAL.EQ.1) THEN
               CMENU = 'Any string'
            ELSE IF (IAL.EQ.2) THEN
               IP = IP+1
               NITE = NINT(PIR(IP))
               CALL SPLIST(PIR(IP+1),NITE,CMENU,SPOINT,PANSTR)
               IP = IP+NITE
            END IF
            IP = IP+1
            IDEF = NINT(PIR(IP))
            IST = SPOINT(IDEF)
            IEND = SPOINT(IDEF+1)
            LES = IEND-IST+1
            LEC = LEN(CVAL)
            CVAL = PANSTR(IST:IEND)
            IF (LES.GT.LEC) THEN
               CVAL(LEC-2:LEC) = '...'
               LES = LEC
            END IF
            CALL SCENT(CVAL,LES)
C
            IP = IP+1
            IDESC = NINT(PIR(IP))
            IST = SPOINT(IDESC)
            IEND = SPOINT(IDESC+1)
            LES = IEND-IST+1
            LEC = LEN(CDESC)
            CDESC = PANSTR(IST:IEND)
            IF (LES.GT.LEC) CDESC(LEC-2:LEC) = '...'
         END IF
         WRITE (LINE,50) CIDX, CKEY, CDESC, CVAL, CMENU
         WRITE (IUOUT,'($(FC)A)') LINE
C
30    CONTINUE
      CALL DASH
C
50    FORMAT (A,1X,A,1X,A,2X,A,2X,A)
      END
C  ---------------------------------------------------------------------
      SUBROUTINE IPMENU ( IL, NLE, IR, NRI, CMENU )
C  ---------------------------------------------------------------------
C
C  Description:
C    Called by subroutine PPR. Constructs a string describing the
C    allowed values for an integer keyword when these were declared as
C    an interval. Eg. [-1,1]  (-inf,100)  etc.
C
C  Input arguments:
C    IL         Determines whether the interval is open/closed/infinity
C               on the left:
C               IL =  1 -> An open interval.
C               IL =  2 -> A closed interval.
C               IL = -3 -> -Infinity was specified.
C    NLE        The left limit of the interval. 0 for -infinity.
C    IR         Determines whether the interval is open/closed/infinity
C               on the right:
C               IR =  1 -> An open interval.
C               IR =  2 -> A closed interval.
C               IR = -3 -> Infinity was specified.
C    NRI        The right limit of the interval. 0 for infinity.
C
C  Output arguments:
C    CMENU      The string describing the allowed values.
C
C  ---------------------------------------------------------------------
C  Arguments:
      CHARACTER*(*) CMENU
C  ---------------------------------------------------------------------
C  Local variables:
      CHARACTER*20 SCR1, SCR2
C  ---------------------------------------------------------------------
C
      IF (IL.EQ.1 .OR. IL.EQ.2) CALL I2STR(NLE,SCR1,LES1)
      IF (IR.EQ.1 .OR. IR.EQ.2) CALL I2STR(NRI,SCR2,LES2)
C
      IF (IL.EQ.1) THEN
         IF (IR.EQ.1) THEN
            CMENU = SCR1(1:LES1)//' < int < '//SCR2(1:LES2)
         ELSE IF (IR.EQ.2) THEN
            CMENU = SCR1(1:LES1)//' < int <= '//SCR2(1:LES2)
         ELSE IF (IR.EQ.3) THEN
            CMENU = 'Any int > '//SCR1(1:LES1)
         END IF
      ELSE IF (IL.EQ.2) THEN
         IF (IR.EQ.1) THEN
            CMENU = SCR1(1:LES1)//' <= int < '//SCR2(1:LES2)
         ELSE IF (IR.EQ.2) THEN
            CMENU = SCR1(1:LES1)//' <= int <= '//SCR2(1:LES2)
         ELSE IF (IR.EQ.3) THEN
            CMENU = 'Any int >= '//SCR1(1:LES1)
         END IF
      ELSE IF (ABS(IL).EQ.3) THEN
         IF (IR.EQ.1) THEN
            CMENU = 'Any int < '//SCR2(1:LES2)
         ELSE IF (IR.EQ.2) THEN
            CMENU = 'Any int <= '//SCR2(1:LES2)
         ELSE IF (IR.EQ.3) THEN
            CMENU = 'Any int'
         END IF
      END IF
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE RPMENU ( IL, XLE, IR, XRI, CMENU )
C  ---------------------------------------------------------------------
C
C  Description:
C    Called by subroutine PPR. Constructs a string describing the
C    allowed values for a real keyword when these were declared as an
C    interval. Eg. [-1.,1.]  (-inf,100.)  etc.
C
C  Input arguments:
C    IL         Determines whether the interval is open/closed/infinity
C               on the left:
C               IL =  1 -> An open interval.
C               IL =  2 -> A closed interval.
C               IL = -3 -> -Infinity was specified.
C    XLE        The left limit of the interval. 0 for -infinity.
C    IR         Determines whether the interval is open/closed/infinity
C               on the right:
C               IR =  1 -> An open interval.
C               IR =  2 -> A closed interval.
C               IR = -3 -> Infinity was specified.
C    XRI        The right limit of the interval. 0 for infinity.
C
C  Output arguments:
C    CMENU      The string describing the allowed values.
C
C  ---------------------------------------------------------------------
      IMPLICIT $(TYPE) (A-H,O-Z)
C  Arguments:
      CHARACTER*(*) CMENU
C  ---------------------------------------------------------------------
C  Local variables:
      CHARACTER*20 SCR1, SCR2
C  ---------------------------------------------------------------------
C
      IF (IL.EQ.1 .OR. IL.EQ.2) CALL R2STR(XLE,SCR1,LES1)
      IF (IR.EQ.1 .OR. IR.EQ.2) CALL R2STR(XRI,SCR2,LES2)
C
      IF (IL.EQ.1) THEN
         IF (IR.EQ.1) THEN
            CMENU = SCR1(1:LES1)//' < real < '//SCR2(1:LES2)
         ELSE IF (IR.EQ.2) THEN
            CMENU = SCR1(1:LES1)//' < real <= '//SCR2(1:LES2)
         ELSE IF (IR.EQ.3) THEN
            CMENU = 'Any real > '//SCR1(1:LES1)
         END IF
      ELSE IF (IL.EQ.2) THEN
         IF (IR.EQ.1) THEN
            CMENU = SCR1(1:LES1)//' <= real < '//SCR2(1:LES2)
         ELSE IF (IR.EQ.2) THEN
            CMENU = SCR1(1:LES1)//' <= real <= '//SCR2(1:LES2)
         ELSE IF (IR.EQ.3) THEN
            CMENU = 'Any real >= '//SCR1(1:LES1)
         END IF
      ELSE IF (ABS(IL).EQ.3) THEN
         IF (IR.EQ.1) THEN
            CMENU = 'Any real < '//SCR2(1:LES2)
         ELSE IF (IR.EQ.2) THEN
            CMENU = 'Any real <= '//SCR2(1:LES2)
         ELSE IF (IR.EQ.3) THEN
            CMENU = 'Any real'
         END IF
      END IF
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE IPLIST ( PIR, NITE, CMENU )
C  ---------------------------------------------------------------------
C
C  Description:
C    Called by subroutine PPR. Constructs a string describing thei
C    allowed values for an integer keyword when these were declared as
C    a set. Eg. {1,2,3} {-1,0,4}  etc.
C
C  Input arguments:
C    PIR        Part of the PIR table. The first entry in PIR that
C               corresponds to the current keyword, is always passed
C               here. Hence, this routine addresses PIR entries as
C               PIR(1), PIR(2),... PIR(NITE).
C               See the description of subroutine PARSP for a discussion
C               of panel related tables.
C    NITE       Number of items (allowed values) in the PIR table for
C               the current keyword.
C
C  Output arguments:
C    CMENU      The string describing the allowed values.
C
C  ---------------------------------------------------------------------
      IMPLICIT $(TYPE) (A-H,O-Z)
C  Arguments:
      DIMENSION PIR(NITE)
      CHARACTER*(*) CMENU
C  ---------------------------------------------------------------------
C  Local variables:
      CHARACTER*20 SCR1
      CHARACTER C
C  ---------------------------------------------------------------------
C
      LEM = LEN(CMENU)
      CMENU = '{'
      LE = 1
C
C  Loop over all allowed values.
      DO 10,I=1,NITE
         ITEM = NINT(PIR(I))
         CALL I2STR(ITEM,SCR1,LES)
         IF (LE+LES+1.LE.LEM) THEN
            CMENU(LE+1:LE+LES) = SCR1(1:LES)
            LE = LE+LES
         ELSE
            IF (LE+3.LE.LEM) THEN
               CMENU(LE+1:) = '...'
               LE = LE+3
            ELSE
               KSTART = LE-1
               DO 20,J=KSTART,1,-1
                  C = CMENU(J:J)
                  IF (C.EQ.',' .OR. C.EQ.'{') THEN
                     LE = J
                     CMENU(LE+1:) = '...'
                     LE = LE+3
                     RETURN
                  END IF
20             CONTINUE
C  This should never happen.
               CALL MABORT('Left brace missing.','IPLIST')
            END IF
            RETURN
         END IF
         LE = LE+1
         IF (I.EQ.NITE) THEN
            CMENU(LE:LE) = '}'
         ELSE
            CMENU(LE:LE) = ','
         END IF
10    CONTINUE
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE RPLIST ( PIR, NITE, CMENU )
C  ---------------------------------------------------------------------
C
C  Description:
C    Called by subroutine PPR. Constructs a string describing the
C    allowed values for a real keyword when these were declared as a
C    set. Eg. {1.4,2.2,3.} {-1.,0.,4.}  etc.
C
C  Input arguments:
C    PIR        Part of the PIR table. The first entry in PIR that
C               corresponds to the current keyword, is always passed
C               here. Hence, this routine addresses PIR entries as
C               PIR(1), PIR(2),... PIR(NITE).
C               See the description of subroutine PARSP for a discussion
C               of panel related tables.
C    NITE       Number of items (allowed values) in the PIR table for
C               the current keyword.
C
C  Output arguments:
C    CMENU      The string describing the allowed values.
C
C  ---------------------------------------------------------------------
      IMPLICIT $(TYPE) (A-H,O-Z)
C  Arguments:
      DIMENSION PIR(NITE)
      CHARACTER*(*) CMENU
C  ---------------------------------------------------------------------
C  Local variables:
      CHARACTER*11 SCR1
      CHARACTER C
C  ---------------------------------------------------------------------
C
      LEM = LEN(CMENU)
      CMENU = '{'
      LE = 1
C
C  Loop over all allowed values.
      DO 10,I=1,NITE
         RITEM = PIR(I)
         CALL R2STR(RITEM,SCR1,LES)
         IF (LE+LES+1.LE.LEM) THEN
            CMENU(LE+1:LE+LES) = SCR1(1:LES)
            LE = LE+LES
         ELSE
            IF (LE+3.LE.LEM) THEN
               CMENU(LE+1:) = '...'
            ELSE
               KSTART = LE-1
               DO 20,J=KSTART,1,-1
                  C = CMENU(J:J)
                  IF (C.EQ.',' .OR. C.EQ.'{') THEN
                     LE = J
                     CMENU(LE+1:) = '...'
                     LE = LE+3
                     RETURN
                  END IF
20             CONTINUE
C  This should never happen.
               CALL MABORT('Left brace missing.','RPLIST')
            END IF
            RETURN
         END IF
         LE = LE+1
         IF (I.EQ.NITE) THEN
            CMENU(LE:LE) = '}'
         ELSE
            CMENU(LE:LE) = ','
         END IF
10    CONTINUE
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE SPLIST ( PIR, NITE, CMENU, SPOINT, PANSTR )
C  ---------------------------------------------------------------------
C
C  Description:
C    Called by subroutine PPR. Constructs a string describing the
C    allowed values for a real keyword when these were declared as a
C    set. Eg. {'a','b'} {'hello','goodbye'}  etc.
C
C  Input arguments:
C    PIR        Part of the PIR table. The first entry in PIR that
C               corresponds to the current keyword, is always passed
C               here. Hence, this routine addresses PIR entries as
C               PIR(1), PIR(2),... PIR(NITE).
C               See the description of subroutine PARSP for a discussion
C               of panel related tables.
C    NITE       Number of items (allowed values) in the PIR table for
C               the current keyword.
C    SPOINT     The SPOINT table. See the description of subroutine
C               PARSP for a discussion of panel related tables.
C    PANSTR     String storage for panel related strings. See the
C               description of subroutine PARSP for a discussion of
C               panel related tables.
C
C  Output arguments:
C    CMENU      The string describing the allowed values.
C
C  ---------------------------------------------------------------------
      IMPLICIT $(TYPE) (A-H,O-Z)
C  Arguments:
      DIMENSION PIR(NITE)
      INTEGER SPOINT(*)
C  ---------------------------------------------------------------------
C  Local variables:
      CHARACTER*(*) CMENU, PANSTR
C  ---------------------------------------------------------------------
C
      LEM = LEN(CMENU)
      CMENU = '{'
      LE = 1
C
C  Loop over all allowed values.
      DO 10,I=1,NITE
         NS = NINT(PIR(I))
         IST = SPOINT(NS)
         IEND = SPOINT(NS+1)
         LES = IEND-IST+1
         IF (LE+LES+1.LE.LEM) THEN
            CMENU(LE+1:LE+LES) = PANSTR(IST:IEND)
            LE = LE+LES
         ELSE
            CMENU(LE+1:LEM-3) = PANSTR(IST:IEND)
            CMENU(LEM-2:LEM) = '...'
            LE = LEM
            RETURN
         END IF
         LE = LE+1
         IF (I.EQ.NITE) THEN
            CMENU(LE:LE) = '}'
         ELSE
            CMENU(LE:LE) = ','
         END IF
10    CONTINUE
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE PPRO
C  ---------------------------------------------------------------------
C
C  Description:
C    Subroutine PPRO prints the values of the MCL-returned keywords
C    upon completion of a panel command.
C
C  ---------------------------------------------------------------------
      IMPLICIT $(TYPE) (A-H,O-Z)
C  Common blocks:
C  Maximum length of a keyword.
      PARAMETER ( KEYLEN = 10 )
C  Maximum number of panel commands.
      PARAMETER ( MXPCOM = 80 )
C  Maximum number of keywords in all panel commands.
      PARAMETER ( MXPKEY = 1000 )
C  Maximum storage for integer/real keywords.
      PARAMETER ( MXPIR = 5000 )
C  Maximum storage for string keywords and descriptions.
      PARAMETER ( MXPSTR = 20000 )
C  Maximum number of pointers in the above string storage.
      PARAMETER ( MXPPST = 5000 )
C
      CHARACTER*(KEYLEN) PCOM(MXPCOM)
      INTEGER            PONF(MXPCOM)
      INTEGER            KPOINT(MXPCOM)
      INTEGER            NCHELP(2*MXPCOM)
      CHARACTER*(KEYLEN) PKEY(MXPKEY)
      INTEGER            NKHELP(2*MXPKEY)
      CHARACTER          KTYPE(MXPKEY)
      INTEGER            TPOINT(MXPKEY)
      CHARACTER          KIO(MXPKEY)
      $(TYPE)            PIR(MXPIR)
      INTEGER            SPOINT(MXPPST)
      CHARACTER*(MXPSTR) PANSTR
      COMMON /PCOMS1/ KPOINT, NCHELP, NKHELP, TPOINT, SPOINT, PONF
      COMMON /PCOMS2/ PIR
      COMMON /PCOMS3/ PKEY, KTYPE, PANSTR, PCOM, KIO
C
C  NPCOM: Number of panel commands in the panel description file.
C  NPKEY: Number of keywords in all panel commands.
C  NPIR:  Number of positions filled in array PIR.
C  NPSP:  Number of positions filled in the char-pointer array SPOINT.
C  NPCH:  Number of character positions filled in PANSTR.
      COMMON /PCOMS4/ NPCOM, NPKEY, NPIR, NPSP, NPCH
      COMMON /UNITS/ IUINP, IUOUT
      CHARACTER*10 CURCOM
      COMMON /CURC/ CURCOM
C  ---------------------------------------------------------------------
C  Local variables:
      CHARACTER CKEY*10, CVAL*60
C  ---------------------------------------------------------------------
C
C  Locate the current command in the panel table.
      CALL PLOCP(CURCOM,IPC)
      IF (IPC.EQ.0) CALL MABORT('No panel.','PPRO')
C
      KP1 = KPOINT(IPC)
      NKEY = 0
C
C  Loop over all keywords.
      DO 30,I=KP1,KPOINT(IPC+1)-1
         IF (KIO(I).EQ.' ') GOTO 30
C
C  This is definitely an MCL returned keyword.
         NKEY = NKEY+1
         CKEY = PKEY(I)
         IF (NKEY.EQ.1) WRITE (IUOUT,40)
C
         IP = TPOINT(I)
         IF (ISCOMP(KTYPE(I),'I').NE.0) THEN
C  An integer keyword.
            IRL = NINT(PIR(IP))
            IF (IRL.EQ.1) THEN
               IP = IP+4
            ELSE IF (IRL.EQ.2) THEN
               IP = IP+1
               NITE = NINT(PIR(IP))
               IP = IP+NITE
            END IF
            IP = IP+1
            IDEF = NINT(PIR(IP))
            CALL I2STR(IDEF,CVAL,LES)
         ELSE IF (ISCOMP(KTYPE(I),'R').NE.0) THEN
C  A real keyword.
            IRL = NINT(PIR(IP))
            IF (IRL.EQ.1) THEN
               IP = IP+4
            ELSE IF (IRL.EQ.2) THEN
               IP = IP+1
               NITE = NINT(PIR(IP))
               IP = IP+NITE
            END IF
            IP = IP+1
            XDEF = PIR(IP)
            CALL R2NRTF(XDEF,CVAL,LES)
         ELSE IF (ISCOMP(KTYPE(I),'S').NE.0 .OR.
     &               ISCOMP(KTYPE(I),'C').NE.0) THEN
C  A string keyword.
            IAL = NINT(PIR(IP))
            IF (IAL.EQ.2) THEN
               IP = IP+1
               NITE = NINT(PIR(IP))
               IP = IP+NITE
            END IF
            IP = IP+1
            IDEF = NINT(PIR(IP))
            IST = SPOINT(IDEF)
            IEND = SPOINT(IDEF+1)
            LES = IEND-IST+3
            CVAL = ''''//PANSTR(IST:IEND)//''''
            LEC = LEN(CVAL)
            IF (LES.GT.LEC) THEN
               CVAL(LEC-2:LEC) = '...'
               LES = LEC
            END IF
         END IF
         WRITE (IUOUT,50) CKEY, CVAL(1:LES)
C
30    CONTINUE
      IF (NKEY.NE.0) CALL DASH
C
40    FORMAT ($(FC)'The following values would be returned to ',
     &       'an MCL program:')
50    FORMAT ($(FC)A,' = ',A)
      END
C  ---------------------------------------------------------------------
      SUBROUTINE SCENT ( STR, LES )
C  ---------------------------------------------------------------------
C
C  Description:
C    Subroutine SCENT centers the contents of STR. For example the input
C    STR = 'Hello    ' will produce the output STR = '  Hello  '.
C
C  Input arguments:
C    LES        Effective length of STR.
C
C  Input / Output arguments:
C    STR        On input contains a string in its first LES positions.
C               On output contains the centered string.
C
C  ---------------------------------------------------------------------
C  Arguments:
      CHARACTER*(*) STR
C  ---------------------------------------------------------------------
C
      LEA = LEN(STR)
      IBL = LEA-LES
      IRI = IBL/2
      IF (IRI.GT.0) THEN
         DO 10,I=LES+IRI,IRI+1,-1
            STR(I:I) = STR(I-IRI:I-IRI)
10       CONTINUE
         STR(1:IRI) = ' '
      END IF
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE SETPI ( KEY, IVAL, IERR )
C  ---------------------------------------------------------------------
C
C  Description:
C    Subroutine SETPI is part of a set of subroutines that manipulate
C    the values assigned to the keywords of the various panels. It
C    assigns the integer value IVAL to keyword KEY in the current
C    panel. See the description of subroutine PARSP for a discussion
C    of the complex internal structure storing panel related
C    information.
C
C    This routine will abort if:
C       1) The panel name (CURCOM) is not defined in the panel
C          description file.
C       2) KEY is not a valid keyword for this panel.
C       3) KEY is not an integer keyword.
C    Normally none of these should happen.
C
C  Input arguments:
C    KEY        The keyword.
C    IVAL       The value to be assigned to KEY.
C
C  Output arguments:
C    IERR       Error indicator:
C               IERR = 0  -> IVAL has been assigned to KEY.
C               IERR = -1 -> IVAL is not allowed as a value for the
C                            KEY keyword.
C
C  ---------------------------------------------------------------------
      IMPLICIT $(TYPE) (A-H,O-Z)
C  Arguments:
      CHARACTER*(*) KEY
C  ---------------------------------------------------------------------
C  Common blocks:
C  Maximum length of a keyword.
      PARAMETER ( KEYLEN = 10 )
C  Maximum number of panel commands.
      PARAMETER ( MXPCOM = 80 )
C  Maximum number of keywords in all panel commands.
      PARAMETER ( MXPKEY = 1000 )
C  Maximum storage for integer/real keywords.
      PARAMETER ( MXPIR = 5000 )
C  Maximum storage for string keywords and descriptions.
      PARAMETER ( MXPSTR = 20000 )
C  Maximum number of pointers in the above string storage.
      PARAMETER ( MXPPST = 5000 )
C
      CHARACTER*(KEYLEN) PCOM(MXPCOM)
      INTEGER            PONF(MXPCOM)
      INTEGER            KPOINT(MXPCOM)
      INTEGER            NCHELP(2*MXPCOM)
      CHARACTER*(KEYLEN) PKEY(MXPKEY)
      INTEGER            NKHELP(2*MXPKEY)
      CHARACTER          KTYPE(MXPKEY)
      INTEGER            TPOINT(MXPKEY)
      CHARACTER          KIO(MXPKEY)
      $(TYPE)            PIR(MXPIR)
      INTEGER            SPOINT(MXPPST)
      CHARACTER*(MXPSTR) PANSTR
      COMMON /PCOMS1/ KPOINT, NCHELP, NKHELP, TPOINT, SPOINT, PONF
      COMMON /PCOMS2/ PIR
      COMMON /PCOMS3/ PKEY, KTYPE, PANSTR, PCOM, KIO
C
C  NPCOM: Number of panel commands in the panel description file.
C  NPKEY: Number of keywords in all panel commands.
C  NPIR:  Number of positions filled in array PIR.
C  NPSP:  Number of positions filled in the char-pointer array SPOINT.
C  NPCH:  Number of character positions filled in PANSTR.
      COMMON /PCOMS4/ NPCOM, NPKEY, NPIR, NPSP, NPCH
      CHARACTER*10 CURCOM
      COMMON /CURC/ CURCOM
      CHARACTER*(KEYLEN) PREVPA
      COMMON /PRPA1/ PREVPA
      COMMON /PRPA2/ IPREVP
      LOGICAL STRO, WEAK
      COMMON /PRILOG/ STRO, WEAK
      COMMON /UNITS/ IUINP, IUOUT
C  ---------------------------------------------------------------------
C  Local variables:
      CHARACTER*(KEYLEN) PNAME, KNAME
C  ---------------------------------------------------------------------
C
      PNAME = CURCOM
      KNAME = KEY
C
C  Look at the cache first.
      IF (ISCOMP(PREVPA,PNAME).NE.0) THEN
         I = IPREVP
      ELSE
C  Search for the panel.
         LE1 = LENGTH(PNAME)
         IF (LE1.EQ.0) CALL MABORT('Zero panel length.','SETPI')
         CALL PLOCP(PNAME(1:LE1),I)
         IF (I.EQ.0)
     &          CALL MABORT('Panel not found: '//PNAME(1:LE1),'SETPI')
C  Store in the cache.
         PREVPA = PNAME
         IPREVP = I
      END IF
C
C  Search for the keyword.
      LE1 = LENGTH(KNAME)
      IF (LE1.EQ.0) CALL MABORT('Zero keyword length.','SETPI')
      CALL PLOCK(I,KNAME(1:LE1),I)
      IF (I.EQ.0)
     &    CALL MABORT('Keyword not found: '//KNAME(1:LE1),'SETPI')
C
C  Get keyword type.
      IF (ISCOMP(KTYPE(I),'I').EQ.0)
     &    CALL MABORT('Not an integer keyword '//KNAME(1:LE1),'SETPI')
C
C  Is IVAL an allowed value ?
      CALL PRANI(IVAL,I,IOK)
      IF (IOK.EQ.0) THEN
         IF (WEAK) WRITE (IUOUT,40) KNAME(1:LE1)
         IERR = -1
         RETURN
      END IF
C
C  Set the value.
      IP = TPOINT(I)
      IRL = NINT(PIR(IP))
      IF (IRL.EQ.1) THEN
         IP = IP+4
      ELSE IF (IRL.EQ.2) THEN
         IP = IP+1
         NITE = NINT(PIR(IP))
         IP = IP+NITE
      END IF
      IP = IP+1
      PIR(IP) = IVAL
      IERR = 0
C
40    FORMAT (/$(FC)'The specified value for the ',A,' keyword ',
     &        'is not allowed.')
      END
C  ---------------------------------------------------------------------
      SUBROUTINE SETPR ( KEY, XVAL, IERR )
C  ---------------------------------------------------------------------
C
C  Description:
C    Subroutine SETPR is part of a set of subroutines that manipulate
C    the values assigned to the keywords of the various panels. It
C    assigns the real value XVAL to keyword KEY in the current
C    panel. See the description of subroutine PARSP for a discussion
C    of the complex internal structure storing panel related
C    information.
C
C    This routine will abort if:
C       1) The panel name (CURCOM) is not defined in the panel
C          description file.
C       2) KEY is not a valid keyword for this panel.
C       3) KEY is not a real keyword.
C    Normally none of these should happen.
C
C  Input arguments:
C    KEY        The keyword.
C    XVAL       The value to be assigned to KEY.
C
C  Output arguments:
C    IERR       Error indicator:
C               IERR = 0  -> IVAL has been assigned to KEY.
C               IERR = -1 -> IVAL is not allowed as a value for the
C                            KEY keyword.
C
C  ---------------------------------------------------------------------
      IMPLICIT $(TYPE) (A-H,O-Z)
C  Arguments:
      CHARACTER*(*) KEY
C  ---------------------------------------------------------------------
C  Common blocks:
C  Maximum length of a keyword.
      PARAMETER ( KEYLEN = 10 )
C  Maximum number of panel commands.
      PARAMETER ( MXPCOM = 80 )
C  Maximum number of keywords in all panel commands.
      PARAMETER ( MXPKEY = 1000 )
C  Maximum storage for integer/real keywords.
      PARAMETER ( MXPIR = 5000 )
C  Maximum storage for string keywords and descriptions.
      PARAMETER ( MXPSTR = 20000 )
C  Maximum number of pointers in the above string storage.
      PARAMETER ( MXPPST = 5000 )
C
      CHARACTER*(KEYLEN) PCOM(MXPCOM)
      INTEGER            PONF(MXPCOM)
      INTEGER            KPOINT(MXPCOM)
      INTEGER            NCHELP(2*MXPCOM)
      CHARACTER*(KEYLEN) PKEY(MXPKEY)
      INTEGER            NKHELP(2*MXPKEY)
      CHARACTER          KTYPE(MXPKEY)
      INTEGER            TPOINT(MXPKEY)
      CHARACTER          KIO(MXPKEY)
      $(TYPE)            PIR(MXPIR)
      INTEGER            SPOINT(MXPPST)
      CHARACTER*(MXPSTR) PANSTR
      COMMON /PCOMS1/ KPOINT, NCHELP, NKHELP, TPOINT, SPOINT, PONF
      COMMON /PCOMS2/ PIR
      COMMON /PCOMS3/ PKEY, KTYPE, PANSTR, PCOM, KIO
C
C  NPCOM: Number of panel commands in the panel description file.
C  NPKEY: Number of keywords in all panel commands.
C  NPIR:  Number of positions filled in array PIR.
C  NPSP:  Number of positions filled in the char-pointer array SPOINT.
C  NPCH:  Number of character positions filled in PANSTR.
      COMMON /PCOMS4/ NPCOM, NPKEY, NPIR, NPSP, NPCH
      CHARACTER*10 CURCOM
      COMMON /CURC/ CURCOM
      CHARACTER*(KEYLEN) PREVPA
      COMMON /PRPA1/ PREVPA
      COMMON /PRPA2/ IPREVP
      LOGICAL STRO, WEAK
      COMMON /PRILOG/ STRO, WEAK
      COMMON /UNITS/ IUINP, IUOUT
C  ---------------------------------------------------------------------
C  Local variables:
      CHARACTER*(KEYLEN) PNAME, KNAME
C  ---------------------------------------------------------------------
C
      PNAME = CURCOM
      KNAME = KEY
C
C  Look at the cache first.
      IF (ISCOMP(PREVPA,PNAME).NE.0) THEN
         I = IPREVP
      ELSE
C  Search for the panel.
         LE1 = LENGTH(PNAME)
         IF (LE1.EQ.0) CALL MABORT('Zero panel length.','SETPR')
         CALL PLOCP(PNAME(1:LE1),I)
         IF (I.EQ.0)
     &          CALL MABORT('Panel not found: '//PNAME(1:LE1),'SETPR')
C  Store in the cache.
         PREVPA = PNAME
         IPREVP = I
      END IF
C
C  Search for the keyword.
      LE1 = LENGTH(KNAME)
      IF (LE1.EQ.0) CALL MABORT('Zero keyword length.','SETPR')
      CALL PLOCK(I,KNAME(1:LE1),I)
      IF (I.EQ.0)
     &    CALL MABORT('Keyword not found: '//KNAME(1:LE1),'SETPR')
C
C  Get keyword type.
      IF (ISCOMP(KTYPE(I),'R').EQ.0)
     &    CALL MABORT('Not a real keyword '//KNAME(1:LE1),'SETPR')
C
C  Is XVAL an allowed value ?
      CALL PRANR(XVAL,I,IOK)
      IF (IOK.EQ.0) THEN
         IF (WEAK) WRITE (IUOUT,40) KNAME(1:LE1)
         IERR = -1
         RETURN
      END IF
C
C  Set the value.
      IP = TPOINT(I)
      IRL = NINT(PIR(IP))
      IF (IRL.EQ.1) THEN
         IP = IP+4
      ELSE IF (IRL.EQ.2) THEN
         IP = IP+1
         NITE = NINT(PIR(IP))
         IP = IP+NITE
      END IF
      IP = IP+1
      PIR(IP) = XVAL
      IERR = 0
C
40    FORMAT (/$(FC)'The specified value for the ',A,' keyword ',
     &        'is not allowed.')
      END
C  ---------------------------------------------------------------------
      SUBROUTINE SETPS ( KEY, SVAL, LES, IERR )
C  ---------------------------------------------------------------------
C
C  Description:
C    Subroutine SETPS is part of a set of subroutines that manipulate
C    the values assigned to the keywords of the various panels. It
C    assigns the string value SVAL to keyword KEY in the current
C    panel. See the description of subroutine PARSP for a discussion
C    of the complex internal structure storing panel related
C    information.
C
C    This routine will abort if:
C       1) The panel name (CURCOM) is not defined in the panel
C          description file.
C       2) KEY is not a valid keyword for this panel.
C       3) KEY is not a real keyword.
C    Normally none of these should happen.
C
C  Input arguments:
C    KEY        The keyword.
C    SVAL       The value to be assigned to KEY.
C    LES        Number of useful characters in SVAL.
C
C  Output arguments:
C    IERR       Error indicator:
C               IERR = 0  -> IVAL has been assigned to KEY.
C               IERR = -1 -> IVAL is not allowed as a value for the
C                            KEY keyword or there is not enough string
C                            storage available to hold SVAL.
C
C  ---------------------------------------------------------------------
      IMPLICIT $(TYPE) (A-H,O-Z)
C  Arguments:
      CHARACTER*(*) KEY, SVAL
C  ---------------------------------------------------------------------
C  Common blocks:
C  Maximum length of a keyword.
      PARAMETER ( KEYLEN = 10 )
C  Maximum number of panel commands.
      PARAMETER ( MXPCOM = 80 )
C  Maximum number of keywords in all panel commands.
      PARAMETER ( MXPKEY = 1000 )
C  Maximum storage for integer/real keywords.
      PARAMETER ( MXPIR = 5000 )
C  Maximum storage for string keywords and descriptions.
      PARAMETER ( MXPSTR = 20000 )
C  Maximum number of pointers in the above string storage.
      PARAMETER ( MXPPST = 5000 )
C
      CHARACTER*(KEYLEN) PCOM(MXPCOM)
      INTEGER            PONF(MXPCOM)
      INTEGER            KPOINT(MXPCOM)
      INTEGER            NCHELP(2*MXPCOM)
      CHARACTER*(KEYLEN) PKEY(MXPKEY)
      INTEGER            NKHELP(2*MXPKEY)
      CHARACTER          KTYPE(MXPKEY)
      INTEGER            TPOINT(MXPKEY)
      CHARACTER          KIO(MXPKEY)
      $(TYPE)            PIR(MXPIR)
      INTEGER            SPOINT(MXPPST)
      CHARACTER*(MXPSTR) PANSTR
      COMMON /PCOMS1/ KPOINT, NCHELP, NKHELP, TPOINT, SPOINT, PONF
      COMMON /PCOMS2/ PIR
      COMMON /PCOMS3/ PKEY, KTYPE, PANSTR, PCOM, KIO
C
C  NPCOM: Number of panel commands in the panel description file.
C  NPKEY: Number of keywords in all panel commands.
C  NPIR:  Number of positions filled in array PIR.
C  NPSP:  Number of positions filled in the char-pointer array SPOINT.
C  NPCH:  Number of character positions filled in PANSTR.
      COMMON /PCOMS4/ NPCOM, NPKEY, NPIR, NPSP, NPCH
      CHARACTER*10 CURCOM
      COMMON /CURC/ CURCOM
      CHARACTER*(KEYLEN) PREVPA
      COMMON /PRPA1/ PREVPA
      COMMON /PRPA2/ IPREVP
      LOGICAL STRO, WEAK
      COMMON /PRILOG/ STRO, WEAK
      COMMON /UNITS/ IUINP, IUOUT
C  ---------------------------------------------------------------------
C  Local variables:
      CHARACTER*(KEYLEN) PNAME, KNAME
C  ---------------------------------------------------------------------
C
      PNAME = CURCOM
      KNAME = KEY
C
C  Look at the cache first.
      IF (ISCOMP(PREVPA,PNAME).NE.0) THEN
         I = IPREVP
      ELSE
C  Search for the panel.
         LE1 = LENGTH(PNAME)
         IF (LE1.EQ.0) CALL MABORT('Zero panel length.','SETPS')
         CALL PLOCP(PNAME(1:LE1),I)
         IF (I.EQ.0)
     &          CALL MABORT('Panel not found: '//PNAME(1:LE1),'SETPS')
C  Store in the cache.
         PREVPA = PNAME
         IPREVP = I
      END IF
C
C  Search for the keyword.
      LE1 = LENGTH(KNAME)
      IF (LE1.EQ.0) CALL MABORT('Zero keyword length.','SETPS')
      CALL PLOCK(I,KNAME(1:LE1),I)
      IF (I.EQ.0)
     &    CALL MABORT('Keyword not found: '//KNAME(1:LE1),'SETPS')
C
C  Get keyword type.
      IF (ISCOMP(KTYPE(I),'S').EQ.0 .AND. ISCOMP(KTYPE(I),'C').EQ.0)
     &    CALL MABORT('Not a string keyword '//KNAME(1:LE1),'SETPS')
C
C Is SVAL an allowed value ?
      CALL PRANS(SVAL,LES,I,IOK)
      IF (IOK.EQ.0) THEN
         IF (WEAK) WRITE (IUOUT,42) KNAME(1:LE1)
         IERR = -1
         RETURN
      END IF
C
C  Set the value.
      IP = TPOINT(I)
      IAL = NINT(PIR(IP))
      IF (IAL.EQ.1) THEN
      ELSE IF (IAL.EQ.2) THEN
         IP = IP+1
         NITE = NINT(PIR(IP))
         IP = IP+NITE
      END IF
      IP = IP+1
      IDEF = NINT(PIR(IP))
C
      IST = SPOINT(IDEF)
      IEND = SPOINT(IDEF+1)
      LESBEF = IEND-IST+1
      IDIFF = LES-LESBEF
      IF (NPCH+IDIFF.GT.MXPSTR) THEN
         IF (WEAK) WRITE (IUOUT,41) KNAME(1:LE1)
         IERR = -1
         RETURN
      END IF
C
      IF (LES.GT.LESBEF) THEN
         DO 40,I=NPCH,IEND+1,-1
            IGO = I+IDIFF
            PANSTR(IGO:IGO) = PANSTR(I:I)
40       CONTINUE
         NPCH = NPCH+IDIFF
         DO 45,I=IDEF+1,NPSP
            SPOINT(I) = SPOINT(I)+IDIFF
45       CONTINUE
      ELSE IF (LES.LT.LESBEF) THEN
         DO 50,I=IEND+1,NPCH
            IGO = I+IDIFF
            PANSTR(IGO:IGO) = PANSTR(I:I)
50       CONTINUE
         DO 55,I=IDEF+1,NPSP
            SPOINT(I) = SPOINT(I)+IDIFF
55       CONTINUE
         NPCH = NPCH+IDIFF
      END IF
      IF (LES.GT.0) PANSTR(IST:SPOINT(IDEF+1)) = SVAL(1:LES)
      IERR = 0
C
41    FORMAT (/$(FC)'There is not enough space available to hold'
     &        /$(FC)'the value for the ',A,' keyword.')
42    FORMAT (/$(FC)'The specified value for the ',A,' keyword ',
     &        'is not allowed.')
      END
C  ---------------------------------------------------------------------
      SUBROUTINE SETPPI ( PANEL, KEY, IVAL, IERR )
C  ---------------------------------------------------------------------
C
C  Description:
C    Subroutine SETPPI is part of a set of subroutines that manipulate
C    the values assigned to the keywords of the various panels. It
C    assigns the integer value IVAL to keyword KEY in the panel
C    specified by PANEL. See the description of subroutine PARSP for
C    a discussion of the complex internal structure storing panel
C    related information.
C
C    This routine will abort if:
C       1) The panel name PANEL is not defined in the panel
C          description file.
C       2) KEY is not a valid keyword for this panel.
C       3) KEY is not an integer keyword.
C    Normally none of these should happen.
C
C  Input arguments:
C    PANEL      The panel.
C    KEY        The keyword.
C    IVAL       The value to be assigned to KEY.
C
C  Output arguments:
C    IERR       Error indicator:
C               IERR = 0  -> IVAL has been assigned to KEY.
C               IERR = -1 -> IVAL is not allowed as a value for the
C                            KEY keyword.
C
C  ---------------------------------------------------------------------
      IMPLICIT $(TYPE) (A-H,O-Z)
C  Arguments:
      CHARACTER*(*) PANEL, KEY
C  ---------------------------------------------------------------------
C  Common blocks:
C  Maximum length of a keyword.
      PARAMETER ( KEYLEN = 10 )
C  Maximum number of panel commands.
      PARAMETER ( MXPCOM = 80 )
C  Maximum number of keywords in all panel commands.
      PARAMETER ( MXPKEY = 1000 )
C  Maximum storage for integer/real keywords.
      PARAMETER ( MXPIR = 5000 )
C  Maximum storage for string keywords and descriptions.
      PARAMETER ( MXPSTR = 20000 )
C  Maximum number of pointers in the above string storage.
      PARAMETER ( MXPPST = 5000 )
C
      CHARACTER*(KEYLEN) PCOM(MXPCOM)
      INTEGER            PONF(MXPCOM)
      INTEGER            KPOINT(MXPCOM)
      INTEGER            NCHELP(2*MXPCOM)
      CHARACTER*(KEYLEN) PKEY(MXPKEY)
      INTEGER            NKHELP(2*MXPKEY)
      CHARACTER          KTYPE(MXPKEY)
      INTEGER            TPOINT(MXPKEY)
      CHARACTER          KIO(MXPKEY)
      $(TYPE)            PIR(MXPIR)
      INTEGER            SPOINT(MXPPST)
      CHARACTER*(MXPSTR) PANSTR
      COMMON /PCOMS1/ KPOINT, NCHELP, NKHELP, TPOINT, SPOINT, PONF
      COMMON /PCOMS2/ PIR
      COMMON /PCOMS3/ PKEY, KTYPE, PANSTR, PCOM, KIO
C
C  NPCOM: Number of panel commands in the panel description file.
C  NPKEY: Number of keywords in all panel commands.
C  NPIR:  Number of positions filled in array PIR.
C  NPSP:  Number of positions filled in the char-pointer array SPOINT.
C  NPCH:  Number of character positions filled in PANSTR.
      COMMON /PCOMS4/ NPCOM, NPKEY, NPIR, NPSP, NPCH
      CHARACTER*(KEYLEN) PREVPA
      COMMON /PRPA1/ PREVPA
      COMMON /PRPA2/ IPREVP
      LOGICAL STRO, WEAK
      COMMON /PRILOG/ STRO, WEAK
      COMMON /UNITS/ IUINP, IUOUT
C  ---------------------------------------------------------------------
C  Local variables:
      CHARACTER*(KEYLEN) PNAME, KNAME
C  ---------------------------------------------------------------------
C
      PNAME = PANEL
      KNAME = KEY
C
C  Look at the cache first.
      IF (ISCOMP(PREVPA,PNAME).NE.0) THEN
         I = IPREVP
      ELSE
C  Search for the panel.
         LE1 = LENGTH(PNAME)
         IF (LE1.EQ.0) CALL MABORT('Zero panel length.','SETPPI')
         CALL PLOCP(PNAME(1:LE1),I)
         IF (I.EQ.0)
     &          CALL MABORT('Panel not found: '//PNAME(1:LE1),'SETPPI')
C  Store in the cache.
         PREVPA = PNAME
         IPREVP = I
      END IF
C
C  Search for the keyword.
      LE1 = LENGTH(KNAME)
      IF (LE1.EQ.0) CALL MABORT('Zero keyword length.','SETPPI')
      CALL PLOCK(I,KNAME(1:LE1),I)
      IF (I.EQ.0)
     &    CALL MABORT('Keyword not found: '//KNAME(1:LE1),'SETPPI')
C
C  Get keyword type.
      IF (ISCOMP(KTYPE(I),'I').EQ.0)
     &    CALL MABORT('Not an integer keyword '//KNAME(1:LE1),'SETPPI')
C
C Is IVAL an allowed value ?
      CALL PRANI(IVAL,I,IOK)
      IF (IOK.EQ.0) THEN
         IF (WEAK) WRITE (IUOUT,40) KNAME(1:LE1)
         IERR = -1
         RETURN
      END IF
C
C  Set the value.
      IP = TPOINT(I)
      IRL = NINT(PIR(IP))
      IF (IRL.EQ.1) THEN
         IP = IP+4
      ELSE IF (IRL.EQ.2) THEN
         IP = IP+1
         NITE = NINT(PIR(IP))
         IP = IP+NITE
      END IF
      IP = IP+1
      PIR(IP) = IVAL
      IERR = 0
C
40    FORMAT (/$(FC)'The specified value for the ',A,' keyword ',
     &        'is not allowed.')
      END
C  ---------------------------------------------------------------------
      SUBROUTINE SETPPR ( PANEL, KEY, XVAL, IERR )
C  ---------------------------------------------------------------------
C
C  Description:
C    Subroutine SETPPR is part of a set of subroutines that manipulate
C    the values assigned to the keywords of the various panels. It
C    assigns the real value XVAL to keyword KEY in the panel
C    specified by PANEL. See the description of subroutine PARSP for
C    a discussion of the complex internal structure storing panel
C    related information.
C
C    This routine will abort if:
C       1) The panel name PANEL is not defined in the panel
C          description file.
C       2) KEY is not a valid keyword for this panel.
C       3) KEY is not a real keyword.
C    Normally none of these should happen.
C
C  Input arguments:
C    PANEL      The panel.
C    KEY        The keyword.
C    IVAL       The value to be assigned to KEY.
C
C  Output arguments:
C    IERR       Error indicator:
C               IERR = 0  -> IVAL has been assigned to KEY.
C               IERR = -1 -> IVAL is not allowed as a value for the
C                            KEY keyword.
C
C  ---------------------------------------------------------------------
      IMPLICIT $(TYPE) (A-H,O-Z)
C  Arguments:
      CHARACTER*(*) PANEL, KEY
C  ---------------------------------------------------------------------
C  Common blocks:
C  Maximum length of a keyword.
      PARAMETER ( KEYLEN = 10 )
C  Maximum number of panel commands.
      PARAMETER ( MXPCOM = 80 )
C  Maximum number of keywords in all panel commands.
      PARAMETER ( MXPKEY = 1000 )
C  Maximum storage for integer/real keywords.
      PARAMETER ( MXPIR = 5000 )
C  Maximum storage for string keywords and descriptions.
      PARAMETER ( MXPSTR = 20000 )
C  Maximum number of pointers in the above string storage.
      PARAMETER ( MXPPST = 5000 )
C
      CHARACTER*(KEYLEN) PCOM(MXPCOM)
      INTEGER            PONF(MXPCOM)
      INTEGER            KPOINT(MXPCOM)
      INTEGER            NCHELP(2*MXPCOM)
      CHARACTER*(KEYLEN) PKEY(MXPKEY)
      INTEGER            NKHELP(2*MXPKEY)
      CHARACTER          KTYPE(MXPKEY)
      INTEGER            TPOINT(MXPKEY)
      CHARACTER          KIO(MXPKEY)
      $(TYPE)            PIR(MXPIR)
      INTEGER            SPOINT(MXPPST)
      CHARACTER*(MXPSTR) PANSTR
      COMMON /PCOMS1/ KPOINT, NCHELP, NKHELP, TPOINT, SPOINT, PONF
      COMMON /PCOMS2/ PIR
      COMMON /PCOMS3/ PKEY, KTYPE, PANSTR, PCOM, KIO
C
C  NPCOM: Number of panel commands in the panel description file.
C  NPKEY: Number of keywords in all panel commands.
C  NPIR:  Number of positions filled in array PIR.
C  NPSP:  Number of positions filled in the char-pointer array SPOINT.
C  NPCH:  Number of character positions filled in PANSTR.
      COMMON /PCOMS4/ NPCOM, NPKEY, NPIR, NPSP, NPCH
      CHARACTER*(KEYLEN) PREVPA
      COMMON /PRPA1/ PREVPA
      COMMON /PRPA2/ IPREVP
      LOGICAL STRO, WEAK
      COMMON /PRILOG/ STRO, WEAK
      COMMON /UNITS/ IUINP, IUOUT
C  ---------------------------------------------------------------------
C  Local variables:
      CHARACTER*(KEYLEN) PNAME, KNAME
C  ---------------------------------------------------------------------
C
      PNAME = PANEL
      KNAME = KEY
C
C  Look at the cache first.
      IF (ISCOMP(PREVPA,PNAME).NE.0) THEN
         I = IPREVP
      ELSE
C  Search for the panel.
         LE1 = LENGTH(PNAME)
         IF (LE1.EQ.0) CALL MABORT('Zero panel length.','SETPPR')
         CALL PLOCP(PNAME(1:LE1),I)
         IF (I.EQ.0)
     &          CALL MABORT('Panel not found: '//PNAME(1:LE1),'SETPPR')
C  Store in the cache.
         PREVPA = PNAME
         IPREVP = I
      END IF
C
C  Search for the keyword.
      LE1 = LENGTH(KNAME)
      IF (LE1.EQ.0) CALL MABORT('Zero keyword length.','SETPPR')
      CALL PLOCK(I,KNAME(1:LE1),I)
      IF (I.EQ.0)
     &    CALL MABORT('Keyword not found: '//KNAME(1:LE1),'SETPPR')
C
C  Get keyword type.
      IF (ISCOMP(KTYPE(I),'R').EQ.0)
     &    CALL MABORT('Not a real keyword '//KNAME(1:LE1),'SETPPR')
C
C Is XVAL an allowed value ?
      CALL PRANR(XVAL,I,IOK)
      IF (IOK.EQ.0) THEN
         IF (WEAK) WRITE (IUOUT,40) KNAME(1:LE1)
         IERR = -1
         RETURN
      END IF
C
C  Set the value.
      IP = TPOINT(I)
      IRL = NINT(PIR(IP))
      IF (IRL.EQ.1) THEN
         IP = IP+4
      ELSE IF (IRL.EQ.2) THEN
         IP = IP+1
         NITE = NINT(PIR(IP))
         IP = IP+NITE
      END IF
      IP = IP+1
      PIR(IP) = XVAL
      IERR = 0
C
40    FORMAT (/$(FC)'The specified value for the ',A,' keyword ',
     &        'is not allowed.')
      END
C  ---------------------------------------------------------------------
      SUBROUTINE SETPPS ( PANEL, KEY, SVAL, LES, IERR )
C  ---------------------------------------------------------------------
C
C  Description:
C    Subroutine SETPPS is part of a set of subroutines that manipulate
C    the values assigned to the keywords of the various panels. It
C    assigns the string value SVAL to keyword KEY in the panel
C    specified by PANEL. See the description of subroutine PARSP for
C    a discussion of the complex internal structure storing panel
C    related information.
C
C    This routine will abort if:
C       1) The panel name PANEL is not defined in the panel
C          description file.
C       2) KEY is not a valid keyword for this panel.
C       3) KEY is not a string keyword.
C    Normally none of these should happen.
C
C  Input arguments:
C    PANEL      The panel.
C    KEY        The keyword.
C    IVAL       The value to be assigned to KEY.
C
C  Output arguments:
C    IERR       Error indicator:
C               IERR = 0  -> IVAL has been assigned to KEY.
C               IERR = -1 -> IVAL is not allowed as a value for the
C                            KEY keyword or there is not enough string
C                            storage available to hold SVAL.
C
C  ---------------------------------------------------------------------
      IMPLICIT $(TYPE) (A-H,O-Z)
C  Arguments:
      CHARACTER*(*) PANEL, KEY, SVAL
C  ---------------------------------------------------------------------
C  Common blocks:
C  Maximum length of a keyword.
      PARAMETER ( KEYLEN = 10 )
C  Maximum number of panel commands.
      PARAMETER ( MXPCOM = 80 )
C  Maximum number of keywords in all panel commands.
      PARAMETER ( MXPKEY = 1000 )
C  Maximum storage for integer/real keywords.
      PARAMETER ( MXPIR = 5000 )
C  Maximum storage for string keywords and descriptions.
      PARAMETER ( MXPSTR = 20000 )
C  Maximum number of pointers in the above string storage.
      PARAMETER ( MXPPST = 5000 )
C
      CHARACTER*(KEYLEN) PCOM(MXPCOM)
      INTEGER            PONF(MXPCOM)
      INTEGER            KPOINT(MXPCOM)
      INTEGER            NCHELP(2*MXPCOM)
      CHARACTER*(KEYLEN) PKEY(MXPKEY)
      INTEGER            NKHELP(2*MXPKEY)
      CHARACTER          KTYPE(MXPKEY)
      INTEGER            TPOINT(MXPKEY)
      CHARACTER          KIO(MXPKEY)
      $(TYPE)            PIR(MXPIR)
      INTEGER            SPOINT(MXPPST)
      CHARACTER*(MXPSTR) PANSTR
      COMMON /PCOMS1/ KPOINT, NCHELP, NKHELP, TPOINT, SPOINT, PONF
      COMMON /PCOMS2/ PIR
      COMMON /PCOMS3/ PKEY, KTYPE, PANSTR, PCOM, KIO
C
C  NPCOM: Number of panel commands in the panel description file.
C  NPKEY: Number of keywords in all panel commands.
C  NPIR:  Number of positions filled in array PIR.
C  NPSP:  Number of positions filled in the char-pointer array SPOINT.
C  NPCH:  Number of character positions filled in PANSTR.
      COMMON /PCOMS4/ NPCOM, NPKEY, NPIR, NPSP, NPCH
      CHARACTER*(KEYLEN) PREVPA
      COMMON /PRPA1/ PREVPA
      COMMON /PRPA2/ IPREVP
      LOGICAL STRO, WEAK
      COMMON /PRILOG/ STRO, WEAK
      COMMON /UNITS/ IUINP, IUOUT
C  ---------------------------------------------------------------------
C  Local variables:
      CHARACTER*(KEYLEN) PNAME, KNAME
C  ---------------------------------------------------------------------
C
      PNAME = PANEL
      KNAME = KEY
C
C  Look at the cache first.
      IF (ISCOMP(PREVPA,PNAME).NE.0) THEN
         I = IPREVP
      ELSE
C  Search for the panel.
         LE1 = LENGTH(PNAME)
         IF (LE1.EQ.0) CALL MABORT('Zero panel length.','SETPPS')
         CALL PLOCP(PNAME(1:LE1),I)
         IF (I.EQ.0)
     &          CALL MABORT('Panel not found: '//PNAME(1:LE1),'SETPPS')
C  Store in the cache.
         PREVPA = PNAME
         IPREVP = I
      END IF
C
C  Search for the keyword.
      LE1 = LENGTH(KNAME)
      IF (LE1.EQ.0) CALL MABORT('Zero keyword length.','SETPPS')
      CALL PLOCK(I,KNAME(1:LE1),I)
      IF (I.EQ.0)
     &    CALL MABORT('Keyword not found: '//KNAME(1:LE1),'SETPPS')
C
C  Get keyword type.
      IF (ISCOMP(KTYPE(I),'S').EQ.0 .AND. ISCOMP(KTYPE(I),'C').EQ.0)
     &    CALL MABORT('Not a string keyword '//KNAME(1:LE1),'SETPPS')
C
C  Is SVAL an allowed value ?
      CALL PRANS(SVAL,LES,I,IOK)
      IF (IOK.EQ.0) THEN
         IF (WEAK) WRITE (IUOUT,42) KNAME(1:LE1)
         IERR = -1
         RETURN
      END IF
C
C  Set the value.
      IP = TPOINT(I)
      IAL = NINT(PIR(IP))
      IF (IAL.EQ.1) THEN
      ELSE IF (IAL.EQ.2) THEN
         IP = IP+1
         NITE = NINT(PIR(IP))
         IP = IP+NITE
      END IF
      IP = IP+1
      IDEF = NINT(PIR(IP))
C
      IST = SPOINT(IDEF)
      IEND = SPOINT(IDEF+1)
      LESBEF = IEND-IST+1
      IDIFF = LES-LESBEF
      IF (NPCH+IDIFF.GT.MXPSTR) THEN
         IF (STRO) WRITE (IUOUT,41) KNAME(1:LE1)
         IERR = -1
         RETURN
      END IF
C
      IF (LES.GT.LESBEF) THEN
         DO 40,I=NPCH,IEND+1,-1
            IGO = I+IDIFF
            PANSTR(IGO:IGO) = PANSTR(I:I)
40       CONTINUE
         NPCH = NPCH+IDIFF
         DO 45,I=IDEF+1,NPSP
            SPOINT(I) = SPOINT(I)+IDIFF
45       CONTINUE
      ELSE IF (LES.LT.LESBEF) THEN
         DO 50,I=IEND+1,NPCH
            IGO = I+IDIFF
            PANSTR(IGO:IGO) = PANSTR(I:I)
50       CONTINUE
         DO 55,I=IDEF+1,NPSP
            SPOINT(I) = SPOINT(I)+IDIFF
55       CONTINUE
         NPCH = NPCH+IDIFF
      END IF
      IF (LES.GT.0) PANSTR(IST:SPOINT(IDEF+1)) = SVAL(1:LES)
      IERR = 0
C
41    FORMAT (/$(FC)'There is not enough space available to hold'
     &        /$(FC)'the value for the ',A,' keyword.')
42    FORMAT (/$(FC)'The specified value for the ',A,' keyword ',
     &        'is not allowed.')
      END
C  ---------------------------------------------------------------------
      SUBROUTINE R2NRTF ( X, STR, LES )
C  ---------------------------------------------------------------------
C
C  Description:
C    This subroutine converts a real number X to its string
C    representation using subroutine STORIC. Leading/trailing blanks
C    and zeroes are removed. The number of significant digits in the
C    output is controlled by the machine accuracy.
C
C  Input arguments:
C    X          Real number to be converted.
C
C  Output arguments:
C    STR        Contains the string representation of X.
C    LES        Effective length of STR.
C
C  ---------------------------------------------------------------------
      IMPLICIT $(TYPE) (A-H,O-Z)
C  Arguments:
      CHARACTER*(*) STR
C  ---------------------------------------------------------------------
C  Common blocks:
      COMMON /ACCUR1/ NDRTF
C  ---------------------------------------------------------------------
C  Local variables:
      CHARACTER*64 CH
C  ---------------------------------------------------------------------
C
      CALL STORIC(X,NDRTF,CH,L)
      LES = MIN(L,LEN(STR))
      STR = CH
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE R2RTF ( X, STR, LES )
C  ---------------------------------------------------------------------
C
C  Description:
C    This subroutine will write a real number (X) on a string (STR),
C    using the run time format 1PG__.__ constructed by subroutine MACAC.
C
C  Input arguments:
C    X          The number to be written on STR.
C
C  Output arguments:
C    STR        Upon return contains the real number X. Since a
C               maximum of 16 significant digits is imposed on all
C               formats, the actual length of STR should be no less
C               than 23 characters.
C    LES        Effective length of STR.
C
C  ---------------------------------------------------------------------
      IMPLICIT $(TYPE) (A-H,O-Z)
C  Arguments:
      CHARACTER*(*) STR
C  ---------------------------------------------------------------------
C  Common blocks:
      COMMON /ACCUR1/ NDRTF
      CHARACTER RTFORM*10, RRFORM*7, REFORM*10
      COMMON /ACCUR2/ RTFORM, RRFORM, REFORM
C  ---------------------------------------------------------------------
C
      WRITE (STR,RTFORM) X
      LES = NDRTF+7
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE R2ERTF ( X, STR, LES )
C  ---------------------------------------------------------------------
C
C  Description:
C    This subroutine will write a real number (X) on a string (STR),
C    using the run time format 1PE__.__ constructed by subroutine MACAC.
C
C  Input arguments:
C    X          The number to be written on STR.
C
C  Output arguments:
C    STR        Upon return contains the real number X. Since a
C               maximum of 16 significant digits is imposed on all
C               formats, the actual length of STR should be no less
C               than 23 characters.
C    LES        Effective length of STR.
C
C  ---------------------------------------------------------------------
      IMPLICIT $(TYPE) (A-H,O-Z)
C  Arguments:
      CHARACTER*(*) STR
C  ---------------------------------------------------------------------
C  Common blocks:
      COMMON /ACCUR1/ NDRTF
      CHARACTER RTFORM*10, RRFORM*7, REFORM*10
      COMMON /ACCUR2/ RTFORM, RRFORM, REFORM
C  ---------------------------------------------------------------------
C
      WRITE (STR,REFORM) X
      LES = NDRTF+7
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE R2STR ( X, STR, LES )
C  ---------------------------------------------------------------------
C
C  Description:
C    Converts a real number X to its string representation. This routine
C    is just a convenience wrapper around the main conversion subroutine
C    STORIC. Only 5 significant digits are returned, while
C    leading/trailing spaces and zeroes are removed.
C
C  Input arguments:
C    X          Real number to be converted.
C
C  Output arguments:
C    STR        String representation of X.
C    LES        Number of useful characters in STR.
C
C  ---------------------------------------------------------------------
      IMPLICIT $(TYPE) (A-H,O-Z)
C  Arguments:
      CHARACTER*(*) STR
C  ---------------------------------------------------------------------
C  Local variables:
      CHARACTER*30 CH
C  ---------------------------------------------------------------------
C
C  Call the actual conversion routine. Request 5 significant digits.
      CALL STORIC(X,5,CH,L)
      LES = MIN(L,LEN(STR))
      STR = CH
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE STORIC ( X, NDIG, CH, LE )
C  ---------------------------------------------------------------------
C
C  Description:
C    Subroutine storic converts a real number X in its string
C    representation (CH). The number of significant digits in the
C    output can be set by NDIG. The format used, depends on the
C    magnidute of X. For small values of X:
C       10**(-NDIG) <= ABS(X) < 10**NDIG
C    an F format is used. Larger values require an E format.
C    A trailing blank is always added at the end of CH.
C
C  Input arguments:
C    X          The number to be converted.
C    NDIG       Number of significant digits in the output string.
C               If NDIG <= 0 then the full machine accuracy will
C               be used.
C
C  Output arguments:
C    CH         Contains the converted number.
C    LE         Effective length of STR.
C
C  ---------------------------------------------------------------------
      IMPLICIT $(TYPE) (A-H,O-Z)
C  Arguments:
      CHARACTER*(*) CH
C  ---------------------------------------------------------------------
C  Common blocks:
      COMMON /ACCUR1/ NDRTF
C  ---------------------------------------------------------------------
C  Local variables:
      CHARACTER*30 FORM, CNUM
      CHARACTER*6 CEXP
      CHARACTER C
C  ---------------------------------------------------------------------
C
C  Treat a zero.
      IF (X.EQ.0.) THEN
         CH = '0.'
         LE = 2
         RETURN
      END IF
C
C  Determine number of significant digits.
      IF (NDIG.LE.0) THEN
         ND = NDRTF
      ELSE
         ND = NDIG
      END IF
C
      TEN = 10.
      AX = ABS(X)
      A = TEN**(-ND)
      B = TEN**ND
      IF (AX.GE.A .AND. AX.LT.B) THEN
C
C  Use an F format.
         IDIG = INT(LOG10(AX))
         IF (IDIG.LT.0) THEN
            IDIG = 0
         ELSE
            IDIG = IDIG+1
         END IF
         LENF = ND+3
         WRITE (FORM,'(''(F'',I2.2,''.'',I2.2,'')'')') LENF, ND-IDIG
         WRITE (CNUM,FORM) X
C
C  Remove trailing zeros.
         DO 10,I=LENF,1,-1
            C = CNUM(I:I)
            IF (C.NE.' ' .AND. C.NE.'0') THEN
               IEND = I
               GOTO 11
            END IF
10       CONTINUE
         CALL MABORT('All digits are zeros.','STORIC')
11       CONTINUE
      ELSE
C
C  Use an E format. Allow room for 3 digits in the exponent.
         LENF = ND+8
         WRITE (FORM,'(''(1PE'',I2.2,''.'',I2.2,''E3)'')') LENF, ND
         WRITE (CNUM,FORM) X
C
C  Remove trailing zeros.
         IDOT = INDEX(CNUM,'.')
         IZST = IDOT+ND
         DO 40,I=IZST,IDOT,-1
            IF (CNUM(I:I).NE.'0') THEN
               CEXP = CNUM(IZST+1:LENF)
               CNUM(I+1:LENF) = CEXP
               IEND = LENF-(IZST-I)
               GOTO 41
            END IF
40       CONTINUE
         IEND = LENF
41       CONTINUE
      END IF
C
C  Remove leading zeroes in the exponent.
      IE = INDEX(CNUM,'E')
      IE = IE+1
      IF (CNUM(IE:IE).EQ.'+' .OR. CNUM(IE:IE).EQ.'-') IE = IE+1
      DO 30,I=IE,LENF
         IF (CNUM(IE:IE).EQ.'0') THEN
            CEXP = CNUM(IE+1:LENF)
            CNUM(IE:LENF) = CEXP
            IEND = IEND-1
         ELSE
            GOTO 31
         END IF
30    CONTINUE
      CALL MABORT('Exponent is zero.','STORIC')
31    CONTINUE
C
C  Remove leading blanks.
      DO 20,I=1,LENF
         IF (CNUM(I:I).NE.' ') THEN
            ISTART = I
            GOTO 21
         END IF
20    CONTINUE
      CALL MABORT('All digits are blanks.','STORIC')
21    CONTINUE
C
      LE = IEND-ISTART+1
      CH = CNUM(ISTART:IEND)
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE VNAME ( IND, NAME, LENA )
C  ---------------------------------------------------------------------
C
C  Description:
C    Returns a string that contains the index IND of a minimization
C    variable, and its name (if one is defined). The output is used
C    mainly for producing readable messages.
C
C  Input arguments:
C    IND        Index of the variable.
C
C  Output arguments:
C    NAME       The output string. This is of the form: index (name)
C               Eg: 2 (alpha)
C    LENA       Number of useful characters in NAME.
C
C  ---------------------------------------------------------------------
C  Arguments:
      CHARACTER*(*) NAME
C  ---------------------------------------------------------------------
C  Common blocks:
      PARAMETER ( MXV = $(MXV) )
      CHARACTER*10 NICKNA
      COMMON /NAMES/ NICKNA(MXV)
C  ---------------------------------------------------------------------
C  Local variables:
      CHARACTER*10 STR
C  ---------------------------------------------------------------------
C
      LACT = LEN(NAME)
      CALL I2STR(IND,STR,LES)
      LE1 = LENGTH(NICKNA(IND))
      IF (LE1.EQ.0) THEN
         NAME = STR(1:LES)
         LENA = MIN(LACT,LES)
      ELSE
         NAME = STR(1:LES)//' ('//NICKNA(IND)(1:LE1)//')'
         LENA = MIN(LACT,LES+LE1+3)
      END IF
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE ACCUM ( MERR )
C  ---------------------------------------------------------------------
C
C  Description:
C    Implements the ACCUM command.
C
C  Output arguments:
C    MERR       Error indicator.
C
C  ---------------------------------------------------------------------
      IMPLICIT $(TYPE) (A-H,O-Z)
C  Common blocks:
      PARAMETER ( MXV = $(MXV) )
      COMMON /PARAMS/ POINT(MXV), VAL
      COMMON /MODE3/ MODEV, IFFORM, INMCL
      COMMON /DIMEN/ NOD, NOF
      COMMON /TOTCAL/ ITNOCS, IPARTC, ITDER, IPDER, ITHESS, IPHESS,
     &                ITJAC, IPJAC
C  ---------------------------------------------------------------------
C  Local variables:
      PARAMETER ( MXFILE = 80 )
      CHARACTER*(MXFILE) FILE
      CHARACTER*5 TYPE
C  ---------------------------------------------------------------------
C
C  Set MCL-returned values.
      CALL SETPI('INFO',0,IERR)
      CALL SETPI('POINTS',0,IERR)
      CALL SETPI('FCALLS',0,IERR)
C
C  Default target value is always the current value.
      IF (ITNOCS.NE.0) CALL SETPR('TARGET',VAL,IERR)
      CALL SETPI('CANCEL',1,IERR)
      CALL CHANGE(MERR)
      IF (MERR.NE.0) GOTO 100
      CALL GETPI('CANCEL',ICANC)
      IF (ICANC.EQ.0 .OR. ITNOCS.EQ.0 .OR. MODEV.EQ.0) GOTO 100
C
C  Get panel parameters.
      CALL GETPI('NOC',NOCAC)
      CALL GETPI('NOP',NOPAC)
      CALL GETPR('TARGET',TARAC)
      CALL GETPS('FILE',FILE,LENF)
      CALL GETPS('TYPE',TYPE,LET)
C
C  Open the output file.
      CALL OPENIT(FILE,TYPE,1,1,IUINI,NREC)
      IF (IUINI.EQ.-1) THEN
         MERR = -2
         GOTO 100
      END IF
      IF (ISCOMP(TYPE,'TEXT').NE.0) THEN
         MODE = 1
      ELSE
         MODE = 2
      END IF
C
C  If text mode was specified, position the unit at End-of-File.
      IF (MODE.EQ.1) CALL POSIT(IUINI)
C
C  Add a backup record.
      CALL BACKUP('M')
      CALL ACCUMP(NOPAC,NOCAC,TARAC,IERFL,IPOI,IUINI,FILE,MODE,NREC,
     &            NOCEX)
C
C  If binary mode was specified, write the total number of records
C  in the file.
      IF (MODE.EQ.2 .AND. IPOI.NE.0) WRITE (IUINI,REC=1) NOD, NREC+IPOI
      CLOSE (IUINI)
C
C  Set MCL-returned values.
100   CONTINUE
      CALL SETPI('INFO',MERR,IERR)
      CALL SETPI('POINTS',IPOI,IERR)
      CALL SETPI('FCALLS',NOCEX,IERR)
      END
C  ---------------------------------------------------------------------
      SUBROUTINE ACCUMP ( NOPOI, NOCAL, VALTA, IERFL, IPOI, IUINI,
     &                    FILE, MODE, NREC, NOC )
C  ---------------------------------------------------------------------
C
C  Description:
C    Part of the ACCUM command. Accumulates at random points that
C    satisfy the criterion: ACSQ(X,N) <= VALTA.
C    This routine terminates normally if either NOPOI points are
C    acumulated or the objective function is called NOCAL times,
C    whatever comes first.
C
C  Input arguments:
C    NOPOI      The desired number of points to be accumulated.
C    NOCAL      Maximum number of calls to the objective function.
C    VALTA      The target value. Points with value less or equal than
C               VALTA are accepted and disposed in unit IUINI.
C    IUINI      The unit number where the accumulated points are
C               disposed.
C    FILE       File name to accept the accumulated points. It should
C               be opened according to MODE, as unit IUINI.
C    MODE       File type. 1 for a text file, 2 for a binary file.
C    NREC       Number of existing records in unit IUINI. Ignored
C               if mode = 1.
C
C  Output arguments:
C    IERFL      Set to 1 if there are non-fixed variables without both
C               margins; 0 otherwise.
C    IPOI       Number of points accumulated and written to the
C               specified file.
C    NOC        Number of calls performed to function ACSQ.
C
C  ---------------------------------------------------------------------
      IMPLICIT $(TYPE) (A-H,O-Z)
C  Arguments:
      CHARACTER*(*) FILE
C  ---------------------------------------------------------------------
C  Common blocks:
      PARAMETER ( MXV = $(MXV) )
      COMMON /PARAMS/ POINT(MXV), VAL
      COMMON /DETAIL/ IXAT(MXV), MARG(MXV)
      COMMON /XLIMS/ XLL(MXV), XRL(MXV)
      COMMON /DIMEN/ NOD, NOF
      LOGICAL STRO, WEAK
      COMMON /PRILOG/ STRO, WEAK
      COMMON /UNITS/ IUINP, IUOUT
      COMMON /RWORK/ XT1(MXV), XT2(MXV), XT3(MXV), XT4(MXV), XT5(MXV),
     &               XT6(MXV), XT7(MXV+1)
C  ---------------------------------------------------------------------
C  Local variables:
      CHARACTER*20 STR
C  ---------------------------------------------------------------------
C
C  Constants.
      ZERO = 0.
      ONE = 1.
C  Initialize counters.
      NOC = 0
      IPOI = 0
      IERFL = 0
C
C  Check for proper framing of the non-fixed variables.
      DO 1,I=1,NOD
         IF (MARG(I).NE.2 .AND. IXAT(I).NE.0) THEN
            IF (WEAK) THEN
               CALL VNAME(I,STR,LES)
               WRITE (IUOUT,6) STR(1:LES)
            END IF
            IERFL = 1
         END IF
1     CONTINUE
      IF (IERFL.NE.0) RETURN
C
C  The following loop samples points at random.
      ITER = 1
C  ---------------------------------------------------------------------
4     CONTINUE
C
C  Sample a point in between its margins.
      DO 2,I=1,NOD
         XT1(I) = POINT(I)
         XLDIF = (XRL(I)-XLL(I))*RANDI(ZERO,ONE)+XLL(I)-XT1(I)
         XT1(I) = XT1(I)+IXAT(I)*XLDIF
2     CONTINUE
      VALT = ACSQ(XT1,NOD)
      NOC = NOC+1
C
C  Check if the trial value is less than the target value.
C  If the criterion is satisfied dispose point and value to unit IUINI.
      IF (VALT.LE.VALTA) THEN
         IPOI = IPOI+1
         CALL WRITER(XT1,VALT,IUINI,MODE,NREC+IPOI)
      END IF
C
C  Keep point and value it if less than the current point.
      IF (VALT.LT.VAL) THEN
         DO 10,I=1,NOD
10          POINT(I) = XT1(I)
         VAL = VALT
      END IF
C
C  Need to terminate ?
      IF (NOC.GE.NOCAL .OR. IPOI.GE.NOPOI) GOTO 5
      ITER = ITER+1
      GOTO 4
C  ---------------------------------------------------------------------
C
5     CONTINUE
      IF (STRO) THEN
         IF (IPOI.EQ.0) THEN
            WRITE (IUOUT,7)
         ELSE
            CALL I2STR(IPOI,STR,LES)
            WRITE (IUOUT,8) FILE(1:LENGTH(FILE)), STR(1:LES)
         END IF
         CALL I2STR(NOC,STR,LES)
         WRITE (IUOUT,20) STR(1:LES)
      END IF
C
6     FORMAT ($(FC)'Variable ',A,' is neither framed nor fixed.')
7     FORMAT (/$(FC)'Failed to accumulate any new points.')
8     FORMAT (/$(FC)'Number of points appended to file "',A,'": ',A)
20    FORMAT ($(FC)'Function calls: ',A)
      END
C  ---------------------------------------------------------------------
      SUBROUTINE NEWALI ( MERR )
C  ---------------------------------------------------------------------
C
C  Description:
C    Implements the ALIAS command.
C    Syntax: ALIAS alias_name command
C    Aliases are small (up to 5 characters) substitutes for the actual
C    Merlin commands.
C
C  Output arguments:
C    MERR       Error indicator.
C
C  ---------------------------------------------------------------------
C  Common blocks:
      LOGICAL STRO, WEAK
      COMMON /PRILOG/ STRO, WEAK
      COMMON /UNITS/ IUINP, IUOUT
      PARAMETER ( MAXALI = 20 )
      CHARACTER*5 ALIAS(MAXALI)
      CHARACTER*10 ALICOM(MAXALI)
      COMMON /ALICO1/ ALIAS, ALICOM
      COMMON /ALICO2/ NALIAS
      PARAMETER ( NMCOM = 98, MXPLUG = $(MXPLUG), NCOM = NMCOM+MXPLUG )
      CHARACTER*10 COMMAN(NCOM)
      COMMON /COM1/ COMMAN
C  ---------------------------------------------------------------------
C  Local variables:
      PARAMETER ( LINLEN = 80 )
      CHARACTER*(LINLEN) ALI, COM
C  ---------------------------------------------------------------------
C
      MERR = 0
C
C  Too many aliases ?
      IF (NALIAS.EQ.MAXALI) THEN
         IF (WEAK) WRITE (IUOUT,30)
         MERR = -2
         RETURN
      END IF
C
      CALL ARGNO(NA)
      IF (NA.EQ.0) THEN
         IF (STRO) WRITE (IUOUT,10)
         CALL INNOCR(ALI,.FALSE.,MERR)
         IF (MERR.NE.0) RETURN
C
         IF (STRO) WRITE (IUOUT,20)
         CALL INNOCR(COM,.FALSE.,MERR)
         IF (MERR.NE.0) RETURN
      ELSE IF (NA.EQ.2) THEN
         CALL CMDARG(1,ALI,LEALI)
         CALL CMDARG(2,COM,LECOM)
      ELSE
         CALL ERRNAR
         MERR = -2
         RETURN
      END IF
      LEALI = LENGTH(ALI)
      LECOM = LENGTH(COM)
      IF (LEALI.EQ.0 .OR. LECOM.EQ.0) THEN
         IF (WEAK) WRITE (IUOUT,90)
         MERR = -2
         RETURN
      END IF
C
C  Is this alias already defined ?
      DO 40,I=1,NALIAS
         IF (ISCOMP(ALI,ALIAS(I)).NE.0) THEN
            IF (WEAK) WRITE (IUOUT,50) ALI(1:LEALI)
            MERR = -2
            RETURN
         END IF
40    CONTINUE
C
C  Check whether this alias would override any of the Merlin commands.
      CALL MATCH(COMMAN,NCOM,ALI,LEALI,MINF,ICOM,IEND)
      IF (MINF.EQ.0 .AND. WEAK)
     &    WRITE (IUOUT,60) COMMAN(ICOM)(1:LENGTH(COMMAN(ICOM)))
C
C  Insert the alias in the alias-table.
80    CONTINUE
      NALIAS = NALIAS+1
      ALIAS(NALIAS) = ALI
      ALICOM(NALIAS) = COM
      IF (STRO) WRITE (IUOUT,70) ALI(1:LEALI), COM(1:LECOM)
C
10    FORMAT (/$(FC)'Enter alias:')
20    FORMAT (/$(FC)'Enter actual command:')
30    FORMAT (/$(FC)'You have too many aliases. Cannot define ',
     &        'any more.')
50    FORMAT (/$(FC)'Alias "',A,'" exists already.')
60    FORMAT (/$(FC)'Warning: This alias will override the "',A,
     &        '" command.')
70    FORMAT (/$(FC)'Alias "',A,'" has been defined as the command "',
     &        A,'".')
90    FORMAT (/$(FC)'The alias name, and/or the actual command, ',
     &        'is missing.')
      END
C  ---------------------------------------------------------------------
      SUBROUTINE ALIDIS
C  ---------------------------------------------------------------------
C
C  Description:
C    Implements the ALIASDIS command.
C    Syntax: ALIASDIS
C    Displays all aliases that are currently defined.
C
C  ---------------------------------------------------------------------
C  Common blocks:
      LOGICAL STRO, WEAK
      COMMON /PRILOG/ STRO, WEAK
      COMMON /UNITS/ IUINP, IUOUT
      PARAMETER ( MAXALI = 20 )
      CHARACTER*5 ALIAS(MAXALI)
      CHARACTER*10 ALICOM(MAXALI)
      COMMON /ALICO1/ ALIAS, ALICOM
      COMMON /ALICO2/ NALIAS
C  ---------------------------------------------------------------------
C
      IF (STRO) THEN
         IF (NALIAS.EQ.0) THEN
            WRITE (IUOUT,40)
         ELSE
            WRITE (IUOUT,10)
            DO 20,I=1,NALIAS
20             WRITE (IUOUT,30) ALIAS(I), ALICOM(I)
         END IF
      END IF
C
10    FORMAT (/$(FC)'Current aliases are:',
     &        /$(FC)'--------------------')
30    FORMAT ($(FC)A,' = ',A)
40    FORMAT (/$(FC)'There are no aliases defined.')
      END
C  ---------------------------------------------------------------------
      SUBROUTINE UNALI ( MERR )
C  ---------------------------------------------------------------------
C
C  Description:
C    Implements the UNALIAS command.
C    Syntax: UNALIAS alias_name
C    Removes a previously defined alias.
C
C  Output arguments:
C    MERR       Error indicator.
C
C  ---------------------------------------------------------------------
C  Common blocks:
      LOGICAL STRO, WEAK
      COMMON /PRILOG/ STRO, WEAK
      COMMON /UNITS/ IUINP, IUOUT
      PARAMETER ( MAXALI = 20 )
      CHARACTER*5 ALIAS(MAXALI)
      CHARACTER*10 ALICOM(MAXALI)
      COMMON /ALICO1/ ALIAS, ALICOM
      COMMON /ALICO2/ NALIAS
C  ---------------------------------------------------------------------
C  Local variables:
      CHARACTER*80 ALI
C  ---------------------------------------------------------------------
C
      MERR = 0
      CALL ARGNO(NA)
      IF (NA.EQ.0) THEN
         IF (STRO) WRITE (IUOUT,10)
         CALL INNOCR(ALI,.FALSE.,MERR)
         IF (MERR.NE.0) RETURN
      ELSE IF (NA.EQ.1) THEN
         CALL CMDARG(1,ALI,LEALI)
      ELSE
         MERR = -2
         CALL ERRNAR
         RETURN
      END IF
      LEALI = LENGTH(ALI)
      IF (LEALI.EQ.0) THEN
         IF (WEAK) WRITE (IUOUT,90)
         MERR = -2
         RETURN
      END IF
C
C  Locate the alias in the alias-table.
      DO 40,I=1,NALIAS
         IF (ISCOMP(ALI,ALIAS(I)).NE.0) THEN
C  Alias located. Now remove it.
            DO 50,J=I+1,NALIAS
               ALIAS(J-1) = ALIAS(J)
               ALICOM(J-1) = ALICOM(J)
50          CONTINUE
            NALIAS = NALIAS-1
            IF (STRO) WRITE (IUOUT,30) ALI(1:LEALI)
            RETURN
         END IF
40    CONTINUE
      IF (WEAK) WRITE (IUOUT,20) ALI(1:LEALI)
      MERR = -2
C
10    FORMAT (/$(FC)'Enter alias to remove:')
20    FORMAT (/$(FC)'There is no alias named "',A,'".')
30    FORMAT (/$(FC)'Alias "',A,'" has been removed.')
90    FORMAT (/$(FC)'The alias name is missing.')
      END
C  ---------------------------------------------------------------------
      SUBROUTINE BACK ( BMODE, MERR )
C  ---------------------------------------------------------------------
C
C  Description:
C    Implements the BACKUP command.
C    Syntax: BACKUP (Has panel)
C
C  Input arguments:
C    BMODE      If BMODE is empty, the backup panel is presented.
C               Otherwise the backup mode is set to BMODE and backup
C               starts immediately.
C
C  Output arguments:
C    MERR       Error indicator.
C
C  ---------------------------------------------------------------------
C  Arguments:
      CHARACTER*(*) BMODE
C  ---------------------------------------------------------------------
C  Common blocks:
      COMMON /BACKIT/ IUBA, IUBAMO, IBASA
      PARAMETER ( MXFILE = 80 )
      CHARACTER*10 WHEBAK, MODBAK
      CHARACTER*(MXFILE) FILBAK
      COMMON /BAKIT2/ WHEBAK, MODBAK, FILBAK
      LOGICAL STRO, WEAK
      COMMON /PRILOG/ STRO, WEAK
      COMMON /UNITS/ IUINP, IUOUT
C  ---------------------------------------------------------------------
C  Local variables:
      CHARACTER*(MXFILE) FILE
      CHARACTER*80 WHEN, SAFE
      CHARACTER*5 CTYPE, CMODE
C  ---------------------------------------------------------------------
C
      MERR = 0
      LEM = LENGTH(BMODE)
      IF (LEM.EQ.0) THEN
         CALL SETPI('CANCEL',1,IERR)
         CALL CHANGE(MERR)
         IF (MERR.NE.0) RETURN
         CALL GETPI('CANCEL',ICANC)
         IF (ICANC.EQ.0) RETURN
         CALL GETPS('MODE',CMODE,LEM)
      ELSE
         CMODE = BMODE
         CALL SETPPS('BACKUP','MODE',CMODE,LEM,IERR)
      END IF
C
C  Get panel parameters.
      CALL GETPPS('BACKUP','FILE',FILE,LEF)
      CALL GETPPS('BACKUP','TYPE',CTYPE,LET)
      CALL GETPPS('BACKUP','WHEN',WHEN,LEW)
      CALL GETPPS('BACKUP','SAFE',SAFE,LEW)
      WHEBAK = WHEN
      CALL UPPER(WHEBAK,LEN(WHEBAK))
      MODBAK = CMODE
      CALL UPPER(MODBAK,LEN(MODBAK))
      IF (ISCOMP(SAFE,'YES').NE.0) THEN
         IBASA = 1
      ELSE
         IBASA = 0
      END IF
C
C  Proceed according to the backup mode CMODE.
      IF (ISCOMP(CMODE,'NO').NE.0) THEN
         IF (IUBA.NE.-1) THEN
            CALL U2NAME(IUBA,'backup',FILE,LEF)
            CLOSE (IUBA)
            IUBA = -1
            IF (LEF.EQ.0) THEN
               IF (STRO) WRITE (IUOUT,50)
            ELSE
               IF (STRO) WRITE (IUOUT,60) FILE(1:LEF)
            END IF
         END IF
         RETURN
      END IF
C
C  Is the backup file alredy open ?
      IF (IUBA.NE.-1 .AND. STRO) THEN
         CALL U2NAME(IUBA,'backup',FILE,LEF)
         IF (LEF.NE.0) THEN
            WRITE (IUOUT,70) FILE(1:LEF)
         ELSE
            WRITE (IUOUT,75)
         END IF
         RETURN
      END IF
C
C  Backup file not open.
      IF (ISCOMP(CTYPE,'TEXT').NE.0) THEN
         IUBAMO = 1
      ELSE
         IUBAMO = 2
      END IF
      CALL OPENIT(FILE,CTYPE,1,1,IUBA,NREC)
      FILBAK = FILE
      IF (IUBA.EQ.-1) THEN
         MERR = -2
      ELSE
         IF (STRO) THEN
            IF (IBASA.EQ.0) THEN
               WRITE (IUOUT,100) FILE(1:LEF)
            ELSE
               WRITE (IUOUT,110) FILE(1:LEF)
            END IF
         END IF
      END IF
C
50    FORMAT (/$(FC)'Backup has been stopped.')
60    FORMAT (/$(FC)'Backup to file "',A,'" has been stopped.')
70    FORMAT (/$(FC)'Backup is already in effect. Backup file is "',
     &        A,'".')
75    FORMAT (/$(FC)'Backup is already in effect.')
100   FORMAT (/$(FC)'Backup to file "',A,'" started.')
110   FORMAT (/$(FC)'Safe backup to file "',A,'" started.')
      END
C  ---------------------------------------------------------------------
      SUBROUTINE BACKUP ( WHEN )
C  ---------------------------------------------------------------------
C
C  Description:
C    Adds a record to the backup file according to the modes specified
C    in the backup panel. Assumes the backup file is correctly
C    positioned for FULLBACK.
C
C  Input arguments:
C    WHEN       Character variable, indicating the reason for the
C               backup. If it matches the ones defined in the
C               backup panel, then bckup will proceed.
C
C  ---------------------------------------------------------------------
      IMPLICIT $(TYPE) (A-H,O-Z)
C  Arguments:
      CHARACTER*(*) WHEN
C  ---------------------------------------------------------------------
C  Common blocks:
      PARAMETER ( MXV = $(MXV) )
      COMMON /PARAMS/ POINT(MXV), VAL
      COMMON /DIMEN/ NOD, NOF
      COMMON /BACKIT/ IUBA, IUBAMO, IBASA
      PARAMETER ( MXFILE = 80 )
      CHARACTER*10 WHEBAK, MODBAK
      CHARACTER*(MXFILE) FILBAK
      COMMON /BAKIT2/ WHEBAK, MODBAK, FILBAK
      COMMON /TOTCAL/ ITNOCS, IPARTC, ITDER, IPDER, ITHESS, IPHESS,
     &                ITJAC, IPJAC
C  ---------------------------------------------------------------------
C  Local variables:
      CHARACTER W
      CHARACTER*4 CTYPE
C  ---------------------------------------------------------------------
C
      IF (IUBA.EQ.-1 .OR. ITNOCS.EQ.0) RETURN
C
      W = WHEN(1:1)
      CALL UPPER(W,1)
C
      DO 10,I=1,LEN(WHEBAK)
         IF (W.EQ.WHEBAK(I:I)) GOTO 20
10    CONTINUE
      RETURN
C
20    CONTINUE
      IF (IUBAMO.EQ.1) THEN
         IF (MODBAK.EQ.'LAST') THEN
            REWIND IUBA
            WRITE (IUBA,'(I6)') NOD
         END IF
         CALL WRITER(POINT,VAL,IUBA,IUBAMO,0)
      ELSE
         IF (MODBAK.EQ.'LAST') THEN
            NBAREC = 1
         ELSE
            READ (IUBA,REC=1) NVAR, NBAREC
            NBAREC = NBAREC+1
         END IF
         CALL WRITER(POINT,VAL,IUBA,IUBAMO,NBAREC)
         WRITE (IUBA,REC=1) NOD, NBAREC
      END IF
C
C  Safe backup ?
      IF (IBASA.NE.0) THEN
         CLOSE (IUBA)
         IF (IUBAMO.EQ.1) THEN
            CTYPE = 'TEXT'
         ELSE
            CTYPE = 'BIN'
         END IF
         CALL OPENIT(FILBAK,CTYPE,1,1,IUBA,NREC)
      END IF
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE CHCFLA ( MERR )
C  ---------------------------------------------------------------------
C
C  Description:
C    Implements the CFLAG command.
C    Syntax: CFLAG index_specification value ...
C
C  Output arguments:
C    MERR       Error indicator.
C
C  ---------------------------------------------------------------------
      IMPLICIT $(TYPE) (A-H,O-Z)
C  Common blocks:
      PARAMETER ( MXV = $(MXV) )
      COMMON /IWORK/ INDE(MXV)
      PARAMETER ( MAXCFL = 10, MAXCLE = 30 )
      CHARACTER*(MAXCLE) CFLAG(MAXCFL)
      COMMON /CFLAGS/ CFLAG
C  ---------------------------------------------------------------------
C  Local variables:
      PARAMETER ( LINLEN = 80 )
      CHARACTER*(LINLEN) LINE
      INTEGER IS(LINLEN/2), IE(LINLEN/2), ITEM(MAXCFL)
C  ---------------------------------------------------------------------
C
C  Prompt the user, or parse command line arguments for index/value
C  specifications.
      CALL CHISX(INDE,ITEM,NCH,MAXCFL,LINE,IS,IE,1,MERR)
      IF (MERR.NE.0) RETURN
C
      DO 100,I=1,NCH
         IND = INDE(I)
         N = ITEM(I)
         CFLAG(IND) = LINE(IS(N):IE(N))
100   CONTINUE
      END
C  ---------------------------------------------------------------------
      SUBROUTINE CFLADI ( MERR )
C  ---------------------------------------------------------------------
C
C  Description:
C    Implements the CFLAGDIS command.
C    Syntax: CFLAGDIS [ index_specification ... ]
C
C  Output arguments:
C    MERR       Error indicator.
C
C  ---------------------------------------------------------------------
      IMPLICIT $(TYPE) (A-H,O-Z)
C  Arguments:
C  ---------------------------------------------------------------------
C  Common blocks:
      PARAMETER ( MXV = $(MXV) )
      COMMON /IWORK/ INDE(MXV)
      LOGICAL STRO, WEAK
      COMMON /PRILOG/ STRO, WEAK
      COMMON /UNITS/ IUINP, IUOUT
      PARAMETER ( MAXCFL = 10, MAXCLE = 30 )
      CHARACTER*(MAXCLE) CFLAG(MAXCFL)
      COMMON /CFLAGS/ CFLAG
C  ---------------------------------------------------------------------
C
C  Parse any command line arguments for index specifications.
      CALL CHIX(MAXCFL,INDE,NCH,-2,MERR)
      IF (MERR.NE.0) RETURN
C
      IF (STRO) THEN
         WRITE (IUOUT,30)
         DO 20,I=1,NCH
            IND = INDE(I)
            WRITE (IUOUT,10) IND, CFLAG(IND)
20       CONTINUE
      END IF
C
10    FORMAT ($(FC)'Flag(',I2,')= ''',A,'''')
30    FORMAT (/$(FC)'Current character flag values:'
     &        /$(FC)'------------------------------')
      END
C  ---------------------------------------------------------------------
      SUBROUTINE CHPOI ( NCH, IERR )
C  ---------------------------------------------------------------------
C
C  Description:
C    Part of the POINT command. Accepts and processes a one-line-input.
C
C  Output arguments:
C    NCH        Number of pairs "index value" that were processed.
C    MERR       Error indicator.
C
C  ---------------------------------------------------------------------
      IMPLICIT $(TYPE) (A-H,O-Z)
C  Common blocks:
      PARAMETER ( MXV = $(MXV) )
      COMMON /PARAMS/ POINT(MXV), VAL
      COMMON /DETAIL/ IXAT(MXV), MARG(MXV)
      COMMON /INDEFI/ INDEF(MXV)
      COMMON /IWORK/ INDE(MXV)
      COMMON /RWORK/ XT1(MXV), XT2(MXV), XT3(MXV), XT4(MXV), XT5(MXV),
     &               XT6(MXV), XT7(MXV+1)
C  ---------------------------------------------------------------------
C
      CALL CHIV(INDE,XT1,NCH,1,IERR)
      IF (IERR.NE.0) RETURN
C
C  Add a backup record.
      IF (NCH.NE.0) CALL BACKUP('X')
C
      DO 10,I=1,NCH
         IND = INDE(I)
         POINT(IND) = XT1(I)
         IXAT(IND) = 1
         INDEF(IND) = 1
10    CONTINUE
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE CHSTEP ( MERR )
C  ---------------------------------------------------------------------
C
C  Description:
C    Implements the STEP command.
C    Syntax: STEP index_specification value ...
C
C  Output arguments:
C    MERR       Error indicator.
C
C  ---------------------------------------------------------------------
      IMPLICIT $(TYPE) (A-H,O-Z)
C  Common blocks:
      PARAMETER ( MXV = $(MXV) )
      COMMON /ROSTEP/ STEP(MXV)
      COMMON /IWORK/ INDE(MXV)
      COMMON /RWORK/ XT1(MXV), XT2(MXV), XT3(MXV), XT4(MXV), XT5(MXV),
     &               XT6(MXV), XT7(MXV+1)
C  ---------------------------------------------------------------------
C
C  Prompt the user, or parse command line arguments for index/value
C  specifications.
      CALL CHIV(INDE,XT1,NPAIRS,1,MERR)
      IF (MERR.NE.0) RETURN
C
      DO 10,I=1,NPAIRS
10       STEP(INDE(I)) = XT1(I)
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE CLEAR ( MERR )
C  ---------------------------------------------------------------------
C
C  Description:
C    Implements the CLEAR command.
C
C  Output arguments:
C    MERR       Error indicator.
C
C  ---------------------------------------------------------------------
C  Common blocks:
      LOGICAL STRO, WEAK
      COMMON /PRILOG/ STRO, WEAK
      COMMON /CNTRL/ IDISK, MACECH, LIMCLK, IECOMA, IPROEC, MCLUBU
      PARAMETER ( MAXINL = 20 )
      INTEGER BACKIN(MAXINL), WHATIN(MAXINL), REDIN(MAXINL)
      COMMON /INNER1/ LEVIN, BACKIN, WHATIN, REDIN
      CHARACTER*10 MCPREV(MAXINL)
      COMMON /INNER2/ MCPREV
      COMMON /UNITS/ IUINP, IUOUT
C  ---------------------------------------------------------------------
C  Local variables:
      CHARACTER*80 LINE
C  ---------------------------------------------------------------------
C
      MERR = 0
C
C  LEVIN=0 means we are reading commands from the Merlin input file.
      IF (LEVIN.EQ.0) THEN
         IF (WEAK) WRITE (IUOUT,10)
         MERR = -2
      ELSE
         IF (STRO .AND. MACECH.NE.0) THEN
            LE = LENGTH(MCPREV(LEVIN))
            LINE = 'Execution of macro '//MCPREV(LEVIN)(1:LE)//
     &             ' has been completed'
            CALL DECILE(2)
            CALL MID1(LINE)
         ELSE
            CALL DECILE(2)
         END IF
      END IF
C
10    FORMAT (/$(FC)'CLEAR has no effect outside a macro.')
      END
C  ---------------------------------------------------------------------
      SUBROUTINE CONFID ( H, MAXW, MERR )
C  ---------------------------------------------------------------------
C
C  Description:
C    Implements the CONFIDENCE command.
C    Syntax: CONFIDENCE index_specification ...
C
C  Input arguments:
C    MAXW       Number of available storage positions in array H.
C
C  Output arguments:
C    MERR       Error indicator.
C
C  Input / Output arguments:
C    H          The 2-dimensional Merlin work array.
C
C  Notes:
C    This command uses the covariance matrix as calculated by the
C    COVARIANCE command. Fixed variables are ignored in the
C    Hessian and subsequently in the covariance matrix calculations.
C    Hence you should invoke CONFIDENCE with the same fix statuses as
C    for the COVARIANCE command.
C
C  ---------------------------------------------------------------------
      IMPLICIT $(TYPE) (A-H,O-Z)
C  Arguments:
      DIMENSION H(NOD,NOD)
C  ---------------------------------------------------------------------
C  Common blocks:
      PARAMETER ( MXV = $(MXV) )
      COMMON /IWORK/ INDE(MXV)
      COMMON /DIMEN/ NOD, NOF
      COMMON /ACCURA/ COMACC
      COMMON /TOTCAL/ ITNOCS, IPARTC, ITDER, IPDER, ITHESS, IPHESS,
     &                ITJAC, IPJAC
      COMMON /MODE3/ MODEV, IFFORM, INMCL
      COMMON /RWORK/ XT1(MXV), XT2(MXV), XT3(MXV), XT4(MXV), XT5(MXV),
     &               XT6(MXV), XT7(MXV+1)
      LOGICAL STRO, WEAK
      COMMON /PRILOG/ STRO, WEAK
      COMMON /UNITS/ IUINP, IUOUT
C  ---------------------------------------------------------------------
C  Local variables:
      PARAMETER ( MXFILE = 80 )
C  MXDI controls the number of significant digits that are displayed
C  for the elipsoid-related printouts.
      PARAMETER ( MXDI = 6 )
      CHARACTER*(MXFILE) FILE
      CHARACTER*79 LINE, STR
      CHARACTER*10 STR2, FORM
      LOGICAL WROW, SORTED
      DATA FORM / '(1PG__.__)' /
C  ---------------------------------------------------------------------
C
C  Get indices for the ellipsoid.
      CALL CHI(INDE,NSEL,1,MERR)
      IF (MERR.NE.0) RETURN
      IF (NSEL.EQ.0) RETURN
C
C  Did the user supply enough storage ?
      NEED = NOD**2
      CALL VSIM('N**2',MAXW,NEED,MERR)
      IF (MERR.NE.0) RETURN
      IF (ITNOCS.EQ.0 .OR. MODEV.EQ.0) RETURN
C
C  Read in covariance matrix.
      CALL GETPPS('COVARIANCE','FILE',FILE,LEF)
      CALL OPENGF(FILE,1,NU)
      IF (NU.EQ.-1) THEN
         MERR = -2
         RETURN
      END IF
C
C  Determine free (non-fixed) variables.
      CALL FREEVA(NFREE)
      NNUR = 0
      DO 200,J=1,NFREE
         DO 230,I=1,NFREE
            READ (NU,*,END=220) CVR
            NNUR = NNUR+1
            H(J,I) = CVR
230      CONTINUE
200   CONTINUE
      CLOSE (NU)
      CALL I2STR(NFREE,STR,LES)
      IF (STRO) WRITE (IUOUT,250) STR(1:LES), STR(1:LES),
     &                            FILE(1:LENGTH(FILE))
C
C  Print standard error.
      CALL GETPPR('COVARIANCE','PROB',PROB)
      CALL PRSTDE(H,NOD,PROB)
C
C  Sort the command line indices.
      DO 15,K=1,NSEL-1
         SORTED = .TRUE.
         DO 35,I=NSEL,K+1,-1
            IF (INDE(I).LT.INDE(I-1)) THEN
               SORTED = .FALSE.
               IM1 = I-1
               ITMP = INDE(I)
               INDE(I) = INDE(IM1)
               INDE(IM1) = ITMP
            END IF
35       CONTINUE
         IF (SORTED) GOTO 16
15    CONTINUE
16    CONTINUE
C
C  Selcect rows/columns form the covariance matrix.
      CALL CVSEL(H,NOD,INDE,NSEL,NFREE)
      IF (NSEL.EQ.0) THEN
         IF (WEAK) WRITE (IUOUT,300)
         RETURN
      END IF
C
C  Decompose.
      CALL HPDD(H,XT1,NOD,NSEL,COMACC,ISPD)
C
C  Invert.
      DO 737,I=1,NSEL
         DO 721,J=1,NSEL
721         XT1(J) = 0.
         XT1(I) = 1.
         CALL LLSOLV(H,XT1,XT1,NSEL,NOD)
         XT2(I) = XT1(I)
         DO 722,J=I+1,NSEL
722         H(J,I) = XT1(J)
737   CONTINUE
C
      DO 731,I=1,NSEL
         H(I,I) = XT2(I)
         DO 732,J=I+1,NSEL
732         H(I,J) = H(J,I)
731   CONTINUE
C
C  Final output.
      IF (.NOT.STRO) RETURN
C
      NP = NSEL
      ACC = SQRT(COMACC)
      CALL FDELP(NP,PROB,DCHI,ACC)
      CALL I2STR(NSEL,STR2,LES2)
      WRITE (FORM(5:6),'(I2.2)') MXDI+7
      WRITE (FORM(8:9),'(I2.2)') MXDI
      WRITE (STR,FORM) DCHI
      LES = MXDI+7
C
      CALL DASH
      WRITE (IUOUT,30)
      WRITE (IUOUT,40) STR(1:LES), STR2(1:LES2), STR2(1:LES2)
      WRITE (IUOUT,*)
      LL = LEN(LINE)
      WROW = INT((LL-2)/(MXDI+7+2)).LT.NSEL
      DO 44,I=1,NSEL
         IF (WROW) THEN
            CALL I2STR(I,STR,LES)
            WRITE (IUOUT,50) STR(1:LES)
         END IF
         NL = 0
         DO 45,J=1,NSEL
            WRITE (STR,FORM) H(I,J)
            LES = MXDI+7
            IEND = NL+LES
            IF (NL.NE.0) IEND = IEND+2
            IF (IEND.GT.LL) THEN
               WRITE (IUOUT,'($(FC)A)') LINE(1:NL)
               LINE(1:LES) = STR(1:LES)
               NL = LES
            ELSE
               IF (NL.EQ.0) THEN
                  LINE(NL+1:NL+LES) = STR(1:LES)
                  NL = NL+LES
               ELSE
                  LINE(NL+1:NL+LES+2) = '  '//STR(1:LES)
                  NL = NL+LES+2
               END IF
            END IF
45       CONTINUE
         WRITE (IUOUT,'($(FC)A)') LINE(1:NL)
44    CONTINUE
      RETURN
C
C  EOF on the covariance file.
220   CONTINUE
      IF (WEAK) THEN
         CALL I2STR(NNUR,STR,LES)
         WRITE (IUOUT,210) FILE(1:LENGTH(FILE)), STR(1:LES)
      END IF
      CLOSE (NU)
      MERR = -2
C
30    FORMAT ($(FC)'The confidence region is defined by the ',
     &        'equation: Delta = (x-x*)^T C (x-x*)')
40    FORMAT ($(FC)'Where: Delta = ',A,
     &       /$(FC)'          x* = The current minimum'
     &       /$(FC)'    (x-x*)^T = Traspose of (x-x*)'
     &       /$(FC)'and C the following ',A,' x ',A,' matrix:')
50    FORMAT ($(FC)'Row ',A)
210   FORMAT (/$(FC)'Error: File "',A,'" contains only ',A,
     &        ' real numbers.')
250   FORMAT (/$(FC)'The ',A,' x ',A,' covariance marix has been ',
     &        'read in from file "',A,'".')
300   FORMAT (/$(FC)'Warning: All variables are fixed.')
      END
C  ---------------------------------------------------------------------
      SUBROUTINE CONTRL ( MERR )
C  ---------------------------------------------------------------------
C
C  Description:
C    Implements the CONTROL command.
C    Syntax: CONTROL (Has panel)
C
C  Output arguments:
C    MERR       Error indicator.
C
C  ---------------------------------------------------------------------
C  Common blocks:
      COMMON /CNTRL/ IDISK, MACECH, LIMCLK, IECOMA, IPROEC, MCLUBU
C  ---------------------------------------------------------------------
C
C  Present the panel and process any user changes.
      CALL CHANGE(MERR)
      IF (MERR.NE.0) RETURN
C
C  Get panel parameters.
      CALL GETPI('DISKMAC',IDISK)
      CALL GETPI('ECHOMAC',MACECH)
      CALL GETPI('ECHOPRO',IPROEC)
      CALL GETPI('MCLKEYS',LIMCLK)
      CALL GETPI('ECHOCOM',IECOMA)
      CALL GETPI('MCLBUF',MCLUBU)
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE COVAR ( H, MAXW, MERR )
C  ---------------------------------------------------------------------
C
C  Description:
C    Implements the COVARIANCE command.
C    Syntax: COVARIANCE (Has panel)
C
C  Input arguments:
C    MAXW       Number of available storage positions in array H.
C
C  Output arguments:
C    MERR       Error indicator.
C
C  Input / Output arguments:
C    H          The 2-dimensional Merlin work array.
C
C  ---------------------------------------------------------------------
      IMPLICIT $(TYPE) (A-H,O-Z)
C  Arguments:
      DIMENSION H(NOD,NOD)
C  ---------------------------------------------------------------------
C  Common blocks:
      PARAMETER ( MXV = $(MXV) )
      COMMON /PARAMS/ POINT(MXV), VAL
      COMMON /DIMEN/ NOD, NOF
      COMMON /ACCURA/ COMACC
      COMMON /TOTCAL/ ITNOCS, IPARTC, ITDER, IPDER, ITHESS, IPHESS,
     &                ITJAC, IPJAC
      COMMON /MODE3/ MODEV, IFFORM, INMCL
      COMMON /RWORK/ XT1(MXV), XT2(MXV), XT3(MXV), XT4(MXV), XT5(MXV),
     &               XT6(MXV), XT7(MXV+1)
      LOGICAL STRO, WEAK
      COMMON /PRILOG/ STRO, WEAK
      COMMON /UNITS/ IUINP, IUOUT
C  ---------------------------------------------------------------------
C  Local variables:
      PARAMETER ( MXFILE = 80 )
      CHARACTER*(MXFILE) FILE
      CHARACTER COV
      CHARACTER*30 CSTR
      CHARACTER*10 STR
C  ---------------------------------------------------------------------
C
C  Did the user supply enough storage ?
      NEED = NOD**2
      CALL VSIM('N**2',MAXW,NEED,MERR)
      IF (MERR.NE.0) RETURN
C
C  Present the panel.
      CALL SETPI('CANCEL',1,IERR)
      CALL CHANGE(MERR)
      IF (MERR.NE.0) RETURN
      CALL GETPI('CANCEL',ICANC)
      IF (ICANC.NE.1 .OR. ITNOCS.EQ.0 .OR. MODEV.EQ.0) RETURN
C
      CALL GETPS('DO',COV,LEC)
      IF (ISCOMP(COV,'C').NE.0 .OR. ISCOMP(COV,'W').NE.0) THEN
C
C  Calculate the Hessian matrix. Use function values.
         CALL SECDER(H,NOD,NOD,POINT,VAL,'F',XT1,XT2,NC)
C
C  Remove fixed variables.
         CALL SHRINK(H,NOD,NFREE)
         IF (NFREE.EQ.0) THEN
            IF (WEAK) WRITE (IUOUT,110)
            RETURN
         END IF
C
C  Choleski decomposition of the Hessian matrix.
         CALL HPDD(H,XT1,NOD,NFREE,COMACC,ISPD)
C
C  H now contains the Choleski factor L[t] in its upper triangular part.
C  Get the inverse, column by column. Save it in the lower part of H.
C  Diagonal elements of the inverse in XT2.
         DO 30,I=1,NFREE
            DO 21,J=1,NFREE
21             XT1(J) = 0.
            XT1(I) = 1.
            CALL LLSOLV(H,XT1,XT1,NFREE,NOD)
            XT2(I) = XT1(I)
            DO 22,J=I+1,NFREE
22             H(J,I) = XT1(J)
30       CONTINUE
C
C  Restore the full, symmetric inverse.
         DO 31,I=1,NFREE
            H(I,I) = XT2(I)
            DO 32,J=I+1,NFREE
32             H(I,J) = H(J,I)
31       CONTINUE
C
C  Convert into a covariance matrix.
         DO 42,I=1,NFREE
            DO 43,J=1,NFREE
43             H(I,J) = 2*H(I,J)
42       CONTINUE
C
C  Covariance matrix is ready.
         IF (ISCOMP(COV,'W').NE.0) THEN
            CALL GETPS('FILE',FILE,LEF)
            CALL OPENGF(FILE,0,NU)
            IF (NU.EQ.-1) THEN
               MERR = -2
               RETURN
            END IF
            DO 100,J=1,NFREE
               DO 101,I=1,NFREE
                  CVR = H(I,J)
                  CALL R2RTF(CVR,CSTR,LEC)
                  WRITE (NU,'(A)') CSTR(1:LEC)
101            CONTINUE
100         CONTINUE
            CLOSE (NU)
            CALL I2STR(NFREE,STR,LES)
            IF (STRO) WRITE (IUOUT,230) STR(1:LES),
     &                     STR(1:LES), FILE(1:LENGTH(FILE))
         ELSE
            CALL I2STR(NFREE,STR,LES)
            IF (STRO) WRITE (IUOUT,240) STR(1:LES), STR(1:LES)
         END IF
      END IF
C
      IF (ISCOMP(COV,'R').NE.0) THEN
         CALL GETPS('FILE',FILE,LEF)
         CALL OPENGF(FILE,1,NU)
         IF (NU.EQ.-1) THEN
            MERR = -2
            RETURN
         END IF
         CALL FREEVA(NFREE)
         NNUR = 0
         DO 200,J=1,NFREE
            DO 205,I=1,NFREE
               READ (NU,*,END=220) CVR
               NNUR = NNUR+1
               H(J,I) = CVR
205         CONTINUE
200      CONTINUE
         CLOSE (NU)
         CALL I2STR(NFREE,STR,LES)
         IF (STRO) WRITE (IUOUT,250) STR(1:LES), STR(1:LES),
     &                FILE(1:LENGTH(FILE))
      END IF
C
C  Give out standard errors for the specified probability.
      CALL GETPR('PROB',PROB)
      CALL PRSTDE(H,NOD,PROB)
C
C  Print quality of fit.
      IF (NOF-NFREE.GT.0) THEN
         TWO = 2.
         DD = (NOF-NFREE)/TWO
         VV = VAL/2.
         QQ = GAMINC(DD,VV)
         WRITE (IUOUT,58) QQ
      END IF
      RETURN
C
C  EOF while reading the covariance matrix.
220   CONTINUE
      IF (WEAK) THEN
         CALL I2STR(NNUR,STR,LES)
         WRITE (IUOUT,210) FILE(1:LENGTH(FILE)), STR(1:LES)
      END IF
      MERR = -2
C
58    FORMAT (/$(FC)'Quality of fit: ',F6.4)
110   FORMAT (/$(FC)'Warning: All variables are fixed.')
210   FORMAT (/$(FC)'Error: File "',A,'" contains only ',A,
     &        ' real numbers.')
230   FORMAT (/$(FC)'The ',A,' x ',A,' covariance marix has been ',
     &        'calculated ',
     &        /$(FC)'and written to file "',A,'".')
240   FORMAT (/$(FC)'The ',A,' x ',A,' covariance marix has been ',
     &        'calculated.')
250   FORMAT (/$(FC)'The ',A,' x ',A,' covariance marix has been ',
     &        'read in from file "',A,'".')
      END
C  ---------------------------------------------------------------------
      SUBROUTINE PRSTDE ( H, LD, PROB )
C  ---------------------------------------------------------------------
C
C  Description:
C    Prints the standard errors and confidence intervals ala SHORTDIS.
C
C  Input arguments:
C    H          The covariance matrix.
C    LD         Leading dimension of H.
C    PROB       Confidence probability.
C
C  ---------------------------------------------------------------------
      IMPLICIT $(TYPE) (A-H,O-Z)
C  Arguments:
      DIMENSION H(LD,*)
C  ---------------------------------------------------------------------
C  Common blocks:
      PARAMETER ( MXV = $(MXV) )
      COMMON /PARAMS/ POINT(MXV), VAL
      COMMON /DETAIL/ IXAT(MXV), MARG(MXV)
      CHARACTER*10 NICKNA
      COMMON /NAMES/ NICKNA(MXV)
      COMMON /DIMEN/ NOD, NOF
      COMMON /ACCURA/ COMACC
      LOGICAL STRO, WEAK
      COMMON /PRILOG/ STRO, WEAK
      COMMON /UNITS/ IUINP, IUOUT
C  ---------------------------------------------------------------------
C  Local variables:
      CHARACTER*7 CFIX
      CHARACTER*30 CSTR
C  ---------------------------------------------------------------------
C
      IF (.NOT.STRO) RETURN
C
      NP = 1
      ACC = SQRT(COMACC)
C
C  Given the specified PROB, solve for delta-Chi.
      CALL FDELP(NP,PROB,DCHI,ACC)
      WRITE (IUOUT,57) PROB
      CALL DASH
      WRITE (IUOUT,56)
      CALL DASH
      NF = 0
      SQD = SQRT(DCHI)
      DO 50,I=1,NOD
         CALL FIXSTR(I,CFIX)
         CALL R2RTF(POINT(I),CSTR,LEP)
         IF (IXAT(I).EQ.0) THEN
            WRITE (IUOUT,55) I, NICKNA(I), CFIX, CSTR(1:LEP)
         ELSE
            NF = NF+1
            SHII = SQRT(H(NF,NF))
            XIVAL = SQD*SHII
            WRITE (IUOUT,55) I, NICKNA(I), CFIX, CSTR(1:LEP),
     &                             SHII, XIVAL
         END IF
50    CONTINUE
      CALL WVSDAT(IUOUT,VAL)
C
55    FORMAT ($(FC)I5,') ',A10,1X,A,1X,A:,1PG15.7,1X,1PG15.7)
56    FORMAT ($(FC)'Index',2X,'Name',7X,'Fix',7X,'Parameter value',
     &        8X,'Std error',6X,'Conf. interv.')
57    FORMAT (/$(FC)'The following confidence intervals are computed ',
     &        'for probability = ',F6.4)
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE FDELP ( NP, PROB, DCHI, ACC )
C  ---------------------------------------------------------------------
C
C  Description:
C    Simple root finder. It is used to determine the change DCHI in a
C    sum-of-squares function, such that the probability of a
C    chi-square variable with NP degrees of freedom being less than
C    DCHI is PROB. This problem reduces to finding a root of the
C    equation:
C      GAMINC(NP/2,X/2)+PROB-1 = 0
C    where GAMINC is the incomplete Gamma function.
C    Initialy it brackets the minimum, and then uses bisection to
C    isolate the root, to within fractional accuracy ACC.
C
C  Input arguments:
C    NP         Parameter NP of the equation to be solved.
C    PROB       Parameter PROB of the equation to be solved.
C    ACC        Fractional accuracy of the root.
C
C  Output arguments:
C    DCHI       Root of the equation.
C
C  ---------------------------------------------------------------------
      IMPLICIT $(TYPE) (A-H,O-Z)
C  Local variables:
      PARAMETER ( ENH = 1.5, NBRACK = 50, NBISE = 50 )
      FUN(X) = GAMINC(XNPBY2,X/TWO)+PROB-ONE
C  ---------------------------------------------------------------------
C
C  Some constants.
      ONE = 1.
      TWO = 2.
      XNPBY2 = NP/TWO
C
C  Guess an initial bracketing interval. We know that the root is
C  always > 0.
      X1 = 0.
      X2 = 20.
      F1 = FUN(X1)
      F2 = FUN(X2)
C
C  Bracket the root.
      DO 10,J=1,NBRACK
         IF (F1*F2.LT.0.) GOTO 20
         IF (ABS(F1).LT.ABS(F2)) THEN
            X1 = X1+ENH*(X1-X2)
            F1 = FUN(X1)
         ELSE
            X2 = X2+ENH*(X2-X1)
            F2 = FUN(X2)
         END IF
10    CONTINUE
      CALL MABORT('Cannot determine bracket.','FDELP')
C
C  Use bisection.
20    CONTINUE
      F = F1
      FM = F2
      IF (F.LT.0.) THEN
         X = X1
         DX = X2-X1
      ELSE
         X = X2
         DX = X1-X2
      END IF
      DO 30,J=1,NBISE
         DX = DX/2.
         XM = X+DX
         FM = FUN(XM)
         IF (FM.LE.0.) X = XM
         IF (ABS(DX)/MAX(ONE,ABS(X)).LT.ACC .OR. FM.EQ.0.) THEN
            DCHI = X
            RETURN
         END IF
30    CONTINUE
      CALL MABORT('Bisection failed.','FDELP')
C
      END
C  ---------------------------------------------------------------------
      FUNCTION GAMINC ( A, X )
C  ---------------------------------------------------------------------
C
C  Description:
C    Calculates the incomplete Gamma function Q(a,x) using either the
C    series representation, or the continued fraction representation.
C
C  Input arguments:
C    A          Argument A of the incomplete Gamma function.
C    X          Argument X of the incomplete Gamma function.
C
C  ---------------------------------------------------------------------
      IMPLICIT $(TYPE) (A-H,O-Z)
C  Local variables:
      PARAMETER ( ITMAX=100, EPS=3.E-7, FPMIN=1.E-30 )
C  ---------------------------------------------------------------------
C
C  Incorrect arguments ?
      IF (X.LT.0. .OR. A.LE.0.)
     &    CALL MABORT('Incorrect arguments.','GAMINC')
C
C  Some constants.
      ONE = 1.
C
      IF (X.LT.A+1.) THEN
C  Use the series representation.
         IF (X.EQ.0.) THEN
            G = 0.
            GOTO 100
         END IF
         AP = A
         SUM = ONE/A
         DEL = SUM
         DO 10,N=1,ITMAX
            AP = AP+1.
            DEL = DEL*X/AP
            SUM = SUM+DEL
            IF (ABS(DEL).LT.ABS(SUM)*EPS) GOTO 20
10       CONTINUE
         CALL MABORT('More iterations are in need.','GAMINC')
C
20       CONTINUE
         GLN = GAMMLN(A)
         G = SUM*EXP(-X+A*LOG(X)-GLN)
100      CONTINUE
         GAMINC = ONE-G
      ELSE
C  Use the continued fraction represenation.
         B = X+1.-A
         C = ONE/FPMIN
         D = ONE/B
         H = D
         DO 30,I=1,ITMAX
            AN = -I*(I-A)
            B = B+2.
            D = AN*D+B
            IF (ABS(D).LT.FPMIN) D = FPMIN
            C = B+AN/C
            IF (ABS(C).LT.FPMIN) C = FPMIN
            D = ONE/D
            DEL = D*C
            H = H*DEL
            IF (ABS(DEL-1.).LT.EPS) GOTO 40
30       CONTINUE
         CALL MABORT('More iterations are in need.','GAMINC')
40       CONTINUE
         GLN = GAMMLN(A)
         GAMINC = EXP(-X+A*LOG(X)-GLN)*H
      ENDIF
C
      END
C  ---------------------------------------------------------------------
      FUNCTION GAMMLN ( XARG )
C  ---------------------------------------------------------------------
C
C  Description:
C    Returns the logarithm of the Gamma function.
C
C  Input arguments:
C    XARG       Number of variables of the objective function.
C
C  ---------------------------------------------------------------------
      IMPLICIT $(TYPE) (A-H,O-Z)
C  Local variables:
      DOUBLE PRECISION SER, STP, TMP, X, Y, COF
      DIMENSION COF(6)
      DATA COF / 76.18009172947146D0, -86.50532032941677D0,
     &           24.01409824083091D0, -1.231739572450155D0,
     &          .1208650973866179D-2, -0.5395239384953D-5 /
      DATA STP / 2.5066282746310005D0 /
C  ---------------------------------------------------------------------
C
      X = XARG
      Y = X
      TMP = X+5.5D0
      TMP = (X+0.5D0)*LOG(TMP)-TMP
      SER = 1.000000000190015D0
      DO 10,J=1,6
         Y = Y+1.D0
         SER = SER+COF(J)/Y
10    CONTINUE
      GAMMLN = TMP+LOG(STP*SER/X)
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE DELETE ( MERR )
C  ---------------------------------------------------------------------
C
C  Description:
C    Implements the DELETE command.
C    Syntax: DELETE file ...
C
C  Output arguments:
C    MERR       Error indicator.
C
C  ---------------------------------------------------------------------
C  Common blocks:
      LOGICAL STRO, WEAK
      COMMON /PRILOG/ STRO, WEAK
      COMMON /UNITS/ IUINP, IUOUT
C  ---------------------------------------------------------------------
C  Local variables:
      PARAMETER ( MXFILE = 80 )
      CHARACTER*(MXFILE) FILE
C  ---------------------------------------------------------------------
C
      MERR = 0
C
C  Did the user enter any file names to delete ?
      CALL ARGNO(NA)
      IF (NA.EQ.0) THEN
C
C  No arguments in the command line. Prompt the user.
         IF (STRO) WRITE (IUOUT,10)
         CALL INNOCR(FILE,.FALSE.,MERR)
         IF (MERR.NE.0) RETURN
         CALL DEL1F(FILE,MERR)
      ELSE
C
C  Loop over all files specified in the command line.
         DO 100,I=1,NA
C  Get the name of the file to delete.
            CALL CMDARG(I,FILE,LENF)
            CALL DEL1F(FILE,MERR)
100      CONTINUE
      END IF
C
10    FORMAT (/$(FC)'Enter file name to delete:')
      END
C  ---------------------------------------------------------------------
      SUBROUTINE DEL1F ( FILE, MERR )
C  ---------------------------------------------------------------------
C
C  Description:
C    Part of the DELETE command. Deletes one file only.
C    The file is opened and then closed with STATUS='DELETE'.
C    This works on all machines we know of.
C
C  Input arguments:
C    FILE       The file to be deleted.
C
C  Input / Output arguments:
C    MERR       Error indicator. Its value is changed only if an error
C               occurs while in this routine.
C
C  ---------------------------------------------------------------------
C  Arguments:
      CHARACTER*(*) FILE
C  ---------------------------------------------------------------------
C  Common blocks:
      LOGICAL STRO, WEAK
      COMMON /PRILOG/ STRO, WEAK
      COMMON /UNITS/ IUINP, IUOUT
C  ---------------------------------------------------------------------
C  Local variables:
      LOGICAL OPE, EXI
C  ---------------------------------------------------------------------
C
C  A null file name is not allowed.
      LE1 = LENGTH(FILE)
      IF (LE1.EQ.0) THEN
         CALL ERRZFI
         MERR = -2
         RETURN
      END IF
C
C  Get info for the file.
      INQUIRE (FILE=FILE(1:LE1),EXIST=EXI,OPENED=OPE,NUMBER=NUN,
     &         IOSTAT=IERR)
C
C  Any errors during INQUIRE ?
      IF (IERR.NE.0) THEN
         CALL ERRINQ(FILE,IERR)
         MERR = -2
         RETURN
      END IF
C
C  Does the file exist ? (Attempting to delete a non existent file is
C  not an error)
      IF (.NOT.EXI) THEN
         CALL WRNEXI(FILE)
         RETURN
      END IF
C
C  Is the file open ? Cannot delete open files.
      IF (OPE) THEN
         CALL ERRBUS(FILE,NUN)
         MERR = -2
         RETURN
      END IF
C
C  Allocate unit number for the file.
      NU = NUNIT()
      IF (NU.EQ.-1) THEN
         CALL ERRUAL(FILE)
         MERR = -2
         RETURN
      END IF
C
C  Open the file.
      OPEN (NU,FILE=FILE(1:LE1),IOSTAT=IOE)
      IF (IOE.NE.0) THEN
         CALL ERROPE(FILE,IOE)
         MERR = -2
         RETURN
      END IF
C
C  Close/delete the file.
      CLOSE (NU,STATUS='DELETE')
C
C  Make sure the file has been deleted.
      INQUIRE (FILE=FILE(1:LE1),EXIST=EXI,IOSTAT=IERR)
C
C  Any errors during INQUIRE ?
      IF (IERR.NE.0) THEN
         CALL ERRINQ(FILE,IERR)
         MERR = -2
         RETURN
      END IF
C
C  Is the file still around ?
      IF (EXI) THEN
         IF (WEAK) WRITE (IUOUT,70) FILE(1:LE1)
      ELSE
         IF (STRO) WRITE (IUOUT,60) FILE(1:LE1)
      END IF
C
60    FORMAT (/$(FC)'File "',A,'" has been deleted.')
70    FORMAT (/$(FC)'Error: File "',A,'" could not be deleted.')
      END
C  ---------------------------------------------------------------------
      SUBROUTINE DEMAR ( IOPT, MERR )
C  ---------------------------------------------------------------------
C
C  Description:
C    Implements the LDEMARGIN and RDEMARGIN commands.
C    Syntax: LDEMARGIN index_specification ...
C            RDEMARGIN index_specification ...
C
C  Input arguments:
C    IOPT       Specifies the operation to be performed:
C               IOPT = -1 -> LDEMARGIN
C               IOPT =  1 -> RDEMARGIN
C
C  Output arguments:
C    MERR       Error indicator.
C
C  ---------------------------------------------------------------------
      IMPLICIT $(TYPE) (A-H,O-Z)
C  Common blocks:
      PARAMETER ( MXV = $(MXV) )
      COMMON /XLIMS/ XLL(MXV), XRL(MXV)
      COMMON /DETAIL/ IXAT(MXV), MARG(MXV)
      COMMON /IWORK/ INDE(MXV)
      COMMON /BIGCOM/ BIGNUM, BLIM, DWARF, DWALOG
      LOGICAL STRO, WEAK
      COMMON /PRILOG/ STRO, WEAK
      COMMON /UNITS/ IUINP, IUOUT
C  ---------------------------------------------------------------------
C
      CALL ARGNO(NA)
C
C  Prompt the user if no arguments were entered.
      IF (NA.EQ.0 .AND. STRO) THEN
         IF (IOPT.EQ.-1) THEN
            WRITE (IUOUT,20)
         ELSE
            WRITE (IUOUT,30)
         END IF
      END IF
C
C  Read an one line input, or parse command line arguments for
C  index/value specifications.
      CALL CHI(INDE,NIT,1,MERR)
      IF (MERR.NE.0) RETURN
C
C  Add a backup record.
      CALL BACKUP('D')
C
      DO 10,I=1,NIT
         IND = INDE(I)
         IF (IOPT.EQ.-1) THEN
            IF (MARG(IND).EQ.-1) MARG(IND) = 0
            IF (MARG(IND).EQ.2) MARG(IND) = 1
         ELSE
            IF (MARG(IND).EQ.1) MARG(IND) = 0
            IF (MARG(IND).EQ.2) MARG(IND) = -1
         END IF
C  Reset the values for the cleared margins to default.
         IF (MARG(IND).EQ.0) THEN
            XLL(IND) = -BLIM
            XRL(IND) = BLIM
         ELSE IF (MARG(IND).EQ.-1) THEN
            XRL(IND) = BLIM
         ELSE IF (MARG(I).EQ.1) THEN
            XLL(IND) = -BLIM
         END IF
10    CONTINUE
C
20    FORMAT (/$(FC)'Enter left margins sequence:')
30    FORMAT (/$(FC)'Enter right margins sequence:')
      END
C  ---------------------------------------------------------------------
      SUBROUTINE DISCA ( MERR )
C  ---------------------------------------------------------------------
C
C  Description:
C    Implements the DISCARD command.
C    Syntax: DISCARD file [ type ]
C            type can be one of TEXT or BIN.
C    Open files such as the backup and memo files cannot be discarded.
C
C  Output arguments:
C    MERR       Error indicator.
C
C  ---------------------------------------------------------------------
      IMPLICIT $(TYPE) (A-H,O-Z)
C  Common blocks:
      PARAMETER ( MXV = $(MXV) )
      COMMON /PARAMS/ POINT(MXV), VAL
      COMMON /DIMEN/ NOD, NOF
      LOGICAL STRO, WEAK
      COMMON /PRILOG/ STRO, WEAK
      COMMON /UNITS/ IUINP, IUOUT
C  ---------------------------------------------------------------------
C  Local variables:
      PARAMETER ( MXFILE = 80 )
      CHARACTER*(MXFILE) FILE
      CHARACTER*80 TYPE
      CHARACTER*10 STR
      LOGICAL EXI, OPE
C  ---------------------------------------------------------------------
C
      MERR = 0
      CALL ARGNO(NA)
      IF (NA.EQ.0) THEN
C  No command line arguments. Ask for the file.
         IF (STRO) WRITE (IUOUT,10)
         CALL INNOCR(FILE,.FALSE.,MERR)
         IF (MERR.NE.0) RETURN
         TYPE = 'TEXT'
      ELSE IF (NA.EQ.1) THEN
C  Only the file name was entered in the command line.
C  Default type is TEXT.
         CALL CMDARG(1,FILE,LENF)
         TYPE = 'TEXT'
      ELSE IF (NA.EQ.2) THEN
C  FIle name and file type were entered in the command line.
         CALL CMDARG(1,FILE,LENF)
         CALL CMDARG(2,TYPE,LENC)
      ELSE
C  Too many arguments.
         CALL ERRNAR
         MERR = -2
         RETURN
      END IF
C
      LENF = LENGTH(FILE)
      IF (LENF.EQ.0) THEN
         IF (WEAK) WRITE (IUOUT,40)
         MERR = -2
         RETURN
      END IF
C
C  Check file type.
      IF (ISCOMP(TYPE,'TEXT').NE.0) THEN
         MODE = 1
      ELSE IF (ISCOMP(TYPE,'BIN').NE.0) THEN
         MODE = 2
      ELSE
         CALL ERRARG(TYPE)
         MERR = -2
         RETURN
      END IF
C
C  Get file info.
      INQUIRE (FILE=FILE,EXIST=EXI,OPENED=OPE,NUMBER=NUM,IOSTAT=IOE)
      IF (IOE.NE.0) THEN
         CALL ERRINQ(FILE,IOE)
         MERR = -2
         RETURN
      END IF
C
C  If the file exists, it must be deleted first.
      IF (EXI) THEN
C
C  Cannot discard an open file.
         IF (OPE) THEN
            CALL ERRBUS(FILE,NUM)
            MERR = -2
            RETURN
         END IF
C
C  Allocate unit for the file.
         NU = NUNIT()
         IF (NU.EQ.-1) THEN
            CALL ERRUAL(FILE)
            MERR = -2
            RETURN
         END IF
C
C  Open/close the file in order to do a safe delete.
         OPEN (NU,FILE=FILE,IOSTAT=IOE)
         IF (IOE.NE.0) THEN
C  The file refuses to open properly.
            CALL ERROPE(FILE,IOE)
            MERR = -2
            RETURN
         END IF
         CLOSE (NU,STATUS='DELETE',IOSTAT=IOE)
         IF (IOE.NE.0) THEN
C  The file refuses to be deleted. The user should check with the
C  machines's OS.
            CALL I2STR(IOE,STR,LES)
            IF (WEAK) WRITE (IUOUT,30) STR(1:LES)
C  Try to just close the file.
            CLOSE (NU,IOSTAT=IOE)
            IF (IOE.NE.0) CALL ERRCLO(FILE,IOE)
            MERR = -2
            RETURN
         END IF
      END IF
C
C  Open the file according to the specified file-type.
      CALL OPENIT(FILE,TYPE,0,1,NU,NREC)
      IF (NU.EQ.-1) THEN
         MERR = -2
         RETURN
      END IF
C
C  Write one Merlin record.
      NREC = NREC+1
      CALL WRITER(POINT,VAL,NU,MODE,NREC)
      IF (MODE.EQ.2) WRITE (NU,REC=1) NOD, NREC
      CLOSE (NU)
C
C  Informative printout.
      IF (STRO) WRITE (IUOUT,20) FILE(1:LENF)
C
10    FORMAT (/$(FC)'Enter file name to discard:')
20    FORMAT (/$(FC)'The previous contents (if any) of file "',A,
     &        '" have been discarded.')
30    FORMAT (/$(FC)'Cannot delete file "',A,'". IOSTAT = ',A)
40    FORMAT (/$(FC)'You must supply a file name.')
      END
C  ---------------------------------------------------------------------
      SUBROUTINE DUMP ( MERR)
C  ---------------------------------------------------------------------
C
C  Description:
C    Implements the DUMP command.
C    Syntax: DUMP (Has panel)
C
C  Output arguments:
C    MERR       Error indicator.
C
C  ---------------------------------------------------------------------
      IMPLICIT $(TYPE) (A-H,O-Z)
C  Common blocks:
      PARAMETER ( MXV = $(MXV) )
      COMMON /PARAMS/ POINT(MXV), VAL
      COMMON /DETAIL/ IXAT(MXV), MARG(MXV)
      COMMON /XLIMS/ XLL(MXV), XRL(MXV)
      CHARACTER*10 NICKNA
      COMMON /NAMES/ NICKNA(MXV)
      COMMON /ROSTEP/ STEP(MXV)
      COMMON /DIMEN/ NOD, NOF
      LOGICAL STRO, WEAK
      COMMON /PRILOG/ STRO, WEAK
      COMMON /UNITS/ IUINP, IUOUT
      COMMON /BIGCOM/ BIGNUM, BLIM, DWARF, DWALOG
      COMMON /RWORK/ XT1(MXV), XT2(MXV), XT3(MXV), XT4(MXV), XT5(MXV),
     &               XT6(MXV), XT7(MXV+1)
C  ---------------------------------------------------------------------
C  Local variables:
      PARAMETER ( MXFILE = 80 )
      CHARACTER*(MXFILE) FILE
      CHARACTER*80 FORM, USEF
      CHARACTER WHAT
      CHARACTER*25 MSG
      CHARACTER*10 STR, CVAL, CAPP
      LOGICAL EXI, OPE, FREE
C  ---------------------------------------------------------------------
C
      SAVE IFIRST
      DATA IFIRST / 0 /
C
C  On the first entry set the TO keyword to the number of variables.
      IF (IFIRST.EQ.0) THEN
         CALL SETPI('TO',NOD,IERR)
         IFIRST = 1
      END IF
C
C  Present the panel and process any user changes.
      CALL SETPI('CANCEL',1,MERR)
      CALL CHANGE(MERR)
      IF (MERR.NE.0) RETURN
      CALL GETPI('CANCEL',ICANC)
      IF (ICANC.EQ.0) RETURN
C
C  Get panel parameters.
      CALL GETPS('FILE',FILE,LEF)
      CALL GETPS('WHAT',WHAT,LEW)
      CALL GETPI('NPL',NPL)
      CALL GETPS('APPEND',CAPP,LECA)
      CALL GETPI('FROM',IFROM)
      CALL GETPI('TO',ITO)
      CALL GETPS('VALUE',CVAL,LEV)
      CALL GETPS('FORMAT',USEF,LER)
C
C  We need a file name.
      LEF = LENGTH(FILE)
      IF (LEF.EQ.0) THEN
         IF (WEAK) WRITE (IUOUT,200)
         MERR = -2
         RETURN
      END IF
C
C  FROM should be <= NOD.
      IF (IFROM.GT.NOD) THEN
         CALL I2STR(NOD,STR,LES)
         IF (WEAK) WRITE (IUOUT,240) STR(1:LES)
         MERR = -2
         RETURN
      END IF
C
C  TO should be <= NOD.
      IF (ITO.GT.NOD) THEN
         CALL I2STR(NOD,STR,LES)
         IF (WEAK) WRITE (IUOUT,245) STR(1:LES)
         MERR = -2
         RETURN
      END IF
C
C  FROM should be <= TO.
      IF (IFROM.GT.ITO) THEN
         IF (WEAK) WRITE (IUOUT,220)
         MERR = -2
         RETURN
      END IF
C
C  Determine whether a free format is to be used.
      FREE = ISCOMP(USEF,'*').NE.0 .OR. NPL.EQ.0
      IF (.NOT.FREE) THEN
         LER = LENGTH(USEF)
         FORM = '('//USEF(1:LER)//')'
         LER = LER+2
      END IF
C
C  Get file status.
      INQUIRE (FILE=FILE,EXIST=EXI,OPENED=OPE,NUMBER=NUM,IOSTAT=IOE)
      IF (IOE.NE.0) THEN
         CALL ERRINQ(FILE,IOE)
         MERR = -2
         RETURN
      END IF
C
C  Cannot dump on an open file.
      IF (OPE) THEN
         CALL ERRBUS(FILE,NUM)
         MERR = -2
         RETURN
      END IF
C
C  Allocate unit for the file.
      NU = NUNIT()
      IF (NU.EQ.-1) THEN
         CALL ERRUAL(FILE)
         MERR = -2
         RETURN
      END IF
C
C  Open the file. Take advantage of the APPEND extension if present
C  on this system.
      IAPP = ISCOMP(CAPP,'YES')
      IF (EXI .AND. IAPP.NE.0) THEN
         CALL OPAPPE(NU,FILE,LEF,IOE)
      ELSE
         OPEN (NU,FILE=FILE,IOSTAT=IOE)
      END IF
C
      IF (IOE.NE.0) THEN
         CALL ERROPE(FILE,IOE)
         MERR = -2
         RETURN
      END IF
C
      IF (IAPP.EQ.0) REWIND NU
C
C  Should we write the current value before the rest of the data ?
      IF (ISCOMP(CVAL,'BEFORE').NE.0) THEN
         IF (FREE) THEN
            WRITE (NU,*) VAL
         ELSE
            WRITE (NU,FORM,IOSTAT=IOE) VAL
            IF (IOE.NE.0) GOTO 155
         END IF
         IF (STRO) THEN
            IF (IAPP.EQ.0) THEN
               WRITE (IUOUT,230) FILE(1:LEF)
            ELSE
               WRITE (IUOUT,235) FILE(1:LEF)
            END IF
         END IF
      END IF
C
C  Begin the dump. For all possible WHAT values (except for WHAT='F'
C  and WHAT='N') copy the array to be dumped to a work-array.
C  Then write the work array to the file using the specified
C  FORMAT, NPL etc.
      NTOT = ITO-IFROM+1
      CALL I2STR(NTOT,STR,LES)
      IF (ISCOMP(WHAT,'N').NE.0) THEN
C  Dump names. Names are always written one per line.
         DO 10,I=1,NTOT
10          WRITE (NU,'(A)') NICKNA(IFROM+I-1)
         MSG = 'name(s)'
      ELSE IF (ISCOMP(WHAT,'F').NE.0) THEN
C  Dump fix statuses. Since these are integer variables, we must
C  duplicate the code here, using array IXAT instead of XT1.
         IF (NPL.EQ.0) THEN
            WRITE (NU,*) (IXAT(I),I=IFROM,ITO)
         ELSE
            NLI = NTOT/NPL
            DO 80,I=1,NLI
               IST = (I-1)*NPL+IFROM
               IEND = IST+NPL-1
               IF (FREE) THEN
                  WRITE (NU,*) (IXAT(J),J=IST,IEND)
               ELSE
                  WRITE (NU,FORM,IOSTAT=IOE) (IXAT(J),J=IST,IEND)
                  IF (IOE.NE.0) GOTO 155
               END IF
80          CONTINUE
            NREM = MOD(NTOT,NPL)
            IF (NREM.NE.0) THEN
               IST = NLI*NPL+IFROM
               IEND = ITO
               IF (FREE) THEN
                  WRITE (NU,*) (IXAT(J),J=IST,IEND)
               ELSE
                  WRITE (NU,FORM,IOSTAT=IOE) (IXAT(J),J=IST,IEND)
                  IF (IOE.NE.0) GOTO 155
               END IF
            END IF
         END IF
         MSG = 'fix status(es)'
      ELSE
         IF (ISCOMP(WHAT,'X').NE.0) THEN
C  Dump the current point.
            DO 120,I=IFROM,ITO
120            XT1(I) = POINT(I)
            MSG = 'variable(s)'
         ELSE IF (ISCOMP(WHAT,'L').NE.0) THEN
C  Dump the left margins.
            DO 130,I=IFROM,ITO
               MA = MARG(I)
               IF (MA.EQ.0 .OR. MA.EQ.1) THEN
                  XMA = -2*BLIM
               ELSE
                  XMA = XLL(I)
               END IF
               XT1(I) = XMA
130         CONTINUE
            MSG = 'left margin(s)'
         ELSE IF (ISCOMP(WHAT,'R').NE.0) THEN
C  Dump the right margins.
            DO 140,I=IFROM,ITO
               MA = MARG(I)
               IF (MA.EQ.0 .OR. MA.EQ.-1) THEN
                  XMA = 2*BLIM
               ELSE
                  XMA = XRL(I)
               END IF
               XT1(I) = XMA
140         CONTINUE
            MSG = 'right margin(s)'
         ELSE IF (ISCOMP(WHAT,'S').NE.0) THEN
C  Dump the ROLL steps.
            DO 110,I=IFROM,ITO
110            XT1(I) = STEP(I)
            MSG = 'step(s)'
         ELSE
            CALL MABORT('Incorrect "what".','DUMP')
         END IF
C
C  The work array XT1 now contains the values to be written.
         IF (NPL.EQ.0) THEN
            WRITE (NU,*) (XT1(I),I=IFROM,ITO)
         ELSE
C
C  Write a number of lines containing NPL numbers each.
            NLI = NTOT/NPL
            DO 90,I=1,NLI
               IST = (I-1)*NPL+IFROM
               IEND = IST+NPL-1
               IF (FREE) THEN
                  WRITE (NU,*) (XT1(J),J=IST,IEND)
               ELSE
                  WRITE (NU,FORM,IOSTAT=IOE) (XT1(J),J=IST,IEND)
                  IF (IOE.NE.0) GOTO 155
               END IF
90          CONTINUE
C
C  Write the remainder. (A full line with NPL numbers is not possible.)
            NREM = MOD(NTOT,NPL)
            IF (NREM.NE.0) THEN
               IST = NLI*NPL+IFROM
               IEND = ITO
               IF (FREE) THEN
                  WRITE (NU,*) (XT1(J),J=IST,IEND)
               ELSE
                  WRITE (NU,FORM,IOSTAT=IOE) (XT1(J),J=IST,IEND)
                  IF (IOE.NE.0) GOTO 155
               END IF
            END IF
         END IF
      END IF
C
C  Print informative message.
      LEM = LENGTH(MSG)
      IF (STRO) THEN
         IF (IAPP.EQ.0) THEN
            WRITE (IUOUT,150) STR(1:LES), MSG(1:LEM), FILE(1:LEF)
         ELSE
            WRITE (IUOUT,160) STR(1:LES), MSG(1:LEM), FILE(1:LEF)
         END IF
      END IF
C
C  Should we write the current value after the rest of the data ?
      IF (ISCOMP(CVAL,'AFTER').NE.0) THEN
         IF (FREE) THEN
            WRITE (NU,*) VAL
         ELSE
            WRITE (NU,FORM,IOSTAT=IOE) VAL
            IF (IOE.NE.0) GOTO 155
         END IF
         IF (STRO) THEN
            IF (IAPP.EQ.0) THEN
               WRITE (IUOUT,230) FILE(1:LEF)
            ELSE
               WRITE (IUOUT,235) FILE(1:LEF)
            END IF
         END IF
      END IF
      CLOSE (NU)
      RETURN
C
C  Some error during write. Probably an incorrect format.
155   CONTINUE
      CALL ERRWR(FILE,IOE)
      CLOSE (NU)
      MERR = -2
C
200   FORMAT (/$(FC)'You must enter a filename.')
240   FORMAT (/$(FC)'The specified value for the FROM keyword ',
     &        'is incorrect.'
     &        /$(FC)'There are only ',A,' variables.')
245   FORMAT (/$(FC)'The specified value for the TO keyword ',
     &        'is incorrect.'
     &        /$(FC)'There are only ',A,' variables.')
220   FORMAT (/$(FC)'The values of the "FROM" and "TO" keywords ',
     &        'are incorrect.')
150   FORMAT (/$(FC)A,' ',A,' have been written to file "',A,'".')
160   FORMAT (/$(FC)A,' ',A,' have been appended to file "',A,'".')
230   FORMAT (/$(FC)'The current value has been written to "',A,'".')
235   FORMAT (/$(FC)'The current value has been appended to "',A,'".')
      END
C  ---------------------------------------------------------------------
      SUBROUTINE ECHO
C  ---------------------------------------------------------------------
C
C  Description:
C    Implements the ECHO command.
C    Syntax: ECHO anything
C    ECHO will print its arguments in the output unit IUOUT, allowing
C    custom messages to be issued.
C
C  ---------------------------------------------------------------------
C  Common blocks:
      LOGICAL STRO, WEAK
      COMMON /PRILOG/ STRO, WEAK
      COMMON /UNITS/ IUINP, IUOUT
      PARAMETER ( LINLEN = 80 )
      INTEGER IS(LINLEN/2), IE(LINLEN/2)
      COMMON /INDAT1/ IS, IE, NIT, LEC
      CHARACTER*(LINLEN) INPLIN, QINPL
      COMMON /INDAT2/ INPLIN, QINPL
C  ---------------------------------------------------------------------
C  Local variables:
      CHARACTER*80 LINE
C  ---------------------------------------------------------------------
C
      IF (.NOT.STRO) RETURN
C
      CALL ARGNO(NA)
      IF (NA.EQ.0) THEN
C
C  No command line arguments. Print an empty line.
         WRITE (IUOUT,*)
      ELSE
C
C  Construct message from the command line arguments.
         LL = 0
         DO 20,I=2,NIT
            LE = IE(I)-IS(I)+1
            IF (LE.GT.0) THEN
               NLL = LL+LE
               LINE(LL+1:NLL) = QINPL(IS(I):IE(I))
               LL = NLL
            END IF
            LL = LL+1
            LINE(LL:LL) = ' '
20       CONTINUE
         LL = LL-1
         IF (LL.EQ.0) THEN
            WRITE (IUOUT,*)
         ELSE
            WRITE (IUOUT,10) LINE(1:LL)
         END IF
      END IF
C
10    FORMAT ($(FC)A)
      END
C  ---------------------------------------------------------------------
      SUBROUTINE EPILOG ( MERR )
C  ---------------------------------------------------------------------
C
C  Description:
C    Implements the EPILOG command.
C    Syntax: EPILOG command
C
C    EPILOG sets the command to be executed as epilog when a STOP,
C    RETURN or QUIT command is issued by the user. An empty input
C    (eg EPILOG ' ') clears the epilog.
C    The epilog command can be set in the configuration file as well.
C
C  Output arguments:
C    MERR       Error indicator.
C
C  ---------------------------------------------------------------------
C  Common blocks:
      LOGICAL STRO, WEAK
      COMMON /PRILOG/ STRO, WEAK
      COMMON /UNITS/ IUINP, IUOUT
      PARAMETER ( MXFILE = 80 )
      CHARACTER*(MXFILE) PROLO, EPILO, IN, OUT, HFILE, MFILE, PDFILE
      COMMON /CP2/ PROLO, EPILO, IN, OUT, HFILE, MFILE, PDFILE
C  ---------------------------------------------------------------------
C  Local variables:
      CHARACTER*80 EP
C  ---------------------------------------------------------------------
C
      MERR = 0
      CALL ARGNO(NA)
      LEPI = LENGTH(EPILO)
      IF (NA.EQ.0) THEN
C  No command line arguments. Prompt the user.
         IF (STRO) THEN
            IF (LEPI.EQ.0) THEN
               WRITE (IUOUT,10)
            ELSE
               WRITE (IUOUT,20) EPILO(1:LEPI)
            END IF
            WRITE (IUOUT,30)
         END IF
         CALL INNOCR(EP,.FALSE.,MERR)
         IF (MERR.EQ.-1) RETURN
      ELSE IF (NA.EQ.1) THEN
         CALL CMDARG(1,EP,LEE)
      ELSE
         CALL ERRNAR
         MERR = -2
         RETURN
      END IF
      LEE = LENGTH(EP)
C
C  A zero length input clears the epilog.
      IF (LEE.EQ.0) THEN
         EPILO = ' '
         IF (STRO .AND. LEPI.NE.0) WRITE (IUOUT,50)
      ELSE
         EPILO = EP
         IF (STRO) WRITE (IUOUT,40) EPILO(1:LEE)
      END IF
C
10    FORMAT (/$(FC)'Currently there is no epilog defined.')
20    FORMAT (/$(FC)'The following command is defined ',
     &                          'as epilog: "',A,'".')
30    FORMAT ($(FC)'Enter command to be executed as epilog:')
40    FORMAT (/$(FC)'Epilog is set to "',A,'".')
50    FORMAT (/$(FC)'Epilog has been cleared.')
      END
C  ---------------------------------------------------------------------
      SUBROUTINE EVAL
C  ---------------------------------------------------------------------
C
C  Description:
C    Implements the EVALUATE command.
C    Syntax: EVALUATE
C
C    Evaluates the objective function at the current point. Moreover
C    if a NOEVAL command was issued, it reverses its effect, thus
C    allowing subsequent calls to the objective function to occur.
C
C  ---------------------------------------------------------------------
      IMPLICIT $(TYPE) (A-H,O-Z)
C  Common blocks:
      PARAMETER ( MXV = $(MXV) )
      COMMON /PARAMS/ POINT(MXV), VAL
      COMMON /MODE3/ MODEV, IFFORM, INMCL
      COMMON /TOTCAL/ ITNOCS, IPARTC, ITDER, IPDER, ITHESS, IPHESS,
     &                ITJAC, IPJAC
      LOGICAL STRO, WEAK
      COMMON /PRILOG/ STRO, WEAK
      COMMON /UNITS/ IUINP, IUOUT
      COMMON /DIMEN/ NOD, NOF
C  ---------------------------------------------------------------------
C
      MOP = MODEV
      MODEV = 1
C
C  Check if there are any undefined variables.
      IF (ITNOCS.EQ.0) THEN
         CALL UNDEF(IGO,1)
         IF (IGO.EQ.0) RETURN
      END IF
C
C  Evaluate the function.
      VALNEW = ACSQ(POINT,NOD)
      IF (VALNEW.NE.VAL .AND. MOP.EQ.1 .AND. WEAK) WRITE (IUOUT,10)
      VAL = VALNEW
      IF (MOP.EQ.0 .AND. STRO) WRITE (IUOUT,20)
C
10    FORMAT (/$(FC)'Warning: Function evaluation is inconsistent.')
20    FORMAT (/$(FC)'Function evaluation has been turned on.')
      END
C  ---------------------------------------------------------------------
      SUBROUTINE FIXALL
C  ---------------------------------------------------------------------
C
C  Description:
C    Implements the FIXALL command.
C    Syntax: FIXALL
C
C  ---------------------------------------------------------------------
C  Common blocks:
      PARAMETER ( MXV = $(MXV) )
      COMMON /DETAIL/ IXAT(MXV), MARG(MXV)
      COMMON /DIMEN/ NOD, NOF
      COMMON /UPDATE/ IGRADU, IHESSU
      LOGICAL STRO, WEAK
      COMMON /PRILOG/ STRO, WEAK
      COMMON /UNITS/ IUINP, IUOUT
      COMMON /TOTCAL/ ITNOCS, IPARTC, ITDER, IPDER, ITHESS, IPHESS,
     &                ITJAC, IPJAC
C  ---------------------------------------------------------------------
C
      IF (ITNOCS.EQ.0) RETURN
C
      DO 10,I=1,NOD
         IXAT(I) = 0
         IHESSU = 0
10    CONTINUE
      IF (STRO) WRITE (IUOUT,20)
C
20    FORMAT (/$(FC)'All variables have been fixed.')
      END
C  ---------------------------------------------------------------------
      SUBROUTINE FIXLOO ( LL, MERR )
C  ---------------------------------------------------------------------
C
C  Description:
C    Implements the FIX and LOOSE commands.
C    Syntax: FIX index_specification ...
C            LOOSE index_specification ...
C
C  Input arguments:
C    LL         Determines the operation to be performed:
C                 LL = 0 -> FIX
C                 LL = 1 -> LOOSE
C
C  Output arguments:
C    MERR       Error indicator.
C
C  ---------------------------------------------------------------------
      IMPLICIT $(TYPE) (A-H,O-Z)
C  Common blocks:
      PARAMETER ( MXV = $(MXV) )
      COMMON /DETAIL/ IXAT(MXV), MARG(MXV)
      COMMON /IWORK/ INDE(MXV)
      COMMON /UPDATE/ IGRADU, IHESSU
      COMMON /TOTCAL/ ITNOCS, IPARTC, ITDER, IPDER, ITHESS, IPHESS,
     &                ITJAC, IPJAC
C  ---------------------------------------------------------------------
C
      MERR = 0
      IF (ITNOCS.EQ.0) RETURN
C
C  Prompt the user, or parse command line arguments for index
C  specifications.
      CALL CHI(INDE,NCH,1,MERR)
      IF (MERR.NE.0) RETURN
C
C  Add a backup record.
      IF (NCH.NE.0) CALL BACKUP('D')
      DO 20,I=1,NCH
         IXAT(INDE(I)) = LL
         IF (LL.EQ.1) IGRADU = 0
         IHESSU = 0
20    CONTINUE
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE CHFLAG ( MERR )
C  ---------------------------------------------------------------------
C
C  Description:
C    Implements the FLAG command.
C    Syntax: FLAG index_specification ...
C
C  Output arguments:
C    MERR       Error indicator.
C
C  ---------------------------------------------------------------------
      IMPLICIT $(TYPE) (A-H,O-Z)
C  Common blocks:
      PARAMETER ( MXV = $(MXV) )
      COMMON /IWORK/ INDE(MXV)
      COMMON /RWORK/ XT1(MXV), XT2(MXV), XT3(MXV), XT4(MXV), XT5(MXV),
     &               XT6(MXV), XT7(MXV+1)
      PARAMETER ( MAXFLA = 20 )
      COMMON /FLAGS/ FLAG(MAXFLA)
C  ---------------------------------------------------------------------
C
C  Prompt the user, or parse command line arguments for index/value
C  specifications.
      CALL CHIVX(MAXFLA,INDE,XT1,NPAIRS,1,MERR)
      IF (MERR.NE.0) RETURN
C
      DO 20,I=1,NPAIRS
20       FLAG(INDE(I)) = XT1(I)
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE FLADIS ( MERR )
C  ---------------------------------------------------------------------
C
C  Description:
C    Implements the FLAGDIS command.
C    Syntax: FLAGDIS [ index_specification ... ]
C
C  Output arguments:
C    MERR       Error indicator.
C
C  ---------------------------------------------------------------------
      IMPLICIT $(TYPE) (A-H,O-Z)
C  Common blocks:
      PARAMETER ( MXV = $(MXV) )
      COMMON /IWORK/ INDE(MXV)
      LOGICAL STRO, WEAK
      COMMON /PRILOG/ STRO, WEAK
      COMMON /UNITS/ IUINP, IUOUT
      PARAMETER ( MAXFLA = 20 )
      COMMON /FLAGS/ FLAG(MAXFLA)
C  ---------------------------------------------------------------------
C  Local variables:
      PARAMETER ( NCOL = 4 )
      INTEGER IX(NCOL)
C  ---------------------------------------------------------------------
C
C  Parse any command line arguments for index specifications.
      CALL CHIX(MAXFLA,INDE,NCH,-2,MERR)
      IF (MERR.NE.0) RETURN
C
      IF (.NOT.STRO) RETURN
C
C  Print flags columnwise, NCOL columns.
      IF (STRO) WRITE (IUOUT,*)
      CALL CPRINT('Current flag values')
      CALL CPRINT('-------------------')
      NROW = NCH/NCOL
      NFULL = NROW
      NREM = MOD(NCH,NCOL)
      IF (NREM.NE.0) NROW = NROW+1
      DO 40,IR=1,NROW
         ITO = NCOL
         IF (IR.GT.NFULL) ITO = NREM
         DO 45,I=1,ITO
            IF (I.LE.NREM) THEN
               IX(I) = NROW*(I-1)+IR
            ELSE
               IX(I) = NROW*NREM+(I-NREM-1)*NFULL+IR
            END IF
45       CONTINUE
         WRITE (IUOUT,10) (INDE(IX(I)),FLAG(INDE(IX(I))),I=1,ITO)
40    CONTINUE
C
10    FORMAT ($(FC)4(:'Flag(',I2,')= ',F6.2,3X))
      END
C  ---------------------------------------------------------------------
      SUBROUTINE GNORM
C  ---------------------------------------------------------------------
C
C  Description:
C    Implements the GNORM command. Calculates the L1, L2 and Infinity
C    norms of the gradient vector at the current point. The RMS gradient
C    defined as RMS = SQRT(L2**2/NOD) is calculated as well.
C    The gradient is evaluated using the current derivative mode.
C
C  ---------------------------------------------------------------------
      IMPLICIT $(TYPE) (A-H,O-Z)
C  Common blocks:
      PARAMETER ( MXV = $(MXV) )
      COMMON /PARAMS/ POINT(MXV), VAL
      COMMON /RWORK/ XT1(MXV), XT2(MXV), XT3(MXV), XT4(MXV), XT5(MXV),
     &               XT6(MXV), XT7(MXV+1)
      COMMON /TOTCAL/ ITNOCS, IPARTC, ITDER, IPDER, ITHESS, IPHESS,
     &                ITJAC, IPJAC
      COMMON /MODE3/ MODEV, IFFORM, INMCL
      COMMON /DIMEN/ NOD, NOF
      LOGICAL STRO, WEAK
      COMMON /PRILOG/ STRO, WEAK
      COMMON /UNITS/ IUINP, IUOUT
C  ---------------------------------------------------------------------
C  Local variables:
      CHARACTER*30 C1, C2, CI, CRMS
C  ---------------------------------------------------------------------
C
      IF (ITNOCS.EQ.0 .OR. MODEV.EQ.0) RETURN
C
      CALL DERMER(POINT,VAL,XT1,NOC,NGR,0)
      G1 = 0.
      GSQ = 0.
      GI = 0.
      DO 10,I=1,NOD
         GRABS = ABS(XT1(I))
         G1 = G1+GRABS
         GSQ = GSQ+GRABS**2
         GI = MAX(GI,GRABS)
10    CONTINUE
      G2 = SQRT(GSQ)
      GRMS = SQRT(GSQ/NOD)
C
      IF (STRO) THEN
         WRITE (IUOUT,*)
         CALL DASH
         CALL R2RTF(G1,C1,LE1)
         CALL R2RTF(G2,C2,LE2)
         CALL R2RTF(GI,CI,LEI)
         CALL R2RTF(GRMS,CRMS,LER)
         WRITE (IUOUT,20) C1(1:LE1), C2(1:LE2), CI(1:LEI), CRMS(1:LER)
         CALL DASH
      END IF
C
20    FORMAT ($(FC)'Gradient norms: L1    = ',A/
     &        $(FC)'                L2    = ',A/
     &        $(FC)'                L-inf = ',A/
     &        $(FC)'RMS Gradient          = ',A)
      END
C  ---------------------------------------------------------------------
      SUBROUTINE GODFA ( MERR )
C  ---------------------------------------------------------------------
C
C  Description:
C    Implements the GODFATHER command.
C    Syntax: GODFATHER index name ...
C
C  Output arguments:
C    MERR       Error indicator.
C
C  ---------------------------------------------------------------------
      IMPLICIT $(TYPE) (A-H,O-Z)
C  Common blocks:
      PARAMETER ( MXV = $(MXV) )
      COMMON /IWORK/ INDE(MXV)
C  ---------------------------------------------------------------------
C  Local variables:
      PARAMETER ( LINLEN = 80 )
      CHARACTER*(LINLEN) LINE
      INTEGER IS(LINLEN/2), IE(LINLEN/2), ITEM(MXV)
C  ---------------------------------------------------------------------
C
C  Prompt the user, or parse command line arguments for index/value
C  specifications.
      CALL CHIS(INDE,ITEM,NCH,LINE,IS,IE,1,MERR)
      IF (MERR.NE.0) RETURN
C
C  Add a backup record.
      IF (NCH.EQ.0) CALL BACKUP('D')
C
C  Proceed with the changes. Subroutine SETNAM will actualy assign
C  the name, check for duplicates and issue error messages.
      NERR = 0
      DO 100,I=1,NCH
         IND = INDE(I)
         N = ITEM(I)
         CALL SETNAM(IND,LINE(IS(N):IE(N)),IERR)
         IF (IERR.NE.0) NERR = NERR+1
100   CONTINUE
      IF (NERR.NE.0) MERR = -2
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE GRADC ( MERR )
C  ---------------------------------------------------------------------
C
C  Description:
C    Implements the GRADDIS command:
C    Syntax: GRADDIS mode1 [ mode2 ] [ index_specification ]
C
C  Output arguments:
C    MERR       Error indicator.
C
C  Notes:
C    Due to the syntax of the command, the name of a minimization
C    variable may be mistaken for a derivative mode. Hence you cannot
C    name any minimization variables as FAST, QUAD, NUMER or ANAL and
C    specify them by name in a GRADCHECK command.
C    Suppose variable 2 is named ANAL. You wish to check a FAST
C    gradient against your current mode and display the result for
C    variable #2. The following command will fail to do it.
C
C    GRADCHECK  FAST   ANAL
C               ^      ^
C               |      |
C             mode1  display gradients for
C                    the variable named ANAL.
C    Instead use: GRADCHECK  FAST  2
C    Other pitfalls exist as well.
C
C  ---------------------------------------------------------------------
      IMPLICIT $(TYPE) (A-H,O-Z)
C  Common blocks:
      PARAMETER ( MXV = $(MXV) )
      COMMON /RWORK/ XT1(MXV), XT2(MXV), XT3(MXV), XT4(MXV), XT5(MXV),
     &               XT6(MXV), XT7(MXV+1)
      COMMON /PARAMS/ POINT(MXV), VAL
      CHARACTER NICKNA*10
      COMMON /NAMES/ NICKNA(MXV)
      CHARACTER*5 GRADI
      COMMON /DERIMO/ GRADI(MXV)
      COMMON /DIMEN/ NOD, NOF
      CHARACTER*10 DERIVA, PRINTO, PROCES, JACOMO
      COMMON /MODEIN/ DERIVA, PRINTO, PROCES, JACOMO
      COMMON /TOTCAL/ ITNOCS, IPARTC, ITDER, IPDER, ITHESS, IPHESS,
     &                ITJAC, IPJAC
      COMMON /MODE3/ MODEV, IFFORM, INMCL
      LOGICAL STRO, WEAK
      COMMON /PRILOG/ STRO, WEAK
      COMMON /UNITS/ IUINP, IUOUT
      PARAMETER ( LINLEN = 80 )
      INTEGER IS(LINLEN/2), IE(LINLEN/2)
      COMMON /INDAT1/ IS, IE, NIT, LEC
      COMMON /IWORK/ INDE(MXV)
C  ---------------------------------------------------------------------
C  Local variables:
      CHARACTER*5 OLDMO, OLDGR(MXV), HELPMO
      CHARACTER*10 DMODE, DMODE2, TEMP
      CHARACTER*7 CFIX
      LOGICAL ISMIX
C  ---------------------------------------------------------------------
C
      MERR = 0
      IF (ITNOCS.EQ.0 .OR. MODEV.EQ.0) RETURN
C
      CALL ARGNO(NA)
      IF (NA.EQ.0) THEN
C  No arguments. Ask for a derivative mode.
         IF (STRO) WRITE (IUOUT,10)
         CALL INNOCR(DMODE,.TRUE.,MERR)
         IF (MERR.NE.0) RETURN
         IF (ISOKDM(DMODE).EQ.0) THEN
            IF (WEAK) WRITE (IUOUT,20)
            MERR = -2
            RETURN
         END IF
         DMODE2 = DERIVA
      ELSE
C  Something is present in the comand line.
         CALL CMDARG(1,DMODE,LED)
         CALL UPPER(DMODE,LED)
         IF (ISOKDM(DMODE).NE.0) THEN
C  DMODE is a valid derivative mode. Remove this argument from the
C  command line.
            NIT = NIT-1
            NA = NA-1
            DO 40,I=2,NIT
               IS(I) = IS(I+1)
               IE(I) = IE(I+1)
40          CONTINUE
         ELSE
C  DMODE is not a valid derivative mode. Maybe its an index
C  specification. Ask for a derivative mode.
            IF (STRO) WRITE (IUOUT,10)
            CALL INNOCR(DMODE,.TRUE.,MERR)
            IF (MERR.NE.0) RETURN
            IF (ISOKDM(DMODE).EQ.0) THEN
               IF (WEAK) WRITE (IUOUT,20)
               MERR = -2
               RETURN
            END IF
         END IF
C
C  Are there any other arguments ?
         IF (NA.GE.1) THEN
            CALL CMDARG(1,DMODE2,LED2)
            CALL UPPER(DMODE2,LED2)
            IF (ISOKDM(DMODE2).NE.0) THEN
C  DMODE2 is a valid derivative mode. Remove it from the command line.
               NIT = NIT-1
               DO 50,I=2,NIT
                  IS(I) = IS(I+1)
                  IE(I) = IE(I+1)
50             CONTINUE
            ELSE
               DMODE2 = DERIVA
            END IF
         ELSE
            DMODE2 = DERIVA
         END IF
      END IF
C
C  Check whether the derivative modes coincide.
      IF (DMODE.EQ.DMODE2) THEN
         IF (DMODE.EQ.DERIVA) THEN
            LED = LENGTH(DERIVA)
            IF (WEAK) WRITE (IUOUT,60) DERIVA(1:LED)
         ELSE
            IF (WEAK) WRITE (IUOUT,70)
         END IF
         MERR = -2
         RETURN
      END IF
C
C  Make sure that DMODE2 always contains the current derivative mode
C  (if the user specifies it).
      IF (DMODE.EQ.DERIVA) THEN
         TEMP = DMODE
         DMODE = DMODE2
         DMODE2 = TEMP
      END IF
C
C  Get index specifications.
      CALL CHI(INDE,NCH,-2,MERR)
      IF (MERR.NE.0) RETURN
C
C  Start check. Calculate derivatives with DMODE2 first.
      OLDMO = DERIVA
      DERIVA = DMODE2
      DO 15,I=1,NOD
15       OLDGR(I) = GRADI(I)
      IF (DMODE2.NE.OLDMO) THEN
         DO 17,I=1,NOD
17          GRADI(I) = DMODE2
      END IF
C
C  If DMODE2 is the current dervative mode, request normal DERMER
C  operation. Otherwise prevent the gradient from being cached.
      IF (DMODE2.EQ.OLDMO) THEN
         IOPT = 0
      ELSE
         IOPT = 1
      END IF
      CALL DERMER(POINT,VAL,XT1,NOC,NGR,IOPT)
C
C  Calculate derivatives with DMODE.
      DERIVA = DMODE
      DO 16,I=1,NOD
16       GRADI(I) = DMODE
C
C  Since DMODE cannot be the current derivative mode, do not cache
C  the gradient.
      CALL DERMER(POINT,VAL,XT2,NOC,NGR,1)
C
C  Restore gradient mode.
      DERIVA = OLDMO
      DO 30,I=1,NOD
30       GRADI(I) = OLDGR(I)
      IF (.NOT.STRO) RETURN
C
C  Print the results.
      WRITE (IUOUT,*)
      CALL DASH
      WRITE (IUOUT,5) DMODE, DMODE2
      CALL DASH
      SUMERR = 0.
      DIFMAX = 0.
      ISMIX = DMODE2.EQ.'MIXED'
      ONE = 1.
      DO 2,K=1,NCH
         I = INDE(K)
         GCURR = XT1(I)
         GOTHER = XT2(I)
         ERR = ABS(GOTHER-GCURR)/
     &            MAX(ONE,MIN(ABS(GOTHER),ABS(GCURR)))
         SUMERR = SUMERR+ERR**2
         IF (ERR.GT.DIFMAX) DIFMAX = ERR
         HELPMO = ' '
         IF (ISMIX) HELPMO = GRADI(I)
         CALL FIXSTR(I,CFIX)
         WRITE (IUOUT,4) I, NICKNA(I), CFIX, GOTHER, GCURR, HELPMO,
     &                      ERR
2     CONTINUE
      RMS = SQRT(SUMERR/NCH)
      WRITE (IUOUT,6) RMS, DIFMAX
C
4     FORMAT ($(FC)I5,') ',A10,1X,A,2X,1PE12.5,3X,1PE12.5,2X,A,
     &        3X,1PE10.3)
5     FORMAT ($(FC)'Index',2X,'Name',7X,'Fix',10X,A,5X,A,8X,
     &        'Rel. Error')
6     FORMAT (/$(FC)'RMS relative error : ',1PE11.4
     &        /$(FC)'Max. relative error: ',1PE11.4)
10    FORMAT (/$(FC)'Enter derivative mode (fast, quad, numer, anal):')
20    FORMAT (/$(FC)'You must enter one of: fast, quad, numer, anal.')
60    FORMAT (/$(FC)'"',A,'" is your current mode.')
70    FORMAT (/$(FC)'You must specify two different derivative modes.')
      END
C  ---------------------------------------------------------------------
      FUNCTION ISOKDM ( DMODE )
C  ---------------------------------------------------------------------
C
C  Description:
C    Part of the GRADCHECK command. Returns 1 if DMODE is one of the
C    four derivative modes: NUMER, QUAD, FAST or ANAl; 0 otherwise.
C
C  Input arguments:
C    DMODE      Derivative mode to check.
C
C  ---------------------------------------------------------------------
C  Arguments:
      CHARACTER*(*) DMODE
C  ---------------------------------------------------------------------
C
      IF (DMODE.EQ.'NUMER' .OR. DMODE.EQ.'QUAD' .OR.
     &    DMODE.EQ.'FAST'  .OR. DMODE.EQ.'ANAL') THEN
         ISOKDM = 1
      ELSE
         ISOKDM = 0
      END IF
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE GRAPH ( MERR )
C  ---------------------------------------------------------------------
C
C  Description:
C    Subroutine GRAPH produces a rough X-Y graph. It is suitable for
C    display on ASCII terminals, or printing on conventional character
C    printers.
C
C  Output arguments:
C    MERR       Error indicator.
C
C  ---------------------------------------------------------------------
      IMPLICIT $(TYPE) (A-H,O-Z)
C  Common blocks:
      COMMON /TOTCAL/ ITNOCS, IPARTC, ITDER, IPDER, ITHESS, IPHESS,
     &                ITJAC, IPJAC
      COMMON /MODE3/ MODEV, IFFORM, INMCL
C  ---------------------------------------------------------------------
C
C  Present the panel and process any user changes.
      CALL SETPI('CANCEL',1,IERR)
      CALL CHANGE(MERR)
      IF (MERR.NE.0) RETURN
      CALL GETPI('CANCEL',ICANC)
      IF (ICANC.EQ.0 .OR. ITNOCS.EQ.0 .OR. MODEV.EQ.0) RETURN
C
C  Get panel parameters.
      CALL GETPI('INDEX',IWVAR)
      CALL GETPI('NOP',NOCGR)
      CALL GETPR('FROM',GXMIN)
      CALL GETPR('TO',GXMAX)
      CALL GETPI('LINES',ILIN)
      CALL GETPI('COLS',ICOL)
C
C  Call the grapher.
      XAX = 0.
      YAX = 0.
      CALL GRA1(NOCGR,ILIN,ICOL,IWVAR,GXMIN,GXMAX,XAX,YAX)
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE GRA1 ( NOC, ILIN, ICOL, IWVAR, XMIN, XMAX, XAX, YAX )
C  ---------------------------------------------------------------------
C
C  Description:
C    Called by subroutine GRAPH. It performs the necessary preparations
C    for the graph.
C
C  Input arguments:
C    NOC        Number of points to be plotted.
C    ILIN       Number of lines to appear in the plot.
C    ICOL       Number of columns to appear in the plot.
C    IWVAR      Variables to be plotted.
C    XMIN       Low bound on the x axis.
C    XMAM       Upper bound on the x axis.
C    XAX        Zero position on x axis.
C    YAX        Zero position on y axis.
C
C  ---------------------------------------------------------------------
      IMPLICIT $(TYPE) (A-H,O-Z)
C  Common blocks:
      PARAMETER ( MXV = $(MXV) )
      COMMON /PARAMS/ POINT(MXV), VAL
      COMMON /DETAIL/ IXAT(MXV), MARG(MXV)
      COMMON /XLIMS/ XLL(MXV), XRL(MXV)
      COMMON /DIMEN/ NOD, NOF
      LOGICAL STRO, WEAK
      COMMON /PRILOG/ STRO, WEAK
      COMMON /UNITS/ IUINP, IUOUT
C  ---------------------------------------------------------------------
C  Local variables:
      CHARACTER*50 MSX, MSY
      PARAMETER (ICMAX = 124)
      DIMENSION X(ICMAX), Y(ICMAX)
C  ---------------------------------------------------------------------
C
      IF (IWVAR.LT.1 .OR. IWVAR.GT.NOD) RETURN
C
C  Check if variable is fixed.
      IF (IXAT(IWVAR).EQ.0) RETURN
C
C  Check XMIN, XMAX (if out of margins).
      IF (XMAX.LT.XMIN) THEN
         XX = XMAX
         XMAX = XMIN
         XMIN = XX
      END IF
      MI = MARG(IWVAR)
      IF (MI.NE.0) THEN
         IF ((MI.EQ.-1 .OR. MI.EQ.2) .AND. XMIN.LT.XLL(IWVAR))
     &           XMIN = XLL(IWVAR)
         IF ((MI.EQ.1 .OR. MI.EQ.2) .AND. XMAX.GT.XRL(IWVAR))
     &           XMAX = XRL(IWVAR)
      END IF
C
C  Check ILIN, ICOL.
      IF (ICOL.LT.2) ICOL = 2
      IF (ICOL.GT.ICMAX) ICOL = ICMAX
      IF (ILIN.LT.2) ILIN = 2
      IF (NOC.LT.2) NOC =2
      IF (NOC.GT.ICOL) NOC = ICOL
      IVAR = IWVAR
      N = NOD
      XSTEP = (XMAX-XMIN)/(NOC-1)
C
C  Prepare X, Y.
      DO 2,I=1,NOC
         XX = XMIN+(I-1)*XSTEP
         X(I) = XX
         Y(I) = ACSQ1(POINT,IWVAR,XX)
2     CONTINUE
C
C  Add current point (CPP) if XMIN <= CPP <= XMAX
      CPP = POINT(IWVAR)
      NCC = NOC
      IF (CPP.GE.XMIN .AND. CPP.LE.XMAX) THEN
         NCC = NOC+1
         Y(NCC) = VAL
         X(NCC) = CPP
      END IF
C
C  Find YMIN.
      IMIN = 1
      DO 3,I=2,NCC
3        IF (Y(I).LT.Y(IMIN)) IMIN = I
C
C  If a better point has been found update POINT, VAL.
      IF (Y(IMIN).LT.VAL) THEN
         POINT(IWVAR) = X(IMIN)
         VAL = Y(IMIN)
      END IF
C
C  Mark current point.
      X0 = POINT(IWVAR)
      Y0 = VAL
C
C  Prepare MSX, MSY.
      MSX = '    Variable '
      CALL VNAME(IWVAR,MSX(14:),LEM)
      MSY = '  FUNCTION  VALUE'
      IT = IUOUT
      IF (STRO)
     &    CALL GRAF(X,Y,NCC,ILIN,ICOL,XAX,YAX,X0,Y0,IT,MSX,MSY,MIS)
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE GRAF ( X, Y, NOV, ILIN, ICOL, XAX, YAX, X0, Y0, ITAPE,
     &                  MSX, MSY, MIS )
C  ---------------------------------------------------------------------
C
C  Description:
C    Plots Y versus X in unit ITAPE.
C
C  Input arguments:
C    X          N
C    Y          N
C    NOV        Number of data points
C    ILIN       Number of lines.
C    ICOL       Number of coloumns.
C    XAX        Origin's X-coordinate.
C    YAX        Origin's Y-coordinate.
C    X0         X-coordinate of a point to be marked.
C    Y0         Y-coordinate of a point to be marked.
C    ITAPE      Unit to receive output.
C    MSX        Message to be displayed on the X-axis.
C    MSY        Message to be displayed on the Y-axis.
C
C  Output arguments:
C    MIS        MIS = 1 -> Input arguments are wrong and no graph
C                          takes place.
C               MIS = 0 -> Input arguments are ok.
C
C  ---------------------------------------------------------------------
      IMPLICIT $(TYPE) (A-H,O-Z)
C  Arguments:
      DIMENSION X(NOV), Y(NOV)
C  ---------------------------------------------------------------------
C  Local variables:
      PARAMETER ( NMAX = 155 )
      DIMENSION IYY(NMAX), XAA(15)
      CHARACTER*124 GRA, BLANK, MSX*50, MSY*50
      CHARACTER*1 PL, ST, DOT, EX, MPR, B
      DATA  PL,  ST, DOT, EX,   B /
     &     '+', '*', '.', 'X', ' '/
C  ---------------------------------------------------------------------
C
C  Check input arguments.
      CALL GRCHE(NOV,ILIN,ICOL,ITAPE,NMAX,MIS)
      IF (MIS.EQ.1) RETURN
C
C  Prepare BLANK.
      DO 1,I=1,124
1        BLANK(I:I) = B
      GRA = BLANK
C
C  Find XMIN-XMAX, YMIN-YMAX.
      XMIN = X(1)
      XMAX = X(1)
      YMIN = Y(1)
      YMAX = Y(1)
      DO 2,I=2,NOV
         IF (X(I).LT.XMIN) XMIN = X(I)
         IF (X(I).GT.XMAX) XMAX = X(I)
         IF (Y(I).LT.YMIN) YMIN = Y(I)
         IF (Y(I).GT.YMAX) YMAX = Y(I)
2     CONTINUE
C
C Find X-step, Y-step.
      STEPX = (XMAX-XMIN)/(ICOL-1)
      STEPY = (YMAX-YMIN)/(ILIN-1)
      IF (STEPX.EQ.0.) STEPX = 1.
      IF (STEPY.EQ.0.) STEPY = 1.
C
C  Prepare Y directives.
      DO 3,I=1,NOV
3        IYY(I) = -INT((Y(I)-YMIN)/STEPY+.5)+ILIN
C
C  Prepare XAX-YAX directives.
      IXAX = 0
      IYAX = 0
      IF (XAX.GE.XMIN .AND. XAX.LE.XMAX)
     &    IXAX = INT((XAX-XMIN)/STEPX+.5)+2
      IF (YAX.GE.YMIN .AND. YAX.LE.YMAX)
     &    IYAX = ILIN-INT((YAX-YMIN)/STEPY+.5)
C
C  Prepare X0 - Y0 directives.
      IX0 = 0
      IY0 = 0
      IF (X0.GE.XMIN .AND. X0.LE.XMAX)
     &    IX0 = INT((X0-XMIN)/STEPX+.5)+2
      IF (Y0.GE.YMIN .AND. Y0.LE.YMAX)
     &    IY0 = ILIN-INT((Y0-YMIN)/STEPY+.5)
      IC2 = ICOL+2
      DO 4,K=1,IC2
4        GRA(K:K) = DOT
      DO 5,K=2,IC2,5
5        GRA(K:K) = PL
      WRITE (ITAPE,17) BLANK
      WRITE (ITAPE,17) GRA(1:IC2)
      DO 8,I=1,ILIN
         IF (I.LE.50) THEN
            MPR = MSY(I:I)
         ELSE
            MSY = B
         END IF
         IF (I.NE.IYAX) THEN
            GRA = BLANK
         ELSE
            DO 6,K=2,ICOL+1
6              GRA(K:K) = DOT
         END IF
         IF (IXAX.NE.0) GRA(IXAX:IXAX) = DOT
         LY = ((I-1)/2)*2-I+1
         IF (LY.NE.0) THEN
            GRA(1:1) = DOT
            GRA(IC2:IC2) = DOT
         ELSE
            GRA(1:1) = PL
            GRA(IC2:IC2) = PL
         END IF
C
C  Check if there is a point at this line.
         DO 7,J=1,NOV
            IF (I.EQ.IYY(J)) THEN
               IX = INT((X(J)-XMIN)/STEPX+.5)+2
               GRA(IX:IX) = ST
            END IF
7        CONTINUE
         IF (I.EQ.IY0) GRA(IX0:IX0) = EX
         IF (LY.NE.0) THEN
            WRITE (ITAPE,13) MPR, GRA(1:IC2)
         ELSE
            YY = YMIN-REAL(I-ILIN)*STEPY
            WRITE (ITAPE,14) MPR, YY, GRA(1:IC2)
         END IF
8     CONTINUE
      DO 9,K=1,IC2
9        GRA(K:K) = DOT
      DO 10,K=2,IC2,5
10       GRA(K:K) = PL
      WRITE (ITAPE,17) GRA(1:IC2)
      NX = 0
      DO 11,I=2,IC2,10
         NX = NX+1
         XAA(NX) = XMIN+REAL(I-2)*STEPX
11    CONTINUE
      WRITE (BLANK,15) (XAA(I),I=1,NX,2)
      ICN = IC2+13
      IF (ICN.GT.74 .AND. ICN.LT.80) THEN
         DO 12,I=60,72
12          BLANK(I:I) = B
         WRITE (ITAPE,18) BLANK(1:71)
      ELSE
         WRITE (ITAPE,18) BLANK(1:IC2)
      END IF
      IF (NX.GE.2) THEN
         WRITE (ITAPE,16) (XAA(I),I=2,NX,2)
      END IF
      WRITE (ITAPE,17) MSX(1:50)
C
13    FORMAT ($(FC)1X,A1,11X,A)
14    FORMAT ($(FC)1X,A1,1X,1PG10.3,A)
15    FORMAT (6(1PE13.6,'   I   '))
16    FORMAT ($(FC)18X,5(1PE13.6,7X))
17    FORMAT ($(FC)13X,A)
18    FORMAT ($(FC)8X,A)
      END
C  ---------------------------------------------------------------------
      SUBROUTINE GRCHE ( NOV, ILIN, ICOL, ITAPE, NMAX, MIS )
C  ---------------------------------------------------------------------
C
C  Description:
C    Called by subroutine GRAF. It checks the input arguments for
C    validity and resets the values for ILIN, ICOL, if not apropriate.
C
C  Input arguments:
C    NOV        Number of data points
C    ILIN       Number of lines.
C    ICOL       Number of coloumns.
C    ITAPE      Unit to receive output.
C    NMAX       Maximum number of data points.
C
C  Output arguments:
C    MIS        MIS = 1 if an error occurs, else MIS = 0.
C
C  ---------------------------------------------------------------------
      IMPLICIT $(TYPE) (A-H,O-Z)
C
      MIS = 0
      IF (NOV.LE.1 .OR. NOV.GT.NMAX .OR. ITAPE.LT.0) MIS = 1
      IF (ILIN.LT.2) ILIN = 2
      IF (ICOL.LT.2) ICOL = 2
      IF (ICOL.GT.122) ICOL = 122
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE GRDIS ( MERR )
C  ---------------------------------------------------------------------
C
C  Description:
C    Implements the GRADDIS command.
C    Syntax: GRADDIS [ index_specification ... ]
C
C  Output arguments:
C    MERR       Error indicator.
C  ---------------------------------------------------------------------
      IMPLICIT $(TYPE) (A-H,O-Z)
C  Common blocks:
      PARAMETER ( MXV = $(MXV) )
      COMMON /PARAMS/ POINT(MXV), VAL
      CHARACTER*5 GRADI
      COMMON /DERIMO/ GRADI(MXV)
      CHARACTER NICKNA*10
      COMMON /NAMES/ NICKNA(MXV)
      COMMON /TOTCAL/ ITNOCS, IPARTC, ITDER, IPDER, ITHESS, IPHESS,
     &                ITJAC, IPJAC
      COMMON /MODE3/ MODEV, IFFORM, INMCL
      COMMON /IWORK/ INDE(MXV)
      COMMON /RWORK/ XT1(MXV), XT2(MXV), XT3(MXV), XT4(MXV), XT5(MXV),
     &               XT6(MXV), XT7(MXV+1)
      LOGICAL STRO, WEAK
      COMMON /PRILOG/ STRO, WEAK
      COMMON /UNITS/ IUINP, IUOUT
C  ---------------------------------------------------------------------
C  Local variables:
      CHARACTER*23 STR
      CHARACTER*7 CFIX
C  ---------------------------------------------------------------------
C
C  Parse any command line arguments for index specifications.
      CALL CHI(INDE,NCH,-2,MERR)
      IF (MERR.NE.0) RETURN
C
      IF (ITNOCS.EQ.0 .OR. MODEV.EQ.0) RETURN
C
C  Evaluate the gradient at the current point. If subroutine DERMER
C  has the gradient available in its internal cache, use it.
      CALL DERMER(POINT,VAL,XT1,NOC,NGR,0)
C
      IF (.NOT.STRO) RETURN
      IF (STRO) WRITE (IUOUT,10)
      DO 20,I=1,NCH
         IND = INDE(I)
         CALL R2ERTF(XT1(IND),STR,LES)
         CALL FIXSTR(IND,CFIX)
         WRITE (IUOUT,30) IND, NICKNA(IND), CFIX, STR, GRADI(IND)
20    CONTINUE
C
10    FORMAT (/$(FC)'Index',2X,'Name',6X,'Fix',7X,'Gradient',16X,
     &        'Gradient-Mode'/$(FC)64('-'))
30    FORMAT ($(FC)I5,') ',A10,A,1X,A,7X,A)
      END
C  ---------------------------------------------------------------------
      SUBROUTINE HELP
C  ---------------------------------------------------------------------
C
C  Description:
C    Implements the HELP command.
C    Syntax: HELP            -> Displays the standard help message.
C            HELP command    -> Displays help for the specified command.
C            HELP subject    -> Displays help for the specified subject.
C
C    Help is taken from the help file, or, for a panel command, from the
C    panel description file.
C
C  ---------------------------------------------------------------------
      IMPLICIT $(TYPE) (A-H,O-Z)
C  Common blocks:
      PARAMETER ( MXFILE = 80 )
      CHARACTER*(MXFILE) PROLO, EPILO, IN, OUT, HFILE, MFILE, PDFILE
      COMMON /CP2/ PROLO, EPILO, IN, OUT, HFILE, MFILE, PDFILE
      LOGICAL STRO, WEAK
      COMMON /PRILOG/ STRO, WEAK
      COMMON /UNITS/ IUINP, IUOUT
C  Maximum length of a keyword.
      PARAMETER ( KEYLEN = 10 )
C  Maximum number of panel commands.
      PARAMETER ( MXPCOM = 80 )
C  Maximum number of keywords in all panel commands.
      PARAMETER ( MXPKEY = 1000 )
C  Maximum storage for integer/real keywords.
      PARAMETER ( MXPIR = 5000 )
C  Maximum storage for string keywords and descriptions.
      PARAMETER ( MXPSTR = 20000 )
C  Maximum number of pointers in the above string storage.
      PARAMETER ( MXPPST = 5000 )
C
      CHARACTER*(KEYLEN) PCOM(MXPCOM)
      INTEGER            PONF(MXPCOM)
      INTEGER            KPOINT(MXPCOM)
      INTEGER            NCHELP(2*MXPCOM)
      CHARACTER*(KEYLEN) PKEY(MXPKEY)
      INTEGER            NKHELP(2*MXPKEY)
      CHARACTER          KTYPE(MXPKEY)
      INTEGER            TPOINT(MXPKEY)
      CHARACTER          KIO(MXPKEY)
      $(TYPE)            PIR(MXPIR)
      INTEGER            SPOINT(MXPPST)
      CHARACTER*(MXPSTR) PANSTR
      COMMON /PCOMS1/ KPOINT, NCHELP, NKHELP, TPOINT, SPOINT, PONF
      COMMON /PCOMS2/ PIR
      COMMON /PCOMS3/ PKEY, KTYPE, PANSTR, PCOM, KIO
C
C  NPCOM: Number of panel commands in the panel description file.
C  NPKEY: Number of keywords in all panel commands.
C  NPIR:  Number of positions filled in array PIR.
C  NPSP:  Number of positions filled in the char-pointer array SPOINT.
C  NPCH:  Number of character positions filled in PANSTR.
      COMMON /PCOMS4/ NPCOM, NPKEY, NPIR, NPSP, NPCH
C  ---------------------------------------------------------------------
C  Local variables:
      CHARACTER*80 ARG
C  ---------------------------------------------------------------------
C
      CALL ARGNO(NA)
      IF (NA.EQ.0) THEN
C  Display the standard help message.
         CALL OPHELP(HFILE,NU,1)
         IF (NU.EQ.-1) RETURN
         CALL HCOPY(NU)
         CLOSE (NU)
         RETURN
      END IF
C
      CALL CMDARG(1,ARG,LEA)
      IF (NA.GT.1 .OR. LEA.EQ.0) THEN
C  If the command line arguments are incorrect, issue help for the
C  HELP command itself.
         ARG = 'HELP'
         LEA = 4
      END IF
C
C  Locate the command in the command-table.
      CALL LOCC(ARG,IC)
C
C  Check if the command has a panel.
      IF (IC.NE.0) THEN
         CALL PLOCP(ARG,IP)
      ELSE
         IP = 0
      END IF
C
      IF (IP.EQ.0) THEN
C  No panel. Use the help file.
         CALL OPHELP(HFILE,NU,1)
         IF (NU.EQ.-1) RETURN
C  Locate the command in the help file.
         CALL HFIND(ARG,LEA,IC,NU,IFO)
         IF (IFO.NE.0) THEN
            CALL HCOPY(NU)
         ELSE
            IF (IC.NE.0) THEN
               IF (STRO) THEN
                  LEF = LENGTH(HFILE)
                  WRITE (IUOUT,20) ARG(1:LEA), HFILE(1:LEF)
               END IF
            ELSE
               IF (STRO) WRITE (IUOUT,30) ARG(1:LEA)
            END IF
         END IF
         CLOSE (NU)
      ELSE
C  The command has a panel. Use the panel description file.
         CALL OPHELP(PDFILE,NU,2)
         IF (NU.EQ.-1) RETURN
         LISTA = NCHELP(2*IP-1)
         LIEND = NCHELP(2*IP)
         IF (LISTA.EQ.0) THEN
            LEP = LENGTH(PDFILE)
            IF (STRO) WRITE (IUOUT,40) ARG(1:LEA), PDFILE(1:LEP)
         ELSE
            CALL DASH
            CALL PCOPY(NU,LISTA,LIEND)
            CALL DASH
         END IF
         CLOSE (NU)
      END IF
C
20    FORMAT (/$(FC)'Sorry, there is no help available for the "',A,
     &        '" command'/$(FC)'in the help file "',A,'".')
30    FORMAT (/$(FC)'There is no command or topic named "',A,'".')
40    FORMAT (/$(FC)'Sorry, there is no help available for the "',A,
     &        '" command'/$(FC)'in the panel description file "',A,'".')
      END
C  ---------------------------------------------------------------------
      SUBROUTINE HCOPY ( NU )
C  ---------------------------------------------------------------------
C
C  Description:
C    Part of the HELP command. Copies the contents of the help file
C    to the Merlin output, until the description of a new command or
C    subject is reached, or until End-of-File is encountered.
C
C  Input arguments:
C    NU         Unit number of the help file.
C
C  ---------------------------------------------------------------------
C  Common blocks:
      LOGICAL STRO, WEAK
      COMMON /PRILOG/ STRO, WEAK
      COMMON /UNITS/ IUINP, IUOUT
C  ---------------------------------------------------------------------
C  Local variables:
      CHARACTER*80 LINE
      CHARACTER*3 PRE
C  ---------------------------------------------------------------------
C
      CALL DASH
30    READ (NU,'(A)',END=20) LINE
      PRE = LINE(1:3)
      IF (PRE.EQ.'***' .OR. PRE.EQ.'===') GOTO 20
      LL = LENGTH(LINE)
      IF (LL.EQ.0) THEN
         IF (STRO) WRITE (IUOUT,*)
      ELSE
         IF (STRO) WRITE (IUOUT,'($(FC)A)') LINE(1:LL)
      END IF
      GOTO 30
C
20    CONTINUE
      CALL DASH
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE HFIND ( CMD, LEC, IC, NU, IFO )
C  ---------------------------------------------------------------------
C
C  Description:
C    Part of the HELP command. Searches the help file for a description
C    of the command or subject, as specified in CMD.
C
C  Input arguments:
C    CMD        Command or subject to search for.
C    LEC        Effective length of CMD.
C    IC         What are we searching for ?
C               IC = 0    -> We are searching for a subject.
C               IC .NE. 0 -> We are searching for a Merlin command.
C    NU         Unit number of the help file.
C
C  Output arguments:
C    IFO        Indicates whether a description for CMD exists in the
C               help file:
C               IFO = 1 -> Description found. The help file is
C                          positioned before the description of CMD.
C               IFO = 0 -> Description not found.
C
C  ---------------------------------------------------------------------
C  Arguments:
      CHARACTER*(*) CMD
C  ---------------------------------------------------------------------
C  Local variables:
      CHARACTER*80 LINE
      CHARACTER*3 PRE
C  ---------------------------------------------------------------------
C
30    READ (NU,'(A)',END=20) LINE
      PRE = LINE(1:3)
      IF (PRE.EQ.'***' .AND. IC.NE.0  .OR.
     &    PRE.EQ.'===' .AND. IC.EQ.0) THEN
         IF (ISCOMP(LINE(4:),CMD(1:LEC)).NE.0) THEN
            IFO = 1
            RETURN
         END IF
      END IF
      GOTO 30
C
20    CONTINUE
      IFO = 0
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE OPHELP ( FILE, NU, IFILE )
C  ---------------------------------------------------------------------
C
C  Description:
C    Part of the HELP command. Opens the help or the panel description
C    file.
C
C  Input arguments:
C    FILE       Name of the file to open.
C    IFILE      IFILE = 1 -> Open the help file.
C               IFILE = 2 -> Open the panel description file.
C
C  Output arguments:
C    NU         Unit number of the opened file. NU is set to -1 if
C               any errors occur.
C
C  ---------------------------------------------------------------------
C  Arguments:
      CHARACTER*(*) FILE
C  ---------------------------------------------------------------------
C  Common blocks:
      LOGICAL STRO, WEAK
      COMMON /PRILOG/ STRO, WEAK
      COMMON /UNITS/ IUINP, IUOUT
C  ---------------------------------------------------------------------
C  Local variables:
      LOGICAL EXI, OPE
C  ---------------------------------------------------------------------
C
      NU = -1
      LEF = LENGTH(FILE)
      IF (LEF.EQ.0) THEN
         IF (IFILE.EQ.1) THEN
            IF (STRO) WRITE (IUOUT,10)
         ELSE
            IF (STRO) WRITE (IUOUT,20)
         END IF
         RETURN
      END IF
C
C  Get info for the file.
      INQUIRE (FILE=FILE,EXIST=EXI,OPENED=OPE,NUMBER=NUM,IOSTAT=IOE)
      IF (IOE.NE.0) THEN
         CALL ERRINQ(FILE,IOE)
         RETURN
      END IF
C
C  Does the file exist ?
      IF (.NOT.EXI) THEN
         CALL ERREXI(FILE)
         RETURN
      END IF
C
C  Is the file opened ?
      IF (OPE) THEN
         CALL ERRBUS(FILE,NUM)
         RETURN
      END IF
C
C  Allocate a unit number.
      NU = NUNIT()
      IF (NU.EQ.-1) THEN
         CALL ERRUAL(FILE)
         RETURN
      END IF
C
C  Open the file.
      OPEN (NU,FILE=FILE,IOSTAT=IOE)
      IF (IOE.NE.0) THEN
         CALL ERROPE(FILE,IOE)
         NU = -1
         RETURN
      END IF
      REWIND NU
C
10    FORMAT (/$(FC)'Sorry, there is no help file.')
20    FORMAT (/$(FC)'Sorry, there is no panel description file.')
      END
C  ---------------------------------------------------------------------
      SUBROUTINE PCOPY ( NU, NFROM, NTO )
C  ---------------------------------------------------------------------
C
C  Description:
C    Part of the HELP command. Copies a range of lines (NFROM..NTO) from
C    the panel descrition file to the Merlin output.
C
C  Input arguments:
C    NU         Unit number of the panel description file.
C    NFROM      The first line to copy.
C    NTO        The last line to copy.
C
C  ---------------------------------------------------------------------
C  Common blocks:
      LOGICAL STRO, WEAK
      COMMON /PRILOG/ STRO, WEAK
      COMMON /UNITS/ IUINP, IUOUT
C  ---------------------------------------------------------------------
C  Local variables:
      CHARACTER CH
      CHARACTER*80 LINE
C  ---------------------------------------------------------------------
C
C  Skip NFROM-1 lines.
      REWIND NU
      DO 10,I=1,NFROM-1
10       READ (NU,'(A)',END=30) CH
C
C  Copy.
      DO 20,I=NFROM,NTO
         READ (NU,'(a)',END=30) LINE
         LL = LENGTH(LINE)
         IF (LL.EQ.0) THEN
            IF (STRO) WRITE (IUOUT,*)
         ELSE
            IF (STRO) WRITE (IUOUT,'($(FC)A)') LINE(1:LL)
         END IF
20    CONTINUE
      RETURN
C
C  End-of-File in the panel description file.
30    CONTINUE
      IF (STRO) WRITE (IUOUT,40)
C
40    FORMAT (/$(FC)'A serious error has occurred:'
     &        /$(FC)'EOF in the panel description file while ',
     &        'copying help.')
      END
C  ---------------------------------------------------------------------
      SUBROUTINE HESCOM ( H, MAXW, MERR )
C  ---------------------------------------------------------------------
C
C  Description:
C    Implements the HESSIAN command:
C    Syntax: HESSIAN (Has panel)
C
C  Input arguments:
C    MAXW       Number of available storage positions in array H.
C
C  Output arguments:
C    MERR       Error indicator.
C
C  Input / Output arguments:
C    V          The 2-dimensional Merlin work array.
C
C  ---------------------------------------------------------------------
      IMPLICIT $(TYPE) (A-H,O-Z)
C  Arguments:
      DIMENSION H(NOD,*)
C  ---------------------------------------------------------------------
C  Common blocks:
      PARAMETER ( MXV = $(MXV) )
      COMMON /PARAMS/ POINT(MXV), VAL
      COMMON /DETAIL/ IXAT(MXV), MARG(MXV)
      COMMON /TOTCAL/ ITNOCS, IPARTC, ITDER, IPDER, ITHESS, IPHESS,
     &                ITJAC, IPJAC
      COMMON /MODE3/ MODEV, IFFORM, INMCL
      COMMON /ACCURA/ COMACC
      COMMON /DIMEN/ NOD, NOF
      COMMON /UPDATE/ IGRADU, IHESSU
      COMMON /RWORK/ XT1(MXV), XT2(MXV), XT3(MXV), XT4(MXV), XT5(MXV),
     &               XT6(MXV), XT7(MXV+1)
      LOGICAL STRO, WEAK
      COMMON /PRILOG/ STRO, WEAK
      COMMON /UNITS/ IUINP, IUOUT
C  ---------------------------------------------------------------------
C  Local variables:
      PARAMETER ( MXFILE = 80 )
      CHARACTER*(MXFILE) FILE
      CHARACTER*10 STR
      CHARACTER*30 HSTR
      CHARACTER WHAT, USE, HM1, HM2, CTMP
C  ---------------------------------------------------------------------
C
C  Is there enough storage ?
      NEED = NOD**2
      CALL VSIM('N**2',MAXW,NEED,MERR)
      IF (MERR.NE.0) RETURN
C
C  Present the panel.
      CALL SETPI('CANCEL',1,IERR)
      CALL CHANGE(MERR)
      IF (MERR.NE.0) RETURN
      CALL GETPI('CANCEL',ICANC)
      IF (ICANC.EQ.0 .OR. ITNOCS.EQ.0 .OR. MODEV.EQ.0) RETURN
C
C  What should we do ?
      CALL GETPS('DO',WHAT,LEW)
      IF (ISCOMP(WHAT,'C').NE.0) THEN
C
C  Calculate the Hessian matrix.
         CALL GETPS('USE',USE,LEU)
         CALL SECDER(H,NOD,NOD,POINT,VAL,USE,XT2,XT3,NC)
         IF (ISCOMP(USE,'F').NE.0) THEN
            IF (STRO) THEN
               CALL I2STR(NC,STR,LES)
               WRITE (IUOUT,20) STR(1:LES)
            END IF
         ELSE IF (ISCOMP(USE,'G').NE.0) THEN
            IF (STRO) THEN
               CALL I2STR(NC,STR,LES)
               WRITE (IUOUT,30) STR(1:LES)
            END IF
         ELSE IF (ISCOMP(USE,'A').NE.0) THEN
            IF (STRO) WRITE (IUOUT,40)
         END IF
C
C  Fill in upper part of the Hessian.
         DO 10,I=1,NOD
            DO 11,J=I+1,NOD
11             H(I,J) = H(J,I)
10       CONTINUE
C
C  Shrink the Hessian matrix. Rows and columns corresponding to fixed
C  variables, are removed.
         CALL SHRINK(H,NOD,NFREE)
         IHESSU = 0
      ELSE IF (ISCOMP(WHAT,'D').NE.0) THEN
C
C  Decompose the Hessian matrix.
         CALL FREEVA(NFREE)
         CALL HPDD(H,XT2,NOD,NFREE,COMACC,ISPD)
C
C  Fill in lower part.
         DO 160,I=1,NFREE
            DO 165,J=1,I-1
165            H(I,J) = H(J,I)
160      CONTINUE
         IF (STRO) THEN
            WRITE (IUOUT,50)
            IF (ISPD.EQ.0) THEN
               WRITE (IUOUT,60)
            ELSE
               WRITE (IUOUT,70)
            END IF
         END IF
         IHESSU = 1
      ELSE IF (ISCOMP(WHAT,'M').NE.0) THEN
C
C  Compose the Hessian matrix from its Choleski decomposition.
         CALL LLCOMP(H,NOD,NOD)
         IF (STRO) WRITE (IUOUT,240)
      ELSE IF (ISCOMP(WHAT,'P').NE.0) THEN
C
C  Compare two different Hessian calculation modes.
         CALL GETPS('MODE1',HM1,LE1)
         CALL GETPS('MODE2',HM2,LE1)
         IF (ISCOMP(HM1,HM2).NE.0) THEN
            IF (STRO) WRITE (IUOUT,120)
            MERR = -2
            RETURN
         END IF
C
C  A Hessian calculation, using gradient values, should be performed
C  first, since it uses the whole array H.
         IF (ISCOMP(HM2,'G').NE.0) THEN
            CTMP = HM1
            HM1 = HM2
            HM2 = CTMP
         END IF
C
C  Calculate the Hessian using the first mode.
         CALL SECDER(H,NOD,NOD,POINT,VAL,HM1,XT2,XT3,NC)
C
C  Move lower to upper part.
         DO 130,I=1,NOD
            XT1(I) = H(I,I)
            DO 135,J=I+1,NOD
135            H(I,J) = H(J,I)
130      CONTINUE
C
C  Calculate the Hessian using the second mode.
         CALL SECDER(H,NOD,NOD,POINT,VAL,HM2,XT2,XT3,NC)
C
C  Compare.
         DIFMAX = 0.
         RMS = 0.
         ONE = 1.
         NCOMP = 0
         DO 140,I=1,NOD
            IF (IXAT(I).EQ.0) GOTO 140
            HES1 = H(I,I)
            HES2 = XT1(I)
            DIFF = ABS(HES1-HES2)/MAX(ONE,MIN(ABS(HES1),ABS(HES2)))
            RMS = RMS+DIFF**2
            IF (DIFF.GT.DIFMAX) DIFMAX = DIFF
            NCOMP = NCOMP+1
            DO 170,J=1,I-1
               IF (IXAT(J).EQ.0) GOTO 170
               HES1 = H(I,J)
               HES2 = H(J,I)
               DIFF = ABS(HES1-HES2)/
     &                         MAX(ONE,MIN(ABS(HES1),ABS(HES2)))
               RMS = RMS+DIFF**2
               IF (DIFF.GT.DIFMAX) DIFMAX = DIFF
               NCOMP = NCOMP+1
170         CONTINUE
140      CONTINUE
         IF (NCOMP.NE.0) THEN
            RMS = SQRT(RMS/NCOMP)
            IF (STRO) WRITE (IUOUT,150) RMS, DIFMAX
         END IF
         CALL SHRINK(H,NOD,NFREE)
         IHESSU = 0
      ELSE IF (ISCOMP(WHAT,'W').NE.0) THEN
C
C  Write the Hessian in a file.
         CALL GETPS('FILE',FILE,LEF)
         CALL OPENGF(FILE,0,NU)
         IF (NU.EQ.-1) THEN
            MERR = -2
            RETURN
         END IF
         CALL FREEVA(NFREE)
         DO 100,J=1,NFREE
            DO 105,I=1,NFREE
               HES = H(I,J)
               IF (I.LT.J) HES = H(J,I)
               CALL R2RTF(HES,HSTR,LEH)
               WRITE (NU,'(A)') HSTR(1:LEH)
105         CONTINUE
100      CONTINUE
         CLOSE (NU)
         IF (STRO) THEN
            CALL I2STR(NFREE,STR,LES)
            WRITE (IUOUT,110) FILE(1:LEF), STR(1:LES), STR(1:LES)
         END IF
      ELSE IF (ISCOMP(WHAT,'R').NE.0) THEN
C
C  Read the Hessian from a file.
         CALL GETPS('FILE',FILE,LEF)
         CALL OPENGF(FILE,0,NU)
         IF (NU.EQ.-1) THEN
            MERR = -2
            RETURN
         END IF
         CALL FREEVA(NFREE)
         NNUR = 0
         DO 200,J=1,NFREE
            DO 205,I=1,NFREE
               READ (NU,*,END=220) HES
               NNUR = NNUR+1
               H(J,I) = HES
205         CONTINUE
200      CONTINUE
         CLOSE (NU)
         IF (STRO) THEN
            CALL I2STR(NFREE,STR,LES)
            WRITE (IUOUT,230) STR(1:LES), STR(1:LES), FILE(1:LEF)
         END IF
         IHESSU = 0
      END IF
      RETURN
C
C  EOF while reading the Hessian.
220   CONTINUE
      IF (WEAK) THEN
         CALL I2STR(NNUR,STR,LES)
         WRITE (IUOUT,210) FILE(1:LEF), STR(1:LES)
      END IF
      MERR = -2
C
20    FORMAT (/$(FC)'The Hessian matrix has been calculated. ',A,
     &        ' function calls were used.')
30    FORMAT (/$(FC)'The Hessian matrix has been calculated. ',A,
     &        ' gradient calls were used.')
40    FORMAT (/$(FC)'The Hessian matrix has been calculated. ',
     &        '1 Hessian call was used.')
50    FORMAT (/$(FC)'The Choleski factorization of the Hessian ',
     &        'matrix has been calculated.')
60    FORMAT (/$(FC)'The Hessian matrix was not positive definite.')
70    FORMAT (/$(FC)'The Hessian matrix was positive definite.')
110   FORMAT (/$(FC)'File "',A,'" contains the ',A,' x ',A,
     &        ' Hessian matrix.')
120   FORMAT (/$(FC)'You must enter two different modes.')
150   FORMAT (/$(FC)'RMS relative error : ',1PG12.5/
     &         $(FC)'Max. relative error: ',1PG12.5)
210   FORMAT (/$(FC)'Error: File "',A,'" contains only ',A,
     &        ' real numbers.')
230   FORMAT (/$(FC)'The ',A,' x ',A,' Hessian matrix has been ',
     &        'read from file "',A,'".')
240   FORMAT (/$(FC)'The Hessian matrix has been composed from its ',
     &        'Choleski factorization.')
      END
C  ---------------------------------------------------------------------
      SUBROUTINE HPDD ( A, ADIAG, LD, N, COMACC, ISPD )
C  ---------------------------------------------------------------------
C
C  Description:
C    Subroutine HPDD finds a non-negative constant m, such that
C    A + m*I is safely positive definite, where m=0 if A is already
C    positive definite. It then sets A = A+m*I and calculates the
C    Choleski decomposition of the new A.
C    Actualy only the lower triangular part of A is referenced,
C    while the Cholesky factor L[t] is stored in the upper
C    triangular part of A.
C    For details on the implementation see:
C
C      Dennis J.E & Schnabel R.B.
C      Numerical Methods for Unconstrained Optimization and Nonlinear
C      Equations, Prentice-Hall, 1983.
C
C  Input arguments:
C    LD         Leading dimension of matrix A.
C    N          Dimensionality of matrix A.
C    COMACC     Machine accuracy.
C
C  Output arguments:
C    ADIAG      Diagonal elements of the matrix A+m*I.
C    ISPD       Indicates whether the original matrix was positive
C               definite:
C               ISPD = 0 -> The original matrix A was not positive
C                           definite.
C               ISPD = 1 -> The original matrix was positive definite.
C
C  Input / Output arguments:
C    A          On input an NxN symmetrix matrix. Only the diagonal and
C               lower triangular elements are referenced.
C               On output the diagonal and upper triangular part
C               contains the Choleski factor L[t].
C
C  ---------------------------------------------------------------------
      IMPLICIT $(TYPE) (A-H,O-Z)
C  Arguments:
      DIMENSION A(LD,N), ADIAG(N)
C  ---------------------------------------------------------------------
C
      ZERO = 0.
      ONE = 1.
      COMTOL = SQRT(COMACC)
C
      DO 30,I=1,N
30       ADIAG(I) = A(I,I)
C
      DMA = ADIAG(1)
      DMI = ADIAG(1)
      DO 10,I=2,N
         AII = ADIAG(I)
         DMA = MAX(DMA,AII)
         DMI = MIN(DMI,AII)
10    CONTINUE
      PDM = MAX(ZERO,DMA)
      IF (DMI.LE.COMTOL*PDM) THEN
         XMU = 2*(PDM-DMI)*COMTOL-DMI
         DMA = DMA+XMU
      ELSE
         XMU = 0.
      END IF
      OFFMAX = 0.
      DO 80,I=1,N
         DO 85,J=1,I-1
85          OFFMAX = MAX(OFFMAX,ABS(A(J,I)))
80    CONTINUE
      CON = ONE+2*COMTOL
      IF (OFFMAX*CON.GT.DMA) THEN
         XMU = XMU-DMA+OFFMAX*CON
         DMA = DMA*CON
      END IF
      IF (DMA.EQ.0.) THEN
         XMU = 1.
         DMA = 1.
      END IF
      ISPD = 1
      IF (XMU.GT.0) THEN
         ISPD = 0
         DO 20,I=1,N
20          ADIAG(I) = ADIAG(I)+XMU
      END IF
      OFFMAX = SQRT(MAX(DMA,OFFMAX/N))
C
      CALL LLDEC(A,ADIAG,LD,N,OFFMAX,ADDMAX,COMACC)
      IF (ADDMAX.GT.0.) THEN
         EMA = ADIAG(1)
         EMI = ADIAG(1)
         DO 40,I=1,N
            SOFF = 0.
            DO 50,J=1,I-1
50             SOFF = SOFF+ABS(A(I,J))
            DO 55,J=I+1,N
55             SOFF = SOFF+ABS(A(J,I))
            EMA = MAX(EMA,ADIAG(I)+SOFF)
            EMI = MIN(EMI,ADIAG(I)-SOFF)
40       CONTINUE
C
         SDD = (EMA-EMI)*COMTOL-EMI
         SDD = MAX(ZERO,SDD)
         XMU = MIN(ADDMAX,SDD)
         DO 60,I=1,N
60          ADIAG(I) = ADIAG(I)+XMU
         CALL LLDEC(A,ADIAG,LD,N,ZERO,ADDMAX,COMACC)
         ISPD = 0
      END IF
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE LLDEC ( A, ADIAG, LD, N, OFLMAX, ADDMAX, COMACC )
C  ---------------------------------------------------------------------
C
C  Description:
C    Calculates the perturbed Choleski decomposition of matrix A.
C    It actualy decomposes A+D, where D is a diagonal matrix, with
C    non-negative elements, large enough to allow the decomposition
C    to proceed. For details on the algorithm, see:
C
C      Dennis J.E & Schnabel R.B.
C      Numerical Methods for Unconstrained Optimization and Nonlinear
C      Equations, Prentice-Hall, 1983.
C
C      Gill P.E., Murray W. and Wright M.H.,
C      Practical Optimization, Academic Press, 1989.
C
C  Input arguments:
C    ADIAG      Diagonal elements of A.
C    LD         Leading dimension of matrix A.
C    N          Dimensionality of matrix A.
C    OFLMAX     On output, all elements L(i,j) of the Choleski
C               factor L, are less than, or equal to OFLMAX.
C    COMACC     Machine accuracy.
C
C  Output arguments:
C    ADDMAX     The maximum element, added to the diagonal of A.
C
C  Input / Output arguments:
C    A          On input the lower triangular part of A (excluding the
C               diagonal emements, which are input on ADIAG), contains
C               the matrix to be decomposed.
C               On output, the input part is intact, while the upper
C               triangular part plus the diagonal elements of A, contain
C               the Choleski factor L[t].
C
C  ---------------------------------------------------------------------
      IMPLICIT $(TYPE) (A-H,O-Z)
C  Arguments:
      DIMENSION A(LD,N), ADIAG(N)
C  ---------------------------------------------------------------------
C
      AMINL = SQRT(SQRT(COMACC))
      AMINL2 = 0.
      IF (OFLMAX.EQ.0.) THEN
         OFLMAX = 0.
         DO 50,I=1,N
50          OFLMAX = MAX(OFLMAX,ABS(ADIAG(I)))
         AMINL2 = SQRT(COMACC)*OFLMAX
      END IF
      ADDMAX = 0.
C
      DO 8,I=1,N
         DO 9,J=1,I
9           A(J,I) = 0.
8     CONTINUE
C
      DO 10,J=1,N
         SUM = 0.
         DO 20,I=1,J-1
20          SUM = SUM+A(I,J)**2
         A(J,J) = ADIAG(J)-SUM
         AMINJJ = 0.
         DO 30,I=J+1,N
            SUM = 0.
            DO 40,K=1,J-1
40             SUM = SUM+A(K,I)*A(K,J)
            A(J,I) = A(I,J)-SUM
            AMINJJ = MAX(AMINJJ,ABS(A(J,I)))
30       CONTINUE
         AMINJJ = MAX(AMINJJ/OFLMAX,AMINL)
         IF (A(J,J).GT.AMINJJ**2) THEN
            A(J,J) = SQRT(A(J,J))
         ELSE
            IF (AMINJJ.LT.AMINL2) AMINJJ = AMINL2
            ADDMAX = MAX(ADDMAX,AMINJJ**2-A(J,J))
            A(J,J) = AMINJJ
         END IF
         DO 60,I=J+1,N
60          A(J,I) = A(J,I)/A(J,J)
10    CONTINUE
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE HIDE ( MERR )
C  ---------------------------------------------------------------------
C
C  Description:
C    Implements the HIDEOUT command.
C    Syntax: HIDEOUT file [ append ]
C
C  Output arguments:
C    MERR       Error indicator.
C
C  ---------------------------------------------------------------------
C  Common blocks:
      LOGICAL STRO, WEAK
      COMMON /PRILOG/ STRO, WEAK
      COMMON /UNITS/ IUINP, IUOUT
C  ---------------------------------------------------------------------
C  Local variables:
      PARAMETER ( MXFILE = 80 )
      CHARACTER*(MXFILE) FILE
      CHARACTER*80 HMODE
      LOGICAL OPE
C  ---------------------------------------------------------------------
C
      MERR = 0
      CALL ARGNO(NA)
      IF (NA.EQ.0) THEN
         IF (STRO) WRITE (IUOUT,30)
30       FORMAT (/$(FC)'Enter file name for output disposal:')
         CALL INNOCR(FILE,.FALSE.,MERR)
         IF (MERR.NE.0) RETURN
         LEF = LENGTH(FILE)
         HMODE=' '
      ELSE IF (NA.EQ.1) THEN
         CALL CMDARG(1,FILE,LEF)
         HMODE=' '
      ELSE IF (NA.EQ.2) THEN
         CALL CMDARG(1,FILE,LEF)
         CALL CMDARG(2,HMODE,LEH)
      ELSE
         CALL ERRNAR
         MERR = -2
         RETURN
      END IF
C
C  IREW is set to 0 if output has to be appended.
      IF (HMODE.EQ.' ') THEN
         IREW = 1
      ELSE
         IF (ISCOMP(HMODE,'APPEND').EQ.0) THEN
            IF (WEAK) WRITE (IUOUT,10)
            MERR = -2
            RETURN
         END IF
         IREW = 0
      END IF
C
      INQUIRE (FILE=FILE,OPENED=OPE,NUMBER=IU,IOSTAT=IOE)
      IF (IOE.NE.0) THEN
         CALL ERRINQ(FILE,IOE)
         MERR = -2
         RETURN
      END IF
      IF (OPE) THEN
         CALL ERRBUS(FILE,IU)
         MERR = -2
         RETURN
      END IF
C
C  Allocate unit for the file.
      NU = NUNIT()
      IF (IREW.NE.0) THEN
         OPEN (NU,FILE=FILE,IOSTAT=IOE)
      ELSE
         CALL OPAPPE(NU,FILE,LEF,IOE)
      END IF
      IF (IOE.NE.0) THEN
         CALL ERROPE(FILE,IOE)
         MERR = -2
         RETURN
      END IF
      IF (STRO) WRITE (IUOUT,20) FILE(1:LEF)
C
C  Increment the output level.
      CALL INCOLE(NU,4,IERR)
      IF (IERR.NE.0) THEN
         CLOSE (NU)
         MERR = -2
      END IF
C
10    FORMAT (/$(FC)'Incorrect argument. Only "append" can follow.')
20    FORMAT (/$(FC)'Output is now beeing redirected to file "',A,'".')
      END
C  ---------------------------------------------------------------------
      SUBROUTINE HISTOR ( IOPT, MERR )
C  ---------------------------------------------------------------------
C
C  Description:
C    Implements the HISTORY command.
C    Syntax: HISTORY (Has panel)
C
C  Input arguments:
C    IOPT       Determines whether the HISTORY panel should be
C               presented:
C               IOPT = 0 -> Present the HISTORY panel, prompt the user
C                           for changes and proceed accordingly.
C                           This value of IOPT is used when the user
C                           enter the HISTORY command.
C               IOPT = 1 -> Do not present the HISTORY panel. Use the
C                           default panel values to (possibly) start
C                           the history mechanism. This value of IOPT
C                           is used at Merlin startup, to automaticaly
C                           start history.
C
C  Output arguments:
C    MERR       Error indicator.
C
C  ---------------------------------------------------------------------
C  Common blocks:
      LOGICAL STRO, WEAK
      COMMON /PRILOG/ STRO, WEAK
      COMMON /UNITS/ IUINP, IUOUT
      COMMON /DIMEN/ NOD, NOF
      COMMON /LOGIT/ IULOG
C  ---------------------------------------------------------------------
C  Local variables:
      PARAMETER ( MXFILE = 80 )
      CHARACTER*(MXFILE) FILE
      CHARACTER*10 STR
      CHARACTER*3 ONOFF
      CHARACTER OLD
C  ---------------------------------------------------------------------
C
      MERR = 0
      IF (IOPT.EQ.0) THEN
C  IOPT=0 signifies a normal HISTORY command. Present the panel.
         CALL SETPI('CANCEL',1,IERR)
         CALL CHANGE(MERR)
         IF (MERR.NE.0) RETURN
         CALL GETPI('CANCEL',ICANC)
         IF (ICANC.EQ.0) RETURN
      END IF
      CALL GETPPS('HISTORY','STATUS',ONOFF,LES)
C
      IF (IULOG.EQ.-1) THEN
C
C  The history mechanism is not active.
         IF (ISCOMP(ONOFF,'ON').NE.0) THEN
C  Get panel parameters.
            CALL GETPPS('HISTORY','FILE',FILE,LEF)
            CALL GETPPS('HISTORY','OLD',OLD,LEO)
C  Open the hisrory file.
            CALL INILOG(FILE,IULOG,OLD)
C  Write initial messages.
            IF (IULOG.NE.-1) THEN
               CALL LOGGER('%  Merlin log started.',22)
               CALL I2STR(NOD,STR,LES)
               CALL LOGGER('%  Variables: '//STR(1:LES),14+LES)
               CALL I2STR(NOF,STR,LES)
               CALL LOGGER('%  Terms    : '//STR(1:LES),14+LES)
               IF (STRO) WRITE (IUOUT,10) FILE(1:LEF)
            ELSE
               MERR = -2
            END IF
         END IF
      ELSE
C
C  The history mechanism is already active.
         CALL U2NAME(IULOG,'history',FILE,LEF)
         IF (ISCOMP(ONOFF,'ON').NE.0) THEN
            IF (STRO) THEN
               IF (LEF.EQ.0) THEN
                  WRITE (IUOUT,25)
               ELSE
                  WRITE (IUOUT,20) FILE(1:LEF)
               END IF
            END IF
         ELSE
C  History is active, we need to stop it however.
            CLOSE (IULOG)
            IULOG = -1
            IF (STRO) THEN
               IF (LEF.EQ.0) THEN
                  WRITE (IUOUT,35)
               ELSE
                  WRITE (IUOUT,30) FILE(1:LEF)
               END IF
            END IF
         END IF
      END IF
C
10    FORMAT (/$(FC)'Logging to file "',A,'" started.')
20    FORMAT (/$(FC)'Logging is already active. Log file is "',A,'".')
25    FORMAT (/$(FC)'Logging is already active.')
30    FORMAT (/$(FC)'Logging to file "',A,'" has been stopped.')
35    FORMAT (/$(FC)'Logging has been stopped.')
      END
C  ---------------------------------------------------------------------
      SUBROUTINE INILOG ( FILE, NU, OLD )
C  ---------------------------------------------------------------------
C
C  Description:
C    Part of the HISTORY command. Opens the history file as specified
C    by FILE with unit NU.
C
C  Input arguments:
C    FILE       The history file name.
C    OLD        If a file with the same name exists alerady, it is
C               handled according to OLD:
C               OLD = 'A' -> Append to the file.
C               OLD = 'D' -> Delete the file. (Actually overwrite its
C                            contents.)
C
C  Output arguments:
C    NU         Unit number of the history file. On any error NU = -1
C               is returned.
C
C  ---------------------------------------------------------------------
C  Arguments:
      CHARACTER*(*) FILE, OLD
C  ---------------------------------------------------------------------
C  Common blocks:
      LOGICAL STRO, WEAK
      COMMON /PRILOG/ STRO, WEAK
      COMMON /UNITS/ IUINP, IUOUT
C  ---------------------------------------------------------------------
C  Local variables:
      LOGICAL EXI, OPE
C  ---------------------------------------------------------------------
C
      NU = -1
      LEF = LENGTH(FILE)
      IF (LEF.EQ.0) THEN
         IF (STRO) WRITE (IUOUT,10)
         RETURN
      END IF
C
C  Get file info.
      INQUIRE (FILE=FILE,EXIST=EXI,OPENED=OPE,NUMBER=NUM,IOSTAT=IERR)
      IF (IERR.NE.0) THEN
         CALL ERRINQ(FILE,IERR)
         RETURN
      END IF
C
C  Cannot use an open file.
      IF (OPE) THEN
         CALL ERRBUS(FILE,NUM)
         RETURN
      END IF
C
C  Allocate unit for the history file.
      NU = NUNIT()
      IF (NU.EQ.-1) THEN
         CALL ERRUAL(FILE)
         RETURN
      END IF
C
C  Open the file.
      IF (EXI .AND. ISCOMP(OLD,'A').NE.0) THEN
         CALL OPAPPE(NU,FILE,LEF,IERR)
      ELSE
         OPEN (NU,FILE=FILE,IOSTAT=IERR)
      END IF
C
      IF (IERR.NE.0) THEN
         CALL ERROPE(FILE,IERR)
         NU = -1
         RETURN
      END IF
C
10    FORMAT (/$(FC)'You must specify a history file in order ',
     &        'to start the Melrin log.')
      END
C  ---------------------------------------------------------------------
      SUBROUTINE LOGGER ( LINE, LL )
C  ---------------------------------------------------------------------
C
C  Description:
C    Subroutine LOGGER will add the LINE with length LL in the Merlin
C    history file. Only top level commands/input is added; commands
C    inside macros and MCL instuctions are ignored.
C    If the history mechanism is not active, this subroutine just
C    returns.
C
C  Input arguments:
C    LINE       Line to be added to the history file.
C    LL         Effective length of the input LINE.
C
C  ---------------------------------------------------------------------
C  Arguments:
      CHARACTER*(*) LINE
C  ---------------------------------------------------------------------
C  Common blocks:
      PARAMETER ( MAXINL = 20 )
      INTEGER BACKIN(MAXINL), WHATIN(MAXINL), REDIN(MAXINL)
      COMMON /INNER1/ LEVIN, BACKIN, WHATIN, REDIN
      COMMON /LOGIT/ IULOG
      COMMON /MODE3/ MODEV, IFFORM, INMCL
C  ---------------------------------------------------------------------
C
      IF (IULOG.EQ.-1) RETURN
      IF (INMCL.NE.0) RETURN
      IF (LEVIN.NE.0) RETURN
C
      IF (LL.EQ.0) THEN
         WRITE (IULOG,*)
      ELSE
         WRITE (IULOG,'(A)') LINE(1:LL)
      END IF
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE INITCO ( MERR )
C  ---------------------------------------------------------------------
C
C  Description:
C    Implements the INIT command.
C    Syntax: INIT (Has panel)
C
C  Output arguments:
C    MERR       Error indicator.
C  ---------------------------------------------------------------------
      IMPLICIT $(TYPE) (A-H,O-Z)
C  Common blocks:
      PARAMETER ( MXV = $(MXV) )
      COMMON /PARAMS/ POINT(MXV), VAL
      COMMON /DETAIL/ IXAT(MXV), MARG(MXV)
      COMMON /ROSTEP/ STEP(MXV)
      COMMON /INDEFI/ INDEF(MXV)
      COMMON /DIMEN/ NOD, NOF
      LOGICAL STRO, WEAK
      COMMON /PRILOG/ STRO, WEAK
      COMMON /UNITS/ IUINP, IUOUT
      COMMON /BIGCOM/ BIGNUM, BLIM, DWARF, DWALOG
      COMMON /IWORK/ INDE(MXV)
      COMMON /MODE3/ MODEV, IFFORM, INMCL
      COMMON /RWORK/ XT1(MXV), XT2(MXV), XT3(MXV), XT4(MXV), XT5(MXV),
     &               XT6(MXV), XT7(MXV+1)
C  ---------------------------------------------------------------------
C  Local variables:
      PARAMETER ( MXFILE = 80 )
      CHARACTER*(MXFILE) FILE
      CHARACTER*80 FORM, USEF
      CHARACTER*10 NAME, STR1, STR2
      CHARACTER*25 MSG
      CHARACTER CH, WHAT
      LOGICAL EXI, OPE, FREE
C  ---------------------------------------------------------------------
C
      SAVE IFIRST
      DATA IFIRST / 0 /
C
C  On the first entry set the TO keyword to the number of variables.
      IF (IFIRST.EQ.0) THEN
         CALL SETPI('TO',NOD,IERR)
         IFIRST = 1
      END IF
C
C  Present the panel and process any user changes.
      CALL SETPI('CANCEL',1,IERR)
      CALL CHANGE(MERR)
      IF (MERR.NE.0) RETURN
      CALL GETPI('CANCEL',ICANC)
      IF (ICANC.EQ.0) RETURN
C
C  Get panel parameters.
      CALL GETPS('FILE',FILE,LEF)
      CALL GETPS('WHAT',WHAT,LEW)
      CALL GETPI('NPL',NPL)
      CALL GETPI('SKIP',ISKIP)
      CALL GETPI('FROM',IFROM)
      CALL GETPI('TO',ITO)
      CALL GETPS('FORMAT',USEF,LER)
C
C  We need a file name.
      LEF = LENGTH(FILE)
      IF (LEF.EQ.0) THEN
         IF (WEAK) WRITE (IUOUT,200)
         MERR = -2
         RETURN
      END IF
C
C  FROM should be <= NOD.
      IF (IFROM.GT.NOD) THEN
         CALL I2STR(NOD,STR1,LES1)
         IF (WEAK) WRITE (IUOUT,240) STR1(1:LES1)
         MERR = -2
         RETURN
      END IF
C
C  TO should be <= NOD.
      IF (ITO.GT.NOD) THEN
         CALL I2STR(NOD,STR1,LES1)
         IF (WEAK) WRITE (IUOUT,245) STR1(1:LES1)
         MERR = -2
         RETURN
      END IF
C
C  FROM should be <= TO.
      IF (IFROM.GT.ITO) THEN
         IF (WEAK) WRITE (IUOUT,220)
         MERR = -2
         RETURN
      END IF
C
C  Determine whether a free format is to be used.
      FREE = ISCOMP(USEF,'*').NE.0 .OR. NPL.EQ.0
      IF (.NOT.FREE) THEN
         LER = LENGTH(USEF)
         FORM = '('//USEF(1:LER)//')'
         LER = LER+2
      END IF
C
C  Get file status.
      INQUIRE (FILE=FILE,EXIST=EXI,OPENED=OPE,NUMBER=NUM,IOSTAT=IOE)
      IF (IOE.NE.0) THEN
         CALL ERRINQ(FILE,IOE)
         MERR = -2
         RETURN
      END IF
C
C  Cannot init from an non-existent file.
      IF (.NOT.EXI) THEN
         MERR = -2
         CALL ERREXI(FILE)
         RETURN
      END IF
C
C  Cannot init from an open file.
      IF (OPE) THEN
         MERR = -2
         CALL ERRBUS(FILE,NUM)
         RETURN
      END IF
C
C  Allocate unit for the file.
      NU = NUNIT()
      IF (NU.EQ.-1) THEN
         MERR = -2
         CALL ERRUAL(FILE)
         RETURN
      END IF
C
C  Open the file.
      OPEN (NU,FILE=FILE,IOSTAT=IOE)
      IF (IOE.NE.0) THEN
         MERR = -2
         CALL ERROPE(FILE,IOE)
         RETURN
      END IF
C
      REWIND NU
C
C  Skip lines.
      DO 30,IS=1,ISKIP
30       READ (NU,'(A)',END=40) CH
C
C  Begin initialization. For all possible WHAT values (except for
C  WHAT='F' and WHAT='N') read in the work array XT1 using FORMAT,
C  NPL etc. Then copy the work array to the proper destination.
      NTOT = ITO-IFROM+1
      IF (ISCOMP(WHAT,'N').NE.0) THEN
C  Initialize names. Names are always read one per line.
         CALL BACKUP('D')
         DO 10,I=1,NTOT
            READ (NU,'(A)',END=20) NAME
            CALL SETNAM(IFROM+I-1,NAME,IERR)
10       CONTINUE
         MSG = 'name(s)'
      ELSE IF (ISCOMP(WHAT,'F').NE.0) THEN
C  Initialize fix statuses. Since these are integer variables, we must
C  duplicate the code here, using an integer work array instead of XT1.
         CALL BACKUP('D')
         IF (NPL.EQ.0) THEN
            READ (NU,*,END=50,IOSTAT=IOE) (INDE(I),I=IFROM,ITO)
            IF (IOE.NE.0) GOTO 55
         ELSE
            NLI = NTOT/NPL
            DO 80,I=1,NLI
               IST = (I-1)*NPL+IFROM
               IEND = IST+NPL-1
               IF (FREE) THEN
                  READ (NU,*,END=50,IOSTAT=IOE) (INDE(J),J=IST,IEND)
               ELSE
                  READ (NU,FORM,END=50,IOSTAT=IOE)
     &                                (INDE(J),J=IST,IEND)
               END IF
               IF (IOE.NE.0) GOTO 55
80          CONTINUE
            NREM = MOD(NTOT,NPL)
            IF (NREM.NE.0) THEN
               IST = NLI*NPL+IFROM
               IEND = ITO
               IF (FREE) THEN
                  READ (NU,*,END=50,IOSTAT=IOE) (INDE(J),J=IST,IEND)
               ELSE
                  READ (NU,FORM,END=50,IOSTAT=IOE)
     &                                (INDE(J),J=IST,IEND)
               END IF
               IF (IOE.NE.0) GOTO 55
            END IF
         END IF
         DO 100,I=IFROM,ITO
            IX = INDE(I)
            IF (IX.EQ.0 .OR. IX.EQ.1) THEN
               IXAT(I) = IX
            ELSE
               IF (WEAK) WRITE (IUOUT,210) FILE(1:LEF)
               CLOSE (NU)
               MERR = -2
               RETURN
            END IF
100      CONTINUE
         MSG = 'fix status(es)'
      ELSE
         IF (NPL.EQ.0) THEN
            READ (NU,*,END=50,IOSTAT=IOE) (XT1(I),I=IFROM,ITO)
            IF (IOE.NE.0) GOTO 55
         ELSE
C
C  Read a number of lines containing NPL numbers each.
            NLI = NTOT/NPL
            DO 85,I=1,NLI
               IST = (I-1)*NPL+IFROM
               IEND = IST+NPL-1
               IF (FREE) THEN
                  READ (NU,*,END=50,IOSTAT=IOE) (XT1(J),J=IST,IEND)
               ELSE
                  READ (NU,FORM,END=50,IOSTAT=IOE)
     &                                (XT1(J),J=IST,IEND)
               END IF
               IF (IOE.NE.0) GOTO 55
85          CONTINUE
C
C  Read the remainder. (A full line with NPL numbers is not possible.)
            NREM = MOD(NTOT,NPL)
            IF (NREM.NE.0) THEN
               IST = NLI*NPL+IFROM
               IEND = ITO
               IF (FREE) THEN
                  READ (NU,*,END=50,IOSTAT=IOE) (XT1(J),J=IST,IEND)
               ELSE
                  READ (NU,FORM,END=50,IOSTAT=IOE)
     &                                (XT1(J),J=IST,IEND)
               END IF
               IF (IOE.NE.0) GOTO 55
            END IF
         END IF
C
C  The work array XT1 now contains the values to be used for
C  initialization.
         IF (ISCOMP(WHAT,'X').NE.0) THEN
C  Initialize the current point.
            CALL BACKUP('X')
            DO 120,I=IFROM,ITO
               POINT(I) = XT1(I)
               IXAT(I) = 1
               INDEF(I) = 1
120         CONTINUE
            CALL POIMAR(IFEV)
C  If POIMAR did not call ACSQ, do it here.
            IF (IFEV.EQ.0 .AND. MODEV.EQ.1) VAL = ACSQ(POINT,NOD)
            MSG = 'variable(s)'
         ELSE IF (ISCOMP(WHAT,'L').NE.0) THEN
C  Initialize left margins.
            CALL BACKUP('D')
            DO 130,I=IFROM,ITO
               XMA = XT1(I)
130            IF (XMA.GE.-BLIM) CALL SETLMA(I,XMA)
            CALL POIMAR(IFEV)
C  If POIMAR did not call ACSQ, do it here.
            IF (IFEV.EQ.0 .AND. MODEV.EQ.1) VAL = ACSQ(POINT,NOD)
            MSG = 'left margin(s)'
         ELSE IF (ISCOMP(WHAT,'R').NE.0) THEN
C  Initialize right margins.
            CALL BACKUP('D')
            DO 140,I=IFROM,ITO
               XMA = XT1(I)
140            IF (XMA.LE.BLIM) CALL SETRMA(I,XMA)
            CALL POIMAR(IFEV)
C  If POIMAR did not call ACSQ, do it here.
            IF (IFEV.EQ.0 .AND. MODEV.EQ.1) VAL = ACSQ(POINT,NOD)
            MSG = 'right margin(s)'
         ELSE IF (ISCOMP(WHAT,'S').NE.0) THEN
C  Initialize ROLL steps.
            DO 110,I=IFROM,ITO
110            STEP(I) = XT1(I)
            MSG = 'step(s)'
         ELSE
            CALL MABORT('Incorrect "what".','INITCO')
         END IF
      END IF
      CLOSE (NU)
C
C  Print informative message.
      LEM = LENGTH(MSG)
      IF (STRO) THEN
         CALL I2STR(NTOT,STR1,LES1)
         WRITE (IUOUT,150) STR1(1:LES1), MSG(1:LEM), FILE(1:LEF)
      END IF
      RETURN
C
C  End-of-File while reading names.
20    CONTINUE
      IF (I.EQ.1) THEN
         IF (WEAK) WRITE (IUOUT,160) FILE(1:LEF)
      ELSE
         CALL I2STR(I-1,STR1,LES1)
         IF (WEAK) WRITE (IUOUT,180) FILE(1:LEF), STR1(1:LES1)
      END IF
      CLOSE (NU)
      MERR = -2
      RETURN
C
C  End-of-File while skiping lines.
40    CONTINUE
      IF (IS.EQ.1) THEN
         IF (WEAK) WRITE (IUOUT,160) FILE(1:LEF)
      ELSE
         IF (WEAK) THEN
            CALL I2STR(ISKIP,STR1,LES1)
            CALL I2STR(IS-1,STR2,LES2)
            WRITE (IUOUT,170) STR1(1:LES1), FILE(1:LEF), STR2(1:LES2)
         END IF
      END IF
      CLOSE (NU)
      MERR = -2
      RETURN
C
C  End-of-File while reading.
50    CONTINUE
      IF (WEAK) WRITE (IUOUT,190) FILE(1:LEF)
      CLOSE (NU)
      MERR = -2
      RETURN
C
C  Some error while reading (Incorrect format, non-numeric data, etc.)
55    CONTINUE
      CALL ERRRD(FILE,IOE)
      CLOSE (NU)
      MERR = -2
      RETURN
C
150   FORMAT (/$(FC)A,' ',A,' have been initialized from file "',
     &        A,'".')
160   FORMAT (/$(FC)'File "',A,'" appears to be empty.')
170   FORMAT (/$(FC)'Cannot skip ',A,' lines. File "',A,
     &        '" only has ',A,' line(s).')
180   FORMAT (/$(FC)'Not enough names in file "',A,'". Only ',A,
     &        ' have been initialized.')
190   FORMAT (/$(FC)'Not enough data in file "',A,
     &        '". No initialization.')
200   FORMAT (/$(FC)'You must enter a filename.')
210   FORMAT (/$(FC)'Incorrect data in file "',A,'". Fix statuses ',
     &        'must be 0 or 1.')
220   FORMAT (/$(FC)'The values of the "FROM" and "TO" keywords ',
     &        'are incorrect.')
240   FORMAT (/$(FC)'The specified value for the FROM keyword ',
     &        'is incorrect.'
     &        /$(FC)'There are only ',A,' variables.')
245   FORMAT (/$(FC)'The specified value for the TO keyword ',
     &        'is incorrect.'
     &        /$(FC)'There are only ',A,' variables.')
      END
C  ---------------------------------------------------------------------
      SUBROUTINE INSPE ( MERR )
C  ---------------------------------------------------------------------
C
C  Description:
C    Implements the INSPECT command.
C    Syntax: INSPECT (Has panel)
C
C  Output arguments:
C    MERR       Error indicator.
C
C  ---------------------------------------------------------------------
      IMPLICIT $(TYPE) (A-H,O-Z)
C  Common blocks:
      PARAMETER ( MAXINL = 20 )
      INTEGER BACKIN(MAXINL), WHATIN(MAXINL), REDIN(MAXINL)
      COMMON /INNER1/ LEVIN, BACKIN, WHATIN, REDIN
      COMMON /DIMEN/ NOD, NOF
      LOGICAL STRO, WEAK
      COMMON /PRILOG/ STRO, WEAK
      COMMON /UNITS/ IUINP, IUOUT
C  ---------------------------------------------------------------------
C  Local variables:
      PARAMETER ( NPAGE = 20 )
      PARAMETER ( MXFILE = 80 )
      CHARACTER*(MXFILE) FILE
      CHARACTER*10 NAME, TYPE, STR
      CHARACTER CH
C  ---------------------------------------------------------------------
C
      MERR = 0
C
C  Set MCL-returned values.
C  REC will remain zero if an error accurs. On successful completion
C  however, it will contain the last record inspected. This is the
C  same value as the one set in the REC keyword in the PICK panel.
C  While in MCL or in a macro, REC will be set to the total
C  number of Merlin records in the file under inspection.
      CALL SETPI('REC',0,IERR)
C
C  Present the panel and process any user changes.
      CALL SETPI('CANCEL',1,IERR)
      CALL CHANGE(MERR)
      IF (MERR.NE.0) RETURN
      CALL GETPI('CANCEL',ICANC)
      IF (ICANC.EQ.0) RETURN
C
C  Get panel parameters.
      CALL GETPS('FILE',FILE,LEF)
      CALL GETPS('TYPE',TYPE,LET)
      CALL GETPI('DIR',IDIR)
      CALL GETPI('PRINT',IPRINT)
      CALL GETPI('FROM',ISTART)
      CALL GETPI('TO',IEND)
C
      LEF = LENGTH(FILE)
      IF (ISCOMP(TYPE,'TEXT').NE.0) THEN
         MODE = 1
      ELSE
         MODE = 2
      END IF
C
C  DIR determines the direction. IDIR=1 -> From top to bottom,
C  IDIR = -1 -> From bottom to top of file. If the FROM, TO keywords
C  do not agree, rearrange them.
      IF (IDIR.EQ.1) THEN
         IF (ISTART.NE.0 .AND. IEND.NE.0 .AND. IEND.LT.ISTART) THEN
            ITMP = ISTART
            ISTART = IEND
            IEND = ITMP
         END IF
      ELSE
         IF (ISTART.NE.0 .AND. IEND.NE.0 .AND. IEND.GT.ISTART) THEN
            ITMP = ISTART
            ISTART = IEND
            IEND = ITMP
         END IF
      END IF
C
C  Open the file.
      CALL OPENIT(FILE,TYPE,0,0,NU,INSREC)
      IF (NU.EQ.-1) THEN
         MERR = -2
         RETURN
      END IF
C
C  Position the file according to ISTART. A binary file does not need
C  positioning.
      IF (ISTART.NE.0) THEN
         NSR = ISTART
         IF (MODE.EQ.1) THEN
C  Read/skip ISTART-1 records.
            DO 40,I=1,(NOD+1)*(ISTART-1)
40             READ (NU,'(A)',END=50) CH
         ELSE
            IF (ISTART.GE.INSREC) GOTO 50
         END IF
      ELSE
         IF (IDIR.EQ.-1) THEN
            IF (MODE.EQ.1) THEN
C  For backward inspection, we need to be at EOF.
               CALL POSITC(NU,NREC,IEOF)
               IF (IEOF.NE.0) THEN
                  IF (WEAK) WRITE (IUOUT,130) FILE(1:LEF)
                  CLOSE (NU)
                  MERR = -2
                  RETURN
               ELSE IF (NREC.EQ.0) THEN
                  GOTO 50
               END IF
C  Get ready for the last record.
               DO 60,I=1,NOD+1
60                BACKSPACE NU
               NSR = NREC
            ELSE
               NSR = INSREC
            END IF
         ELSE
            NSR = 1
         END IF
      END IF
C
C  IEND is the last record to be inspected. Set it.
      IF (IDIR.EQ.-1 .AND. IEND.EQ.0) IEND = 1
C
C  File is now correctly positioned at end of record NSR-1.
C  Once more: Binary files do need positioning.
C
C  Read this record. IND is the current record that is being read.
      IND = NSR
C  NP1 is the counter used to stop scrolling every NPAGE records
C  if IPRINT = 1.
      NP1 = 0
      IF (IPRINT.NE.0 .AND. STRO) WRITE (IUOUT,*)
C
100   CONTINUE
C
C  Read and print the variables.
      DO 80,I=1,NOD
         IF (MODE.EQ.1) THEN
C  Read data for 1 variable from a text file.
            CALL ROLDAT(NU,NAME,IXA,POI,MA,XL,XR,IEOF)
            IF (IEOF.EQ.1) THEN
               IF (IND.EQ.NSR) GOTO 50
               IND = IND-1
               GOTO 150
            ELSE IF (IEOF.EQ.2) THEN
               IF (WEAK) WRITE (IUOUT,130) FILE(1:LEF)
               CLOSE (NU)
               MERR = -2
               RETURN
            END IF
         ELSE
C  Read data for 1 variable from a binary file.
            CALL RDIDAT(NU,IND,I,NAME,IXA,POI,MA,XL,XR)
         END IF
         IF (IPRINT.EQ.2) CALL WOLDAT(IUOUT,I,NAME,POI,MA,XL,XR)
80    CONTINUE
C
C  Read and print the functions value.
      IF (MODE.EQ.1) THEN
         CALL RVSDAT(NU,VA,IEOF)
         IF (IEOF.NE.0) THEN
            IF (WEAK) WRITE (IUOUT,130) FILE(1:LEF)
            CLOSE (NU)
            MERR = -2
            RETURN
         END IF
      ELSE
         CALL RVDDAT(NU,IND,VA)
      END IF
      IF (IPRINT.NE.0) THEN
         CALL INSPVA(IND,VA)
         NP1 = NP1+1
      END IF
C
C  Stop scrolling every 1 record if IPRINT = 2; every NPAGE records
C  if IPRINT = 1.
C  Sorry, but MCL and macros have no interaction with the user.
      IF ((IPRINT.EQ.1 .AND. NP1+1.GT.NPAGE .OR. IPRINT.EQ.2) .AND.
     &     LEVIN.EQ.0) THEN
         IF (STRO) WRITE (IUOUT,20)
         CALL INNOCR(CH,.FALSE.,IERR)
         IF (IERR.EQ.-1) THEN
            CLOSE (NU)
            MERR = -1
            RETURN
         END IF
         IF (ISCOMP(CH,'Q').NE.0) GOTO 150
         NP1 = 0
      END IF
C
C  Get ready for the next record. Position the file according to IDIR
C  and check for termination.
      IF (IDIR.EQ.1) THEN
         NEXT = IND+1
         IF ((MODE.EQ.2 .AND. NEXT.GT.INSREC) .OR.
     &          (IEND.NE.0 .AND. NEXT.GT.IEND)) GOTO 150
         IND = NEXT
      ELSE
         IF (IND.EQ.IEND) GOTO 150
         IF (MODE.EQ.1) THEN
            DO 90,I=1,2*(NOD+1)
90             BACKSPACE NU
         END IF
         IND = IND-1
      END IF
      GOTO 100
C
C  Successful completion.
150   CONTINUE
      CLOSE (NU)
C
C  Informative printout if IPRINT = 0.
      CALL I2STR(IND,STR,LES)
      IF (IPRINT.EQ.0 .AND. STRO)
     &    WRITE (IUOUT,10) FILE(1:LEF), STR(1:LES)
C
C  Set the MCL-returned values.
      CALL SETPI('REC',IND,IERR)
C
C  Set default values to the PICK panel. Any subsequent PICK will
C  read the record last inspected.
      CALL SETPPS('PICK','FILE',FILE,LEF,IERR)
      CALL SETPPS('PICK','TYPE',TYPE,LET,IERR)
      CALL SETPPI('PICK','REC',IND,IERR)
      RETURN
C
C  End-of-File while reading a text file, or a binary file does not
C  contain that many records.
50    CONTINUE
      IF (WEAK) THEN
         IF (ISTART.EQ.0) ISTART = 1
         CALL I2STR(ISTART,STR,LES)
         WRITE (IUOUT,120) FILE(1:LEF), STR(1:LES)
      END IF
      CLOSE (NU)
      MERR = -2
      RETURN
C
10    FORMAT (/$(FC)'Inspection completed. File "',A,'" contains ',
     &       A,' records.')
20    FORMAT (/$(FC)'Enter q to terminate file inspection, anything ',
     &        'else to continue.')
110   FORMAT (/$(FC)'File "',A,'" is probably empty.')
120   FORMAT (/$(FC)'File "',A,'" does not contain ',A,' record(s) ',
     &        'or the file is corrupt.')
130   FORMAT (/$(FC)'File "',A,'" is corrupt.')
      END
C  ---------------------------------------------------------------------
      SUBROUTINE INSPVA ( IREC, VAL )
C  ---------------------------------------------------------------------
C
C  Description:
C    Part of the INSPECT command. Displays the record number IREC
C    and the value of the objective function VAL.
C
C  Input arguments:
C    IREC       Record number to be displayed.
C    VAL        The value to be displayed.
C
C  ---------------------------------------------------------------------
      IMPLICIT $(TYPE) (A-H,O-Z)
C  Common blocks:
      LOGICAL STRO, WEAK
      COMMON /PRILOG/ STRO, WEAK
      COMMON /UNITS/ IUINP, IUOUT
C  ---------------------------------------------------------------------
C  Local variables:
      CHARACTER*23 CVAL
C  ---------------------------------------------------------------------
C
      IF (STRO) THEN
         CALL R2RTF(VAL,CVAL,LEV)
         WRITE (IUOUT,10) IREC, CVAL(1:LEV)
      END IF
C
10    FORMAT ($(FC)4X,'Record ',I4,6X,'Value ',7('.'),4X,A)
      END
C  ---------------------------------------------------------------------
      SUBROUTINE JCOMP ( FJ, MAXW, MERR )
C  ---------------------------------------------------------------------
C
C  Description:
C    Implements the JCOMPARE command:
C    Syntax: HESSIAN
C
C  Input arguments:
C    MAXW       Number of available storage positions in array FJ.
C
C  Output arguments:
C    MERR       Error indicator.
C
C  Input / Output arguments:
C    FJ         The 2-dimensional Merlin work array.
C  ---------------------------------------------------------------------
      IMPLICIT $(TYPE) (A-H,O-Z)
C  Arguments:
      DIMENSION FJ(NOF,NOD)
C  ---------------------------------------------------------------------
C  Common blocks:
      PARAMETER ( MXV = $(MXV) )
      PARAMETER ( MXT = $(MXT) )
      COMMON /DIMEN/ NOD, NOF
      COMMON /PARAMS/ POINT(MXV), VAL
      COMMON /DETAIL/ IXAT(MXV), MARG(MXV)
      COMMON /RWORKT/ FT1(MXT), FT2(MXT), FT3(MXT)
      LOGICAL STRO, WEAK
      COMMON /PRILOG/ STRO, WEAK
      COMMON /UNITS/ IUINP, IUOUT
      CHARACTER*10 DERIVA, PRINTO, PROCES, JACOMO
      COMMON /MODEIN/ DERIVA, PRINTO, PROCES, JACOMO
C  ---------------------------------------------------------------------
C  Local variables:
      CHARACTER*10 JTMP
C  ---------------------------------------------------------------------
C
C  Is there enough storage ?
      NEED = NOD*NOF
      CALL VSIM('M*N',MAXW,NEED,MERR)
      IF (MERR.NE.0) RETURN
C
C  Calculate the terms at the given point.
      MERR = 0
      CALL LSQFCN(NOF,NOD,POINT,FT1)
C
C  Save the current Jacobian mode.
      JTMP = JACOMO
C
C  Get the Jacobian from the user supplied SUBROUTINE JANAL.
      JACOMO = 'JANAL'
      CALL MJACOB(NOF,NOD,POINT,FT1,FJ,NOF)
C
C  Store the Jacobian in a temporary file.
      IU = NUNIT()
      IF (IU.EQ.-1) THEN
         CALL ERRUAL('<SCRATCH>')
         MERR = -2
         RETURN
      END IF
      OPEN (IU,STATUS='SCRATCH',FORM='UNFORMATTED')
      DO 10,J=1,NOD
         DO 20,I=1,NOF
            WRITE (IU) FJ(I,J)
20       CONTINUE
10    CONTINUE
C
C  Approximate the Jacobian using forward differences.
      JACOMO = 'JNUMER'
      CALL MJACOB(NOF,NOD,POINT,FT1,FJ,NOF)
C
C  Compare the two Jacobians.
      REWIND IU
      DIFMAX = 0.
      RMS = 0.
      ONE = 1.
      NCOMP = 0
      DO 30,J=1,NOD
         IX = IXAT(J)
         DO 40,I=1,NOF
            READ (IU) F1
            F2 = FJ(I,J)
            IF (IX.NE.0) THEN
               DIFF = ABS(F1-F2)/MAX(ONE,MIN(ABS(F1),ABS(F2)))
               RMS = RMS+DIFF**2
               IF (DIFF.GT.DIFMAX) DIFMAX = DIFF
               NCOMP = NCOMP+1
            END IF
40       CONTINUE
30    CONTINUE
C
      CLOSE (IU,STATUS='DELETE')
C
C  Restore the Jacobian mode.
      JACOMO = JTMP
      RMS = SQRT(RMS/NCOMP)
      IF (STRO) WRITE (IUOUT,150) RMS, DIFMAX
C
150   FORMAT (/$(FC)'RMS relative error : ',1PG12.5/
     &         $(FC)'Max. relative error: ',1PG12.5)
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE LIMITS
C  ---------------------------------------------------------------------
C
C  Description:
C    Implements the LIMITS command.
C
C  ---------------------------------------------------------------------
C  Common blocks:
      LOGICAL STRO, WEAK
      COMMON /PRILOG /STRO, WEAK
      COMMON /UNITS/ IUINP, IUOUT
C  ---------------------------------------------------------------------
C  Local variables:
      PARAMETER ( MXV = $(MXV) )
      PARAMETER ( MXT = $(MXT) )
      PARAMETER ( MAXALI = 20 )
      PARAMETER ( MAXINL = 20 )
      PARAMETER ( MAXONL = 20 )
      PARAMETER ( MAXFLA = 20 )
      PARAMETER ( MAXCFL = 10, MAXCLE = 30 )
      PARAMETER ( NMCOM = 98, MXPLUG = $(MXPLUG), NCOM = NMCOM+MXPLUG )
      PARAMETER ( LSTA = $(MCLMEM) )
      PARAMETER ( MAXMFI = 20 )
      PARAMETER ( MXBUF = $(MCLBUF), MXPTR = MXBUF/5 )
      CHARACTER*80 LINE, BLANK
      DATA BLANK / ' ' /
C  ---------------------------------------------------------------------
C
      IF (.NOT.STRO) RETURN
C
      WRITE (IUOUT,10)
      WRITE (IUOUT,20) MXV
      WRITE (IUOUT,30) MXT
      WRITE (IUOUT,40) MAXALI
      WRITE (IUOUT,50) MIN(MAXINL,MAXONL)
      WRITE (IUOUT,60) MAXFLA
      WRITE (IUOUT,70) MAXCFL
      WRITE (IUOUT,80) MAXCLE
      WRITE (IUOUT,90) MXPLUG
      WRITE (IUOUT,*)
      WRITE (IUOUT,100) LSTA
      WRITE (IUOUT,110) MAXMFI
C
      CALL I2STR(MXBUF,LINE,LES)
      LINE(LES+1:) = ' characters '
      LES = LES+12
      CALL I2STR(MXPTR,LINE(LES+1:),LEL)
      LES = LES+LEL
      LINE(LES+1:) = ' lines'
      LES = LES+6
      WRITE (IUOUT,120) BLANK(1:36-LES), LINE(1:LES)
C
10    FORMAT (/$(FC)'Current limits are:'
     &        /$(FC)56('-'))
20    FORMAT ($(FC)'Maximum number of minimization parameters:',T50,I7)
30    FORMAT ($(FC)'Maximum number of terms in the sum of squares:',
     &        T50,I7)
40    FORMAT ($(FC)'Maximum number of aliases:',T50,I7)
50    FORMAT ($(FC)'Maximum nesting level:',T50,I7)
60    FORMAT ($(FC)'Number of numerical flags:',T50,I7)
70    FORMAT ($(FC)'Number of character flags:',T50,I7)
80    FORMAT ($(FC)'Length of a character flag:',T50,I7)
90    FORMAT ($(FC)'Maximum number of plug-in modules:',T50,I7)
100   FORMAT ($(FC)'Current MCL memory:',T50,I7)
110   FORMAT ($(FC)'Maximum number of MCL files:',T50,I7)
120   FORMAT ($(FC)'Maximum length of an MCL object code,',
     &        /$(FC)'that fits in memory:',A,A)
      END
C  ---------------------------------------------------------------------
      SUBROUTINE LIST
C  ---------------------------------------------------------------------
C
C  Description:
C    Implements the LIST command.
C    Syntax: LIST
C
C  ---------------------------------------------------------------------
C  Common blocks:
      PARAMETER ( NMCOM = 98, MXPLUG = $(MXPLUG), NCOM = NMCOM+MXPLUG )
      CHARACTER*10 COMMAN(NCOM)
      COMMON /COM1/ COMMAN
      LOGICAL STRO, WEAK
      COMMON /PRILOG/ STRO, WEAK
      COMMON /UNITS/ IUINP, IUOUT
C  ---------------------------------------------------------------------
C  Local variables:
      CHARACTER*80 LINE
C  ---------------------------------------------------------------------
C
      IF (.NOT.STRO) RETURN
      CALL DASH
      CALL CPRINT('Merlin commands')
      CALL DASH
C
      NC = 0
      LL = 0
      DO 10,I=1,NCOM
         IF (COMMAN(I)(1:4).NE.'PLUG') THEN
            LINE(LL+1:LL+13) = '   '//COMMAN(I)
            NC = NC+1
            LL = LL+13
            IF (NC.EQ.6) THEN
               WRITE (IUOUT,20) LINE(1:LL)
               NC = 0
               LL = 0
            END IF
         END IF
10    CONTINUE
      IF (NC.NE.0) WRITE (IUOUT,20) LINE(1:LL)
      WRITE (IUOUT,30)
C
20    FORMAT ($(FC)A)
30    FORMAT (/$(FC)'Use the HELP command to obtain more information.')
      END
C  ---------------------------------------------------------------------
      SUBROUTINE LOOSAL
C  ---------------------------------------------------------------------
C
C  Description:
C    Implements the LOOSALL command.
C    Syntax: LOOSALL
C
C  ---------------------------------------------------------------------
      IMPLICIT $(TYPE) (A-H,O-Z)
C  Common blocks:
      PARAMETER ( MXV = $(MXV) )
      COMMON /PARAMS/ POINT(MXV), VAL
      COMMON /DETAIL/ IXAT(MXV), MARG(MXV)
      COMMON /ROSTEP/ STEP(MXV)
      COMMON /DIMEN/ NOD, NOF
      COMMON /ACCURA/ COMACC
      COMMON /UPDATE/ IGRADU, IHESSU
      COMMON /TOTCAL/ ITNOCS, IPARTC, ITDER, IPDER, ITHESS, IPHESS,
     &                ITJAC, IPJAC
      LOGICAL STRO, WEAK
      COMMON /PRILOG/ STRO, WEAK
      COMMON /UNITS/ IUINP, IUOUT
C  ---------------------------------------------------------------------
C
      ONE = 1.
      COMTOL = SQRT(COMACC)
C
C  Add a backup record.
      CALL BACKUP('D')
      DO 10,I=1,NOD
C  Set an arbitrary value for the search step.
         IF (IXAT(I).EQ.0 .AND. ITNOCS.NE.0)
     &                  STEP(I) = (ONE+ABS(POINT(I)))*COMTOL
C  Loose the variable.
         IXAT(I) = 1
         IGRADU = 0
         IHESSU = 0
10    CONTINUE
      IF (STRO) WRITE (IUOUT,20)
C
20    FORMAT (/$(FC)'All variables are now free.')
      END
C  ---------------------------------------------------------------------
      SUBROUTINE MACRO ( MERR )
C  ---------------------------------------------------------------------
C
C  Description:
C    Subroutine MACRO implements the MACRO command. It helps the user
C    construct a macro (command package) and dispose it in the Merlin
C    macro file.
C
C  Output arguments:
C    MERR       Error indicator.
C
C  ---------------------------------------------------------------------
      IMPLICIT $(TYPE) (A-H,O-Z)
C  Common blocks:
      PARAMETER ( MXFILE = 80 )
      CHARACTER*(MXFILE) PROLO, EPILO, IN, OUT, HFILE, MFILE, PDFILE
      COMMON /CP2/ PROLO, EPILO, IN, OUT, HFILE, MFILE, PDFILE
      LOGICAL STRO, WEAK
      COMMON /PRILOG/ STRO, WEAK
      COMMON /UNITS/ IUINP, IUOUT
C  ---------------------------------------------------------------------
C  Local variables:
      PARAMETER ( LINLEN = 80 )
      CHARACTER*(LINLEN) COM, TMACNM, MACNM
C  ---------------------------------------------------------------------
C
C  Get macro name. It should start with a dot.
      IUMAC = -1
4     IF (STRO) WRITE (IUOUT,10)
      CALL INNOCR(TMACNM,.FALSE.,MERR)
      IF (MERR.NE.0) THEN
         IF (IUMAC.NE.-1) CLOSE (IUMAC)
         RETURN
      END IF
C
C  Prepend the macro dot.
      MACNM = '.'//TMACNM
C
C  Search for the macro.
      IF (IUMAC.EQ.-1) THEN
C
C  Open the macro file.
         CALL OPMAC(1,INFO,IUMAC)
C
C  Any errors ?
         IF (INFO.NE.0) THEN
            MERR = -2
            RETURN
         END IF
      END IF
C
C  Search in the macro file.
      CALL LOCMAC(MACNM,IUMAC,ISU)
      IF (ISU.EQ.1) THEN
         IF (WEAK) WRITE (IUOUT,30)
         IF (ISIAF().NE.0) GOTO 4
         CLOSE (IUMAC)
         MERR = -2
         RETURN
      ELSE IF (ISU.EQ.-1) THEN
         IF (WEAK) THEN
            LEM = LENGTH(MFILE)
            WRITE (IUOUT,40) MFILE(1:LEM)
         END IF
         CLOSE (IUMAC)
         MERR = -2
         RETURN
      END IF
C
C  Position the macro file at EOF.
      CALL POSIT(IUMAC)
C
C  Write preliminaries.
      WRITE (IUMAC,'(''*'')')
      WRITE (IUMAC,'(A)') MACNM(1:LENGTH(MACNM))
C
C Loop, get command, write to macrofile.
7     IF (STRO) WRITE (IUOUT,50)
      CALL INNOCR(COM,.FALSE.,MERR)
      IF (MERR.EQ.-1) THEN
         CLOSE (IUMAC)
         RETURN
      END IF
C
      LEC = LENGTH(COM)
      IF (LEC.EQ.0) THEN
         WRITE (IUMAC,*)
      ELSE
         WRITE (IUMAC,'(A)') COM(1:LEC)
      END IF
      IF (ISCOMP(COM,'CLEAR').EQ.0) GOTO 7
      IF (STRO) WRITE (IUOUT,60)
      CLOSE (IUMAC)
      MERR = 0
C
10    FORMAT (/$(FC)'** Enter macro name. Exit with a blank line **')
30    FORMAT (/$(FC)'This macro exists already in the macro-file. ',
     &        'Use a different name.')
40    FORMAT (/$(FC)'The Merlin macro file "',A,'" is corrupt.')
50    FORMAT (/$(FC)'** Enter desired command. Exit with CLEAR **')
60    FORMAT (/$(FC)'** Macro editing complete **')
      END
C  ---------------------------------------------------------------------
      SUBROUTINE MAD ( MERR )
C  ---------------------------------------------------------------------
C
C  Description:
C    Implements the MAD command.
C    Syntax: MAD (Has panel)
C
C  Output arguments:
C    MERR       Error indicator.
C
C  ---------------------------------------------------------------------
C
      CALL CHANGE(MERR)
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE MARGI ( IOPT, MERR )
C  ---------------------------------------------------------------------
C
C  Description:
C    Implements the LMARGIN and RMARGIN commands.
C    Syntax: LMARGIN index_specification value ...
C            RMARGIN index_specification value ...
C
C  Input arguments:
C    IOPT       Specifies the operation to be performed:
C               IOPT = -1 -> LMARGIN
C               IOPT =  1 -> RMARGIN
C
C  Output arguments:
C    MERR       Error indicator.
C
C  ---------------------------------------------------------------------
      IMPLICIT $(TYPE) (A-H,O-Z)
C  Common blocks:
      PARAMETER ( MXV = $(MXV) )
      LOGICAL STRO, WEAK
      COMMON /PRILOG /STRO, WEAK
      COMMON /UNITS/ IUINP, IUOUT
      COMMON /IWORK/ INDE(MXV)
      COMMON /RWORK/ XT1(MXV), XT2(MXV), XT3(MXV), XT4(MXV), XT5(MXV),
     &               XT6(MXV), XT7(MXV+1)
C  ---------------------------------------------------------------------
C
      CALL ARGNO(NA)
C
C  Prompt the user if no arguments were entered.
      IF (NA.EQ.0 .AND. STRO) THEN
         IF (IOPT.EQ.-1) THEN
            WRITE (IUOUT,20)
         ELSE
            WRITE (IUOUT,30)
         END IF
      END IF
C
C  Read an one line input, or parse command line arguments for
C  index/value specifications.
      CALL CHIV(INDE,XT1,NPAIRS,1,MERR)
      IF (MERR.NE.0) RETURN
C
C  Add a backup record.
      IF (NPAIRS.NE.0) CALL BACKUP('D')
C
      DO 10,I=1,NPAIRS
         IND = INDE(I)
         XM = XT1(I)
         IF (IOPT.EQ.-1) THEN
            CALL SETLMA(IND,XM)
         ELSE
            CALL SETRMA(IND,XM)
         END IF
10    CONTINUE
C
C  Check if the current point is inside the margins.
C  If not readjust it.
      CALL POIMAR(IFEV)
C
20    FORMAT (/$(FC)'Enter left margins sequence:')
30    FORMAT (/$(FC)'Enter right margins sequence:')
      END
C  ---------------------------------------------------------------------
      SUBROUTINE MEMO ( MERR )
C  ---------------------------------------------------------------------
C
C  Description:
C    Implements the MEMO command.
C    Syntax: MEMO (Has panel)
C
C  Output arguments:
C    MERR       Error indicator.
C
C  ---------------------------------------------------------------------
C  Common blocks:
      PARAMETER ( MXFILE = 80 )
      CHARACTER*(MXFILE) MEMOFI
      COMMON /MEMO1/ MEMOFI
      COMMON /MEMO2/ MEMOMO, MEMOKE, MEMREC, IUMEM
      LOGICAL STRO, WEAK
      COMMON /PRILOG/ STRO, WEAK
      COMMON /UNITS/IUINP, IUOUT
      COMMON /TOTCAL/ ITNOCS, IPARTC, ITDER, IPDER, ITHESS, IPHESS,
     &                ITJAC, IPJAC
C  ---------------------------------------------------------------------
C  Local variables:
      CHARACTER*(MXFILE) ORGFIL
      CHARACTER*5 TYPE
      SAVE IUMO
C  ---------------------------------------------------------------------
C
      CALL SETPI('CANCEL',1,MERR)
      CALL CHANGE(MERR)
      IF (MERR.NE.0) RETURN
      CALL GETPI('CANCEL',ICANC)
      IF (ICANC.EQ.0) RETURN
      IF (ITNOCS.EQ.0) RETURN
C
C  Get panel parameters.
      CALL GETPS('FILE',MEMOFI,MEMOLE)
      CALL GETPS('TYPE',TYPE,LET)
      CALL GETPI('OPEN',MEMOKE)
      IF (ISCOMP(TYPE,'TEXT').NE.0) THEN
         MEMOMO = 1
      ELSE
         MEMOMO = 2
      END IF
C
C  Is the memo file already open ?
      IF (IUMEM.NE.-1) THEN
         CALL U2NAME(IUMEM,'MEMO',ORGFIL,LEOF)
         IF (STRO) WRITE (IUOUT,70) ORGFIL(1:LEOF)
      END IF
      MO = MEMOMO
      IF (IUMEM.NE.-1) MO = IUMO
C
C  Open the memo file.
      IUMEMP = IUMEM
      IF (IUMEM.EQ.-1) THEN
         CALL OPENIT(MEMOFI,TYPE,1,1,IUMEM,MEMREC)
         IF (IUMEM.EQ.-1) THEN
            MERR = -2
            RETURN
         END IF
      END IF
C
C  Add a record to the memo file.
      CALL WMEMO(MO)
      IF (STRO) WRITE (IUOUT,10) MEMOFI(1:LENGTH(MEMOFI))
C
C  Leave the file open ?
      IF (MEMOKE.EQ.0) THEN
         CLOSE (IUMEM)
         IUMEM = -1
         IF (IUMEMP.NE.-1 .AND. STRO) WRITE (IUOUT,100) ORGFIL(1:LEOF)
      ELSE
         IF (STRO) WRITE (IUOUT,90) MEMOFI(1:LENGTH(MEMOFI))
         IUMO = MO
      END IF
C
10    FORMAT (/$(FC)'The current point has been appended to file "',
     &        A,'".')
70    FORMAT (/$(FC)'Warning: The memo file "',A,'" is open. ',
     &        'Panel parameters ignored.')
90    FORMAT (/$(FC)'File "',A,'" is left open.')
100   FORMAT (/$(FC)'The previous memo file "',A,'" has been closed.')
      END
C  ---------------------------------------------------------------------
      SUBROUTINE WMEMO ( MODE )
C  ---------------------------------------------------------------------
C
C  Description:
C    Called by subroutine MEMO. Adds a record to the memo file.
C
C  Input arguments:
C    MODE       MODE = 1 -> The memo file is a text file.
C               MODE = 2 -> The memo file is a binary file.
C
C  ---------------------------------------------------------------------
      IMPLICIT $(TYPE) (A-H,O-Z)
C  Arguments:
C  ---------------------------------------------------------------------
C  Common blocks:
      PARAMETER ( MXV = $(MXV) )
      COMMON /PARAMS/ POINT(MXV), VAL
      COMMON /DIMEN/ NOD, NOF
      COMMON /MEMO2/ MEMOMO, MEMOKE, MEMREC, IUMEM
C  ---------------------------------------------------------------------
C
      IF (MODE.EQ.1) THEN
         CALL WRITER(POINT,VAL,IUMEM,MODE,0)
      ELSE
         IST = MEMREC+1
         CALL WRITER(POINT,VAL,IUMEM,MODE,IST)
         MEMREC = MEMREC+1
         WRITE (IUMEM,REC=1) NOD, MEMREC
      END IF
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE MIXED ( MERR )
C  ---------------------------------------------------------------------
C
C  Description:
C    Implements the MIXED command.
C    Syntax: MIXED index_specification mode ...
C            where mode may be any one of: FAST, QUAD, NUMER, ANAL
C                                      or: F,    Q,    N,     A.
C            in obvious correspondence.
C
C  Output arguments:
C    MERR       Error indicator.
C
C  ---------------------------------------------------------------------
      IMPLICIT $(TYPE) (A-H,O-Z)
C  Common blocks:
      PARAMETER ( MXV = $(MXV) )
      CHARACTER*5 GRADI
      COMMON /DERIMO/ GRADI(MXV)
      COMMON /UPDATE/ IGRADU, IHESSU
      COMMON /IWORK/ INDE(MXV)
      CHARACTER*10 DERIVA, PRINTO, PROCES, JACOMO
      COMMON /MODEIN/ DERIVA, PRINTO, PROCES, JACOMO
      LOGICAL STRO, WEAK
      COMMON /PRILOG/ STRO, WEAK
      COMMON /UNITS/ IUINP, IUOUT
C  ---------------------------------------------------------------------
C  Local variables:
      PARAMETER ( LINLEN = 80 )
      CHARACTER*(LINLEN) LINE, GRMODE
      INTEGER IS(LINLEN/2), IE(LINLEN/2), ITEM(MXV)
C  ---------------------------------------------------------------------
C
C  Prompt the user, or parse command line arguments for index/value
C  specifications.
      CALL CHIS(INDE,ITEM,NCH,LINE,IS,IE,1,MERR)
      IF (MERR.NE.0) RETURN
C
      NTERR = 0
      NOK = 0
      DO 100,I=1,NCH
         IND = INDE(I)
         N = ITEM(I)
         GRMODE = LINE(IS(N):IE(N))
C
         IF (ISCOMP(GRMODE,'F').NE.0) THEN
            GRMODE = 'FAST'
         ELSE IF (ISCOMP(GRMODE,'Q').NE.0) THEN
            GRMODE = 'QUAD'
         ELSE IF (ISCOMP(GRMODE,'N').NE.0) THEN
            GRMODE = 'NUMER'
         ELSE IF (ISCOMP(GRMODE,'A').NE.0) THEN
            GRMODE = 'ANAL'
         END IF
C
         IF (ISCOMP(GRMODE,'FAST').NE.0 .OR.
     &          ISCOMP(GRMODE,'QUAD').NE.0 .OR.
     &          ISCOMP(GRMODE,'NUMER').NE.0 .OR.
     &          ISCOMP(GRMODE,'ANAL').NE.0) THEN
            CALL UPPER(GRMODE,LEN(GRMODE))
            IF (GRADI(IND).NE.GRMODE) THEN
               IGRADU = 0
               GRADI(IND) = GRMODE
               NOK = NOK+1
            END IF
         ELSE
            IF (WEAK) WRITE (IUOUT,10) LINE(IS(N):IE(N))
            NTERR = NTERR+1
         END IF
100   CONTINUE
C
C  Enter the MIXED mode if at least one gradient mode was changed.
      IF (NOK.NE.0) DERIVA = 'MIXED'
      IF (NTERR.EQ.0) THEN
         MERR = 0
      ELSE
         MERR = -2
      END IF
C
10    FORMAT (/$(FC)'Derivatime mode "',A,'" is incorrect.')
      END
C  ---------------------------------------------------------------------
      SUBROUTINE MODIS
C  ---------------------------------------------------------------------
C
C  Description:
C    Implements the MODEDIS command.
C    Syntax: MODEDIS
C
C    The MODEDIS command prints the Merlin operation modes, currently
C    in effect.
C
C  ---------------------------------------------------------------------
      IMPLICIT $(TYPE) (A-H,O-Z)
C  Common blocks:
      PARAMETER ( NMCOM = 98, MXPLUG = $(MXPLUG), NCOM = NMCOM+MXPLUG )
      CHARACTER*10 COMMAN(NCOM)
      COMMON /COM1/ COMMAN
      COMMON /DIMEN/ NOD, NOF
      COMMON /BIGCOM/ BIGNUM, BLIM, DWARF, DWALOG
      COMMON /ACCURA/ COMACC
      CHARACTER*10 DERIVA, PRINTO, PROCES, JACOMO
      COMMON /MODEIN/ DERIVA, PRINTO, PROCES, JACOMO
      PARAMETER ( MXFILE = 80 )
      CHARACTER*(MXFILE) PROLO, EPILO, IN, OUT, HFILE, MFILE, PDFILE
      COMMON /CP2/ PROLO, EPILO, IN, OUT, HFILE, MFILE, PDFILE
      COMMON /MODE3/ MODEV, IFFORM, INMCL
      LOGICAL AIMON
      COMMON /AIM/ TARGET, AIMON
      LOGICAL STRO, WEAK
      COMMON /PRILOG/ STRO, WEAK
      COMMON /UNITS/ IUINP, IUOUT
C  ---------------------------------------------------------------------
C  Local variables:
      CHARACTER*(MXFILE) CIN, COUT
      CHARACTER*20 STR
      CHARACTER*10 STR2, SAFE
C  ---------------------------------------------------------------------
C
      IF (.NOT.STRO) RETURN
C
      WRITE (IUOUT,10)
      CALL DASH
      CALL I2STR(NOD,STR,LES)
      CALL R2STR(BIGNUM,STR2,LES2)
      WRITE (IUOUT,20) STR(1:LES), STR2(1:LES2)
C
      CALL I2STR(NOF,STR,LES)
      CALL R2STR(DWARF,STR2,LES2)
      WRITE (IUOUT,30) STR(1:LES), STR2(1:LES2)
C
      CALL GETPPS('BACKUP','MODE',STR,LES)
      CALL GETPPS('BACKUP','SAFE',SAFE,LESF)
      STR(LES+1:) = 'BACK'
      LES = LES+4
      CALL UPPER(STR,LES)
      IF (ISCOMP(SAFE,'YES').NE.0) THEN
         STR(LES+1:) = ' (safe)'
         LES = LES+7
      END IF
      CALL R2STR(COMACC,STR2,LES2)
      WRITE (IUOUT,40) STR(1:LES), STR2(1:LES2)
C
      IF (AIMON) THEN
         WRITE (IUOUT,50) DERIVA, TARGET
      ELSE
         WRITE (IUOUT,60) DERIVA
      END IF
C
      NP = MXPLUG
      DO 200,I=1,NCOM
200      IF (COMMAN(I)(1:4).EQ.'PLUG') NP = NP-1
      WRITE (IUOUT,70) PRINTO, NP
      WRITE (IUOUT,80) PROCES, JACOMO
C
      IF (MODEV.EQ.0) THEN
         WRITE (IUOUT,110) 'OFF'
      ELSE
         WRITE (IUOUT,110) 'ON'
      END IF
C
      STR = '$(TYPE)'
      LL = LENGTH(STR)
      CALL UPPER(STR,LL)
      WRITE (IUOUT,115) STR(1:LL)
C
      IF (IFFORM.EQ.0) THEN
         WRITE (IUOUT,120) 'General function'
      ELSE
         WRITE (IUOUT,120) 'Sum of squares'
      END IF
C
      LEP = LENGTH(PROLO)
      IF (LEP.EQ.0) THEN
         WRITE (IUOUT,90) '- NONE -'
      ELSE
         WRITE (IUOUT,90) '"'//PROLO(1:LEP)//'"'
      END IF
C
      LEE = LENGTH(EPILO)
      IF (LEE.EQ.0) THEN
         WRITE (IUOUT,100) '- NONE -'
      ELSE
         WRITE (IUOUT,100) '"'//EPILO(1:LEE)//'"'
      END IF
C
      LIN = LENGTH(IN)
      IF (LIN.EQ.0) THEN
         CIN = '- NONE -'
         LIN = 8
      ELSE
         CIN = '"'//IN(1:LIN)//'"'
         LIN = LIN+2
      END IF
      WRITE (IUOUT,130) CIN(1:LIN)
C
      LOUT = LENGTH(OUT)
      IF (LOUT.EQ.0) THEN
         COUT = '- NONE -'
         LOUT = 8
      ELSE
         COUT = '"'//OUT(1:LOUT)//'"'
         LOUT = LOUT+2
      END IF
      WRITE (IUOUT,140) COUT(1:LOUT)
      CALL DASH
C
10    FORMAT (/$(FC)'Current operation modes')
20    FORMAT ($(FC)'Number of variables:      ',A,
     &        T45,'Largest real:          ',A)
30    FORMAT ($(FC)'Number of terms:          ',A,
     &        T45,'Smallest nonzero real: ',A)
40    FORMAT ($(FC)'Backup mode: ',T27,A,
     &        T45,'Machine''s accuracy:    ',A)
50    FORMAT ($(FC)'Derivative mode: ',T27,A,
     &        T45,'Target value:          ',1PE10.2E3)
60    FORMAT ($(FC)'Derivative mode: ',T27,A,
     &        T45,'Target value:          ','- NONE -')
70    FORMAT ($(FC)'Printout mode: ',T27,A,
     &        T45,'User plug-in modules:  ',I5)
80    FORMAT ($(FC)'Processing mode: ',T27,A,
     &        T45,'Jacobian mode:         ',A)
90    FORMAT ($(FC)'Prolog command: ',T27,A)
100   FORMAT ($(FC)'Epilog command: ',T27,A)
110   FORMAT ($(FC)'Function evaluation: ',T27,A)
115   FORMAT ($(FC)'Arithmetic:          ',T27,A)
120   FORMAT ($(FC)'Functional form: ',T27,A)
130   FORMAT ($(FC)'Input file: ',T27,A)
140   FORMAT ($(FC)'Output file: ',T27,A)
      END
C  ---------------------------------------------------------------------
      SUBROUTINE NONAME ( MERR )
C  ---------------------------------------------------------------------
C
C  Description:
C    Implements the NONAME command.
C    Syntax: NONAME index_specification ...
C
C  Output arguments:
C    MERR       Error indicator.
C
C  ---------------------------------------------------------------------
C  Common blocks:
      PARAMETER ( MXV = $(MXV) )
      CHARACTER*10 NICKNA
      COMMON /NAMES/ NICKNA(MXV)
      COMMON /IWORK/ INDE(MXV)
C  ---------------------------------------------------------------------
C
C  Prompt the user, or parse command line arguments for index
C  specifications.
      CALL CHI(INDE,NCH,1,MERR)
      IF (MERR.NE.0) RETURN
C
C  Add a backup record.
      CALL BACKUP('D')
C
      DO 20,I=1,NCH
20       NICKNA(INDE(I)) = ' '
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE NOTARG
C  ---------------------------------------------------------------------
C
C  Description:
C    Implements the NOTARGET command.
C    Syntax: NOTARGET
C
C  ---------------------------------------------------------------------
      IMPLICIT $(TYPE) (A-H,O-Z)
C  Common blocks:
      LOGICAL AIMON
      COMMON /AIM/ TARGET, AIMON
      LOGICAL STRO, WEAK
      COMMON /PRILOG/ STRO, WEAK
      COMMON /UNITS/ IUINP, IUOUT
C  ---------------------------------------------------------------------
C  Local variables:
      CHARACTER*30 STR
C  ---------------------------------------------------------------------
C
      IF (AIMON) THEN
         IF (STRO) THEN
            CALL R2STR(TARGET,STR,LES)
            WRITE (IUOUT,10) STR(1:LES)
         END IF
         AIMON = .FALSE.
      ELSE
         IF (STRO) WRITE (IUOUT,20)
      END IF
C
10    FORMAT (/$(FC)'The previous target value ',A,' has been ',
     &        'cleared.')
20    FORMAT (/$(FC)'There is no target value in effect.')
      END
C  ---------------------------------------------------------------------
      SUBROUTINE PDUMP ( MERR )
C  ---------------------------------------------------------------------
C
C  Description:
C    Implements the PDUMP command.
C    Syntax: PDUMP file [ panel ... ]
C
C  Output arguments:
C    MERR       Error indicator.
C
C  ---------------------------------------------------------------------
      IMPLICIT $(TYPE) (A-H,O-Z)
C  Common blocks:
      PARAMETER ( NMCOM = 98, MXPLUG = $(MXPLUG), NCOM = NMCOM+MXPLUG )
      CHARACTER*10 COMMAN(NCOM)
      COMMON /COM1/ COMMAN
      LOGICAL STRO, WEAK
      COMMON /PRILOG/ STRO, WEAK
      COMMON /UNITS/ IUINP, IUOUT
C  ---------------------------------------------------------------------
C  Local variables:
      PARAMETER ( MXFILE = 80 )
      CHARACTER*(MXFILE) FILE
      CHARACTER*80 ARG
      LOGICAL OPE, ALL
C  ---------------------------------------------------------------------
C
      MERR = 0
      CALL ARGNO(NA)
      IF (NA.EQ.0) THEN
C  No arguments were entered. Ask for the file.
         IF (STRO) WRITE (IUOUT,10)
         CALL INNOCR(FILE,.FALSE.,MERR)
         IF (MERR.NE.0) RETURN
      ELSE
         CALL CMDARG(1,FILE,LEF)
      END IF
      LEF = LENGTH(FILE)
      ALL = NA.LE.1
C
C  Get info for the file.
      INQUIRE (FILE=FILE,OPENED=OPE,NUMBER=NUM,IOSTAT=IOE)
      IF (IOE.NE.0) THEN
         CALL ERRINQ(FILE,IOE)
         MERR = -2
         RETURN
      END IF
C
C  Cannot dump in an open file.
      IF (OPE) THEN
         CALL ERRBUS(FILE,NUM)
         MERR = -2
         RETURN
      END IF
C
C  Allocate unit for the file.
      NU = NUNIT()
      IF (NU.EQ.-1) THEN
         CALL ERRUAL(FILE)
         MERR = -2
         RETURN
      END IF
C
C  Open the file.
      OPEN (NU,FILE=FILE,IOSTAT=IOE)
      IF (IOE.NE.0) THEN
         CALL ERROPE(FILE,IOE)
         MERR = -2
         RETURN
      END IF
      REWIND NU
C
      IF (ALL) THEN
C  We need to dump all panels.
         DO 20,I=1,NCOM
            CALL PLOCP(COMMAN(I),IP)
            IF (IP.NE.0) CALL PDUMP1(NU,COMMAN(I),IP)
20       CONTINUE
      ELSE
C  Dump only panels that were entered in the command line.
         DO 30,I=2,NA
            CALL CMDARG(I,ARG,LEA)
C  Is this a valid Merlin command ?
            CALL LOCC(ARG,IC)
            IF (IC.EQ.0) THEN
               IF (STRO) WRITE (IUOUT,50) ARG(1:LEA)
               MERR = -2
            ELSE
C  Does the command have a panel ?
               CALL PLOCP(COMMAN(IC),IP)
               IF (IP.EQ.0) THEN
                  IF (STRO) WRITE (IUOUT,40) ARG(1:LEA)
                  MERR = -2
               ELSE
                  CALL PDUMP1(NU,COMMAN(IC),IP)
               END IF
            END IF
30       CONTINUE
      END IF
C
C  Informative printout.
      IF (STRO) WRITE (IUOUT,60) FILE(1:LEF)
      CLOSE (NU)
C
10    FORMAT (/$(FC)'Enter file name:')
50    FORMAT (/$(FC)'There is no such command: "',A,'".')
40    FORMAT (/$(FC)'Command "',A,'" has no panel.')
60    FORMAT (/$(FC)'Panel parameters dumped to file "',A,'".')
      END
C  ---------------------------------------------------------------------
      SUBROUTINE PDUMP1 ( NU, COM, IP )
C  ---------------------------------------------------------------------
C
C  Description:
C    Part of the PDUMP command. Writes on unit NU all panel parameters
C    of panel COM.
C
C  Input arguments:
C    NU         Unit number to write to.
C    COM        Name of the panel.
C    IP         Index of the panel in the panel table.
C
C  ---------------------------------------------------------------------
      IMPLICIT $(TYPE) (A-H,O-Z)
C  Arguments:
      CHARACTER*(*) COM
C  ---------------------------------------------------------------------
C  Common blocks:
C  Maximum length of a keyword.
      PARAMETER ( KEYLEN = 10 )
C  Maximum number of panel commands.
      PARAMETER ( MXPCOM = 80 )
C  Maximum number of keywords in all panel commands.
      PARAMETER ( MXPKEY = 1000 )
C  Maximum storage for integer/real keywords.
      PARAMETER ( MXPIR = 5000 )
C  Maximum storage for string keywords and descriptions.
      PARAMETER ( MXPSTR = 20000 )
C  Maximum number of pointers in the above string storage.
      PARAMETER ( MXPPST = 5000 )
C
      CHARACTER*(KEYLEN) PCOM(MXPCOM)
      INTEGER            PONF(MXPCOM)
      INTEGER            KPOINT(MXPCOM)
      INTEGER            NCHELP(2*MXPCOM)
      CHARACTER*(KEYLEN) PKEY(MXPKEY)
      INTEGER            NKHELP(2*MXPKEY)
      CHARACTER          KTYPE(MXPKEY)
      INTEGER            TPOINT(MXPKEY)
      CHARACTER          KIO(MXPKEY)
      $(TYPE)            PIR(MXPIR)
      INTEGER            SPOINT(MXPPST)
      CHARACTER*(MXPSTR) PANSTR
      COMMON /PCOMS1/ KPOINT, NCHELP, NKHELP, TPOINT, SPOINT, PONF
      COMMON /PCOMS2/ PIR
      COMMON /PCOMS3/ PKEY, KTYPE, PANSTR, PCOM, KIO
C
C  NPCOM: Number of panel commands in the panel description file.
C  NPKEY: Number of keywords in all panel commands.
C  NPIR:  Number of positions filled in array PIR.
C  NPSP:  Number of positions filled in the char-pointer array SPOINT.
C  NPCH:  Number of character positions filled in PANSTR.
      COMMON /PCOMS4/ NPCOM, NPKEY, NPIR, NPSP, NPCH
C  ---------------------------------------------------------------------
C  Local variables:
      CHARACTER*(KEYLEN) KEY
      CHARACTER*80 STR
      CHARACTER KT
C  ---------------------------------------------------------------------
C
      KSTA = KPOINT(IP)
      KEND = KPOINT(IP+1)-1
      DO 30,J=KSTA,KEND
C  Do not write MCL-returned keywords.
         IF (KIO(J).NE.' ') GOTO 30
         KEY = PKEY(J)
         KT = KTYPE(J)
         IF (ISCOMP(KT,'I').NE.0) THEN
C  Keyword that accepts integer values.
            CALL GETPPI(COM,KEY,IPARAM)
            CALL I2STR(IPARAM,STR,LES)
            WRITE (NU,40) COM, KEY, STR(1:LES)
         ELSE IF (ISCOMP(KT,'R').NE.0) THEN
C  Keyword that accepts real values.
            CALL GETPPR(COM,KEY,RPARAM)
            CALL R2NRTF(RPARAM,STR,LES)
            WRITE (NU,40) COM, KEY, STR(1:LES)
         ELSE IF (ISCOMP(KT,'S').NE.0) THEN
C  Keyword that accepts string values.
            CALL GETPPS(COM,KEY,STR,LES)
            WRITE (NU,50) COM, KEY, STR(1:LES)
         ELSE IF (ISCOMP(KT,'C').NE.0) THEN
C  Keyword that accepts character values.
            CALL GETPPS(COM,KEY,STR,LES)
            WRITE (NU,50) COM, KEY, STR(1:LES)
         END IF
30    CONTINUE
C
40    FORMAT (A,'  ',A,'  ',A,'   !')
50    FORMAT (A,'  ',A,'  ''',A,'''   !')
      END
C  ---------------------------------------------------------------------
      SUBROUTINE PICK ( MERR )
C  ---------------------------------------------------------------------
C
C  Description:
C    Implements the PICK command.
C    Syntax: PICK (Has panel)
C
C  Output arguments:
C    MERR       Error indicator.
C
C  ---------------------------------------------------------------------
      IMPLICIT $(TYPE) (A-H,O-Z)
C  Common blocks:
      PARAMETER ( MXV = $(MXV) )
      COMMON /PARAMS/ POINT(MXV), VAL
      COMMON /DIMEN/ NOD, NOF
      COMMON /MODE3/ MODEV, IFFORM, INMCL
      LOGICAL STRO, WEAK
      COMMON /PRILOG/ STRO, WEAK
      COMMON /UNITS/ IUINP, IUOUT
C  ---------------------------------------------------------------------
C  Local variables:
      PARAMETER ( MXFILE = 80 )
      CHARACTER*(MXFILE) FILE
      CHARACTER*10 TYPE, STR
      CHARACTER*23 CVAL
      CHARACTER CH
C  ---------------------------------------------------------------------
C
      MERR = 0
C
C  Present the panel and process any user changes.
      CALL SETPI('CANCEL',1,IERR)
      CALL CHANGE(MERR)
      IF (MERR.NE.0) RETURN
      CALL GETPI('CANCEL',ICANC)
      IF (ICANC.EQ.0) RETURN
C
C  Get panel parameters.
      CALL GETPS('FILE',FILE,LEF)
      CALL GETPS('TYPE',TYPE,LET)
      CALL GETPI('REC',IREC)
C
      IF (IREC.EQ.0) THEN
         IF (WEAK) WRITE (IUOUT,100)
         MERR = -2
         RETURN
      END IF
C
      LEF = LENGTH(FILE)
      IF (ISCOMP(TYPE,'TEXT').NE.0) THEN
         MODE = 1
      ELSE
         MODE = 2
      END IF
C
C  Open the file.
      CALL OPENIT(FILE,TYPE,0,0,NU,INSREC)
      IF (NU.EQ.-1) THEN
         MERR = -2
         RETURN
      END IF
C
C  Position the file according to IREC. A binary file does not need
C  positioning.
      IF (MODE.EQ.1) THEN
C  Read/skip IREC-1 records.
         DO 40,I=1,(NOD+1)*(IREC-1)
40          READ (NU,'(A)',END=50) CH
      ELSE
         IF (IREC.GT.INSREC) GOTO 50
      END IF
C
C  File is now correctly positioned at end of record IREC-1.
C  Read the record. Subroutine READER will take care of INDEF.
C  A non zero IEOF indicates EOF or a corrupt file.
      CALL BACKUP('X')
      CALL READER(NU,MODE,IREC,IEOF)
      IF (IEOF.NE.0) THEN
         IF (WEAK) THEN
            IF (ISTART.EQ.0) ISTART = 1
            CALL I2STR(ISTART,STR,LES)
            WRITE (IUOUT,10) FILE(1:LEF), STR(1:LES)
         END IF
         CLOSE (NU)
         MERR = -2
         RETURN
      END IF
      CLOSE (NU)
C
C  Check point in respect to margins (if any) and readjust it
C  if it is out of margins range.
      CALL POIMAR(IFEV)
C
C  Informative printout.
      IF (STRO) THEN
         CALL I2STR(IREC,STR,LES)
         WRITE (IUOUT,80) STR(1:LES), FILE(1:LEF)
         IF (MODEV.NE.0) THEN
            CALL R2RTF(VAL,CVAL,LEV)
            WRITE (IUOUT,90) CVAL(1:LEV)
         END IF
      END IF
      RETURN
C
C  End-of-File while reading a text file, or a binary file does not
C  contain that many records.
50    CONTINUE
      IF (WEAK) THEN
         IF (ISTART.EQ.0) ISTART = 1
         CALL I2STR(ISTART,STR,LES)
         WRITE (IUOUT,10) FILE(1:LEF), STR(1:LES)
      END IF
      CLOSE (NU)
      MERR = -2
C
10    FORMAT (/$(FC)'File "',A,'" does not contain ',A,' record(s) ',
     &        'or the file is corrupt.')
80    FORMAT (/$(FC)'Successfully picked record ',A,' from file "',
     &        A,'".')
90    FORMAT (/$(FC)'Current value: ',A)
100   FORMAT (/$(FC)'You must supply a record number or use the ',
     &        'inspect command.')
      END
C  ---------------------------------------------------------------------
      SUBROUTINE POICOM ( MERR )
C  ---------------------------------------------------------------------
C
C  Description:
C    Implements the POINT command.
C    Syntax: POINT index_specification value ...
C
C  Output arguments:
C    MERR       Error indicator.
C
C  Notes:
C    POINT is the only Merlin command that accepts one line input, asks
C    however for multiple lines.
C
C  ---------------------------------------------------------------------
      IMPLICIT $(TYPE) (A-H,O-Z)
C  Common blocks:
      PARAMETER ( MXV = $(MXV) )
      COMMON /PARAMS/ POINT(MXV), VAL
      COMMON /DIMEN/ NOD, NOF
      COMMON /MODE3/ MODEV, IFFORM, INMCL
      COMMON /TOTCAL/ ITNOCS, IPARTC, ITDER, IPDER, ITHESS, IPHESS,
     &                ITJAC, IPJAC
      LOGICAL STRO, WEAK
      COMMON /PRILOG/ STRO, WEAK
      COMMON /UNITS/ IUINP, IUOUT
C  ---------------------------------------------------------------------
C
      CALL ARGNO(NAR)
C
C  NTOT is the total number of variables that were assigned values in
C  this session of the POINT ommand.
      NTOT = 0
10    CONTINUE
C
C  Get and process the one line input.
      CALL CHPOI(NPOI,MERR)
      NTOT = NTOT+NPOI
      IF (NAR.EQ.0 .AND. INMCL.EQ.0) THEN
         IF (MERR.EQ.0) THEN
            IF (NPOI.NE.0) THEN
               IF (STRO) WRITE (IUOUT,20)
               GOTO 10
            END IF
         ELSE
            IF (STRO) WRITE (IUOUT,20)
            IF (ISIAF().NE.0) GOTO 10
         END IF
      END IF
C
      IF (NTOT.NE.0) THEN
         IF (ITNOCS.EQ.0) THEN
C  Check for undefined variables.
            CALL UNDEF(IGO,1)
            IF (IGO.EQ.0) RETURN
         END IF
C
C  Check if the point lies inside the margins.
         CALL POIMAR(IFEV)
C
C  If POIMAR did not call ACSQ, do it here.
         IF (IFEV.EQ.0 .AND. MODEV.EQ.1) VAL = ACSQ(POINT,NOD)
      END IF
C
20    FORMAT ($(FC)'POINT')
      END
C  ---------------------------------------------------------------------
      SUBROUTINE PONOFF ( IONOFF, MERR )
C  ---------------------------------------------------------------------
C
C  Description:
C    Implements the PANELON and PANELOFF commands.
C    Syntax: PANELON [ panel ... ]
C            PANELOFF [ panel ... ]
C
C  Input arguments:
C    IONOFF     Selects the panel status:
C               IONOFF = 0 -> Turns panels off.
C               IONOFF = 1 -> Turns panels on.
C
C  Output arguments:
C    MERR       Error indicator.
C
C  ---------------------------------------------------------------------
      IMPLICIT $(TYPE) (A-H,O-Z)
C  Common blocks:
      LOGICAL STRO, WEAK
      COMMON /PRILOG/ STRO, WEAK
      COMMON /UNITS/ IUINP, IUOUT
C  Maximum length of a keyword.
      PARAMETER ( KEYLEN = 10 )
C  Maximum number of panel commands.
      PARAMETER ( MXPCOM = 80 )
C  Maximum number of keywords in all panel commands.
      PARAMETER ( MXPKEY = 1000 )
C  Maximum storage for integer/real keywords.
      PARAMETER ( MXPIR = 5000 )
C  Maximum storage for string keywords and descriptions.
      PARAMETER ( MXPSTR = 20000 )
C  Maximum number of pointers in the above string storage.
      PARAMETER ( MXPPST = 5000 )
C
      CHARACTER*(KEYLEN) PCOM(MXPCOM)
      INTEGER            PONF(MXPCOM)
      INTEGER            KPOINT(MXPCOM)
      INTEGER            NCHELP(2*MXPCOM)
      CHARACTER*(KEYLEN) PKEY(MXPKEY)
      INTEGER            NKHELP(2*MXPKEY)
      CHARACTER          KTYPE(MXPKEY)
      INTEGER            TPOINT(MXPKEY)
      CHARACTER          KIO(MXPKEY)
      $(TYPE)            PIR(MXPIR)
      INTEGER            SPOINT(MXPPST)
      CHARACTER*(MXPSTR) PANSTR
      COMMON /PCOMS1/ KPOINT, NCHELP, NKHELP, TPOINT, SPOINT, PONF
      COMMON /PCOMS2/ PIR
      COMMON /PCOMS3/ PKEY, KTYPE, PANSTR, PCOM, KIO
C
C  NPCOM: Number of panel commands in the panel description file.
C  NPKEY: Number of keywords in all panel commands.
C  NPIR:  Number of positions filled in array PIR.
C  NPSP:  Number of positions filled in the char-pointer array SPOINT.
C  NPCH:  Number of character positions filled in PANSTR.
      COMMON /PCOMS4/ NPCOM, NPKEY, NPIR, NPSP, NPCH
C  ---------------------------------------------------------------------
C  Local variables:
      CHARACTER*80 PAN
      CHARACTER*4 ONOFF(0:1)
      DATA ONOFF / 'off.', 'on. ' /
C  ---------------------------------------------------------------------
C
      MERR = 0
      CALL ARGNO(NA)
      IF (NA.EQ.0) THEN
C
C  No command line arguments. Process all panels.
         DO 10,I=1,NPCOM
10          PONF(I) = IONOFF
C
C  Informative printout.
         IF (STRO) WRITE (IUOUT,20) ONOFF(IONOFF)
      ELSE
C
C  Some command line arguments were entered. Process these panels
C  only.
         DO 30,I=1,NA
            CALL CMDARG(I,PAN,LEP)
C  Is it a valid Merlin command ?
            CALL LOCC(PAN,IC)
            IF (IC.EQ.0) THEN
               IF (WEAK) WRITE (IUOUT,40) PAN(1:LEP)
               MERR = -2
               RETURN
            END IF
C  Does it have a panel ?
            CALL PLOCP(PAN,IP)
            IF (IP.EQ.0) THEN
               IF (WEAK) WRITE (IUOUT,50) PAN(1:LEP)
               MERR = -2
            ELSE
               PONF(IP) = IONOFF
               IF (STRO) WRITE (IUOUT,60)
     &                            PAN(1:LEP), ONOFF(IONOFF)
            END IF
30       CONTINUE
      END IF
C
20    FORMAT (/$(FC)'All panels are now ',A)
40    FORMAT (/$(FC)'There is no such Merlin command: "',A,'".')
50    FORMAT (/$(FC)'The "',A,'" command has no panel.')
60    FORMAT (/$(FC)'The "',A,'" panel is now ',A)
      END
C  ---------------------------------------------------------------------
      SUBROUTINE PSGRAF ( MERR )
C  ---------------------------------------------------------------------
C
C  Description:
C    Implements the PSGRAPH command.
C    Syntax: PSGRAPH (Has panel).
C
C  Output arguments:
C    MERR       Error indicator.
C
C  ---------------------------------------------------------------------
      IMPLICIT $(TYPE) (A-H,O-Z)
C  Common blocks:
      PARAMETER ( MXV = $(MXV) )
      COMMON /PARAMS/ POINT(MXV), VAL
      COMMON /DETAIL/ IXAT(MXV), MARG(MXV)
      COMMON /XLIMS/ XLL(MXV), XRL(MXV)
      COMMON /RWORK/ XT1(MXV), XT2(MXV), XT3(MXV), XT4(MXV), XT5(MXV),
     &               XT6(MXV), XT7(MXV+1)
      COMMON /DIMEN/ NOD, NOF
      LOGICAL STRO, WEAK
      COMMON /PRILOG/ STRO, WEAK
      COMMON /UNITS/ IUINP, IUOUT
      COMMON /TOTCAL/ ITNOCS, IPARTC, ITDER, IPDER, ITHESS, IPHESS,
     &                ITJAC, IPJAC
      COMMON /MODE3/ MODEV, IFFORM, INMCL
C  ---------------------------------------------------------------------
C  Local variables:
      PARAMETER ( LINLEN = 80 )
      CHARACTER*(LINLEN) XTITLE, YTITLE
      PARAMETER ( MXFILE = 80 )
      CHARACTER*(MXFILE) FILE, DAFILE
      CHARACTER WHAT
      CHARACTER*20 STR, STR2
      PARAMETER ( MAXDAT = 1000, MAXCRV = 10 )
      DIMENSION ISARR(MAXCRV), ICARR(MAXCRV), LTARR(MAXCRV)
      DIMENSION COL(MAXCRV,3), COLL(2,3), DAT(MAXDAT,MAXCRV+1)
      DIMENSION XPS(MAXDAT,MAXCRV+1), W1(MAXCRV), W2(MAXCRV)
      DIMENSION W3(MAXCRV), W4(MAXCRV)
      LOGICAL OPE
C  ---------------------------------------------------------------------
C
C  Present the panel and process any user changes.
      CALL SETPI('CANCEL',1,IERR)
      CALL CHANGE(MERR)
      IF (MERR.NE.0) RETURN
      CALL GETPI('CANCEL',ICANC)
      IF (ICANC.EQ.0) RETURN
C
C  Get panel parameters.
      CALL GETPS('WHAT',WHAT,LEW)
      CALL GETPI('SYMBOL',ISYM)
      XTITLE = ' '
      CALL GETPS('XTITLE',XTITLE,LENX)
      YTITLE = ' '
      CALL GETPS('YTITLE',YTITLE,LENY)
      CALL GETPI('FONTSIZE',IFONT)
      CALL GETPI('GRID',IGRID)
      CALL GETPI('FRAME',IFRAME)
      CALL GETPI('CONNECT',ICONN)
      CALL GETPS('FILE',FILE,LEF)
      LEF = LENGTH(FILE)
      CALL GETPI('XTICKS',NXT)
      CALL GETPI('YTICKS',NYT)
      CALL GETPI('PAPER',IPAP)
C
      NXT = NXT+1
      NYT = NYT+1
C
C  Prepare the data points.
      IF (ISCOMP(WHAT,'F').NE.0) THEN
C
C  Plot the objective function.
         IF (ITNOCS.EQ.0 .OR. MODEV.EQ.0) RETURN
         CALL GETPI('INDEX',IDX)
         CALL GETPI('NOP',NOP)
         CALL GETPR('FROM',FROM)
         CALL GETPR('TO',TO)
C
C  FROM <= TO ?
         IF (FROM.GT.TO) THEN
            TMP = FROM
            FROM = TO
            TO = TMP
         END IF
C
C  Is there enough storage for the data points ?
         IF (NOP.GT.MAXDAT) THEN
            IF (WEAK) THEN
               CALL I2STR(MAXDAT,STR,LES)
               WRITE (IUOUT,40) STR(1:LES)
            END IF
            MERR = -2
            RETURN
         END IF
C
C  Perform some checks on the input parameters.
C  Is IDX valid index ?
         IF (IDX.LT.1 .OR. IDX.GT.NOD) THEN
            IF (WEAK) THEN
               CALL I2STR(NOD,STR,lES)
               WRITE (IUOUT,30) STR(1:LES)
            END IF
            MERR = -2
            RETURN
         END IF
C
C  Is it a fixed variable ?
         IF (IXAT(IDX).EQ.0) THEN
            IF (WEAK) THEN
               CALL VNAME(IDX,STR,LES)
               WRITE (IUOUT,50) STR(1:LES)
            END IF
            MERR = -2
            RETURN
         END IF
C
C  Adjust FROM, TO if out of margins.
         MA = MARG(IDX)
         IF (MA.EQ.-1 .OR. MA.EQ.2) THEN
            IF (FROM.LT.XLL(IDX)) THEN
               FROM = XLL(IDX)
               CALL VNAME(IDX,STR,LES)
               WRITE (IUOUT,60) STR(1:LES)
            END IF
         ELSE IF (MA.EQ.1 .OR. MA.EQ.2) THEN
            IF (TO.GT.XRL(IDX)) THEN
               TO = XRL(IDX)
               CALL VNAME(IDX,STR,LES)
               WRITE (IUOUT,70) STR(1:LES)
            END IF
         END IF
C
         ST = (TO-FROM)/(NOP-1)
         DO 100,I=1,NOD
100         XT1(I) = POINT(I)
         DO 110,I=1,NOP
            XX = FROM+(I-1)*ST
            XT1(IDX) = XX
            DAT(I,1) = XX
            DAT(I,2) = ACSQ(XT1,NOD)
110      CONTINUE
         NOFCOL = 2
C
C  Symbol for the data points.
         ISARR(1) = ISYM
C
C  Line type.
         LTARR(1) = 1
      ELSE IF (ISCOMP(WHAT,'D').NE.0) THEN
C
C  Plot a datafile.
         CALL GETPS('DATAFILE',DAFILE,LEDF)
         LEDF = LENGTH(DAFILE)
         CALL OPENGF(DAFILE,1,NU)
         IF (NU.EQ.-1) THEN
            MERR = -2
            RETURN
         END IF
C
C  Determine number of rows and columns.
         CALL DFESTI(NU,NROW,NCOL)
C
C  Empty data file ?
         IF (NROW.EQ.0 .OR. NCOL.EQ.0) THEN
            IF (WEAK) THEN
               WRITE (IUOUT,90) DAFILE(1:LEDF)
               MERR = -2
               CLOSE (NU)
               RETURN
            END IF
         END IF
C
C  Printout rows, columns.
         IF (STRO) THEN
            CALL I2STR(NROW,STR,LES)
            CALL I2STR(NCOL,STR2,LES2)
            WRITE (IUOUT,80) STR(1:LES), STR2(1:LES2)
         END IF
C
C  Too many points or columns ?
         MERR = 0
         IF (NROW.GT.MAXDAT) THEN
            CALL I2STR(MAXDAT,STR,LES)
            IF (WEAK) WRITE (IUOUT,130) STR(1:LES)
            MERR = -2
         END IF
         IF (NCOL.GT.MAXCRV+1) THEN
            CALL I2STR(MAXCRV,STR,LES)
            IF (WEAK) WRITE (IUOUT,120) STR(1:LES)
            MERR = -2
         END IF
         IF (MERR.NE.0) THEN
            CLOSE (NU)
            RETURN
         END IF
C
C  Read in the data points.
         CALL DFRIN(NU,DAT,MAXDAT,NCOL,MERR)
         IF (MERR.NE.0) THEN
            CLOSE (NU)
            RETURN
         END IF
         NOFCOL = NCOL
         NOP = NROW
C
C  Sort with respect to X.
         CALL PSGSO(DAT,MAXDAT,NROW,NCOL)
         CLOSE (NU)
C
C  Symbol for the data points.
         DO 140,I=1,NCOL-1
            IF (ISYM.NE.0) THEN
               ISARR(I) = MOD(I,6)
            ELSE
               ISARR(I) = 0
            END IF
140      CONTINUE
C
C  Line type.
         DO 150,I=1,NCOL-1
150         LTARR(I) = MOD(I,3)
      END IF
C
C  Connect the data points ?
      DO 180,I=1,NOFCOL-1
180      ICARR(I) = ICONN
      DO 10,I=1,3
C  Color for the curve and data points (black).
         DO 160,J=1,NOFCOL-1
160         COL(J,I) = 0.
C  Color for the X-label (black).
         COLL(1,I) = 0.
C  Color for the Y-label (black).
         COLL(2,I) = 0.
10    CONTINUE
C
C  Allocate unit for the file.
      IU = NUNIT()
      IF (IU.EQ.-1) THEN
         CALL ERRUAL(FILE)
         MERR = -2
         RETURN
      END IF
C
C  Get info for the file.
      INQUIRE (FILE=FILE,OPENED=OPE,NUMBER=NUM,IOSTAT=IOE)
      IF (IOE.NE.0) THEN
         CALL ERRINQ(FILE,IOE)
         MERR = -2
         RETURN
      END IF
C
C  Is it open ?
      IF (OPE) THEN
         CALL ERRBUS(FILE,NUM)
         MERR = -2
         RETURN
      END IF
C
C  Open the file.
      OPEN (UNIT=IU,FILE=FILE,IOSTAT=IOE)
      IF (IOE.NE.0) THEN
         CALL ERROPE(FILE,IOE)
         MERR = -2
         RETURN
      END IF
      CALL PSGR1(XTITLE(1:LENX),YTITLE(1:LENY),ISARR,IFRAME,IGRID,IFONT,
     &           NXT,NYT,NOFCOL,DAT,NOP,ICARR,LTARR,COL,COL,COLL,IU,
     &           MAXDAT,MAXCRV+1,IPAP,XPS,W1,W2,W3,W4)
      CLOSE (IU)
      IF (STRO) WRITE (IUOUT,20) FILE(1:LEF)
C
20    FORMAT (/$(FC)'The Postscript graph has been disposed in ',
     &        'file "',A,'".')
30    FORMAT (/$(FC)'Error: Index should be in between 1 and ',A,'.')
40    FORMAT (/$(FC)'Error: Too many data points. Maximum is ',A,'.')
50    FORMAT (/$(FC)'Error: Variable ',A,' is fixed. Not plotted.')
60    FORMAT (/$(FC)'"FROM" was less than the left margin of ',
     &        'variable ',A,'.'
     &        /$(FC)'It has been adjusted.')
70    FORMAT (/$(FC)'"TO" was greater than the right margin of ',
     &        'variable ',A,'.'
     &        /$(FC)'It has been adjusted.')
80    FORMAT (/$(FC)A,' data points, ',A,' data column(s).')
90    FORMAT (/$(FC)'The first line of file "',A,'" is empty.')
120   FORMAT (/$(FC)'Too many data columns. Maximum is ',A,'.')
130   FORMAT (/$(FC)'Too many data points. Maximum is ',A,'.')
      END
C  ---------------------------------------------------------------------
      SUBROUTINE PSGR1 ( XLABEL, YLABEL, SYMBOL, FRAME, GRIDLI, MFSIZE,
     &                   NX, NY, NOFCOL, GDAT, NOFORW, LINE, DASHTY,
     &                   COLOR, SCOLOR, RCOLOR, IUNI, MAXROW, MAXCOL,
     &                   IPGSIZ, PSDATA, PY, YN, FMAX, FMIN )
C  ---------------------------------------------------------------------
C
C  Description:
C    This is the main routine that creates the Postscript graph. It
C    can simultaneously plot several data sets (curves), using different
C    options for each one of them. Merlin however, always plots only one
C    curve and does not exploit color.
C    The PSGRAPH engine is due to P. Kiriakakis.
C
C  Input arguments:
C    XLABEL     X-axis label.
C    YLABEL     Y-axis label.
C    SYMBOL     Array of integers. SYMBOL(I) is the symbol to be used
C               for the points of data set I.
C               SYMBOL(I) = 0 -> No symbol.
C               SYMBOL(I) = 1 -> Star.
C               SYMBOL(I) = 2 -> Square.
C               SYMBOL(I) = 3 -> Triangle.
C               SYMBOL(I) = 4 -> Plus.
C               SYMBOL(I) = 5 -> Filled square.
C               SYMBOL(I) = 6 -> Filled triangle.
C    FRAME      FRAME = 0 -> No frame around the graph.
C               FRAME = 1 -> Plot a frame around the graph.
C    GRIDLI     GRIDLI = 0 -> Don't plot any gridlines.
C               GRIDLI = 1 -> Plot gridlines.
C    MFSIZE     Font size for the labels.
C    NX         Number of grid lines on X-axis
C    NY         Number of grid lines on Y-axis.
C    NOFCOL     Number of columns in the data array GDAT.
C    GDAT       The data array. The first coloumn contains the X-points,
C               while the others are the
C    NOFORW     Number of rows (data points) in the GDAT array.
C    LINE       LINE(I) determines whether the points of data set I
C               will be connected.
C               LINE(I) = 0 -> Do not connect the data points of set I.
C               LINE(I) = 1 -> Connect the data points of set I.
C    DASHTY     DASHTY(I) determines the line style to be used when
C               connecting the points of data set I.
C               DASHTY(I) = 1 -> Solid line.
C               DASHTY(I) = 2 -> Dashed line.
C               DASHTY(I) = 3 -> Another dashed line.
C    COLOR      COLOR(I,J) J=1,3 determines the color to be used for
C               curve I. Colors are specified in the RGB system, with
C               each component being a real number in between 0 and 1.
C               (R,G,B) = (0,0,0) is black,, whereas (R,G,B) = (1,1,1)
C               is white.
C               COLOR(I,1) -> Red.
C               COLOR(I,2) -> Green.
C               COLOR(I,3) -> Blue.
C    SCOLOR     SCOLOR(I,J) J=1,3 determines the color to obe used for
C               the points of curve I. See the description of COLOR for
C               allowed values.
C    RCOLOR     RCOLOR(I,) J=1,3 determines the color to be used for
C               the labels. See the description of COLOR for allowed
C               values.
C               RCOLOR(1,J) J=1,3 -> Color for the X-label.
C               RCOLOR(2,J) J=1,3 -> Color for the Y-label.
C    IUNI       The unit number of a file to receive the Postscript
C               code. The file must be opened and is not closed on
C               return.
C    MAXROW     Leading dimension of matrices GDAT and PSDATA.
C    MAXCOL     Maximum number of columns is matrices GDAT and PSDATA.
C    IPGSIZ     Determines the paper size:
C               IPGSIZ = 0 -> A4      8.26" x 11.69"
C               IPGSIZ = 1 -> B5      7.16" x 10.12"
C               IPGSIZ = 2 -> Letter  8.5"  x 11"
C               IPGSIZ = 3 -> Legal   8.5"  x 14"
C    PSDATA     Work array of length NOFROW x NOFCOL.
C    PY         Work array of length NOFCOL-1.
C    YN         Work array of length NOFCOL-1.
C    FMAX       Work array of length NOFCOL-1.
C    FMIN       Work array of length NOFCOL-1.
C
C  ---------------------------------------------------------------------
      IMPLICIT $(TYPE) (A-H,O-Z)
C  Arguments:
      CHARACTER*(*) XLABEL, YLABEL
      INTEGER SYMBOL(MAXCOL-1), FRAME, GRIDLI, LINE(MAXCOL-1)
      INTEGER DASHTY(MAXCOL-1)
      DIMENSION COLOR(MAXCOL-1,3), SCOLOR(MAXCOL-1,3), RCOLOR(2,3)
      DIMENSION GDAT(MAXROW,MAXCOL)
      DIMENSION PY(MAXCOL-1), YN(MAXCOL-1)
      DIMENSION FMAX(MAXCOL-1), FMIN(MAXCOL-1)
      DIMENSION PSDATA(MAXROW,MAXCOL)
C  ---------------------------------------------------------------------
C  Local variables:
      INTEGER XAXLEN, YAXLEN
      INTEGER XCOUNT
C  ---------------------------------------------------------------------
C
      IF (IPGSIZ.EQ.0) THEN
         XAXLEN = 560
         YAXLEN = 420
         MXBASE = 140
         MYBASE = 60
         ITRANS = 539
      ELSE IF (IPGSIZ.EQ.1) THEN
         XAXLEN = 520
         YAXLEN = 400
         MXBASE = 120
         MYBASE = 50
         ITRANS = 490
      ELSE IF (IPGSIZ.EQ.2) THEN
         XAXLEN = 560
         YAXLEN = 420
         MXBASE = 140
         MYBASE = 60
         ITRANS = 550
      ELSE IF (IPGSIZ.EQ.3) THEN
         XAXLEN = 640
         YAXLEN = 460
         MXBASE = 180
         MYBASE = 60
         ITRANS = 580
      END IF
C
      NOFY = NOFCOL-1
      XCOUNT = NOFORW
C
      CALL PSGR2(GDAT,PSDATA,XCOUNT,NOFY,PXMAX,PXMIN,PYMAX,PYMIN,
     &           XAXLEN,YAXLEN,MAXROW,MAXCOL,PY,YN,FMAX,FMIN)
C
      WRITE (IUNI,1000) '%!PS-Adobe-2.0'
      WRITE (IUNI,1000) '%%Version: 3.15'
      WRITE (IUNI,1000) '%%DocumentFonts: (atend)'
      WRITE (IUNI,1000) '%%Pages: (atend)'
      WRITE (IUNI,1000) '%%EndComments'
      WRITE (IUNI,1010) ITRANS
      CALL DRAWAX(XAXLEN,YAXLEN,MXBASE,MYBASE,IUNI)
      IF (FRAME.EQ.1) CALL PUTFRA(XAXLEN,YAXLEN,MXBASE,MYBASE,IUNI)
      IF (GRIDLI.EQ.1)
     &    CALL PUTGRI(NX,NY,XAXLEN,YAXLEN,MXBASE,MYBASE,IUNI)
      CALL TICMRK(NX,NY,PXMAX,PXMIN,PYMAX,PYMIN,XAXLEN,YAXLEN,MXBASE,
     &            MYBASE,IUNI)
      CALL SETFNT(MFSIZE,IUNI)
      CALL YAXLAB(YLABEL,YAXLEN,RCOLOR(1,1),RCOLOR(1,2),RCOLOR(1,3),
     &            MXBASE,MYBASE,IUNI)
      CALL XAXLAB(XLABEL,XAXLEN,RCOLOR(2,1),RCOLOR(2,2),RCOLOR(2,3),
     &            MXBASE,MYBASE,IUNI)
C
      DO 200,MCURCO=1,NOFY
         IF (SYMBOL(MCURCO).NE.0)
     &          CALL SHOWPT(PSDATA,XCOUNT,MCURCO+1,SYMBOL(MCURCO),
     &                      SCOLOR(MCURCO,1),SCOLOR(MCURCO,2),
     &                      SCOLOR(MCURCO,3),MXBASE,MYBASE,IUNI,
     &                      MAXROW,MAXCOL)
         IF (LINE(MCURCO).NE.0)
     &          CALL CONPTS(PSDATA,XCOUNT,MCURCO+1,DASHTY(MCURCO),
     &                      COLOR(MCURCO,1),COLOR(MCURCO,2),
     &                      COLOR(MCURCO,3),MXBASE,MYBASE,IUNI,
     &                      MAXROW,MAXCOL)
200   CONTINUE
      CALL SHOWPG(IUNI)
C
1000  FORMAT (A)
1010  FORMAT (I4,' 0 translate')
      END
C  ---------------------------------------------------------------------
      SUBROUTINE PSGR2 ( GDAT, PSDATA, XCOUNT, NOFY, PXMAX, PXMIN,
     &                   PYMAX, PYMIN, XAXLEN, YAXLEN, MAXROW, MAXCOL,
     &                   PY, YN, FMAX, FMIN )
C  ---------------------------------------------------------------------
C
C  Description:
C    Called by subroutine PSGR2. Prepares the Postscript coordinates.
C
C  ---------------------------------------------------------------------
      IMPLICIT $(TYPE) (A-H,O-Z)
C  Arguments:
      INTEGER XCOUNT, XAXLEN, YAXLEN
      DIMENSION PY(MAXCOL-1), YN(MAXCOL-1), GDAT(MAXROW,MAXCOL)
      DIMENSION FMIN(MAXCOL-1), FMAX(MAXCOL-1)
      DIMENSION PSDATA(MAXROW,MAXCOL)
C  ---------------------------------------------------------------------
C
      CALL GETXMM(GDAT,XCOUNT,XMAX,XMIN,MAXROW,MAXCOL)
      CALL GETYMM(GDAT,XCOUNT,NOFY,YMAX,YMIN,MAXROW,MAXCOL,FMAX,FMIN)
      PXMAX = XMAX
      PXMIN = XMIN
      PYMAX = YMAX
      PYMIN = YMIN
      IF (PXMIN.EQ.PXMAX) THEN
         IF (PXMIN.EQ.0.) THEN
            PXMAX = 1
            PXMIN = -1
         ELSE
            PXMAX = PXMIN + 0.1*PXMIN
            PXMIN = PXMIN - 0.1*PXMIN
         END IF
      END IF
      IF (PYMIN.EQ.PYMAX) THEN
         IF (PYMIN.EQ.0.) THEN
            PYMAX = 1
            PYMIN = -1
         ELSE
            PYMAX = PYMIN + 0.1*PYMIN
            PYMIN = PYMIN - 0.1*PYMIN
         END IF
      END IF
      AY = (PYMAX-PYMIN)/YAXLEN
      AX = (PXMAX-PXMIN)/XAXLEN
C
      DO 40,I1=1,XCOUNT
         PX = GDAT(I1,1)
         DO 27,I2=1,NOFY
27          PY(I2) = GDAT(I1,I2+1)
         XN = ((PX-PXMIN)/AX)
         DO 30,I3=1,NOFY
30          YN(I3) = ((PY(I3)-PYMIN)/AY)
         PSDATA(I1,1) = XN
         DO 37,I4=1,NOFY
37          PSDATA(I1,I4+1) = YN(I4)
40    CONTINUE
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE GETYMM ( GDAT, XCOUNT, NOFY, YMAX, YMIN, MAXROW,
     &                    MAXCOL, FMAX, FMIN )
C  ---------------------------------------------------------------------
C
C  Description:
C    Part of the Postscript graph routine PSGR1. It returns the minimum
C    and maximum Y-value of all data sets.
C
C  ---------------------------------------------------------------------
      IMPLICIT $(TYPE) (A-H,O-Z)
C  Arguments:
      INTEGER XCOUNT
      DIMENSION FMAX(MAXCOL-1), FMIN(MAXCOL-1), GDAT(MAXROW,MAXCOL)
C  ---------------------------------------------------------------------
C
      DO 20,I=1,NOFY
         TYMAX = GDAT(1,2)
         TYMIN = GDAT(1,2)
         DO 10,II=1,XCOUNT
            IF (GDAT(II,I+1).GT.TYMAX) TYMAX = GDAT(II,I+1)
            IF (GDAT(II,I+1).LT.TYMIN) TYMIN = GDAT(II,I+1)
10       CONTINUE
         FMAX(I) = TYMAX
         FMIN(I) = TYMIN
20    CONTINUE
      T = FMAX(1)
      TT = FMIN(1)
      DO 30,I=1,NOFY
         IF (FMAX(I).GT.T) T = FMAX(I)
         IF (FMIN(I).LT.TT) TT = FMIN(I)
30    CONTINUE
      YMAX = T
      YMIN = TT
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE GETXMM ( GDAT, XCOUNT, XMAX, XMIN, MAXROW, MAXCOL )
C  ---------------------------------------------------------------------
C
C  Description:
C    Part of the Postscript graph routine PSGR1. It returns the limits
C    of the X-axis.
C
C  ---------------------------------------------------------------------
      IMPLICIT $(TYPE) (A-H,O-Z)
C  Arguments:
      INTEGER XCOUNT
      DIMENSION GDAT(MAXROW,MAXCOL)
C  ---------------------------------------------------------------------
C
      XMIN = GDAT(1,1)
      XMAX = GDAT(XCOUNT,1)
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE DRAWAX ( XAXLEN, YAXLEN, MXBASE, MYBASE, IUNI )
C  ---------------------------------------------------------------------
C
C  Description:
C    Part of the Postscript graph routine PSGR1. It creates postscript
C    code that draws the X and Y-axis.
C
C  ---------------------------------------------------------------------
C  Arguments:
      INTEGER XAXLEN, YAXLEN
C  ---------------------------------------------------------------------
C
      WRITE (IUNI,10) '% draws axis'
      WRITE (IUNI,10) '90 rotate'
      WRITE (IUNI,10) 'newpath'
      WRITE (IUNI,10) '% x-axis'
      WRITE (IUNI,1030) MXBASE,MYBASE
      WRITE (IUNI,1000) XAXLEN
      WRITE (IUNI,10) '% y-axis'
      WRITE (IUNI,1030) MXBASE,MYBASE
      WRITE (IUNI,1020) YAXLEN
      WRITE (IUNI,10) 'closepath'
      WRITE (IUNI,10) 'stroke'
C
10    FORMAT (A)
1000  FORMAT (I3,' 0 rlineto')
1020  FORMAT ('0 ',I3,' rlineto')
1030  FORMAT (I3,' ',I3,' moveto')
      END
C  ---------------------------------------------------------------------
      SUBROUTINE YJUST ( IUNI )
C  ---------------------------------------------------------------------
C
C  Description:
C    Part of the Postscript graph routine PSGR1.
C
C  Input arguments:
C    IUNI       Unit number to receive output.
C
C  ---------------------------------------------------------------------
C
      WRITE (IUNI,10) '/checkyj'
      WRITE (IUNI,10) '{'
      WRITE (IUNI,10) ' /yc exch def'
      WRITE (IUNI,10) ' /xc exch def'
      WRITE (IUNI,10) ' dup'
      WRITE (IUNI,10) ' /st1 exch def'
      WRITE (IUNI,10) ' /st2 exch def'
      WRITE (IUNI,10) ' st1 ( ) search'
      WRITE (IUNI,10) ' {'
      WRITE (IUNI,10) '   /s1 exch def'
      WRITE (IUNI,10) '   /s2 exch def'
      WRITE (IUNI,10) '   /s3 exch def'
      WRITE (IUNI,10) '   s1 () eq'
      WRITE (IUNI,10) '   {s3 xc yc yjustify}'
      WRITE (IUNI,10) '   {st2 xc yc yjustify} ifelse'
      WRITE (IUNI,10) '  }'
      WRITE (IUNI,10) '  {st2 xc yc yjustify} ifelse'
      WRITE (IUNI,10) '} def'
      WRITE (IUNI,10) '% subroutine yjustify'
      WRITE (IUNI,10) '/yjustify'
      WRITE (IUNI,10) '{'
      WRITE (IUNI,10) ' /ycor exch def'
      WRITE (IUNI,10) ' /xcor exch def'
      WRITE (IUNI,10) ' dup'
      WRITE (IUNI,10) ' /string1 exch def'
      WRITE (IUNI,10) ' /string2 exch def'
      WRITE (IUNI,10) ' string1 ( ) search'
      WRITE (IUNI,10) ' {'
      WRITE (IUNI,10) '  dup'
      WRITE (IUNI,10) '  /str1 exch def'
      WRITE (IUNI,10) '  /str exch def'
      WRITE (IUNI,10) '  /str2 def'
      WRITE (IUNI,10) '  /str3 def'
      WRITE (IUNI,10) '  str1 stringwidth'
      WRITE (IUNI,10) '  pop xcor exch sub ycor moveto'
      WRITE (IUNI,10) '  str show'
      WRITE (IUNI,10) ' }'
      WRITE (IUNI,10) ' {'
      WRITE (IUNI,10) '  string1 stringwidth'
      WRITE (IUNI,10) '  pop xcor exch sub ycor moveto'
      WRITE (IUNI,10) '  string2 show'
      WRITE (IUNI,10) ' } ifelse'
      WRITE (IUNI,10) '} def'
C
10    FORMAT (A)
      END
C  ---------------------------------------------------------------------
      SUBROUTINE XJUST ( IUNI )
C  ---------------------------------------------------------------------
C
C  Description:
C    Part of the Postscript graph routine PSGR1.
C
C  Input arguments:
C    IUNI       Unit number to receive output.
C
C  ---------------------------------------------------------------------
C
      WRITE (IUNI,10) '/checkxj'
      WRITE (IUNI,10) '{'
      WRITE (IUNI,10) ' /yc exch def'
      WRITE (IUNI,10) ' /xc exch def'
      WRITE (IUNI,10) ' dup'
      WRITE (IUNI,10) ' /st1 exch def'
      WRITE (IUNI,10) ' /st2 exch def'
      WRITE (IUNI,10) ' st1 ( ) search'
      WRITE (IUNI,10) ' {'
      WRITE (IUNI,10) '   /s1 exch def'
      WRITE (IUNI,10) '   /s2 exch def'
      WRITE (IUNI,10) '   /s3 exch def'
      WRITE (IUNI,10) '   s1 () eq'
      WRITE (IUNI,10) '   {s3 xc yc xjustify}'
      WRITE (IUNI,10) '   {st2 xc yc xjustify} ifelse'
      WRITE (IUNI,10) '  }'
      WRITE (IUNI,10) '  {st2 xc yc xjustify} ifelse'
      WRITE (IUNI,10) '} def'
      WRITE (IUNI,10) '% subroutine xjustify'
      WRITE (IUNI,10) '/xjustify'
      WRITE (IUNI,10) '{'
      WRITE (IUNI,10) ' /ycor exch def'
      WRITE (IUNI,10) ' /xcor exch def'
      WRITE (IUNI,10) ' dup'
      WRITE (IUNI,10) ' /string1 exch def'
      WRITE (IUNI,10) ' /string2 exch def'
      WRITE (IUNI,10) ' string1 ( ) search'
      WRITE (IUNI,10) ' {'
      WRITE (IUNI,10) '  dup'
      WRITE (IUNI,10) '  /str1 exch def'
      WRITE (IUNI,10) '  /str exch def'
      WRITE (IUNI,10) '  /str2 def'
      WRITE (IUNI,10) '  /str3 def'
      WRITE (IUNI,10) '  str1 stringwidth'
      WRITE (IUNI,10) '  pop 2 div xcor exch sub ycor moveto'
      WRITE (IUNI,10) '  str show'
      WRITE (IUNI,10) ' }'
      WRITE (IUNI,10) ' {'
      WRITE (IUNI,10) '  string1 stringwidth'
      WRITE (IUNI,10) '  pop 2 div xcor exch sub ycor moveto'
      WRITE (IUNI,10) '  string2 show'
      WRITE (IUNI,10) ' } ifelse'
      WRITE (IUNI,10) '} def'
C
10    FORMAT (A)
      END
C  ---------------------------------------------------------------------
      SUBROUTINE TICMRK ( NX, NY, PXMAX, PXMIN, PYMAX, PYMIN, XAXLEN,
     &                    YAXLEN, MXBASE, MYBASE, IUNI )
C  ---------------------------------------------------------------------
C
C  Description:
C    Part of the Postscript graph routine PSGR1. It writes postscript
C    code that draws the tickmarks on the X and Y-axis.
C
C  ---------------------------------------------------------------------
      IMPLICIT $(TYPE) (A-H,O-Z)
C  Arguments:
C  ---------------------------------------------------------------------
      INTEGER XAXLEN, YAXLEN
C  Local variables:
      CHARACTER*20 NUM
C  ---------------------------------------------------------------------
C
      STEPX = (PXMAX-PXMIN)/NX
      STEPY = (PYMAX-PYMIN)/NY
      DMX = REAL(XAXLEN)/NX
      DMY = REAL(YAXLEN)/NY
      MX = INT(DMX)
      MY = INT(DMY)
      WRITE (IUNI,30) 'newpath'
      WRITE (IUNI,30) '% x-axis ticks'
      WRITE (IUNI,1000) MXBASE,MYBASE
      WRITE (IUNI,1010) MXBASE,MX,XAXLEN+MXBASE,MYBASE
      WRITE (IUNI,30) '0 -6 rlineto}for'
      WRITE (IUNI,30) '% y-axis ticks'
      WRITE (IUNI,1000) MXBASE, MYBASE
      WRITE (IUNI,1030) MYBASE, MY, YAXLEN+MYBASE, MXBASE
      WRITE (IUNI,30) '-6 0 rlineto}for'
      WRITE (IUNI,30) 'closepath'
      WRITE (IUNI,30) 'stroke'
      CALL SETFNT(10,IUNI)
      CALL XJUST(IUNI)
      WRITE (IUNI,30) '% x-axis marks'
      CALL NUM2ST(NUM,PXMIN,IS,IE)
      WRITE (IUNI,1020) NUM(IS:IE), MXBASE, MYBASE-20
      CALL NUM2ST(NUM,PXMAX,IS,IE)
      WRITE (IUNI,1020) NUM(IS:IE), XAXLEN+MXBASE, MYBASE-20
      DO 10,II=1,NX-1
         CALL NUM2ST(NUM,PXMIN+STEPX*II,IS,IE)
         WRITE (IUNI,1020) NUM(IS:IE), MXBASE+MX*II, MYBASE-20
10    CONTINUE
      CALL YJUST(IUNI)
      WRITE (IUNI,30) '% y-axis marks'
      CALL NUM2ST(NUM,PYMIN,IS,IE)
      WRITE (IUNI,1040) NUM(IS:IE), MXBASE-8, MYBASE
      CALL NUM2ST(NUM,PYMAX,IS,IE)
      WRITE (IUNI,1040) NUM(IS:IE), MXBASE-8, YAXLEN+MYBASE
      DO 20,II=1,NY-1
         CALL NUM2ST(NUM,PYMIN+STEPY*II,IS,IE)
         WRITE (IUNI,1040) NUM(IS:IE), MXBASE-8, MY*II+MYBASE
20    CONTINUE
C
30    FORMAT (A)
1000  FORMAT (I3,' ',I3,' moveto')
1010  FORMAT (I3,' ',I3,' ',I3,' { ',I3,' moveto')
1020  FORMAT ('(',A,') ',I3,' ',I3,' checkxj')
1025  FORMAT ('(',A,') ',I3,' ',I3,' checkxj')
1030  FORMAT (I3,' ',I3,' ',I3,' { ',I3,' exch moveto')
1040  FORMAT ('(',A,') ',I3,' ',I3,' checkyj')
1045  FORMAT ('(',A,') ',I3,' ',I3,' checkyj')
      END
C  ---------------------------------------------------------------------
      SUBROUTINE PUTFRA (XAXL, YAXL, MXBASE, MYBASE, IUNI )
C  ---------------------------------------------------------------------
C
C  Description:
C    Part of the Postscript graph routine PSGR1. It writes postscript
C    code that draws a frame around the graph.
C
C  ---------------------------------------------------------------------
C  Arguments:
      INTEGER XAXL, YAXL
C  ---------------------------------------------------------------------
C
      WRITE (IUNI,10) '% draw frame'
      WRITE (IUNI,10) 'newpath'
      WRITE (IUNI,1000) MXBASE, MYBASE+YAXL
      WRITE (IUNI,1010) XAXL,0
      WRITE (IUNI,1000) MXBASE+XAXL, MYBASE+YAXL
      WRITE (IUNI,1010) 0,-YAXL
      WRITE (IUNI,10) 'closepath'
      WRITE (IUNI,10) 'stroke'
C
10    FORMAT (A)
1000  FORMAT (I4,' ',I4,' moveto')
1010  FORMAT (I4,' ',I4,' rlineto')
      END
C  ---------------------------------------------------------------------
      SUBROUTINE PUTGRI (NX, NY, XAXL, YAXL, MXBASE, MYBASE, IUNI )
C  ---------------------------------------------------------------------
C
C  Description:
C    Part of the Postscript graph routine PSGR1. It writes postscript
C    code that draws the horizontal and vertical gridlines.
C
C  ---------------------------------------------------------------------
C  Arguments:
      INTEGER XAXL, YAXL
C  ---------------------------------------------------------------------
C
      DMX = REAL(XAXL)/NX
      DMY = REAL(YAXL)/NY
      MX = INT(DMX)
      MY = INT(DMY)
      WRITE (IUNI,10) 'newpath'
      WRITE (IUNI,10) '% x-axis gridli'
      WRITE (IUNI,1000) MXBASE, MYBASE
      WRITE (IUNI,1010) MXBASE, MX, XAXL+MXBASE, MYBASE
      WRITE (IUNI,1020) 0, YAXL
      WRITE (IUNI,10) '% y-axis gridli'
      WRITE (IUNI,1000) MXBASE, MYBASE
      WRITE (IUNI,1030) MYBASE, MY, YAXL+MYBASE, MXBASE
      WRITE (IUNI,1020) XAXL, 0
      WRITE (IUNI,10) 'closepath'
      WRITE (IUNI,10) 'stroke'
C
10    FORMAT (A)
1000  FORMAT (I3,' ',I3,' moveto')
1010  FORMAT (I3,' ',I3,' ',I3,' { ',I3,' moveto')
1020  FORMAT (I4,' ',I4,' rlineto}for')
1030  FORMAT (I3,' ',I3,' ',I3,' { ',I3,' exch moveto')
      END
C  ---------------------------------------------------------------------
      SUBROUTINE SHOWPG ( IUNI )
C  ---------------------------------------------------------------------
C
C  Description:
C    Part of the Postscript graph routine PSGR1. It writes the
C    "showpage" instruction.
C
C  ---------------------------------------------------------------------
      WRITE (IUNI,10) 'showpage'
C
10    FORMAT (A)
      END
C  ---------------------------------------------------------------------
      SUBROUTINE SETFNT ( KS, IUNI )
C  ---------------------------------------------------------------------
C
C  Description:
C    Part of the Postscript graph routine PSGR1. It writes code that
C    sets the Times-Roman font, in fontsize KS.
C
C  Input arguments:
C    KS         The fontsize.
C    IUNI       Unit number to receive output.
C
C  ---------------------------------------------------------------------
C
      WRITE (IUNI,10) '% seting font'
      WRITE (IUNI,10) '/Times-Roman findfont'
      WRITE (IUNI,1000) KS
      WRITE (IUNI,10) 'setfont'
C
10    FORMAT (A)
1000  FORMAT (I2,' scalefont')
      END
C  ---------------------------------------------------------------------
      SUBROUTINE XAXLAB ( XLABEL, XAXL, XCL1, XCL2, XCL3, MXBASE,
     &                    MYBASE, IUNI )
C  ---------------------------------------------------------------------
C
C  Description:
C    Part of the Postscript graph routine PSGR1. It writes code that
C    plots the X-axis label.
C
C  ---------------------------------------------------------------------
      IMPLICIT $(TYPE) (A-H,O-Z)
C  Arguments:
      CHARACTER*(*) XLABEL
      INTEGER XAXL
C  ---------------------------------------------------------------------
C  Local variables:
      INTEGER YPOS
C  ---------------------------------------------------------------------
C
      LL = LENGTH(XLABEL)
      IF (LL.EQ.0) RETURN
      YPOS = MYBASE - 45
      WRITE (IUNI,10) '% set color'
      WRITE (IUNI,1060) XCL1, XCL2, XCL3
      WRITE (IUNI,10) '% x axis label'
      WRITE (IUNI,10) '0 0 moveto'
      WRITE (IUNI,1050) XLABEL(1:LL)
      WRITE (IUNI,10) 'pop'
      WRITE (IUNI,1030) XAXL
      WRITE (IUNI,1040) MXBASE
      WRITE (IUNI,1000) YPOS
      WRITE (IUNI,1010) XLABEL(1:LL)
      WRITE (IUNI,10) '% set color to normal'
      WRITE (IUNI,10) '0 0 0 setrgbcolor'
C
10    FORMAT (A)
1000  FORMAT (I3,' moveto')
1010  FORMAT ('(',A,') show')
1030  FORMAT (I3,' exch sub 2 div')
1040  FORMAT (I3,' add')
1050  FORMAT ('(',A,') stringwidth')
1060  FORMAT (F3.1,' ',F3.1,' ',F3.1,' setrgbcolor')
      END
C  ---------------------------------------------------------------------
      SUBROUTINE YAXLAB ( YLABEL, YAXL, YLCOL1, YLCOL2, YLCOL3, MXBASE,
     &                    MYBASE, IUNI )
C  ---------------------------------------------------------------------
C
C  Description:
C    Part of the Postscript graph routine PSGR1. It writes code that
C    plots the Y-axis label.
C
C  ---------------------------------------------------------------------
      IMPLICIT $(TYPE) (A-H,O-Z)
C  Arguments:
      CHARACTER*(*) YLABEL
      INTEGER YAXL
C  ---------------------------------------------------------------------
C  Local variables:
      INTEGER XPOS
C  ---------------------------------------------------------------------
C
      LL = LENGTH(YLABEL)
      IF (LL.EQ.0) RETURN
      XPOS = MXBASE-60
      WRITE (IUNI,10) '% set color'
      WRITE (IUNI,1060) YLCOL1, YLCOL2, YLCOL3
      WRITE (IUNI,10) '% y axis label'
      WRITE (IUNI,10) '0 0 moveto'
      WRITE (IUNI,1050) YLABEL(1:LL)
      WRITE (IUNI,10) 'pop'
      WRITE (IUNI,1030) YAXL
      WRITE (IUNI,1040) MYBASE
      WRITE (IUNI,1000) XPOS
      WRITE (IUNI,10) '90 rotate'
      WRITE (IUNI,1010) YLABEL(1:LL)
      WRITE (IUNI,10) '-90 rotate'
      WRITE (IUNI,10) '0 0 moveto'
      WRITE (IUNI,10) '% set color to normal'
      WRITE (IUNI,10) '0 0 0 setrgbcolor'
C
10    FORMAT (A)
1000  FORMAT (I3,' exch moveto')
1010  FORMAT ('(',A,') show')
1030  FORMAT (I3,' exch sub abs 2 div round')
1040  FORMAT (I3,' add')
1050  FORMAT ('(',A,') stringwidth')
1060  FORMAT (F3.1,' ',F3.1,' ',F3.1,' setrgbcolor')
      END
C  ---------------------------------------------------------------------
      SUBROUTINE STAR ( IUNI )
C  ---------------------------------------------------------------------
C
C  Description:
C    Part of the Postscript graph routine PSGR1. It creates postscript
C    code for ploting the star symbol.
C
C  Input arguments:
C    IUNI       Unit number to receive output.
C
C  ---------------------------------------------------------------------
C
      WRITE (IUNI,10) '% the star making routine'
      WRITE (IUNI,10) '/star'
      WRITE (IUNI,10) '{newpath'
      WRITE (IUNI,10) ' moveto'
      WRITE (IUNI,10) ' 2   2  rlineto'
      WRITE (IUNI,10) ' -4  -4 rlineto'
      WRITE (IUNI,10) ' 2   2  rlineto'
      WRITE (IUNI,10) ' -2  2  rlineto'
      WRITE (IUNI,10) ' 4  -4  rlineto'
      WRITE (IUNI,10) 'closepath'
      WRITE (IUNI,10) 'stroke}def'
C
10    FORMAT (A)
      END
C  ---------------------------------------------------------------------
      SUBROUTINE SQUARE ( IUNI )
C  ---------------------------------------------------------------------
C
C  Description:
C    Part of the Postscript graph routine PSGR1. It creates postscript
C    code for ploting a square.
C
C  Input arguments:
C    IUNI       Unit number to receive output.
C
C  ---------------------------------------------------------------------
C
      WRITE (IUNI,10) '% the square making routine'
      WRITE (IUNI,10) '/square'
      WRITE (IUNI,10) '{newpath'
      WRITE (IUNI,10) ' moveto'
      WRITE (IUNI,10) ' 2   2  rmoveto'
      WRITE (IUNI,10) ' 0  -4 rlineto'
      WRITE (IUNI,10) ' -4   0  rlineto'
      WRITE (IUNI,10) ' 0  4  rlineto'
      WRITE (IUNI,10) ' 4  0  rlineto'
      WRITE (IUNI,10) 'closepath'
      WRITE (IUNI,10) 'stroke}def'
C
10    FORMAT (A)
      END
C  ---------------------------------------------------------------------
      SUBROUTINE TRIANG ( IUNI )
C  ---------------------------------------------------------------------
C
C  Description:
C    Part of the Postscript graph routine PSGR1. It creates postscript
C    code for ploting a triangle.
C
C  Input arguments:
C    IUNI       Unit number to receive output.
C
C  ---------------------------------------------------------------------
C
      WRITE (IUNI,10) '% the triangle making routine'
      WRITE (IUNI,10) '/triangle'
      WRITE (IUNI,10) '{newpath'
      WRITE (IUNI,10) ' moveto'
      WRITE (IUNI,10) ' 0   2  rmoveto'
      WRITE (IUNI,10) ' 2  -4 rlineto'
      WRITE (IUNI,10) ' -4   0  rlineto'
      WRITE (IUNI,10) ' 2  4  rlineto'
      WRITE (IUNI,10) 'closepath'
      WRITE (IUNI,10) 'stroke}def'
C
10    FORMAT (A)
      END
C  ---------------------------------------------------------------------
      SUBROUTINE PLUS ( IUNI )
C  ---------------------------------------------------------------------
C
C  Description:
C    Part of the Postscript graph routine PSGR1. It creates postscript
C    code for ploting the plus symbol.
C
C  Input arguments:
C    IUNI       Unit number to receive output.
C
C  ---------------------------------------------------------------------
C
      WRITE (IUNI,10) '% the plus making routine'
      WRITE (IUNI,10) '/plus'
      WRITE (IUNI,10) '{newpath'
      WRITE (IUNI,10) ' moveto'
      WRITE (IUNI,10) ' 0   2  rmoveto'
      WRITE (IUNI,10) ' 0  -4 rlineto'
      WRITE (IUNI,10) ' -2  2  rmoveto'
      WRITE (IUNI,10) ' 4  0  rlineto'
      WRITE (IUNI,10) 'closepath'
      WRITE (IUNI,10) 'stroke}def'
C
10    FORMAT (A)
      END
C  ---------------------------------------------------------------------
      SUBROUTINE FSQUAR ( IUNI )
C  ---------------------------------------------------------------------
C
C  Description:
C    Part of the Postscript graph routine PSGR1. It creates postscript
C    code for ploting the filled square.
C
C  Input arguments:
C    IUNI       Unit number to receive output.
C
C  ---------------------------------------------------------------------
C
      WRITE (IUNI,10) '% the filled square making routine'
      WRITE (IUNI,10) '/fsquare'
      WRITE (IUNI,10) '{newpath'
      WRITE (IUNI,10) ' moveto'
      WRITE (IUNI,10) ' 2   2  rmoveto'
      WRITE (IUNI,10) ' 0  -4 rlineto'
      WRITE (IUNI,10) ' -4   0  rlineto'
      WRITE (IUNI,10) ' 0  4  rlineto'
      WRITE (IUNI,10) ' 4  0  rlineto'
      WRITE (IUNI,10) 'closepath'
      WRITE (IUNI,10) 'fill}def'
C
10    FORMAT (A)
      END
C  ---------------------------------------------------------------------
      SUBROUTINE FTRIAN ( IUNI )
C  ---------------------------------------------------------------------
C
C  Description:
C    Part of the Postscript graph routine PSGR1. It creates postscript
C    code for ploting the filled triangle.
C
C  Input arguments:
C    IUNI       Unit number to receive output.
C
C  ---------------------------------------------------------------------
C
      WRITE (IUNI,10) '% the filled triangle making routine'
      WRITE (IUNI,10) '/ftriangle'
      WRITE (IUNI,10) '{newpath'
      WRITE (IUNI,10) ' moveto'
      WRITE (IUNI,10) ' 0   2  rmoveto'
      WRITE (IUNI,10) ' 2  -4 rlineto'
      WRITE (IUNI,10) ' -4   0  rlineto'
      WRITE (IUNI,10) ' 2  4  rlineto'
      WRITE (IUNI,10) 'closepath'
      WRITE (IUNI,10) 'fill}def'
C
10    FORMAT (A)
      END
C  ---------------------------------------------------------------------
      SUBROUTINE SHOWPT ( PSDATA, XCOUNT, MCURCO, MSYMB, SCOL1, SCOL2,
     &                    SCOL3, MXBASE, MYBASE, IUNI, MAXROW, MAXCOL )
C  ---------------------------------------------------------------------
C
C  Description:
C    Part of the Postscript graph routine PSGR1. It writes code that
C    plots the data points according to the given symbol MSYMB.
C
C  ---------------------------------------------------------------------
      IMPLICIT $(TYPE) (A-H,O-Z)
C  Arguments:
      INTEGER XCOUNT
      DIMENSION PSDATA(MAXROW,MAXCOL)
C  ---------------------------------------------------------------------
C
      IF (MSYMB.EQ.1) CALL STAR(IUNI)
      IF (MSYMB.EQ.2) CALL SQUARE(IUNI)
      IF (MSYMB.EQ.3) CALL TRIANG(IUNI)
      IF (MSYMB.EQ.4) CALL PLUS(IUNI)
      IF (MSYMB.EQ.5) CALL FSQUAR(IUNI)
      IF (MSYMB.EQ.6) CALL FTRIAN(IUNI)
      WRITE (IUNI,10) '% set color'
      WRITE (IUNI,1000) SCOL1, SCOL2, SCOL3
      WRITE (IUNI,10) '% puting the points'
      DO 100,III=1,XCOUNT
         XCOR = PSDATA(III,1)
         YCOR = PSDATA(III,MCURCO)
         IF (MSYMB.EQ.1) WRITE (IUNI,1010) XCOR+MXBASE, YCOR+MYBASE
         IF (MSYMB.EQ.2) WRITE (IUNI,1020) XCOR+MXBASE, YCOR+MYBASE
         IF (MSYMB.EQ.3) WRITE (IUNI,1030) XCOR+MXBASE, YCOR+MYBASE
         IF (MSYMB.EQ.4) WRITE (IUNI,1040) XCOR+MXBASE, YCOR+MYBASE
         IF (MSYMB.EQ.5) WRITE (IUNI,1050) XCOR+MXBASE, YCOR+MYBASE
         IF (MSYMB.EQ.6) WRITE (IUNI,1060) XCOR+MXBASE, YCOR+MYBASE
100   CONTINUE
      WRITE (IUNI,10) '% set color to normal'
      WRITE (IUNI,10) '0 0 0 setrgbcolor'
C
10    FORMAT (A)
1000  FORMAT (F3.1,' ',F3.1,' ',F3.1,' setrgbcolor')
1010  FORMAT (F8.3,' ',F8.3,' star')
1020  FORMAT (F8.3,' ',F8.3,' square')
1030  FORMAT (F8.3,' ',F8.3,' triangle')
1040  FORMAT (F8.3,' ',F8.3,' plus')
1050  FORMAT (F8.3,' ',F8.3,' fsquare')
1060  FORMAT (F8.3,' ',F8.3,' ftriangle')
      END
C  ---------------------------------------------------------------------
      SUBROUTINE CONPTS ( PSDATA, XCOUNT, MCURCO, DASH, COLOR1, COLOR2,
     &                    COLOR3, MXBASE, MYBASE, IUNI, MAXROW, MAXCOL )
C  ---------------------------------------------------------------------
C
C  Description:
C    Part of the Postscript graph routine PSGR1. It writes code to
C    connect subsequent points of the same curve.
C
C  ---------------------------------------------------------------------
      IMPLICIT $(TYPE) (A-H,O-Z)
C  Arguments:
      INTEGER XCOUNT, DASH
      DIMENSION PSDATA(MAXROW,MAXCOL)
C  ---------------------------------------------------------------------
C
      WRITE (IUNI,10) '% set color'
      WRITE (IUNI,1000) COLOR1, COLOR2, COLOR3
      WRITE (IUNI,10) '% dash type'
      IF (DASH.EQ.1) WRITE (IUNI,10) '[] 0 setdash'
      IF (DASH.EQ.2) WRITE (IUNI,10) '[6 4 2 4] 0 setdash'
      IF (DASH.EQ.3) WRITE (IUNI,10) '[4 4] 0 setdash'
C
      WRITE (IUNI,10) '% line connecting points'
      WRITE (IUNI,10) 'newpath'
      XCOR = PSDATA(1,1)
      YCOR = PSDATA(1,MCURCO)
      WRITE (IUNI,1030) XCOR+MXBASE, YCOR+MYBASE
      DO 110,III=2,XCOUNT-1
         XCOR = PSDATA(III,1)
         YCOR = PSDATA(III,MCURCO)
         WRITE (IUNI,1020) XCOR+MXBASE, YCOR+MYBASE
110   CONTINUE
      WRITE (IUNI,1030) XCOR+MXBASE, YCOR+MYBASE
      XCOR = PSDATA(XCOUNT,1)
      YCOR = PSDATA(XCOUNT,MCURCO)
      WRITE (IUNI,1020) XCOR+MXBASE, YCOR+MYBASE
      WRITE (IUNI,10) 'closepath'
      WRITE (IUNI,10) 'stroke'
C
10    FORMAT (A)
1000  FORMAT (F3.1,' ',F3.1,' ',F3.1,' setrgbcolor')
1020  FORMAT (F7.3,' ',F7.3,' lineto')
1030  FORMAT (F7.3,' ',F7.3,' moveto')
      END
C  ---------------------------------------------------------------------
      SUBROUTINE NUM2ST ( NUM, XX, IS, IE )
C  ---------------------------------------------------------------------
C
C  Description:
C    Simple minded routine to convert a real into a string. This
C    is subsequently used by the Postscript graph routine PSGR1.
C
C  Input arguments:
C    XX         The number to be converted.
C
C  Output arguments:
C    NUM        The output string.
C    IS         Character position of NUM, where the converted number
C               starts. The whole number is in NUM(IS:IE).
C    IE         Character position of NUM, where the converted number
C               ends. The whole number is in NUM(IS:IE).
C
C  ---------------------------------------------------------------------
      IMPLICIT $(TYPE) (A-H,O-Z)
C  Arguments:
      CHARACTER*(*) NUM
C  ---------------------------------------------------------------------
C
      IF (XX.EQ.0.) THEN
         NUM = '0.0'
         IS = 1
         IE = 3
         RETURN
      END IF
C
      WRITE (NUM,'(1PG9.2)') XX
      IE = LENGTH(NUM)
      DO 10,I=1,IE
         IF (NUM(I:I).NE.' ') THEN
            IS = I
            GOTO 20
         END IF
10    CONTINUE
      CALL MABORT('Blank string.','NUM2ST')
20    CONTINUE
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE DFESTI ( NU, NROW, NCOL )
C  ---------------------------------------------------------------------
C
C  Description:
C    Called by subroutine PSGRAF. It determines the number of rows
C    and columns, present in the data file.
C
C  Input arguments:
C    NU         Unit number of the open datafile.
C
C  Output arguments:
C    NROW       Number of rows.
C    NCOL       Number of columns.
C
C  ---------------------------------------------------------------------
C  Local variables:
      CHARACTER TMP
      CHARACTER*80 LINE, NEW
      INTEGER IS(40), IE(40)
C  ---------------------------------------------------------------------
C
      NROW = 0
      NCOL = 0
      READ (NU,'(A)',END=10) LINE
      LE = LENGTH(LINE)
      CALL QLCR(LINE,LE,NEW,IS,IE,NIT,IERR)
      NROW = 1
      NCOL = NIT
C
20    READ (NU,'(A)',END=10) TMP
      NROW = NROW+1
      GOTO 20
C
10    CONTINUE
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE DFRIN ( NU, DAT, LD, NCOL, MERR )
C  ---------------------------------------------------------------------
C
C  Description:
C    Called by subroutine PSGRAF. It reads the rows and columns of
C    the data file, and disposes the numbers in the two dimensional
C    array DAT.
C
C  Input arguments:
C    NU         Unit number of the open data file.
C    LD         Leading dimension of array DAT.
C    NCOL       Number of columns in the data file.
C
C  Output arguments:
C    DAT        Two dimensional array. Its first column stores the
C               X-values, while the Y-values start at column 2.
C    MERR       Error indicator.
C
C  ---------------------------------------------------------------------
      IMPLICIT $(TYPE) (A-H,O-Z)
C  Arguments:
      DIMENSION DAT(LD,*)
C  ---------------------------------------------------------------------
C  Common blocks:
      LOGICAL STRO, WEAK
      COMMON /PRILOG/ STRO, WEAK
      COMMON /UNITS/ IUINP, IUOUT
C  ---------------------------------------------------------------------
C  Local variables:
      CHARACTER*10 STR, STR2
      CHARACTER*80 LINE, NEW
      INTEGER IS(40), IE(40)
C  ---------------------------------------------------------------------
C
      MERR = 0
      REWIND NU
      NL = 0
C
C  Read a line.
10    READ (NU,'(A)',END=20) LINE
      NL = NL+1
      LE = LENGTH(LINE)
C
C  Crack into items.
      CALL QLCR(LINE,LE,NEW,IS,IE,NIT,IERR)
      IF (IERR.NE.0) THEN
         CALL I2STR(NL,STR,LES)
         IF (WEAK) THEN
            WRITE (IUOUT,40) STR(1:LES)
            MERR = -2
            RETURN
         END IF
      END IF
      IF (NIT.NE.NCOL) THEN
         CALL I2STR(NL,STR,LES)
         CALL I2STR(NIT,STR2,LES2)
         IF (WEAK) THEN
            WRITE (IUOUT,50) STR(1:LES), STR2(1:LES2)
            MERR = -2
            RETURN
         END IF
      END IF
C
C  Convert to real numbers.
      DO 30,I=1,NIT
         LE = IE(I)-IS(I)+1
         CALL TOREAL(NEW(IS(I):IE(I)),LE,XX,IERR)
         IF (IERR.NE.0) THEN
            CALL I2STR(NL,STR,LES)
            IF (WEAK) THEN
               WRITE (IUOUT,40) STR(1:LES)
               MERR = -2
               RETURN
            END IF
         END IF
         DAT(NL,I) = XX
30    CONTINUE
      GOTO 10
C
20    CONTINUE
C
40    FORMAT (/$(FC)'Error: Row ',A,' contains non numeric ',
     &        'characters.')
50    FORMAT (/$(FC)'Error: Row ',A,' has ',A,' columns.')
      END
C  ---------------------------------------------------------------------
      SUBROUTINE PSGSO ( DAT, LD, NROW, NCOL )
C  ---------------------------------------------------------------------
C
C  Description:
C    Called by subroutine PSGRAF. It sorts the first column of the
C    data array DAT, in ascending order, rearranging the rest of the
C    columns, as necessary. A bubble sort is used.
C
C  Input arguments:
C    LD         Leading dimension of array DAT.
C    NROW       Number of data rows in array DAT.
C    NCOL       Number of data columns in array DAT.
C
C  Output arguments:
C    DAT        On input the data array, as read from the data file.
C               On output DAT is sorted.
C
C  ---------------------------------------------------------------------
      IMPLICIT $(TYPE) (A-H,O-Z)
C  Arguments:
      DIMENSION DAT(LD,*)
C  ---------------------------------------------------------------------
C  Local variables:
      LOGICAL SORTED
C  ---------------------------------------------------------------------
C
      DO 10,K=1,NROW-1
         SORTED = .TRUE.
         DO 20,I=NROW,K+1,-1
            IF (DAT(I,1).LT.DAT(I-1,1)) THEN
               SORTED = .FALSE.
               IM1 = I-1
               DO 30,J=1,NCOL
                  TMP = DAT(I,J)
                  DAT(I,J) = DAT(IM1,J)
                  DAT(IM1,J) = TMP
30             CONTINUE
            END IF
20       CONTINUE
         IF (SORTED) RETURN
10    CONTINUE
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE PSTAT ( MERR )
C  ---------------------------------------------------------------------
C
C  Description:
C    Implements the PSTATUS command.
C    Syntax: PSTATUS [ panel ... ]
C
C  Output arguments:
C    MERR       Error indicator.
C
C  ---------------------------------------------------------------------
      IMPLICIT $(TYPE) (A-H,O-Z)
C  Arguments:
C  ---------------------------------------------------------------------
C  Common blocks:
      PARAMETER ( NMCOM = 98, MXPLUG = $(MXPLUG), NCOM = NMCOM+MXPLUG )
      CHARACTER*10 COMMAN(NCOM)
      COMMON /COM1/ COMMAN
      LOGICAL STRO, WEAK
      COMMON /PRILOG/ STRO, WEAK
      COMMON /UNITS/ IUINP, IUOUT
C  Maximum length of a keyword.
      PARAMETER ( KEYLEN = 10 )
C  Maximum number of panel commands.
      PARAMETER ( MXPCOM = 80 )
C  Maximum number of keywords in all panel commands.
      PARAMETER ( MXPKEY = 1000 )
C  Maximum storage for integer/real keywords.
      PARAMETER ( MXPIR = 5000 )
C  Maximum storage for string keywords and descriptions.
      PARAMETER ( MXPSTR = 20000 )
C  Maximum number of pointers in the above string storage.
      PARAMETER ( MXPPST = 5000 )
C
      CHARACTER*(KEYLEN) PCOM(MXPCOM)
      INTEGER            PONF(MXPCOM)
      INTEGER            KPOINT(MXPCOM)
      INTEGER            NCHELP(2*MXPCOM)
      CHARACTER*(KEYLEN) PKEY(MXPKEY)
      INTEGER            NKHELP(2*MXPKEY)
      CHARACTER          KTYPE(MXPKEY)
      INTEGER            TPOINT(MXPKEY)
      CHARACTER          KIO(MXPKEY)
      $(TYPE)            PIR(MXPIR)
      INTEGER            SPOINT(MXPPST)
      CHARACTER*(MXPSTR) PANSTR
      COMMON /PCOMS1/ KPOINT, NCHELP, NKHELP, TPOINT, SPOINT, PONF
      COMMON /PCOMS2/ PIR
      COMMON /PCOMS3/ PKEY, KTYPE, PANSTR, PCOM, KIO
C
C  NPCOM: Number of panel commands in the panel description file.
C  NPKEY: Number of keywords in all panel commands.
C  NPIR:  Number of positions filled in array PIR.
C  NPSP:  Number of positions filled in the char-pointer array SPOINT.
C  NPCH:  Number of character positions filled in PANSTR.
      COMMON /PCOMS4/ NPCOM, NPKEY, NPIR, NPSP, NPCH
C  ---------------------------------------------------------------------
C  Local variables:
      CHARACTER*80 PAN
      PARAMETER ( NPL = 4 )
      CHARACTER*10 PNAM(NPL)
      CHARACTER*3 PSTA(NPL)
      CHARACTER*3 ONOFF(0:1)
      CHARACTER*4 ONOFFD(0:1)
      DATA ONOFF / 'off', 'on ' /
      DATA ONOFFD / 'off.', 'on. ' /
C  ---------------------------------------------------------------------
C
      MERR = 0
      CALL ARGNO(NA)
      IF (NA.EQ.0) THEN
C
C  No arguments entered. Report status of all panels.
C  Use NPL panel-names per line.
         CALL DASH
         CALL CPRINT('Panel status')
         CALL DASH
         N = 0
         DO 10,I=1,NCOM
C  Check if the command has a panel.
            CALL PLOCP(COMMAN(I),IP)
            IF (IP.NE.0) THEN
C  Accumulate panel name and panel status in the current line.
               N = N+1
               PNAM(N) = COMMAN(I)
               PSTA(N) = ONOFF(PONF(IP))
C  Flush line if NPL panels are accumulated.
               IF (N.EQ.NPL) THEN
                  IF (STRO) WRITE (IUOUT,50)
     &                                  (PNAM(K), PSTA(K), K=1,N)
                  N = 0
               END IF
            END IF
10       CONTINUE
C  Flush line if not empty.
         IF (N.NE.0 .AND. STRO) WRITE (IUOUT,50)
     &                            (PNAM(K), PSTA(K), K=1,N)
         CALL DASH
      ELSE
C
C  Some arguments entered in the command line.
         DO 80,I=1,NA
            CALL CMDARG(I,PAN,LEP)
C  Check if this is a valid Merlin command.
            CALL LOCC(PAN,IC)
            IF (IC.EQ.0) THEN
               IF (WEAK) WRITE (IUOUT,20) PAN(1:LEP)
               MERR = -2
               RETURN
            END IF
C  Check if the command has a panel.
            CALL PLOCP(COMMAN(IC),IP)
C  Report panel status.
            IF (IP.EQ.0) THEN
               IF (WEAK) WRITE (IUOUT,30) PAN(1:LEP)
               MERR = -2
            ELSE
               IF (STRO) WRITE (IUOUT,40)
     &                            PAN(1:LEP), ONOFFD(PONF(IP))
            END IF
80       CONTINUE
      END IF
C
20    FORMAT (/$(FC)'There is no such Merlin command: "',A,'".')
30    FORMAT (/$(FC)'The "',A,'" command has no panel.')
40    FORMAT (/$(FC)'The "',A,'" panel is ',A)
50    FORMAT ($(FC)3X,4(A,2X,A:4X))
      END
C  ---------------------------------------------------------------------
      SUBROUTINE QUIT ( MERR )
C  ---------------------------------------------------------------------
C
C  Description:
C    Implements the QUIT command.
C    Syntax: QUIT quit_flag [ noepilog ]
C
C  Output arguments:
C    MERR       Error indicator.
C
C  ---------------------------------------------------------------------
      IMPLICIT $(TYPE) (A-H,O-Z)
C  Common blocks:
      PARAMETER ( LSTA = $(MCLMEM) )
      COMMON /MCLV1/ STACK(LSTA)
      PARAMETER ( MAXMFI = 20 )
      COMMON /MCLV2/ ISTACK, INFILE, MBASE, INAD, MPC, MCLLNO, LENBUF,
     &               IBUF, MCLUNI(MAXMFI), NUMFI
      COMMON /GETOUT/ IRET, IRET2, IQU
      PARAMETER ( MXFILE = 80 )
      CHARACTER*(MXFILE) PROLO, EPILO, IN, OUT, HFILE, MFILE, PDFILE
      COMMON /CP2/ PROLO, EPILO, IN, OUT, HFILE, MFILE, PDFILE
      COMMON /MODE3/ MODEV, IFFORM, INMCL
      LOGICAL STRO, WEAK
      COMMON /PRILOG/ STRO, WEAK
      COMMON /UNITS/ IUINP, IUOUT
C  ---------------------------------------------------------------------
C  Local variables:
      CHARACTER*80 ARG
      LOGICAL SUPRES
C  ---------------------------------------------------------------------
C
      CALL ARGNO(NA)
      IF (INMCL.EQ.0) THEN
C  Not in MCL.
         IF (NA.EQ.0) THEN
            IF (STRO) WRITE (IUOUT,10)
            CALL ININ(IQU,MERR)
            IF (MERR.EQ.-2) THEN
               IF (WEAK) WRITE (IUOUT,35)
               RETURN
            END IF
            IF (MERR.NE.0) RETURN
            SUPRES = .FALSE.
         ELSE IF (NA.EQ.1 .OR. NA.EQ.2) THEN
            CALL CMDARG(1,ARG,LENARG)
            CALL TOINT(ARG,LENARG,IQU,IERR)
            IF (IERR.NE.0) THEN
               IF (WEAK) WRITE (IUOUT,35)
               MERR = -2
               RETURN
            END IF
            IF (NA.EQ.2) THEN
               CALL CMDARG(2,ARG,LENARG)
               IF (ISCOMP(ARG(1:LENARG),'NOEPILOG').NE.0) THEN
                  SUPRES = .TRUE.
               ELSE
                  IF (WEAK) WRITE (IUOUT,25)
                  MERR = -2
                  RETURN
               END IF
            ELSE
               SUPRES = .FALSE.
            END IF
         ELSE
            CALL ERRNAR
            MERR = -2
            RETURN
         END IF
      ELSE
C  We are running MCL.
         IF (NA.EQ.0) THEN
            SUPRES = .FALSE.
         ELSE IF (NA.EQ.1) THEN
            CALL CMDARG(1,ARG,LENARG)
            IF (ISCOMP(ARG(1:LENARG),'NOEPILOG').NE.0) THEN
               SUPRES = .TRUE.
            ELSE
               IF (WEAK) WRITE (IUOUT,25)
               MERR = -2
               RETURN
            END IF
         ELSE
            CALL ERRNAR
            MERR = -2
            RETURN
         END IF
         IQU = NINT(STACK(ISTACK))
         ISTACK = ISTACK-1
      END IF
C
C  The quit-flag must be strictly positive. Negative values are
C  reserved for Merlin error returns, while 0 indicates that RETURN
C  has been used to exit Merlin.
      IF (IQU.LE.0) THEN
         IF (WEAK) WRITE (IUOUT,20)
         MERR = -2
         RETURN
      ELSE
         IF (STRO) THEN
            CALL I2STR(IQU,ARG,LEA)
            WRITE (IUOUT,30) ARG(1:LEA)
         END IF
      END IF
C
      IRET = 1
      IRET2 = 1
C
C  Supress epilog if "noepilog" is entered in the command line.
      LEE = LENGTH(EPILO)
      IF (SUPRES .AND. LEE.NE.0) THEN
         IF (STRO) WRITE (IUOUT,40) EPILO(1:LEE)
         EPILO = ' '
      END IF
      MERR = 0
C
10    FORMAT (/$(FC)'Enter return flag:')
20    FORMAT (/$(FC)'Quit flag must be strictly positive.')
25    FORMAT (/$(FC)'Incorrect argument. Only "noepilog" can follow.')
30    FORMAT (/$(FC)'Quit flag is set equal to ',A,'.')
35    FORMAT (/$(FC)'Incorrect quit flag. You must supply a ',
     &        'positive integer.')
40    FORMAT (/$(FC)'Your epilog command "',A,'" will not be ',
     &        'executed.')
      END
C  ---------------------------------------------------------------------
      SUBROUTINE RESET
C  ---------------------------------------------------------------------
C
C  Description:
C    Implements the RESET command.
C    Sets to 0 the Merlin partial counters for:
C      1) Calls to the objecive function. ( FUNCTION FUNMIN or
C         SUBROUTINE SUBSUM )
C      2) Calls to the user supplied routine that evaluates the
C         gradient vector. ( SUBROUTINE GRANAL )
C      3) Calls to the user supplied routine that evaluates the
C         second derivative matrix. ( SUBROUTINE HANAL )
C
C  ---------------------------------------------------------------------
C  Common blocks:
      LOGICAL STRO, WEAK
      COMMON /PRILOG/ STRO, WEAK
      COMMON /UNITS/ IUINP, IUOUT
      COMMON /TOTCAL/ ITNOCS, IPARTC, ITDER, IPDER, ITHESS, IPHESS,
     &                ITJAC, IPJAC
C  ---------------------------------------------------------------------
C
      IPARTC = 0
      IPDER = 0
      IPHESS = 0
      IPJAC = 0
      IF (STRO) WRITE (IUOUT,10)
C
10    FORMAT (/$(FC)'Partial counters have been reset.')
      END
C  ---------------------------------------------------------------------
      SUBROUTINE RETSTP ( ISEL, MERR )
C  ---------------------------------------------------------------------
C
C  Description:
C    Implements the RETURN and STOP commands.
C    Syntax: STOP [ noepilog]
C            RETURN [ noepilog ]
C
C  Input arguments:
C    ISEL       Selects the RETURN (ISEL=1) or the STOP (ISEL=2)
C               command.
C
C  Output arguments:
C    MERR       Error indicator.
C
C  ---------------------------------------------------------------------
C  Common blocks:
      PARAMETER ( MXFILE = 80 )
      CHARACTER*(MXFILE) PROLO, EPILO, IN, OUT, HFILE, MFILE, PDFILE
      COMMON /CP2/ PROLO, EPILO, IN, OUT, HFILE, MFILE, PDFILE
      COMMON /GETOUT/ IRET, IRET2, IQU
      LOGICAL STRO, WEAK
      COMMON /PRILOG/ STRO, WEAK
      COMMON /UNITS/ IUINP, IUOUT
C  ---------------------------------------------------------------------
C  Local variables:
      CHARACTER*80 ARG
      LOGICAL SUPRES
C  ---------------------------------------------------------------------
C
      MERR = 0
      CALL ARGNO(NA)
      IF (NA.EQ.0) THEN
         SUPRES = .FALSE.
      ELSE IF (NA.EQ.1) THEN
         CALL CMDARG(1,ARG,LENARG)
         IF (ISCOMP(ARG(1:LENARG),'NOEPILOG').NE.0) THEN
            SUPRES = .TRUE.
         ELSE
            IF (WEAK) WRITE (IUOUT,20)
            MERR = -2
            RETURN
         END IF
      ELSE
         CALL ERRNAR
         MERR = -2
         RETURN
      END IF
C
      IRET = ISEL
      IRET2 = ISEL
      IQU = 0
C
C  Supress epilog if "noepilog" is entered in the command line.
      LEE = LENGTH(EPILO)
      IF (SUPRES .AND. LEE.NE.0) THEN
         IF (STRO) WRITE (IUOUT,30) EPILO(1:LEE)
         EPILO = ' '
      END IF
C
20    FORMAT (/$(FC)'Incorrect argument. Only "noepilog" can follow.')
30    FORMAT (/$(FC)'Your epilog command "',A,'" will not be ',
     &        'executed.')
      END
C  ---------------------------------------------------------------------
      SUBROUTINE REVEAL ( MERR )
C  ---------------------------------------------------------------------
C
C  Description:
C    Implements the REVEAL command.
C
C  Output arguments:
C    MERR       Error indicator.
C
C  ---------------------------------------------------------------------
C  Common blocks:
      PARAMETER ( MAXONL = 20 )
      INTEGER BACOUT(MAXONL), FROUT(MAXONL)
      COMMON /OUTER/ LEVOUT, BACOUT, FROUT
      LOGICAL STRO, WEAK
      COMMON /PRILOG/ STRO, WEAK
      COMMON /UNITS/ IUINP, IUOUT
C  ---------------------------------------------------------------------
C  Local variables:
      PARAMETER ( MXFILE = 80 )
      CHARACTER*(MXFILE) FILE
C  ---------------------------------------------------------------------
C
C  Search for a HIDEOUT command.
      DO 10,I=LEVOUT,1,-1
         IF (FROUT(I).EQ.4) GOTO 20
10    CONTINUE
      IF (WEAK) WRITE (IUOUT,30)
      MERR = -2
      RETURN
C
20    CONTINUE
C
C  Get the name of the "HIDEOUT" file.
      CALL U2NAME(BACOUT(I),'hideout',FILE,LEF)
C
C  Decrement the output level.
      CALL DECOLE(4)
      IF (LEF.NE.0 .AND. STRO) WRITE (IUOUT,40) FILE(1:LEF)
      MERR = 0
C
30    FORMAT (/$(FC)'A HIDEOUT command is not in effect.')
40    FORMAT (/$(FC)'Output resumes after hideout to file "',A,'".')
      END
C  ---------------------------------------------------------------------
      SUBROUTINE REWEOF ( IOPT, MERR )
C  ---------------------------------------------------------------------
C
C  Description:
C    Implements the REWIND and GOEOF commands.
C    Syntax: REWIND file ...
C            GOEOF file ...
C    Open Merlin files (backup, memo and history) cannot be processed.
C
C  Input arguments:
C    IOPT       Selects the operation to be performed.
C               IOPT = 1 -> REWIND
C               IOPT = 2 -> GOEOF
C
C  Output arguments:
C    MERR       Error indicator.
C
C  ---------------------------------------------------------------------
C  Common blocks:
      LOGICAL STRO, WEAK
      COMMON /PRILOG/ STRO, WEAK
      COMMON /UNITS/ IUINP, IUOUT
C  ---------------------------------------------------------------------
C  Local variables:
      PARAMETER ( MXFILE = 80 )
      CHARACTER*(MXFILE) FILE
C  ---------------------------------------------------------------------
C
      MERR = 0
      CALL ARGNO(NA)
      IF (NA.EQ.0) THEN
C  No command line arguments. Prompt the user.
         IF (STRO) THEN
            IF (IOPT.EQ.1) THEN
               WRITE (IUOUT,10)
            ELSE
               WRITE (IUOUT,20)
            END IF
         END IF
         CALL INNOCR(FILE,.FALSE.,MERR)
         IF (MERR.NE.0) RETURN
         CALL REOF1F(FILE,IOPT,MERR)
      ELSE
C  Get file names entered in the command line.
         DO 100,I=1,NA
            CALL CMDARG(I,FILE,LENF)
            CALL REOF1F(FILE,IOPT,MERR)
            IF (MERR.NE.0) RETURN
100      CONTINUE
      END IF
C
10    FORMAT (/$(FC)'Enter file name to rewind:')
20    FORMAT (/$(FC)'Enter file name to position at End-Of-File:')
      END
C  ---------------------------------------------------------------------
      SUBROUTINE REOF1F ( FILE, IOPT, MERR )
C  ---------------------------------------------------------------------
C
C  Description:
C    Part of the REWIND/GOEOF commands. According to IOPT rewinds or
C    positions at EOF the file specified by FILE.
C
C  Input arguments:
C    FILE       The file name to position.
C    IOPT       Selects the operation to be performed.
C               IOPT = 1 -> The file is rewound.
C               IOPT = 2 -> The file is positioned at End-of-File.
C
C  Output arguments:
C    MERR       Error indicator.
C
C  ---------------------------------------------------------------------
C  Arguments:
      CHARACTER*(*) FILE
C  ---------------------------------------------------------------------
C  Common blocks:
      PARAMETER ( MAXMFI = 20 )
      COMMON /MCLV2/ ISTACK, INFILE, MBASE, INAD, MPC, MCLLNO, LENBUF,
     &               IBUF, MCLUNI(MAXMFI), NUMFI
      COMMON /MODE3/ MODEV, IFFORM, INMCL
      LOGICAL STRO, WEAK
      COMMON /PRILOG/ STRO, WEAK
      COMMON /UNITS/ IUINP, IUOUT
C  ---------------------------------------------------------------------
C  Local variables:
      CHARACTER*10 SEQ, STR, DESC
      LOGICAL EXI, OPE
C  ---------------------------------------------------------------------
C
C  Get info for the file.
      LENF = LENGTH(FILE)
      INQUIRE (FILE=FILE,EXIST=EXI,OPENED=OPE,NUMBER=NUM,
     &         SEQUENTIAL=SEQ,IOSTAT=IOE)
      IF (IOE.NE.0) THEN
         CALL ERRINQ(FILE,IOE)
         MERR = -2
         RETURN
      END IF
C
C  The file does not exist.
      IF (.NOT.EXI) THEN
         CALL WRNEXI(FILE)
         RETURN
      END IF
C
      IF (OPE) THEN
C
C  The file is open.
C  Is is one of the open Merlin files ?
         CALL ISAMEF(NUM,DESC)
         IF (DESC.NE.' ') THEN
            IF (WEAK) THEN
               WRITE (IUOUT,60) FILE(1:LENF), DESC(1:LENGTH(DESC))
               IF (IOPT.EQ.1) THEN
                  WRITE (IUOUT,61)
               ELSE
                  WRITE (IUOUT,62)
               END IF
            END IF
            MERR = -2
            RETURN
         END IF
C
C  Access mode should be SEQUENTIAL. DIRECT access files cannot be
C  positioned.
         IF (ISCOMP(SEQ,'NO').NE.0) THEN
            CALL I2STR(NUM,STR,LES)
            IF (WEAK) WRITE (IUOUT,20) FILE(1:LENF), STR(1:LES)
            MERR = -2
            RETURN
         ELSE IF (ISCOMP(SEQ,'UNKNOWN').NE.0) THEN
            CALL I2STR(NUM,STR,LES)
            IF (WEAK) WRITE (IUOUT,25) FILE(1:LENF), STR(1:LES)
            MERR = -2
            RETURN
         END IF
         NU = NUM
      ELSE
C
C  The file is not open. Open it.
         NU = NUNIT()
         IF (NU.EQ.-1) THEN
            CALL ERRUAL(FILE)
            MERR = -2
            RETURN
         END IF
         OPEN (NU,FILE=FILE,IOSTAT=IOE)
         IF (IOE.NE.0) THEN
            CALL ERROPE(FILE,IOE)
            MERR = -2
            RETURN
         END IF
      END IF
C
      IF (IOPT.EQ.1) THEN
C  Rewind the file.
         REWIND (NU,IOSTAT=IOE)
         IF (IOE.NE.0) THEN
            CALL ERRREW(FILE,IOE)
            MERR = -2
         END IF
      ELSE
C  Position the file at End-of-File.
         CALL POSIT(NU)
      END IF
C
C  If the file initialy was not open, close it.
      IF (.NOT.OPE) THEN
         IF (INMCL.EQ.0) THEN
C  Not in MCL.
            CLOSE (NU)
         ELSE
C  In MCL leave the file open.
            NUMFI = NUMFI+1
            IF (NUMFI.GT.MAXMFI) THEN
C  Too many MCL files.
               IF (WEAK) THEN
                  CALL I2STR(MAXMFI,STR,LES)
                  WRITE (IUOUT,70) STR(1:LES), FILE(1:LENF)
               END IF
               MERR = -2
               CLOSE (NU)
               RETURN
            END IF
            MCLUNI(NUMFI) = NU
         END IF
      END IF
C
C  Informative printout.
      IF (MERR.EQ.0 .AND. STRO) THEN
         IF (OPE) THEN
            CALL I2STR(NU,STR,LES)
            IF (IOPT.EQ.1) THEN
               WRITE (IUOUT,30) FILE(1:LENF), STR(1:LES)
            ELSE
               WRITE (IUOUT,35) FILE(1:LENF), STR(1:LES)
            END IF
         ELSE
            IF (IOPT.EQ.1) THEN
               WRITE (IUOUT,40) FILE(1:LENF)
            ELSE
               WRITE (IUOUT,45) FILE(1:LENF)
            END IF
         END IF
      END IF
C
20    FORMAT (/$(FC)'Error: File "',A,'" (UNIT = ',A,') is open, ',
     &        'not as a sequential file however.')
25    FORMAT (/$(FC)'Error: Cannot determine the access mode of ',
     &        'file "',A,'" (UNIT = ',A,').')
30    FORMAT (/$(FC)'File "',A,'" (UNIT = ',A,') has been rewound.')
35    FORMAT (/$(FC)'File "',A,'" (UNIT = ',A,') has ',
     &        'been positioned at End-of-File.')
40    FORMAT (/$(FC)'File "',A,'" has been rewound.')
45    FORMAT (/$(FC)'File "',A,'" has been positioned ',
     &        'at End-of-File.')
60    FORMAT (/$(FC)'File "',A,'" is your current ',A,' file.')
61    FORMAT ($(FC)'It cannot be rewound.')
62    FORMAT ($(FC)'It cannot be positioned at End-of-File.')
70    FORMAT (/$(FC)'Error: You have too many (',A,') MCL files open.'
     &        /$(FC)'       File "',A,'" cannot be opened.')
      END
C  ---------------------------------------------------------------------
      SUBROUTINE RUNMCL ( ISREDI, MERR )
C  ---------------------------------------------------------------------
C
C  Description:
C    Implements the RUNMCL command.
C    Syntax: RUNMCL file
C
C  Output arguments:
C    MERR       Error indicator.
C
C  Input / Output arguments:
C    ISREDI     On input indicates whether output redirection
C               (">" or ">>") is in effect for the current command.
C               ISREDI = .TRUE.  -> Redirection in effect.
C               ISREDI = .FALSE. -> No redirection.
C               On output is always set to .FALSE., indicating that
C               the effect of an output redirection will not be undone
C               by the Merlin operating system.
C
C  ---------------------------------------------------------------------
      IMPLICIT $(TYPE) (A-H,O-Z)
C  Arguments:
      LOGICAL ISREDI
C  ---------------------------------------------------------------------
C  Common blocks:
      PARAMETER ( MAXINL = 20 )
      INTEGER BACKIN(MAXINL), WHATIN(MAXINL), REDIN(MAXINL)
      COMMON /INNER1/ LEVIN, BACKIN, WHATIN, REDIN
      PARAMETER ( MAXONL = 20 )
      INTEGER BACOUT(MAXONL), FROUT(MAXONL)
      COMMON /OUTER/ LEVOUT, BACOUT, FROUT
      PARAMETER ( MAXMFI = 20 )
      COMMON /MCLV2/ ISTACK, INFILE, MBASE, INAD, MPC, MCLLNO, LENBUF,
     &               IBUF, MCLUNI(MAXMFI), NUMFI
      LOGICAL STRO, WEAK
      COMMON /PRILOG/ STRO, WEAK
      COMMON /UNITS/ IUINP, IUOUT
      PARAMETER ( MXBUF = $(MCLBUF), MXPTR = MXBUF/5 )
      CHARACTER*(MXBUF) MCLBUF
      INTEGER MCLPTR(MXPTR)
      COMMON /MCLBU1/ MCLBUF
      COMMON /MCLBU2/ NPTR, ICPTR, MCLPTR
      COMMON /CNTRL/ IDISK, MACECH, LIMCLK, IECOMA, IPROEC, MCLUBU
      COMMON /MODE3/ MODEV, IFFORM, INMCL
C  ---------------------------------------------------------------------
C  Local variables:
      PARAMETER ( MXFILE = 80 )
      CHARACTER*(MXFILE) FILE
      CHARACTER*80 LINE
      CHARACTER*10 STR
      LOGICAL EXI, OPE
C  ---------------------------------------------------------------------
C
      MERR = 0
C
C  Check number of arguments.
      CALL ARGNO(NARGS)
      IF (NARGS.GT.1) THEN
         IF (WEAK) WRITE (IUOUT,17)
         MERR = -2
         RETURN
      END IF
C
C  Get MOC file name.
      IF (NARGS.EQ.0) THEN
         WRITE (IUOUT,30)
         CALL INNOCR (FILE,.FALSE., MERR)
         IF (MERR.NE.0) RETURN
         LEF = LENGTH(FILE)
      ELSE
         CALL CMDARG(1,FILE,LEF)
      END IF
C
C  Get info for the file.
      INQUIRE (FILE=FILE(1:LEF),EXIST=EXI,OPENED=OPE,NUMBER=NUN,
     &         IOSTAT=IERR)
      IF (IERR.NE.0) THEN
         CALL ERRINQ(FILE,IERR)
         MERR = -2
         RETURN
      END IF
C
      IF (.NOT.EXI) THEN
C  The file does not exist.
         CALL ERREXI(FILE)
         MERR = -2
         RETURN
      END IF
C
      IF (OPE) THEN
C  The file is open.
         CALL ERRBUS(FILE,NUN)
         MERR = -2
         RETURN
      END IF
C
C  Allocate unit.
      INFILE = NUNIT()
      IF (INFILE.EQ.-1) THEN
         CALL ERRUAL(FILE)
         MERR = -2
         RETURN
      END IF
C
C  Open the file.
      OPEN (INFILE,FILE=FILE(1:LEF),IOSTAT=IERR)
      IF (IERR.NE.0) THEN
         CALL ERROPE(FILE,IERR)
         MERR = -2
         RETURN
      END IF
C
C  Is this a MOC file ?
      REWIND INFILE
      READ (INFILE,'(A)',END=40) LINE
      IF (ISCOMP(LINE,'MCL-3.0').EQ.0) THEN
         IF (WEAK) WRITE (IUOUT,50) FILE(1:LEF)
         CLOSE (INFILE)
         MERR = -2
         RETURN
      END IF
C
C  Determine whether buffered or unbuffered MOC is to be used.
C  NPRT is the number of pointers in the MOC buffer. It is set to 0,
C  when the buffer is not to be used.
      NPTR = 0
      IF (MCLUBU.NE.0) THEN
         NCH = 0
         MXPTR1 = MXPTR-1
10       READ (INFILE,'(A)',END=20) LINE
         LL = LENGTH(LINE)
         IF (LL.EQ.0) LL = 1
         NPTR = NPTR+1
         NCHLL = NCH+LL
         IF (NPTR.GT.MXPTR1 .OR. NCHLL.GT.MXBUF) THEN
C  Too many characters or pointers. Revert to unfuffered.
            REWIND INFILE
            READ (INFILE,'(A)') LINE
            NPTR = 0
            CALL I2STR(MXBUF,STR,LES)
            IF (STRO) WRITE (IUOUT,80) STR(1:LES)
            GOTO 70
         END IF
         MCLPTR(NPTR) = NCH+1
         MCLBUF(NCH+1:NCHLL) = LINE(1:LL)
         NCH = NCHLL
         GOTO 10
C
20       CONTINUE
         NPTR = NPTR+1
         MCLPTR(NPTR) = NCH+1
         ICPTR = 0
      END IF
C
C  Increment the input level.
70    CONTINUE
      CALL INCILE(INFILE,' ',0,MERR)
      IF (MERR.NE.0) THEN
C
C  Is the input nesting too deep ?
         CLOSE (INFILE)
         INFILE = -1
         RETURN
      END IF
C
C  Has the RUNMCL command been redirected ?
      IF (ISREDI) THEN
         ISREDI = .FALSE.
         REDIN(LEVIN) = 1
         FROUT(LEVOUT) = 3
      END IF
C
C  Assign initial values for several MCL variables.
C  Number of open MCL files.
      NUMFI = 0
C  Stack pointer.
      ISTACK = 0
C  Memory base for the current routine.
      MBASE = 10000000
C  Local storage for the current routine (indirect addressing).
      INAD = 10000000
C  Program counter.
      MPC = 0
C  MCL source line number for debug purposes.
      MCLLNO = -1
C  Length of the current line in the output buffer.
      LENBUF = 0
C  Current line in the buffer.
      IBUF = 1
C  We are running MCL.
      INMCL = 1
      RETURN
C
C  End-of-File in the MOC file.
40    CONTINUE
      IF (WEAK) WRITE (IUOUT,60) FILE(1:LEF)
      CLOSE (INFILE)
      MERR = -2
C
17    FORMAT (/$(FC)'Only one file name may be specified.')
30    FORMAT (/$(FC)'Enter file name:')
50    FORMAT (/$(FC)'File "',A,'" does not appear to be a valid ',
     &        'Merlin Object Code file.')
60    FORMAT (/$(FC)'File "',A,'" appears to be empty.')
80    FORMAT (/$(FC)'This MCL program is too large for the ',
     &        'current buffer (',A,' chars).'
     &        /$(FC)'The program will run unbuffered, which may ',
     &        'affect its performance.'/)
      END
C  ---------------------------------------------------------------------
      SUBROUTINE SETDER ( MODE )
C  ---------------------------------------------------------------------
C
C  Description:
C    Sets the derivative mode, according to the input argument MODE.
C
C  Input arguments:
C    MODE       The derivative mode to set.
C
C  ---------------------------------------------------------------------
C  Arguments:
      CHARACTER*(*) MODE
C  ---------------------------------------------------------------------
C  Common blocks:
      PARAMETER ( MXV = $(MXV) )
      CHARACTER*5 GRADI
      COMMON /DERIMO/ GRADI(MXV)
      CHARACTER*10 DERIVA, PRINTO, PROCES, JACOMO
      COMMON /MODEIN/ DERIVA, PRINTO, PROCES, JACOMO
      COMMON /UPDATE/ IGRADU, IHESSU
      COMMON /DIMEN/ NOD, NOF
C  ---------------------------------------------------------------------
C
      DERIVA = MODE
      DO 10,I=1,NOD
10       GRADI(I) = MODE
      IGRADU = 0
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE SETFFO ( IOPT, MERR )
C  ---------------------------------------------------------------------
C
C  Description:
C    Implements the SOS and GENERAL commands.
C    Syntax: SOS
C            GENERAL
C    Sets the functional form of the objective function, which in turn
C    determines the appropriate
C
C  Input arguments:
C    IOPT       Determines the form of the objective function.
C               IOPT = 0 -> Implements the GENERAL command and sets the
C                           functional form to general. The use must
C                           provide FUNCTION FUNMIN.
C               IOPT = 1 -> Implements the SOS command and sets the
C                           functional form to Sum-of-Squares. The
C                           user must provide SUBROUTINE SUBSUM.
C
C  Output arguments:
C    MERR       Error indicator.
C
C  ---------------------------------------------------------------------
C  Common blocks:
      LOGICAL STRO, WEAK
      COMMON /PRILOG/ STRO, WEAK
      COMMON /UNITS/ IUINP, IUOUT
      COMMON /MODE3/ MODEV, IFFORM, INMCL
      COMMON /DIMEN/ NOD, NOF
C  ---------------------------------------------------------------------
C
      MERR = 0
      IF (IOPT.EQ.0) THEN
         IF (STRO) WRITE (IUOUT,10)
         IFFORM = 0
      ELSE
         IF (NOF.EQ.0) THEN
            IF (WEAK) WRITE (IUOUT,20)
            MERR = -2
         ELSE
            IF (STRO) WRITE (IUOUT,30)
            IFFORM = 1
         END IF
      END IF
C
10    FORMAT (/$(FC)'Functional form has been set to "general".'
     &        /$(FC)'Merlin will be calling "function funmin".')
20    FORMAT (/$(FC)'You cannot enter the Sum-of-Squares mode because'
     &        /$(FC)'The number of terms specified in the call to'
     &        /$(FC)'subroutine Merlin was 0.')
30    FORMAT (/$(FC)'Functional form has been set to "sum-of-squares".'
     &        /$(FC)'Merlin will be calling "subroutine subsum".')
      END
C  ---------------------------------------------------------------------
      SUBROUTINE SHDIS ( MERR )
C  ---------------------------------------------------------------------
C
C  Description:
C    Implements the SHORTDIS command.
C    Syntax: SHORTDIS [ index_specification ... ]
C
C  Output arguments:
C    MERR       Error indicator.
C
C  ---------------------------------------------------------------------
C  Common blocks:
      PARAMETER ( MXV = $(MXV) )
      COMMON /IWORK/ INDE(MXV)
      LOGICAL STRO, WEAK
      COMMON /PRILOG/ STRO, WEAK
      COMMON /UNITS/ IUINP, IUOUT
      COMMON /TOTCAL/ ITNOCS, IPARTC, ITDER, IPDER, ITHESS, IPHESS,
     &                ITJAC, IPJAC
C  ---------------------------------------------------------------------
C
      MERR = 0
      IF (ITNOCS.EQ.0) RETURN
C
C  Parse any command line arguments for index specifications.
      CALL CHI(INDE,NCH,-2,MERR)
      IF (MERR.NE.0) RETURN
      IF (.NOT.STRO) RETURN
C
C  Display title.
      WRITE (IUOUT,*)
      CALL TITDIS
C
C  Display partial counters.
      CALL CNTDIS
C
C  Display column headers.
      IF (NCH.NE.0) WRITE (IUOUT,10)
C
C  Display variables.
      CALL WRITE2(INDE,NCH)
C
10    FORMAT (1X,'Index',2X,'Name',7X,'Fix',7X,'Parameter value',7X,
     &        'Left margin',3X,'Right margin')
      END
C  ---------------------------------------------------------------------
      SUBROUTINE STPDIS ( MERR )
C  ---------------------------------------------------------------------
C
C  Description:
C    Implements the STEPDIS command.
C    Syntax: STEPDIS [ index_specification ... ]
C
C  Output arguments:
C    MERR       Error indicator.
C
C  ---------------------------------------------------------------------
      IMPLICIT $(TYPE) (A-H,O-Z)
C  Common blocks:
      PARAMETER ( MXV = $(MXV) )
      COMMON /ROSTEP/ STEP(MXV)
      CHARACTER*10 NICKNA
      COMMON /NAMES/ NICKNA(MXV)
      COMMON /IWORK/ INDE(MXV)
      LOGICAL STRO, WEAK
      COMMON /PRILOG/ STRO, WEAK
      COMMON /UNITS/ IUINP, IUOUT
C  ---------------------------------------------------------------------
C  Local variables:
      CHARACTER*30 STR
      CHARACTER*7 CFIX
C  ---------------------------------------------------------------------
C
C  Parse any command line arguments for index specifications.
      CALL CHI(INDE,NCH,-2,MERR)
      IF (MERR.NE.0) RETURN
C
      IF (STRO) THEN
         WRITE (IUOUT,30)
         DO 10,I=1,NCH
            IND = INDE(I)
            CALL R2STR(STEP(IND),STR,LES)
            CALL FIXSTR(IND,CFIX)
            WRITE (IUOUT,20) IND, NICKNA(IND), CFIX, STEP(IND)
10       CONTINUE
      END IF
C
30    FORMAT (/$(FC)'Index',2X,'Name',6X,'Fix',7X,'Step'
     &        /$(FC)37('-'))
20    FORMAT ($(FC)I5,') ',A10,A,1X,1PG12.5)
      END
C  ---------------------------------------------------------------------
      SUBROUTINE TARG ( MERR )
C  ---------------------------------------------------------------------
C
C  Description:
C    Implemets the TARGET command.
C    Syntax: TARGET target_value
C
C  Output arguments:
C    MERR       Error indicator.
C
C  ---------------------------------------------------------------------
      IMPLICIT $(TYPE) (A-H,O-Z)
C  Common blocks:
      PARAMETER ( LSTA = $(MCLMEM) )
      COMMON /MCLV1/ STACK(LSTA)
      PARAMETER ( MAXMFI = 20 )
      COMMON /MCLV2/ ISTACK, INFILE, MBASE, INAD, MPC, MCLLNO, LENBUF,
     &               IBUF, MCLUNI(MAXMFI), NUMFI
      LOGICAL AIMON
      COMMON /AIM/ TARGET, AIMON
      COMMON /MODE3/ MODEV, IFFORM, INMCL
      LOGICAL STRO, WEAK
      COMMON /PRILOG/ STRO, WEAK
      COMMON /UNITS/ IUINP, IUOUT
C  ---------------------------------------------------------------------
C  Local variables:
      CHARACTER*80 CTA
      CHARACTER*30 STR
C  ---------------------------------------------------------------------
C
      CALL ARGNO(NA)
      IF (INMCL.EQ.0) THEN
C  Not in MCL.
         IF (NA.EQ.0) THEN
            IF (STRO) THEN
               IF (AIMON) THEN
                  CALL R2NRTF(TARGET,STR,LES)
                  WRITE (IUOUT,10) STR(1:LES)
               ELSE
                  WRITE (IUOUT,20)
               END IF
               WRITE (IUOUT,30)
            END IF
            CALL INREAL(TA,MERR)
            IF (MERR.EQ.1) RETURN
            IF (MERR.EQ.-1) RETURN
         ELSE IF (NA.EQ.1) THEN
            CALL CMDARG(1,CTA,LEC)
            CALL TOREAL(CTA,LEC,TA,MERR)
         ELSE
            CALL ERRNAR
            MERR = -2
            RETURN
         END IF
      ELSE
C  We are running MCL.
         TA = STACK(ISTACK)
         ISTACK = ISTACK-1
         MERR = 0
      END IF
C
      IF (MERR.EQ.-2) THEN
         IF (WEAK) WRITE (IUOUT,40)
      ELSE
         TARGET = TA
         AIMON = .TRUE.
         IF (STRO) THEN
            CALL R2NRTF(TARGET,STR,LES)
            WRITE (IUOUT,50) STR(1:LES)
         END IF
      END IF
C
10    FORMAT (/$(FC)'Current target value is ',A)
20    FORMAT (/$(FC)'Currently there is no taget value defined.')
30    FORMAT ($(FC)'Enter target value:')
40    FORMAT (/$(FC)'You must supply a real number as the ',
     &        'target value.')
50    FORMAT (/$(FC)'Target value has been set to ',A)
      END
C  ---------------------------------------------------------------------
      SUBROUTINE TRMDIS ( MERR )
C  ---------------------------------------------------------------------
C
C  Description:
C    Implements the TERMDIS command.
C    Syntax: TERMDIS [ index_specification ... ]
C
C  Output arguments:
C    MERR       Error indicator.
C
C  ---------------------------------------------------------------------
      IMPLICIT $(TYPE) (A-H,O-Z)
C  Common blocks:
      PARAMETER ( MXV = $(MXV) )
      PARAMETER ( MXT = $(MXT) )
      COMMON /PARAMS/ POINT(MXV), VAL
      COMMON /DIMEN/ NOD, NOF
      COMMON /MODE3/ MODEV, IFFORM, INMCL
      LOGICAL STRO, WEAK
      COMMON /PRILOG/ STRO, WEAK
      COMMON /UNITS/ IUINP, IUOUT
      COMMON /TOTCAL/ ITNOCS, IPARTC, ITDER, IPDER, ITHESS, IPHESS,
     &                ITJAC, IPJAC
      COMMON /RWORKT/ FT1(MXT), FT2(MXT), FT3(MXT)
      COMMON /IWORKT/ INDET(MXT)
C  ---------------------------------------------------------------------
C  Local variables:
      CHARACTER*23 CPOI
C  ---------------------------------------------------------------------
C
      MERR = 0
C
C  Check for sum-of-squares.
      IF (NOF.EQ.0) THEN
         IF (WEAK) WRITE (IUOUT,40)
         MERR = -2
         RETURN
      END IF
C
C  Check if in SOS mode.
      IF (IFFORM.NE.1) THEN
         IF (WEAK) WRITE (IUOUT,50)
         MERR = -2
         RETURN
      END IF
      IF (ITNOCS.EQ.0) RETURN
C
C  Parse any command line arguments for index specifications.
      CALL CHIX(NOF,INDET,NCH,-2,MERR)
      IF (MERR.NE.0) RETURN
      IF (.NOT.STRO .OR. NCH.EQ.0) RETURN
C
C  Display header.
      WRITE (IUOUT,10)
C
C  Call the objective function.
      CALL LSQFCN(NOF,NOD,POINT,FT1)
C
C  Display terms.
      DO 20,I=1,NCH
         IND = INDET(I)
         CALL R2RTF(FT1(IND),CPOI,LES)
         WRITE (IUOUT,30) IND, CPOI(1:LES)
20    CONTINUE
C
10    FORMAT (/$(FC)'Index',15X,'Term'
     &        /$(FC)35('-'))
30    FORMAT ($(FC)I5,' ....... ',A)
40    FORMAT (/$(FC)'You have specified TERMS=0 in the Merlin call.')
50    FORMAT (/$(FC)'You must enter the Sum-of-Squares (SOS) mode.')
      END
C  ---------------------------------------------------------------------
      SUBROUTINE TITLE ( MERR )
C  ---------------------------------------------------------------------
C
C  Description:
C    Implements the TITLE command.
C    Syntax: TITLE descriptive_title_for_this_session
C
C  Output arguments:
C    MERR       Error indicator.
C
C  ---------------------------------------------------------------------
C  Common blocks:
      LOGICAL STRO, WEAK
      COMMON /PRILOG/ STRO, WEAK
      COMMON /UNITS/ IUINP, IUOUT
      PARAMETER ( LINLEN = 80 )
      INTEGER IS(LINLEN/2), IE(LINLEN/2)
      COMMON /INDAT1/ IS, IE, NIT, LEC
      CHARACTER*(LINLEN) INPLIN, QINPL
      COMMON /INDAT2/ INPLIN, QINPL
      CHARACTER*80 MTITLE
      COMMON /MTIT1/ MTITLE
      COMMON /MTIT2/ LEMTIT
C  ---------------------------------------------------------------------
C  Local variables:
      CHARACTER*80 LINE
C  ---------------------------------------------------------------------
C
      MERR = 0
      LEM1 = LEMTIT
      CALL ARGNO(NA)
      IF (NA.EQ.0) THEN
C
C  Prompt the user for the title.
         IF (STRO) WRITE (IUOUT,10)
         CALL INNOCR(LINE,.FALSE.,MERR)
         IF (MERR.EQ.-1) RETURN
         MTITLE = LINE
      ELSE
C
C  Construct title from the command line arguments.
         LL = 0
         DO 20,I=2,NIT
            LE = IE(I)-IS(I)+1
            IF (LE.GT.0) THEN
               NLL = LL+LE
               LINE(LL+1:NLL) = QINPL(IS(I):IE(I))
               LL = NLL
            END IF
            LL = LL+1
            LINE(LL:LL) = ' '
20       CONTINUE
         LL = LL-1
         MTITLE = LINE(1:LL)
         LEMTIT = LL
      END IF
C
      LEMTIT = LENGTH(MTITLE)
      IF (LEMTIT.EQ.0) THEN
         IF (STRO .AND. LEM1.NE.0) WRITE (IUOUT,30)
      ELSE
         IF (LEMTIT.LE.60) THEN
            IF (STRO) WRITE (IUOUT,40) MTITLE(1:LEMTIT)
         ELSE
            IF (STRO) WRITE (IUOUT,40) MTITLE(1:57)//'...'
         END IF
      END IF
C
10    FORMAT (/$(FC)'Enter descriptive title for this minimization ',
     &        'session:')
30    FORMAT (/$(FC)'Title has been removed.')
40    FORMAT (/$(FC)'Title is set to: "',A,'"')
      END
C  ---------------------------------------------------------------------
      SUBROUTINE VALDIS
C  ---------------------------------------------------------------------
C
C  Description:
C    Implements the VALDIS command.
C    Syntax: VALDIS
C
C  ---------------------------------------------------------------------
      IMPLICIT $(TYPE) (A-H,O-Z)
C  Common blocks:
      PARAMETER ( MXV = $(MXV) )
      COMMON /PARAMS/ POINT(MXV), VAL
      COMMON /TOTCAL/ ITNOCS, IPARTC, ITDER, IPDER, ITHESS, IPHESS,
     &                ITJAC, IPJAC
      LOGICAL STRO, WEAK
      COMMON /PRILOG/ STRO, WEAK
      COMMON /UNITS/ IUINP, IUOUT
C  ---------------------------------------------------------------------
C  Local variables:
      CHARACTER*23 CVAL
C  ---------------------------------------------------------------------
C
      IF (ITNOCS.EQ.0) RETURN
C
C  Display title for this session.
      CALL TITDIS
C
C  Display total and partial counters.
      CALL CNTDIS
C
C  Display current value.
      IF (STRO) THEN
         CALL R2RTF(VAL,CVAL,LEV)
         WRITE (IUOUT,10) CVAL(1:LEV)
      END IF
C
10    FORMAT (/$(FC)15x,'Current value: ',A)
      END
C  ---------------------------------------------------------------------
      SUBROUTINE CHI ( INDE, NCH, IASK, IERR )
C  ---------------------------------------------------------------------
C
C  Description:
C    Subroutine CHI (CHange Index) is called whenever a Merlin command
C    requires a one-line-input that contains indices of variables.
C    FIX, LOOSE LDEMARGIN etc. are examples of such commands.
C    When an MCL program is running, indices are popped directly from
C    the MCL stack. During normal operation, command line arguments
C    are considered, and/or the user is prompted for the one-line-input,
C    according to the input option IASK. For a full explanation of
C    the various ways of index specification see the description of
C    SUBROUTINE VSPEC.
C
C  Input arguments:
C    IASK       Determines the action of this routine when no command
C               line arguments are entered.
C               IASK = 0  ->  Do not ask for an one-line-input. The user
C                             MUST supply command line arguments.
C               IASK = 1  ->  If the user enters command line arguments,
C                             use them. Otherwise ask for the one-line-
C                             input.
C               IASK = 2  ->  If the user does not enter command line
C                             arguments, assume the whole valid index
C                             space: 1...NOD
C               IASK =-2  ->  Same as IASK=2, but if running MCL, don't
C                             pop indices from the MCL stack. (Naturaly
C                             in MCL mode commands are not expected to
C                             have any arguments.)
C
C  Output arguments:
C    INDE       Array, containing in its first NCH positions the
C               requested indices.
C    NCH        Number of indices in the INDE array.
C    IERR       Error indicator.
C
C  ---------------------------------------------------------------------
C  Arguments:
      INTEGER INDE(*)
C  ---------------------------------------------------------------------
C  Common blocks:
      PARAMETER ( LINLEN = 80 )
      INTEGER IS(LINLEN/2), IE(LINLEN/2)
      COMMON /INDAT1/ IS, IE, NIT, LEC
      CHARACTER*(LINLEN) INPLIN, QINPL
      COMMON /INDAT2/ INPLIN, QINPL
      COMMON /DIMEN/ NOD, NOF
      COMMON /MODE3/ MODEV, IFFORM, INMCL
C  ---------------------------------------------------------------------
C  Local variables:
      CHARACTER*(LINLEN) LINE, NEW
      INTEGER ISL(LINLEN/2), IEL(LINLEN/2)
C  ---------------------------------------------------------------------
C
      IF (INMCL.EQ.0) THEN
C  Not in MCL mode.
         IF (NIT.EQ.1) THEN
C  No command line arguments.
            IF (IASK.EQ.0) THEN
C  This should never happen in this version of Merlin.
               CALL MABORT('No args, IASK = 0.','CHI')
            ELSE IF (IASK.EQ.1) THEN
C  Ask the user for the one-line-input.
               CALL INLINE(LINE,.TRUE.,.FALSE.,NEW,ISL,IEL,NITL,IERR)
               IF (IERR.NE.0) RETURN
               CALL CHI1(NEW,ISL,IEL,NITL,INDE,NCH,IERR)
            ELSE IF (IASK.EQ.2 .OR. IASK.EQ.-2) THEN
C  Use the whole index space: 1...NOD.
               NCH = NOD
               DO 10,I=1,NCH
10                INDE(I) = I
               IERR = 0
            ELSE
               CALL MABORT('Incorrect "iask".','CHI')
            END IF
         ELSE
C  Use command line arguments.
            CALL CHI1(QINPL,IS(2),IE(2),NIT-1,INDE,NCH,IERR)
         END IF
      ELSE
C  In MCL mode.
         IF (IASK.EQ.-2) THEN
            NCH = NOD
            DO 20,I=1,NCH
20             INDE(I) = I
            IERR = 0
         ELSE
            CALL CHI2(INDE,NCH,IERR)
         END IF
      END IF
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE CHI1 ( LINE, IS, IE, NIT, INDE, NCH, IERR )
C  ---------------------------------------------------------------------
C
C  Description:
C    Subroutine CHI1 is called by subroutine CHI to decode the indices
C    specified in the character variable LINE.
C
C  Input arguments:
C    LINE       Character variable containing the indices.
C    IS         IS(I) contains the starting character of item I
C               in LINE.
C    IE         IE(I) contains the ending character of item I in LINE.
C    NIT        Number of items in LINE.
C
C  Output arguments:
C    INDE       Array, containing in its first NCH positions the
C               requested indices.
C    NCH        Number of indices in the INDE array.
C    IERR       Error indicator.
C
C  ---------------------------------------------------------------------
C  Arguments:
      CHARACTER*(*) LINE
      INTEGER IS(*), IE(*), INDE(*)
C  ---------------------------------------------------------------------
C  Common blocks:
      PARAMETER ( MXV = $(MXV) )
      LOGICAL SPEC(MXV)
      COMMON /LWORK/ SPEC
      COMMON /DIMEN/ NOD, NOF
C  ---------------------------------------------------------------------
C
C  Pass the input to the more general, index-cracking routine VSPEC.
      CALL VSPEC(LINE,IS,IE,NIT,SPEC,IERR)
      IF (IERR.NE.0) RETURN
C
C  SPEC(I) is .TRUE. is index I was specified in the input LINE.
      NCH = 0
      DO 10,I=1,NOD
         IF (SPEC(I)) THEN
            NCH = NCH+1
            INDE(NCH) = I
         END IF
10    CONTINUE
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE CHI2 ( INDE, NIT, ICODE )
C  ---------------------------------------------------------------------
C
C  Description:
C    Subroutine CHI2 is called by subroutine CHI to pop a number of
C    indices from the MCL stack. The top of the stack contains the
C    number of indices to be popped.
C
C  Output arguments:
C    INDE       Array, containing in its first NCH positions the
C               requested indices.
C    NCH        Number of indices in the INDE array.
C    ICODE      Error indicator.
C
C  ---------------------------------------------------------------------
      IMPLICIT $(TYPE) (A-H,O-Z)
C  Arguments:
      INTEGER INDE(*)
C  ---------------------------------------------------------------------
C  Common blocks:
      PARAMETER ( LSTA = $(MCLMEM) )
      COMMON /MCLV1/ STACK(LSTA)
      PARAMETER ( MAXMFI = 20 )
      COMMON /MCLV2/ ISTACK, INFILE, MBASE, INAD, MPC, MCLLNO, LENBUF,
     &               IBUF, MCLUNI(MAXMFI), NUMFI
C  ---------------------------------------------------------------------
C
C  How many indices should we pop ?
      NIT = NINT(STACK(ISTACK))
      ISTACK = ISTACK-1
      NTERR = 0
C
C  Pop indices and check their validity: 1 <= index <= NOD
      DO 20,I=1,NIT
         CALL MPARIN(STACK(ISTACK),IND,IERR)
         IF (IERR.EQ.0) THEN
            INDE(I) = IND
         ELSE
            NTERR = NTERR+1
         END IF
         ISTACK = ISTACK-1
20    CONTINUE
C
C  Set return code according to the number of errors.
      IF (NTERR.EQ.0) THEN
         ICODE = 0
      ELSE
         ICODE = -2
      END IF
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE CHIS2 ( IOPT, MAXIT, CMD, INDE, ITEM, LINEO, IS, IE,
     &                   NCH, IERR )
C  ---------------------------------------------------------------------
C
C  Description:
C    Subroutine CHIS2 is called by subroutines CHIS and CHISX to return
C    index-string pairs when in MCL mode. Indices are popped from
C    the stack. Stirngs and their corresponding effective lengths
C    are read from the MOC file.
C
C  Input arguments:
C    IOPT       IOPT = 1 -> Called from subroutine CHIS.
C               IOPT = 2 -> Called from subroutine CHISX.
C    MAXIT      The allowed range of indices is: 1 <= index <= MAXIT
C    CMD        The Merlin command that requested the index-string
C               pairs.
C
C  Output arguments:
C    INDE       Array, containing in its first NCH positions the
C               requested indices.
C    ITEM       Integer array. The string corresponding to INDE(I)
C               is the ITEM(I)-th item in LINEO.
C    LINEO      Character variable containing the string in the one-
C               line-input as returned by the line cracking subroutine
C               QLCR.
C    IS         IS(I) contains the starting character of item I
C               in LINEO.
C    IE         IE(I) contains the ending character of item I in LINEO.
C    NCH        Number of index-string pairs.
C    IERR       Error indicator.
C
C  ---------------------------------------------------------------------
      IMPLICIT $(TYPE) (A-H,O-Z)
C  Arguments:
      INTEGER INDE(*), ITEM(*), IS(*), IE(*)
      CHARACTER*(*) LINEO, CMD
C  ---------------------------------------------------------------------
C  Common blocks:
      PARAMETER ( LSTA = $(MCLMEM) )
      COMMON /MCLV1/ STACK(LSTA)
      LOGICAL STRO, WEAK
      COMMON /PRILOG/ STRO, WEAK
      COMMON /UNITS/ IUINP, IUOUT
      PARAMETER ( MAXMFI = 20 )
      COMMON /MCLV2/ ISTACK, INFILE, MBASE, INAD, MPC, MCLLNO, LENBUF,
     &               IBUF, MCLUNI(MAXMFI), NUMFI
C  ---------------------------------------------------------------------
C  Local variables:
      CHARACTER C
C  ---------------------------------------------------------------------
C
C  How many indices should we pop ?
      CALL ININ(NCH,IERR)
      IF (IERR.NE.0) RETURN
      NTERR = 0
      NCLI = 0
      LENACT = LEN(LINEO)
      DO 20,I=1,NCH
C
C  Pop index and check its validity: 1 <= INDEX <= MAXIT
         IF (IOPT.EQ.1) THEN
            CALL MPARIN(STACK(ISTACK-I+1),IND,IERR)
         ELSE
            CALL MPARIX(MAXIT,STACK(ISTACK-I+1),IND,IERR)
         END IF
         IF (IERR.EQ.0) THEN
            INDE(I) = IND
         ELSE
            NTERR = NTERR+1
         END IF
C
C  Read in the string value.
         CALL ININ(LL,IERR)
         IF (IERR.NE.0) RETURN
         ISTA = NCLI+1
         IEND = NCLI+LL
         IF (IEND.GT.LENACT) THEN
            IF (WEAK) WRITE (IUOUT,40) CMD
            CALL INNOCR(C,.FALSE.,IERR)
            DO 30,J=I+1,NCH
               CALL ININ(LL,IERR)
               IF (IERR.NE.0) RETURN
               CALL INNOCR(C,.FALSE.,IERR)
               IF (IERR.NE.0) RETURN
30          CONTINUE
            NCH = I-1
            GOTO 21
         END IF
         CALL INNOCR(LINEO(ISTA:IEND),.FALSE.,IERR)
         IF (IERR.NE.0) RETURN
         IS(I) = ISTA
         IE(I) = IEND
         NCLI = IEND
         ITEM(I) = I
20    CONTINUE
21    CONTINUE
      ISTACK = ISTACK-NCH
C
C  Set return code according to the number of errors.
      IF (NTERR.EQ.0) THEN
         IERR = 0
      ELSE
         IERR = -2
      END IF
C
40    FORMAT (/$(FC)'Warning: This ',A,' command has too many ',
     &        'index-string pairs.',
     &        /$(FC)'         Some of the changes were ignored.')
      END
C  ---------------------------------------------------------------------
      SUBROUTINE CHIS ( INDE, ITEM, NCH, LINEO, ISO, IEO, IASK, IERR )
C  ---------------------------------------------------------------------
C
C  Description:
C    Subroutine CHIS (CHange Index-String) is called whenever a Merlin
C    command requires a one-line-input that contains index-string pairs.
C    GODFATHER, CFLAG etc. are examples of such commands.
C    When an MCL program is running, indices are popped directly from
C    the MCL stack.
C    During normal operation, command line arguments are considered,
C    and/or the user is prompted for the one-line-input, according to
C    the input option IASK. For a full explanation of the various ways
C    of specifying index-string pairs see the description of SUBROUTINE
C    VSPEC.
C
C  Input arguments:
C    IASK       Determines the action of this routine when no command
C               line arguments are entered.
C               IASK = 0  ->  Do not ask for an one-line-input. The user
C                             MUST supply command line arguments.
C               IASK = 1  ->  If the user enters command line arguments,
C                             use them. Otherwise ask for the one-line-
C                             input.
C
C  Output arguments:
C    INDE       Array, containing in its first NCH positions the
C               requested indices.
C    ITEM       Integer array. The string corresponding to INDE(I)
C               is the ITEM(I)-th item in LINEO.
C    NCH        Number of index-string pairs.
C    LINEO      Character variable containing the string in the one-
C               line-input as returned by the line cracking subroutine
C               QLCR.
C    ISO        ISO(I) contains the starting character of item I
C               in LINEO.
C    IEO        IEO(I) contains the ending character of item I in LINEO.
C    IERR       Error indicator.
C
C  ---------------------------------------------------------------------
C  Arguments:
      INTEGER INDE(*), ITEM(*), ISO(*), IEO(*)
      CHARACTER*(*) LINEO
C  ---------------------------------------------------------------------
C  Common blocks:
      PARAMETER ( LINLEN = 80 )
      INTEGER IS(LINLEN/2), IE(LINLEN/2)
      COMMON /INDAT1/ IS, IE, NIT, LEC
      CHARACTER*(LINLEN) INPLIN, QINPL
      COMMON /INDAT2/ INPLIN, QINPL
      COMMON /DIMEN/ NOD, NOF
      COMMON /MODE3/ MODEV, IFFORM, INMCL
C  ---------------------------------------------------------------------
C  Local variables:
      CHARACTER*(LINLEN) LINE, NEW
      INTEGER ISL(LINLEN/2), IEL(LINLEN/2)
C  ---------------------------------------------------------------------
C
      IF (INMCL.EQ.0) THEN
C  Not in MCL mode.
         IF (NIT.EQ.1) THEN
C  No command line arguments.
            IF (IASK.EQ.1) THEN
C  Ask the user for the one-line-input.
               CALL INLINE(LINE,.TRUE.,.FALSE.,NEW,ISL,IEL,NITL,
     &                              IERR)
               IF (IERR.NE.0) RETURN
               CALL CHIS1(NEW,ISL,IEL,NITL,INDE,ITEM,NCH,IERR)
               IF (IERR.NE.0) RETURN
               LINEO = NEW
               DO 30,I=1,NITL
                  ISO(I) = ISL(I)
                  IEO(I) = IEL(I)
30             CONTINUE
            ELSE
C  This should never happen in this version of Merlin.
               CALL MABORT('No args, IASK = 0.','CHIS')
            END IF
         ELSE
C  Use command line arguments.
            CALL CHIS1(QINPL,IS(2),IE(2),NIT-1,INDE,ITEM,NCH,IERR)
            IF (IERR.NE.0) RETURN
            LINEO = QINPL
            DO 20,I=1,NIT-1
               I1 = I+1
               ISO(I) = IS(I1)
               IEO(I) = IE(I1)
20          CONTINUE
         END IF
      ELSE
C  In MCL mode.
            CALL CHIS2(1,NOD,QINPL(IS(1):IE(1)),INDE,ITEM,LINEO,ISO,
     &                 IEO,NCH,IERR)
      END IF
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE CHIS1 ( LINE, IS, IE, NIT, INDE, ITEM, NCH, ICODE )
C  ---------------------------------------------------------------------
C
C  Description:
C    Subroutine CHIS1 is called by subroutine CHIS to decode the
C    index-string pairs specified in the character variable LINE.
C
C  Input arguments:
C    LINE       Character variable containing the index-string pairs.
C    IS         IS(I) contains the starting character of item I
C               in LINE.
C    IE         IE(I) contains the ending character of item I in LINE.
C    NIT        Number of items in LINE.
C
C  Output arguments:
C    INDE       Array, containing in its first NCH positions the
C               requested indices.
C    ITEM       Integer array. The string corresponding to INDE(I)
C               is the ITEM(I)-th item in LINE.
C    NCH        Number of index-string pairs.
C    ICODE      Error indicator.
C
C  ---------------------------------------------------------------------
C  Arguments:
      CHARACTER*(*) LINE
      INTEGER IS(*), IE(*), INDE(*), ITEM(*)
C  ---------------------------------------------------------------------
C  Common blocks:
      PARAMETER ( MXV = $(MXV) )
      LOGICAL STRO, WEAK
      COMMON /PRILOG/ STRO, WEAK
      COMMON /UNITS/ IUINP, IUOUT
      LOGICAL SPEC(MXV)
      COMMON /LWORK/ SPEC
      COMMON /DIMEN/ NOD, NOF
C  ---------------------------------------------------------------------
C
      DO 10,I=1,NOD
10       INDE(I) = 0
C
      NTERR = 0
      I = 0
      NCH = 0
C
C  Loop over all items in the input line.
20    CONTINUE
      I = I+1
      IF (I.GT.NIT) GOTO 40
C
C  Call the more general, index-cracking routive VSPEC. Pass only
C  one item of LINE.
      CALL VSPEC(LINE,IS(I),IE(I),1,SPEC,IERR)
      IF (IERR.NE.0) NTERR = NTERR+1
      I = I+1
      IF (I.GT.NIT) THEN
         IF (WEAK) WRITE (IUOUT,60)
         ICODE = -2
         RETURN
      END IF
C
C  Fill in INDE, so that INDE(J) = 1 if index J was specified in the
C  input LINE.
      DO 30,J=1,NOD
         IF (SPEC(J)) THEN
            INDE(J) = 1
            ITEM(J) = I
         END IF
30    CONTINUE
      GOTO 20
C
C  Any errors ?
40    CONTINUE
      IF (NTERR.NE.0) THEN
         ICODE = -2
         RETURN
      END IF
C
C  At this point INDE(I) = 0 or INDE(I) = 1. The array must be
C  compacted to satisfy its description as an output argument.
      NCH = 0
      DO 50,I=1,NOD
         IF (INDE(I).NE.0) THEN
            NCH = NCH+1
            INDE(NCH) = I
            ITEM(NCH) = ITEM(I)
         END IF
50    CONTINUE
      ICODE = 0
C
60    FORMAT (/$(FC)'Last item in input line is obsolete.')
      END
C  ---------------------------------------------------------------------
      SUBROUTINE CHISX ( INDE, ITEM, NCH, MAXIT, LINEO, ISO, IEO, IASK,
     &                   IERR )
C  ---------------------------------------------------------------------
C
C  Description:
C    Subroutine CHISX is called whenever a Merlin command requires an
C    one-line-input that contains index-string pairs. While subroutine
C    CHIS considers index-string pairs for the minimization variables,
C    this subroutine returns a general index to be used in commands
C    such as CFLAG.
C    Nevertheless, due to workspace restrictions, the index range
C    (as specified by MAXIT) cannot extend beyond the maximum number of
C    variables MXV.
C    When an MCL program is running, indices are popped directly from
C    the MCL stack.
C    During normal operation, command line arguments are considered,
C    and/or the user is prompted for the one-line-input, according
C    to the input option IASK. For a full explanation of the various
C    ways of index specification see the description of SUBROUTINE
C    ISPEC.
C
C  Input arguments:
C    MAXIT      The allowed range of indices is: 1 <= index <= MAXIT
C    IASK       Determines the action of this routine when no command
C               line arguments are entered.
C               IASK = 0  ->  Do not ask for an one-line-input. The user
C                             MUST supply command line arguments.
C               IASK = 1  ->  If the user enters command line arguments,
C                             use them. Otherwise ask for the one-line-
C                             input.
C
C  Output arguments:
C    INDE       Array, containing in its first NCH positions the
C               requested indices.
C    ITEM       Integer array. The string corresponding to INDE(I)
C               is the ITEM(I)-th item in LINEO.
C    NCH        Number of index-string pairs.
C    LINEO      Character variable containing the string in the one-
C               line-input as returned by the line cracking subroutine
C               QLCR.
C    ISO        ISO(I) contains the starting character of item I
C               in LINEO.
C    IEO        IEO(I) contains the ending character of item I in LINEO.
C    IERR       Error indicator.
C
C  ---------------------------------------------------------------------
C  Arguments:
      INTEGER INDE(*), ITEM(*), ISO(*), IEO(*)
      CHARACTER*(*) LINEO
C  ---------------------------------------------------------------------
C  Common blocks:
      PARAMETER ( LINLEN = 80 )
      INTEGER IS(LINLEN/2), IE(LINLEN/2)
      COMMON /INDAT1/ IS, IE, NIT, LEC
      CHARACTER*(LINLEN) INPLIN, QINPL
      COMMON /INDAT2/ INPLIN, QINPL
      COMMON /MODE3/ MODEV, IFFORM, INMCL
C  ---------------------------------------------------------------------
C  Local variables:
      CHARACTER*(LINLEN) LINE, NEW
      INTEGER ISL(LINLEN/2), IEL(LINLEN/2)
C  ---------------------------------------------------------------------
C
      IF (INMCL.EQ.0) THEN
C  Not in MCL mode.
         IF (NIT.EQ.1) THEN
C  No command line arguments.
            IF (IASK.EQ.1) THEN
C  Ask the user for the one-line-input.
               CALL INLINE(LINE,.TRUE.,.FALSE.,NEW,ISL,IEL,NITL,
     &                              IERR)
               IF (IERR.NE.0) RETURN
               CALL CHISX1(MAXIT,NEW,ISL,IEL,NITL,INDE,ITEM,NCH,
     &                              IERR)
               IF (IERR.NE.0) RETURN
               LINEO = NEW
               DO 30,I=1,NITL
                  ISO(I) = ISL(I)
                  IEO(I) = IEL(I)
30             CONTINUE
            ELSE
C  This should never happen in this version of Merlin.
               CALL MABORT('No args, IASK = 0.','CHISX')
            END IF
         ELSE
C  Use command line arguments.
            CALL CHISX1(MAXIT,QINPL,IS(2),IE(2),NIT-1,INDE,ITEM,
     &                         NCH,IERR)
            IF (IERR.NE.0) RETURN
            LINEO = QINPL
            DO 20,I=1,NIT-1
               I1 = I+1
               ISO(I) = IS(I1)
               IEO(I) = IE(I1)
20          CONTINUE
         END IF
      ELSE
C  In MCL mode.
         CALL CHIS2(2,MAXIT,QINPL(IS(1):IE(1)),INDE,ITEM,LINEO,ISO,
     &                 IEO,NCH,IERR)
      END IF
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE CHISX1 ( MAXIT, LINE, IS, IE, NIT, INDE, ITEM, NCH,
     &                    ICODE )
C  ---------------------------------------------------------------------
C
C  Description:
C    Subroutine CHISX1 is called by subroutine CHISX to decode the
C    index-string pairs specified in the character variable LINE.
C
C  Input arguments:
C    MAXIT      The allowed range of indices is: 1 <= index <= MAXIT
C    LINE       Character variable containing the index-string pairs.
C    IS         IS(I) contains the starting character of item I
C               in LINE.
C    IE         IE(I) contains the ending character of item I in LINE.
C    NIT        Number of items in LINE.
C
C  Output arguments:
C    INDE       Array, containing in its first NCH positions the
C               requested indices.
C    ITEM       Integer array. The string corresponding to INDE(I)
C               is the ITEM(I)-th item in LINE.
C    NCH        Number of index-string pairs.
C    ICODE      Error indicator.
C
C  ---------------------------------------------------------------------
C  Arguments:
      CHARACTER*(*) LINE
      INTEGER IS(*), IE(*), INDE(*), ITEM(*)
C  ---------------------------------------------------------------------
C  Common blocks:
      PARAMETER ( MXV = $(MXV) )
      LOGICAL SPEC(MXV)
      COMMON /LWORK/ SPEC
      LOGICAL STRO, WEAK
      COMMON /PRILOG/ STRO, WEAK
      COMMON /UNITS/ IUINP, IUOUT
C  ---------------------------------------------------------------------
C
      DO 10,I=1,MAXIT
10       INDE(I) = 0
C
      NTERR = 0
      I = 0
      NCH = 0
C
C  Loop over all items in the input line.
20    CONTINUE
      I = I+1
      IF (I.GT.NIT) GOTO 40
C
C  Call the more index-cracking routive ISPEC. Pass only
C  one item of LINE.
      CALL ISPEC(LINE,IS(I),IE(I),1,MAXIT,SPEC,IERR)
      IF (IERR.NE.0) NTERR = NTERR+1
      I = I+1
      IF (I.GT.NIT) THEN
         IF (WEAK) WRITE (IUOUT,60)
         ICODE = -2
         RETURN
      END IF
C
C  Fill in INDE, so that INDE(J) = 1 if index J was specified in the
C  input LINE.
      DO 30,J=1,MAXIT
         IF (SPEC(J)) THEN
            INDE(J) = 1
            ITEM(J) = I
         END IF
30    CONTINUE
      GOTO 20
C
C  Any errors ?
40    CONTINUE
      IF (NTERR.NE.0) THEN
         ICODE = -2
         RETURN
      END IF
C
C  At this point INDE(I) = 0 or INDE(I) = 1. The array must be
C  compacted to satisfy its description as an output argument.
      NCH = 0
      DO 50,I=1,MAXIT
         IF (INDE(I).NE.0) THEN
            NCH = NCH+1
            INDE(NCH) = I
            ITEM(NCH) = ITEM(I)
         END IF
50    CONTINUE
      ICODE = 0
C
60    FORMAT (/$(FC)'Last item in input line is obsolete.')
      END
C  ---------------------------------------------------------------------
      SUBROUTINE CHIV ( INDE, VALUE, NCH, IASK, IERR )
C  ---------------------------------------------------------------------
C
C  Description:
C    Subroutine CHIV (CHange Index-Value) is called whenever a Merlin
C    command requires a one-line-input that contains index-value pairs.
C    POINT, STEP, LMARGIN etc. are examples of such commands.
C    When an MCL program is running, indices and values are popped
C    directly from the MCL stack. During normal operation, command line
C    arguments are considered, and/or the user is prompted for the
C    one-line-input, according to the input option IASK. For a full
C    explanation of the various ways of specifying index-value pairs
C    see the description of SUBROUTINE VSPEC.
C
C  Input arguments:
C    IASK       Determines the action of this routine when no command
C               line arguments are entered.
C               IASK = 0  ->  Do not ask for an one-line-input. The user
C                             MUST supply command line arguments.
C               IASK = 1  ->  If the user enters command line arguments,
C                             use them. Otherwise ask for the one-line-
C                             input.
C
C  Output arguments:
C    INDE       Array, containing in its first NCH positions the
C               requested indices.
C    VALUE      Array, containing in its first NCH positions the
C               requested values.
C    NCH        Number of index-value pairs.
C    IERR       Error indicator.
C
C  ---------------------------------------------------------------------
      IMPLICIT $(TYPE) (A-H,O-Z)
C  Arguments:
      INTEGER INDE(*)
      DIMENSION VALUE(*)
C  ---------------------------------------------------------------------
C  Common blocks:
      PARAMETER ( LINLEN = 80 )
      INTEGER IS(LINLEN/2), IE(LINLEN/2)
      COMMON /INDAT1/ IS, IE, NIT, LEC
      CHARACTER*(LINLEN) INPLIN, QINPL
      COMMON /INDAT2/ INPLIN, QINPL
      COMMON /MODE3/ MODEV, IFFORM, INMCL
C  ---------------------------------------------------------------------
C  Local variables:
      CHARACTER*(LINLEN) LINE, NEW
      INTEGER ISL(LINLEN/2), IEL(LINLEN/2)
C  ---------------------------------------------------------------------
C
      IF (INMCL.EQ.0) THEN
C  Not in MCL mode.
         IF (NIT.EQ.1) THEN
C  No command line arguments.
            IF (IASK.EQ.1) THEN
C  Ask the user for the one-line-input.
               CALL INLINE(LINE,.TRUE.,.FALSE.,NEW,ISL,IEL,
     &                              NITL,IERR)
               IF (IERR.NE.0) RETURN
               CALL CHIV1(NEW,ISL,IEL,NITL,INDE,VALUE,NCH,IERR)
            ELSE
C  This should never happen in this version of Merlin.
               CALL MABORT('No args, IASK = 0.','CHIV')
            END IF
         ELSE
C  Use command line arguments.
            CALL CHIV1(QINPL,IS(2),IE(2),NIT-1,INDE,VALUE,NCH,IERR)
         END IF
      ELSE
C  In MCL mode.
         CALL CHIV2(INDE,VALUE,NCH,IERR)
      END IF
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE CHIV1 ( LINE, IS, IE, NIT, INDE, VALUE, NCH, ICODE )
C  ---------------------------------------------------------------------
C
C  Description:
C    Subroutine CHIV1 is called by subroutine CHIV to decode the
C    index-value pairs specified in the character variable LINE.
C
C  Input arguments:
C    LINE       Character variable containing the index-value pairs.
C    IS         IS(I) contains the starting character of item I
C               in LINE.
C    IE         IE(I) contains the ending character of item I in LINE.
C    NIT        Number of items in LINE.
C
C  Output arguments:
C    INDE       Array, containing in its first NCH positions the
C               requested indices.
C    VALUE      Array, containing in its first NCH positions the
C               requested values.
C    NCH        Number of index-value pairs.
C    ICODE      Error indicator.
C
C  ---------------------------------------------------------------------
      IMPLICIT $(TYPE) (A-H,O-Z)
C  Arguments:
      CHARACTER*(*) LINE
      INTEGER IS(*), IE(*), INDE(*)
      DIMENSION VALUE(*)
C  ---------------------------------------------------------------------
C  Common blocks:
      PARAMETER ( MXV = $(MXV) )
      LOGICAL SPEC(MXV)
      COMMON /LWORK/ SPEC
      LOGICAL STRO, WEAK
      COMMON /PRILOG/ STRO, WEAK
      COMMON /UNITS/ IUINP, IUOUT
      COMMON /DIMEN/ NOD, NOF
C  ---------------------------------------------------------------------
C
      DO 10,I=1,NOD
10       INDE(I) = 0
C
      NTERR = 0
      I = 0
      NCH = 0
C
C  Loop over all items in the input line.
20    CONTINUE
      I = I+1
      IF (I.GT.NIT) GOTO 40
C
C  Call the more general, index-cracking routive VSPEC. Pass only
C  one item of LINE.
      CALL VSPEC(LINE,IS(I),IE(I),1,SPEC,IERR)
      IF (IERR.NE.0) NTERR = NTERR+1
      I = I+1
      IF (I.GT.NIT) THEN
         IF (WEAK) WRITE (IUOUT,60)
         ICODE = -2
         RETURN
      END IF
C
C  Try to decode the "value". A proper real number is expected.
      LE = IE(I)-IS(I)+1
      CALL TOREAL(LINE(IS(I):IE(I)),LE,VL,IVERR)
      NCH = NCH+1
      IF (IVERR.EQ.0) THEN
C
C  Fill in INDE, so that INDE(J) = 1 if index J was specified in the
C  input LINE.
         DO 30,J=1,NOD
            IF (SPEC(J)) THEN
               INDE(J) = 1
               VALUE(J) = VL
            END IF
30       CONTINUE
      ELSE
         NTERR = NTERR+1
         IF (WEAK) WRITE (IUOUT,70) LINE(IS(I):IE(I))
      END IF
      GOTO 20
C
C  Any errors ?
40    CONTINUE
      IF (NTERR.NE.0) THEN
         ICODE = -2
         RETURN
      END IF
C
C  At this point INDE(I) = 0 or INDE(I) = 1. The array must be
C  compacted to satisfy its description as an output argument.
      NCH = 0
      DO 50,I=1,NOD
         IF (INDE(I).NE.0) THEN
            NCH = NCH+1
            INDE(NCH) = I
            VALUE(NCH) = VALUE(I)
         END IF
50    CONTINUE
      ICODE = 0
C
60    FORMAT (/$(FC)'Last item in input line is obsolete.')
70    FORMAT (/$(FC)'"',A,'" is not a number.')
      END
C  ---------------------------------------------------------------------
      SUBROUTINE CHIV2 ( INDE, VALUE, NCH, ICODE )
C  ---------------------------------------------------------------------
C
C  Description:
C    Subroutine CHIV2 is called by subroutine CHIV to pop a number of
C    index-value pairs from the MCL stack. The top of the stack
C    contains the number of index-value pairs to be popped.
C
C  Output arguments:
C    INDE       Array, containing in its first NCH positions the
C               requested indices.
C    VALUE      Array, containing in its first NCH positions the
C               requested values.
C    NCH        Number of index-value pairs.
C    ICODE      Error indicator.
C
C  ---------------------------------------------------------------------
      IMPLICIT $(TYPE) (A-H,O-Z)
C  Arguments:
      INTEGER INDE(*)
      DIMENSION VALUE(*)
C  ---------------------------------------------------------------------
C  Common blocks:
      PARAMETER ( LSTA = $(MCLMEM) )
      COMMON /MCLV1/ STACK(LSTA)
      PARAMETER ( MAXMFI = 20 )
      COMMON /MCLV2/ ISTACK, INFILE, MBASE, INAD, MPC, MCLLNO, LENBUF,
     &               IBUF, MCLUNI(MAXMFI), NUMFI
C  ---------------------------------------------------------------------
C
C  How many index-value pairs should we pop ?
      NCH = NINT(STACK(ISTACK))
      ISTACK = ISTACK-1
      NTERR = 0
C
C  Pop indices and check their validity: 1 <= index <= NOD
      DO 20,I=NCH,1,-1
         VL = STACK(ISTACK)
         ISTACK=ISTACK-1
         CALL MPARIN(STACK(ISTACK),IND,IERR)
         IF (IERR.EQ.0) THEN
            INDE(I) = IND
            VALUE(I) = VL
         ELSE
            NTERR = NTERR+1
         END IF
         ISTACK = ISTACK-1
20    CONTINUE
C
C  Set return code according to the number of errors.
      IF (NTERR.EQ.0) THEN
         ICODE = 0
      ELSE
         ICODE = -2
      END IF
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE CHIVX ( MAXIT, INDE, VALUE, NCH, IASK, IERR )
C  ---------------------------------------------------------------------
C
C  Description:
C    Subroutine CHIVX is called whenever a Merlin command requires an
C    one-line-input that contains index-value pairs. While subroutine
C    CHIV considers index-value pairs for the minimization variables,
C    this subroutine returns a general index to be used in commands
C    such as FLAG.
C    Nevertheless, due to workspace restrictions, the index range
C    (as specified by MAXIT) cannot extend beyond the maximum number of
C    variables MXV.
C    When an MCL program is running, indices are popped directly from
C    the MCL stack. During normal operation, command line arguments
C    are considered, and/or the user is prompted for the one-line-input,
C    according to the input option IASK. For a full explanation of
C    the various ways of index specification see the description of
C    SUBROUTINE ISPEC.
C
C  Input arguments:
C    MAXIT      The allowed range of indices is: 1 <= index <= MAXIT
C    IASK       Determines the action of this routine when no command
C               line arguments are entered.
C               IASK = 0  ->  Do not ask for an one-line-input. The user
C                             MUST supply command line arguments.
C               IASK = 1  ->  If the user enters command line arguments,
C                             use them. Otherwise ask for the one-line-
C                             input.
C
C  Output arguments:
C    INDE       Array, containing in its first NCH positions the
C               requested indices.
C    VALUE      Array, containing in its first NCH positions the
C               requested values.
C    NCH        Number of indices in the INDE array.
C    IERR       Error indicator.
C
C  ---------------------------------------------------------------------
      IMPLICIT $(TYPE) (A-H,O-Z)
C  Arguments:
      INTEGER INDE(*)
      DIMENSION VALUE(*)
C  ---------------------------------------------------------------------
C  Common blocks:
      PARAMETER ( LINLEN = 80 )
      INTEGER IS(LINLEN/2), IE(LINLEN/2)
      COMMON /INDAT1/ IS, IE, NIT, LEC
      CHARACTER*(LINLEN) INPLIN, QINPL
      COMMON /INDAT2/ INPLIN, QINPL
      COMMON /MODE3/ MODEV, IFFORM, INMCL
C  ---------------------------------------------------------------------
C  Local variables:
      CHARACTER*(LINLEN) LINE, NEW
      INTEGER ISL(LINLEN/2), IEL(LINLEN/2)
C  ---------------------------------------------------------------------
C
      IF (INMCL.EQ.0) THEN
C  Not in MCL mode.
         IF (NIT.EQ.1) THEN
C  No command line arguments.
            IF (IASK.EQ.1) THEN
C  Ask the user for the one-line-input.
               CALL INLINE(LINE,.TRUE.,.FALSE.,NEW,ISL,IEL,NITL,
     &                              IERR)
               IF (IERR.NE.0) RETURN
               CALL CHIVX1(MAXIT,NEW,ISL,IEL,NITL,INDE,VALUE,
     &                              NCH,IERR)
            ELSE
C  This should never happen in this version of Merlin.
               CALL MABORT('No args, IASK = 0.','CHIVX')
            END IF
         ELSE
C  Use command line arguments.
            CALL CHIVX1(MAXIT,QINPL,IS(2),IE(2),NIT-1,INDE,
     &                         VALUE,NCH,IERR)
         END IF
      ELSE
C  In MCL mode.
         CALL CHIVX2(MAXIT,INDE,VALUE,NCH,IERR)
      END IF
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE CHIVX1 ( MAXIT, LINE, IS, IE, NIT, INDE, VALUE, NCH,
     &                    ICODE )
C  ---------------------------------------------------------------------
C
C  Description:
C    Subroutine CHIVX1 is called by subroutine CHIVX to decode the
C    index-value pairs specified in the character variable LINE.
C
C  Input arguments:
C    MAXIT      The allowed range of indices is: 1 <= index <= MAXIT
C    LINE       Character variable containing the index-value pairs.
C    IS         IS(I) contains the starting character of item I
C               in LINE.
C    IE         IE(I) contains the ending character of item I in LINE.
C    NIT        Number of items in LINE.
C
C  Output arguments:
C    INDE       Array, containing in its first NCH positions the
C               requested indices.
C    VALUE      Array, containing in its first NCH positions the
C               requested values.
C    NCH        Number of index-value pairs.
C    ICODE      Error indicator.
C
C  ---------------------------------------------------------------------
      IMPLICIT $(TYPE) (A-H,O-Z)
C  Arguments:
      CHARACTER*(*) LINE
      INTEGER IS(*), IE(*), INDE(*)
      DIMENSION VALUE(*)
C  ---------------------------------------------------------------------
C  Common blocks:
      PARAMETER ( MXV = $(MXV) )
      LOGICAL STRO, WEAK
      COMMON /PRILOG/ STRO, WEAK
      COMMON /UNITS/ IUINP, IUOUT
      LOGICAL SPEC(MXV)
      COMMON /LWORK/ SPEC
C  ---------------------------------------------------------------------
C
      DO 10,I=1,MAXIT
10       INDE(I) = 0
C
      NTERR = 0
      I = 0
      NCH = 0
C
C  Loop over all items in the input line.
20    CONTINUE
      I = I+1
      IF (I.GT.NIT) GOTO 40
C
C  Call the more index-cracking routive ISPEC. Pass only
C  one item of LINE.
      CALL ISPEC(LINE,IS(I),IE(I),1,MAXIT,SPEC,IERR)
      IF (IERR.NE.0) NTERR = NTERR+1
      I = I+1
      IF (I.GT.NIT) THEN
         IF (WEAK) WRITE (IUOUT,60)
         ICODE = -2
         RETURN
      END IF
C
C  Try to decode the "value". A proper real number is expected.
      LE = IE(I)-IS(I)+1
      CALL TOREAL(LINE(IS(I):IE(I)),LE,VL,IVERR)
      NCH = NCH+1
      IF (IVERR.EQ.0) THEN
C
C  Fill in INDE, so that INDE(J) = 1 if index J was specified in the
C  input LINE.
         DO 30,J=1,MAXIT
            IF (SPEC(J)) THEN
               INDE(J) = 1
               VALUE(J) = VL
            END IF
30       CONTINUE
      ELSE
         NTERR = NTERR+1
         IF (WEAK) WRITE (IUOUT,70) LINE(IS(I):IE(I))
      END IF
      GOTO 20
C
C  Any errors ?
40    CONTINUE
      IF (NTERR.NE.0) THEN
         ICODE = -2
         RETURN
      END IF
C
C  At this point INDE(I) = 0 or INDE(I) = 1. The array must be
C  compacted to satisfy its description as an output argument.
      NCH = 0
      DO 50,I=1,MAXIT
         IF (INDE(I).NE.0) THEN
            NCH = NCH+1
            INDE(NCH) = I
            VALUE(NCH) = VALUE(I)
         END IF
50    CONTINUE
      ICODE = 0
C
60    FORMAT (/$(FC)'Last item in input line is obsolete.')
70    FORMAT (/$(FC)'"',A,'" is not a number.')
      END
C  ---------------------------------------------------------------------
      SUBROUTINE CHIVX2 ( MAXIT, INDE, VALUE, NCH, ICODE )
C  ---------------------------------------------------------------------
C
C  Description:
C    Subroutine CHIVX2 is called by subroutine CHIVX to pop a number of
C    index-value pairs from the MCL stack. The top of the stack
C    contains the number of index-value pairs to be popped.
C
C  Input arguments:
C    MAXIT      The allowed range of indices is: 1 <= index <= MAXIT
C
C  Output arguments:
C    INDE       Array, containing in its first NCH positions the
C               requested indices.
C    VALUE      Array, containing in its first NCH positions the
C               requested values.
C    NCH        Number of index-value pairs.
C    ICODE      Error indicator.
C
C  ---------------------------------------------------------------------
      IMPLICIT $(TYPE) (A-H,O-Z)
C  Arguments:
      INTEGER INDE(*)
      DIMENSION VALUE(*)
C  ---------------------------------------------------------------------
C  Common blocks:
      PARAMETER ( LSTA = $(MCLMEM) )
      COMMON /MCLV1/ STACK(LSTA)
      PARAMETER ( MAXMFI = 20 )
      COMMON /MCLV2/ ISTACK, INFILE, MBASE, INAD, MPC, MCLLNO, LENBUF,
     &               IBUF, MCLUNI(MAXMFI), NUMFI
C  ---------------------------------------------------------------------
C
C  How many index-value pairs should we pop ?
      NCH = NINT(STACK(ISTACK))
      ISTACK = ISTACK-1
      NTERR = 0
C
C  Pop indices and check their validity: 1 <= index <= MAXIT
      DO 20,I=NCH,1,-1
         VL = STACK(ISTACK)
         ISTACK=ISTACK-1
         CALL MPARIX(MAXIT,STACK(ISTACK),IND,IERR)
         IF (IERR.EQ.0) THEN
            INDE(I) = IND
            VALUE(I) = VL
         ELSE
            NTERR = NTERR+1
         END IF
         ISTACK = ISTACK-1
20    CONTINUE
C
C  Set return code according to the number of errors.
      IF (NTERR.EQ.0) THEN
         ICODE = 0
      ELSE
         ICODE = -2
      END IF
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE CHIX ( MAXIT, INDE, NCH, IASK, IERR )
C  ---------------------------------------------------------------------
C
C  Description:
C    Subroutine CHIX is called whenever a Merlin command requires an
C    one-line-input that contains indices. While subroutine CHI
C    considers indices for the minimization variables, this subroutine
C    returns a general index to be used in commands such as FLAGDIS.
C    Nevertheless, due to workspace restrictions, the index range
C    (as specified by MAXIT) cannot extend beyond NOD.
C    When an MCL program is running, indices are popped directly from
C    the MCL stack. During normal operation, command line arguments
C    are considered, and/or the user is prompted for the one-line-input,
C    according to the input option IASK. For a full explanation of
C    the various ways of index specification see the description of
C    SUBROUTINE ISPEC.
C
C  Input arguments:
C    MAXIT      The allowed range of indices is: 1 <= index <= MAXIT
C    IASK       Determines the action of this routine when no command
C               line arguments are entered.
C               IASK = 0  ->  Do not ask for an one-line-input. The user
C                             MUST supply command line arguments.
C               IASK = 1  ->  If the user enters command line arguments,
C                             use them. Otherwise ask for the one-line-
C                             input.
C               IASK = 2  ->  If the user does not enter command line
C                             arguments, assume the whole valid index
C                             space: 1...MAXIT
C               IASK =-2  ->  Same as IASK=2, but if running MCL, don't
C                             pop indices from the MCL stack. (Naturaly
C                             in MCL mode commands are not expected to
C                             have any arguments.)
C
C  Output arguments:
C    INDE       Array, containing in its first NCH positions the
C               requested indices.
C    NCH        Number of indices in the INDE array.
C    IERR       Error indicator.
C
C  ---------------------------------------------------------------------
C  Arguments:
      INTEGER INDE(*)
C  ---------------------------------------------------------------------
C  Common blocks:
      PARAMETER ( LINLEN = 80 )
      INTEGER IS(LINLEN/2), IE(LINLEN/2)
      COMMON /INDAT1/ IS, IE, NIT, LEC
      CHARACTER*(LINLEN) INPLIN, QINPL
      COMMON /INDAT2/ INPLIN, QINPL
      COMMON /MODE3/ MODEV, IFFORM, INMCL
C  ---------------------------------------------------------------------
C  Local variables:
      CHARACTER*(LINLEN) LINE, NEW
      INTEGER ISL(LINLEN/2), IEL(LINLEN/2)
C  ---------------------------------------------------------------------
C
      IF (INMCL.EQ.0) THEN
C  Not in MCL mode.
         IF (NIT.EQ.1) THEN
C  No command line arguments.
            IF (IASK.EQ.0) THEN
C  This should never happen in this version of Merlin.
               CALL MABORT('No args, IASK = 0.','CHIX')
            ELSE IF (IASK.EQ.1) THEN
C  Ask the user for the one-line-input.
               CALL INLINE(LINE,.TRUE.,.FALSE.,NEW,ISL,IEL,NITL,IERR)
               IF (IERR.NE.0) RETURN
               CALL CHIX1(MAXIT,NEW,ISL,IEL,NITL,INDE,NCH,IERR)
            ELSE IF (IASK.EQ.2 .OR. IASK.EQ.-2) THEN
C  Use the whole index space: 1...MAXIT.
               NCH = MAXIT
               DO 10,I=1,NCH
10                INDE(I) = I
               IERR = 0
            ELSE
               CALL MABORT('Incorrect "iask".','CHI')
            END IF
         ELSE
C  Use command line arguments.
            CALL CHIX1(MAXIT,QINPL,IS(2),IE(2),NIT-1,INDE,NCH,IERR)
         END IF
      ELSE
C  In MCL mode.
         IF (IASK.EQ.-2) THEN
            NCH = MAXIT
            DO 20,I=1,NCH
20             INDE(I) = I
            IERR = 0
         ELSE
            CALL CHIX2(MAXIT,INDE,NCH,IERR)
         END IF
      END IF
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE CHIX1 ( MAXIT, LINE, IS, IE, NIT, INDE, NCH, IERR )
C  ---------------------------------------------------------------------
C
C  Description:
C    Subroutine CHIX1 is called by subroutine CHIX to decode the indices
C    specified in the character variable LINE.
C
C  Input arguments:
C    MAXIT      The allowed range of indices is: 1 <= index <= MAXIT
C    LINE       Character variable containing the indices.
C    IS         IS(I) contains the starting character of item I
C               in LINE.
C    IE         IE(I) contains the ending character of item I in LINE.
C    NIT        Number of items in LINE.
C
C  Output arguments:
C    INDE       Array, containing in its first NCH positions the
C               requested indices.
C    NCH        Number of indices in the INDE array.
C    IERR       Error indicator.
C
C  ---------------------------------------------------------------------
C  Arguments:
      CHARACTER*(*) LINE
      INTEGER IS(*), IE(*), INDE(*)
C  ---------------------------------------------------------------------
C  Common blocks:
      PARAMETER ( MXT = $(MXT) )
      LOGICAL SPECT(MXT)
      COMMON /LWORKT/ SPECT
C  ---------------------------------------------------------------------
C
C  Pass the input to the index-cracking routine ISPEC.
      CALL ISPEC(LINE,IS,IE,NIT,MAXIT,SPECT,IERR)
      IF (IERR.NE.0) RETURN
C
C  SPECT(I) is .TRUE. is index I was specified in the input LINE.
      NCH = 0
      DO 10,I=1,MAXIT
         IF (SPECT(I)) THEN
            NCH = NCH+1
            INDE(NCH) = I
         END IF
10    CONTINUE
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE CHIX2 ( MAXIT, INDE, NIT, ICODE )
C  ---------------------------------------------------------------------
C
C  Description:
C    Subroutine CHIX2 is called by subroutine CHI to pop a number of
C    indices from the MCL stack. The top of the stack contains the
C    number of indices to be popped.
C
C  Input arguments:
C    MAXIT      The allowed range of indices is: 1 <= index <= MAXIT
C
C  Output arguments:
C    INDE       Array, containing in its first NCH positions the
C               requested indices.
C    NCH        Number of indices in the INDE array.
C    ICODE      Error indicator.
C
C  ---------------------------------------------------------------------
      IMPLICIT $(TYPE) (A-H,O-Z)
C  Arguments:
      INTEGER INDE(*)
C  ---------------------------------------------------------------------
C  Common blocks:
      PARAMETER ( LSTA = $(MCLMEM) )
      COMMON /MCLV1/ STACK(LSTA)
      PARAMETER ( MAXMFI = 20 )
      COMMON /MCLV2/ ISTACK, INFILE, MBASE, INAD, MPC, MCLLNO, LENBUF,
     &               IBUF, MCLUNI(MAXMFI), NUMFI
C  ---------------------------------------------------------------------
C
C  How many indices should we pop ?
      NIT = NINT(STACK(ISTACK))
      ISTACK = ISTACK-1
      NTERR = 0
C
C  Pop indices and check their validity: 1 <= index <= MAXIT
      DO 20,I=1,NIT
         CALL MPARIX(MAXIT,STACK(ISTACK),IND,IERR)
         IF (IERR.EQ.0) THEN
            INDE(I) = IND
         ELSE
            NTERR = NTERR+1
         END IF
         ISTACK = ISTACK-1
20    CONTINUE
C
C  Set return code according to the number of errors.
      IF (NTERR.EQ.0) THEN
         ICODE = 0
      ELSE
         ICODE = -2
      END IF
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE ISPEC ( LINE, IS, IE, NIT, MAXIT, SPEC, IERR )
C  ---------------------------------------------------------------------
C
C  Description:
C    Subroutine ISPEC is a simple version of subroutine VSPEC. It parses
C    a sequence of characters (LINE) that contain index specifications.
C    While subroutine VSPEC considers indices for the minimization
C    variables, this subroutine uses a more general index in the
C    range 1...MAXIT. Hence, attribute specifications and names of
C    variables are not allowed in LINE. The input argument LINE should
C    be separated in items by the line cracking routine QLCR.
C    Each item in LINE may be one of the following:
C
C      1) An index: 1 2 5 etc. Indices should be in the range 1...MAXIT
C      2) A range of the form: START-END where START and END are
C         indices. Eg. 7-9 2-8 etc.
C         START should be less than or equal to END.
C         Either one of START, END may be omitted, but not both.
C         If omitted, START defaults to 1 and END to the maximum
C         number of indices MAXIT.
C
C    In addition if the negation symbol ! is put in front of an
C    index specification, the meaning of the specification is
C    reversed:
C
C      1) In front of an index means all indices except the given
C         one: !2 !4
C      2) In fron of a range means all indices that do not
C         belong to the range: !2-7  !-5
C
C    Note that no spaces or tab characters are allowed in between
C    the negation symbol ! and the index specification that follows.
C
C  Input arguments:
C    LINE       Character variable. This is the sequence of characters
C               to be parsed.
C    IS         Starting character positions of each item in LINE,
C               as returned by subroutine QLCR.
C    IE         Ending character positions of each item in LINE,
C               as returned by subroutine QLCR.
C    NIT        Number of items in LINE.
C    MAXIT      The allowed range for indices is 1...MAXIT
C
C  Output arguments:
C    SPEC       Logical array.
C               SPEC(I) = .TRUE. if index I was included in the index
C                                specifications in LINE.
C               SPEC(I) = .FALSE. if index I was NOT included in the
C                                 index specifications in LINE.
C    IERR       Error indicator.
C
C  ---------------------------------------------------------------------
C  Arguments:
      CHARACTER*(*) LINE
      INTEGER IS(*), IE(*)
      LOGICAL SPEC(*)
C  ---------------------------------------------------------------------
C  Common blocks:
      LOGICAL STRO, WEAK
      COMMON /PRILOG/ STRO, WEAK
      COMMON /UNITS/ IUINP, IUOUT
C  ---------------------------------------------------------------------
C
      IERR = 0
C
C  Assign initial values.
      DO 10,I=1,MAXIT
10       SPEC(I) = .FALSE.
C
C  Total number of errors.
      NTERR = 0
C
C  Loop through all items.
      DO 100,I=1,NIT
C
C  Check for negation.
         IF (LINE(IS(I):IS(I)).EQ.'!') THEN
            IS1 = IS(I)+1
            INEG = 1
         ELSE
            IS1 = IS(I)
            INEG = 0
         END IF
         IE1 = IE(I)
         LE1 = IE1-IS1+1
         IF (LE1.EQ.0) THEN
            IF (WEAK) WRITE (IUOUT,210)
            NTERR = NTERR+1
            GOTO 100
         END IF
C
C  Check for a range of numbers.
         IDA = INDEX(LINE(IS1:IE1),'-')
         IF (IDA.NE.0) THEN
            LEB = IDA-1
            LEA = LE1-IDA
            IF (LEB.EQ.0 .AND. LEA.EQ.0) THEN
               IF (WEAK) WRITE (IUOUT,220)
               NTERR = NTERR+1
               GOTO 100
            END IF
            IF (LEB.NE.0) THEN
               CALL PARINX(LINE(IS1:IS1+LEB-1),LEB,MAXIT,NNB,IERR)
               IF (IERR.NE.0) THEN
                  NTERR = NTERR+1
                  GOTO 100
               END IF
            ELSE
               NNB = 1
            END IF
            IF (LEA.NE.0) THEN
               CALL PARINX(LINE(IS1+IDA:IE1),LEA,MAXIT,NNA,IERR)
               IF (IERR.NE.0) THEN
                  NTERR = NTERR+1
                  GOTO 100
               END IF
            ELSE
               NNA = MAXIT
            END IF
            IF (NNB.GT.NNA) THEN
               IF (WEAK) WRITE (IUOUT,230) LINE(IS1:IE1)
               NTERR = NTERR+1
               GOTO 100
            END IF
            IF (INEG.EQ.0) THEN
               DO 20,J=NNB,NNA
20                SPEC(J) = .TRUE.
            ELSE
               DO 30,J=1,NNB-1
30                SPEC(J) = .TRUE.
               DO 40,J=NNA+1,MAXIT
40                SPEC(J) = .TRUE.
            END IF
            GOTO 100
         END IF
C
C  No other possibility. Assume a single number.
         CALL PARINX(LINE(IS1:IE1),LE1,MAXIT,NN,IERR)
         IF (IERR.EQ.0) THEN
            IF (INEG.EQ.0) THEN
               SPEC(NN) = .TRUE.
            ELSE
               DO 50,J=1,MAXIT
50                IF (J.NE.NN) SPEC(J) = .TRUE.
            END IF
         ELSE
            NTERR = NTERR+1
         END IF
C
100   CONTINUE
C
C  Set the error code according to the number of errors NTERR.
      IF (NTERR.EQ.0) THEN
         IERR = 0
      ELSE
         IERR = -2
      END IF
C
210   FORMAT (/$(FC)'Nothing follows the negation symbol !')
220   FORMAT (/$(FC)'A - only is not accepted.')
230   FORMAT (/$(FC)'The range "',A,'" is in reverse order.')
      END
C  ---------------------------------------------------------------------
      SUBROUTINE MPARIN ( VAL, IDX, IERR )
C  ---------------------------------------------------------------------
C
C  Description:
C    Translates the top of the MCL stack into an index of a minimization
C    variable and checks for its validity.
C
C  Input arguments:
C    VAL        Contents of the top of the MCL stack.
C
C  Output arguments:
C    IDX        Index of a minimization variable.
C    IERR       Error indicator.
C
C  ---------------------------------------------------------------------
      IMPLICIT $(TYPE) (A-H,O-Z)
C  Common blocks:
      LOGICAL STRO, WEAK
      COMMON /PRILOG/ STRO, WEAK
      COMMON /UNITS/ IUINP, IUOUT
      COMMON /DIMEN/ NOD, NOF
C  ---------------------------------------------------------------------
C  Local variables:
      CHARACTER*10 TMP, TMP2
C  ---------------------------------------------------------------------
C
      IERR = 0
C
C  Translate to an integer.
      IDX = NINT(VAL)
C
C  Is it a valid index ?
      IF (IDX.LT.1) THEN
         IF (WEAK) THEN
            CALL I2STR(IDX,TMP,LET)
            WRITE (IUOUT,30) TMP(1:LET)
         END IF
         IDX = 0
         IERR = -2
      ELSE IF (IDX.GT.NOD) THEN
         IF (WEAK) THEN
            CALL I2STR(IDX,TMP,LET)
            CALL I2STR(NOD,TMP2,LET2)
            WRITE (IUOUT,40) TMP(1:LET), TMP2(1:LET2)
         END IF
         IDX = 0
         IERR = -2
      END IF
C
30    FORMAT (/$(FC)'The index specification "',A,
     &        '" is incorrect (Too small).')
40    FORMAT (/$(FC)'The index specification "',A,'" is incorrect. ',
     &        'There are only ',A,' variables.')
      END
C  ---------------------------------------------------------------------
      SUBROUTINE MPARIX ( MAXIT, VAL, IDX, IERR )
C  ---------------------------------------------------------------------
C
C  Description:
C    Translates the top of the MCL stack into an index and checks for
C    its validity. While subroutine MPARIN considers indices for
C    minimization variables, this subroutine returns a genereal index
C    to be used with commands such as FLAG.
C
C  Input arguments:
C    MAXIT      Allowed range of IDX is: 1 <= IDX <= MAXIT
C    VAL        Contents of the top of the MCL stack.
C
C  Output arguments:
C    IDX        Translated index.
C    IERR       Error indicator.
C
C  ---------------------------------------------------------------------
      IMPLICIT $(TYPE) (A-H,O-Z)
C  Common blocks:
      LOGICAL STRO, WEAK
      COMMON /PRILOG/ STRO, WEAK
      COMMON /UNITS/ IUINP, IUOUT
C  ---------------------------------------------------------------------
C  Local variables:
      CHARACTER*10 TMP, TMP2
C  ---------------------------------------------------------------------
C
      IERR = 0
C
C  Translate to an integer.
      IDX = NINT(VAL)
C
C  Is it a valid index ?
      IF (IDX.LT.1) THEN
         IF (WEAK) THEN
            CALL I2STR(IDX,TMP,LET)
            WRITE (IUOUT,30) TMP(1:LET)
         END IF
         IDX = 0
         IERR = -2
      ELSE IF (IDX.GT.MAXIT) THEN
         IF (WEAK) THEN
            CALL I2STR(IDX,TMP,LET)
            CALL I2STR(MAXIT,TMP2,LET2)
            WRITE (IUOUT,40) TMP(1:LET), TMP2(1:LET2)
         END IF
         IDX = 0
         IERR = -2
      END IF
C
30    FORMAT (/$(FC)'The index specification "',A,
     &        '" is incorrect (Too small).')
40    FORMAT (/$(FC)'The index specification "',A,'" is incorrect. ',
     &        'There are only ',A,' indices.')
      END
C  ---------------------------------------------------------------------
      SUBROUTINE PARIN ( STR, LE, IDX, IERR )
C  ---------------------------------------------------------------------
C
C  Description:
C    Translates a string into the index of a minimization variable.
C    The variable may be specified in STR either by its name (as
C    assigned by a GODFATHER command), or by its index.
C
C  Input arguments:
C    STR        The string to be translated. Contains the name of a
C               variable or an index.
C    LE         Number of useful characters in STR.
C
C  Output arguments:
C    IDX        Index of the variable.
C    IERR       Error indicator.
C
C  ---------------------------------------------------------------------
C  Arguments:
      CHARACTER*(*) STR
C  ---------------------------------------------------------------------
C  Common blocks:
      PARAMETER ( MXV = $(MXV) )
      CHARACTER*10 NICKNA
      COMMON /NAMES/ NICKNA(MXV)
      COMMON /DIMEN/ NOD, NOF
      LOGICAL STRO, WEAK
      COMMON /PRILOG/ STRO, WEAK
      COMMON /UNITS/ IUINP, IUOUT
C  ---------------------------------------------------------------------
C  Local variables:
      CHARACTER*10 TMP
C  ---------------------------------------------------------------------
C
C  Cannot translate an empty string.
      IF (STR(1:LE).EQ.' ') THEN
         IF (WEAK) WRITE (IUOUT,20) STR(1:LE)
         IERR = -2
         RETURN
      END IF
C
C  Initialy treat the input string as an index.
      CALL TOINT(STR,LE,IDX,IERR)
      IF (IERR.EQ.0) THEN
C
C  Translation to an integer succeded. Is it a valid index ?
         IF (IDX.LT.1) THEN
            IF (WEAK) WRITE (IUOUT,30) STR(1:LE)
            IDX = 0
            IERR = -2
         ELSE IF (IDX.GT.NOD) THEN
            IF (WEAK) THEN
               CALL I2STR(NOD,TMP,LET)
               WRITE (IUOUT,40) STR(1:LE), TMP(1:LET)
            END IF
            IDX = 0
            IERR = -2
         END IF
      ELSE
C
C  Translation to an integer failed. Search the names.
         DO 10,I=1,NOD
            IF (ISCOMP(STR(1:LE),NICKNA(I)).NE.0) THEN
               IDX = I
               IERR = 0
               RETURN
            END IF
10       CONTINUE
C
C  STR is not the name of a variable.
         IF (WEAK) WRITE (IUOUT,20) STR(1:LE)
         IDX = 0
         IERR = -2
      END IF
C
20    FORMAT (/$(FC)'There is no such minimization variable: "',A,'".')
30    FORMAT (/$(FC)'The index specification "',A,
     &        '" is incorrect (Too small).')
40    FORMAT (/$(FC)'The index specification "',A,'" is incorrect. ',
     &        'There are only ',A,' variables.')
      END
C  ---------------------------------------------------------------------
      SUBROUTINE PARINX ( STR, LE, MAXIT, IDX, IERR )
C  ---------------------------------------------------------------------
C
C  Description:
C    Translates a string into an index. While subroutine PARIN
C    considers indices for the minimization variables, this subroutine
C    returns a general index to be used in commands such as FLAG.
C
C  Input arguments:
C    STR        The string to be translated.
C    LE         Number of useful characters in STR.
C    MAXIT      Allowed range for IDX is: 1 <= IDX <= MAXIT
C
C  Output arguments:
C    IDX        Index contained in STR.
C    IERR       Error indicator.
C
C  ---------------------------------------------------------------------
C  Arguments:
      CHARACTER*(*) STR
C  ---------------------------------------------------------------------
C  Common blocks:
      LOGICAL STRO, WEAK
      COMMON /PRILOG/ STRO, WEAK
      COMMON /UNITS/ IUINP, IUOUT
C  ---------------------------------------------------------------------
C  Local variables:
      CHARACTER*10 TMP
C  ---------------------------------------------------------------------
C
C  Cannot translate an empty string.
      IF (STR(1:LE).EQ.' ') THEN
         IF (WEAK) WRITE (IUOUT,20) STR(1:LE)
         IERR = -2
         RETURN
      END IF
C
C  Translate STR to integer.
      CALL TOINT(STR,LE,IDX,IERR)
      IF (IERR.EQ.0) THEN
C
C  Translation to an integer succeded. Is it a valid index ?
         IF (IDX.LT.1) THEN
            IF (WEAK) WRITE (IUOUT,30) STR(1:LE)
            IDX = 0
            IERR = -2
         ELSE IF (IDX.GT.MAXIT) THEN
            IF (WEAK) THEN
               CALL I2STR(MAXIT,TMP,LET)
               WRITE (IUOUT,40) STR(1:LE), TMP(1:LET)
            END IF
            IDX = 0
            IERR = -2
         END IF
      ELSE
C
C  Translation to an integer failed.
         IF (WEAK) WRITE (IUOUT,20) STR(1:LE)
         IDX = 0
         IERR = -2
      END IF
C
20    FORMAT (/$(FC)'There is no such index: "',A,'".')
30    FORMAT (/$(FC)'The index specification "',A,
     &        '" is incorrect (Too small).')
40    FORMAT (/$(FC)'The index specification "',A,'" is incorrect. ',
     &        'There are only ',A,' indices.')
      END
C  ---------------------------------------------------------------------
      SUBROUTINE VSPEC ( LINE, IS, IE, NIT, SPEC, IERR )
C  ---------------------------------------------------------------------
C
C  Description:
C    Subroutine VSPEC parses a sequence of characters (LINE) that
C    contain specifications of minimization variables. The input
C    argument LINE should be separated in items by the line cracking
C    routine QLCR.
C    Each item in LINE may be one of the following:
C
C      1) An index: 1 2 5 etc. Indices should be in the range 1...NOD
C      2) The name of a variable, as assigned by a GODFATHER command.
C         Upper and lower case letters are equivalent in names.
C      3) A range of the form: START-END where START and END may be
C         either an index or a name. Eg. 7-9 alpha-8 sigma-bc1
C         (Assuming alpha, sigma and bc1 are valid names)
C         START should be less than or equal to END.
C         Either one of START, END may be omitted, but not both.
C         If omitted, START defaults to 1 and END to the number of
C         variables NOD. Eg. -7  alpha-   2-
C      4) An attribute of the form /xx. All variables that possess the
C         specified attribute are included. Upper and lower case
C         letters are equivalent in an attribute. The following
C         is a list of all attributes and their meaning.
C           /f   or  /fix         Fixed variables.
C           /l   or  /loose       Loose variables.
C           /lm  or  /left        Variables with left margin.
C           /rm  or  /right       Variables with right margin.
C           /m   or  /margin      Variables with both margins.
C           /fd  or  /fast        Variables whose derivative mode
C                                 is set to FAST.
C           /qd  or  /quad        Variables whose derivative mode
C                                 is set to QUAD.
C           /nd  or  /numer       Variables whose derivative mode
C                                 is set to NUMER.
C           /ad  or  /anal        Variables whose derivative mode
C                                 is set to ANAL.
C           /n   or  /named       Named variables.
C         Examples: /fd /f /lm
C
C    In addition if the negation symbol ! is put in front of an
C    index specification, the meaning of the specification is
C    reversed:
C
C      1) In front of an index means all variables except the one
C         with the given index: !2 !4
C      2) In front of a name means all variables except the one
C         with the given name: !alpha !sigma
C      3) In fron of a range means all variables that do not
C         belong to the range: !2-7  !-5  !alpha-
C      4) In front of an attribute means all variables that do
C         not possess the given attribute: !/fd !/f !/m
C
C    Note that no spaces or tab characters are allowed in between
C    the negation symbol ! and the index specification that follows.
C
C  Input arguments:
C    LINE       Character variable. This is the sequence of characters
C               to be parsed.
C    IS         Starting character positions of each item in LINE,
C               as returned by subroutine QLCR.
C    IE         Ending character positions of each item in LINE,
C               as returned by subroutine QLCR.
C    NIT        Number of items in LINE.
C
C  Output arguments:
C    SPEC       Logical array.
C               SPEC(I) = .TRUE. if the variable with index I was
C                                included in the index specifications
C                                in LINE.
C               SPEC(I) = .FALSE. if the variable with index I was NOT
C                                 included in the index specifications
C                                 in LINE.
C    IERR       Error indicator.
C
C  ---------------------------------------------------------------------
C  Arguments:
      CHARACTER*(*) LINE
      INTEGER IS(*), IE(*)
      LOGICAL SPEC(*)
C  ---------------------------------------------------------------------
C  Common blocks:
      PARAMETER ( MXV = $(MXV) )
      COMMON /DETAIL/ IXAT(MXV), MARG(MXV)
      CHARACTER*5 GRADI
      COMMON /DERIMO/ GRADI(MXV)
      CHARACTER*10 NICKNA
      COMMON /NAMES/ NICKNA(MXV)
      COMMON /DIMEN/ NOD, NOF
      COMMON /UNITS/ IUINP, IUOUT
      LOGICAL STRO, WEAK
      COMMON /PRILOG/ STRO, WEAK
C  ---------------------------------------------------------------------
C
      IERR = 0
C
C  Assign initial values.
      DO 10,I=1,NOD
10       SPEC(I) = .FALSE.
C
C  Total number of errors.
      NTERR = 0
C
C  Loop through all items.
      DO 100,I=1,NIT
C
C  Check for negation.
         IF (LINE(IS(I):IS(I)).EQ.'!') THEN
            IS1 = IS(I)+1
            INEG = 1
         ELSE
            IS1 = IS(I)
            INEG = 0
         END IF
         IE1 = IE(I)
         LE1 = IE1-IS1+1
         IF (LE1.EQ.0) THEN
            IF (WEAK) WRITE (IUOUT,210)
            NTERR = NTERR+1
            GOTO 100
         END IF
C
C  Check for a range of numbers.
         IDA = INDEX(LINE(IS1:IE1),'-')
         IF (IDA.NE.0) THEN
            LEB = IDA-1
            LEA = LE1-IDA
            IF (LEB.EQ.0 .AND. LEA.EQ.0) THEN
               IF (WEAK) WRITE (IUOUT,220)
               NTERR = NTERR+1
               GOTO 100
            END IF
            IF (LEB.NE.0) THEN
               CALL PARIN(LINE(IS1:IS1+LEB-1),LEB,NNB,IERR)
               IF (IERR.NE.0) THEN
                  NTERR = NTERR+1
                  GOTO 100
               END IF
            ELSE
               NNB = 1
            END IF
            IF (LEA.NE.0) THEN
               CALL PARIN(LINE(IS1+IDA:IE1),LEA,NNA,IERR)
               IF (IERR.NE.0) THEN
                  NTERR = NTERR+1
                  GOTO 100
               END IF
            ELSE
               NNA = NOD
            END IF
            IF (NNB.GT.NNA) THEN
               IF (WEAK) WRITE (IUOUT,230) LINE(IS1:IE1)
               NTERR = NTERR+1
               GOTO 100
            END IF
            IF (INEG.EQ.0) THEN
               DO 20,J=NNB,NNA
20                SPEC(J) = .TRUE.
            ELSE
               DO 30,J=1,NNB-1
30                SPEC(J) = .TRUE.
               DO 40,J=NNA+1,NOD
40                SPEC(J) = .TRUE.
            END IF
            GOTO 100
         END IF
C
C  Check for attribute specification.
         IF (LINE(IS1:IS1).EQ.'/') THEN
            LE1 = IE1-IS1
            IF (LE1.EQ.0) THEN
               IF (WEAK) WRITE (IUOUT,240)
               NTERR = NTERR+1
               GOTO 100
            END IF
C  Fixed variables.
            IF (ISCOMP(LINE(IS1+1:IE1),'F').NE.0 .OR.
     &                ISCOMP(LINE(IS1+1:IE1),'FIX').NE.0) THEN
               DO 50,J=1,NOD
                  IF (INEG.EQ.0) THEN
                     IF (IXAT(J).EQ.0) SPEC(J) = .TRUE.
                  ELSE
                     IF (IXAT(J).EQ.1) SPEC(J) = .TRUE.
                  END IF
50             CONTINUE
               GOTO 100
            END IF
C  Loose variables.
            IF (ISCOMP(LINE(IS1+1:IE1),'L').NE.0 .OR.
     &                ISCOMP(LINE(IS1+1:IE1),'LOOSE').NE.0) THEN
               DO 60,J=1,NOD
                  IF (INEG.EQ.0) THEN
                     IF (IXAT(J).EQ.1) SPEC(J) = .TRUE.
                  ELSE
                     IF (IXAT(J).EQ.0) SPEC(J) = .TRUE.
                  END IF
60             CONTINUE
               GOTO 100
            END IF
C  Varibles with both margins.
            IF (ISCOMP(LINE(IS1+1:IE1),'M').NE.0 .OR.
     &                ISCOMP(LINE(IS1+1:IE1),'MARGIN').NE.0) THEN
               DO 70,J=1,NOD
                  IF (INEG.EQ.0) THEN
                     IF (MARG(J).EQ.2) SPEC(J) = .TRUE.
                  ELSE
                     IF (MARG(J).NE.2) SPEC(J) = .TRUE.
                  END IF
70             CONTINUE
               GOTO 100
            END IF
C  Variables with left margin only.
            IF (ISCOMP(LINE(IS1+1:IE1),'LM').NE.0 .OR.
     &                ISCOMP(LINE(IS1+1:IE1),'LEFT').NE.0) THEN
               DO 80,J=1,NOD
                  IF (INEG.EQ.0) THEN
                     IF (MARG(J).EQ.-1) SPEC(J) = .TRUE.
                  ELSE
                     IF (MARG(J).NE.-1) SPEC(J) = .TRUE.
                  END IF
80             CONTINUE
               GOTO 100
            END IF
C  Variables with right margin only.
            IF (ISCOMP(LINE(IS1+1:IE1),'RM').NE.0 .OR.
     &                ISCOMP(LINE(IS1+1:IE1),'RIGHT').NE.0) THEN
               DO 90,J=1,NOD
                  IF (INEG.EQ.0) THEN
                     IF (MARG(J).EQ.1) SPEC(J) = .TRUE.
                  ELSE
                     IF (MARG(J).NE.1) SPEC(J) = .TRUE.
                  END IF
90             CONTINUE
               GOTO 100
            END IF
C  Varibles using FAST derivatives.
            IF (ISCOMP(LINE(IS1+1:IE1),'FD').NE.0 .OR.
     &                ISCOMP(LINE(IS1+1:IE1),'FAST').NE.0) THEN
               DO 110,J=1,NOD
                  ISAME = ISCOMP(GRADI(J),'FAST')
                  IF (INEG.EQ.0) THEN
                     IF (ISAME.NE.0) SPEC(J) = .TRUE.
                  ELSE
                     IF (ISAME.EQ.0) SPEC(J) = .TRUE.
                  END IF
110            CONTINUE
               GOTO 100
            END IF
C  Variables using QUAD derivatives.
            IF (ISCOMP(LINE(IS1+1:IE1),'QD').NE.0 .OR.
     &                ISCOMP(LINE(IS1+1:IE1),'QUAD').NE.0) THEN
               DO 120,J=1,NOD
                  ISAME = ISCOMP(GRADI(J),'QUAD')
                  IF (INEG.EQ.0) THEN
                     IF (ISAME.NE.0) SPEC(J) = .TRUE.
                  ELSE
                     IF (ISAME.EQ.0) SPEC(J) = .TRUE.
                  END IF
120            CONTINUE
               GOTO 100
            END IF
C  Variables using NUMER derivatives.
            IF (ISCOMP(LINE(IS1+1:IE1),'ND').NE.0 .OR.
     &                ISCOMP(LINE(IS1+1:IE1),'NUMER').NE.0) THEN
               DO 130,J=1,NOD
                  ISAME = ISCOMP(GRADI(J),'NUMER')
                  IF (INEG.EQ.0) THEN
                     IF (ISAME.NE.0) SPEC(J) = .TRUE.
                  ELSE
                     IF (ISAME.EQ.0) SPEC(J) = .TRUE.
                  END IF
130            CONTINUE
               GOTO 100
            END IF
C  Variables using ANAL derivatives.
            IF (ISCOMP(LINE(IS1+1:IE1),'AD').NE.0 .OR.
     &                ISCOMP(LINE(IS1+1:IE1),'ANAL').NE.0) THEN
               DO 140,J=1,NOD
                  ISAME = ISCOMP(GRADI(J),'ANAL')
                  IF (INEG.EQ.0) THEN
                     IF (ISAME.NE.0) SPEC(J) = .TRUE.
                  ELSE
                     IF (ISAME.EQ.0) SPEC(J) = .TRUE.
                  END IF
140            CONTINUE
               GOTO 100
            END IF
C  Named variables.
            IF (ISCOMP(LINE(IS1+1:IE1),'N').NE.0 .OR.
     &                ISCOMP(LINE(IS1+1:IE1),'NAMED').NE.0) THEN
               DO 150,J=1,NOD
                  ISAME = 0
                  IF (NICKNA(J).NE.' ') ISAME = 1
                  IF (INEG.EQ.0) THEN
                     IF (ISAME.NE.0) SPEC(J) = .TRUE.
                  ELSE
                     IF (ISAME.EQ.0) SPEC(J) = .TRUE.
                  END IF
150            CONTINUE
               GOTO 100
            END IF
C  Incorrect attribute.
            IF (WEAK) WRITE (IUOUT,250) LINE(IS1:IE1)
            NTERR = NTERR+1
            GOTO 100
         END IF
C
C  No other possibility. Assume a single number.
         CALL PARIN(LINE(IS1:IE1),LE1,NN,IERR)
         IF (IERR.EQ.0) THEN
            IF (INEG.EQ.0) THEN
               SPEC(NN) = .TRUE.
            ELSE
               DO 160,J=1,NOD
160               IF (J.NE.NN) SPEC(J) = .TRUE.
            END IF
         ELSE
            NTERR = NTERR+1
         END IF
C
100   CONTINUE
C
C  Set the error code according to the number of errors NTERR.
      IF (NTERR.EQ.0) THEN
         IERR = 0
      ELSE
         IERR = -2
      END IF
C
210   FORMAT (/$(FC)'Nothing follows the negation symbol !')
220   FORMAT (/$(FC)'A - only is not accepted.')
230   FORMAT (/$(FC)'The range "',A,'" is in reverse order.')
240   FORMAT (/$(FC)'Nothing follows the attribute symbol /')
250   FORMAT (/$(FC)'The attribute "',A,'" is incorrect.')
      END
C  ---------------------------------------------------------------------
      SUBROUTINE CONECP ( CONFIL, IERRU )
C  ---------------------------------------------------------------------
C
C  Description:
C    Copies the error file produced by the configuration parsing
C    routine MCONF to the standard output (*). The error file is then
C    removed.
C    This routine is called whenever there are configuration problems
C    concerning the output file.
C
C  Input arguments:
C    CONFIL     Name of the configuration file.
C    IERRU      Unit number of the error file.
C
C  ---------------------------------------------------------------------
C  Arguments:
      CHARACTER*(*) CONFIL
C  ---------------------------------------------------------------------
C  Local variables:
      CHARACTER*80 LINE
C  ---------------------------------------------------------------------
C
C  Initial message.
      CALL DASHO
      WRITE (*,150) CONFIL(1:LENGTH(CONFIL))
150   FORMAT ($(FC)'The following error(s) have been detected while '/
     &        $(FC)'processing the Merlin configuration file "',A,'":')
      CALL DASHO
C
C  Copy error file.
      REWIND IERRU
10    READ (IERRU,'(A)',END=20) LINE
      LE = LENGTH(LINE)
      IF (LE.EQ.0) THEN
         WRITE (*,*)
      ELSE
         WRITE (*,'($(FC)A)') LINE(1:LE)
      END IF
      GOTO 10
C
C  Remove error file.
20    CONTINUE
      CLOSE (IERRU,STATUS='DELETE')
      CALL DASHO
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE CONEOU ( CONFIL, IUOUT, IERRU )
C  ---------------------------------------------------------------------
C
C  Description:
C    Copies the error file produced by the configuration parsing
C    routine MCONF to the Merlin output file. The error file is then
C    removed.
C
C  Input arguments:
C    CONFIL     Name of the configuration file.
C    IUOUT      The Merlin output unit.
C    IERRU      Unit number of the error file.
C
C  ---------------------------------------------------------------------
C  Arguments:
      CHARACTER*(*) CONFIL
C  ---------------------------------------------------------------------
C  Local variables:
      CHARACTER*80 LINE
C  ---------------------------------------------------------------------
C
C  Initial message.
      CALL DASH
      WRITE (IUOUT,150) CONFIL(1:LENGTH(CONFIL))
150   FORMAT ($(FC)'The following error(s) have been detected while '/
     &        $(FC)'processing the Merlin configuration file "',A,'":')
      CALL DASH
C
C  Copy error file.
      REWIND IERRU
10    READ (IERRU,'(A)',END=20) LINE
      LE = LENGTH(LINE)
      IF (LE.EQ.0) THEN
         WRITE (IUOUT,*)
      ELSE
         WRITE (IUOUT,'($(FC)A)') LINE(1:LE)
      END IF
      GOTO 10
C
C  Remove error file.
20    CONTINUE
      CLOSE (IERRU,STATUS='DELETE')
      CALL DASH
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE DASHO
C  ---------------------------------------------------------------------
C
C  Description:
C    Prints a horizontal line of dashes. Subroutine DASHO is used before
C    the output unit is opened; hence it cannot write to IUOUT.
C
C  ---------------------------------------------------------------------
C
      WRITE (*,10)
C
10    FORMAT ($(FC)79('-'))
      END
C  ---------------------------------------------------------------------
      SUBROUTINE I2STR ( N, STR, LES )
C  ---------------------------------------------------------------------
C
C  Description:
C    Converts an integer (N) to its string representation (STR).
C
C  Input arguments:
C    N          The integer to be converted.
C
C  Output arguments:
C    STR        The first LES characters of STR contain the string
C               representation of N.
C    LES        Effective length of STR.
C
C  ---------------------------------------------------------------------
C  Arguments:
      CHARACTER*(*) STR
C  ---------------------------------------------------------------------
C  Local variables:
      CHARACTER*5 FORM
      DATA FORM / '(I__)' /
C  ---------------------------------------------------------------------
C
      LE = LEN(STR)
      WRITE (FORM(3:4),'(I2.2)') LE
      WRITE (STR,FORM) N
C
      DO 10,I=1,LE
         IF (STR(I:I).NE.' ') THEN
            DO 20,J=I,LE
               JI = J-I+1
               STR(JI:JI) = STR(J:J)
20          CONTINUE
            STR(LE-I+2:) = ' '
            LES = LE-I+1
            RETURN
         END IF
10    CONTINUE
      LES = LE
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE IRANGE ( TOKEN, NUM, ALPHA, NT, NN, IERR, PIR, NPIR,
     &                    MXPIR, PDESCF, NPL, COMM, KEY, IUOUT, LEP,
     &                    WEAK )
C  ---------------------------------------------------------------------
C
C  Description:
C    Subroutine IRANGE parses the definition of a range or set of
C    integers and fills in the appropriate data structures.
C    A range of integers has the form:
C      L S,E R
C    where: L is ( or [ indicating an open/closed interval on the left.
C           R is ) or ] indicating an open/closed interval on the right.
C           S is an integer (start of the interval) or -inf (for
C             infinity).
C           E is an integer (end of the interval) or inf (for infinity).
C    Examples: (-1,1)  (-inf,0]  [10,100)  [70,80]
C
C    A set (list) of integers has the form:
C      { I1, I2, I3, ... }
C    where I1, I2,... etc. are integers.
C
C    Panel related data structures are described in subroutine PARSP.
C    The lexical analysis tables (TOKEN, NUM, ALPHA) are described in
C    subroutine LEX.
C
C  Input arguments:
C    TOKEN      Tokens in the current line (see subroutine LEX).
C    NUM        Numbers in the current line (see subroutine LEX).
C    ALPHA      Identifiers/strings in the current line (see subroutine
C               LEX).
C    MXPIR      Maximum allowed number of entries in the PIR table.
C    PDESCF     Name of the panel description file.
C    NPL        Number of lines read so far from the panel description
C               file.
C    COMM       The panel currently under processing.
C    KEY        The keyword currently under processing.
C    IUOUT      The Merlin output unit.
C    LEP        Effective length of PDESCF.
C    WEAK       Error messages are printed only of WEAK = .TRUE.
C
C  Output arguments:
C    IERR       Error indicator.
C
C  Input / Output arguments:
C    NT         Number of tokens (TOKEN) used so far.
C    NN         Number of numbers (NUM) used so far.
C    PIR        The PIR table (see subroutine PARSP).
C    NPIR       Number of entries in the PIR array.
C
C  ---------------------------------------------------------------------
C  Arguments:
      INTEGER TOKEN(*)
      $(TYPE) NUM(*), PIR(MXPIR)
      CHARACTER*(*) ALPHA, PDESCF, COMM, KEY
      LOGICAL WEAK
C  ---------------------------------------------------------------------
C  Common blocks:
C  Codes used by the lexical analysis routine LEX.
      INTEGER LEFTP, RIGHTP, PERIOD, COMMA, GT, LT, NE, EQ, GE,
     &        LE, PLUS, MINUS, STAR, SLASH, BISTAR, COLON, EOL,
     &        IDENT, UNSI, NOTREC, STRING, QUOTE, LBRA, RBRA,
     &        SEMI, PERCE, AND, OR, XOR, NOT, EQUAL, LBRACE,
     &        RBRACE, QMARK
      COMMON /MCLSCD/ LEFTP, RIGHTP, PERIOD, COMMA, GT, LT, NE, EQ, GE,
     &                LE, PLUS, MINUS, STAR, SLASH, BISTAR, COLON, EOL,
     &                IDENT, UNSI, NOTREC, STRING, QUOTE, LBRA, RBRA,
     &                SEMI, PERCE, AND, OR, XOR, NOT, EQUAL, LBRACE,
     &                RBRACE, QMARK
C  ---------------------------------------------------------------------
C  Local variables:
      LOGICAL OK
      $(TYPE) XLE, XLI, XRI
C  ---------------------------------------------------------------------
C
      IERR = 0
      NT = NT+1
      INIT = TOKEN(NT)
      IF (INIT.EQ.LBRACE) THEN
C  A list of values is expected.
         NPIR = NPIR+1
         IF (NPIR.GT.MXPIR)
     &          CALL OVER('pir','PIR','IRANGE',MXPIR,'MXPIR')
         PIR(NPIR) = 2
C  Allocate PIR space for the # of items.
         NPIR = NPIR+1
         IF (NPIR.GT.MXPIR)
     &          CALL OVER('pir','PIR','IRANGE',MXPIR,'MXPIR')
         ISTLI = NPIR
10       CONTINUE
         NT = NT+1
         ITOK = TOKEN(NT)
         ISI = 1
         IF (ITOK.EQ.PLUS) THEN
            ISI = 1
            NT = NT+1
         ELSE IF (ITOK.EQ.MINUS) THEN
            ISI = -1
            NT = NT+1
         END IF
         IF (TOKEN(NT).NE.UNSI) THEN
            LE1 = LENGTH(KEY)
            LE2 = LENGTH(COMM)
            CALL PERRMS(PDESCF,LEP,WEAK,NPL,IUOUT)
            IF (WEAK) WRITE (IUOUT,270) KEY(1:LE1), COMM(1:LE2)
            IERR = 1
            RETURN
         END IF
         NN = NN+1
         XLI = NUM(NN)
         IF (XLI.NE.INT(XLI)) THEN
            LE1 = LENGTH(KEY)
            LE2 = LENGTH(COMM)
            CALL PERRMS(PDESCF,LEP,WEAK,NPL,IUOUT)
            IF (WEAK) WRITE (IUOUT,270) KEY(1:LE1), COMM(1:LE2)
            IERR = 1
            RETURN
         END IF
         NPIR = NPIR+1
         IF (NPIR.GT.MXPIR)
     &          CALL OVER('pir','PIR','IRANGE',MXPIR,'MXPIR')
         PIR(NPIR) = ISI*NINT(XLI)
         NT = NT+1
         IF (TOKEN(NT).EQ.COMMA) GOTO 10
         IF (TOKEN(NT).EQ.RBRACE) THEN
C  End of list ok.
            NLI = NPIR-ISTLI
            PIR(ISTLI) = NLI
            RETURN
         END IF
         LE1 = LENGTH(KEY)
         LE2 = LENGTH(COMM)
         CALL PERRMS(PDESCF,LEP,WEAK,NPL,IUOUT)
         IF (WEAK) WRITE (IUOUT,272) KEY(1:LE1), COMM(1:LE2)
         IERR = 1
         RETURN
      ELSE IF (INIT.EQ.LEFTP .OR. INIT.EQ.LBRA) THEN
C  A left open/closed interval is expected.
         LEF = 0
         IRI = 0
         IF (INIT.EQ.LEFTP) ITPL = 1
         IF (INIT.EQ.LBRA) ITPL = 2
         NT = NT+1
         ISI = 1
         IF (TOKEN(NT).EQ.PLUS) THEN
            ISI = 1
            NT = NT+1
         ELSE IF (TOKEN(NT).EQ.MINUS) THEN
            ISI = -1
            NT = NT+1
         END IF
         IF (TOKEN(NT).EQ.IDENT) THEN
            NN = NN+1
            IST = NINT(NUM(NN))
            NN = NN+1
            IEND = NINT(NUM(NN))
            IF (ISCOMP(ALPHA(IST:IEND),'inf').NE.0) THEN
               ITPL = 3*ISI
            ELSE
               LE1 = LENGTH(KEY)
               LE2 = LENGTH(COMM)
               CALL PERRMS(PDESCF,LEP,WEAK,NPL,IUOUT)
               IF (WEAK) WRITE (IUOUT,273) KEY(1:LE1), COMM(1:LE2)
               IERR = 1
               RETURN
            END IF
         ELSE IF (TOKEN(NT).EQ.UNSI) THEN
            NN = NN+1
            XLE = NUM(NN)
            IF (XLE.NE.INT(XLE)) THEN
               LE1 = LENGTH(KEY)
               LE2 = LENGTH(COMM)
               CALL PERRMS(PDESCF,LEP,WEAK,NPL,IUOUT)
               IF (WEAK) WRITE (IUOUT,274) KEY(1:LE1), COMM(1:LE2)
               IERR = 1
               RETURN
            END IF
            LEF = ISI*NINT(XLE)
         ELSE
            LE1 = LENGTH(KEY)
            LE2 = LENGTH(COMM)
            CALL PERRMS(PDESCF,LEP,WEAK,NPL,IUOUT)
            IF (WEAK) WRITE (IUOUT,273) KEY(1:LE1), COMM(1:LE2)
            IERR = 1
            RETURN
         END IF
         NT = NT+1
         IF (TOKEN(NT).NE.COMMA) THEN
            LE1 = LENGTH(KEY)
            LE2 = LENGTH(COMM)
            CALL PERRMS(PDESCF,LEP,WEAK,NPL,IUOUT)
            IF (WEAK) WRITE (IUOUT,276) KEY(1:LE1), COMM(1:LE2)
            IERR = 1
            RETURN
         END IF
C
         ITPR = 0
         NT = NT+1
         ISI = 1
         IF (TOKEN(NT).EQ.PLUS) THEN
            ISI = 1
            NT = NT+1
         ELSE IF (TOKEN(NT).EQ.MINUS) THEN
            ISI = -1
            NT = NT+1
         END IF
         IF (TOKEN(NT).EQ.IDENT) THEN
            NN = NN+1
            IST = NINT(NUM(NN))
            NN = NN+1
            IEND = NINT(NUM(NN))
            IF (ISCOMP(ALPHA(IST:IEND),'inf').NE.0) THEN
               ITPR = 3*ISI
            ELSE
               LE1 = LENGTH(KEY)
               LE2 = LENGTH(COMM)
               CALL PERRMS(PDESCF,LEP,WEAK,NPL,IUOUT)
               IF (WEAK) WRITE (IUOUT,277) KEY(1:LE1), COMM(1:LE2)
               IERR = 1
               RETURN
            END IF
         ELSE IF (TOKEN(NT).EQ.UNSI) THEN
            NN = NN+1
            XRI = NUM(NN)
            IF (XRI.NE.INT(XRI)) THEN
               LE1 = LENGTH(KEY)
               LE2 = LENGTH(COMM)
               CALL PERRMS(PDESCF,LEP,WEAK,NPL,IUOUT)
               IF (WEAK) WRITE (IUOUT,278) KEY(1:LE1), COMM(1:LE2)
               IERR = 1
               RETURN
            END IF
            IRI = ISI*NINT(XRI)
         ELSE
            LE1 = LENGTH(KEY)
            LE2 = LENGTH(COMM)
            CALL PERRMS(PDESCF,LEP,WEAK,NPL,IUOUT)
            IF (WEAK) WRITE (IUOUT,277) KEY(1:LE1), COMM(1:LE2)
            IERR = 1
            RETURN
         END IF
C
         NT = NT+1
         IF (TOKEN(NT).EQ.RIGHTP) THEN
            IF (ITPR.EQ.0) ITPR = 1
         ELSE IF (TOKEN(NT).EQ.RBRA) THEN
            IF (ITPR.EQ.0) ITPR = 2
         ELSE
            LE1 = LENGTH(KEY)
            LE2 = LENGTH(COMM)
            CALL PERRMS(PDESCF,LEP,WEAK,NPL,IUOUT)
            IF (WEAK) WRITE (IUOUT,280) KEY(1:LE1), COMM(1:LE2)
            IERR = 1
            RETURN
         END IF
C
         IF (NPIR+5.GT.MXPIR)
     &          CALL OVER('pir','PIR','IRANGE',MXPIR,'MXPIR')
         PIR(NPIR+1) = 1
         PIR(NPIR+2) = ITPL
         PIR(NPIR+3) = LEF
         PIR(NPIR+4) = ITPR
         PIR(NPIR+5) = IRI
         NPIR = NPIR+5
         IF (ITPL.EQ. 1 .OR.ITPL.EQ.2) THEN
            IF (ITPR.EQ.1 .OR. ITPR.EQ.2) THEN
               OK = LEF.LE.IRI
            ELSE
               OK = ITPR.EQ.3
            END IF
         ELSE IF (ITPL.EQ.-3) THEN
            IF (ITPR.EQ.1 .OR. ITPR.EQ.2) THEN
               OK = .TRUE.
            ELSE
               OK = ITPL.LT.ITPR
            END IF
         ELSE
            OK = .FALSE.
         END IF
         IF (.NOT.OK) THEN
            LE1 = LENGTH(KEY)
            LE2 = LENGTH(COMM)
            CALL PERRMS(PDESCF,LEP,WEAK,NPL,IUOUT)
            IF (WEAK) WRITE (IUOUT,281) KEY(1:LE1), COMM(1:LE2)
            IERR = 1
            RETURN
         END IF
      ELSE IF (INIT.EQ.EOL) THEN
         LE1 = LENGTH(KEY)
         LE2 = LENGTH(COMM)
         CALL PERRMS(PDESCF,LEP,WEAK,NPL,IUOUT)
         IF (WEAK) WRITE (IUOUT,282) KEY(1:LE1), COMM(1:LE2)
         IERR = 1
         RETURN
      ELSE
         LE1 = LENGTH(KEY)
         LE2 = LENGTH(COMM)
         CALL PERRMS(PDESCF,LEP,WEAK,NPL,IUOUT)
         IF (WEAK) WRITE (IUOUT,283) KEY(1:LE1), COMM(1:LE2)
         IERR = 1
         RETURN
      END IF
C
270   FORMAT ($(FC)'Integers only should be present in the ',
     &        'list of allowed values'/
     &        $(FC)'for the ',A,' keyword in the ',A,' panel.')
272   FORMAT ($(FC)'The list of integer values for the ',A,
     &        ' keyword in the ',A,' panel,'
     &        /$(FC)'is incorrect.')
273   FORMAT ($(FC)'The left limit for the ',A,
     &        ' keyword in the ',A,' panel,'
     &        /$(FC)'is incorrect.')
274   FORMAT ($(FC)'The left limit for the ',A,
     &        ' keyword in the ',A,' panel,'
     &        /$(FC)'should be an integer.')
276   FORMAT ($(FC)'A comma is expected after the left',
     &        ' limit for the ',A,' keyword'
     &        /$(FC)'in the ',A,' panel.')
277   FORMAT ($(FC)'The right limit for the ',A,
     &        ' keyword in the ',A,' panel,'
     &        /$(FC)'is incorrect.')
278   FORMAT ($(FC)'The right limit for the ',A,
     &        ' keyword in the ',A,' panel,'
     &        /$(FC)'should be an integer.')
280   FORMAT ($(FC)'The interval specification for the ',A,
     &        ' keyword in the ',A,' panel,'
     &        /$(FC)'is incorrect.')
281   FORMAT ($(FC)'The interval specification for the ',A,
     &        ' keyword in the ',A,' panel,'
     &        /$(FC)'is incorrect. Left limit > right limit.')
282   FORMAT ($(FC)'There is no description of allowed values ',
     &        'for the ',A,' keyword'
     &        /$(FC)'in the ',A,' panel.')
283   FORMAT ($(FC)'The specification of allowed values for the ',A,
     &        ' keyword'
     &        /$(FC)'in the ',A,' panel, is incorrect.')
      END
C  ---------------------------------------------------------------------
      LOGICAL FUNCTION ISCAP ( C )
C  ---------------------------------------------------------------------
C
C  Description:
C    Returns .TRUE. if the input character C is a capital letter;
C    .FALSE. otherwise.
C
C  Input arguments:
C    C          Character to check.
C
C  ---------------------------------------------------------------------
C  Arguments:
      CHARACTER C
C  ---------------------------------------------------------------------
C  Local variables:
      CHARACTER*26 LETT
      DATA LETT / 'ABCDEFGHIJKLMNOPQRSTUVWXYZ' /
C  ---------------------------------------------------------------------
C
      ISCAP = INDEX(LETT,C).NE.0
      END
C  ---------------------------------------------------------------------
      FUNCTION ISCOMP ( A, B )
C  ---------------------------------------------------------------------
C
C  Description:
C    Performs a case insensitive comparison of the two character strings
C    A and B. Returns 0 if they don't match, non-zero if they match.
C
C  Input arguments:
C    A          The first of the two strings to be compared.
C    B          The second of the two strings to be compared.
C
C  ---------------------------------------------------------------------
C  Arguments:
      CHARACTER*(*) A, B
C  ---------------------------------------------------------------------
C  Local variables:
      CHARACTER CA, CB
      LOGICAL BLANK
C  ---------------------------------------------------------------------
C
      LA = LEN(A)
      LB = LEN(B)
C
C  Compare all characters in A and B, one by one.
      LE = MIN(LA,LB)
      DO 10,I=1,LE
         CA = A(I:I)
         CB = B(I:I)
C  Convert to upper case before comparison.
         CALL UPPER(CA,1)
         CALL UPPER(CB,1)
         IF (CA.NE.CB) THEN
            ISCOMP = 0
            RETURN
         END IF
10    CONTINUE
C
C  Blanks at the end of a string are ignored.
      IF (LA.GT.LB) THEN
         BLANK = A(LB+1:) .EQ. ' '
      ELSE IF (LB.GT.LA) THEN
         BLANK = B(LA+1:) .EQ. ' '
      ELSE
         BLANK = .TRUE.
      END IF
C
      IF (BLANK) THEN
         ISCOMP = 1
      ELSE
         ISCOMP = 0
      END IF
C
      END
C  ---------------------------------------------------------------------
      FUNCTION LENGTH ( STRING )
C  ---------------------------------------------------------------------
C
C  Description:
C    Returns the position of the last non-blank non-tab character in
C    the input string, ie. its effective length.
C
C  Input arguments:
C    STRING    A character variable whose length we seek.
C
C  ---------------------------------------------------------------------
C  Arguments:
      CHARACTER STRING*(*)
C  ---------------------------------------------------------------------
C  Common blocks:
      CHARACTER TAB, ESC
      COMMON /SPECHA/ TAB, ESC
C  ---------------------------------------------------------------------
C  Local variables:
      CHARACTER C
C  ---------------------------------------------------------------------
C
      DO 10,I=LEN(STRING),1,-1
         C = STRING(I:I)
         IF (C.NE.' ' .AND. C.NE.TAB) THEN
            LENGTH = I
            RETURN
         END IF
10    CONTINUE
      LENGTH = 0
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE MCONF ( CFILE, ICONU, IFILEU, IERRU, NMCOM, MXPLUG,
     &                   COMMAN, EXCLU, NEXCLU, NURANG, IERROR, HEADER,
     &                   MODE, PROLOG, EPILOG, FILIN, FILOUT, PDESCF,
     &                   STRO, WEAK, IPRECO, OPRECO, BI, SM, IUINP,
     &                   IUOUT, MINU, MAXU, HASAPP, OUTAPP, NSCHA,
     &                   NSINT, NSREAL, HFILE, MFILE, EFILE, OFILE,
     &                   NDIG, NEOFA )
C  ---------------------------------------------------------------------
C
C  Purpose:
C    Reads the Merlin configuration file and changes the defaults
C    according to the configuration directives. Errors are written
C    to a scratch file with unit number IERRU. If any errors
C    actually occur (IERROR.NE.0) this file remains open when
C    subroutine MCONF returns.
C
C  Input arguments:
C    CFILE     The name of the configuration file to be read.
C
C    ICONU     Unit number to be assigned to the configuration file
C              in the corresponding OPEN statement.
C
C    IFILEU    Unit number to be assigned to the file specified in
C              a "FILE" configuration directive.
C
C    IERRU     Unit number to be assigned to the a scratch file
C              that receives all error messages.
C
C    NMCOM     Number of Merlin commands in the COMMAN array. The
C              actual length of this array is NMCOM+MXPLUG
C
C    MXPLUG    Maximum number of plugs that may be defined by the user.
C
C    COMMAN    Character array, containing all the Merlin commands and
C              plug-in names. This array may or may not be sorted.
C
C    EXCLU     Character array. Contains additional commands that must
C              be excluded form being valid command names in a PLUG
C              directive. Mainly used by the MCL compiler to exclude
C              IF, WHEN, LOOP etc. from being used as plug-in names.
C
C    NEXCLU    Number of entries in the EXCLU array.
C
C    NURANG    Minimum acceptable number of units that are assigned to
C              the unit allocator.
C
C  Output arguments:
C    IERROR    Indicates whether any errors occurred while processing
C              the configuration file. Possible values are:
C                IERROR = 0  -->  No error.
C                IERROR = 1  -->  Some error (not related to the
C                                 configuration directives that
C                                 affect the Merlin output file/unit).
C                IERROR = 2  -->  An error occurred while processing
C                                 configuration directives that affect
C                                 the Merlin output file/unit).
C              When IERROR = 2 Merlin must not open the output file.
C
C  Input / Output arguments:
C    The arguments listed below, on entry have the default values
C    assigned by Merlin/MCL. They are not referenced in any way, other
C    than by assignment statements that change their values when a
C    corresponding configuration directive is encountered.
C
C    Variable:      Configuration directive:
C    HEADER         HEADER
C    MODE           MODE
C    PROLOG         PLOLOG
C    EPILOG         EPILOG
C    FILIN          INPUT_FILE
C    FILOUT         OUTPUT_FILE
C    PDESCF         PDESC_FILE
C    STRO           PRINTOUT
C    WEAK           PRINTOUT
C    INPRECO        INPUT_PRECONN
C    OPRECO         OUTPUT_PRECONN
C    BI             BIGGER
C    SM             SMALLER
C    IUINP          INPUT_UNIT
C    IUOUT          OUTPUT_UNIT
C    MINU           UNIT_RANGE
C    MAXU           UNIT_RANGE
C    HASAPP         HAS_APPEND
C    OUTAPP         OUTPUT_FILE
C    NSCHA          SIZE_CHAR
C    NSINT          SIZE_INT
C    NSREAL         SIZE_REAL
C    HFILE          HELP_FILE
C    EFILE          MCL_ERROR_FILE
C    OFILE          MCL_OBJECT_FILE
C    NDIG           MACHINE_DIGITS
C    IEOF           ONEOF
C
C  ---------------------------------------------------------------------
      IMPLICIT $(TYPE) (A-H,O-Z)
      CHARACTER*(*) CFILE, MODE, PROLOG, EPILOG, FILIN, FILOUT, PDESCF
      CHARACTER*(*) COMMAN(*), EXCLU(NEXCLU), HFILE, MFILE, EFILE, OFILE
      LOGICAL HEADER, STRO, WEAK, IPRECO, OPRECO, HASAPP, OUTAPP
C
      PARAMETER ( LCLEN = 80 )
      CHARACTER*(LCLEN) LINE, ORIG, NAME, FL, ITEM
      INTEGER IS(LCLEN/2), IE(LCLEN/2)
      CHARACTER*20 STR
      LOGICAL EOF, EXI, OPE
C
      PARAMETER ( NDIR = 27 )
      CHARACTER*15 DIR(NDIR)
      INTEGER NITAL(NDIR)
C
      DATA (DIR(I),NITAL(I),I=1,10) /
     &     'HAS_APPEND     ', 2,
     &     'HEADER         ', 2,
     &     'SIZE_CHAR      ', 2,
     &     'MODE           ', 2,
     &     'FILE           ', 2,
     &     'PROLOG         ', 2,
     &     'EPILOG         ', 2,
     &     'INPUT_FILE     ', 2,
     &     'INPUT_UNIT     ', 2,
     &     'OUTPUT_FILE    ', -1 /
      DATA (DIR(I),NITAL(I),I=11,20) /
     &     'OUTPUT_UNIT    ', -1,
     &     'INPUT_PRECONN  ', 2,
     &     'BIGGER         ', 2,
     &     'OUTPUT_PRECONN ', -1,
     &     'UNIT_RANGE     ', 3,
     &     'SIZE_INT       ', 2,
     &     'SIZE_REAL      ', 2,
     &     'HELP_FILE      ', 2,
     &     'SMALLER        ', 2,
     &     'PRINTOUT       ', 2 /
      DATA (DIR(I),NITAL(I),I=21,NDIR) /
     &     'PLUG           ', 3 ,
     &     'PDESC_FILE     ', 2 ,
     &     'MACRO_FILE     ', 2 ,
     &     'MCL_ERROR_FILE ', 2 ,
     &     'MCL_OBJECT_FILE', 2 ,
     &     'MACHINE_DIGITS ', 2 ,
     &     'ONEOF          ', 2 /
C
      IERROR = 0
      IF (CFILE.EQ.' ') RETURN
C
      INQUIRE (FILE=CFILE,EXIST=EXI,OPENED=OPE,NUMBER=ICOI,IOSTAT=IERR)
      IF (IERR.NE.0) THEN
         OPEN (IERRU,STATUS='SCRATCH')
         REWIND IERRU
         CALL I2STR(IERR,STR,LES)
         CALL CONERR(IERRU,0,LINE,ORIG,IS,IE,0,
     &           'Cannot inquire the Merlin configuration file. '//
     &           'IOSTAT = '//STR(1:LES)//'.')
         IERROR = 1
         RETURN
      END IF
      IF (.NOT.EXI) RETURN
      IF (OPE) THEN
         OPEN (IERRU,STATUS='SCRATCH')
         REWIND IERRU
         CALL I2STR(ICOI,STR,LES)
         CALL CONERR(IERRU,0,LINE,ORIG,IS,IE,0,
     &           'The Merlin configuration file is already open.'//
     &           ' UNIT = '//STR(1:LES)//'.')
         IERROR = 1
         RETURN
      END IF
C
      OPEN (ICONU,FILE=CFILE,IOSTAT=IERR)
      IF (IERR.NE.0) THEN
         OPEN (IERRU,STATUS='SCRATCH')
         REWIND IERRU
         CALL I2STR(IERR,STR,LES)
         CALL CONERR(IERRU,0,LINE,ORIG,IS,IE,0,
     &           'The Merlin configuration file cannot be opened. '//
     &           'IOSTAT = '//STR(1:LES)//'.')
         IERROR = 1
         RETURN
      END IF
      REWIND ICONU
      OPEN (IERRU,STATUS='SCRATCH')
      REWIND IERRU
C
      EOF = .FALSE.
      NL = 0
100   CONTINUE
      IF (EOF) GOTO 200
      READ (ICONU,'(A)',END=200) ORIG
      NL = NL+1
      LL = LENGTH(LINE)
      CALL QLCR(ORIG,LL,LINE,IS,IE,NIT,IERR)
      IF (IERR.NE.0) THEN
         CALL CONERR(IERRU,NL,LINE,ORIG,IS,IE,0,
     &                  'Syntax error (Probably a missing quote).')
         IERROR = 1
         GOTO 100
      END IF
C
C  Is this an empty line with no items ?
      IF (NIT.EQ.0) GOTO 100
C
C  Is this a comment line ?
      IF (LINE(IS(1):IS(1)).EQ.'%') GOTO 100
C
C  Check if this is a correct configuration directive.
      ITEM = LINE(IS(1):IE(1))
      LEI = IE(1)-IS(1)+1
C
      DO 500,I=1,NDIR
         IF (ISCOMP(ITEM,DIR(I)).NE.0) THEN
            NI = NITAL(I)
            IF (NI.NE.-1 .AND. NIT.NE.NI) THEN
               CALL CONERR(IERRU,NL,LINE,ORIG,IS,IE,0,
     &            'The number of parameters of "'//ITEM(1:LEI)//
     &            '" is incorrect.')
               IERROR = 1
               GOTO 100
            ELSE
C  Check for arguments with zero length.
               DO 91,J=1,NIT
                  LEP = IE(J)-IS(J)+1
                  IF (LEP.EQ.0) THEN
                     CALL CONERR(IERRU,NL,LINE,ORIG,IS,IE,0,
     &            'One or more parameters of "'//ITEM(1:LEI)//
     &            '" have zero length.')
                     IERROR = 1
                     GOTO 100
                  END IF
91             CONTINUE
               GOTO (1,2,3,4,5,6,7,8,9,10,11,12,13,14,15,16,17,18,19,
     &                        20,21,22,23,24,25,26,27),I
            END IF
         END IF
500   CONTINUE
C
      CALL CONERR(IERRU,NL,LINE,ORIG,IS,IE,0,
     &            'There is no such configuration directive: "'//
     &            ITEM(1:LEI)//'".')
      IERROR = 1
      GOTO 100
C
C  ..................................................... HAS_APPEND
C
C  Syntax: HAS_APPEND yes | no
C
1     CONTINUE
      IF (ISCOMP(LINE(IS(2):IE(2)),'YES').NE.0) THEN
         HASAPP = .TRUE.
      ELSE IF (ISCOMP(LINE(IS(2):IE(2)),'NO').NE.0) THEN
         HASAPP = .FALSE.
      ELSE
         CALL CONERR(IERRU,NL,LINE,ORIG,IS,IE,2,
     &                  'Must be "yes" or "no".')
         IERROR = 1
      END IF
      GOTO 100
C
C  ..................................................... HEADER
C
C  Syntax: HEADER on | off
C
2     CONTINUE
      IF (ISCOMP(LINE(IS(2):IE(2)),'ON').NE.0) THEN
         HEADER = .TRUE.
      ELSE IF (ISCOMP(LINE(IS(2):IE(2)),'OFF').NE.0) THEN
         HEADER = .FALSE.
      ELSE
         CALL CONERR(IERRU,NL,LINE,ORIG,IS,IE,2,
     &                  'Must be "on" or "off".')
         IERROR = 1
      END IF
      GOTO 100
C
C  ..................................................... SIZE_CHAR
C
C  Syntax: SIZE_CHAR int
C
3     CONTINUE
      LE = IE(2)-IS(2)+1
      CALL TOINT(LINE(IS(2):IE(2)),LE,NSIZE,IERR)
      IF (IERR.NE.0) THEN
         CALL CONERR(IERRU,NL,LINE,ORIG,IS,IE,2,
     &                  'Must an integer > 0.')
         IERROR = 1
      ELSE IF (NSIZE.LE.0) THEN
         CALL CONERR(IERRU,NL,LINE,ORIG,IS,IE,2,
     &                  'Must be an integer > 0.')
         IERROR = 1
      ELSE
         NSCHA = NSIZE
      END IF
      GOTO 100
C
C  ..................................................... MODE
C
C  Syntax: MODE iaf | batch
C
4     CONTINUE
      LL = IE(2)-IS(2)+1
      IF (ISCOMP(LINE(IS(2):IE(2)),'IAF').NE.0) THEN
         MODE = LINE(IS(2):IE(2))
         CALL UPPER(MODE,LL)
      ELSE IF (ISCOMP(LINE(IS(2):IE(2)),'BATCH').NE.0) THEN
         MODE = LINE(IS(2):IE(2))
         CALL UPPER(MODE,LL)
      ELSE
         CALL CONERR(IERRU,NL,LINE,ORIG,IS,IE,2,
     &                  'Must be "iaf" or "batch".')
         IERROR = 1
      END IF
      GOTO 100
C
C  ..................................................... FILE
C
C  Syntax: FILE filename
C                  data lines
C          ...END
C
5     CONTINUE
      NAME = LINE(IS(2):IE(2))
      LEF = IE(2)-IS(2)+1
      INQUIRE (FILE=NAME(1:LEF),OPENED=OPE,NUMBER=IUFI,IOSTAT=IERR)
      IF (OPE) THEN
         CALL I2STR(IUFI,STR,LES)
         CALL CONERR(IERRU,NL,LINE,ORIG,IS,IE,0,'File "'//
     &                  NAME(1:LEF)//'" is already open. UNIT = '//
     &                  STR(1:LES)//'.')
         IERR = -1
      ELSE
         OPEN (IFILEU,FILE=NAME(1:LEF),IOSTAT=IERR)
         IF (IERR.NE.0) THEN
            CALL I2STR(IERR,STR,LES)
            CALL CONERR(IERRU,NL,LINE,ORIG,IS,IE,0,'File "'//
     &                        NAME(1:LEF)//'" could not be opened. '//
     &                        'IOSTAT = '//STR(1:LES)//'.')
         END IF
      END IF
      IF (IERR.EQ.0) REWIND IFILEU
185   READ (ICONU,'(A)',END=195) FL
      IF (ISCOMP(FL(1:6),'...END').NE.0) GOTO 196
      IF (IERR.EQ.0) THEN
         LE = LENGTH(FL)
         IF (LE.EQ.0) THEN
            WRITE (IFILEU,*)
         ELSE
            WRITE (IFILEU,'(A)') FL(1:LE)
         END IF
      END IF
      GOTO 185
C
195   EOF = .TRUE.
196   CONTINUE
      IF (IERR.EQ.0) THEN
         REWIND IFILEU
         CLOSE (IFILEU)
      ELSE
         IERROR = 1
      END IF
      GOTO 100
C
C  ..................................................... PROLOG
C
C  Syntax: PROLOG command
C
6     CONTINUE
      PROLOG = LINE(IS(2):IE(2))
      GOTO 100
C
C  ..................................................... EPILOG
C
C  Syntax: EPILOG command
C
7     CONTINUE
      EPILOG = LINE(IS(2):IE(2))
      GOTO 100
C
C  ..................................................... INPUT_FILE
C
C  Syntax: INPUT_FILE file
C
8     CONTINUE
      FILIN = LINE(IS(2):IE(2))
      GOTO 100
C
C  ..................................................... INPUT_UNIT
C
C  Syntax: INPUT_UNIT unit_number
C
9     CONTINUE
      LE = IE(2)-IS(2)+1
      CALL TOINT(LINE(IS(2):IE(2)),LE,NU,IERR)
      IF (IERR.NE.0) THEN
         CALL CONERR(IERRU,NL,LINE,ORIG,IS,IE,2,
     &                  'Must be a unit number >= 0.')
         IERROR = 1
      ELSE IF (NU.LT.0) THEN
         CALL CONERR(IERRU,NL,LINE,ORIG,IS,IE,2,
     &                  'Must be a unit number >= 0.')
         IERROR = 1
      ELSE
         IUINP = NU
      END IF
      GOTO 100
C
C  ..................................................... OUTPUT_FILE
C
C  Syntax: OUTPUT_FILE file [ append ]
C
10    CONTINUE
      IF (NIT.NE.2 .AND. NIT.NE.3) THEN
         CALL CONERR(IERRU,NL,LINE,ORIG,IS,IE,0,
     &            'The number of parameters of "'//ITEM(1:LEI)//
     &            '" is incorrect.')
         IERROR = 2
      ELSE
         FILOUT = LINE(IS(2):IE(2))
         IF (NIT.EQ.3) THEN
            FL = LINE(IS(3):IE(3))
            IF (ISCOMP(FL,'APPEND').NE.0) THEN
               OUTAPP = .TRUE.
            ELSE
               CALL CONERR(IERRU,NL,LINE,ORIG,IS,IE,3,
     &                              'Only "append" is allowed.')
               IERROR = 2
            END IF
         END IF
      END IF
      GOTO 100
C
C  ..................................................... OUTPUT_UNIT
C
C  Syntax: OUTPUT_UNIT unit_number
C
11    CONTINUE
      IF (NIT.NE.2) THEN
         CALL CONERR(IERRU,NL,LINE,ORIG,IS,IE,0,
     &            'The number of parameters of "'//ITEM(1:LEI)//
     &            '" is incorrect.')
         IERROR = 2
         GOTO 100
      END IF
C
      LE = IE(2)-IS(2)+1
      CALL TOINT(LINE(IS(2):IE(2)),LE,NU,IERR)
      IF (IERR.NE.0) THEN
         CALL CONERR(IERRU,NL,LINE,ORIG,IS,IE,2,
     &                  'Must be a unit number >= 0.')
         IERROR = 2
      ELSE IF (NU.LT.0) THEN
         CALL CONERR(IERRU,NL,LINE,ORIG,IS,IE,2,
     &                  'Must be a unit number >= 0.')
         IERROR = 2
      ELSE
         IUOUT = NU
      END IF
      GOTO 100
C
C  ..................................................... INPUT_PRECONN
C
C  Syntax: INPUT_PRECONN yes | no
C
12    CONTINUE
      IF (ISCOMP(LINE(IS(2):IE(2)),'YES').NE.0) THEN
         IPRECO = .TRUE.
      ELSE IF (ISCOMP(LINE(IS(2):IE(2)),'NO').NE.0) THEN
         IPRECO = .FALSE.
      ELSE
         CALL CONERR(IERRU,NL,LINE,ORIG,IS,IE,2,
     &                  'Must be "yes" or "no".')
         IERROR = 1
      END IF
      GOTO 100
C
C  ..................................................... BIGGER
C
C  Syntax: BIGGER real
C
13    CONTINUE
      LE = IE(2)-IS(2)+1
      CALL TOREAL(LINE(IS(2):IE(2)),LE,X,IERR)
      IF (IERR.NE.0) THEN
         CALL CONERR(IERRU,NL,LINE,ORIG,IS,IE,2,
     &                  'Must be a positive real.')
         IERROR = 1
      ELSE IF (X.LE.0.) THEN
         CALL CONERR(IERRU,NL,LINE,ORIG,IS,IE,2,
     &                  'Must be a positive real.')
         IERROR = 1
      ELSE
         BI = X
      END IF
      GOTO 100
C
C  ..................................................... OUTPUT_PRECONN
C
C  Syntax: OUTPUT_PRECONN yes | no
C
14    CONTINUE
      IF (NIT.NE.2) THEN
         CALL CONERR(IERRU,NL,LINE,ORIG,IS,IE,0,
     &            'The number of parameters of "'//ITEM(1:LEI)//
     &            '" is incorrect.')
         IERROR = 2
         GOTO 100
      END IF
C
      IF (ISCOMP(LINE(IS(2):IE(2)),'YES').NE.0) THEN
         OPRECO = .TRUE.
      ELSE IF (ISCOMP(LINE(IS(2):IE(2)),'NO').NE.0) THEN
         OPRECO = .FALSE.
      ELSE
         CALL CONERR(IERRU,NL,LINE,ORIG,IS,IE,2,
     &                  'Must be "yes" or "no".')
         IERROR = 2
      END IF
      GOTO 100
C
C  ..................................................... UNIT_RANGE
C
C  Syntax: UNIT_RANGE start end
C
15    CONTINUE
      LE = IE(2)-IS(2)+1
      CALL TOINT(LINE(IS(2):IE(2)),LE,NU1,IERR)
      IF (IERR.NE.0) THEN
         CALL CONERR(IERRU,NL,LINE,ORIG,IS,IE,2,
     &                  'Must be a unit number: 0 <= n <= 999.')
         IERROR = 1
         GOTO 100
      ELSE IF (NU1.LT.0 .OR. NU1.GT.999) THEN
         CALL CONERR(IERRU,NL,LINE,ORIG,IS,IE,2,
     &                  'Must be a unit number: 0 <= n <= 999.')
         IERROR = 1
         GOTO 100
      END IF
C
      LE = IE(3)-IS(3)+1
      CALL TOINT(LINE(IS(3):IE(3)),LE,NU2,IERR)
      IF (IERR.NE.0) THEN
         CALL CONERR(IERRU,NL,LINE,ORIG,IS,IE,3,
     &                  'Must be a unit number: 0 <= n <= 999.')
         IERROR = 1
         GOTO 100
      ELSE IF (NU2.LT.0 .OR. NU2.GT.999) THEN
         CALL CONERR(IERRU,NL,LINE,ORIG,IS,IE,3,
     &                  'Must be a unit number: 0 <= n <= 999.')
         IERROR = 1
         GOTO 100
      ELSE IF (NU2-NU1+1.LT.NURANG) THEN
         CALL I2STR(NURANG,STR,LES)
         CALL CONERR(IERRU,NL,LINE,ORIG,IS,IE,0,
     &                  'You must provide a range of at least '//
     &                  STR(1:LES)//' unit numbers.')
         IERROR = 1
         GOTO 100
      ELSE
         MINU = NU1
         MAXU = NU2
      END IF
      GOTO 100
C
C  ..................................................... SIZE_INT
C
C  Syntax: SIZE_INT int
C
16    CONTINUE
      LE = IE(2)-IS(2)+1
      CALL TOINT(LINE(IS(2):IE(2)),LE,NSIZE,IERR)
      IF (IERR.NE.0) THEN
         CALL CONERR(IERRU,NL,LINE,ORIG,IS,IE,2,
     &                  'Must an integer > 0.')
         IERROR = 1
      ELSE IF (NSIZE.LE.0) THEN
         CALL CONERR(IERRU,NL,LINE,ORIG,IS,IE,2,
     &                  'Must be an integer > 0.')
         IERROR = 1
      ELSE
         NSINT = NSIZE
      END IF
      GOTO 100
C
C  ..................................................... SIZE_REAL
C
C  Syntax: SIZE_REAL int
C
17    CONTINUE
      LE = IE(2)-IS(2)+1
      CALL TOINT(LINE(IS(2):IE(2)),LE,NSIZE,IERR)
      IF (IERR.NE.0) THEN
         CALL CONERR(IERRU,NL,LINE,ORIG,IS,IE,2,
     &                  'Must an integer > 0.')
         IERROR = 1
      ELSE IF (NSIZE.LE.0) THEN
         CALL CONERR(IERRU,NL,LINE,ORIG,IS,IE,2,
     &                  'Must be an integer > 0.')
         IERROR = 1
      ELSE
         NSREAL = NSIZE
      END IF
      GOTO 100
C
C  ..................................................... HELP_FILE
C
C  Syntax: HELP_FILE file
C
18    CONTINUE
      HFILE = LINE(IS(2):IE(2))
      GOTO 100
C
C  ..................................................... SMALLER
C
C  Syntax: SMALLER real
C
19    CONTINUE
      LE = IE(2)-IS(2)+1
      CALL TOREAL(LINE(IS(2):IE(2)),LE,X,IERR)
      IF (IERR.NE.0) THEN
         CALL CONERR(IERRU,NL,LINE,ORIG,IS,IE,2,
     &                  'Must be a real number: 0 < x < 1.')
         IERROR = 1
      ELSE IF (X.LE.0. .OR. X.GT.1.) THEN
         CALL CONERR(IERRU,NL,LINE,ORIG,IS,IE,2,
     &                  'Must be a real number: 0 < x < 1.')
         IERROR = 1
      ELSE
         SM = X
      END IF
      GOTO 100
C
C  ..................................................... PRINTOUT
C
C  Syntax: PRINTOUT full | half | no
C
20    CONTINUE
      IF (ISCOMP(LINE(IS(2):IE(2)),'FULL').NE.0) THEN
         WEAK = .TRUE.
         STRO = .TRUE.
      ELSE IF (ISCOMP(LINE(IS(2):IE(2)),'HALF').NE.0) THEN
         STRO = .FALSE.
         WEAK = .TRUE.
      ELSE IF (ISCOMP(LINE(IS(2):IE(2)),'NO').NE.0) THEN
         WEAK = .FALSE.
         STRO = .FALSE.
      ELSE
         CALL CONERR(IERRU,NL,LINE,ORIG,IS,IE,2,
     &                  'Must be "no", "half" or "full".')
         IERROR = 1
      END IF
      GOTO 100
C
C  ..................................................... PLUG
C
C  Syntax: PLUG number name
C
21    CONTINUE
      LE = IE(2)-IS(2)+1
      CALL TOINT(LINE(IS(2):IE(2)),LE,NPLU,IERR)
      IF (IERR.NE.0) THEN
         CALL CONERR(IERRU,NL,LINE,ORIG,IS,IE,2,
     &                  'Must be an integer: n > 0.')
         IERROR = 1
         GOTO 100
      ELSE IF (NPLU.LT.1) THEN
         CALL CONERR(IERRU,NL,LINE,ORIG,IS,IE,2,
     &                  'Must be an integer: n > 0.')
         IERROR = 1
         GOTO 100
      ELSE IF (NPLU.GT.MXPLUG) THEN
         CALL I2STR(MXPLUG,STR,LES)
         CALL CONERR(IERRU,NL,LINE,ORIG,IS,IE,2,'There are only '//
     &                  STR(1:LES)//'plugs in the command list.')
         IERROR = 1
         GOTO 100
      END IF
C
C  Search the command array for the position of the plug.
      CALL I2STR(NPLU,STR,LES)
      NAME = 'PLUG'//STR(1:LES)
      NCOM = NMCOM+MXPLUG
      DO 600,ICM=1,NCOM
         IF (ISCOMP(COMMAN(ICM),NAME).NE.0) GOTO 603
600   CONTINUE
      CALL MABORT('Plug not found in command list.','MCONF')
C
603   CONTINUE
      LE = IE(3)-IS(3)+1
      LECO = LEN(COMMAN(1))
      IF (LE.GT.LECO) THEN
         CALL I2STR(LECO,STR,LES)
         CALL CONERR(IERRU,NL,LINE,ORIG,IS,IE,3,
     &            'The command name is too long. Maximum is '//
     &            STR(1:LES)//' characters.')
         IERROR = 1
         GOTO 100
      END IF
      NAME = LINE(IS(3):IE(3))
      LENA = IE(3)-IS(3)+1
      IF (ISCOMP(NAME(1:4),'PLUG').NE.0) THEN
         CALL I2STR(LECO,STR,LES)
         CALL CONERR(IERRU,NL,LINE,ORIG,IS,IE,3,
     &            'The command cannot start with "plug".')
         IERROR = 1
         GOTO 100
      END IF
      DO 604,I=1,NCOM
         IF (ISCOMP(NAME,COMMAN(I)).NE.0) THEN
            CALL CONERR(IERRU,NL,LINE,ORIG,IS,IE,3,
     &            'This command exists already.')
            IERROR = 1
            GOTO 100
         END IF
604   CONTINUE
      DO 605,I=1,NEXCLU
         IF (ISCOMP(NAME,EXCLU(I)).NE.0) THEN
            CALL CONERR(IERRU,NL,LINE,ORIG,IS,IE,3,
     &            'This command exists already.')
            IERROR = 1
            GOTO 100
         END IF
605   CONTINUE
      CALL UPPER(NAME,LENA)
      COMMAN(ICM) = NAME
      GOTO 100
C
C  ..................................................... PDESC_FILE
C
C  Syntax: PDESC_FILE file
C
22    CONTINUE
      PDESCF = LINE(IS(2):IE(2))
      GOTO 100
C
C  ..................................................... MACRO_FILE
C
C  Syntax: MACRO_FILE file
C
23    CONTINUE
      MFILE = LINE(IS(2):IE(2))
      GOTO 100
C
C  ..................................................... MCL_ERROR_FILE
C
C  Syntax: MCL_ERROR_FILE file
C
24    CONTINUE
      EFILE = LINE(IS(2):IE(2))
      GOTO 100
C
C  ..................................................... MCL_OBJECT_FILE
C
C  Syntax: MCL_OBJECT_FILE file
C
25    CONTINUE
      OFILE = LINE(IS(2):IE(2))
      GOTO 100
C
C  ..................................................... ONEOF
C
C  Syntax: MACHINE_DIGITS int
C
26    CONTINUE
      LE = IE(2)-IS(2)+1
      CALL TOINT(LINE(IS(2):IE(2)),LE,NN,IERR)
      IF (IERR.NE.0) THEN
         CALL CONERR(IERRU,NL,LINE,ORIG,IS,IE,2,
     &                  'Must an integer > 0.')
         IERROR = 1
      ELSE IF (NN.LT.0) THEN
         CALL CONERR(IERRU,NL,LINE,ORIG,IS,IE,2,
     &                  'Must an integer > 0.')
         IERROR = 1
      ELSE
         NDIG = NN
      END IF
      GOTO 100
C
C  ..................................................... MACHINE_DIGITS
C
C  Syntax: ONEOF rewind | ignore | return
C
27    CONTINUE
      IF (ISCOMP(LINE(IS(2):IE(2)),'REWIND').NE.0) THEN
         NEOFA = 1
      ELSE IF (ISCOMP(LINE(IS(2):IE(2)),'IGNORE').NE.0) THEN
         NEOFA = 2
      ELSE IF (ISCOMP(LINE(IS(2):IE(2)),'RETURN').NE.0) THEN
         NEOFA = 3
      ELSE
         CALL CONERR(IERRU,NL,LINE,ORIG,IS,IE,2,
     &                  'Must be "rewind", "ignore" or "return".')
         IERROR = 1
      END IF
      GOTO 100
C
200   CONTINUE
C
C  An End-of-File has been encountered.
      CLOSE (ICONU)
      IF (IERROR.EQ.0) CLOSE (IERRU,STATUS='DELETE')
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE CONERR ( IERRU, NL, LINE, ORIG, IS, IE, IPARNO, MSG )
C  ---------------------------------------------------------------------
C
C  Description:
C    Called by subroutine MCONF. It writes messages to the error file,
C    for incorrect configuration directives.
C
C  Input arguments:
C    IERRU      Error messages are written to this unit.
C    NL         The line number of the configuration file, that is
C               currently being processed.
C    LINE       Items of the current line (see subroutine QLCR).
C    ORIG       The line of the configuration file that is being
C               processed.
C    IS         IS(I) is the starting character position of item I
C               in LINE.
C    IE         IE(I) is the ending character position of item I
C               in LINE.
C    IPARNO     The item # that is in error.
C    MSG        Short message to be displayed.
C
C  ---------------------------------------------------------------------
C  Arguments:
      INTEGER IS(*), IE(*)
      CHARACTER*(*) LINE, ORIG, MSG
C  ---------------------------------------------------------------------
C  Local variables:
      CHARACTER*15 STR, BLANK
      DATA BLANK / '               ' /
C  ---------------------------------------------------------------------
C
      IF (NL.NE.0) THEN
         CALL I2STR(NL,STR,LES)
         WRITE (IERRU,10) STR(1:LES), ORIG(1:LENGTH(ORIG))
C
         LEB = LES+7
         IF (IPARNO.NE.0) WRITE (IERRU,20) BLANK(1:LEB),
     &              LINE(IS(IPARNO):IE(IPARNO)), LINE(IS(1):IE(1))
         LEM = LENGTH(MSG)
         IF (LEM.NE.0) WRITE (IERRU,30) BLANK(1:LEB), MSG(1:LEM)
      ELSE
         LEM = LENGTH(MSG)
         IF (LEM.NE.0) THEN
            WRITE (IERRU,40) MSG(1:LEM)
         ELSE
            CALL MABORT('Some error in config. No msg available.',
     &                        'CONERR')
         END IF
      END IF
C
10    FORMAT ($(FC)/'Line ',A,': ',A)
20    FORMAT ($(FC)A,'Parameter "',A,'" is incorrect for ',
     &        'the directive "',A,'".')
30    FORMAT ($(FC)A,A)
40    FORMAT ($(FC)A)
      END
C  ---------------------------------------------------------------------
      SUBROUTINE LEX ( LINE, LENG, TOKEN, NTOK, NUM, ALPHA )
C  ---------------------------------------------------------------------
C
C  Description:
C    Subroutine LEX is the lexical analyzer of the Merlin/MCL package.
C    It scans the input line LINE and determines its structure. LINE is
C    analyzed in a sequence of lexical items, or tokens. Tokens are
C    stored in succession in array TOKEN, along with some additional
C    information:
C      For an identifier (TOKEN(I)=IDENT)
C        The identifier is converted to upper case and stored in ALPHA.
C        Two successive entries in array NUM point to the starting and
C        ending character positions of the identifier in ALPHA.
C      For an unsigned number (TOKEN(I)=UNSI)
C        The number is stored in array NUM.
C      For a string (TOKEN(I)=STRING)
C        The string is stored in ALPHA. Two successive entries in array
C        NUM point to the starting and ending character position of the
C        string in ALPHA. Single quotes can be embedded in a string
C        using the escape sequence \'
C
C    The complete list of all possible tokens is kept in common
C    block MCLSCD. The last entry in array TOKEN is always the EOL
C    token. Ie: TOKEN(NTOK) = EOL
C    Underscores in an identifier are ignored. (HELLO_THERE is stored
C    as HELLOTHERE.)
C
C  Input arguments:
C    LINE       The line to be analyzed.
C    LENG       Effective length of LINE.
C
C  Output arguments:
C    TOKEN      Integer array. Contains the numeric codes of the tokens
C    NTOK       Number of tokens in TOKEN.
C    NUM        Real array. It contains all unsigned numbers, starting
C               and ending character positions for identifiers and
C               strings.
C    ALPHA      Stores all identifiers and strings. It should be at
C               least as long as the input line.
C
C  ---------------------------------------------------------------------
      IMPLICIT INTEGER (A-Z)
C  Arguments:
      INTEGER TOKEN(*)
      CHARACTER*(*) LINE, ALPHA
      $(TYPE) NUM(*)
C  ---------------------------------------------------------------------
C  Common blocks:
C  Codes used by the lexical analysis routine LEX.
      INTEGER LEFTP, RIGHTP, PERIOD, COMMA, GT, LT, NE, EQ, GE,
     &        LE, PLUS, MINUS, STAR, SLASH, BISTAR, COLON, EOL,
     &        IDENT, UNSI, NOTREC, STRING, QUOTE, LBRA, RBRA,
     &        SEMI, PERCE, AND, OR, XOR, NOT, EQUAL, LBRACE,
     &        RBRACE, QMARK
      COMMON /MCLSCD/ LEFTP, RIGHTP, PERIOD, COMMA, GT, LT, NE, EQ, GE,
     &                LE, PLUS, MINUS, STAR, SLASH, BISTAR, COLON, EOL,
     &                IDENT, UNSI, NOTREC, STRING, QUOTE, LBRA, RBRA,
     &                SEMI, PERCE, AND, OR, XOR, NOT, EQUAL, LBRACE,
     &                RBRACE, QMARK
      CHARACTER TAB, ESC
      COMMON /SPECHA/ TAB, ESC
C  ---------------------------------------------------------------------
C  Local variables:
      CHARACTER UCHA, UPPER1, CHA
      LOGICAL DOTFND, EOLIN, ALPHAB, NMERIC, SPCIAL, UNDER, LIT
      $(TYPE) NUMBER, RE, TEN
      EOLIN() = (FINGER.GE.LENG)
      ALPHAB(CHA) = ('A'.LE.CHA) .AND. (CHA.LE.'Z')
      NMERIC() = ('0'.LE.CHA) .AND. (CHA.LE.'9')
      SPCIAL() = (CHA.EQ.'=') .OR. (CHA.EQ.'(') .OR. (CHA.EQ.')').OR.
     &           (CHA.EQ.'.') .OR. (CHA.EQ.',') .OR. (CHA.EQ.';').OR.
     &           (CHA.EQ.'+') .OR. (CHA.EQ.'-') .OR. (CHA.EQ.'/').OR.
     &           (CHA.EQ.'[') .OR. (CHA.EQ.']') .OR. (CHA.EQ.':').OR.
     &           (CHA.EQ.'#') .OR. (CHA.EQ.'{') .OR. (CHA.EQ.'}').OR.
     &           (CHA.EQ.'?')
      UNDER() = (CHA.EQ.'_')
C  ---------------------------------------------------------------------
C
      IC0 = ICHAR('0')
      TEN = 10.
      FINGER = 0
      LLEN = LEN(LINE)
C
C  NT is the number of entries in TOKEN.
      NT = 0
C  NN is the number of entries in NUM.
      NN = 0
C  NAL is the effective length of ALPHA.
      NAL = 0
C
100   CONTINUE
      IF (EOLIN()) THEN
         GOTO 9000
      ELSE
         FINGER = FINGER+1
      END IF
C
C  CHA is always the character under processing.
      CHA = LINE(FINGER:FINGER)
      UCHA = UPPER1(CHA)
C
C  We jump to 200, when there is further processing to
C  be done from other parts of LEX during scanning.
200   CONTINUE
      IF (ALPHAB(UCHA)) THEN
C
C  Alphabetic character located.
C  The logical operators (and, or, xor, not) are being identified
C  here too.
         NAL = NAL+1
         NALST = NAL
         ALPHA(NALST:NALST) = UCHA
300      CONTINUE
         IF (.NOT.EOLIN()) THEN
            FINGER = FINGER+1
            CHA = LINE(FINGER:FINGER)
            UCHA = UPPER1(CHA)
            IF ((ALPHAB(UCHA) .OR. NMERIC() .OR. UNDER())) THEN
C  Ignore underscores when they appear into identifiers.
               IF (UNDER()) GOTO 300
               NAL = NAL+1
               ALPHA(NAL:NAL) = UCHA
               GOTO 300
            ELSE
C  Obtain operator code (if operator).
               IF (ALPHA(NALST:NAL).EQ.'AND') THEN
                  CALL PSHSTR(AND,TOKEN,NT,LLEN)
                  NAL = NALST-1
               ELSE IF (ALPHA(NALST:NAL).EQ.'OR') THEN
                  CALL PSHSTR(OR,TOKEN,NT,LLEN)
                  NAL = NALST-1
               ELSE IF (ALPHA(NALST:NAL).EQ.'XOR') THEN
                  CALL PSHSTR(XOR,TOKEN,NT,LLEN)
                  NAL = NALST-1
               ELSE IF (ALPHA(NALST:NAL).EQ.'NOT') THEN
                  CALL PSHSTR(NOT,TOKEN,NT,LLEN)
                  NAL = NALST-1
               ELSE
                  CALL PSHSTR(IDENT,TOKEN,NT,LLEN)
                  RE = NALST
                  CALL PSHNUM(RE,NUM,NN,LLEN)
                  RE = NAL
                  CALL PSHNUM(RE,NUM,NN,LLEN)
               END IF
               GOTO 200
            END IF
         ELSE
            IF (ALPHA(NALST:NAL).EQ.'AND') THEN
               CALL PSHSTR(AND,TOKEN,NT,LLEN)
               NAL = NALST-1
            ELSE IF (ALPHA(NALST:NAL).EQ.'OR') THEN
               CALL PSHSTR(OR,TOKEN,NT,LLEN)
               NAL = NALST-1
            ELSE IF (ALPHA(NALST:NAL).EQ.'XOR') THEN
               CALL PSHSTR(XOR,TOKEN,NT,LLEN)
               NAL = NALST-1
            ELSE IF (ALPHA(NALST:NAL).EQ.'NOT') THEN
               CALL PSHSTR(NOT,TOKEN,NT,LLEN)
               NAL = NALST-1
            ELSE
               CALL PSHSTR(IDENT,TOKEN,NT,LLEN)
               RE = NALST
               CALL PSHNUM(RE,NUM,NN,LLEN)
               RE = NAL
               CALL PSHNUM(RE,NUM,NN,LLEN)
            END IF
         END IF
C
      ELSE IF (NMERIC()) THEN
C
C  Numeric character located. Proceed to transform
C  the sequence of characters to real number.
         DOTFND = .FALSE.
         NUMBER = ICHAR(CHA)-IC0
         EXPO = 0
         POWER = -1
500      CONTINUE
         IF (.NOT.EOLIN()) THEN
            FINGER = FINGER+1
            CHA = LINE(FINGER:FINGER)
            UCHA = UPPER1(CHA)
            IF (NMERIC()) THEN
               IF (POWER.EQ.-1) THEN
                  RE = ICHAR(CHA)-IC0
                  NUMBER = NUMBER*TEN+RE
                  GOTO 500
               ELSE
                  POWER = POWER+1
                  RE = ICHAR(CHA)-IC0
                  NUMBER = NUMBER+RE/(TEN**POWER)
                  GOTO 500
               END IF
            ELSE IF (CHA.EQ.'.') THEN
               POWER = 0
               IF (DOTFND) THEN
                  CALL PSHSTR(UNSI,TOKEN,NT,LLEN)
                  CALL PSHNUM(NUMBER,NUM,NN,LLEN)
                  CALL PSHSTR(PERIOD,TOKEN,NT,LLEN)
                  GOTO 100
               ELSE
                  DOTFND = .TRUE.
                  GOTO 500
               END IF
            ELSE IF (UCHA.EQ.'E') THEN
               IF (.NOT.EOLIN()) THEN
                  FINGER = FINGER+1
                  CHA = LINE(FINGER:FINGER)
                  UCHA = UPPER1(CHA)
                  IF (NMERIC()) THEN
                     EXPO = EXPO*10+(ICHAR(CHA)-IC0)
600                  CONTINUE
                     IF (.NOT.EOLIN()) THEN
                        FINGER = FINGER+1
                        CHA = LINE(FINGER:FINGER)
                        UCHA = UPPER1(CHA)
                        IF (NMERIC()) THEN
                           EXPO = EXPO*10+(ICHAR(CHA)-IC0)
                           GOTO 600
                        ELSE
                           NUMBER = NUMBER*(TEN**EXPO)
                           CALL PSHSTR(UNSI,TOKEN,NT,LLEN)
                           CALL PSHNUM(NUMBER,NUM,NN,LLEN)
                           GOTO 200
                        END IF
                     ELSE
                        NUMBER = NUMBER*(TEN**EXPO)
                        CALL PSHSTR(UNSI,TOKEN,NT,LLEN)
                        CALL PSHNUM(NUMBER,NUM,NN,LLEN)
                     END IF
                  ELSE IF (CHA.EQ.'+') THEN
                     IF (.NOT.EOLIN()) THEN
                        FINGER = FINGER+1
                        CHA = LINE(FINGER:FINGER)
                        UCHA = UPPER1(CHA)
                        IF (NMERIC()) THEN
                           EXPO = EXPO*10+(ICHAR(CHA)-IC0)
800                        CONTINUE
                           IF (.NOT.EOLIN()) THEN
                              FINGER = FINGER+1
                              CHA = LINE(FINGER:FINGER)
                              UCHA = UPPER1(CHA)
                              IF (NMERIC()) THEN
                                 EXPO = EXPO*10+
     &                                                  (ICHAR(CHA)-IC0)
                                 GOTO 800
                              ELSE
                                 NUMBER = NUMBER*(TEN**EXPO)
                                 CALL PSHSTR(UNSI,
     &                                      TOKEN,NT,LLEN)
                                 CALL PSHNUM(NUMBER,
     &                                      NUM,NN,LLEN)
                                 GOTO 200
                              END IF
                           ELSE
                              NUMBER = NUMBER*(TEN**EXPO)
                              CALL PSHSTR(UNSI,TOKEN,NT,LLEN)
                              CALL PSHNUM(NUMBER,NUM,NN,LLEN)
                           END IF
                        ELSE
                           CALL PSHSTR(UNSI,TOKEN,NT,LLEN)
                           CALL PSHNUM(NUMBER,NUM,NN,LLEN)
                           CALL PSHSTR(IDENT,TOKEN,NT,LLEN)
                           CALL PSHALP('E',ALPHA,NAL,NUM,NN,LLEN)
                           CALL PSHSTR(PLUS,TOKEN,NT,LLEN)
                           GOTO 200
                        END IF
                     ELSE
                        CALL PSHSTR(UNSI,TOKEN,NT,LLEN)
                        CALL PSHNUM(NUMBER,NUM,NN,LLEN)
                        CALL PSHSTR(IDENT,TOKEN,NT,LLEN)
                        CALL PSHALP('E',ALPHA,NAL,NUM,NN,LLEN)
                        CALL PSHSTR(PLUS,TOKEN,NT,LLEN)
                     END IF
                  ELSE IF (CHA.EQ.'-') THEN
                     IF (.NOT.EOLIN()) THEN
                        FINGER = FINGER+1
                        CHA = LINE(FINGER:FINGER)
                        UCHA = UPPER1(CHA)
                        IF (NMERIC()) THEN
                           EXPO = EXPO*10+(ICHAR(CHA)-IC0)
820                        CONTINUE
                           IF (.NOT.EOLIN()) THEN
                              FINGER = FINGER+1
                              CHA = LINE(FINGER:FINGER)
                              UCHA = UPPER1(CHA)
                              IF (NMERIC()) THEN
                                 EXPO = EXPO*10+
     &                                               (ICHAR(CHA)-IC0)
                                 GOTO 820
                              ELSE
                                 NUMBER = NUMBER*(TEN**(-EXPO))
                                 CALL PSHSTR(UNSI,
     &                                                TOKEN,NT,LLEN)
                                 CALL PSHNUM(NUMBER,
     &                                                NUM,NN,LLEN)
                                 GOTO 200
                              END IF
                           ELSE
                              NUMBER = NUMBER*(TEN**(-EXPO))
                              CALL PSHSTR(UNSI,TOKEN,NT,LLEN)
                              CALL PSHNUM(NUMBER,NUM,NN,LLEN)
                           END IF
                        ELSE
                           CALL PSHSTR(UNSI,TOKEN,NT,LLEN)
                           CALL PSHNUM(NUMBER,NUM,NN,LLEN)
                           CALL PSHSTR(IDENT,TOKEN,NT,LLEN)
                           CALL PSHALP('E',ALPHA,NAL,NUM,NN,LLEN)
                           CALL PSHSTR(MINUS,TOKEN,NT,LLEN)
                           GOTO 200
                        END IF
                     ELSE
                        CALL PSHSTR(UNSI,TOKEN,NT,LLEN)
                        CALL PSHNUM(NUMBER,NUM,NN,LLEN)
                        CALL PSHSTR(IDENT,TOKEN,NT,LLEN)
                        CALL PSHALP('E',ALPHA,NAL,NUM,NN,LLEN)
                        CALL PSHSTR(MINUS,TOKEN,NT,LLEN)
                     END IF
                  ELSE
                     FINGER = FINGER-1
C  Minus because we unget CHA.
                     CALL PSHSTR(UNSI,TOKEN,NT,LLEN)
                     CALL PSHNUM(NUMBER,NUM,NN,LLEN)
                     GOTO 200
                  END IF
               ELSE
                  CALL PSHSTR(UNSI,TOKEN,NT,LLEN)
                  CALL PSHNUM(NUMBER,NUM,NN,LLEN)
                  CALL PSHSTR(IDENT,TOKEN,NT,LLEN)
                  CALL PSHALP(CHA,ALPHA,NAL,NUM,NN,LLEN)
               END IF
            ELSE
               CALL PSHSTR(UNSI,TOKEN,NT,LLEN)
               CALL PSHNUM(NUMBER,NUM,NN,LLEN)
               GOTO 200
            END IF
         ELSE
            CALL PSHSTR(UNSI,TOKEN,NT,LLEN)
            CALL PSHNUM(NUMBER,NUM,NN,LLEN)
         END IF
C
      ELSE IF (SPCIAL()) THEN
C
C  Special situation.
         IF (CHA.EQ.'=') THEN
            IF (.NOT. EOLIN()) THEN
               FINGER = FINGER+1
               CHA = LINE (FINGER:FINGER)
               UCHA = UPPER1(CHA)
               IF (UCHA.EQ.'=') THEN
C  Equality operator detected.
                  CALL PSHSTR(EQUAL,TOKEN,NT,LLEN)
                  GOTO 100
               ELSE
                  CALL PSHSTR(EQ,TOKEN,NT,LLEN)
                  GOTO 200
               END IF
            ELSE
               CALL PSHSTR(EQ,TOKEN,NT,LLEN)
            END IF
         ELSE
C  Special character detected.
            SPCODE = LXTRAN(CHA)
            CALL PSHSTR(SPCODE,TOKEN,NT,LLEN)
            GOTO 100
         END IF
C
      ELSE IF (CHA.EQ.' ' .OR. CHA.EQ.TAB) THEN
C
C  Ignore spaces and tabs.
         GOTO 100
      ELSE IF (CHA.EQ.'*') THEN
C
C  Star located.
         IF (.NOT.EOLIN()) THEN
            FINGER = FINGER+1
            CHA = LINE(FINGER:FINGER)
            UCHA = UPPER1(CHA)
            IF (UCHA.EQ.'*') THEN
               CALL PSHSTR(BISTAR,TOKEN,NT,LLEN)
               GOTO 100
            ELSE
               CALL PSHSTR(STAR,TOKEN,NT,LLEN)
               GOTO 200
            END IF
         ELSE
            CALL PSHSTR(STAR,TOKEN,NT,LLEN)
         END IF
C
      ELSE IF (CHA.EQ.'>') THEN
C
C  GT located.
         IF (.NOT.EOLIN()) THEN
            FINGER = FINGER+1
            CHA = LINE(FINGER:FINGER)
            UCHA = UPPER1(CHA)
            IF (UCHA.EQ.'=') THEN
               CALL PSHSTR(GE,TOKEN,NT,LLEN)
               GOTO 100
            ELSE
               CALL PSHSTR(GT,TOKEN,NT,LLEN)
               GOTO 200
            END IF
         ELSE
            CALL PSHSTR(GT,TOKEN,NT,LLEN)
         END IF
C
      ELSE IF (CHA.EQ.'<') THEN
C
C  LT located.
         IF (.NOT.EOLIN()) THEN
            FINGER = FINGER+1
            CHA = LINE(FINGER:FINGER)
            UCHA = UPPER1(CHA)
            IF (UCHA.EQ.'=') THEN
               CALL PSHSTR(LE,TOKEN,NT,LLEN)
               GOTO 100
            ELSE
               CALL PSHSTR(LT,TOKEN,NT,LLEN)
               GOTO 200
            END IF
         ELSE
            CALL PSHSTR(LT,TOKEN,NT,LLEN)
         END IF
C
      ELSE IF (CHA.EQ.'''') THEN
C
C  Quote located. Process string.
         STFING = 0
         CALL PSHSTR(QUOTE,TOKEN,NT,LLEN)
         NALST = NAL+1
         LIT = .FALSE.
987      CONTINUE
         IF (.NOT.EOLIN()) THEN
            FINGER = FINGER+1
            CHA = LINE(FINGER:FINGER)
988         CONTINUE
            IF (LIT) THEN
               LIT = .FALSE.
               IF (CHA.EQ.'''') THEN
                  NAL = NAL+1
                  ALPHA(NAL:NAL) = CHA
                  STFING = STFING+1
               ELSE
                  NAL = NAL+1
                  ALPHA(NAL:NAL) = ESC
                  STFING = STFING+1
                  GOTO 988
               END IF
               GOTO 987
            ELSE
               IF (CHA.EQ.ESC) THEN
                  LIT = .TRUE.
                  GOTO 987
               ELSE IF (CHA.EQ.'''') THEN
                  IF (STFING.GT.0) THEN
                     RE = NALST
                     CALL PSHNUM(RE,NUM,NN,LLEN)
                     RE = NAL
                     CALL PSHNUM(RE,NUM,NN,LLEN)
                     CALL PSHSTR(STRING,TOKEN,NT,LLEN)
                  END IF
                  CALL PSHSTR(QUOTE,TOKEN,NT,LLEN)
C  In case of a null string we sent just two quote codes.
                  GOTO 100
               ELSE
                  NAL = NAL+1
                  ALPHA(NAL:NAL) = CHA
                  STFING = STFING+1
                  GOTO 987
               END IF
            END IF
         ELSE
C
C  End-of-Line reached.
            IF (LIT) THEN
               NAL = NAL+1
               ALPHA(NAL:NAL) = ESC
               STFING = STFING+1
            END IF
            IF (STFING.GT.0) THEN
               RE = NALST
               CALL PSHNUM(RE,NUM,NN,LLEN)
               RE = NAL
               CALL PSHNUM(RE,NUM,NN,LLEN)
               CALL PSHSTR(STRING,TOKEN,NT,LLEN)
            END IF
         END IF
      ELSE IF (CHA.EQ.'%') THEN
C
C  % is used as a comment character. Line scanning is terminated here.
      ELSE
C
C  Unrecognizable character.
         CALL PSHSTR(NOTREC,TOKEN,NT,LLEN)
C  Continue scanning.
         GOTO 100
C
      END IF
C
C  End of scanning.
9000  CONTINUE
C
C  Add the End-of-Line token.
      CALL PSHSTR(EOL,TOKEN,NT,LLEN)
      NTOK = NT
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE PSHALP ( ITEM, ALPHA, NAL, NUM, NN, LINLEN )
C  ---------------------------------------------------------------------
C
C  Description:
C    Stores ITEM in the character storage ALPHA.
C
C  Input arguments:
C    ITEM       Character item to be stored in ALPHA.
C    LINLEN     Length of the line under processing, as declared
C               in the program unit that called subroutine LEX.
C               IE: LINLEN = LEN(LINE)
C
C  Input / Output arguments:
C    ALPHA      The ALPHA character storage (see subroutine LEX).
C    NAL        Effective length of ALPHA.
C    NUM        The NUM array. (see subroutine LEX).
C    NN         Number of entries in array NUM.
C
C  ---------------------------------------------------------------------
C  Arguments:
      CHARACTER*(*) ALPHA, ITEM
      $(TYPE) NUM(*)
C  ---------------------------------------------------------------------
C  Local variables:
      $(TYPE) RE
C  ---------------------------------------------------------------------
C
      LES = LENGTH(ITEM)
      RE = NAL+1
      CALL PSHNUM(RE,NUM,NN,LINLEN)
      RE = NAL+LES
      CALL PSHNUM(RE,NUM,NN,LINLEN)
      ALPHA(NAL+1:NAL+LES) = ITEM
      NAL = NAL+LES
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE PSHNUM ( ITEM, NUM, NN, LINLEN )
C  ---------------------------------------------------------------------
C
C  Description:
C    Stores ITEM in array NUM.
C
C  Input arguments:
C    ITEM       The item to be stored.
C    LINLEN     Length of the line under processing, as declared
C               in the program unit that called subroutine LEX.
C               IE: LINLEN = LEN(LINE)
C
C  Input / Output arguments:
C    NUM        The NUM array. (see subroutine LEX).
C    NN         Number of entries in array NUM.
C
C  ---------------------------------------------------------------------
C  Arguments:
      $(TYPE) NUM(*), ITEM
C  ---------------------------------------------------------------------
C
      NN = NN+1
      IF (NN.LE.LINLEN) THEN
         NUM(NN) = ITEM
      ELSE
C  This should never happen.
         CALL MABORT ('Stack overflow.','PSHNUM')
      END IF
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE PSHSTR ( ITEM, TOKEN, NT, LINLEN )
C  ---------------------------------------------------------------------
C
C  Description:
C    Stores the token code ITEM in the array of tokens TOKEN.
C
C  Input arguments:
C    ITEM       The item to be stored.
C    LINLEN     Length of the line under processing, as declared
C               in the program unit that called subroutine LEX.
C               IE: LINLEN = LEN(LINE)
C
C  Input / Output arguments:
C    TOKEN      Tokens of the line under processing (see subroutine
C               LEX).
C    NT         Number of entries in TOKEN.
C
C  ---------------------------------------------------------------------
C  Arguments:
      INTEGER TOKEN(LINLEN+1)
C  ---------------------------------------------------------------------
C
      NT = NT+1
      IF (NT.LE.LINLEN+1) THEN
         TOKEN(NT) = ITEM
      ELSE
C  This should never happen.
         CALL MABORT ('Stack overflow.','PSHSTR')
      END IF
C
      END
C  ---------------------------------------------------------------------
      FUNCTION LXTRAN ( C )
C  ---------------------------------------------------------------------
C
C  Description:
C    Transforms some special characters to their corresponding token
C    codes to be used in subroutine LEX.
C
C  Input arguments:
C    C          The character to be transformed.
C
C  ---------------------------------------------------------------------
      IMPLICIT INTEGER (A-Z)
C  Arguments:
      CHARACTER*(*) C
C  ---------------------------------------------------------------------
C  Common blocks:
C  Codes used by the lexical analysis routine LEX.
      INTEGER LEFTP, RIGHTP, PERIOD, COMMA, GT, LT, NE, EQ, GE,
     &        LE, PLUS, MINUS, STAR, SLASH, BISTAR, COLON, EOL,
     &        IDENT, UNSI, NOTREC, STRING, QUOTE, LBRA, RBRA,
     &        SEMI, PERCE, AND, OR, XOR, NOT, EQUAL, LBRACE,
     &        RBRACE, QMARK
      COMMON /MCLSCD/ LEFTP, RIGHTP, PERIOD, COMMA, GT, LT, NE, EQ, GE,
     &                LE, PLUS, MINUS, STAR, SLASH, BISTAR, COLON, EOL,
     &                IDENT, UNSI, NOTREC, STRING, QUOTE, LBRA, RBRA,
     &                SEMI, PERCE, AND, OR, XOR, NOT, EQUAL, LBRACE,
     &                RBRACE, QMARK
C  ---------------------------------------------------------------------
C
      IF (C.EQ.'(') THEN
         LXTRAN = LEFTP
      ELSE IF (C.EQ.')') THEN
         LXTRAN = RIGHTP
      ELSE IF (C.EQ.'.') THEN
         LXTRAN = PERIOD
      ELSE IF (C.EQ.',') THEN
         LXTRAN = COMMA
      ELSE IF (C.EQ.'#') THEN
         LXTRAN = NE
      ELSE IF (C.EQ.':') THEN
         LXTRAN = COLON
      ELSE IF (C.EQ.'+') THEN
         LXTRAN = PLUS
      ELSE IF (C.EQ.'-') THEN
         LXTRAN = MINUS
      ELSE IF (C.EQ.'/') THEN
         LXTRAN = SLASH
      ELSE IF (C.EQ.'[') THEN
         LXTRAN = LBRA
      ELSE IF (C.EQ.']') THEN
         LXTRAN = RBRA
      ELSE IF (C.EQ.';') THEN
         LXTRAN = SEMI
      ELSE IF (C.EQ.'{') THEN
         LXTRAN = LBRACE
      ELSE IF (C.EQ.'}') THEN
         LXTRAN = RBRACE
      ELSE IF (C.EQ.'?') THEN
         LXTRAN = QMARK
      ELSE
         CALL MABORT ('Illegal input character.','LXTRAN')
      END IF
C
      END
C  ---------------------------------------------------------------------
      BLOCK DATA LEXINI
C  ---------------------------------------------------------------------
C
C  Description:
C    Assigns initial values to the integer token codes in common
C    block MCLSCD.
C
C  ---------------------------------------------------------------------
C  Common blocks:
C  Codes used by the lexical analysis routine LEX.
      INTEGER LEFTP, RIGHTP, PERIOD, COMMA, GT, LT, NE, EQ, GE,
     &        LE, PLUS, MINUS, STAR, SLASH, BISTAR, COLON, EOL,
     &        IDENT, UNSI, NOTREC, STRING, QUOTE, LBRA, RBRA,
     &        SEMI, PERCE, AND, OR, XOR, NOT, EQUAL, LBRACE,
     &        RBRACE, QMARK
      COMMON /MCLSCD/ LEFTP, RIGHTP, PERIOD, COMMA, GT, LT, NE, EQ, GE,
     &                LE, PLUS, MINUS, STAR, SLASH, BISTAR, COLON, EOL,
     &                IDENT, UNSI, NOTREC, STRING, QUOTE, LBRA, RBRA,
     &                SEMI, PERCE, AND, OR, XOR, NOT, EQUAL, LBRACE,
     &                RBRACE, QMARK
C  ---------------------------------------------------------------------
C
      DATA LEFTP, RIGHTP, PERIOD, COMMA, GT, LT, NE, EQ, GE, LE, PLUS /
     &         1,      2,      3,     4,  5,  6,  7,  8,  9, 10,   11 /
      DATA MINUS, STAR, SLASH, BISTAR, COLON, EOL, IDENT, UNSI, NOTREC /
     &        12,   13,    14,     15,    16,  17,    18,   19,     20 /
      DATA STRING, QUOTE, LBRA, RBRA, SEMI, PERCE, AND, OR, XOR, NOT /
     &         21,    22,   23,   24,   25,    26,  27, 28,  29,  30 /
      DATA EQUAL, LBRACE, RBRACE, QMARK /
     &        31,     32,     33,    34 /
C
      END
C  ---------------------------------------------------------------------
      FUNCTION NUNIT ( )
C  ---------------------------------------------------------------------
C
C  Description:
C    Function NUNIT will search a range of unit numbers in order to
C    locate an unused unit. (A unit not assigned to an open file).
C    Function NUNIT is a necessity since Fortran requires both, a
C    unit number and a file name in an OPEN statement. Nevertheless
C    function NUNIT and all of the Merlin file handling routines rely
C    on the INQUIRE statement and especially on the OPENED, EXIST,
C    NUMBER and IOSTAT specifiers.
C    The range of numbers to be searched can be set from the
C    configuration file. For efficiency, the search for the unit number
C    starts from the number returned in the previous NUNIT call, not
C    from the beginning of the available range.
C
C  ---------------------------------------------------------------------
C  Common blocks:
      COMMON /UALLOC/ MINU, MAXU, LASTU
C  ---------------------------------------------------------------------
C  Local variables:
      LOGICAL OPEN
C  ---------------------------------------------------------------------
C
C  Start the search from the previously returned unit.
      DO 10,I=LASTU+1,MAXU
         INQUIRE (UNIT=I,OPENED=OPEN,IOSTAT=IERR)
         IF (IERR.EQ.0) THEN
            IF (.NOT.OPEN) THEN
               LASTU = I
               NUNIT = I
               RETURN
            END IF
         END IF
10    CONTINUE
C
C  Wrap around the end of the available range.
      DO 20,I=MINU,LASTU
         INQUIRE (UNIT=I,OPENED=OPEN,IOSTAT=IERR)
         IF (IERR.EQ.0) THEN
            IF (.NOT.OPEN) THEN
               LASTU = I
               NUNIT = I
               RETURN
            END IF
         END IF
20    CONTINUE
C
C  No more unit numbers are available.
      NUNIT = -1
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE OLREAD ( NU, MAXCON, LINE, LLINE, ILL, NL, IERR )
C  ---------------------------------------------------------------------
C
C  Description:
C    Reads a number of lines (physical lines) from unit NU, until a line
C    without the continuation symbol & is reached, or until EOF. The
C    lines are concatenated and output in the large character variable
C    LLINE (logical line). If the maximum number of continuation lines
C    (MAXCON) is exceeded, subroutine OLREAD will continue reading
C    from unit NU; LLINE will not be enlarged however.
C
C  Input arguments:
C    NU         Unit to read from.
C    MAXCON     Maximum number of continuation lines.
C
C  Output arguments:
C    LINE       Character variable. Should have the length of one
C               physical line. Eg: CHARACTER*80 LINE.
C               LINE is used as a temporary variable and upon return
C               contains the last line read. If ILL = 0 the contents
C               of LINE are undefined.
C    LLINE      The input lines are concatenated and disposed here.
C               LLINE should be at least MAXCON*LEN(LINE) characters
C               long.
C    ILL        Effective length of LLINE.
C    NL         Number of lines read.
C    IERR       Error indicator:
C               IERR = 0 -> Everything ok.
C               IERR = 1 -> Too many continuation lines.
C               IERR = 2 -> EOF on input. If ILL.NE.0 this means that
C                           the continuation symbol & is present in the
C                           last line of the file, which is probably an
C                           error.
C
C  ---------------------------------------------------------------------
C  Arguments:
      CHARACTER*(*) LINE, LLINE
C  ---------------------------------------------------------------------
C  Local variables:
      LOGICAL ADDIT
C  ---------------------------------------------------------------------
C
      ILL = 0
      ADDIT = .TRUE.
      NL = 0
      LLINE = ' '
C
104   READ (NU,'(A)',END=200) LINE
      NL = NL+1
      IF (NL.GT.MAXCON) THEN
         ADDIT = .FALSE.
         IERR = 1
      END IF
      LE1 = LENGTH(LINE)
      IF (LE1.EQ.0) THEN
         IF (ADDIT) IERR = 0
         RETURN
      END IF
C
C  Is this a continuation line ?
      IF (LINE(LE1:LE1).EQ.'&') THEN
         IF (ADDIT) THEN
            LLINE(ILL+1:ILL+LE1-1) = LINE(1:LE1-1)
            ILL = ILL+LE1-1
         END IF
         GOTO 104
      END IF
      IF (ADDIT) THEN
         LLINE(ILL+1:ILL+LE1) = LINE(1:LE1)
         ILL = ILL+LE1
         IF (ILL.LT.LEN(LLINE)) LLINE(ILL+1:) = ' '
         IERR = 0
      END IF
      RETURN
C
200   CONTINUE
      IERR = 2
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE PARSP ( PDESCF, IUOUT, STRO, WEAK, ICODE )
C  ---------------------------------------------------------------------
C
C  Description:
C    Parses the panel description file and fills in the appropriate
C    data structures. This routine is called after the configuration
C    file has been processed, and assumes that the output file (unit
C    IUOUT) is open.
C    Panel related data are kept in a combination of real/integer
C    arrays and character variables. The following paragraphs describe
C    the data structures used for storing panel related data.
C
C    For the I-th panel command the following information is stored:
C      1) The name of the panel PCOM(I). PCOM is the panel table; a
C         CHARACTER*10 array.
C      2) Two integer pointers NCHELP(2*I-1), NCHELP(2*I) indicating
C         the starting and ending help lines for this panel, in the
C         panel description file.
C      3) An integer PONF(I), indicating the default status of the
C         panel: PONF(I) = 0 -> the panel is off.
C                PONF(I) = 1 -> the panel is off.
C      4) A pointer KPOINT(I), indicating the first keyword entry for
C         this panel, in the keyword table PKEY. After all panels
C         have been read in, a dummy entry is added at the end of the
C         KPOINT array, to indicate the "start" of a fake (non-
C         existent) panel. This is necessary since the number of
C         keywords in panel I is determined as: KPOINT(I+1)-KPOINT(I).
C
C    For the J-th keyword the following information is stored:
C      5) The name of the keyword PKEY(J). PKEY is the keyword table;
C         a CHARACTER*10 array. Keywords from all panels are stored
C         here.
C      6) The keyword type KTYPE(J). KTYPE(J) may be:
C           KTYPE(J) = 'I' for a keyword that accepts integer values.
C           KTYPE(J) = 'R' for a keyword that accepts real values.
C           KTYPE(J) = 'S' for a keyword that accepts string values.
C           KTYPE(J) = 'C' for a keyword that accepts string values;
C                          character case not being significant.
C      7) Two integer pointers NKHELP(2*J-1), NKHELP(2*J) indicating
C         the starting and ending help lines for this keyword, in the
C         panel description file.
C      8) A character KIO(J), indicating whether this is a normal or
C         an MCL returned keyword:
C           KIO(J)   =  ' ' for a normal keyword.
C           KIO(J) .NE. ' ' for an MCL returned keyword.
C      9) A pointer TPOINT(J), indicating the first entry for this
C         keyword in the PIR table.
C
C     10) The PIR table stores information related to the allowed
C         values of a keyword. (Assume TPOINT(J) = K)
C           a) For a keyword that accepts integer or real values:
C                PIR(K) = 1 This keyword has a range of allowed values.
C                PIR(K) = 2 This keyword has a set of allowed values.
C
C                If PIR(K) = 1 (range) then
C                  PIR(K+1) = 1 Interval is open at the left.
C                  PIR(K+1) = 2 Interval is closed at the left.
C                  PIR(K+1) = 3 Left limit is -Infinity.
C
C                  PIR(K+2) is the left limit; 0 for -Infinity.
C
C                  PIR(K+3) = 1 Interval is open at the right.
C                  PIR(K+3) = 2 Interval is closed at the right.
C                  PIR(K+3) = 3 Right limit is Infinity.
C
C                  PIR(K+4) is the right limit; 0 for -Infinity.
C                  PIR(K+5) is the current value of the keyword.
C                  PIR(K+6) is a pointer to the SPOINT array
C                           (description of the keyword).
C
C                If PIR(K) = 2 (set of values) then
C                  PIR(K+1) is the number of items in the set.
C                  PIR(K+2)...PIR(K+1+PIR(K+1)) is the set of allowed
C                                    values.
C                  PIR(K+2+PIR(K+1)) is the current value of the
C                                    keyword.
C                  PIR(K+3+PIR(K+1)) is a pointer to the SPOINT array
C                                    (description of the keyword).
C
C           b) For a keyword that accepts string or character values:
C                PIR(K) = 1 Any string is allowed as value.
C                PIR(K) = 2 This keyword has a set of allowed values.
C
C                If PIR(K) = 1 (any string) then
C                  PIR(K+1) is a pointer to the SPOINT array
C                           (current value).
C                  PIR(K+2) is a pointer to the SPOINT array
C                           (description of the keyword).
C
C                If PIR(K) = 2 (set of values) then
C                  PIR(K+1) = Number of items in the set.
C                  PIR(K+2)...PIR(K+1+PIR(K+1)) are pointers to the
C                             SPOINT array (allowed string items)
C                  PIR(K+2+PIR(K+1)) is a pointer to the SPOINT array
C                                    (current value).
C                  PIR(K+3+PIR(K+1)) is a pointer to the SPOINT array
C                                    (description of the keyword).
C
C     11) The SPOINT array stores the starting and ending character
C         positions for all panel related strings. For the L-th string
C         SPOINT(2*L-1) is the starting character position of the
C         string in the character storage PANSTR.
C         SPOINT(2*L) is the ending character position. Hence the
C         contents of the L-th string are:
C         PANSTR(SPOINT(2*L-1):SPOINT(2*L))
C
C     12) PANSTR is a large character, used to store all panel related
C         strings. Allowed values for string keywords, descriptions
C         of keywords, etc. are stored here. PANSTR is always addressed
C         through the pointers in array SPOINT.
C
C  Input arguments:
C    PDESCF     Name of the panel description file.
C    IUOUT      The Merlin output unit.
C    STRO       Informative messages are printed only if STRO = .TRUE.
C    WEAK       Error messages are printed only if WEAK = .TRUE.
C
C  Output arguments:
C    ICODE      Error indicator.
C
C  ---------------------------------------------------------------------
      IMPLICIT $(TYPE) (A-H,O-Z)
C  Arguments:
      CHARACTER*(*) PDESCF
      LOGICAL STRO, WEAK
C  ---------------------------------------------------------------------
C  Common blocks:
C  Codes used by the lexical analysis routine LEX.
      INTEGER LEFTP, RIGHTP, PERIOD, COMMA, GT, LT, NE, EQ, GE,
     &        LE, PLUS, MINUS, STAR, SLASH, BISTAR, COLON, EOL,
     &        IDENT, UNSI, NOTREC, STRING, QUOTE, LBRA, RBRA,
     &        SEMI, PERCE, AND, OR, XOR, NOT, EQUAL, LBRACE,
     &        RBRACE, QMARK
      COMMON /MCLSCD/ LEFTP, RIGHTP, PERIOD, COMMA, GT, LT, NE, EQ, GE,
     &                LE, PLUS, MINUS, STAR, SLASH, BISTAR, COLON, EOL,
     &                IDENT, UNSI, NOTREC, STRING, QUOTE, LBRA, RBRA,
     &                SEMI, PERCE, AND, OR, XOR, NOT, EQUAL, LBRACE,
     &                RBRACE, QMARK
C  Maximum length of a keyword.
      PARAMETER ( KEYLEN = 10 )
C  Maximum number of panel commands.
      PARAMETER ( MXPCOM = 80 )
C  Maximum number of keywords in all panel commands.
      PARAMETER ( MXPKEY = 1000 )
C  Maximum storage for integer/real keywords.
      PARAMETER ( MXPIR = 5000 )
C  Maximum storage for string keywords and descriptions.
      PARAMETER ( MXPSTR = 20000 )
C  Maximum number of pointers in the above string storage.
      PARAMETER ( MXPPST = 5000 )
C
      CHARACTER*(KEYLEN) PCOM(MXPCOM)
      INTEGER            PONF(MXPCOM)
      INTEGER            KPOINT(MXPCOM)
      INTEGER            NCHELP(2*MXPCOM)
      CHARACTER*(KEYLEN) PKEY(MXPKEY)
      INTEGER            NKHELP(2*MXPKEY)
      CHARACTER          KTYPE(MXPKEY)
      INTEGER            TPOINT(MXPKEY)
      CHARACTER          KIO(MXPKEY)
      $(TYPE)            PIR(MXPIR)
      INTEGER            SPOINT(MXPPST)
      CHARACTER*(MXPSTR) PANSTR
      COMMON /PCOMS1/ KPOINT, NCHELP, NKHELP, TPOINT, SPOINT, PONF
      COMMON /PCOMS2/ PIR
      COMMON /PCOMS3/ PKEY, KTYPE, PANSTR, PCOM, KIO
C
C  NPCOM: Number of panel commands in the panel description file.
C  NPKEY: Number of keywords in all panel commands.
C  NPIR:  Number of positions filled in array PIR.
C  NPSP:  Number of positions filled in the char-pointer array SPOINT.
C  NPCH:  Number of character positions filled in PANSTR.
      COMMON /PCOMS4/ NPCOM, NPKEY, NPIR, NPSP, NPCH
C  ---------------------------------------------------------------------
C  Local variables:
      PARAMETER ( LENONE = 120, MXCONT = 10, LPLEN = LENONE*MXCONT )
      CHARACTER*(LENONE) LINE
      CHARACTER*(LPLEN) LLINE
      INTEGER TOKEN(LPLEN+1)
      $(TYPE) NUM(LPLEN)
      CHARACTER*(LPLEN) ALPHA
      LOGICAL EX
      CHARACTER*10 STR
      CHARACTER UPPER1
C  These types can be handled by a panel.
      PARAMETER ( NTYPES = 4 )
      CHARACTER TYPE(NTYPES)
      DATA TYPE / 'I', 'R', 'S', 'C' /
C  ---------------------------------------------------------------------
C
C  Number of panel commands.
      NPCOM = 0
C  Number of keywords.
      NPKEY = 0
C  Number of entries in the "pir" table.
      NPIR = 0
C  Number of string pointers.
      NPSP = 0
C  Number of characters stored.
      NPCH = 0
C
C  Zero help pointers.
      DO 10,I=1,2*MXPCOM
10       NCHELP(I) = 0
      DO 20,I=1,2*MXPKEY
20       NKHELP(I) = 0
C
C  Return if there's no panel description file.
      IF (PDESCF.EQ.' ') THEN
         ICODE = 0
         RETURN
      END IF
C
C  Get info about the panel description file.
      LEP = LENGTH(PDESCF)
      INQUIRE (FILE=PDESCF, EXIST=EX, IOSTAT=IERR)
      IF (IERR.NE.0) THEN
         CALL I2STR(IERR,STR,LE1)
         IF (WEAK) WRITE (IUOUT,250) PDESCF(1:LEP), STR(1:LE1)
         ICODE = -1
         RETURN
      END IF
C
C  Does it exist ?
      IF (.NOT.EX) THEN
         IF (STRO) WRITE (IUOUT,251) PDESCF(1:LEP)
         ICODE = 0
         RETURN
      END IF
C
C  Allocate a unit.
      NU = NUNIT()
      IF (NU.EQ.-1) THEN
         IF (WEAK) WRITE (IUOUT,292) PDESCF(1:LEP)
         ICODE = -1
         RETURN
      END IF
C
C  Open the panel description file.
      OPEN (NU,FILE=PDESCF,IOSTAT=IERR)
      IF (IERR.NE.0) THEN
         CALL I2STR(IERR,STR,LE1)
         IF (WEAK) WRITE (IUOUT,252) PDESCF(1:LEP), STR(1:LE1)
         ICODE = -1
         RETURN
      END IF
C
C  NPL is the number of lines read so far.
      NPL = 0
C
C  Start the main loop. Read a line (along with all continuations).
100   CONTINUE
      CALL OLREAD(NU,MXCONT,LINE,LLINE,ILL,NL,IERR)
      NPL = NPL+NL
      IF (IERR.EQ.2) GOTO 200
      IF (IERR.EQ.1) THEN
         CALL I2STR(MXCONT,STR,LE1)
         IF (WEAK) WRITE (IUOUT,253) PDESCF(1:LEP), STR(1:LE1)
         CLOSE (NU)
         ICODE = -1
         RETURN
      END IF
C
C  Convert the input line LLINE to tokens.
      CALL LEX(LLINE,ILL,TOKEN,NTOK,NUM,ALPHA)
C  Empty line ?
      IF (TOKEN(1).EQ.EOL) GOTO 100
C  Comment line ?
      IF (TOKEN(1).EQ.NE) GOTO 100
C  Start parsing the tokens.
      NT = 0
      NN = 0
C  Check for "." lines. NTOK is at least 1 here.
      NT = NT+1
      IF (TOKEN(NT).EQ.PERIOD) THEN
         NT = NT+1
         IF (TOKEN(NT).NE.IDENT) THEN
            CALL PERRMS(PDESCF,LEP,WEAK,NPL,IUOUT)
            IF (WEAK) WRITE (IUOUT,254)
            CLOSE (NU)
            ICODE = -1
            RETURN
         END IF
         NN = NN+1
         IST = NINT(NUM(NN))
         NN = NN+1
         IEND = NINT(NUM(NN))
C  Is this is a .PANEL ?
         IF (ISCOMP(ALPHA(IST:IEND),'PANEL').NE.0) GOTO 501
C  Is this a .HELP ?
         IF (ISCOMP(ALPHA(IST:IEND),'HELP').NE.0) GOTO 502
      END IF
C
C  ------------------------- Keyword definition ------------------------
C
      IF (TOKEN(NT).NE.IDENT) THEN
         CALL PERRMS(PDESCF,LEP,WEAK,NPL,IUOUT)
         IF (WEAK) WRITE (IUOUT,255)
         CLOSE (NU)
         ICODE = -1
         RETURN
      END IF
      NN = NN+1
      IST = NINT(NUM(NN))
      NN = NN+1
      IEND = NINT(NUM(NN))
      LEK = IEND-IST+1
      IF (NPCOM.EQ.0) THEN
         CALL PERRMS(PDESCF,LEP,WEAK,NPL,IUOUT)
         IF (WEAK) WRITE (IUOUT,249) ALPHA(IST:IEND)
         CLOSE (NU)
         ICODE = -1
         RETURN
      END IF
      IF (LEK.GT.KEYLEN) THEN
         CALL I2STR(KEYLEN,STR,LES)
         LE1 = LENGTH(PCOM(NPCOM))
         CALL PERRMS(PDESCF,LEP,WEAK,NPL,IUOUT)
         IF (WEAK) WRITE (IUOUT,248) ALPHA(IST:IEND),
     &                       PCOM(NPCOM)(1:LE1), STR(1:LES)
         CLOSE (NU)
         ICODE = -1
         RETURN
      END IF
      DO 81,I=KPOINT(NPCOM),NPKEY
         IF (ISCOMP(PKEY(I),ALPHA(IST:IEND)).NE.0) THEN
            LE1 = LENGTH(PCOM(NPCOM))
            CALL PERRMS(PDESCF,LEP,WEAK,NPL,IUOUT)
            IF (WEAK) WRITE (IUOUT,256) ALPHA(IST:IEND),
     &                                        PCOM(NPCOM)(1:LE1)
            CLOSE (NU)
            ICODE = -1
            RETURN
         END IF
81    CONTINUE
      NPKEY = NPKEY+1
      IF (NPKEY.GT.MXPKEY)
     &    CALL OVER('pkey','PKEY','PARSP',MXPKEY,'MXPKEY')
      PKEY(NPKEY) = ALPHA(IST:IEND)
C
C  Get keyword type.
      NT = NT+1
      KIO(NPKEY) = ' '
      IF (TOKEN(NT).EQ.QMARK) THEN
         NT = NT+1
         KIO(NPKEY) = '?'
      END IF
      IF (TOKEN(NT).NE.IDENT) THEN
         LE1 = LENGTH(PKEY(NPKEY))
         LE2 = LENGTH(PCOM(NPCOM))
         CALL PERRMS(PDESCF,LEP,WEAK,NPL,IUOUT)
         IF (WEAK) WRITE (IUOUT,257) PKEY(NPKEY)(1:LE1),
     &                                  PCOM(NPCOM)(1:LE2)
         CLOSE (NU)
         ICODE = -1
         RETURN
      END IF
      NN = NN+1
      IST = NINT(NUM(NN))
      NN = NN+1
      IEND = NINT(NUM(NN))
      LE1 = IEND-IST+1
      DO 60,I=1,NTYPES
         IF (ISCOMP(TYPE(I),ALPHA(IST:IEND)).NE.0) GOTO 61
60    CONTINUE
      LE1 = LENGTH(PKEY(NPKEY))
      LE2 = LENGTH(PCOM(NPCOM))
      CALL PERRMS(PDESCF,LEP,WEAK,NPL,IUOUT)
      IF (WEAK) WRITE (IUOUT,258) PKEY(NPKEY)(1:LE1), PCOM(NPCOM)(1:LE2)
      CLOSE (NU)
      ICODE = -1
      RETURN
C
61    CONTINUE
      ITY = I
      KTYPE(NPKEY) = UPPER1(ALPHA(IST:IEND))
C
C  Get range of allowed values.
      IF (ITY.EQ.1) THEN
         TPOINT(NPKEY) = NPIR+1
         CALL IRANGE(TOKEN,NUM,ALPHA,NT,NN,IERR,PIR,NPIR,MXPIR,
     &                  PDESCF,NPL,PCOM(NPCOM),PKEY(NPKEY),IUOUT,
     &                  LEP,WEAK)
      ELSE IF (ITY.EQ.2) THEN
         TPOINT(NPKEY) = NPIR+1
         CALL RRANGE(TOKEN,NUM,ALPHA,NT,NN,IERR,PIR,NPIR,MXPIR,
     &                  PDESCF,NPL,PCOM(NPCOM),PKEY(NPKEY),IUOUT,
     &                  LEP,WEAK)
      ELSE IF (ITY.EQ.3 .OR. ITY.EQ.4) THEN
         TPOINT(NPKEY) = NPIR+1
         CALL SRANGE(TOKEN,NUM,ALPHA,NT,NN,IERR,PIR,NPIR,MXPIR,
     &                  SPOINT,NPSP,MXPPST,PANSTR,NPCH,MXPSTR,PDESCF,
     &                  NPL,PCOM(NPCOM),PKEY(NPKEY),IUOUT,LEP,WEAK)
      END IF
      IF (IERR.NE.0) THEN
         CLOSE (NU)
         ICODE = -1
         RETURN
      END IF
C
C  Get default value.
      NT = NT+1
      IF (ITY.EQ.1) THEN
C  Default value for an integer.
         ISI = 1
         IF (TOKEN(NT).EQ.PLUS) THEN
            ISI = 1
            NT = NT+1
         ELSE IF (TOKEN(NT).EQ.MINUS) THEN
            ISI = -1
            NT = NT+1
         END IF
         IF (TOKEN(NT).NE.UNSI) THEN
            CALL PERRMS(PDESCF,LEP,WEAK,NPL,IUOUT)
            LE1 = LENGTH(PKEY(NPKEY))
            LE2 = LENGTH(PCOM(NPCOM))
            IF (WEAK) WRITE (IUOUT,259)
     &                      PKEY(NPKEY)(1:LE1), PCOM(NPCOM)(1:LE2)
            CLOSE (NU)
            ICODE = -1
            RETURN
         END IF
         NN = NN+1
         XDE = NUM(NN)
         IF (XDE.NE.INT(XDE)) THEN
            CALL PERRMS(PDESCF,LEP,WEAK,NPL,IUOUT)
            LE1 = LENGTH(PKEY(NPKEY))
            LE2 = LENGTH(PCOM(NPCOM))
            IF (WEAK) WRITE (IUOUT,259)
     &                      PKEY(NPKEY)(1:LE1), PCOM(NPCOM)(1:LE2)
            CLOSE (NU)
            ICODE = -1
            RETURN
         END IF
         IDEF = ISI*NINT(XDE)
         CALL PRANI(IDEF,NPKEY,IOK)
         IF (IOK.EQ.0) THEN
            CALL PERRMS(PDESCF,LEP,WEAK,NPL,IUOUT)
            LE1 = LENGTH(PKEY(NPKEY))
            LE2 = LENGTH(PCOM(NPCOM))
            IF (WEAK) WRITE (IUOUT,261)
     &                      PKEY(NPKEY)(1:LE1), PCOM(NPCOM)(1:LE2)
            CLOSE (NU)
            ICODE = -1
            RETURN
         END IF
         NPIR = NPIR+1
         IF (NPIR.GT.MXPIR)
     &          CALL OVER('pir','PIR','PARSP',MXPIR,'MXPIR')
         PIR(NPIR) = IDEF
      ELSE IF (ITY.EQ.2) THEN
C  Default value for a real.
         ISI = 1
         IF (TOKEN(NT).EQ.PLUS) THEN
            ISI = 1
            NT = NT+1
         ELSE IF (TOKEN(NT).EQ.MINUS) THEN
            ISI = -1
            NT = NT+1
         END IF
         IF (TOKEN(NT).NE.UNSI) THEN
            CALL PERRMS(PDESCF,LEP,WEAK,NPL,IUOUT)
            LE1 = LENGTH(PKEY(NPKEY))
            LE2 = LENGTH(PCOM(NPCOM))
            IF (WEAK) WRITE (IUOUT,262)
     &                      PKEY(NPKEY)(1:LE1), PCOM(NPCOM)(1:LE2)
            CLOSE (NU)
            ICODE = -1
            RETURN
         END IF
         NN = NN+1
         XDE = ISI*NUM(NN)
         CALL PRANR(XDE,NPKEY,IOK)
         IF (IOK.EQ.0) THEN
            CALL PERRMS(PDESCF,LEP,WEAK,NPL,IUOUT)
            LE1 = LENGTH(PKEY(NPKEY))
            LE2 = LENGTH(PCOM(NPCOM))
            IF (WEAK) WRITE (IUOUT,261)
     &                      PKEY(NPKEY)(1:LE1), PCOM(NPCOM)(1:LE2)
            CLOSE (NU)
            ICODE = -1
            RETURN
         END IF
         NPIR = NPIR+1
         IF (NPIR.GT.MXPIR)
     &          CALL OVER('pir','PIR','PARSP',MXPIR,'MXPIR')
         PIR(NPIR) = XDE
      ELSE IF (ITY.EQ.3 .OR. ITY.EQ.4) THEN
C  Default value for a string.
         IF (TOKEN(NT).NE.QUOTE) THEN
            CALL PERRMS(PDESCF,LEP,WEAK,NPL,IUOUT)
            LE1 = LENGTH(PKEY(NPKEY))
            LE2 = LENGTH(PCOM(NPCOM))
            IF (WEAK) WRITE (IUOUT,264)
     &                      PKEY(NPKEY)(1:LE1), PCOM(NPCOM)(1:LE2)
            CLOSE (NU)
            ICODE = -1
            RETURN
         END IF
         NT = NT+1
         IF (TOKEN(NT).NE.STRING) THEN
            CALL PERRMS(PDESCF,LEP,WEAK,NPL,IUOUT)
            LE1 = LENGTH(PKEY(NPKEY))
            LE2 = LENGTH(PCOM(NPCOM))
            IF (WEAK) WRITE (IUOUT,264)
     &                      PKEY(NPKEY)(1:LE1), PCOM(NPCOM)(1:LE2)
            CLOSE (NU)
            ICODE = -1
            RETURN
         END IF
         NT = NT+1
         IF (TOKEN(NT).NE.QUOTE) THEN
            CALL PERRMS(PDESCF,LEP,WEAK,NPL,IUOUT)
            LE1 = LENGTH(PKEY(NPKEY))
            LE2 = LENGTH(PCOM(NPCOM))
            IF (WEAK) WRITE (IUOUT,264)
     &                      PKEY(NPKEY)(1:LE1), PCOM(NPCOM)(1:LE2)
            CLOSE (NU)
            ICODE = -1
            RETURN
         END IF
         NN = NN+1
         IST = NINT(NUM(NN))
         NN = NN+1
         IEND = NINT(NUM(NN))
         LE1 = IEND-IST+1
         CALL PRANS(ALPHA(IST:IEND),LE1,NPKEY,IOK)
         IF (IOK.EQ.0) THEN
            CALL PERRMS(PDESCF,LEP,WEAK,NPL,IUOUT)
            LE1 = LENGTH(PKEY(NPKEY))
            LE2 = LENGTH(PCOM(NPCOM))
            IF (WEAK) WRITE (IUOUT,261)
     &                      PKEY(NPKEY)(1:LE1), PCOM(NPCOM)(1:LE2)
            CLOSE (NU)
            ICODE = -1
            RETURN
         END IF
         NPIR = NPIR+1
         IF (NPIR.GT.MXPIR)
     &          CALL OVER('pir','PIR','PARSP',MXPIR,'MXPIR')
         PIR(NPIR) = NPSP+1
         IF (NPSP+2.GT.MXPPST)
     &          CALL OVER('spoint','SPOINT','PARSP',MXPPST,'MXPPST')
         SPOINT(NPSP+1) = NPCH+1
         SPOINT(NPSP+2) = NPCH+LE1
         NPSP = NPSP+2
         IF (NPCH+LE1.GT.MXPSTR)
     &          CALL OVER('panstr','PANSTR','PARSP',MXPSTR,'MXPSTR')
         PANSTR(NPCH+1:NPCH+LE1) = ALPHA(IST:IEND)
         NPCH = NPCH+LE1
      END IF
C
C  Get description.
      NT = NT+1
      IF (TOKEN(NT).NE.QUOTE) THEN
         CALL PERRMS(PDESCF,LEP,WEAK,NPL,IUOUT)
         LE1 = LENGTH(PKEY(NPKEY))
         LE2 = LENGTH(PCOM(NPCOM))
         IF (WEAK) WRITE (IUOUT,268)
     &                PKEY(NPKEY)(1:LE1), PCOM(NPCOM)(1:LE2)
         CLOSE (NU)
         ICODE = -1
         RETURN
      END IF
      NT = NT+1
      IF (TOKEN(NT).NE.STRING) THEN
         CALL PERRMS(PDESCF,LEP,WEAK,NPL,IUOUT)
         LE1 = LENGTH(PKEY(NPKEY))
         LE2 = LENGTH(PCOM(NPCOM))
         IF (WEAK) WRITE (IUOUT,268)
     &             PKEY(NPKEY)(1:LE1), PCOM(NPCOM)(1:LE2)
         CLOSE (NU)
         ICODE = -1
         RETURN
      END IF
      NT = NT+1
      IF (TOKEN(NT).NE.QUOTE) THEN
         CALL PERRMS(PDESCF,LEP,WEAK,NPL,IUOUT)
         LE1 = LENGTH(PKEY(NPKEY))
         LE2 = LENGTH(PCOM(NPCOM))
         IF (WEAK) WRITE (IUOUT,268)
     &                PKEY(NPKEY)(1:LE1), PCOM(NPCOM)(1:LE2)
         CLOSE (NU)
         ICODE = -1
         RETURN
      END IF
      NN = NN+1
      IST = NINT(NUM(NN))
      NN = NN+1
      IEND = NINT(NUM(NN))
      LE1 = IEND-IST+1
      NPIR = NPIR+1
      IF (NPIR.GT.MXPIR)
     &    CALL OVER('pir','PIR','PARSP',MXPIR,'MXPIR')
      PIR(NPIR) = NPSP+1
      IF (NPSP+2.GT.MXPPST)
     &    CALL OVER('spoint','SPOINT','PARSP',MXPPST,'MXPPST')
      SPOINT(NPSP+1) = NPCH+1
      SPOINT(NPSP+2) = NPCH+LE1
      NPSP = NPSP+2
      IF (NPCH+LE1.GT.MXPSTR)
     &    CALL OVER('panstr','PANSTR','PARSP',MXPSTR,'MXPSTR')
      PANSTR(NPCH+1:NPCH+LE1) = ALPHA(IST:IEND)
      NPCH = NPCH+LE1
      NT = NT+1
      IF (NT.NE.NTOK) THEN
         CALL PERRMS(PDESCF,LEP,WEAK,NPL,IUOUT)
         LE1 = LENGTH(PKEY(NPKEY))
         LE2 = LENGTH(PCOM(NPCOM))
         IF (WEAK) WRITE (IUOUT,271)
     &                PKEY(NPKEY)(1:LE1), PCOM(NPCOM)(1:LE2)
         CLOSE (NU)
         ICODE = -1
         RETURN
      END IF
C
C  Read the next line from the panel description file.
      GOTO 100
C
C  -------------------------- Panel definition -------------------------
501   CONTINUE
      NT = NT+1
      IF (TOKEN(NT).NE.IDENT) THEN
         CALL PERRMS(PDESCF,LEP,WEAK,NPL,IUOUT)
         IF (WEAK) WRITE (IUOUT,272)
         CLOSE (NU)
         ICODE = -1
         RETURN
      END IF
C
      NN = NN+1
      IST = NINT(NUM(NN))
      NN = NN+1
      IEND = NINT(NUM(NN))
      LE1 = IEND-IST+1
C  Check against other panel names. (Should check all Merlin commands)
      CALL PLOCP(ALPHA(IST:IEND),IPA)
      IF (IPA.NE.0) THEN
         CALL PERRMS(PDESCF,LEP,WEAK,NPL,IUOUT)
         IF (WEAK) WRITE (IUOUT,274) ALPHA(IST:IEND)
         CLOSE (NU)
         ICODE = -1
         RETURN
      END IF
      NPCOM = NPCOM+1
      IF (NPCOM.GT.MXPCOM)
     &    CALL OVER('pcom','PCOM','PARSP',MXPCOM,'MXPCOM')
      PCOM(NPCOM) = ALPHA(IST:IEND)
      KPOINT(NPCOM) = NPKEY+1
C
C  Get panel status. Should be "on" or "off".
      NT = NT+1
      IF (TOKEN(NT).NE.IDENT) THEN
         CALL PERRMS(PDESCF,LEP,WEAK,NPL,IUOUT)
         IF (WEAK) WRITE (IUOUT,242)
         CLOSE (NU)
         ICODE = -1
         RETURN
      END IF
C
      NN = NN+1
      IST = NINT(NUM(NN))
      NN = NN+1
      IEND = NINT(NUM(NN))
      LE1 = IEND-IST+1
      IF (ISCOMP(ALPHA(IST:IEND),'ON').NE.0) THEN
         PONF(NPCOM) = 1
      ELSE IF (ISCOMP(ALPHA(IST:IEND),'OFF').NE.0) THEN
         PONF(NPCOM) = 0
      ELSE
         CALL PERRMS(PDESCF,LEP,WEAK,NPL,IUOUT)
         IF (WEAK) WRITE (IUOUT,242)
         CLOSE (NU)
         ICODE = -1
         RETURN
      END IF
C
      NT = NT+1
      IF (NT.NE.NTOK) THEN
         CALL PERRMS(PDESCF,LEP,WEAK,NPL,IUOUT)
         IF (WEAK) WRITE (IUOUT,273)
         CLOSE (NU)
         ICODE = -1
         RETURN
      END IF
C
C  Read the next line from the panel description file.
      GOTO 100
C
C
C  -------------------------- Help definition --------------------------
502   CONTINUE
      IF (NPCOM.EQ.0) THEN
         CALL PERRMS(PDESCF,LEP,WEAK,NPL,IUOUT)
         IF (WEAK) WRITE (IUOUT,279)
         CLOSE (NU)
         ICODE = -1
         RETURN
      END IF
      NT = NT+1
      IF (NT.EQ.NTOK) THEN
         IF (NCHELP(2*NPCOM-1).NE.0) THEN
            CALL PERRMS(PDESCF,LEP,WEAK,NPL,IUOUT)
            LE1 = LENGTH(PCOM(NPCOM))
            IF (WEAK) WRITE (IUOUT,284) PCOM(NPCOM)(1:LE1)
            CLOSE (NU)
            ICODE = -1
            RETURN
         END IF
         IWHAT = 1
      ELSE
C  Get key name here.
         IF (TOKEN(NT).NE.IDENT) THEN
            CALL PERRMS(PDESCF,LEP,WEAK,NPL,IUOUT)
            IF (WEAK) WRITE (IUOUT,281)
            CLOSE (NU)
            ICODE = -1
            RETURN
         END IF
         NN = NN+1
         IST = NINT(NUM(NN))
         NN = NN+1
         IEND = NINT(NUM(NN))
         LE1 = IEND-IST+1
         NT = NT+1
         IF (NT.NE.NTOK) THEN
            CALL PERRMS(PDESCF,LEP,WEAK,NPL,IUOUT)
            IF (WEAK) WRITE (IUOUT,280)
            CLOSE (NU)
            ICODE = -1
            RETURN
         END IF
         DO 82,IK=KPOINT(NPCOM),NPKEY
            IF (ISCOMP(PKEY(IK),ALPHA(IST:IEND)).NE.0) GOTO 83
82       CONTINUE
         CALL PERRMS(PDESCF,LEP,WEAK,NPL,IUOUT)
         LE1 = LENGTH(PCOM(NPCOM))
         IF (WEAK) WRITE (IUOUT,283) ALPHA(IST:IEND),
     &                                  PCOM(NPCOM)(1:LE1)
         CLOSE (NU)
         ICODE = -1
         RETURN
C
83       CONTINUE
         IF (NKHELP(2*IK-1).NE.0) THEN
            CALL PERRMS(PDESCF,LEP,WEAK,NPL,IUOUT)
            LE1 = LENGTH(PKEY(IK))
            LE2 = LENGTH(PCOM(NPCOM))
            IF (WEAK) WRITE (IUOUT,285) PKEY(IK)(1:LE1),
     &                                        PCOM(NPCOM)(1:LE2)
            CLOSE (NU)
            ICODE = -1
            RETURN
         END IF
         IWHAT = 2
      END IF
C
      IF (IWHAT.EQ.1) THEN
         NCHELP(2*NPCOM-1) = NPL+1
      ELSE
         NKHELP(2*IK-1) = NPL+1
      END IF
C
C  Scan file until .end
202   CONTINUE
      READ (NU,'(A)',END=201) LINE
      NPL = NPL+1
      IF (ISCOMP(LINE,'.END').EQ.0) GOTO 202
      IF (IWHAT.EQ.1) THEN
         NCHELP(2*NPCOM) = NPL-1
      ELSE
         NKHELP(2*IK) = NPL-1
      END IF
      GOTO 100
C
201   CONTINUE
      CALL PERRMS(PDESCF,LEP,WEAK,NPL,IUOUT)
      IF (WEAK) WRITE (IUOUT,282)
      CLOSE (NU)
      ICODE = -1
      RETURN
C
200   CONTINUE
      CLOSE (NU)
      IF (ILL.NE.0) THEN
         CALL PERRMS(PDESCF,LEP,WEAK,NPL,IUOUT)
         IF (WEAK) WRITE (IUOUT,275)
         ICODE = -1
         RETURN
      END IF
      IF (NPCOM+1.GT.MXPCOM)
     &    CALL OVER('kpoint','KPOINT','PARSP',MXPCOM,'MXPCOM')
      KPOINT(NPCOM+1) = NPKEY+1
      ICODE = 0
C
C  The following WRITE statements display the amount of storage used
C  by the current panel description file. Used for development purposes.
C     IF (STRO) THEN
C         WRITE (IUOUT,*) 'Number of panel commands. NPCOM = ', NPCOM
C         WRITE (IUOUT,*) 'Number of keywords. NPKEY = ', NPKEY
C         WRITE (IUOUT,*) 'Number of entries in the "pir" table. ',
C     &                    'NPIR = ',NPIR
C         WRITE (IUOUT,*) 'Number of string pointers. NPSP = ', NPSP
C         WRITE (IUOUT,*) 'Number of characters stored. NPCH = ', NPCH
C     END IF
C
250   FORMAT (/$(FC)'An I/O error occurred while inquiring for the ',
     &        'existence'
     &        /$(FC)'of the panel description file "',A,
     &        '". IOSTAT = ',A,'.')
251   FORMAT (/$(FC)'Warning: The panel description file "',A,
     &        '" does not exist.')
292   FORMAT (/$(FC)'Cannot allocate a unit number for the ',
     &        'panel description file "',A,'".')
252   FORMAT (/$(FC)'An I/O error occurred while opening the ',
     &        'panel description ',
     &        /$(FC)'file "',A,'". IOSTAT = ',A)
253   FORMAT (/$(FC)'Too many continuation lines in the panel ',
     &        'description file "',A,'".'
     &        /$(FC)'Maximum number of continuation lines is ',A,'.')
254   FORMAT ($(FC)'An identifier is expected after the dot.')
255   FORMAT ($(FC)'An identifier is expected at the begining ',
     &        'of the line.')
249   FORMAT ($(FC)'The keyword ',A,' was found outside a panel ',
     &        'definition.')
248   FORMAT ($(FC)'The keyword ',A, ' in the ',A,' panel'/
     &        $(FC)'is too long. Maximum is ',A,' characters.')
256   FORMAT ($(FC)'The keyword ',A, ' has already ',
     &        'been defined in the ',A,' panel.')
257   FORMAT ($(FC)'The keyword type (I, R, S or C) is expected ',
     &        'after the ',A,' keyword'
     &         /$(FC)'in the ',A,' panel.')
258   FORMAT ($(FC)'The type for the ',A,' keyword in the ',A,' panel'
     &        /$(FC)'is unrecognized. It should be I, R, S or C.')
259   FORMAT ($(FC)'The default value for the ',A,' keyword in the ',
     &        A,' panel'
     &        /$(FC)'should be an integer.')
261   FORMAT ($(FC)'The default value for the ',A,' keyword in the ',
     &        A,' panel'
     &        /$(FC)'is out of the specified range.')
262   FORMAT ($(FC)'The default value for the ',A,' keyword in the ',
     &        A,' panel'
     &        /$(FC)'should be a real.')
264   FORMAT ($(FC)'The default value for the ',A,' keyword in the ',
     &        A,' panel'
     &        /$(FC)'should be a string.')
268   FORMAT ($(FC)'A string is expected as a description for the ',
     &        A,' keyword'
     &        /$(FC)'in the ',A,' panel.')
271   FORMAT ($(FC)'Too many items in the input line for the ',
     &        A,' keyword'
     &        /$(FC)'in the ',A,' panel.')
272   FORMAT ($(FC)'A panel name is expected after ".panel".')
274   FORMAT ($(FC)'The ',A,' panel has alredy been defined.')
242   FORMAT ($(FC)'The panel status ("on" or "off") is expected ',
     &        'after the panel name.')
273   FORMAT ($(FC)'Too many items in the ".panel" line.')
279   FORMAT ($(FC)'Help was found outside a panel definition.')
284   FORMAT ($(FC)'You already specified ".help" for the ',
     &        A,' panel.')
281   FORMAT ($(FC)'Keywords only should follow ".help".')
280   FORMAT ($(FC)'Too many items in the ".help" line.')
283   FORMAT ($(FC)'There is no such keyword ',A,' in the ',
     &        A,' panel.')
285   FORMAT ($(FC)'You already specified ".help" for the ',
     &        A,' keyword in the ',A,' panel.')
282   FORMAT ($(FC)'A ".help" has no corresponding ".end".')
275   FORMAT ($(FC)'A continuation mark was specified in the last line'
     &        /$(FC)'of the panel description file.')
      END
C  ---------------------------------------------------------------------
      SUBROUTINE PERRMS ( PDESCF, LEP, WEAK, NPL, IUOUT )
C  ---------------------------------------------------------------------
C
C  Description:
C    Called by subroutine PARSP. Prints an informative header
C    before subroutine PARSP issues any error message.
C
C  Input arguments:
C    PDESCF     Name of the panel description file.
C    LEP        Effective length of PDESCF.
C    WEAK       This routine prints only if WEAK = .TRUE.
C    NPL        Number of lines read so far from the panel description
C               file.
C    IUOUT      The Merlin output unit.
C
C  ---------------------------------------------------------------------
C  Arguments:
      CHARACTER*(*) PDESCF
      LOGICAL WEAK
C  ---------------------------------------------------------------------
C  Local variables:
      CHARACTER*10 STR
C  ---------------------------------------------------------------------
C
      IF (WEAK) THEN
         CALL I2STR(NPL,STR,LES)
         CALL DASH
         WRITE (IUOUT,10) PDESCF(1:LEP), STR(1:LES)
      END IF
C
10    FORMAT ($(FC)'An error occurred in the panel description file "',
     &        A,'", line ',A,'.')
      END
C  ---------------------------------------------------------------------
      SUBROUTINE PLOCK ( IPA, KEY, IKE )
C  ---------------------------------------------------------------------
C
C  Description:
C    Determines the position of the keyword KEY in the keyword table
C    PKEY. For more information on the structures that store panel
C    related information, see the discussion in subroutine PARSP.
C
C  Input arguments:
C    IPA        Index of the panel, KEY belongs to.
C    KEY        The keyword to search for.
C
C  Output arguments:
C    IKE        Index of KEY in the PKEY table. IKE is set to 0
C               if KEY is not a valid keyword for the panel with
C               index IPA.
C
C  ---------------------------------------------------------------------
      IMPLICIT $(TYPE) (A-H,O-Z)
C  Arguments:
      CHARACTER*(*) KEY
C  ---------------------------------------------------------------------
C  Common blocks:
C  Maximum length of a keyword.
      PARAMETER ( KEYLEN = 10 )
C  Maximum number of panel commands.
      PARAMETER ( MXPCOM = 80 )
C  Maximum number of keywords in all panel commands.
      PARAMETER ( MXPKEY = 1000 )
C  Maximum storage for integer/real keywords.
      PARAMETER ( MXPIR = 5000 )
C  Maximum storage for string keywords and descriptions.
      PARAMETER ( MXPSTR = 20000 )
C  Maximum number of pointers in the above string storage.
      PARAMETER ( MXPPST = 5000 )
C
      CHARACTER*(KEYLEN) PCOM(MXPCOM)
      INTEGER            PONF(MXPCOM)
      INTEGER            KPOINT(MXPCOM)
      INTEGER            NCHELP(2*MXPCOM)
      CHARACTER*(KEYLEN) PKEY(MXPKEY)
      INTEGER            NKHELP(2*MXPKEY)
      CHARACTER          KTYPE(MXPKEY)
      INTEGER            TPOINT(MXPKEY)
      CHARACTER          KIO(MXPKEY)
      $(TYPE)            PIR(MXPIR)
      INTEGER            SPOINT(MXPPST)
      CHARACTER*(MXPSTR) PANSTR
      COMMON /PCOMS1/ KPOINT, NCHELP, NKHELP, TPOINT, SPOINT, PONF
      COMMON /PCOMS2/ PIR
      COMMON /PCOMS3/ PKEY, KTYPE, PANSTR, PCOM, KIO
C
C  NPCOM: Number of panel commands in the panel description file.
C  NPKEY: Number of keywords in all panel commands.
C  NPIR:  Number of positions filled in array PIR.
C  NPSP:  Number of positions filled in the char-pointer array SPOINT.
C  NPCH:  Number of character positions filled in PANSTR.
      COMMON /PCOMS4/ NPCOM, NPKEY, NPIR, NPSP, NPCH
C  ---------------------------------------------------------------------
C
      KP1 = KPOINT(IPA)
      KP2 = KPOINT(IPA+1)-1
      DO 10,I=KP1,KP2
         IF (ISCOMP(KEY,PKEY(I)).NE.0) THEN
            IKE = I
            RETURN
         END IF
10    CONTINUE
C
C  Keyword not found.
      IKE = 0
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE PLOCP ( PNAME, IPOS )
C  ---------------------------------------------------------------------
C
C  Description:
C    Determines the position IPOS of the panel PNAME in the panel table.
C
C  Input arguments:
C    PNAME      Name of the panel.
C
C  Output arguments:
C    IPOS       Position of PNAME in the panel table. IPOS is set to 0
C               if PNAME is not a valid panel.
C
C  ---------------------------------------------------------------------
      IMPLICIT $(TYPE) (A-H,O-Z)
C  Arguments:
      CHARACTER*(*) PNAME
C  ---------------------------------------------------------------------
C  Common blocks:
C  Maximum length of a keyword.
      PARAMETER ( KEYLEN = 10 )
C  Maximum number of panel commands.
      PARAMETER ( MXPCOM = 80 )
C  Maximum number of keywords in all panel commands.
      PARAMETER ( MXPKEY = 1000 )
C  Maximum storage for integer/real keywords.
      PARAMETER ( MXPIR = 5000 )
C  Maximum storage for string keywords and descriptions.
      PARAMETER ( MXPSTR = 20000 )
C  Maximum number of pointers in the above string storage.
      PARAMETER ( MXPPST = 5000 )
C
      CHARACTER*(KEYLEN) PCOM(MXPCOM)
      INTEGER            PONF(MXPCOM)
      INTEGER            KPOINT(MXPCOM)
      INTEGER            NCHELP(2*MXPCOM)
      CHARACTER*(KEYLEN) PKEY(MXPKEY)
      INTEGER            NKHELP(2*MXPKEY)
      CHARACTER          KTYPE(MXPKEY)
      INTEGER            TPOINT(MXPKEY)
      CHARACTER          KIO(MXPKEY)
      $(TYPE)            PIR(MXPIR)
      INTEGER            SPOINT(MXPPST)
      CHARACTER*(MXPSTR) PANSTR
      COMMON /PCOMS1/ KPOINT, NCHELP, NKHELP, TPOINT, SPOINT, PONF
      COMMON /PCOMS2/ PIR
      COMMON /PCOMS3/ PKEY, KTYPE, PANSTR, PCOM, KIO
C
C  NPCOM: Number of panel commands in the panel description file.
C  NPKEY: Number of keywords in all panel commands.
C  NPIR:  Number of positions filled in array PIR.
C  NPSP:  Number of positions filled in the char-pointer array SPOINT.
C  NPCH:  Number of character positions filled in PANSTR.
      COMMON /PCOMS4/ NPCOM, NPKEY, NPIR, NPSP, NPCH
C  ---------------------------------------------------------------------
C
      DO 10,I=1,NPCOM
         IF (ISCOMP(PNAME,PCOM(I)).NE.0) THEN
            IPOS = I
            RETURN
         END IF
10    CONTINUE
C
C  Panel not found.
      IPOS = 0
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE PRANI ( IVAL, NK, IOK )
C  ---------------------------------------------------------------------
C
C  Description:
C    Checks whether the given integer IVAL is allowed as a value for a
C    keyword.
C
C  Input arguments:
C    IVAL       The integer to check.
C    NK         The index of the corresponding keyword in the keyword
C               table PKEY.
C
C  Output arguments:
C    IOK        IOK = 0 -> Value not allowed for this keyword.
C               IOK = 1 -> Value ok.
C
C  ---------------------------------------------------------------------
      IMPLICIT $(TYPE) (A-H,O-Z)
C  Common blocks:
C  Maximum length of a keyword.
      PARAMETER ( KEYLEN = 10 )
C  Maximum number of panel commands.
      PARAMETER ( MXPCOM = 80 )
C  Maximum number of keywords in all panel commands.
      PARAMETER ( MXPKEY = 1000 )
C  Maximum storage for integer/real keywords.
      PARAMETER ( MXPIR = 5000 )
C  Maximum storage for string keywords and descriptions.
      PARAMETER ( MXPSTR = 20000 )
C  Maximum number of pointers in the above string storage.
      PARAMETER ( MXPPST = 5000 )
C
      CHARACTER*(KEYLEN) PCOM(MXPCOM)
      INTEGER            PONF(MXPCOM)
      INTEGER            KPOINT(MXPCOM)
      INTEGER            NCHELP(2*MXPCOM)
      CHARACTER*(KEYLEN) PKEY(MXPKEY)
      INTEGER            NKHELP(2*MXPKEY)
      CHARACTER          KTYPE(MXPKEY)
      INTEGER            TPOINT(MXPKEY)
      CHARACTER          KIO(MXPKEY)
      $(TYPE)            PIR(MXPIR)
      INTEGER            SPOINT(MXPPST)
      CHARACTER*(MXPSTR) PANSTR
      COMMON /PCOMS1/ KPOINT, NCHELP, NKHELP, TPOINT, SPOINT, PONF
      COMMON /PCOMS2/ PIR
      COMMON /PCOMS3/ PKEY, KTYPE, PANSTR, PCOM, KIO
C
C  NPCOM: Number of panel commands in the panel description file.
C  NPKEY: Number of keywords in all panel commands.
C  NPIR:  Number of positions filled in array PIR.
C  NPSP:  Number of positions filled in the char-pointer array SPOINT.
C  NPCH:  Number of character positions filled in PANSTR.
      COMMON /PCOMS4/ NPCOM, NPKEY, NPIR, NPSP, NPCH
C  ---------------------------------------------------------------------
C  Local variables:
      LOGICAL OK
C  ---------------------------------------------------------------------
C
      IF (ISCOMP(KTYPE(NK),'I').EQ.0)
     &    CALL MABORT('Not an integer key: '//PKEY(NK),'PRANI')
C
      IP = TPOINT(NK)
      IRL = NINT(PIR(IP))
      IF (IRL.EQ.1) THEN
         IL = NINT(PIR(IP+1))
         NLE = NINT(PIR(IP+2))
         IR = NINT(PIR(IP+3))
         NRI = NINT(PIR(IP+4))
         CALL IPI(IVAL,IL,NLE,IR,NRI,OK)
         IP = IP+4
      ELSE IF (IRL.EQ.2) THEN
         IP = IP+1
         NITE = NINT(PIR(IP))
         OK = .FALSE.
         DO 10,I=1,NITE
            IP = IP+1
            ITEM = NINT(PIR(IP))
            OK = OK .OR. ITEM.EQ.IVAL
10       CONTINUE
      END IF
      IOK = 0
      IF (OK) IOK = 1
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE IPI ( IVAL, IL, NLE, IR, NRI, OK )
C  ---------------------------------------------------------------------
C
C  Description:
C    Called by subroutine PRANI. Checks whether IVAL is within the
C    interval defined by NLE and NRI.
C
C  Input arguments:
C    IVAL       The integer to check.
C    IL         IL = 1 -> The interval is open on the left.
C               IL = 2 -> The interval is closed on the left.
C               IL = 3 -> -Infinity was specified as the left limit.
C    NLE        The left limit of the interval.
C    IR         IL = 1 -> The interval is open on the right.
C               IL = 2 -> The interval is closed on the right.
C               IL = 3 -> Infinity was specified as the right limit.
C    NRI        The right limit of the interval.
C
C  Output arguments:
C    OK         OK = .TRUE.  -> IVAL is within the specified interval.
C               OK = .FALSE. -> IVAL is not within the interval.
C
C  ---------------------------------------------------------------------
C  Arguments:
      LOGICAL OK
C  ---------------------------------------------------------------------
C
      IF (IL.EQ.1) THEN
         IF (IR.EQ.1) THEN
            OK = IVAL.GT.NLE .AND. IVAL.LT.NRI
         ELSE IF (IR.EQ.2) THEN
            OK = IVAL.GT.NLE .AND. IVAL.LE.NRI
         ELSE IF (IR.EQ.3) THEN
            OK = IVAL.GT.NLE
         END IF
      ELSE IF (IL.EQ.2) THEN
         IF (IR.EQ.1) THEN
            OK = IVAL.GE.NLE .AND. IVAL.LT.NRI
         ELSE IF (IR.EQ.2) THEN
            OK = IVAL.GE.NLE .AND. IVAL.LE.NRI
         ELSE IF (IR.EQ.3) THEN
            OK = IVAL.GE.NLE
         END IF
      ELSE IF (ABS(IL).EQ.3) THEN
         IF (IR.EQ.1) THEN
            OK = IVAL.LT.NRI
         ELSE IF (IR.EQ.2) THEN
            OK = IVAL.LE.NRI
         ELSE IF (IR.EQ.3) THEN
            OK = .TRUE.
         END IF
      END IF
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE PRANR ( XVAL, NK, IOK )
C  ---------------------------------------------------------------------
C
C  Description:
C    Checks whether the given real XVAL is allowed as a value for a
C    keyword.
C
C  Input arguments:
C    XVAL       The real to check.
C    NK         The index of the corresponding keyword in the keyword
C               table PKEY.
C
C  Output arguments:
C    IOK        IOK = 0 -> Value not allowed for this keyword.
C               IOK = 1 -> Value ok.
C
C  ---------------------------------------------------------------------
      IMPLICIT $(TYPE) (A-H,O-Z)
C  Common blocks:
C  Maximum length of a keyword.
      PARAMETER ( KEYLEN = 10 )
C  Maximum number of panel commands.
      PARAMETER ( MXPCOM = 80 )
C  Maximum number of keywords in all panel commands.
      PARAMETER ( MXPKEY = 1000 )
C  Maximum storage for integer/real keywords.
      PARAMETER ( MXPIR = 5000 )
C  Maximum storage for string keywords and descriptions.
      PARAMETER ( MXPSTR = 20000 )
C  Maximum number of pointers in the above string storage.
      PARAMETER ( MXPPST = 5000 )
C
      CHARACTER*(KEYLEN) PCOM(MXPCOM)
      INTEGER            PONF(MXPCOM)
      INTEGER            KPOINT(MXPCOM)
      INTEGER            NCHELP(2*MXPCOM)
      CHARACTER*(KEYLEN) PKEY(MXPKEY)
      INTEGER            NKHELP(2*MXPKEY)
      CHARACTER          KTYPE(MXPKEY)
      INTEGER            TPOINT(MXPKEY)
      CHARACTER          KIO(MXPKEY)
      $(TYPE)            PIR(MXPIR)
      INTEGER            SPOINT(MXPPST)
      CHARACTER*(MXPSTR) PANSTR
      COMMON /PCOMS1/ KPOINT, NCHELP, NKHELP, TPOINT, SPOINT, PONF
      COMMON /PCOMS2/ PIR
      COMMON /PCOMS3/ PKEY, KTYPE, PANSTR, PCOM, KIO
C
C  NPCOM: Number of panel commands in the panel description file.
C  NPKEY: Number of keywords in all panel commands.
C  NPIR:  Number of positions filled in array PIR.
C  NPSP:  Number of positions filled in the char-pointer array SPOINT.
C  NPCH:  Number of character positions filled in PANSTR.
      COMMON /PCOMS4/ NPCOM, NPKEY, NPIR, NPSP, NPCH
C  ---------------------------------------------------------------------
C  Local variables:
      LOGICAL OK
C  ---------------------------------------------------------------------
C
      IF (ISCOMP(KTYPE(NK),'R').EQ.0)
     &    CALL MABORT('Not a real key: '//PKEY(NK),'PRANR')
C
      IP = TPOINT(NK)
      IRL = NINT(PIR(IP))
      IF (IRL.EQ.1) THEN
         IL = NINT(PIR(IP+1))
         XLE = PIR(IP+2)
         IR = NINT(PIR(IP+3))
         XRI = PIR(IP+4)
         CALL IPR(XVAL,IL,XLE,IR,XRI,OK)
         IP = IP+4
      ELSE IF (IRL.EQ.2) THEN
         IP = IP+1
         NITE = NINT(PIR(IP))
         OK = .FALSE.
         DO 10,I=1,NITE
            IP = IP+1
            XITEM = PIR(IP)
            OK = OK .OR. XITEM.EQ.XVAL
10       CONTINUE
      END IF
      IOK = 0
      IF (OK) IOK = 1
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE IPR ( XVAL, IL, XLE, IR, XRI, OK )
C  ---------------------------------------------------------------------
C
C  Description:
C    Called by subroutine PRANR. Checks whether XVAL is within the
C    interval defined by XLE and XRI.
C
C  Input arguments:
C    XVAL       The real to check.
C    IL         IL = 1 -> The interval is open on the left.
C               IL = 2 -> The interval is closed on the left.
C               IL = 3 -> -Infinity was specified as the left limit.
C    XLE        The left limit of the interval.
C    IR         IL = 1 -> The interval is open on the right.
C               IL = 2 -> The interval is closed on the right.
C               IL = 3 -> Infinity was specified as the right limit.
C    XRI        The right limit of the interval.
C
C  Output arguments:
C    OK         OK = .TRUE.  -> XVAL is within the specified interval.
C               OK = .FALSE. -> XVAL is not within the interval.
C
C  ---------------------------------------------------------------------
      IMPLICIT $(TYPE) (A-H,O-Z)
C  Arguments:
      LOGICAL OK
C  ---------------------------------------------------------------------
C
      IF (IL.EQ.1) THEN
         IF (IR.EQ.1) THEN
            OK = XVAL.GT.XLE .AND. XVAL.LT.XRI
         ELSE IF (IR.EQ.2) THEN
            OK = XVAL.GT.XLE .AND. XVAL.LE.XRI
         ELSE IF (IR.EQ.3) THEN
            OK = XVAL.GT.XLE
         END IF
      ELSE IF (IL.EQ.2) THEN
         IF (IR.EQ.1) THEN
            OK = XVAL.GE.XLE .AND. XVAL.LT.XRI
         ELSE IF (IR.EQ.2) THEN
            OK = XVAL.GE.XLE .AND. XVAL.LE.XRI
         ELSE IF (IR.EQ.3) THEN
            OK = XVAL.GE.XLE
         END IF
      ELSE IF (ABS(IL).EQ.3) THEN
         IF (IR.EQ.1) THEN
            OK = XVAL.LT.XRI
         ELSE IF (IR.EQ.2) THEN
            OK = XVAL.LE.XRI
         ELSE IF (IR.EQ.3) THEN
            OK = .TRUE.
         END IF
      END IF
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE PRANS ( SVAL, LES, NK, IOK )
C  ---------------------------------------------------------------------
C
C  Description:
C    Checks whether the given string SVAL is allowed as a value for a
C    keyword.
C
C  Input arguments:
C    SVAL       The string to check.
C    LES        Effective length of SVAL.
C    NK         The index of the corresponding keyword in the keyword
C               table PKEY.
C
C  Output arguments:
C    IOK        IOK = 0 -> Value not allowed for this keyword.
C               IOK = 1 -> Value ok.
C
C  ---------------------------------------------------------------------
      IMPLICIT $(TYPE) (A-H,O-Z)
C  Arguments:
      CHARACTER*(*) SVAL
C  ---------------------------------------------------------------------
C  Common blocks:
C  Maximum length of a keyword.
      PARAMETER ( KEYLEN = 10 )
C  Maximum number of panel commands.
      PARAMETER ( MXPCOM = 80 )
C  Maximum number of keywords in all panel commands.
      PARAMETER ( MXPKEY = 1000 )
C  Maximum storage for integer/real keywords.
      PARAMETER ( MXPIR = 5000 )
C  Maximum storage for string keywords and descriptions.
      PARAMETER ( MXPSTR = 20000 )
C  Maximum number of pointers in the above string storage.
      PARAMETER ( MXPPST = 5000 )
C
      CHARACTER*(KEYLEN) PCOM(MXPCOM)
      INTEGER            PONF(MXPCOM)
      INTEGER            KPOINT(MXPCOM)
      INTEGER            NCHELP(2*MXPCOM)
      CHARACTER*(KEYLEN) PKEY(MXPKEY)
      INTEGER            NKHELP(2*MXPKEY)
      CHARACTER          KTYPE(MXPKEY)
      INTEGER            TPOINT(MXPKEY)
      CHARACTER          KIO(MXPKEY)
      $(TYPE)            PIR(MXPIR)
      INTEGER            SPOINT(MXPPST)
      CHARACTER*(MXPSTR) PANSTR
      COMMON /PCOMS1/ KPOINT, NCHELP, NKHELP, TPOINT, SPOINT, PONF
      COMMON /PCOMS2/ PIR
      COMMON /PCOMS3/ PKEY, KTYPE, PANSTR, PCOM, KIO
C
C  NPCOM: Number of panel commands in the panel description file.
C  NPKEY: Number of keywords in all panel commands.
C  NPIR:  Number of positions filled in array PIR.
C  NPSP:  Number of positions filled in the char-pointer array SPOINT.
C  NPCH:  Number of character positions filled in PANSTR.
      COMMON /PCOMS4/ NPCOM, NPKEY, NPIR, NPSP, NPCH
C  ---------------------------------------------------------------------
C  Local variables:
      LOGICAL OK
      LOGICAL STRI
C  ---------------------------------------------------------------------
C
      IF (ISCOMP(KTYPE(NK),'S').NE.0) THEN
         STRI = .TRUE.
      ELSE IF (ISCOMP(KTYPE(NK),'C').NE.0) THEN
         STRI = .FALSE.
      ELSE
         CALL MABORT('Not a string key: '//PKEY(NK),'PRANS')
      END IF
C
      IP = TPOINT(NK)
      IAL = NINT(PIR(IP))
      IF (IAL.EQ.1) THEN
         OK = .TRUE.
      ELSE IF (IAL.EQ.2) THEN
         IP = IP+1
         NITE = NINT(PIR(IP))
         OK = .FALSE.
         DO 10,I=1,NITE
            IP = IP+1
            NS = NINT(PIR(IP))
            IST = SPOINT(NS)
            IEND = SPOINT(NS+1)
            LEC = IEND-IST+1
            IF (LES.EQ.0) THEN
               OK = OK .OR. LEC.EQ.0
            ELSE
               IF (LEC.NE.0) THEN
                  IF (STRI) THEN
                     OK = OK .OR. SVAL(1:LES).EQ.
     &                                           PANSTR(IST:IEND)
                  ELSE
                     OK = OK .OR. ISCOMP(SVAL(1:LES),
     &                                           PANSTR(IST:IEND)).NE.0
                  END IF
               END IF
            END IF
10       CONTINUE
      END IF
C
      IOK = 0
      IF (OK) IOK = 1
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE QLCR ( LINE, L, NEW, IS, IE, NIT, IERR )
C  ---------------------------------------------------------------------
C
C  Description:
C    Given a string of characters (LINE), this subroutine will
C    separate ("crack") them into a number of items. An item is
C    a sequence of non-blank non-tab characters. For example the
C    following line:
C      I love Merlin.
C    contains the 3 items: I
C                          love
C                          Merlin.
C    Blanks and/or tabs can be embedded in an item using single quotes.
C    The following line:
C      An 'extra fast' routine.
C    contains the 3 items: An
C                          extra fast
C                          routine.
C    A single quote can be entered using the escape sequence \'
C    (even within a pair of quotes). For example the line:
C      A single quote \'
C    contains the 4 items: A
C                        single
C                        quote
C                        '
C    The characters of all items are concatenated and returned in the
C    output string NEW. In order to separate one item from another
C    two arrays of pointers (IS, IE) are returned as well.
C    IS(I) and IE(I) contain the starting and ending positions of
C    item I, correspondingly.
C    For example the input line:
C      Hello it\'s me again.
C    contains 4 items and sets NEW to:
C      Helloit'smeagain.
C    IS(I), IE(I) are:  I    IS(I)    IE(I)    Item
C                       -----------------------------
C                       1      1        5      Hello
C                       2      6        9      it's
C                       3     10       11      me
C                       4     12       17      again.
C
C  Input arguments:
C    LINE       The line to be cracked.
C    L          Effective length of LINE.
C
C  Output arguments:
C    NEW        All items concatenated.
C    IS         IS(I) is the starting character position of item I in
C               the output character NEW.
C    IE         IE(I) is the ending character position of item I in the
C               output character NEW.
C    NIT        Number of items.
C    IERR       Error indicator.
C               IERR =  0 -> Everything ok.
C               IERR = -2 -> A closing quote is missing.
C
C  ---------------------------------------------------------------------
C  Arguments:
      CHARACTER*(*) LINE, NEW
      INTEGER IS(*), IE(*)
C  ---------------------------------------------------------------------
C  Common blocks:
      CHARACTER TAB, ESC
      COMMON /SPECHA/ TAB, ESC
C  ---------------------------------------------------------------------
C  Local variables:
      CHARACTER QUOTE, SPACE
      PARAMETER ( QUOTE = '''', SPACE = ' ' )
      LOGICAL INWORD, INSTR
      CHARACTER C
      LOGICAL WHITE
      WHITE(C) = C.EQ.SPACE .OR. C.EQ.TAB
C  ---------------------------------------------------------------------
C
      NIT = 0
      IERR = 0
      IF (L.EQ.0) RETURN
C
      NN = 0
C
C  INWORD is set to .TRUE. as long as the character under examination
C  belongs to an item. When examining item separators (blanks/tabs),
C  INWORD is set to .FALSE.
      INWORD = .FALSE.
C
C  INSTR is set to .TRUE. while examining characters after a quote.
C  (not an escaped quote.)
      INSTR = .FALSE.
C
C  The following loop examines all characters in LINE, one by one.
      I = 1
10    CONTINUE
      C = LINE(I:I)
      IF (C.EQ.QUOTE) THEN
         IF (INSTR) THEN
C  Closing quote.
            INSTR = .FALSE.
         ELSE
C  Opening quote.
            INSTR = .TRUE.
            IF (.NOT.INWORD) THEN
               INWORD = .TRUE.
               IF (I.EQ.L) THEN
                  GOTO 20
               ELSE
                  NIT = NIT+1
                  IS(NIT) = NN+1
               END IF
            END IF
         END IF
      ELSE IF (WHITE(C)) THEN
C  Space or tab.
         IF (INWORD) THEN
            IF (INSTR) THEN
C  A new item starts.
               NN = NN+1
               NEW(NN:NN) = C
            ELSE
               INWORD = .FALSE.
               IF (IS(NIT).GT.NN) THEN
                  NIT = NIT-1
               ELSE
                  IE(NIT) = NN
               END IF
            END IF
         END IF
      ELSE IF (C.EQ.ESC) THEN
C  The escape character (backslash).
         IF (I.EQ.L) THEN
C  The escape character is the last one in LINE.
            NN = NN+1
            NEW(NN:NN) = C
            IF (.NOT.INWORD) THEN
               NIT = NIT+1
               IS(NIT) = NN
               INWORD = .TRUE.
            END IF
            GOTO 20
         ELSE
C  Something follows the escape character.
            I = I+1
            C = LINE(I:I)
            IF (.NOT.INWORD) THEN
               NIT = NIT+1
               IS(NIT) = NN+1
               INWORD = .TRUE.
            END IF
            IF (C.EQ.QUOTE) THEN
C  A quote follows the escape character. Add it to the current item.
               NN = NN+1
               NEW(NN:NN) = QUOTE
            ELSE
C  A character other than a quote follows the escape character.
               NN = NN+1
               NEW(NN:NN) = ESC
               GOTO 10
            END IF
         END IF
      ELSE
C  Any other character.
         IF (INWORD) THEN
C  Add the character to the current item.
            NN = NN+1
            NEW(NN:NN) = C
         ELSE
C  Start a new item.
            INWORD = .TRUE.
            NN = NN+1
            NEW(NN:NN) = C
            NIT = NIT+1
            IS(NIT) = NN
         END IF
      END IF
      I = I+1
      IF (I.LE.L) GOTO 10
C
C  All characters in LINE have been processed. Set the ending pointer
C  for the last item. Also check whether a closing quote is missing.
20    CONTINUE
      IF (INWORD) THEN
         IF (IS(NIT).GT.NN) THEN
            NIT = NIT-1
         ELSE
            IE(NIT) = NN
         END IF
         IF (INSTR) IERR = -2
      END IF
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE RRANGE ( TOKEN, NUM, ALPHA, NT, NN, IERR, PIR, NPIR,
     &                    MXPIR, PDESCF, NPL, COMM, KEY, IUOUT, LEP,
     &                    WEAK )
C  ---------------------------------------------------------------------
C
C  Description:
C    Subroutine RRANGE parses the definition of a range or set of
C    reals and fills in the appropriate data structures.
C    A range of reals has the form:
C      L S,E R
C    where: L is ( or [ indicating an open/closed interval on the left.
C           R is ) or ] indicating an open/closed interval on the right.
C           S is a real (start of the interval) or -inf (for
C             infinity).
C           E is a real (end of the interval) or inf (for infinity).
C    Examples: (-1,1)  (-inf,0]  [10,100)  [70,80]
C
C    A set (list) of reals has the form:
C      { R1, R2, R3, ... }
C    where R1, R2,... etc. are reals.
C
C    Panel related data structures are described in subroutine PARSP.
C    The lexical analysis tables (TOKEN, NUM, ALPHA) are described in
C    subroutine LEX.
C
C  Input arguments:
C    TOKEN      Tokens in the current line (see subroutine LEX).
C    NUM        Numbers in the current line (see subroutine LEX).
C    ALPHA      Identifiers/strings in the current line (see subroutine
C               LEX).
C    MXPIR      Maximum allowed number of entries in the PIR table.
C    PDESCF     Name of the panel description file.
C    NPL        Number of lines read so far from the panel description
C               file.
C    COMM       The panel currently under processing.
C    KEY        The keyword currently under processing.
C    IUOUT      The Merlin output unit.
C    LEP        Effective length of PDESCF.
C    WEAK       Error messages are printed only of WEAK = .TRUE.
C
C  Output arguments:
C    IERR       Error indicator.
C
C  Input / Output arguments:
C    NT         Number of tokens (TOKEN) used so far.
C    NN         Number of numbers (NUM) used so far.
C    PIR        The PIR table (see subroutine PARSP).
C    NPIR       Number of entries in the PIR array.
C
C  ---------------------------------------------------------------------
C  Arguments:
      INTEGER TOKEN(*)
      $(TYPE) NUM(*), PIR(MXPIR)
      CHARACTER*(*) ALPHA, PDESCF, COMM, KEY
      LOGICAL WEAK
C  ---------------------------------------------------------------------
C  Common blocks:
C  Codes used by the lexical analysis routine LEX.
      INTEGER LEFTP, RIGHTP, PERIOD, COMMA, GT, LT, NE, EQ, GE,
     &        LE, PLUS, MINUS, STAR, SLASH, BISTAR, COLON, EOL,
     &        IDENT, UNSI, NOTREC, STRING, QUOTE, LBRA, RBRA,
     &        SEMI, PERCE, AND, OR, XOR, NOT, EQUAL, LBRACE,
     &        RBRACE, QMARK
      COMMON /MCLSCD/ LEFTP, RIGHTP, PERIOD, COMMA, GT, LT, NE, EQ, GE,
     &                LE, PLUS, MINUS, STAR, SLASH, BISTAR, COLON, EOL,
     &                IDENT, UNSI, NOTREC, STRING, QUOTE, LBRA, RBRA,
     &                SEMI, PERCE, AND, OR, XOR, NOT, EQUAL, LBRACE,
     &                RBRACE, QMARK
C  ---------------------------------------------------------------------
C  Local variables:
      LOGICAL OK
      $(TYPE) XLI, XLE, XRI
C  ---------------------------------------------------------------------
C
      IERR = 0
      NT = NT+1
      INIT = TOKEN(NT)
      IF (INIT.EQ.LBRACE) THEN
C  A list of values is expected.
         NPIR = NPIR+1
         IF (NPIR.GT.MXPIR)
     &          CALL OVER('pir','PIR','RRANGE',MXPIR,'MXPIR')
         PIR(NPIR) = 2
C  Allocate PIR space for the # of items.
         NPIR = NPIR+1
         IF (NPIR.GT.MXPIR)
     &          CALL OVER('pir','PIR','RRANGE',MXPIR,'MXPIR')
         ISTLI = NPIR
10       CONTINUE
         NT = NT+1
         ITOK = TOKEN(NT)
         ISI = 1
         IF (ITOK.EQ.PLUS) THEN
            ISI = 1
            NT = NT+1
         ELSE IF (ITOK.EQ.MINUS) THEN
            ISI = -1
            NT = NT+1
         END IF
         IF (TOKEN(NT).NE.UNSI) THEN
            LE1 = LENGTH(KEY)
            LE2 = LENGTH(COMM)
            CALL PERRMS(PDESCF,LEP,WEAK,NPL,IUOUT)
            IF (WEAK) WRITE (IUOUT,270) KEY(1:LE1), COMM(1:LE2)
            IERR = 1
            RETURN
         END IF
         NN = NN+1
         XLI = ISI*NUM(NN)
         NPIR = NPIR+1
         IF (NPIR.GT.MXPIR)
     &          CALL OVER('pir','PIR','RRANGE',MXPIR,'MXPIR')
         PIR(NPIR) = XLI
         NT = NT+1
         IF (TOKEN(NT).EQ.COMMA) GOTO 10
         IF (TOKEN(NT).EQ.RBRACE) THEN
C  End of list ok.
            NLI = NPIR-ISTLI
            PIR(ISTLI) = NLI
            RETURN
         END IF
         LE1 = LENGTH(KEY)
         LE2 = LENGTH(COMM)
         CALL PERRMS(PDESCF,LEP,WEAK,NPL,IUOUT)
         IF (WEAK) WRITE (IUOUT,271) KEY(1:LE1), COMM(1:LE2)
         IERR = 1
         RETURN
      ELSE IF (INIT.EQ.LEFTP .OR. INIT.EQ.LBRA) THEN
C  A left open/closed interval is expected.
         XLE = 0.
         XRI = 0.
         IF (INIT.EQ.LEFTP) ITPL = 1
         IF (INIT.EQ.LBRA) ITPL = 2
         NT = NT+1
         ISI = 1
         IF (TOKEN(NT).EQ.PLUS) THEN
            ISI = 1
            NT = NT+1
         ELSE IF (TOKEN(NT).EQ.MINUS) THEN
            ISI = -1
            NT = NT+1
         END IF
         IF (TOKEN(NT).EQ.IDENT) THEN
            NN = NN+1
            IST = NINT(NUM(NN))
            NN = NN+1
            IEND = NINT(NUM(NN))
            IF (ISCOMP(ALPHA(IST:IEND),'inf').NE.0) THEN
               ITPL = 3*ISI
            ELSE
               LE1 = LENGTH(KEY)
               LE2 = LENGTH(COMM)
               CALL PERRMS(PDESCF,LEP,WEAK,NPL,IUOUT)
               IF (WEAK) WRITE (IUOUT,272) KEY(1:LE1), COMM(1:LE2)
               IERR = 1
               RETURN
            END IF
         ELSE IF (TOKEN(NT).EQ.UNSI) THEN
            NN = NN+1
            XLE = NUM(NN)
            XLE = ISI*XLE
         ELSE
            LE1 = LENGTH(KEY)
            LE2 = LENGTH(COMM)
            CALL PERRMS(PDESCF,LEP,WEAK,NPL,IUOUT)
            IF (WEAK) WRITE (IUOUT,272) KEY(1:LE1), COMM(1:LE2)
            IERR = 1
            RETURN
         END IF
         NT = NT+1
         IF (TOKEN(NT).NE.COMMA) THEN
            LE1 = LENGTH(KEY)
            LE2 = LENGTH(COMM)
            CALL PERRMS(PDESCF,LEP,WEAK,NPL,IUOUT)
            IF (WEAK) WRITE (IUOUT,274) KEY(1:LE1), COMM(1:LE2)
            IERR = 1
            RETURN
         END IF
C
         ITPR = 0
         NT = NT+1
         ISI = 1
         IF (TOKEN(NT).EQ.PLUS) THEN
            ISI = 1
            NT = NT+1
         ELSE IF (TOKEN(NT).EQ.MINUS) THEN
            ISI = -1
            NT = NT+1
         END IF
         IF (TOKEN(NT).EQ.IDENT) THEN
            NN = NN+1
            IST = NINT(NUM(NN))
            NN = NN+1
            IEND = NINT(NUM(NN))
            IF (ISCOMP(ALPHA(IST:IEND),'inf').NE.0) THEN
               ITPR = 3*ISI
            ELSE
               LE1 = LENGTH(KEY)
               LE2 = LENGTH(COMM)
               CALL PERRMS(PDESCF,LEP,WEAK,NPL,IUOUT)
               IF (WEAK) WRITE (IUOUT,275) KEY(1:LE1), COMM(1:LE2)
               IERR = 1
               RETURN
            END IF
         ELSE IF (TOKEN(NT).EQ.UNSI) THEN
            NN = NN+1
            XRI = NUM(NN)
            XRI = XRI
         ELSE
            LE1 = LENGTH(KEY)
            LE2 = LENGTH(COMM)
            CALL PERRMS(PDESCF,LEP,WEAK,NPL,IUOUT)
            IF (WEAK) WRITE (IUOUT,275) KEY(1:LE1), COMM(1:LE2)
            IERR = 1
            RETURN
         END IF
C
         NT = NT+1
         IF (TOKEN(NT).EQ.RIGHTP) THEN
            IF (ITPR.EQ.0) ITPR = 1
         ELSE IF (TOKEN(NT).EQ.RBRA) THEN
            IF (ITPR.EQ.0) ITPR = 2
         ELSE
            LE1 = LENGTH(KEY)
            LE2 = LENGTH(COMM)
            CALL PERRMS(PDESCF,LEP,WEAK,NPL,IUOUT)
            IF (WEAK) WRITE (IUOUT,277) KEY(1:LE1), COMM(1:LE2)
            IERR = 1
            RETURN
         END IF
C
         IF (NPIR+5.GT.MXPIR)
     &          CALL OVER('pir','PIR','RRANGE',MXPIR,'MXPIR')
         PIR(NPIR+1) = 1
         PIR(NPIR+2) = ITPL
         PIR(NPIR+3) = XLE
         PIR(NPIR+4) = ITPR
         PIR(NPIR+5) = XRI
         NPIR = NPIR+5
         IF (ITPL.EQ. 1 .OR.ITPL.EQ.2) THEN
            IF (ITPR.EQ.1 .OR. ITPR.EQ.2) THEN
               OK = XLE.LE.XRI
            ELSE
               OK = ITPR.EQ.3
            END IF
         ELSE IF (ITPL.EQ.-3) THEN
            IF (ITPR.EQ.1 .OR. ITPR.EQ.2) THEN
               OK = .TRUE.
            ELSE
               OK = ITPL.LT.ITPR
            END IF
         ELSE
            OK = .FALSE.
         END IF
         IF (.NOT.OK) THEN
            LE1 = LENGTH(KEY)
            LE2 = LENGTH(COMM)
            CALL PERRMS(PDESCF,LEP,WEAK,NPL,IUOUT)
            IF (WEAK) WRITE (IUOUT,278) KEY(1:LE1), COMM(1:LE2)
            IERR = 1
            RETURN
         END IF
      ELSE IF (INIT.EQ.EOL) THEN
         LE1 = LENGTH(KEY)
         LE2 = LENGTH(COMM)
         CALL PERRMS(PDESCF,LEP,WEAK,NPL,IUOUT)
         IF (WEAK) WRITE (IUOUT,280) KEY(1:LE1), COMM(1:LE2)
         IERR = 1
         RETURN
      ELSE
         LE1 = LENGTH(KEY)
         LE2 = LENGTH(COMM)
         CALL PERRMS(PDESCF,LEP,WEAK,NPL,IUOUT)
         IF (WEAK) WRITE (IUOUT,279) KEY(1:LE1), COMM(1:LE2)
         IERR = 1
         RETURN
      END IF
C
270   FORMAT ($(FC)'Reals only should be present in the ',
     &        'list of allowed values'/
     &        $(FC)'for the ',A,' keyword in the ',A,' panel.')
271   FORMAT ($(FC)'The list of real values for the ',A,
     &        ' keyword in the ',A,' panel,'
     &        /$(FC)'is incorrect.')
272   FORMAT ($(FC)'The left limit for the ',A,
     &        ' keyword in the ',A,' panel,'/
     &        $(FC)'is incorrect.')
274   FORMAT ($(FC)'A comma is expected after the left ',
     &        'limit for the ',A,' keyword'
     &        /$(FC)'in the ',A,' panel.')
275   FORMAT ($(FC)'The right limit for the ',A,
     &        ' keyword in the ',A,' panel, is incorrect.')
277   FORMAT ($(FC)'The interval specification for the ',A,
     &        ' keyword in the ',A,' panel,'
     &        /$(FC)'is incorrect.')
278   FORMAT ($(FC)'The interval specification for the ',A,
     &        ' keyword in the ',A,' panel,'
     &        /$(FC)'is incorrect. Left limit > right limit.')
280   FORMAT ($(FC)'There is no description of allowed values ',
     &        'for the ',A,' keyword'
     &        /$(FC)'in the ',A,' panel, is incorrect.')
279   FORMAT ($(FC)'The specification of allowed values for the ',A,
     &        ' keyword'
     &        /$(FC)'in the ',A,' panel.')
      END
C  ---------------------------------------------------------------------
      SUBROUTINE SRANGE ( TOKEN, NUM, ALPHA, NT, NN, IERR, PIR, NPIR,
     &                    MXPIR, SPOINT, NPSP, MXPPST, PANSTR, NPCH,
     &                    MXPSTR, PDESCF, NPL, COMM, KEY, IUOUT, LEP,
     &                    WEAK )
C  ---------------------------------------------------------------------
C
C  Description:
C    Subroutine SRANGE parses the definition of a range of strings and
C    fills in the appropriate data structures.
C    A set (list) of strings has the form:
C      { S1, S2, S3, ... }
C      where S1, S2,... etc. are strings.
C    or
C      { any }
C    The second form causes any string to be accepted as a keyword
C    value.
C    Examples: {'a','b','c'}  {any} {'Hello','Goodbye'}
C
C    Panel related data structures are described in subroutine PARSP.
C    The lexical analysis tables (TOKEN, NUM, ALPHA) are described in
C    subroutine LEX.
C
C  Input arguments:
C    TOKEN      Tokens in the current line (see subroutine LEX).
C    NUM        Numbers in the current line (see subroutine LEX).
C    ALPHA      Identifiers/strings in the current line (see subroutine
C               LEX).
C    MXPIR      Maximum allowed number of entries in the PIR table.
C    MXPPST     Maximum allowed number of entries in the SPOINT array.
C    MXPSTR     Maximum allowed number of characters in PANSTR.
C    PDESCF     Name of the panel description file.
C    NPL        Number of lines read so far from the panel description
C               file.
C    COMM       The panel currently under processing.
C    KEY        The keyword currently under processing.
C    IUOUT      The Merlin output unit.
C    LEP        Effective length of PDESCF.
C    WEAK       Error messages are printed only of WEAK = .TRUE.
C
C  Output arguments:
C    IERR       Error indicator.
C
C  Input / Output arguments:
C    NT         Number of tokens (TOKEN) used so far.
C    NN         Number of numbers (NUM) used so far.
C    PIR        The PIR table (see subroutine PARSP).
C    NPIR       Number of entries in the PIR array.
C    SPOINT     The SPOINT table (see subroutine PARSP).
C    NPSP       Number of entries in SPOINT.
C    PANSTR     The PANSTR character storage (see subroutine PARSP).
C    NPCH       Number of characters in PANSTR.
C
C  ---------------------------------------------------------------------
C  Arguments:
      INTEGER TOKEN(*), SPOINT(MXPPST)
      CHARACTER*(*) PANSTR, ALPHA, PDESCF, COMM, KEY
      $(TYPE) NUM(*), PIR(MXPIR)
      LOGICAL WEAK
C  ---------------------------------------------------------------------
C  Common blocks:
C  Codes used by the lexical analysis routine LEX.
      INTEGER LEFTP, RIGHTP, PERIOD, COMMA, GT, LT, NE, EQ, GE,
     &        LE, PLUS, MINUS, STAR, SLASH, BISTAR, COLON, EOL,
     &        IDENT, UNSI, NOTREC, STRING, QUOTE, LBRA, RBRA,
     &        SEMI, PERCE, AND, OR, XOR, NOT, EQUAL, LBRACE,
     &        RBRACE, QMARK
      COMMON /MCLSCD/ LEFTP, RIGHTP, PERIOD, COMMA, GT, LT, NE, EQ, GE,
     &                LE, PLUS, MINUS, STAR, SLASH, BISTAR, COLON, EOL,
     &                IDENT, UNSI, NOTREC, STRING, QUOTE, LBRA, RBRA,
     &                SEMI, PERCE, AND, OR, XOR, NOT, EQUAL, LBRACE,
     &                RBRACE, QMARK
C  ---------------------------------------------------------------------
C
      IERR = 0
      NT = NT+1
      INIT = TOKEN(NT)
      IF (INIT.NE.LBRACE) THEN
         LE1 = LENGTH(KEY)
         LE2 = LENGTH(COMM)
         CALL PERRMS(PDESCF,LEP,WEAK,NPL,IUOUT)
         IF (WEAK) WRITE (IUOUT,270) KEY(1:LE1), COMM(1:LE2)
         IERR = 1
         RETURN
      END IF
C
C  A list of values is expected.
      NT = NT+1
      IAL = 2
      IF (TOKEN(NT).EQ.IDENT) THEN
         NN = NN+1
         IST = NINT(NUM(NN))
         NN = NN+1
         IEND = NINT(NUM(NN))
         IF (ISCOMP(ALPHA(IST:IEND),'any').NE.0) THEN
            NT = NT+1
            IF (TOKEN(NT).NE.RBRACE) THEN
               LE1 = LENGTH(KEY)
               LE2 = LENGTH(COMM)
               CALL PERRMS(PDESCF,LEP,WEAK,NPL,IUOUT)
               IF (WEAK) WRITE (IUOUT,271) KEY(1:LE1), COMM(1:LE2)
               IERR = 1
            ELSE
C  Any string is accepted.
               IAL = 1
               NPIR = NPIR+1
               IF (NPIR.GT.MXPIR)
     &                  CALL OVER('pir','PIR','SRANGE',MXPIR,'MXPIR')
               PIR(NPIR) = IAL
            END IF
         ELSE
            LE1 = LENGTH(KEY)
            LE2 = LENGTH(COMM)
            CALL PERRMS(PDESCF,LEP,WEAK,NPL,IUOUT)
            IF (WEAK) WRITE (IUOUT,272) KEY(1:LE1), COMM(1:LE2)
            IERR = 1
         END IF
         RETURN
      END IF
      NPIR = NPIR+1
      IF (NPIR.GT.MXPIR)
     &    CALL OVER('pir','PIR','SRANGE',MXPIR,'MXPIR')
      PIR(NPIR) = IAL
C  Allocate PIR space for the # of items.
      NPIR = NPIR+1
      IF (NPIR.GT.MXPIR)
     &    CALL OVER('pir','PIR','SRANGE',MXPIR,'MXPIR')
      ISTLI = NPIR
10    CONTINUE
      IF (TOKEN(NT).NE.QUOTE) THEN
         LE1 = LENGTH(KEY)
         LE2 = LENGTH(COMM)
         CALL PERRMS(PDESCF,LEP,WEAK,NPL,IUOUT)
         IF (WEAK) WRITE (IUOUT,272) KEY(1:LE1), COMM(1:LE2)
         IERR = 1
         RETURN
      END IF
      NT = NT+1
      IF (TOKEN(NT).NE.STRING) THEN
         LE1 = LENGTH(KEY)
         LE2 = LENGTH(COMM)
         CALL PERRMS(PDESCF,LEP,WEAK,NPL,IUOUT)
         IF (WEAK) WRITE (IUOUT,272) KEY(1:LE1), COMM(1:LE2)
         IERR = 1
         RETURN
      END IF
      NT = NT+1
      IF (TOKEN(NT).NE.QUOTE) THEN
         LE1 = LENGTH(KEY)
         LE2 = LENGTH(COMM)
         CALL PERRMS(PDESCF,LEP,WEAK,NPL,IUOUT)
         IF (WEAK) WRITE (IUOUT,272) KEY(1:LE1), COMM(1:LE2)
         IERR = 1
         RETURN
      END IF
      NN = NN+1
      IST = NINT(NUM(NN))
      NN = NN+1
      IEND = NINT(NUM(NN))
      LE1 = IEND-IST+1
      NPIR = NPIR+1
      IF (NPIR.GT.MXPIR)
     &    CALL OVER('pir','PIR','SRANGE',MXPIR,'MXPIR')
      PIR(NPIR) = NPSP+1
      IF (NPSP+2.GT.MXPPST)
     &    CALL OVER('spoint','SPOINT','SRANGE',MXPPST,'MXPPST')
      SPOINT(NPSP+1) = NPCH+1
      SPOINT(NPSP+2) = NPCH+LE1
      IF (NPCH+LE1.GT.MXPSTR)
     &    CALL OVER('panstr','PANSTR','SRANGE',MXPSTR,'MXPSTR')
      PANSTR(NPCH+1:NPCH+LE1) = ALPHA(IST:IEND)
      NPSP = NPSP+2
      NPCH = NPCH+LE1
C
      NT = NT+1
      IF (TOKEN(NT).EQ.COMMA) THEN
         NT = NT+1
         GOTO 10
      END IF
      IF (TOKEN(NT).EQ.RBRACE) THEN
C  List ended ok.
         NLI = NPIR-ISTLI
         PIR(ISTLI) = NLI
         RETURN
      END IF
      LE1 = LENGTH(KEY)
      LE2 = LENGTH(COMM)
      CALL PERRMS(PDESCF,LEP,WEAK,NPL,IUOUT)
      IF (WEAK) WRITE (IUOUT,276) KEY(1:LE1), COMM(1:LE2)
      IERR = 1
C
270   FORMAT ($(FC)'A list of strings is expected as allowed values'
     &        /$(FC)'for the ',A,' keyword in the ',A,' panel.')
271   FORMAT ($(FC)'A right brace is expected after "any"',
     &        ' in the list of allowed'/
     &        $(FC)'values for the ',A,' keyword in the ',A,' panel.')
272   FORMAT ($(FC)'Strings only should be present in the ',
     &        'list of allowed values'/
     &        $(FC)'for the ',A,' keyword in the ',A,' panel.')
276   FORMAT ($(FC)'The list of strings for the ',A,
     &        ' keyword in the ',A,' panel,'
     &        /$(FC)'is incorrect.')
      END
C  ---------------------------------------------------------------------
      SUBROUTINE TOINT ( STR, LE, N, IERR )
C  ---------------------------------------------------------------------
C
C  Description:
C    Converts an integer stored in a string (STR) to its numeric
C    representation. Leading/trailing blanks/tabs in STR are ignored.
C
C  Input arguments:
C    STR        The string to convert.
C    LE         Effective length of STR.
C
C  Output arguments:
C    N          Integer representation of STR.
C    IERR       Error indicator.
C               IERR =  0 -> Conversion succeeded.
C               IERR = -2 -> Conversion failed. STR does not contain
C                            a proper integer.
C
C  ---------------------------------------------------------------------
C  Arguments:
      CHARACTER*(*) STR
C  ---------------------------------------------------------------------
C  Common blocks:
      CHARACTER TAB, ESC
      COMMON /SPECHA/ TAB, ESC
C  ---------------------------------------------------------------------
C  Local variables:
      CHARACTER C
      LOGICAL NUMBER
      NUMBER(C) = C.GE.'0' .AND. C.LE.'9'
C  ---------------------------------------------------------------------
C
C  Skip leading blanks.
      DO 60,I=1,LE
         C = STR(I:I)
         IF (C.NE.' ' .AND. C.NE.TAB) THEN
            NC = I
            GOTO 70
         END IF
60    CONTINUE
C
C  The string is full of blanks/tabs.
      IERR = -2
      RETURN
C
70    CONTINUE
C
C  Remove trailing blanks/tabs.
C  At least one non-white space character is present.
      DO 80,I=LE,NC,-1
         C = STR(I:I)
         IF (C.NE.' ' .AND. C.NE.TAB) THEN
            IEND = I
            GOTO 90
         END IF
80    CONTINUE
      CALL MABORT('String full of blanks.','TOINT')
C
90    CONTINUE
      N = 0
      IOFF = ICHAR('0')
C
C  Check for a sign.
      C = STR(1:1)
      IF (C.EQ.'+') THEN
         ISI = 1
         IPOS = NC+1
      ELSE IF (C.EQ.'-') THEN
         ISI = -1
         IPOS = NC+1
      ELSE
         ISI = 1
         IPOS = NC
      END IF
C
C  Convert the number.
      IF (IPOS.GT.IEND) THEN
         IERR = -2
         RETURN
      END IF
      DO 10,I=IPOS,IEND
         C = STR(I:I)
         IF (NUMBER(C)) THEN
            N = 10*N+ICHAR(C)-IOFF
         ELSE
            IERR = -2
            RETURN
         END IF
10    CONTINUE
C
      N = ISI*N
      IERR = 0
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE TOREAL ( STR, LE, X, IERR )
C  ---------------------------------------------------------------------
C
C  Description:
C    Converts a real, stored in a string (STR) to its numeric
C    representation. Leading/trailing blanks/tabs in STR are ignored.
C    The number in STR should be of the form:
C      [ + | - ] <INT> . <FRAC> <EXP>
C    where <INT> is the integer part: a sequence of digits,
C          <FRAC> is the fractional part: another sequence of digits,
C          and
C          <EXP> is the exponent in the form: E | D [ + | - ] <INT>
C    The integer part <INT> or the fractional part <FRAC> may be
C    omitted, but not both.
C
C  Input arguments:
C    STR        The string to convert.
C    LE         Effective length of STR.
C
C  Output arguments:
C    N          Integer representation of STR.
C    IERR       Error indicator.
C               IERR =  0 -> Conversion succeeded.
C               IERR = -2 -> Conversion failed. STR does not contain
C                            a proper integer.
C
C  ---------------------------------------------------------------------
      IMPLICIT $(TYPE) (A-H,O-Z)
C  Arguments:
      CHARACTER*(*) STR
C  ---------------------------------------------------------------------
C  Common blocks:
      CHARACTER TAB, ESC
      COMMON /SPECHA/ TAB, ESC
C  ---------------------------------------------------------------------
C  Local variables:
      CHARACTER C
      LOGICAL IPART, FPART
      LOGICAL NUMBER
      NUMBER(C) = C.GE.'0' .AND. C.LE.'9'
C  ---------------------------------------------------------------------
C
      IZERO = ICHAR('0')
      V = 0.0
      W = 0.0
      IERR = 0
      POT= 1.0
      IEX = 0
      TEN = 10
      IPART = .FALSE.
      FPART = .FALSE.
C
C  Skip leading blanks/tabs.
      DO 60,I=1,LE
         C = STR(I:I)
         IF (C.NE.' ' .AND. C.NE.TAB) THEN
            NC = I
            GOTO 70
         END IF
60    CONTINUE
C
C  The string is full of blanks/tabs.
      IERR = -2
      RETURN
C
70    CONTINUE
C
C  Remove trailing blanks/tabs.
C  At least one non-white space character is present.
      DO 80,I=LE,NC,-1
         C = STR(I:I)
         IF (C.NE.' ' .AND. C.NE.TAB) THEN
            IEND = I
            GOTO 90
         END IF
80    CONTINUE
      CALL MABORT('String full of blanks.','TOREAL')
C
C  Check for a sign.
90    CONTINUE
      C = STR(NC:NC)
      IF (C.EQ.'+') THEN
         IS = 1
      ELSE IF (C.EQ.'-') THEN
         IS = -1
      ELSE
         IS = 1
         GOTO 10
      END IF
C
      NC = NC+1
      IF (NC.GT.IEND) THEN
         IERR = -2
         RETURN
      END IF
      C = STR(NC:NC)
C
C  Convert the integer part.
10    CONTINUE
      IF (NUMBER(C)) THEN
         V = 10*V + ICHAR(C)-IZERO
         IPART = .TRUE.
         NC = NC+1
         IF (NC.GT.IEND) THEN
            X = IS*V
            RETURN
         END IF
         C = STR(NC:NC)
         GOTO 10
      END IF
C
20    CONTINUE
      IF (C.EQ.'.') GOTO 30
      CALL UPPER(C,1)
      IF (C.EQ.'E' .OR. C.EQ.'D') GOTO 40
      IERR = -2
      RETURN
C
C  Convert the fractional part.
30    CONTINUE
      NC = NC+1
      IF (NC.GT.IEND) THEN
         IF (.NOT.(IPART.OR.FPART)) IERR = -2
         X = IS*(V+W)
         RETURN
      END IF
      C = STR(NC:NC)
      IF (NUMBER(C)) THEN
         POT = 10*POT
         W = W + (ICHAR(C)-IZERO)/POT
         FPART = .TRUE.
         GOTO 30
      END IF
      CALL UPPER(C,1)
      IF (C.EQ.'E' .OR. C.EQ.'D') GOTO 40
      IERR = -2
      RETURN
C
40    NC = NC+1
      IF (NC.GT.IEND) THEN
         IERR = -2
         RETURN
      END IF
      C = STR(NC:NC)
C
C  An exponent without integer or fractional parts is not allowed.
      IF (.NOT.(IPART.OR.FPART)) THEN
         IERR = -2
         RETURN
      END IF
C
C  Check for a sign in the exponent.
      IF (C.EQ.'+') THEN
         IES = 1
      ELSE IF (C.EQ.'-') THEN
         IES = -1
      ELSE
         IES = 1
         GOTO 50
      END IF
      NC = NC+1
      IF (NC.GT.IEND) THEN
         IERR = -2
         RETURN
      END IF
      C = STR(NC:NC)
C
C  Convert the exponent.
50    CONTINUE
      IF (NUMBER(C)) THEN
         IEX = 10*IEX + ICHAR(C)-IZERO
         NC = NC+1
         IF (NC.GT.IEND) THEN
            X = IS*(V+W) * TEN**(IES*IEX)
            RETURN
         END IF
         C = STR(NC:NC)
         GOTO 50
      END IF
C
      IERR = -2
C
      END
C  ---------------------------------------------------------------------
      CHARACTER FUNCTION UPPER1 ( CHA )
C  ---------------------------------------------------------------------
C
C  Description:
C    Converts one character to upper case. It calls the main conversion
C    routine UPPER.
C
C  Input arguments:
C    CHA        The character to be converted.
C
C  ---------------------------------------------------------------------
C  Arguments:
      CHARACTER CHA
C  ---------------------------------------------------------------------
C  Local variables:
      CHARACTER CHA1
C  ---------------------------------------------------------------------
C
      CHA1 = CHA
      CALL UPPER(CHA1,1)
      UPPER1 = CHA1
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE UPPER ( CHA, LL )
C  ---------------------------------------------------------------------
C
C  Description:
C    Converts to upper case a string of characters. Assumes the ASCII
C    character set.
C
C  Input arguments:
C    LL         Effective length of CHA.
C
C  Input / Output arguments:
C    CHA        On input the character to be converted.
C               On output lower case characters are converted to their
C               upper case equivalents.
C
C  ---------------------------------------------------------------------
C  Arguments:
      CHARACTER*(*) CHA
C  ---------------------------------------------------------------------
C  Local variables:
      CHARACTER C
      LOGICAL ISCAP
C  ---------------------------------------------------------------------
C
      DO 10,I=1,LL
         C = CHA(I:I)
         IF (.NOT.ISCAP(C)) THEN
            IC = ICHAR(C)
            IF (IC.GE.97 .AND. IC.LE.122) CHA(I:I) = CHAR(IC-32)
         END IF
10    CONTINUE
C
      END
